(* TODO: Integrate into Refinement Framework *)
theory DFS_Framework_Refine_Aux
imports CAVA_Base DFS_Framework_Misc
begin

(*****************************)
(* Autoref *)

text {* General casting-tag, that allows type-casting on concrete level, while 
  being identity on abstract level. *}
definition [simp]: "CAST \<equiv> id"
lemma [autoref_itype]: "CAST ::\<^sub>i I \<rightarrow>\<^sub>i I" by simp

(* TODO: This idea does currently not work, b/c a homogeneity rule
  will be created from the (\<lambda>x. x, CAST)\<in>R \<rightarrow> R rule, which will always
  be applied first! As a workaround, we make the cast mandatory!
*)
(*lemma [autoref_rules]: 
  assumes "PRIO_TAG_GEN_ALGO"
  shows "(\<lambda>x. x,CAST) \<in> R \<rightarrow> R"
  by auto
*)


(*****************************)
(* Refine-Util *)

(* TODO: Move to Refine_Util *)
attribute_setup zero_var_indexes = {*
  Scan.succeed (Thm.rule_attribute (K Drule.zero_var_indexes))
*} "Set variable indexes to zero, renaming to avoid clashes"

(* TODO: Move, Refine_Util *)
(*
  mk_record_simp: Converts a lemma of the form 
    "f s = x" to the form "r \<equiv> s \<Longrightarrow> f r = x"
  
  This is used to fold the x.simp - lemmas of a record x with a definition
  of the form "r \<equiv> \<lparr> ... \<rparr>".

  Usage example:
    record foo = ...
    definition c :: foo where "c \<equiv> \<lparr> ... \<rparr>"
    lemmas c_simps[simp] = foo.simps[mk_record_simp, OF c_def]
*)
lemma mk_record_simp_thm:
  fixes f :: "'a \<Rightarrow> 'b"
  assumes "f s = x"
  assumes "r \<equiv> s"
  shows "f r = x"
  using assms by simp

ML {*
  fun mk_record_simp thm = let
    val thy = theory_of_thm thm
    val cert = cterm_of thy
  in
    case concl_of thm of
      @{mpat "Trueprop (?f _=_)"} => 
        let
          val cf = cert f
          val r = cterm_instantiate 
            [(@{cpat "?f :: ?'a \<Rightarrow> ?'b"},cf)] 
            @{thm mk_record_simp_thm}
          val r = r OF [thm]
        in r end
    | _ => raise THM("",~1,[thm])

  end
*}

attribute_setup mk_record_simp = 
  {* Scan.succeed (Thm.rule_attribute (K mk_record_simp)) *}
  "Make simplification rule for record definition"


(*****************************)
(* RefineG_Domain *)
(* TODO: Move to RefineG_Domain *)
lemma flat_ge_sup_mono[refine_mono]: "\<And>a a'::'a::complete_lattice. 
  flat_ge a a' \<Longrightarrow> flat_ge b b' \<Longrightarrow> flat_ge (sup a b) (sup a' b')"
  by (auto simp: flat_ord_def)

declare sup_mono[refine_mono]


(*****************************)
(* Refine-Basic *)
(* TODO: Move to Refine_Basic *)
lemma nofail_RES_conv: "nofail m \<longleftrightarrow> (\<exists>M. m=RES M)" by (cases m) auto

(* TODO: Move, near SPEC_nofail  *)
lemma nofail_SPEC: "nofail m \<Longrightarrow> m \<le> SPEC (\<lambda>_. True)"
  by (simp add: pw_le_iff)

lemma nofail_SPEC_iff: "nofail m \<longleftrightarrow> m \<le> SPEC (\<lambda>_. True)"
  by (simp add: pw_le_iff)

lemma nofail_SPEC_triv_refine: "\<lbrakk> nofail m; \<And>x. \<Phi> x \<rbrakk> \<Longrightarrow> m \<le> SPEC \<Phi>"
  by (simp add: pw_le_iff)

(* TODO: Move *)
lemma bind_cong:
  assumes "m=m'"
  assumes "\<And>x. RETURN x \<le> m' \<Longrightarrow> f x = f' x"
  shows "bind m f = bind m' f'"  
  using assms
  by (auto simp: refine_pw_simps pw_eq_iff pw_le_iff)


primrec the_RES where "the_RES (RES X) = X"
lemma the_RES_inv[simp]: "nofail m \<Longrightarrow> RES (the_RES m) = m"
  by (cases m) auto

lemma le_SPEC_UNIV_rule [refine_vcg]: 
  "m \<le> SPEC (\<lambda>_. True) \<Longrightarrow> m \<le> RES UNIV" by auto

definition [refine_pw_simps]: "nf_inres m x \<equiv> nofail m \<and> inres m x"

lemma nf_inres_RES[simp]: "nf_inres (RES X) x \<longleftrightarrow> x\<in>X" 
  by (simp add: refine_pw_simps)
  
lemma nf_inres_SPEC[simp]: "nf_inres (SPEC \<Phi>) x \<longleftrightarrow> \<Phi> x" 
  by (simp add: refine_pw_simps)

lemma bind_refine_abs': (* Only keep nf_inres-information for abstract *)
  fixes R' :: "('a\<times>'b) set" and R::"('c\<times>'d) set"
  assumes R1: "M \<le> \<Down> R' M'"
  assumes R2: "\<And>x x'. \<lbrakk> (x,x')\<in>R'; nf_inres M' x'
  \<rbrakk> \<Longrightarrow> f x \<le> \<Down> R (f' x')"
  shows "bind M (\<lambda>x. f x) \<le> \<Down> R (bind M' (\<lambda>x'. f' x'))"
  using assms
  apply (simp add: pw_le_iff refine_pw_simps)
  apply blast
  done

(* TODO: Move *)
lemma Let_refine':
  assumes "(m,m')\<in>R"
  assumes "(m,m')\<in>R \<Longrightarrow> f m \<le>\<Down>S (f' m')"
  shows "Let m f \<le> \<Down>S (Let m' f')"
  using assms by simp

(* TODO: Move, test as default simp-rule *)
lemma RETURN_refine_iff[simp]: "RETURN x \<le>\<Down>R (RETURN y) \<longleftrightarrow> (x,y)\<in>R"
  by (auto simp: pw_le_iff refine_pw_simps)

lemma RETURN_RES_refine_iff: 
  "RETURN x \<le>\<Down>R (RES Y) \<longleftrightarrow> (\<exists>y\<in>Y. (x,y)\<in>R)"
  by (auto simp: pw_le_iff refine_pw_simps)

lemma in_nres_rel_iff: "(a,b)\<in>\<langle>R\<rangle>nres_rel \<longleftrightarrow> a \<le>\<Down>R b"
  by (auto simp: nres_rel_def)



lemma inf_RETURN_RES: 
  "inf (RETURN x) (RES X) = (if x\<in>X then RETURN x else SUCCEED)"
  "inf (RES X) (RETURN x) = (if x\<in>X then RETURN x else SUCCEED)"
  by (auto simp: pw_eq_iff refine_pw_simps)

(* TODO: MOve, test as default simp-rule *)
lemma inf_RETURN_SPEC[simp]:
  "inf (RETURN x) (SPEC (\<lambda>y. \<Phi> y)) = SPEC (\<lambda>y. y=x \<and> \<Phi> x)"
  "inf (SPEC (\<lambda>y. \<Phi> y)) (RETURN x) = SPEC (\<lambda>y. y=x \<and> \<Phi> x)"
  by (auto simp: pw_eq_iff refine_pw_simps)

lemma RES_sng_eq_RETURN: "RES {x} = RETURN x"
  by simp

lemma nofail_inf_serialize:
  "\<lbrakk>nofail a; nofail b\<rbrakk> \<Longrightarrow> inf a b = do {x\<leftarrow>a; ASSUME (inres b x); RETURN x}"
  by (auto simp: pw_eq_iff refine_pw_simps)

lemma conc_fun_SPEC: 
  "\<Down>R (SPEC (\<lambda>x. \<Phi> x)) = SPEC (\<lambda>y. \<exists>x. (y,x)\<in>R \<and> \<Phi> x)"  
  by (auto simp: pw_eq_iff refine_pw_simps)

lemma conc_fun_RETURN: 
  "\<Down>R (RETURN x) = SPEC (\<lambda>y. (y,x)\<in>R)"  
  by (auto simp: pw_eq_iff refine_pw_simps)



definition lift_assn :: "('a \<times> 'b) set \<Rightarrow> ('b \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool)"
  -- \<open>Lift assertion over refinement relation\<close>
  where "lift_assn R \<Phi> s \<equiv> \<exists>s'. (s,s')\<in>R \<and> \<Phi> s'"
lemma lift_assnI: "\<lbrakk>(s,s')\<in>R; \<Phi> s'\<rbrakk> \<Longrightarrow> lift_assn R \<Phi> s"
  unfolding lift_assn_def by auto

(* TODO: Replace original lemma *)
lemma case_option_refine[refine]:
  assumes "(x,x')\<in>Id"
  assumes "x=None \<Longrightarrow> fn \<le> \<Down>R fn'"
  assumes "\<And>v v'. \<lbrakk>x=Some v; (v,v')\<in>Id\<rbrakk> \<Longrightarrow> fs v \<le> \<Down>R (fs' v')"
  shows "case_option fn (\<lambda>v. fs v) x \<le> \<Down>R (case_option fn' (\<lambda>v'. fs' v') x')"
  using assms by (auto split: option.split)


definition GHOST :: "'a \<Rightarrow> 'a" 
  -- \<open>Ghost tag to mark ghost variables in let-expressions\<close>
  where [simp]: "GHOST \<equiv> \<lambda>x. x"
lemma GHOST_elim_Let: -- \<open>Unfold rule to inline GHOST-Lets\<close>
  shows "(let x=GHOST m in f x) = f m" by simp



text \<open>The following set of rules executes a step on the LHS or RHS of 
  a refinement proof obligation, without changing the other side.
  These kind of rules is useful for performing refinements with 
  invisible steps.\<close>  
lemma lhs_step_If:
  "\<lbrakk> b \<Longrightarrow> t \<le> m; \<not>b \<Longrightarrow> e \<le> m \<rbrakk> \<Longrightarrow> If b t e \<le> m" by simp

lemma lhs_step_RES:
  "\<lbrakk> \<And>x. x\<in>X \<Longrightarrow> RETURN x \<le> m  \<rbrakk> \<Longrightarrow> RES X \<le> m" 
  by (simp add: pw_le_iff)

lemma lhs_step_SPEC:
  "\<lbrakk> \<And>x. \<Phi> x \<Longrightarrow> RETURN x \<le> m \<rbrakk> \<Longrightarrow> SPEC (\<lambda>x. \<Phi> x) \<le> m" 
  by (simp add: pw_le_iff)

lemma lhs_step_bind:
  fixes m :: "'a nres" and f :: "'a \<Rightarrow> 'b nres"
  assumes "nofail m' \<Longrightarrow> nofail m"
  assumes "\<And>x. nf_inres m x \<Longrightarrow> f x \<le> m'"
  shows "do {x\<leftarrow>m; f x} \<le> m'"
  using assms
  by (simp add: pw_le_iff refine_pw_simps) blast

lemma rhs_step_bind:
  assumes "m \<le> \<Down>R m'" "inres m x" "\<And>x'. (x,x')\<in>R \<Longrightarrow> lhs \<le>\<Down>S (f' x')"
  shows "lhs \<le> \<Down>S (m' \<guillemotright>= f')"
  using assms
  by (simp add: pw_le_iff refine_pw_simps) blast

lemma rhs_step_bind_RES:
  assumes "x'\<in>X'"
  assumes "m \<le> \<Down>R (f' x')"
  shows "m \<le> \<Down>R (RES X' \<guillemotright>= f')"
  using assms by (simp add: pw_le_iff refine_pw_simps) blast

lemma rhs_step_bind_SPEC:
  assumes "\<Phi> x'"
  assumes "m \<le> \<Down>R (f' x')"
  shows "m \<le> \<Down>R (SPEC \<Phi> \<guillemotright>= f')"
  using assms by (simp add: pw_le_iff refine_pw_simps) blast

lemma RES_bind_choose:
  assumes "x\<in>X"
  assumes "m \<le> f x"
  shows "m \<le> RES X \<guillemotright>= f"
  using assms by (auto simp: pw_le_iff refine_pw_simps)

lemma pw_RES_bind_choose: 
  "nofail (RES X \<guillemotright>= f) \<longleftrightarrow> (\<forall>x\<in>X. nofail (f x))"
  "inres (RES X \<guillemotright>= f) y \<longleftrightarrow> (\<exists>x\<in>X. inres (f x) y)"
  by (auto simp: refine_pw_simps)


(* TODO: Move to Refine_Basic: Convenience*)
lemma use_spec_rule:
  assumes "m \<le> SPEC \<Psi>"
  assumes "m \<le> SPEC (\<lambda>s. \<Psi> s \<longrightarrow> \<Phi> s)"
  shows "m \<le> SPEC \<Phi>"
  using assms
  by (auto simp: pw_le_iff refine_pw_simps)

lemma strengthen_SPEC: "m \<le> SPEC \<Phi> \<Longrightarrow> m \<le> SPEC(\<lambda>s. inres m s \<and> \<Phi> s)"
  -- "Strengthen SPEC by adding trivial upper bound for result"
  by (auto simp: pw_le_iff refine_pw_simps)

lemma weaken_SPEC:
  "m \<le> SPEC \<Phi> \<Longrightarrow> (\<And>x. \<Phi> x \<Longrightarrow> \<Psi> x) \<Longrightarrow> m \<le> SPEC \<Psi>"
  by (force elim!: order_trans)


lemma ife_FAIL_to_ASSERT_cnv: 
  "(if \<Phi> then m else FAIL) = op_nres_ASSERT_bnd \<Phi> m"
  by (cases \<Phi>, auto)



lemma param_op_nres_ASSERT_bnd[param]:
  assumes "\<Phi>' \<Longrightarrow> \<Phi>"
  assumes "\<lbrakk>\<Phi>'; \<Phi>\<rbrakk> \<Longrightarrow> (m,m')\<in>\<langle>R\<rangle>nres_rel"
  shows "(op_nres_ASSERT_bnd \<Phi> m, op_nres_ASSERT_bnd \<Phi>' m') \<in> \<langle>R\<rangle>nres_rel"
  using assms
  by (auto simp: pw_le_iff refine_pw_simps nres_rel_def)

declare autoref_FAIL[param]


(* TODO: Move *)
method_setup refine_vcg = 
  {* Attrib.thms >> (fn add_thms => fn ctxt => SIMPLE_METHOD' (
    Refine.rcg_tac (add_thms @ Refine.vcg.get ctxt) ctxt THEN_ALL_NEW (TRY o Refine.post_tac ctxt)
  )) *} 
  "Refinement framework: Generate refinement and verification conditions"



(*****************************)
(* Refine_Transfer *)
lemma (in transfer) transfer_sum[refine_transfer]:
  assumes "\<And>l. \<alpha> (fl l) \<le> Fl l"
  assumes "\<And>r. \<alpha> (fr r) \<le> Fr r"
  shows "\<alpha> (case_sum fl fr x) \<le> (case_sum Fl Fr x)"
  using assms by (auto split: sum.split)


(* TODO: Move *)
lemma nres_of_transfer[refine_transfer]: "nres_of x \<le> nres_of x" by simp



(*****************************)
(* Refine_Foreach  *)
(* TODO: Change in Refine_Foreach(?)! *)
declare FOREACH_patterns[autoref_op_pat_def]

(*****************************)
(* Refine_Recursion  *)
theorem param_RECT[param]:
  assumes "(B, B') \<in> (Ra \<rightarrow> \<langle>Rr\<rangle>nres_rel) \<rightarrow> Ra \<rightarrow> \<langle>Rr\<rangle>nres_rel"
    and "trimono B"
  shows "(REC\<^sub>T B, REC\<^sub>T B')\<in> Ra \<rightarrow> \<langle>Rr\<rangle>nres_rel"
  using autoref_RECT assms by simp

definition "REC_annot pre post body x \<equiv> 
  REC (\<lambda>D x. do {ASSERT (pre x); r\<leftarrow>body D x; ASSERT (post x r); RETURN r}) x"

theorem REC_annot_rule[refine_vcg]:
  assumes M: "trimono body"
  and P: "pre x"
  and S: "\<And>f x. \<lbrakk>\<And>x. pre x \<Longrightarrow> f x \<le> SPEC (post x); pre x\<rbrakk> 
          \<Longrightarrow> body f x \<le> SPEC (post x)"
  and C: "\<And>r. post x r \<Longrightarrow> \<Phi> r"
  shows "REC_annot pre post body x \<le> SPEC \<Phi>"
proof -
  from `trimono body` have [refine_mono]:
    "\<And>f g x xa. (\<And>x. flat_ge (f x) (g x)) \<Longrightarrow> flat_ge (body f x) (body g x)"
    "\<And>f g x xa. (\<And>x. f x \<le> g x) \<Longrightarrow> body f x \<le> body g x"
    apply -
    unfolding trimono_def monotone_def fun_ord_def mono_def le_fun_def
    apply (auto)
    done

  show ?thesis
    unfolding REC_annot_def
    apply (rule order_trans[where y="SPEC (post x)"])
    apply (refine_rcg 
      refine_vcg 
      REC_rule[where pre=pre and M="\<lambda>x. SPEC (post x)"]
      order_trans[OF S]
    )
    apply fact
    apply simp
    using C apply (auto) []
    done
qed



(*****************************)
(* Refine_While  *)
context begin interpretation autoref_syn .
(* TODO: Change in Refine_While *)
lemma [autoref_op_pat_def]:
  "WHILEIT I \<equiv> OP (WHILEIT I)"
  "WHILEI I \<equiv> OP (WHILEI I)"
  by auto
end

context begin interpretation autoref_syn .

(* TODO: Move, change original WHILE-rule  *)
lemma autoref_WHILE'[autoref_rules]:
  assumes "\<And>x x'. \<lbrakk> (x,x')\<in>R\<rbrakk> \<Longrightarrow> (c x,c'$x') \<in> Id"
  assumes "\<And>x x'. \<lbrakk> REMOVE_INTERNAL c' x'; (x,x')\<in>R\<rbrakk> 
    \<Longrightarrow> (f x,f'$x') \<in> \<langle>R\<rangle>nres_rel"
  shows "(WHILE c f,
      (OP WHILE ::: (R\<rightarrow>Id) \<rightarrow> (R\<rightarrow>\<langle>R\<rangle>nres_rel) \<rightarrow> R \<rightarrow> \<langle>R\<rangle>nres_rel)$c'$f'
    )\<in>R \<rightarrow> \<langle>R\<rangle>nres_rel"
  using assms
  by (auto simp add: nres_rel_def fun_rel_def intro!: WHILE_refine)

(* TODO: Move, change original WHILE-rule  *)
lemma autoref_WHILEI[autoref_rules]:
  assumes "\<And>x x'. \<lbrakk>I x'; (x,x')\<in>R\<rbrakk> \<Longrightarrow> (c x,c'$x') \<in> Id"
  assumes "\<And>x x'. \<lbrakk>REMOVE_INTERNAL c' x'; I x'; (x,x')\<in>R\<rbrakk> \<Longrightarrow>(f x,f'$x') \<in> \<langle>R\<rangle>nres_rel"
  assumes "I s' \<Longrightarrow> (s,s')\<in>R"
  shows "(WHILE c f s,
      (OP (WHILEI I) ::: (R\<rightarrow>Id) \<rightarrow> (R\<rightarrow>\<langle>R\<rangle>nres_rel) \<rightarrow> R \<rightarrow> \<langle>R\<rangle>nres_rel)$c'$f'$s'
    )\<in>\<langle>R\<rangle>nres_rel"
  using assms
  by (auto simp add: nres_rel_def fun_rel_def intro!: WHILE_refine')


(* TODO: Move, change original WHILE-rule  *)
lemma autoref_WHILEI'[autoref_rules]:
  assumes "\<And>x x'. \<lbrakk> (x,x')\<in>R; I x'\<rbrakk> \<Longrightarrow> (c x,c'$x') \<in> Id"
  assumes "\<And>x x'. \<lbrakk> REMOVE_INTERNAL c' x'; (x,x')\<in>R; I x'\<rbrakk> 
    \<Longrightarrow> (f x,f'$x') \<in> \<langle>R\<rangle>nres_rel"
  shows "(WHILE c f,
      (OP (WHILEI I) ::: (R\<rightarrow>Id) \<rightarrow> (R\<rightarrow>\<langle>R\<rangle>nres_rel) \<rightarrow> R \<rightarrow> \<langle>R\<rangle>nres_rel)$c'$f'
    )\<in>R \<rightarrow> \<langle>R\<rangle>nres_rel"
    unfolding autoref_tag_defs
    by (parametricity 
      add: autoref_WHILEI[unfolded autoref_tag_defs]
      assms[unfolded autoref_tag_defs])


(* TODO: Move, change original WHILE-rule  *)
lemma autoref_WHILEIT[autoref_rules]:
  assumes "\<And>x x'. \<lbrakk>I x'; (x,x')\<in>R\<rbrakk> \<Longrightarrow> (c x,c'$x') \<in> Id"
  assumes "\<And>x x'. \<lbrakk>REMOVE_INTERNAL c' x'; I x'; (x,x')\<in>R\<rbrakk> \<Longrightarrow>(f x,f'$x') \<in> \<langle>R\<rangle>nres_rel"
  assumes "I s' \<Longrightarrow> (s,s')\<in>R"
  shows "(WHILET c f s,
      (OP (WHILEIT I) ::: (R\<rightarrow>Id) \<rightarrow> (R\<rightarrow>\<langle>R\<rangle>nres_rel) \<rightarrow> R \<rightarrow> \<langle>R\<rangle>nres_rel)$c'$f'$s'
    )\<in>\<langle>R\<rangle>nres_rel"
  using assms
  by (auto simp add: nres_rel_def fun_rel_def intro!: WHILET_refine')


(* TODO: Move, change original WHILE-rule  *)
lemma autoref_WHILEIT'[autoref_rules]:
  assumes "\<And>x x'. \<lbrakk> (x,x')\<in>R; I x'\<rbrakk> \<Longrightarrow> (c x,c'$x') \<in> Id"
  assumes "\<And>x x'. \<lbrakk> REMOVE_INTERNAL c' x'; (x,x')\<in>R; I x'\<rbrakk> 
    \<Longrightarrow> (f x,f'$x') \<in> \<langle>R\<rangle>nres_rel"
  shows "(WHILET c f,
      (OP (WHILEIT I) ::: (R\<rightarrow>Id) \<rightarrow> (R\<rightarrow>\<langle>R\<rangle>nres_rel) \<rightarrow> R \<rightarrow> \<langle>R\<rangle>nres_rel)$c'$f'
    )\<in>R \<rightarrow> \<langle>R\<rangle>nres_rel"
    unfolding autoref_tag_defs
    by (parametricity 
      add: autoref_WHILEIT[unfolded autoref_tag_defs]
      assms[unfolded autoref_tag_defs])


end



(*****************************)
(* Relators *)
lemma set_relD: "(s,s')\<in>\<langle>R\<rangle>set_rel \<Longrightarrow> x\<in>s \<Longrightarrow> \<exists>x'\<in>s'. (x,x')\<in>R"
  unfolding set_rel_def by blast

lemma set_relE[consumes 2]: 
  assumes "(s,s')\<in>\<langle>R\<rangle>set_rel" "x\<in>s"
  obtains x' where "x'\<in>s'" "(x,x')\<in>R"
  using set_relD[OF assms] ..

lemma param_prod': "\<lbrakk> 
  \<And>a b a' b'. \<lbrakk>p=(a,b); p'=(a',b')\<rbrakk> \<Longrightarrow> (f a b,f' a' b')\<in>R  
  \<rbrakk> \<Longrightarrow> (case_prod f p, case_prod f' p')\<in>R"
  by (auto split: prod.split)

(* TODO: Move to Relators *)
lemma list_rel_append1: "(as @ bs, l) \<in> \<langle>R\<rangle>list_rel 
  \<longleftrightarrow> (\<exists>cs ds. l = cs@ds \<and> (as,cs)\<in>\<langle>R\<rangle>list_rel \<and> (bs,ds)\<in>\<langle>R\<rangle>list_rel)"
  apply (simp add: list_rel_def list_all2_append1)
  apply auto
  apply (metis list_all2_lengthD)
  done

lemma list_rel_append2: "(l,as @ bs) \<in> \<langle>R\<rangle>list_rel 
  \<longleftrightarrow> (\<exists>cs ds. l = cs@ds \<and> (cs,as)\<in>\<langle>R\<rangle>list_rel \<and> (ds,bs)\<in>\<langle>R\<rangle>list_rel)"
  apply (simp add: list_rel_def list_all2_append2)
  apply auto
  apply (metis list_all2_lengthD)
  done


(* TODO: Move *)

ML {*
  (* Prefix p_ or wrong type supresses generation of relAPP *)

  fun cnv_relAPP t = let
    fun consider (Var ((name,_),T)) =
      if String.isPrefix "p_" name then false   
      else (
        case T of
          Type(@{type_name set},[Type(@{type_name prod},_)]) => true
        | _ => false)
    | consider _ = true

    fun strip_rcomb u : term * term list =
      let 
        fun stripc (x as (f$t, ts)) = 
          if consider t then stripc (f, t::ts) else x
        | stripc  x =  x
      in  stripc(u,[])  end;

    val (f,a) = strip_rcomb t
  in 
    Relators.list_relAPP a f
  end

  fun to_relAPP_conv ctxt = Refine_Util.f_tac_conv ctxt 
    cnv_relAPP 
    (ALLGOALS (simp_tac 
      (put_simpset HOL_basic_ss ctxt addsimps @{thms relAPP_def})))


  val to_relAPP_attr = Thm.rule_attribute (fn context => let
    val ctxt = Context.proof_of context
  in
    Conv.fconv_rule (Conv.arg1_conv (to_relAPP_conv ctxt))
  end)
*}

attribute_setup to_relAPP = {* Scan.succeed (to_relAPP_attr) *} 
  "Convert relator definition to prefix-form"




(*****************************)  
(* Parametricity-HOL *)

lemma dropWhile_param[param]: 
  "(dropWhile, dropWhile) \<in> (a \<rightarrow> bool_rel) \<rightarrow> \<langle>a\<rangle>list_rel \<rightarrow> \<langle>a\<rangle>list_rel"
  unfolding dropWhile_def by parametricity

term takeWhile
lemma takeWhile_param[param]: 
  "(takeWhile, takeWhile) \<in> (a \<rightarrow> bool_rel) \<rightarrow> \<langle>a\<rangle>list_rel \<rightarrow> \<langle>a\<rangle>list_rel"
  unfolding takeWhile_def by parametricity


lemma [relator_props]:  
  "bijective R \<Longrightarrow> single_valued R"
  "bijective R \<Longrightarrow> single_valued (R\<inverse>)"
  by (simp_all add: bijective_alt)

(* TODO: Move *)
declare bijective_Id[relator_props]
declare bijective_Empty[relator_props]

lemma param_subseteq[param]: 
  "\<lbrakk>single_valued (R\<inverse>)\<rbrakk> \<Longrightarrow> (op \<subseteq>, op \<subseteq>) \<in> \<langle>R\<rangle>set_rel \<rightarrow> \<langle>R\<rangle>set_rel \<rightarrow> bool_rel"
  by (clarsimp simp: set_rel_def single_valued_def) blast

lemma param_subset[param]: 
  "\<lbrakk>single_valued (R\<inverse>)\<rbrakk> \<Longrightarrow> (op \<subset>, op \<subset>) \<in> \<langle>R\<rangle>set_rel \<rightarrow> \<langle>R\<rangle>set_rel \<rightarrow> bool_rel"
  (* Fine-tuned proof for speed *)
  apply (simp add: set_rel_def single_valued_def)
  apply safe
  apply blast+
  done

lemma bij_set_rel_for_inj: 
  fixes R
  defines "\<alpha> \<equiv> fun_of_rel R" 
  assumes "bijective R" "(s,s')\<in>\<langle>R\<rangle>set_rel"  
  shows "inj_on \<alpha> s" "s' = \<alpha>`s"
  using assms
  unfolding bijective_def set_rel_def \<alpha>_def fun_of_rel_def[abs_def]
  apply (auto intro!: inj_onI ImageI simp: image_def)
  apply (metis (mono_tags) Domain.simps contra_subsetD tfl_some)
  apply (metis (mono_tags) someI)
  apply (metis DomainE contra_subsetD tfl_some)
  done


(*****************************)
(* Autoref-HOL  *)
lemmas [autoref_rules] = dropWhile_param takeWhile_param


(*****************************)
(* Autoref-Tool *)

method_setup autoref_solve_id_op = {*
  Scan.succeed (fn ctxt => SIMPLE_METHOD' (
    Autoref_Id_Ops.id_tac (Config.put Autoref_Id_Ops.cfg_ss_id_op false ctxt)
  ))
*}


(*****************************)
(* Autoref_Monadic  *)

(* TODO: Replace! *)
text {* Default setup of the autoref-tool for the monadic framework. *}

lemma autoref_monadicI1:
  assumes "(b,a)\<in>\<langle>R\<rangle>nres_rel"
  assumes "RETURN c \<le> b"
  shows "(RETURN c, a)\<in>\<langle>R\<rangle>nres_rel" "RETURN c \<le>\<Down>R a"
  using assms
  unfolding nres_rel_def
  by simp_all

lemma autoref_monadicI2:
  assumes "(b,a)\<in>\<langle>R\<rangle>nres_rel"
  assumes "nres_of c \<le> b"
  shows "(nres_of c, a)\<in>\<langle>R\<rangle>nres_rel" "nres_of c \<le> \<Down>R a"
  using assms
  unfolding nres_rel_def
  by simp_all

lemmas autoref_monadicI = autoref_monadicI1 autoref_monadicI2

ML {*
  structure Autoref_Monadic = struct

    val cfg_plain = Attrib.setup_config_bool @{binding autoref_plain} (K false)

    fun autoref_monadic_tac ctxt = let
      open Autoref_Tacticals
      val ctxt = Autoref_Phases.init_data ctxt
      val plain = Config.get ctxt cfg_plain
      val trans_thms = if plain then [] else @{thms the_resI}
  
    in
      resolve_tac @{thms autoref_monadicI}
      THEN' 
      IF_SOLVED (Autoref_Phases.all_phases_tac ctxt)
        (RefineG_Transfer.post_transfer_tac trans_thms ctxt)
        (K all_tac) (* Autoref failed *)

    end
  end
*}

method_setup autoref_monadic = {* let
    open Refine_Util Autoref_Monadic
    val autoref_flags = 
          parse_bool_config "trace" Autoref_Phases.cfg_trace
      ||  parse_bool_config "debug" Autoref_Phases.cfg_debug
      ||  parse_bool_config "plain" Autoref_Monadic.cfg_plain

  in
    parse_paren_lists autoref_flags 
    >>
    ( fn _ => fn ctxt => SIMPLE_METHOD' (
      let 
        val ctxt = Config.put Autoref_Phases.cfg_keep_goal true ctxt
      in autoref_monadic_tac ctxt end
    ))

  end

 *} 
 "Automatic Refinement and Determinization for the Monadic Refinement Framework"

(* Move to Refine Transfer *)
lemma dres_unit_simps[refine_transfer_post_simp]:
  "dbind (dRETURN (u::unit)) f = f ()"
  by auto

lemma Let_dRETURN_simp[refine_transfer_post_simp]:
  "Let m dRETURN = dRETURN m" by auto

(* TODO: Move *)
lemmas [refine_transfer_post_simp] = dres_monad_laws



(* TODO: Move *)
lemma le_ASSERT_defI1:
  assumes "c \<equiv> do {ASSERT \<Phi>; m}"
  assumes "\<Phi> \<Longrightarrow> m' \<le> c"
  shows "m' \<le> c"
  using assms
  by (simp add: le_ASSERTI)

lemma refine_ASSERT_defI1:
  assumes "c \<equiv> do {ASSERT \<Phi>; m}"
  assumes "\<Phi> \<Longrightarrow> m' \<le> \<Down>R c"
  shows "m' \<le> \<Down>R c"
  using assms
  by (simp, refine_vcg)

lemma le_ASSERT_defI2:
  assumes "c \<equiv> do {ASSERT \<Phi>; ASSERT \<Psi>; m}"
  assumes "\<lbrakk>\<Phi>; \<Psi>\<rbrakk> \<Longrightarrow> m' \<le> c"
  shows "m' \<le> c"
  using assms
  by (simp add: le_ASSERTI)

lemma refine_ASSERT_defI2:
  assumes "c \<equiv> do {ASSERT \<Phi>; ASSERT \<Psi>; m}"
  assumes "\<lbrakk>\<Phi>; \<Psi>\<rbrakk> \<Longrightarrow> m' \<le> \<Down>R c"
  shows "m' \<le> \<Down>R c"
  using assms
  by (simp, refine_vcg)

lemma ASSERT_le_defI:
  assumes "c \<equiv> do { ASSERT \<Phi>; m'}"
  assumes "\<Phi>"
  assumes "\<Phi> \<Longrightarrow> m' \<le> m"
  shows "c \<le> m"
  using assms by (auto)








(*****************************)
(* ICF *)

(* Note, currently select is defined in CAVA_Base! *)
lemma select_correct: 
  "select X \<le> SPEC (\<lambda>r. case r of None \<Rightarrow> X={} | Some x \<Rightarrow> x\<in>X)"
  unfolding select_def
  apply (refine_rcg refine_vcg)
  by auto


definition "prod_bhc bhc1 bhc2 \<equiv> \<lambda>n (a,b). (bhc1 n a * 33 + bhc2 n b) mod n"

lemma prod_bhc_ga[autoref_ga_rules]:
  "\<lbrakk> GEN_ALGO_tag (is_bounded_hashcode R eq1 bhc1); 
     GEN_ALGO_tag (is_bounded_hashcode S eq2 bhc2)\<rbrakk> 
  \<Longrightarrow> is_bounded_hashcode (R\<times>\<^sub>rS) (prod_eq eq1 eq2) (prod_bhc bhc1 bhc2)"
  unfolding is_bounded_hashcode_def prod_bhc_def prod_eq_def[abs_def]
  apply safe
  apply (auto dest: fun_relD)
  done

lemma prod_dhs_ga[autoref_ga_rules]:
  "\<lbrakk> GEN_ALGO_tag (is_valid_def_hm_size TYPE('a) n1);
     GEN_ALGO_tag (is_valid_def_hm_size TYPE('b) n2) \<rbrakk>
   \<Longrightarrow> is_valid_def_hm_size TYPE('a*'b) (n1+n2)"
   unfolding is_valid_def_hm_size_def by auto

(* TODO: Move *)
abbreviation ahs_rel where "ahs_rel bhc \<equiv> (map2set_rel (ahm_rel bhc))"



(*****************************)
(*  *)

end

