theory WHILE_aux
imports 
  "../../Libs/Collections/Refine_Dflt" Refine_Leof DFS_Framework_Refine_Aux
begin

(* TODO: Move. Replace REC_unfold lemma?*)
lemma REC_unfold': "trimono B \<Longrightarrow> REC B = B (REC B)"
  apply (rule ext) by (rule REC_unfold)

lemma RECT_unfold': "trimono B \<Longrightarrow> RECT B = B (RECT B)"
  apply (rule ext) by (rule RECT_unfold)

(* TODO: Replace REC_le_RECT-lemma *)
lemma REC_le_RECT': "REC body x \<le> REC\<^sub>T body x"
  apply (cases "trimono body")
  apply (rule REC_le_RECT, assumption)
  apply (simp add: RECT_def)
  done


thm bind_distrib_sup (* TODO: Rename to bind_distrib_sup1*)

lemma  bind_distrib_sup2: "bind m (\<lambda>x. sup (f x) (g x)) = sup (bind m f) (bind m g)"
  by (auto simp: pw_eq_iff refine_pw_simps)

lemma bind_distrib_Sup1: "bind (Sup M) f = (SUP m:M. bind m f)" 
  by (auto simp: pw_eq_iff refine_pw_simps)

lemma bind_distrib_Sup2: "F\<noteq>{} \<Longrightarrow> bind m (Sup F) = (SUP f:F. bind m f)"
  by (auto simp: pw_eq_iff refine_pw_simps)

lemma bind_le_nofailI:
  assumes "nofail m"
  assumes "\<And>x. RETURN x \<le> m \<Longrightarrow> f x \<le> m'"
  shows "m\<guillemotright>=f \<le> m'"
  using assms 
  by (simp add: refine_pw_simps pw_le_iff) blast

lemma bind_le_shift:
  "bind m f \<le> m' 
  \<longleftrightarrow> m \<le> (if nofail m' then SPEC (\<lambda>x. f x \<le> m') else FAIL)"
  by (auto simp: pw_le_iff refine_pw_simps)


theorem gen_kleene_lfp:
  fixes f:: "'a::complete_lattice \<Rightarrow> 'a"
  assumes CONT: "cont f"
  shows "lfp (\<lambda>x. sup m (f x)) = (SUP i. (f^^i) m)"
proof (rule antisym)
  let ?f = "(\<lambda>x. sup m (f x))"
  let ?K="{ (f^^i) m | i . True}"
  note MONO=cont_is_mono[OF CONT]
  note MONO'[refine_mono] = monoD[OF MONO]
  {
    fix i
    have "(f^^i) m \<le> lfp ?f"
      apply (induct i)
      apply simp
      apply (subst lfp_unfold, tagged_solver)
      apply simp
      
      apply (subst lfp_unfold, tagged_solver)
      apply simp
      by (metis MONO' le_supI2)
  } thus "(SUP i. (f^^i) m) \<le> lfp ?f" by (blast intro: SUP_least)
next
  let ?f = "(\<lambda>x. sup m (f x))"
  show "lfp ?f \<le> (SUP i. (f^^i) m)"
    apply (rule lfp_lowerbound)
    apply (rule sup_least)
    apply (rule order_trans[OF _ SUP_upper[where i=0]], simp_all) []
    unfolding SUP_def
    apply (simp add: contD [OF CONT] del: Sup_image_eq)
    apply (rule Sup_subset_mono)
    apply (auto)
    apply (rule_tac x="Suc i" in range_eqI)
    apply simp
    done
qed
  


(* Alternative proof that re-uses standard Kleene, but is more tedious *)
lemma SUP_funpow_contracting:
  fixes f :: "'a \<Rightarrow> ('a::complete_lattice)"
  assumes C: "cont f" 
  shows "f (SUP i. (f^^i) m) \<le> (SUP i. (f^^i) m)"
proof -
  have 1: "\<And>i x. f ((f^^i) x) = (f^^(Suc i)) x"
    by simp

  have "f (SUP i. (f^^i) m) = (SUP i. f ((f ^^ i) m))"
    unfolding SUP_def by (subst contD[OF C]) (simp_all)
  also have "\<dots> \<le> (SUP i. (f^^i) m)"
    apply (rule SUP_least)
    apply (simp, subst 1)
    apply (rule SUP_upper)
    ..
  finally show ?thesis .
qed

lemma gen_kleene_chain_conv:
  fixes f :: "'a::complete_lattice \<Rightarrow> 'a"
  assumes C: "cont f"
  shows "(SUP i. (f^^i) m) = (SUP i. ((\<lambda>x. sup m (f x))^^i) bot)"
proof -
  let ?f' = "\<lambda>x. sup m (f x)"
  show ?thesis
  proof (intro antisym SUP_least)
    from C have C': "cont ?f'"
      unfolding cont_def
      by (simp add: SUP_sup_distrib[symmetric])
  
    fix i
    show "(f ^^ i) m \<le> (SUP i. (?f' ^^ i) bot)"
    proof (induction i)
      case 0 show ?case
        by (rule order_trans[OF _ SUP_upper[where i=1]]) auto
    next
      case (Suc i)
      from cont_is_mono[OF C, THEN monoD, OF Suc]
      have "(f ^^ (Suc i)) m \<le> f (SUP i. ((\<lambda>x. sup m (f x)) ^^ i) bot)"
        by simp
      also have "\<dots> \<le> sup m \<dots>" by simp
      also note SUP_funpow_contracting[OF C']
      finally show ?case .
    qed
  next
    fix i
    show "(?f'^^i) bot \<le> (SUP i. (f^^i) m)"
    proof (induction i)
      case 0 thus ?case by simp
    next
      case (Suc i)
      from monoD[OF cont_is_mono[OF C] Suc]
      have "(?f'^^Suc i) bot \<le> sup m (f (SUP i. (f ^^ i) m))" 
        by (simp add: le_supI2)
      also have "\<dots> \<le> (SUP i. (f ^^ i) m)"
        apply (rule sup_least)
        apply (rule order_trans[OF _ SUP_upper[where i=0]], simp_all) []
        apply (rule SUP_funpow_contracting[OF C])
        done
      finally show ?case .
    qed
  qed
qed
   
theorem gen_kleene_lfp_alternative_proof:
  assumes C: "cont f"
  shows "lfp (\<lambda>x. sup m (f x)) = (SUP i. (f^^i) m)"
  apply (subst gen_kleene_chain_conv[OF C])
  apply (rule kleene_lfp)
    using C
    unfolding cont_def
    apply (simp add: SUP_sup_distrib[symmetric])
  done



(* Tail recursion transformation *)
lemma tailrec_transform_aux1:
  assumes "RETURN s \<le> m"
  shows "REC (\<lambda>D s. sup (a s \<guillemotright>= D) (b s)) s \<le> lfp (\<lambda>x. sup m (x\<guillemotright>=a)) \<guillemotright>= b"
  (is "REC ?F s \<le> lfp ?f \<guillemotright>= b")
proof (rule REC_rule[where pre = "\<lambda>s. RETURN s \<le> lfp ?f"], refine_mono)
  show "RETURN s \<le> lfp (\<lambda>x. sup m (x \<guillemotright>= a))"
    apply (subst lfp_unfold, tagged_solver)
    using assms apply (simp add: le_supI1)
    done
next
  fix f x
  assume IH: "\<And>x. RETURN x \<le> lfp ?f \<Longrightarrow>
                f x \<le> lfp ?f \<guillemotright>= b"
  and PRE: "RETURN x \<le> lfp ?f"

  show " sup (a x \<guillemotright>= f) (b x) \<le> lfp ?f \<guillemotright>= b"
  proof (rule sup_least)  
    show "b x \<le> lfp ?f \<guillemotright>= b"
    using PRE
      by (simp add: pw_le_iff refine_pw_simps) blast
  next
    from PRE have "a x \<le> lfp ?f \<guillemotright>= a"
      by (simp add: pw_le_iff refine_pw_simps) blast
    also have "\<dots> \<le> lfp ?f"
      apply (subst (2) lfp_unfold, tagged_solver)
      apply (simp add: le_supI2)
      done
    finally show "a x \<guillemotright>= f \<le> lfp ?f \<guillemotright>= b"
      using IH
      by (simp add: pw_le_iff refine_pw_simps) blast
  qed
qed  

corollary tailrec_transform1: 
  fixes m :: "'a nres"
  shows "m\<guillemotright>=REC (\<lambda>D s. sup (a s \<guillemotright>= D) (b s)) \<le> lfp (\<lambda>x. sup m (x\<guillemotright>=a)) \<guillemotright>= b"
  apply (cases "nofail m")
  apply (erule bind_le_nofailI)
  apply (erule tailrec_transform_aux1)
  apply (simp add: not_nofail_iff lfp_const)
  done

          
lemma tailrec_transform_aux2:
  fixes m :: "'a nres"
  shows "lfp (\<lambda>x. sup m (x\<guillemotright>=a)) 
    \<le> m \<guillemotright>= REC (\<lambda>D s. sup (a s \<guillemotright>= D) (RETURN s))"
  (is "lfp ?f \<le> m \<guillemotright>= REC ?F")
  apply (subst gen_kleene_lfp)
  apply (simp add: cont_def pw_eq_iff refine_pw_simps, blast)
  apply (rule SUP_least, simp)
proof -
  fix i
  show "((\<lambda>x. x \<guillemotright>= a) ^^ i) m \<le> m \<guillemotright>= REC (\<lambda>D s. sup (a s \<guillemotright>= D) (RETURN s))"
    apply (induction i arbitrary: m)
    apply simp
    apply (subst REC_unfold', refine_mono)
    apply (simp add: pw_le_iff refine_pw_simps, blast)
    
    apply (subst funpow_Suc_right)
    apply simp
    apply (rule order_trans)
    apply rprems
    apply simp

    apply (subst (2) REC_unfold', refine_mono)
    apply (simp add: bind_distrib_sup2)
    done
qed


lemma tailrec_transform_aux3: "REC (\<lambda>D s. sup (a s \<guillemotright>= D) (RETURN s)) s \<guillemotright>= b 
  \<le> REC (\<lambda>D s. sup (a s \<guillemotright>= D) (b s)) s"
  apply (subst bind_le_shift)
  apply (rule REC_rule, refine_mono)
  apply (rule TrueI)
  apply auto
    apply (subst (asm) (3) REC_unfold, refine_mono)
    apply (rule order_trans[OF bind_mono(1)[OF order_refl]])
    apply rprems
    apply (subst (3) REC_unfold, refine_mono)
    apply (simp add: refine_pw_simps pw_le_iff, blast)

    apply (subst REC_unfold, refine_mono, simp)
  done

lemma tailrec_transform2:
  "lfp (\<lambda>x. sup m (bind x a)) \<guillemotright>= b \<le> m \<guillemotright>= REC (\<lambda>D s. sup (a s \<guillemotright>= D) (b s))"
proof -
  from bind_mono(1)[OF tailrec_transform_aux2 order_refl]
  have "lfp (\<lambda>x. sup m (bind x a)) \<guillemotright>= b 
    \<le> m \<guillemotright>= (\<lambda>x. REC (\<lambda>D s. sup (a s \<guillemotright>= D) (RETURN s)) x \<guillemotright>= b)"
    by simp
  also from bind_mono(1)[OF order_refl tailrec_transform_aux3]
  have "\<dots> \<le> m \<guillemotright>= REC (\<lambda>D s. sup (a s \<guillemotright>= D) (b s))" .
  finally show ?thesis .
qed  

definition "tailrec_body a b \<equiv> (\<lambda>D s. sup (bind (a s) D) (b s))"

theorem tailrec_transform: 
  "m \<guillemotright>= REC (\<lambda>D s. sup (a s \<guillemotright>= D) (b s)) = lfp (\<lambda>x. sup m (bind x a)) \<guillemotright>= b"
  apply (rule antisym)
  apply (rule tailrec_transform1)  
  apply (rule tailrec_transform2)  
  done

theorem tailrec_transform': 
  "m \<guillemotright>= REC (tailrec_body a b) = lfp (\<lambda>x. sup m (bind x a)) \<guillemotright>= b"
  unfolding tailrec_body_def
  by (rule tailrec_transform)

lemma "WHILE c f = 
  REC (tailrec_body 
    (\<lambda>s. do {ASSUME (c s); f s}) 
    (\<lambda>s. do {ASSUME (\<not>c s); RETURN s})
  )"
  unfolding WHILE_def WHILEI_def WHILEI_body_def tailrec_body_def
  apply (fo_rule fun_cong arg_cong)+ apply (intro ext)
  apply auto
  done

lemma WHILEI_tailrec_conv: "WHILEI I c f = 
  REC (tailrec_body 
    (\<lambda>s. do {ASSERT (I s); ASSUME (c s); f s}) 
    (\<lambda>s. do {ASSERT (I s); ASSUME (\<not>c s); RETURN s})
  )"
  unfolding WHILEI_def WHILEI_body_def tailrec_body_def
  apply (fo_rule fun_cong arg_cong)+ apply (intro ext)
  apply auto
  done

lemma WHILEIT_tailrec_conv: "WHILEIT I c f = 
  RECT (tailrec_body 
    (\<lambda>s. do {ASSERT (I s); ASSUME (c s); f s}) 
    (\<lambda>s. do {ASSERT (I s); ASSUME (\<not>c s); RETURN s})
  )"
  unfolding WHILEIT_def WHILEI_body_def tailrec_body_def
  apply (fo_rule fun_cong arg_cong)+ apply (intro ext)
  apply auto
  done


definition "WHILEI_lfp_body I m c f \<equiv> 
  (\<lambda>x. sup m (do {
     s \<leftarrow> x;
     _ \<leftarrow> ASSERT (I s);
     _ \<leftarrow> ASSUME (c s);
     f s
   }))"
  
lemma WHILEI_lfp_conv: "m \<guillemotright>= WHILEI I c f = 
  do { 
    s \<leftarrow> lfp (WHILEI_lfp_body I m c f); 
    ASSERT (I s); 
    ASSUME (\<not>c s); 
    RETURN s 
  }"
  unfolding WHILEI_lfp_body_def
  apply (subst WHILEI_tailrec_conv)
  apply (subst tailrec_transform')
  ..

(* Just a reformulation of lfp_induct *)
lemma lfp_induct':
  assumes M: "mono f"
  assumes IS: "\<And>m. \<lbrakk> m \<le> lfp f; m \<le> P \<rbrakk> \<Longrightarrow> f m \<le> P"
  shows "lfp f \<le> P"
  apply (rule lfp_induct[OF M])
  apply (rule IS)
  by simp_all

lemma lfp_gen_induct:
  -- "Induction lemma for generalized lfps"
  assumes M: "mono f"
  notes [refine_mono] = monoD[OF M]
  assumes I0: "m0 \<le> P"
  assumes IS: "\<And>m. \<lbrakk>
      m \<le> lfp (\<lambda>s. sup m0 (f s));  (* Assume already established invariants *)
      m \<le> P;                       (* Assume invariant *)
      f m \<le> lfp (\<lambda>s. sup m0 (f s)) (* Assume that step preserved est. invars *)
    \<rbrakk> \<Longrightarrow> f m \<le> P"                 (* Show that step preserves invariant *)
  shows "lfp (\<lambda>s. sup m0 (f s)) \<le> P"
  apply (rule lfp_induct', tagged_solver)
  apply (rule sup_least)
  apply (rule I0)
  apply (rule IS, assumption+)
  apply (subst lfp_unfold, tagged_solver)
  apply (rule le_supI2)
  apply (rule monoD[OF M])
  .
  
definition msii -- "Most specific invariant for WHILE-loop"
  where "msii I m c f \<equiv> lfp (WHILEI_lfp_body I m c f)"

lemma [simp, intro!]: "mono (WHILEI_lfp_body I m c f)"
  unfolding WHILEI_lfp_body_def by tagged_solver

definition "filter_ASSUME c m \<equiv> do {x\<leftarrow>m; ASSUME (c x); RETURN x}"
definition "filter_ASSERT c m \<equiv> do {x\<leftarrow>m; ASSERT (c x); RETURN x}"

lemma [refine_pw_simps]: "nofail (filter_ASSUME c m) \<longleftrightarrow> nofail m"
  unfolding filter_ASSUME_def
  by (simp add: refine_pw_simps)

lemma [refine_pw_simps]: "inres (filter_ASSUME c m) x 
  \<longleftrightarrow> (nofail m \<longrightarrow> inres m x \<and> c x)"
  unfolding filter_ASSUME_def
  by (simp add: refine_pw_simps)

lemma msii_is_invar:
  "m \<le> msii I m c f"
  "m \<le> msii I m c f \<Longrightarrow> bind (filter_ASSUME c (filter_ASSERT I m)) f \<le> msii I m c f"
  unfolding  msii_def
  apply -
  apply (subst lfp_unfold, simp)
  apply (simp add: WHILEI_lfp_body_def)

  unfolding filter_ASSUME_def filter_ASSERT_def
  apply (subst lfp_unfold, simp)
  apply (simp add: WHILEI_lfp_body_def)
  apply (simp only: refine_pw_simps pw_le_iff) apply blast
  done

lemma WHILE_msii_conv: "m \<guillemotright>= WHILEI I c f 
  = filter_ASSUME (Not o c) (filter_ASSERT I (msii I m c f))"
  unfolding WHILEI_lfp_conv filter_ASSERT_def filter_ASSUME_def msii_def
  by simp

lemma msii_induct: 
  assumes I0: "m0 \<le> P"
  assumes IS: "\<And>m. \<lbrakk>m \<le> msii I m0 c f; m \<le> P;
    filter_ASSUME c (filter_ASSERT I m) \<guillemotright>= f \<le> msii I m0 c f
      \<rbrakk> \<Longrightarrow> filter_ASSUME c (filter_ASSERT I m) \<guillemotright>= f \<le> P"
  shows "msii I m0 c f \<le> P"
  unfolding msii_def WHILEI_lfp_body_def
  apply (rule lfp_gen_induct, tagged_solver)
  unfolding msii_def[symmetric] WHILEI_lfp_body_def[symmetric]
  apply (rule I0)
  apply (drule IS, assumption)
  unfolding filter_ASSERT_def filter_ASSUME_def
  apply simp_all
  done


  (* Auxiliary WHILE theory: Best invariant predicate *)

subsection {* Reachable without fail *}
text {* Reachable states in a while loop, ignoring failing states *}

  (* (R)eachable states (w)ith(o)ut (f)ail. All non-fail states reachable
    in a while-loop. *)
  inductive rwof :: "'a nres \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a nres) \<Rightarrow> 'a \<Rightarrow> bool"
    for m0 cond step 
    where
      init: "\<lbrakk> m0=RES X; x\<in>X \<rbrakk> \<Longrightarrow> rwof m0 cond step x"
    | step: "\<lbrakk> rwof m0 cond step x; cond x; step x = RES Y; y\<in>Y \<rbrakk> 
      \<Longrightarrow> rwof m0 cond step y"

  (* This lemma establishes consequences of rwof as invariants. *)
  lemma establish_rwof_invar:
    assumes I: "m0 \<le>\<^sub>n SPEC I"
    assumes S: "\<And>s. \<lbrakk> rwof m0 cond step s; I s; cond s \<rbrakk> 
      \<Longrightarrow> step s \<le>\<^sub>n SPEC I"
    assumes "rwof m0 cond step s"
    shows "I s"
    using assms(3)
    apply induct
    using I apply auto []
    using S apply fastforce []
    done

  (* Predicate to express an rwof-invariant *)
  definition "is_rwof_invar m0 cond step I \<equiv> 
      (m0 \<le>\<^sub>n SPEC I)
    \<and> (\<forall>s. rwof m0 cond step s \<and> I s \<and> cond s 
        \<longrightarrow> step s \<le>\<^sub>n SPEC I )"

  lemma is_rwof_invarI[intro?]:
    assumes I: "m0 \<le>\<^sub>n SPEC I"
    assumes S: "\<And>s. \<lbrakk> rwof m0 cond step s; I s; cond s \<rbrakk> 
      \<Longrightarrow> step s \<le>\<^sub>n SPEC I"
    shows "is_rwof_invar m0 cond step I"
    using assms unfolding is_rwof_invar_def by blast

  lemma rwof_cons: "\<lbrakk>is_rwof_invar m0 cond step I; rwof m0 cond step s\<rbrakk> \<Longrightarrow> I s"
    unfolding is_rwof_invar_def
    using establish_rwof_invar[of m0 I cond step s]
    by blast
    

  (* This lemma proves a specification for the while loop, based on rwof
    and a nofail-proof. *)
  lemma rwof_WHILE_rule:
    assumes I0: "m0 \<le> SPEC I"
    assumes S: "\<And>s. \<lbrakk> rwof m0 cond step s; I s; cond s \<rbrakk> \<Longrightarrow> step s \<le> SPEC I"
    shows "m0 \<guillemotright>= WHILE cond step \<le> SPEC (\<lambda>s. rwof m0 cond step s \<and> \<not>cond s \<and> I s)"
  proof -
    from I0 obtain M0 where [simp]: "m0 = RES M0" and "M0 \<subseteq> Collect I"
      by (cases m0) auto

    show ?thesis 
      apply simp
      apply (refine_rcg refine_vcg)
      apply (rule WHILE_rule[where I="\<lambda>s. I s \<and> rwof m0 cond step s"])
    proof -
      fix s assume "s\<in>M0" thus "I s \<and> rwof m0 cond step s"
        using I0 by (auto intro: rwof.init) []

    next
      fix s
      assume A: "I s \<and> rwof m0 cond step s" and C: "cond s"
      hence "step s \<le> SPEC I" by - (rule S, simp_all)
      also then obtain S' where "step s = RES S'" by (cases "step s") auto
      from rwof.step[OF conjunct2[OF A] C this] this 
      have "step s \<le> SPEC (rwof m0 cond step)" by auto
      finally (SPEC_rule_conjI) 
      show "step s \<le> SPEC (\<lambda>s. I s \<and> rwof m0 cond step s)" .
    qed auto
  qed

subsection {* Filtering out states that satisfy the loop condition *}
  (* A shortcut definition for filtering out all results that satisfy
    the loop condition. Intuitively, filtering out these results from the 
    best invariant gives the results of the while loop. *)
  definition filter_nb :: "('a \<Rightarrow> bool) \<Rightarrow> 'a nres \<Rightarrow> 'a nres" where 
    "filter_nb b I \<equiv> do {s\<leftarrow>I; ASSUME (\<not>b s); RETURN s}"
  
  lemma pw_filter_nb[refine_pw_simps]:
    "nofail (filter_nb b I) \<longleftrightarrow> nofail I"
    "inres (filter_nb b I) x \<longleftrightarrow> (nofail I \<longrightarrow> inres I x \<and> \<not>b x)"
    unfolding filter_nb_def
    by (simp_all add: refine_pw_simps)

  lemma filter_nb_mono: "m\<le>m' \<Longrightarrow> filter_nb cond m \<le> filter_nb cond m'"
    unfolding filter_nb_def
    by refine_mono

  lemma filter_nb_cont: 
    "filter_nb cond (Sup M) = Sup {filter_nb cond m | m. m \<in> M}"
    apply (rule antisym)
    apply (simp add: pw_le_iff refine_pw_simps)
    apply (auto intro: not_nofail_inres simp: refine_pw_simps) []
    
    apply (simp add: pw_le_iff refine_pw_simps)
    apply (auto intro: not_nofail_inres simp: refine_pw_simps) []
    done

  lemma filter_nb_FAIL[simp]: "filter_nb cond FAIL = FAIL"
    by (simp add: filter_nb_def)

  lemma filter_nb_RES[simp]: "filter_nb cond (RES X) = RES {x\<in>X. \<not>cond x}"
    by (simp add: pw_eq_iff refine_pw_simps)

subsection {* Bounded while-loop *}

  (* A while rule that establishes an inductive invariant,
    and then filters out the results satisfying the condition. *)
  lemma WHILE_rule_gen_le:
    assumes I0: "m0 \<le> I"
    assumes ISTEP: "\<And>s. \<lbrakk>RETURN s \<le> I; b s\<rbrakk> \<Longrightarrow> f s \<le> I"
    shows "m0 \<guillemotright>= WHILE b f \<le> filter_nb b I"
      apply (unfold WHILE_def WHILEI_def)
      apply (refine_rcg order_trans[OF I0] refine_vcg pw_bind_leI)
      using I0 apply (simp add: pw_le_iff refine_pw_simps)

      apply (rule REC_rule[OF WHILEI_body_trimono, where pre="\<lambda>s. RETURN s \<le> I"])
        using I0 apply (simp add: pw_le_iff refine_pw_simps)
      
        unfolding WHILEI_body_def
        apply (split split_if)+
        apply (intro impI conjI)
        apply (simp_all)
        using ISTEP
        apply (simp (no_asm_use) only: pw_le_iff refine_pw_simps) apply blast
        apply (simp only: pw_le_iff refine_pw_simps) apply metis 
    done


  primrec bounded_WHILE' 
    :: "nat \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a nres) \<Rightarrow> 'a nres \<Rightarrow> 'a nres" 
  where
    "bounded_WHILE' 0 cond step m = m"
  | "bounded_WHILE' (Suc n) cond step m = do {
      x \<leftarrow> m;
      if cond x then bounded_WHILE' n cond step (step x)
      else RETURN x
    }"

  primrec bounded_WHILE
    :: "nat \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a nres) \<Rightarrow> 'a nres \<Rightarrow> 'a nres" 
  where
    "bounded_WHILE 0 cond step m = m"
  | "bounded_WHILE (Suc n) cond step m = do {
      x \<leftarrow> bounded_WHILE n cond step m;
      if cond x then step x
      else RETURN x
    }"

  lemma bounded_WHILE_shift: "do {
             x \<leftarrow> m;
             if cond x then bounded_WHILE n cond step (step x) else RETURN x
           } = do {
             x \<leftarrow> bounded_WHILE n cond step m;
             if cond x then step x else RETURN x
           }"
  proof (induction n arbitrary: m)
    case 0 thus ?case by (simp cong: if_cong)
  next
    case (Suc n) 

    have aux1: "do {
      x \<leftarrow> m;
      if cond x then do {
                       x \<leftarrow> bounded_WHILE n cond step (step x);
                       if cond x then step x else RETURN x
                     }
      else RETURN x
    } = do {
      x \<leftarrow> do {
        x \<leftarrow> m; 
        if cond x then bounded_WHILE n cond step (step x) else RETURN x
      };
      if cond x then step x else RETURN x
    }"
    by (simp add: pw_eq_iff refine_pw_simps)

    show ?case
      apply (simp cong: if_cong)
      apply (subst aux1)
      apply (subst Suc.IH)
      apply simp
      done
  qed


  lemma bounded_WHILE'_eq: 
    "bounded_WHILE' n cond step m = bounded_WHILE n cond step m"
    apply (induct n arbitrary: m)
    apply (auto cong: if_cong simp: bounded_WHILE_shift)
    done

  lemma mWHILE_unfold: "m \<guillemotright>= WHILE cond step = do {
      x \<leftarrow> m;
      if cond x then step x \<guillemotright>= WHILE cond step
      else RETURN x
    }"
    by (subst WHILE_unfold[abs_def]) (rule refl)

  lemma WHILE_bounded_aux1: 
    "filter_nb cond (bounded_WHILE n cond step m) \<le> m \<guillemotright>= WHILE cond step"
    unfolding bounded_WHILE'_eq[symmetric]
    apply (induct n arbitrary: m)
    apply simp
    apply (subst mWHILE_unfold)
    apply (simp add: pw_le_iff refine_pw_simps, blast)
    
    apply simp
    apply (subst mWHILE_unfold)
    apply (auto simp add: pw_le_iff refine_pw_simps)
    done

  lemma WHILE_bounded_aux2:
    "m \<guillemotright>= WHILE cond step 
      \<le> filter_nb cond (Sup {bounded_WHILE n cond step m | n. True})"
    apply (rule WHILE_rule_gen_le)
    apply (metis (mono_tags, lifting) Sup_upper bounded_WHILE.simps(1) 
      mem_Collect_eq)
  proof - 
    fix s
    assume "RETURN s \<le> Sup {bounded_WHILE n cond step m |n. True}"
    then obtain n where "RETURN s \<le> bounded_WHILE n cond step m"
      by (fold inres_def) (auto simp: refine_pw_simps)
    moreover assume "cond s"
    ultimately have "step s \<le> bounded_WHILE (Suc n) cond step m"
      by (simp add: pw_le_iff refine_pw_simps, blast)
    thus "step s \<le> Sup {bounded_WHILE n cond step m |n. True}"
      by (metis (mono_tags, lifting) Sup_upper2 mem_Collect_eq)
  qed

   
  lemma WHILE_bounded: 
    "m \<guillemotright>= WHILE cond step 
    = filter_nb cond (Sup {bounded_WHILE n cond step m | n. True})"
    apply (rule antisym)
    apply (rule WHILE_bounded_aux2)

    apply (simp add: filter_nb_cont)
    apply (rule Sup_least)
    apply (auto simp: WHILE_bounded_aux1)
    done

  subsubsection {* Relation to rwof *}

  lemma rwof_in_bounded_WHILE:
    assumes "rwof m0 cond step s" 
    shows "\<exists>n. RETURN s \<le> (bounded_WHILE n cond step m0)"
    using assms
    apply induction
    apply (rule_tac x=0 in exI)
    apply simp

    apply clarsimp
    apply (rule_tac x="Suc n" in exI)
    apply (auto simp add: pw_le_iff refine_pw_simps) []
    done

  lemma bounded_WHILE_FAIL_rwof:
    assumes "bounded_WHILE n cond step m0 = FAIL"
    assumes M0: "m0 \<noteq> FAIL"
    shows "\<exists>n'<n. \<exists>x X. 
        bounded_WHILE n' cond step m0 = RES X 
      \<and> x\<in>X \<and> cond x \<and> step x = FAIL"
    using assms
  proof (induction n)
    case 0 thus ?case by simp
  next
    case (Suc n)
    assume "bounded_WHILE (Suc n) cond step m0 = FAIL"
    hence "bounded_WHILE n cond step m0 = FAIL 
      \<or> (\<exists>X x. bounded_WHILE n cond step m0 = RES X 
          \<and> x\<in>X \<and> cond x \<and> step x = FAIL)"
      (is "?C1 \<or> ?C2")
      apply (cases "bounded_WHILE n cond step m0")
      apply simp
      apply (simp add: pw_eq_iff refine_pw_simps split: split_if_asm)
      apply (auto intro: not_nofail_inres simp: refine_pw_simps)
      done
    moreover {
      assume ?C1
      from Suc.IH[OF this M0] obtain n' x X where "n' < n" and
        1: "bounded_WHILE n' cond step m0 = RES X \<and> x\<in>X \<and> cond x \<and> step x = FAIL"
        by blast
      hence 2: "n' < Suc n" by simp
      from 1 2 have ?case by blast
    } moreover {
      assume ?C2
      hence ?case by blast
    } ultimately show ?case by blast
  qed

  lemma bounded_WHILE_RES_rwof:
    assumes "bounded_WHILE n cond step m0 = RES X"
    assumes "x\<in>X"
    shows "rwof m0 cond step x"
    using assms
  proof (induction n arbitrary: x X)
    case 0 thus ?case by (simp add: rwof.init)
  next
    case (Suc n)

    from Suc.prems(1) obtain Xh where
      BWN: "bounded_WHILE n cond step m0 = RES Xh"
      and "\<forall>x\<in>Xh. cond x \<longrightarrow> nofail (step x)"
      and "X = {x'. \<exists>x\<in>Xh. cond x \<and> inres (step x) x'} \<union> {x. x\<in>Xh \<and> \<not>cond x} "
      apply (cases "bounded_WHILE n cond step m0")
      apply simp
      apply (rule that, assumption)
      apply (force simp: refine_pw_simps pw_eq_iff) []
      apply (auto simp add: refine_pw_simps pw_eq_iff split: split_if_asm)
      done
    with Suc.prems(2) obtain xh X' where 
        "xh\<in>Xh" and
      C: "cond xh \<longrightarrow> step xh = RES X' \<and> x\<in>X'" and NC:  "\<not>cond xh \<longrightarrow> x=xh"
      by (auto simp: nofail_RES_conv)
    
    show ?case  
      apply (cases "cond xh")
        apply (rule rwof.step[where Y=X'])
          apply (rule Suc.IH[OF BWN `xh\<in>Xh`])
          apply assumption
          apply (simp_all add: C) [2]
        
        apply (rule Suc.IH[OF BWN])
        apply (simp add: NC `xh\<in>Xh`)
      done
  qed
      

  lemma rwof_FAIL_imp_WHILE_FAIL:
    assumes RW: "rwof m0 cond step s" 
    and C: "cond s" 
    and S: "step s = FAIL"
    shows "m0 \<guillemotright>= WHILE cond step = FAIL"
  proof -
    from rwof_in_bounded_WHILE[OF RW] obtain n where 
      "RETURN s \<le> bounded_WHILE n cond step m0" ..
    with C have "step s \<le> bounded_WHILE (Suc n) cond step m0"
      by (auto simp add: pw_le_iff refine_pw_simps)
    with S have "bounded_WHILE (Suc n) cond step m0 = FAIL" by simp
    with WHILE_bounded_aux1[of cond "Suc n" step m0] show ?thesis
      by (simp del: bounded_WHILE.simps)
  qed

  lemma pw_bounded_WHILE_RES_rwof: "\<lbrakk> nofail (bounded_WHILE n cond step m0); 
    inres (bounded_WHILE n cond step m0) x \<rbrakk> \<Longrightarrow> rwof m0 cond step x"
    using bounded_WHILE_RES_rwof[of n cond step m0 _ x]
    by (auto simp add: pw_eq_iff)
    
  corollary WHILE_nofail_imp_rwof_nofail:
    assumes "nofail (m0 \<guillemotright>= WHILE cond step)"
    assumes RW: "rwof m0 cond step s" 
    assumes C: "cond s"
    shows "nofail (step s)"
    apply (rule ccontr) apply (simp add: nofail_def)
    using assms rwof_FAIL_imp_WHILE_FAIL[OF RW C]
    by auto

  (* TODO: Move *)
  lemma WHILE_le_WHILEI: "WHILE b f s \<le> WHILEI I b f s"
    unfolding WHILE_def
    by (rule WHILEI_weaken) simp

  corollary WHILEI_nofail_imp_rwof_nofail:
    assumes NF: "nofail (m0 \<guillemotright>= WHILEI I cond step)"
    assumes RW: "rwof m0 cond step s" 
    assumes C: "cond s"
    shows "nofail (step s)"
  proof -
    from NF have "nofail (m0 \<guillemotright>= WHILE cond step)"
      using WHILE_le_WHILEI 
      by (fastforce simp: pw_le_iff refine_pw_simps)
    from WHILE_nofail_imp_rwof_nofail[OF this RW C] show ?thesis .
  qed

  corollary WHILET_nofail_imp_rwof_nofail:
    assumes NF: "nofail (m0 \<guillemotright>= WHILET cond step)"
    assumes RW: "rwof m0 cond step s" 
    assumes C: "cond s"
    shows "nofail (step s)"
  proof -
    from NF have "nofail (m0 \<guillemotright>= WHILE cond step)"
      using WHILEI_le_WHILEIT unfolding WHILE_def WHILET_def 
      by (fastforce simp: pw_le_iff refine_pw_simps)
    from WHILE_nofail_imp_rwof_nofail[OF this RW C] show ?thesis .
  qed

  corollary WHILEIT_nofail_imp_rwof_nofail:
    assumes NF: "nofail (m0 \<guillemotright>= WHILEIT I cond step)"
    assumes RW: "rwof m0 cond step s" 
    assumes C: "cond s"
    shows "nofail (step s)"
  proof -
    from NF have "nofail (m0 \<guillemotright>= WHILE cond step)"
      using WHILE_le_WHILEI WHILEI_le_WHILEIT unfolding WHILE_def
      by (fastforce simp: pw_le_iff refine_pw_simps)
    from WHILE_nofail_imp_rwof_nofail[OF this RW C] show ?thesis .
  qed

  lemma pw_rwof_in_bounded_WHILE:
    "rwof m0 cond step x \<Longrightarrow> \<exists>n. inres (bounded_WHILE n cond step m0) x"
    using rwof_in_bounded_WHILE[of m0 cond step x]
    by (auto simp add: pw_le_iff intro: not_nofail_inres)

subsection {* WHILE-loops in the nofail-case *}
  lemma nofail_WHILE_eq_rwof: 
    assumes NF: "nofail (m0 \<guillemotright>= WHILE cond step)"
    shows "m0 \<guillemotright>= WHILE cond step = SPEC (\<lambda>s. rwof m0 cond step s \<and> \<not>cond s)"
  proof -
    from NF WHILE_bounded[of m0 cond step] have NF': 
      "nofail (Sup {filter_nb cond m |m. \<exists>n. m = bounded_WHILE n cond step m0})"
      by (auto simp: filter_nb_cont)

    (*from NF WHILE_bounded[of m0 cond step] obtain X 
      where FR:
        "Sup {filter_nb cond m |m. \<exists>n. m = bounded_WHILE n cond step m0} = RES X"
      by (auto simp: nofail_RES_conv filter_nb_cont)*)
    
    show ?thesis
      unfolding WHILE_bounded[of m0 cond step] filter_nb_cont
      apply simp
    proof (rule antisym)
      show "Sup {filter_nb cond m |m. \<exists>n. m = bounded_WHILE n cond step m0}
        \<le> SPEC (\<lambda>s. rwof m0 cond step s \<and> \<not> cond s)"
        using NF' pw_bounded_WHILE_RES_rwof[of _ cond step m0]
        by (fastforce simp: pw_le_iff refine_pw_simps)
    next
      show "SPEC (\<lambda>s. rwof m0 cond step s \<and> \<not> cond s)
        \<le> Sup {filter_nb cond m |m. \<exists>n. m = bounded_WHILE n cond step m0}"      
        using NF' pw_rwof_in_bounded_WHILE[of m0 cond step]
        by (fastforce simp: pw_le_iff refine_pw_simps)
    qed
  qed

  lemma WHILE_refine_rwof:
    assumes "nofail (m \<guillemotright>= WHILE c f) \<Longrightarrow> mi \<le> SPEC (\<lambda>s. rwof m c f s \<and> \<not>c s)"
    shows "mi \<le> m \<guillemotright>= WHILE c f"
    apply (cases "nofail (m \<guillemotright>= WHILE c f)")
    apply (subst nofail_WHILE_eq_rwof, simp, fact)
    apply (simp add: pw_le_iff)
    done


  lemma pw_rwof_init:
    assumes NF: "nofail (m0 \<guillemotright>= WHILE cond step)"
    shows "inres m0 s \<Longrightarrow> rwof m0 cond step s" and "nofail m0"
    apply -
    using NF apply (cases m0, simp)
    apply (rule rwof.init, assumption)
    apply auto []
    
    using NF apply (simp add: refine_pw_simps)
    done

  lemma rwof_init:
    assumes NF: "nofail (m0 \<guillemotright>= WHILE cond step)" 
    shows "m0 \<le> SPEC (rwof m0 cond step)"
    using pw_rwof_init[OF NF]
    by (simp add: pw_le_iff refine_pw_simps)
  
  lemma pw_rwof_step':
    assumes NF: "nofail (step s)"
    assumes R: "rwof m0 cond step s"
    assumes C: "cond s"
    shows "inres (step s) s' \<Longrightarrow> rwof m0 cond step s'"
    using NF
    apply (clarsimp simp: nofail_RES_conv)
    apply (rule rwof.step[OF R C])
    apply (assumption)
    by simp

  lemma rwof_step': 
    "\<lbrakk> nofail (step s); rwof m0 cond step s; cond s \<rbrakk>
    \<Longrightarrow> step s \<le> SPEC (rwof m0 cond step)"
    using pw_rwof_step'[of step s m0 cond]
    by (simp add: pw_le_iff refine_pw_simps)

  lemma pw_rwof_step:
    assumes NF: "nofail (m0 \<guillemotright>= WHILE cond step)"
    assumes R: "rwof m0 cond step s"
    assumes C: "cond s"
    shows "inres (step s) s' \<Longrightarrow> rwof m0 cond step s'"
      and "nofail (step s)"
    using WHILE_nofail_imp_rwof_nofail[OF NF R C] pw_rwof_step'[OF _ assms(2-)]
    by simp_all

  lemma rwof_step: 
    "\<lbrakk> nofail (m0 \<guillemotright>= WHILE cond step); rwof m0 cond step s; cond s \<rbrakk>
    \<Longrightarrow> step s \<le> SPEC (rwof m0 cond step)"
    using pw_rwof_step[of m0 cond step s]
    by (simp add: pw_le_iff refine_pw_simps)

  lemma (in -) rwof_leof_init: "m \<le>\<^sub>n SPEC (rwof m c f)"
    apply rule
    using rwof.init
    apply (fastforce simp: nofail_RES_conv)
    done

  lemma (in -) rwof_leof_step: "\<lbrakk>rwof m c f s; c s\<rbrakk> \<Longrightarrow> f s \<le>\<^sub>n SPEC (rwof m c f)"
    apply rule
    using rwof.step
    apply (fastforce simp: nofail_RES_conv)
    done


  lemma (in -) rwof_step_refine:
    assumes NF: "nofail (m0 \<guillemotright>= WHILE cond step)"  
    assumes A: "rwof m0 cond step' s"
    assumes FR: "\<And>s. \<lbrakk> rwof m0 cond step s; cond s \<rbrakk> \<Longrightarrow> step' s \<le>\<^sub>n step s"
    shows "rwof m0 cond step s"
    apply (rule establish_rwof_invar[OF _ _ A])
      apply (rule rwof_leof_init)

      apply (rule leof_trans_nofail[OF FR], assumption+)
        apply (rule WHILE_nofail_imp_rwof_nofail[OF NF], assumption+)

        apply (simp add: rwof_leof_step)
    done

subsection {* Adding Invariants *}
  lemma WHILE_eq_I_rwof: "m \<guillemotright>= WHILE c f = m \<guillemotright>= WHILEI (rwof m c f) c f"
  proof (rule antisym)
    have "m \<guillemotright>= WHILEI (rwof m c f) c f
      \<le> \<Down>{(s,s) | s. rwof m c f s} 
        (m \<guillemotright>= WHILE c f)"
      unfolding WHILE_def
      apply (rule bind_refine)
      apply (rule intro_prgR[where R="{(s,s) | s. rwof m c f s}"])
      apply (auto simp: pw_le_iff refine_pw_simps) []
      apply (cases m, simp, rule rwof.init, simp_all) []
    
      apply (rule WHILEI_refine)
      apply (auto simp: pw_le_iff refine_pw_simps pw_rwof_step')
      done
    thus "m \<guillemotright>= WHILEI (rwof m c f) c f \<le> m \<guillemotright>= WHILE c f"
      by (simp add: pw_le_iff refine_pw_simps)
  next
    show "m \<guillemotright>= WHILE c f \<le> m \<guillemotright>= WHILEI (rwof m c f) c f"
      unfolding WHILE_def 
      apply (rule bind_mono)
      apply (rule order_refl)
      apply (rule WHILEI_weaken)
      ..
  qed

  lemma WHILET_eq_I_rwof: "m \<guillemotright>= WHILET c f = m \<guillemotright>= WHILEIT (rwof m c f) c f"
  proof (rule antisym)
    have "m \<guillemotright>= WHILEIT (rwof m c f) c f
      \<le> \<Down>{(s,s) | s. rwof m c f s} 
        (m \<guillemotright>= WHILET c f)"
      unfolding WHILET_def
      apply (rule bind_refine)
      apply (rule intro_prgR[where R="{(s,s) | s. rwof m c f s}"])
      apply (auto simp: pw_le_iff refine_pw_simps) []
      apply (cases m, simp, rule rwof.init, simp_all) []
    
      apply (rule WHILEIT_refine)
      apply (auto simp: pw_le_iff refine_pw_simps pw_rwof_step')
      done
    thus "m \<guillemotright>= WHILEIT (rwof m c f) c f \<le> m \<guillemotright>= WHILET c f"
      by (simp add: pw_le_iff refine_pw_simps)
  next
    show "m \<guillemotright>= WHILET c f \<le> m \<guillemotright>= WHILEIT (rwof m c f) c f"
      unfolding WHILET_def 
      apply (rule bind_mono)
      apply (rule order_refl)
      apply (rule WHILEIT_weaken)
      ..
  qed


subsection {* Refinement *}
lemma rwof_refine:
  assumes RW: "rwof m c f s"
  assumes NF: "nofail (m' \<guillemotright>= WHILE c' f')"
  assumes M: "m \<le>\<^sub>n \<Down>R m'"
  assumes C: "\<And>s s'. \<lbrakk>(s,s')\<in>R; rwof m c f s; rwof m' c' f' s'\<rbrakk> \<Longrightarrow> c s = c' s'"
  assumes S: "\<And>s s'. \<lbrakk>(s,s')\<in>R; rwof m c f s; rwof m' c' f' s'; c s; c' s'\<rbrakk> \<Longrightarrow> f s \<le>\<^sub>n \<Down>R (f' s')"
  shows "\<exists>s'. (s,s')\<in>R \<and> rwof m' c' f' s'"
  using RW
  apply (induction rule: establish_rwof_invar[rotated -1,consumes 1])
  using M rwof_init[OF NF] 
  apply (simp only: pw_leof_iff pw_le_iff refine_pw_simps, blast) []

  using C S rwof_step[OF NF] 
  apply (simp only: pw_leof_iff pw_le_iff refine_pw_simps, blast) []
  done


end

