theory WhileT_Wellfounded
imports "../../Modelchecker/CAVA_Automata/Digraph" WHILE_aux
begin
(* TODO: Theory dependency problem: This theory depends 
  on some basic graph concepts (infinite word, path, ipath), which 
  are currently defined in the CAVA automata library. 
  However, conceptually, this theory belongs to the refinement framework, which
  is a prerequisite of CAVA_Automata. *)

text {* In this theory, we show that every non-failing total-correct 
  while loop gives rise to a compatible well-founded relation *}


definition rwof_rel 
  -- "Transitions in a while-loop as relation"
  where "rwof_rel init cond step 
    \<equiv> {(s,s'). rwof init cond step s \<and> cond s \<and> inres (step s) s'}"

lemma rwof_rel_spec: "\<lbrakk>rwof init cond step s; cond s\<rbrakk> 
  \<Longrightarrow> step s \<le>\<^sub>n SPEC (\<lambda>s'. (s,s')\<in>rwof_rel init cond step)"
  unfolding rwof_rel_def
  by (auto simp: pw_leof_iff refine_pw_simps)  


lemma rwof_reachable:
  assumes "rwof init cond step s"
  shows "\<exists>s0. inres init s0 \<and> (s0,s)\<in>(rwof_rel init cond step)\<^sup>*"
  using assms
  apply (induction)
  unfolding rwof_rel_def
  apply (auto intro: rwof.intros) []
  apply clarsimp
  apply (intro exI conjI, assumption)
  apply (rule rtrancl_into_rtrancl, assumption)
  apply (auto intro: rwof.intros) []
  done
  
theorem nofail_WHILEIT_wf_rel: 
  assumes NF: "nofail (init \<guillemotright>= WHILEIT I cond step)"
  shows "wf ((rwof_rel init cond step)\<inverse>)"
proof (rule ccontr)
  assume "\<not>wf ((rwof_rel init cond step)\<inverse>)"
  then obtain f where IP: "ipath (rwof_rel init cond step) f"
    by (auto simp: wf_iff_no_infinite_down_chain ipath_def)
  hence "rwof init cond step (f 0)" by (auto simp: rwof_rel_def ipath_def) 
  then obtain s0 w where 
    "path (rwof_rel init cond step) s0 w (f 0)" 
    "inres init s0"
    apply -
    apply (drule rwof_reachable)
    apply (auto elim: rtrancl_is_path)
    done
  with IP have "ipath (rwof_rel init cond step) (w\<frown>f)" "inres init ((w\<frown>f) 0)"
    apply -
    apply (auto simp: ipath_conc_conv) []
    apply (cases w, auto simp: path_simps conc_def) []
    done
  then obtain f where 
    P: "ipath (rwof_rel init cond step) f" 
    and I: "inres init (f 0)" ..

  from P have [simp]: "\<And>i. cond (f i)"
    unfolding ipath_def rwof_rel_def by auto

  from P have SIR: "\<And>i. inres (step (f i)) (f (Suc i))"
    unfolding ipath_def rwof_rel_def by auto

  def F \<equiv> "(WHILEI_body op \<guillemotright>= RETURN I cond step)"

  {
    assume M: "trimono F"
    def f' \<equiv> "\<lambda>x. if x\<in>range f then FAIL else gfp F x"

    have "f' \<le> F f'"
      unfolding f'_def
      apply (rule le_funI)
      apply (case_tac "x\<in>range f")
      apply simp_all
      defer
      apply (subst gfp_unfold)
      using M apply (simp add: trimono_def)
      apply (unfold F_def WHILEI_body_def) []
      apply (auto simp: pw_le_iff refine_pw_simps) []      

      apply (unfold F_def WHILEI_body_def not_nofail_iff[symmetric]) []
      using SIR
      apply (auto simp: pw_le_iff refine_pw_simps) []      
      done  
    from gfp_upperbound[of f' F,OF this] have "\<And>x. f' x \<le> gfp F x"
      by (auto dest: le_funD)
    from this[of "f 0"] have "gfp F (f 0) = FAIL"
      unfolding f'_def
      by auto

  } note aux=this

  have "FAIL \<le> WHILEIT I cond step (f 0)" 
    unfolding WHILET_def WHILEIT_def RECT_def
    apply clarsimp
    apply (subst gfp_eq_flatf_gfp[symmetric])
    apply (simp_all add: trimono_def) [2]
    apply (unfold F_def[symmetric])
    by (rule aux)
  with NF I show False by (auto simp: refine_pw_simps)
qed


end

