theory Paper_Most_Specific_Invar
imports WHILE_aux
begin


lemma rtrancl_sim:
   assumes "\<And>x y a. \<lbrakk>(x, y) \<in> r; (x,a)\<in>S\<rbrakk> \<Longrightarrow> \<exists>b. (y,b)\<in>S \<and> (a, b) \<in> s"
     and "(x, y) \<in> r\<^sup>*" and "(x,a)\<in>S"
   shows "\<exists>b. (y,b)\<in>S \<and> (a, b) \<in> s\<^sup>*"
   using assms(2, 1,3)
   by (induct) (fastforce intro: rtrancl.rtrancl_into_rtrancl)+

lemma rtrancl_map:
   assumes "\<And>x y. (x, y) \<in> r \<Longrightarrow> (f x, f y) \<in> s"
     and "(x, y) \<in> r\<^sup>*"
   shows "(f x, f y) \<in> s\<^sup>*"
   using rtrancl_sim[where S="{(x,f x) | x. True}"] assms 
   by simp blast

(* Preparation of explanation in Paper *)

(* General setting: Initialization and loop *)
definition "loop init cond step \<equiv> init \<guillemotright>= WHILE cond step"

(* We cannot afford to work under the assumption that the hook-functions do 
  not fail, as we want to use invariants of the basic algorithm to show that
  the hook functions do not fail. Thus, we regard invariants only for the 
  non-failing results. *)

(* Invariant of the algorithm, excluding failing results. *)
definition "is_invar init cond step P 
  \<equiv> init \<le>\<^sub>n SPEC P \<and> (\<forall>s. cond s \<and> P s \<longrightarrow> step s \<le>\<^sub>n SPEC P)"

(* The most specific invariant is the strictest predicate that is an invariant. *)
definition "msi init cond step \<equiv> Inf (Collect (is_invar init cond step))"

lemma msi_invar: "is_invar init cond step (msi init cond step)"
  unfolding is_invar_def[abs_def] msi_def
  by (simp add: pw_leof_iff) metis

lemma msi_least: "is_invar init cond step P \<Longrightarrow> msi init cond step \<le> P"
  unfolding is_invar_def[abs_def] msi_def
  apply (simp add: pw_leof_iff)
  apply (rule Inf_lower)
  apply auto
  done

lemma msi_least_alt: "\<lbrakk>is_invar init cond step P; msi init cond step s\<rbrakk> \<Longrightarrow> P s"
  using msi_least by blast

(* It conincides with the definition that we are actually using in the framework *)
lemma msi_is_rwof: "msi init cond step = rwof init cond step"
  apply (intro antisym)
  apply (rule msi_least)
  unfolding is_invar_def
  apply (metis rwof_leof_init rwof_leof_step)
  by (smt2 establish_rwof_invar is_invar_def le_boolI' le_funI msi_invar)

(*  
  Proving properties of the algorithm is done by three rules.

  First, we establish invariants of the algorithm as consequences of the
  most specific invariant. This is done by showing the invariance of the
  property and the most specific invariant. In particular, this allows to
  use already established invariants when establishing new invariants.
  
  Moreover, new invariants may also be shown by just combining already
  established invariants, without doing an induction proof at all.
*)  
theorem establish_invar: 
  assumes "init \<le>\<^sub>n SPEC P"
  assumes "\<And>s. \<lbrakk> cond s; msi init cond step s; P s \<rbrakk> \<Longrightarrow> step s \<le>\<^sub>n SPEC P"
  shows "msi init cond step \<le> P"
proof -
  have "is_invar init cond step (\<lambda>s. msi init cond step s \<and> P s)"
    using msi_invar unfolding is_invar_def using assms
    apply safe
    apply (blast intro: SPEC_rule_leof_conjI)
    apply (rule SPEC_rule_leof_conjI)
    apply simp apply simp
    done

  with msi_least show ?thesis by clarsimp blast
qed

(* Second, we use the fact that, if the loop terminates and does not fail,
  the result satisfies the most specific invariant and the loop condition
  does not hold. Using the established invariants, we can derive the desired 
  correctness properties from this rule.
*)
theorem loop_correct: 
  "loop init cond step \<le>\<^sub>n SPEC (\<lambda>s. msi init cond step s \<and> \<not>cond s)"
  unfolding loop_def msi_is_rwof
  by (metis (no_types, lifting) SPEC_rule leofI nofail_WHILE_eq_rwof)


(*
  Third, we prove that the loop does not fail, by proving that no
  step from a state satisfying the most specific invariant does fail.
  Again, we can re-use already established invariants.
*)  
theorem prove_nofail: 
  assumes I: "nofail init" 
  assumes S: "\<And>s. \<lbrakk> msi init cond step s; cond s \<rbrakk> \<Longrightarrow> nofail (step s)"
  shows "nofail (loop init cond step)"
  using assms
  unfolding msi_is_rwof nofail_SPEC_iff loop_def WHILE_eq_I_rwof
  apply (cases init, simp_all)

  apply (refine_vcg)
  apply (rule rwof.init, simp_all)
  by (metis S msi_is_rwof rwof_step')

theorem loop_correct':
  assumes "nofail (loop init cond step)"
  shows "loop init cond step \<le> SPEC (\<lambda>s. msi init cond step s \<and> \<not>cond s)"
  apply (rule leofD, fact)
  by (rule loop_correct) 

end

