theory Param_DFS
imports 
  "../Libs/Collections/Refine_Dflt"
  "../Modelchecker/CAVA_Automata/Digraph"
  "Misc/WHILE_aux"
  "Misc/DFS_Framework_Graph_Aux"
begin

(* 
  Peter-aus-dem-Urlaub-Besprechungs-TODO:
    - weaken_SPEC vs use_spec_rule
    - skel_rwof vs rwof vs DFS_invar
      \<longrightarrow> insbesondere bei den param_DFS-Obligationen
     (\<Longrightarrow> nur DFS_invar )
    - I0 zu it_dfs_correct und rwof_WHILE_rule hinzugefügt (ok?)
      \<Longrightarrow> ok
    - rwof s' zu den param_DFS-Obligationen hinzgefügt (ok?)
      \<Longrightarrow> ok
    - abbreviation für 'state.more'?
      \<Longrightarrow> records richtig verwenden
    - was war nochmal der Sinn des rekursiven DFS?
      \<Longrightarrow> gute Frage... (tendenziell bessere Performance)
    - wie mit eingeschränktem Graphen umgehen (zB bei RedDFS)
      Möglichkeiten:
       * rel_restrict E R          (\<Rightarrow> momentan)
          \<longrightarrow> wie (gut) implementieren?
       * R als Parameter der Locale
          \<longrightarrow> lässt sich das wegoptimieren falls R = {}?
          \<longrightarrow> werden die Beweise komplizierter?
       * `succ` statt E als Parameter
          \<longrightarrow> lässt sich einfach einschränken
          \<longrightarrow> liefert Implementierung gleich mit
       * Trick mit discovered initialisieren (wie aus NDFS)
          \<longrightarrow> geht nicht, weil dann die Obligationen nicht mehr stimmen
       (\<Longrightarrow> erstmal rel_restrict, schauen ob Implementierung auf it. Ansatz
           durchgeht)
    - V0 als Menge? 
       \<longrightarrow> was muss geändert werden und wie sieht die Spez. aus?
       \<longrightarrow> nicht einfach mit FOREACH implementierbar (state handling geht kaputt)
       \<longrightarrow> multiple DFS (dfs auf jeden v0\<in>V0 hintereinander 
                         mit korrektem state transfer)
       (\<Longrightarrow> step erweitern )
    - liften auf andere Graph-Darstellung
       \<longrightarrow> notwendig? Wir benutzen keine Grapheigenschaften
                     (nur `path`, was unabh. ist)
       \<Longrightarrow> auf 'v rel belassen
    - BlueDFS mit Lasso-Theorie verheiraten
      (sollte trivial sein, aber das sind mir zu viele verschiedene
       Definitionen in Lasso.thy...)
*)

(* A DFS with timing and a stack, phrased iteratively *)
record 'v state =
  counter :: nat
  discovered :: "'v \<rightharpoonup> nat"
  finished :: "'v \<rightharpoonup> nat"
  pending :: "('v \<times> 'v) set"
  stack :: "'v list"
  tree :: "'v rel" (* TODO: tree_edges *)
  back_edges :: "'v rel"
  cross_edges :: "'v rel"

abbreviation "NOOP s \<equiv> RETURN (state.more s)"


record ('v,'s,'es) gen_parameterization =
  on_init :: "'es nres"
  on_new_root :: "'v \<Rightarrow> 's \<Rightarrow> 'es nres"
  on_discover :: "'v \<Rightarrow> 'v \<Rightarrow> 's \<Rightarrow> 'es nres"
  on_finish :: "'v \<Rightarrow> 's \<Rightarrow> 'es nres"
  on_back_edge :: "'v \<Rightarrow> 'v \<Rightarrow> 's \<Rightarrow> 'es nres"
  on_cross_edge :: "'v \<Rightarrow> 'v \<Rightarrow> 's \<Rightarrow> 'es nres"
  is_break :: "'s \<Rightarrow> bool"

type_synonym ('v,'es) parameterization 
  = "('v,('v,'es) state_scheme,'es) gen_parameterization"

definition "\<And>more init. dflt_parametrization more init \<equiv> \<lparr>  
  on_init = init,
  on_new_root = \<lambda>_. RETURN o more,
  on_discover = \<lambda>_ _. RETURN o more,
  on_finish = \<lambda>_. RETURN o more,
  on_back_edge = \<lambda>_ _. RETURN o more,
  on_cross_edge = \<lambda>_ _. RETURN o more,
  is_break = \<lambda>_. False \<rparr>"
lemmas dflt_parametrization_simp[simp] =
  gen_parameterization.simps[mk_record_simp, OF dflt_parametrization_def]


locale param_DFS_defs =
  graph_defs G
  for G :: "('v, 'more) graph_rec_scheme"
  +
  fixes param :: "('v,'es) parameterization"
begin

  definition is_discovered :: "'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> bool"
    where "is_discovered u s \<equiv> u \<in> dom (discovered s)"

  definition is_finished :: "'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> bool"
    where "is_finished u s \<equiv> u \<in> dom (finished s)"

  definition is_empty_stack :: "('v,'es) state_scheme \<Rightarrow> bool"
    where "is_empty_stack s \<equiv> stack s = []"

  (*definition top_pending :: "('v,'es) state_scheme \<Rightarrow> 'v \<times> 'v set" where
    "top_pending s \<equiv> (hd (stack s), pending s `` {hd (stack s)})"*)

  definition discover 
    :: "'v \<Rightarrow> 'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme"
  where
  "discover u v s \<equiv> let
    d = (discovered s)(v \<mapsto> counter s); c = counter s + 1;
    st = v#stack s;
    p = pending s \<union> {v} \<times> E``{v};
    t = insert (u,v) (tree s)
  in s\<lparr> discovered := d, counter := c, stack := st, pending := p, tree := t\<rparr>"

  lemma discover_simps[simp]:
    "counter (discover u v s) = Suc (counter s)"
    "discovered (discover u v s) = (discovered s)(v \<mapsto> counter s)" 
    "finished (discover u v s) = finished s"
    "stack (discover u v s) = v#stack s"
    "pending (discover u v s) = pending s \<union> {v} \<times> E``{v}"
    "tree (discover u v s) = insert (u,v) (tree s)"
    "cross_edges (discover u v s) = cross_edges s"
    "back_edges (discover u v s) = back_edges s"
    "state.more (discover u v s) = state.more s"
    by (simp_all add: discover_def)

  definition finish 
    :: "'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme" 
  where
  "finish u s \<equiv> let
    f = (finished s)(u \<mapsto> counter s); c = counter s + 1;
    st = tl (stack s)
  in s\<lparr> finished := f, counter := c, stack := st\<rparr>"

  lemma finish_simps[simp]:
    "counter (finish u s) = Suc (counter s)"
    "discovered (finish u s) = discovered s"
    "finished (finish u s) = (finished s)(u \<mapsto> counter s)"
    "stack (finish u s) = tl (stack s)"
    "pending (finish u s) = pending s"
    "tree (finish u s) = tree s"
    "cross_edges (finish u s) = cross_edges s"
    "back_edges (finish u s) = back_edges s"
    "state.more (finish u s) = state.more s"
    by (simp_all add: finish_def)

  definition back_edge
    :: "'v \<Rightarrow> 'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme"
  where
  "back_edge u v s \<equiv> let
    b = insert (u,v) (back_edges s)
   in s\<lparr> back_edges := b \<rparr>"

  lemma back_edge_simps[simp]:
    "counter (back_edge u v s) = counter s"
    "discovered (back_edge u v s) = discovered s"
    "finished (back_edge u v s) = finished s"
    "stack (back_edge u v s) = stack s"
    "pending (back_edge u v s) = pending s"
    "tree (back_edge u v s) = tree s"
    "cross_edges (back_edge u v s) = cross_edges s"
    "back_edges (back_edge u v s) = insert (u,v) (back_edges s)"
    "state.more (back_edge u v s) = state.more s"
    by (simp_all add: back_edge_def)

  definition cross_edge
    :: "'v \<Rightarrow> 'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme"
  where
  "cross_edge u v s \<equiv> let
    c = insert (u,v) (cross_edges s)
   in s\<lparr> cross_edges := c \<rparr>"
  
  lemma cross_edge_simps[simp]:
    "counter (cross_edge u v s) = counter s"
    "discovered (cross_edge u v s) = discovered s"
    "finished (cross_edge u v s) = finished s"
    "stack (cross_edge u v s) = stack s"
    "pending (cross_edge u v s) = pending s"
    "tree (cross_edge u v s) = tree s"
    "cross_edges (cross_edge u v s) = insert (u,v) (cross_edges s)"
    "back_edges (cross_edge u v s) = back_edges s"
    "state.more (cross_edge u v s) = state.more s"
    by (simp_all add: cross_edge_def)
  
  definition do_cross_edge
    :: "'v \<Rightarrow> 'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme nres"
  where
  "do_cross_edge u v s \<equiv> do {
      let s = cross_edge u v s;
      e \<leftarrow> on_cross_edge param u v s;
      RETURN (s\<lparr>state.more := e\<rparr>)
    }"

  definition do_back_edge
    :: "'v \<Rightarrow> 'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme nres"
  where
  "do_back_edge u v s \<equiv> do {
      let s = back_edge u v s;
      e \<leftarrow> on_back_edge param u v s;
      RETURN (s\<lparr>state.more := e\<rparr>)
    }"

  definition do_known_edge
    :: "'v \<Rightarrow> 'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme nres"
  where
  "do_known_edge u v s \<equiv> 
    if is_finished v s then 
      do_cross_edge u v s 
    else 
      do_back_edge u v s"  

  definition do_discover
    :: "'v \<Rightarrow> 'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme nres"
  where
  "do_discover u v s \<equiv> do {
    let s = discover u v s;
    e \<leftarrow> on_discover param u v s;
    RETURN (s\<lparr>state.more := e\<rparr>)
  }"

  definition do_finish
    :: "'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme nres"
  where
  "do_finish u s \<equiv> do {
    let s = finish u s;
    e \<leftarrow> on_finish param u s;
    RETURN (s\<lparr>state.more := e\<rparr>)
  }"

  definition new_root
    :: "'v \<Rightarrow> ('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme"
  where
    "new_root v0 s \<equiv> let
       c = Suc (counter s);
       d = (discovered s)(v0 \<mapsto> counter s);
       p = {v0}\<times>E``{v0};
       st = [v0]
     in s\<lparr>counter := c, discovered := d, pending := p, stack := st\<rparr>"

  lemma new_root_simps[simp]:
    "counter (new_root v0 s) = Suc (counter s)"
    "discovered (new_root v0 s) = (discovered s)(v0 \<mapsto> counter s)"
    "finished (new_root v0 s) = finished s"
    "stack (new_root v0 s) = [v0]"
    "pending (new_root v0 s) = ({v0}\<times>E``{v0})"
    "tree (new_root v0 s) = tree s"
    "cross_edges (new_root v0 s) = cross_edges s"
    "back_edges (new_root v0 s) = back_edges s"
    "state.more (new_root v0 s) = state.more s"
    by (simp_all add: new_root_def)

  definition get_new_root where
    "get_new_root s \<equiv> SPEC (\<lambda>v. v\<in>V0 \<and> \<not>is_discovered v s)"  

  definition do_new_root where 
  "do_new_root v0 s \<equiv> do {
    let s = new_root v0 s;
    e \<leftarrow> on_new_root param v0 s;
    RETURN (s\<lparr>state.more := e\<rparr>)
  }"

  lemmas op_defs = discover_def finish_def back_edge_def cross_edge_def new_root_def
  lemmas do_defs = do_discover_def do_finish_def do_known_edge_def
    do_cross_edge_def do_back_edge_def do_new_root_def
  lemmas pred_defs = is_discovered_def is_finished_def is_empty_stack_def

  definition "empty_state e
    \<equiv> \<lparr> counter = 0,
         discovered = Map.empty,
         finished = Map.empty,
         pending = {},
         stack = [],
         tree = {},
         back_edges = {},
         cross_edges = {},
         \<dots> = e \<rparr>"

  lemma empty_state_simps[simp]:
    "counter (empty_state e) = 0"
    "discovered (empty_state e) = Map.empty"
    "finished (empty_state e) = Map.empty"
    "pending (empty_state e) = {}"
    "stack (empty_state e) = []"
    "tree (empty_state e) = {}"
    "back_edges (empty_state e) = {}"
    "cross_edges (empty_state e) = {}"
    "state.more (empty_state e) = e"
    by (simp_all add: empty_state_def)

  definition "init \<equiv> do {
    e \<leftarrow> on_init param;
    RETURN (empty_state e)
  }"

  definition cond :: "('v,'es) state_scheme \<Rightarrow> bool" where
    "cond s \<longleftrightarrow> (V0 \<subseteq> {v. is_discovered v s} \<longrightarrow> \<not>is_empty_stack s) 
      \<and> \<not>is_break param s"  

  lemma cond_alt:
    "cond = (\<lambda>s. (V0 \<subseteq> dom (discovered s) \<longrightarrow> stack s \<noteq> []) \<and> \<not>is_break param s)"
    apply (rule ext)
    unfolding cond_def is_discovered_def is_empty_stack_def
    by auto


  definition get_pending ::
    "('v, 'es) state_scheme \<Rightarrow> ('v \<times> 'v option \<times> ('v, 'es) state_scheme) nres" 
    -- {* Get topmost stack node and a pending edge if any. The pending
          edge is removed. *}
  where "get_pending s \<equiv> do {
    let u = hd (stack s);
    let Vs = pending s `` {u};

    if Vs = {} then
      RETURN (u,None,s)
    else do {
      v \<leftarrow> RES Vs;
      let s = s\<lparr> pending := pending s - {(u,v)}\<rparr>;
      RETURN (u,Some v,s)
    }
  }"

  definition step :: "('v,'es) state_scheme \<Rightarrow> ('v,'es) state_scheme nres" 
  where
    "step s \<equiv> 
      if is_empty_stack s then do {
        v0 \<leftarrow> get_new_root s;
        do_new_root v0 s
      } else do {
        (u,Vs,s) \<leftarrow> get_pending s;
        case Vs of 
          None \<Rightarrow> do_finish u s 
        | Some v \<Rightarrow> do {
          if is_discovered v s then 
            do_known_edge u v s
          else 
            do_discover u v s
        }
      }"


  definition "it_dfs \<equiv> init \<guillemotright>= WHILE cond step"
  definition "it_dfsT \<equiv> init \<guillemotright>= WHILET cond step"

  (*definition "is_invar I \<equiv> is_rwof_invar init cond step I"*)
  (*definition "skel_rwof \<equiv> rwof init cond step"*)

end

lemma "WHILEI I b f s \<le> 
  do {ASSERT (I s); WHILE b (\<lambda>s. do {s \<leftarrow> f s; ASSERT (I s); RETURN s}) s}"
  unfolding WHILEI_def WHILE_def WHILEI_body_def
  apply (rule refine_IdD)
  apply refine_rcg
  apply (rule introR[where R="br id I"])
  apply (simp_all add: br_def)
  apply (rule intro_prgR[where R="br id I"])
  apply (simp_all add: br_def)
  apply (auto simp: pw_le_iff refine_pw_simps)  
  done

(* TODO: Modify proof in RefineG_Recursion. Remove trimono assumption! *)
lemma RECT_eq_REC: 
  -- "Partial and total correct recursion are equal if total 
    recursion does not fail."
  assumes NT: "RECT body x \<noteq> top"
  shows "RECT body x = REC body x"
  using NT
  unfolding RECT_def REC_def
proof clarsimp
  assume M: "trimono body"
  from lfp_unfold[OF trimonoD_mono[OF M], symmetric]
  have "flatf_ge (body (lfp body)) (lfp body)" by simp
  note flatf_ord.fixp_lowerbound[
    OF trimonoD_flatf_ge[OF M], of "lfp body", OF this]
  moreover assume "flatf_gfp body x \<noteq> top"
  ultimately show "flatf_gfp body x = lfp body x"
    by (auto simp add: fun_ord_def flat_ord_def)
qed

(* TODO: Move to RefineG_While *)
lemma WHILET_eq_WHILE:
  assumes "WHILET b f s0 \<noteq> top"
  shows "WHILET b f s0 = WHILE b f s0"
  using assms
  unfolding WHILET_def WHILE_def WHILEIT_def WHILEI_def
  by (rule RECT_eq_REC)

lemma WHILEIT_eq_WHILEI:
  assumes "WHILEIT I b f s0 \<noteq> top"
  shows "WHILEIT I b f s0 = WHILEI I b f s0"
  using assms
  unfolding WHILEIT_def WHILEI_def
  by (rule RECT_eq_REC)


lemma WHILET_eq_WHILE_tproof:
  assumes "wf V"
  assumes "I s0"
  assumes "\<And>s. \<lbrakk> I s; b s \<rbrakk> \<Longrightarrow> f s \<le> SPEC (\<lambda>s'. I s' \<and> (s',s)\<in>V)"
  shows "WHILET b f s0 = WHILE b f s0"
proof -
  have "WHILET b f s0 \<le> SPEC I"
    by (rule WHILET_rule[where I=I], fact+)
  hence "WHILET b f s0 \<noteq> top" by auto 

  thus ?thesis
    unfolding WHILE_def WHILEI_def WHILET_def WHILEIT_def
    by (subst RECT_eq_REC) auto
qed  

lemma WHILEIT_eq_WHILEI_tproof:
  assumes "wf V"
  (*assumes "I s0"*)
  assumes "\<And>s. \<lbrakk> I s; b s \<rbrakk> \<Longrightarrow> f s \<le> SPEC (\<lambda>s'. I s' \<and> (s',s)\<in>V)"
  (* TODO: The I s' is not necessary here, but right now seems complicated to remove :( *)
  shows "WHILEIT I b f s0 = WHILEI I b f s0"
proof (cases "I s0")
  case True[simp]

  have "WHILEIT I b f s0 \<le> SPEC I"
    by (rule WHILEIT_rule[where I=I], fact+)
  hence "WHILEIT I b f s0 \<noteq> top" by auto 

  thus ?thesis
    unfolding WHILE_def WHILEI_def WHILET_def WHILEIT_def
    by (subst RECT_eq_REC) auto
next    
  case False[simp]
  have "WHILEIT I b f s0 = FAIL"
    by (subst WHILEIT_unfold) simp
  also have "FAIL = WHILEI I b f s0"  
    by (subst WHILEI_unfold) simp
  finally show ?thesis .
qed  



(* This locale is required to establish new invariants.
  We would like to directly establish new invariants in the 
  DFS_invar-locale, unfortunately this causes technical problems:
  When interpreting the DFS_invar locale in a proof inside the 
  DFS_invar-locale itself, we get "duplicate constant" warnings,
  unless we prefix the interpreted locale, which may be quite confusing
  in a proof, as the user has to choose the prefixed lemmas, while the
  unprefixed ones are also available, but for the wrong state.
 *)
locale param_DFS =
  fb_graph G + param_DFS_defs G param
  for G :: "('v, 'more) graph_rec_scheme"
  and param :: "('v,'es) parameterization"
begin

  definition "is_invar I \<equiv> is_rwof_invar init cond step I"

end

locale DFS_invar =
  param_DFS G param
  for G :: "('v, 'more) graph_rec_scheme"
  and param :: "('v,'es) parameterization"
  +
  fixes s :: "('v,'es) state_scheme"
  assumes rwof: "rwof init cond step s"
begin

  lemmas make_invar_thm = rwof_cons[OF _ rwof, folded is_invar_def]

end

(*context param_DFS_defs begin
  lemma rwof_eq_DFS_invar[simp]: 
    "rwof init cond step = DFS_invar G param"
    apply (auto intro: DFS_invar.rwof intro!: ext)
    by unfold_locales

  lemmas establish_invar[intro?] = 
    is_rwof_invarI[of init _ cond step,
      unfolded rwof_eq_DFS_invar, folded is_invar_def]

  lemmas use_invar = rwof_cons[of init cond step,
      unfolded rwof_eq_DFS_invar, folded is_invar_def]

end*)

context param_DFS
begin
  (* Include this into refine-rules to discard any information about 
    parameterization *)
  lemmas indep_invar_rules = 
    leof_True_rule[where m="on_init param"]
    leof_True_rule[where m="on_new_root param v0 s'" for v0 s']
    leof_True_rule[where m="on_discover param u v s'" for u v s']
    leof_True_rule[where m="on_finish param v s'" for v s']
    leof_True_rule[where m="on_cross_edge param u v s'" for u v s']
    leof_True_rule[where m="on_back_edge param u v s'" for u v s']

  (* The DFS-invar locale is equivalent to the strongest invariant of the loop *)
  lemma rwof_eq_DFS_invar[simp]: 
    "rwof init cond step = DFS_invar G param"
    apply (auto intro: DFS_invar.rwof intro!: ext)
    by unfold_locales

  (* A step preserves the (best) invariant *)
  lemma DFS_invar_step: "\<lbrakk>nofail it_dfs; DFS_invar G param s; cond s\<rbrakk> 
    \<Longrightarrow> step s \<le> SPEC (DFS_invar G param)"
    unfolding it_dfs_def rwof_eq_DFS_invar[symmetric]
    by (rule rwof_step)

  lemma DFS_invar_step': "\<lbrakk>nofail (step s); DFS_invar G param s; cond s\<rbrakk> 
    \<Longrightarrow> step s \<le> SPEC (DFS_invar G param)"
    unfolding it_dfs_def rwof_eq_DFS_invar[symmetric]
    by (rule rwof_step')

(*  lemma step_cases_auxE: 
    assumes "nofail (RES X \<guillemotright>= f)" "inres (RES X \<guillemotright>= f) y"
    obtains x where "x\<in>X" "nofail (f x)" "inres (f x) y"
    using pw_RES_bind_choose assms
    by metis
*)

  definition "pre_is_break s \<equiv> DFS_invar G param s"

  definition "pre_on_new_root v0 s' \<equiv> \<exists>s.
    DFS_invar G param s \<and> cond s \<and> 
    stack s = [] \<and> v0 \<in> V0 \<and> v0 \<notin> dom (discovered s) \<and>
    s' = new_root v0 s"

  definition "pre_on_finish u s' \<equiv> \<exists>s.
    DFS_invar G param s \<and> cond s \<and> 
    stack s \<noteq> [] \<and> u = hd (stack s) \<and> pending s `` {u} = {} \<and> s' = finish u s"

  definition "pre_edge_selected u v s \<equiv> 
    DFS_invar G param s \<and> cond s \<and> 
    stack s \<noteq> [] \<and> u = hd (stack s) \<and> (u, v) \<in> pending s"

  definition "pre_on_cross_edge u v s' \<equiv> \<exists>s. pre_edge_selected u v s \<and>
        v \<in> dom (discovered s) \<and> v\<in>dom (finished s) 
        \<and> s' = cross_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)"

  definition "pre_on_back_edge u v s' \<equiv> \<exists>s. pre_edge_selected u v s \<and>
        v \<in> dom (discovered s) \<and> v\<notin>dom (finished s) 
        \<and> s' = back_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)"

  definition "pre_on_discover u v s' \<equiv> \<exists>s. pre_edge_selected u v s \<and>
        v \<notin> dom (discovered s) 
        \<and> s' = discover u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)"

  lemmas pre_on_defs = pre_on_new_root_def pre_on_finish_def 
    pre_edge_selected_def pre_on_cross_edge_def pre_on_back_edge_def
    pre_on_discover_def pre_is_break_def

  lemma establish_invarI[case_names init new_root finish cross_edge back_edge discover]:
    assumes init: "on_init param \<le>\<^sub>n SPEC (\<lambda>x. I (empty_state x))"
    assumes new_root: "\<And>s s' v0. 
      \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
       stack s = []; v0 \<in> V0; v0 \<notin> dom (discovered s);
        s' = new_root v0 s\<rbrakk>
         \<Longrightarrow> on_new_root param v0 s' \<le>\<^sub>n 
             SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                         \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes finish: "\<And>s s' u. 
      \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
       stack s \<noteq> []; u = hd (stack s); 
       pending s `` {u} = {};
       s' = finish u s\<rbrakk>
         \<Longrightarrow> on_finish param u s' \<le>\<^sub>n 
              SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                        \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes cross_edge: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> []; (u, v) \<in> pending s; u = hd (stack s); 
        v \<in> dom (discovered s); v\<in>dom (finished s);
        s' = cross_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)\<rbrakk>
       \<Longrightarrow> on_cross_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes back_edge: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> []; (u, v) \<in> pending s; u = hd (stack s); 
        v \<in> dom (discovered s); v\<notin>dom (finished s);
        s' = back_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)\<rbrakk>
       \<Longrightarrow> on_back_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>) 
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes discover: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> [];  (u, v) \<in> pending s; u = hd (stack s); 
        v \<notin> dom (discovered s);
        s' = discover u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)\<rbrakk>
       \<Longrightarrow> on_discover param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    shows "is_invar I"
    unfolding is_invar_def
  proof
    show "init \<le>\<^sub>n SPEC I"
      unfolding init_def
      by (refine_rcg refine_vcg) (simp add: init)
  next
    fix s
    assume "rwof init cond step s" and IC: "I s" "cond s"
    hence DI: "DFS_invar G param s" by unfold_locales
    then interpret DFS_invar G param s .

    from `cond s` have IB: "\<not> is_break param s" by (simp add: cond_def)

    have B: "step s \<le>\<^sub>n SPEC (DFS_invar G param)"
      by rule (metis DFS_invar_step' DI `cond s`)

    note rule_assms = DI IC IB

    show "step s \<le>\<^sub>n SPEC I"
      apply (rule leof_use_spec_rule[OF B])
      unfolding step_def do_defs pred_defs get_pending_def get_new_root_def
      apply (refine_rcg refine_vcg)
      apply (simp_all)

      apply (blast intro: new_root[OF rule_assms])
      apply (blast intro: finish[OF rule_assms])
      apply (rule cross_edge[OF rule_assms], auto) []
      apply (rule back_edge[OF rule_assms], auto) []
      apply (rule discover[OF rule_assms], auto) []
      done
  qed

  lemma establish_invarI'[case_names init new_root finish cross_edge back_edge discover]:
    assumes init: "on_init param \<le>\<^sub>n SPEC (\<lambda>x. I (empty_state x))"
    assumes new_root: "\<And>s' v0. pre_on_new_root v0 s'
         \<Longrightarrow> on_new_root param v0 s' \<le>\<^sub>n 
             SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                         \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes finish: "\<And>s' u. pre_on_finish u s' 
         \<Longrightarrow> on_finish param u s' \<le>\<^sub>n 
              SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                        \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes cross_edge: "\<And>s' u v. pre_on_cross_edge u v s'
       \<Longrightarrow> on_cross_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes back_edge: "\<And>s' u v. pre_on_back_edge u v s'
       \<Longrightarrow> on_back_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>) 
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes discover: "\<And>s' u v. pre_on_discover u v s'
       \<Longrightarrow> on_discover param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    shows "is_invar I"
    apply (rule establish_invarI)
    using assms
    unfolding pre_on_defs
    apply -
    apply blast
    apply (rprems,blast)+
    done
  
  (* Variant of establish_invarI, where new_root and discover are combined *)
  lemma establish_invarI_ND [case_names prereq init new_discover finish cross_edge back_edge]:
    assumes prereq: "\<And>u v s. on_discover param u v s = on_new_root param v s"
    assumes init: "on_init param \<le>\<^sub>n SPEC (\<lambda>x. I (empty_state x))"
    assumes new_discover: "\<And>s s' v. 
      \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
       v \<notin> dom (discovered s); 
       discovered s' = (discovered s)(v\<mapsto>counter s); finished s' = finished s;
       counter s' = Suc (counter s); stack s' = v#stack s;
       back_edges s' = back_edges s; cross_edges s' = cross_edges s;
       tree s' \<supseteq> tree s;
       state.more s' = state.more s\<rbrakk>
         \<Longrightarrow> on_new_root param v s' \<le>\<^sub>n 
             SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                         \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes finish: "\<And>s s' u. 
      \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
       stack s \<noteq> []; u = hd (stack s); 
       pending s `` {u} = {};
       s' = finish u s\<rbrakk>
         \<Longrightarrow> on_finish param u s' \<le>\<^sub>n 
              SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                        \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes cross_edge: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> []; (u, v) \<in> pending s; u = hd (stack s); 
        v \<in> dom (discovered s); v\<in>dom (finished s);
        s' = cross_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)\<rbrakk>
       \<Longrightarrow> on_cross_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes back_edge: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> []; (u, v) \<in> pending s; u = hd (stack s); 
        v \<in> dom (discovered s); v\<notin>dom (finished s);
        s' = back_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)\<rbrakk>
       \<Longrightarrow> on_back_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>) 
                        \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    shows "is_invar I"
  proof (induct rule: establish_invarI)
    case (new_root s) thus ?case by (auto intro!: new_discover)
  next
    case (discover s s' u v) hence
      "on_new_root param v s' \<le>\<^sub>n 
        SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                   \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
      by (auto intro!: new_discover)
    with prereq show ?case by simp
  qed fact+

  (* Variant of establish_invarI, where cross_edge and back_edge are combined *)
  lemma establish_invarI_CB [case_names prereq init new_root finish cross_back_edge discover]:
    assumes prereq: "\<And>u v s. on_back_edge param u v s = on_cross_edge param u v s"
    assumes init: "on_init param \<le>\<^sub>n SPEC (\<lambda>x. I (empty_state x))"
    assumes new_root: "\<And>s s' v0. 
      \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
       stack s = []; v0 \<in> V0; v0 \<notin> dom (discovered s);
        s' = new_root v0 s\<rbrakk>
         \<Longrightarrow> on_new_root param v0 s' \<le>\<^sub>n 
             SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                         \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes finish: "\<And>s s' u. 
      \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
       stack s \<noteq> []; u = hd (stack s); 
       pending s `` {u} = {};
       s' = finish u s\<rbrakk>
         \<Longrightarrow> on_finish param u s' \<le>\<^sub>n 
              SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                        \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes cross_back_edge: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> []; (u, v) \<in> pending s; u = hd (stack s); 
        v \<in> dom (discovered s);
        discovered s' = discovered s; finished s' = finished s;
        stack s' = stack s; tree s' = tree s; counter s' = counter s;
        pending s' = pending s - {(u,v)};
        cross_edges s' \<union> back_edges s' = cross_edges s \<union> back_edges s \<union> {(u,v)};
        state.more s' = state.more s \<rbrakk>
       \<Longrightarrow> on_cross_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes discover: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> [];  (u, v) \<in> pending s; u = hd (stack s); 
        v \<notin> dom (discovered s);
        s' = discover u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>)\<rbrakk>
       \<Longrightarrow> on_discover param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    shows "is_invar I"
  proof (induct rule: establish_invarI)
    case cross_edge thus ?case by (auto intro!: cross_back_edge)
  next
    case (back_edge s s' u v) hence
      "on_cross_edge param u v s' \<le>\<^sub>n 
             SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                         \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
      by (auto intro!: cross_back_edge)
    with prereq show ?case by simp
  qed fact+

  (* Variant of establish_invarI, where cross_edge and back_edge, and discover and new_root are combined *)
  lemma establish_invarI_ND_CB [case_names prereq_ND prereq_CB init new_discover finish cross_back_edge]:
    assumes prereq: 
        "\<And>u v s. on_discover param u v s = on_new_root param v s"
        "\<And>u v s. on_back_edge param u v s = on_cross_edge param u v s"
    assumes init: "on_init param \<le>\<^sub>n SPEC (\<lambda>x. I (empty_state x))"
    assumes new_discover: "\<And>s s' v. 
     \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
      v \<notin> dom (discovered s); 
      discovered s' = (discovered s)(v\<mapsto>counter s); finished s' = finished s;
      counter s' = Suc (counter s); stack s' = v#stack s;
      back_edges s' = back_edges s; cross_edges s' = cross_edges s;
      tree s' \<supseteq> tree s;
      state.more s' = state.more s\<rbrakk>
        \<Longrightarrow> on_new_root param v s' \<le>\<^sub>n 
            SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                        \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes finish: "\<And>s s' u. 
      \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
       stack s \<noteq> []; u = hd (stack s); 
       pending s `` {u} = {};
       s' = finish u s\<rbrakk>
         \<Longrightarrow> on_finish param u s' \<le>\<^sub>n 
              SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                        \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    assumes cross_back_edge: "\<And>s s' u v.
       \<lbrakk>DFS_invar G param s; I s; cond s; \<not> is_break param s;
        stack s \<noteq> []; (u, v) \<in> pending s; u = hd (stack s); 
        v \<in> dom (discovered s);
        discovered s' = discovered s; finished s' = finished s;
        stack s' = stack s; tree s' = tree s; counter s' = counter s;
        pending s' = pending s - {(u,v)};
        cross_edges s' \<union> back_edges s' = cross_edges s \<union> back_edges s \<union> {(u,v)};
        state.more s' = state.more s \<rbrakk>
       \<Longrightarrow> on_cross_edge param u v s' \<le>\<^sub>n
           SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                      \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
    shows "is_invar I"
  proof (induct rule: establish_invarI_ND)
    case cross_edge thus ?case by (auto intro!: cross_back_edge)
  next
    case (back_edge s s' u v) hence
      "on_cross_edge param u v s' \<le>\<^sub>n 
             SPEC (\<lambda>x. DFS_invar G param (s'\<lparr>state.more := x\<rparr>)
                         \<longrightarrow> I (s'\<lparr>state.more := x\<rparr>))"
      by (auto intro!: cross_back_edge)
    with prereq show ?case by simp
  qed fact+


  (* Variant of establish_invarI w/o the parametrization *)
  lemma is_invarI_full [case_names init new_root finish cross_edge back_edge discover]:
    assumes init: "\<And>e. I (empty_state e)"
    assumes new_root: "\<And>s s' v0 e. 
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s = []; v0 \<notin> dom (discovered s); v0 \<in> V0;
        s' = new_root v0 s\<lparr>state.more := e\<rparr>\<rbrakk> 
      \<Longrightarrow> I s'"
    and finish: "\<And>s s' u e. 
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s \<noteq> []; pending s `` {u} = {};
        u = hd (stack s); s' = finish u s\<lparr>state.more := e\<rparr>\<rbrakk> 
       \<Longrightarrow> I s'"
    and cross_edge: "\<And>s s' u v e.
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s \<noteq> []; v \<in> pending s `` {u}; v \<in> dom (discovered s); 
        v \<in> dom (finished s);
        u = hd (stack s); 
        s' = (cross_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>))\<lparr>state.more := e\<rparr>\<rbrakk>
       \<Longrightarrow> I s'"
    and back_edge: "\<And>s s' u v e.
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s \<noteq> []; v \<in> pending s `` {u}; v \<in> dom (discovered s); v \<notin> dom (finished s); 
        u = hd (stack s); 
        s' = (back_edge u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>))\<lparr>state.more := e\<rparr>\<rbrakk>
       \<Longrightarrow> I s'"
    and discover: "\<And>s s' u v e.
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s \<noteq> []; v \<in> pending s `` {u}; v \<notin> dom (discovered s); 
        u = hd (stack s); 
        s' = (discover u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>))\<lparr>state.more := e\<rparr>\<rbrakk>
       \<Longrightarrow> I s'"
  shows "is_invar I"
  apply (rule establish_invarI) 
  apply (blast intro: indep_invar_rules assms)+
  done

  (* Variant of establish_invarI w/o the parametrization and with cross_ + back_edge combined *)
  lemma is_invarI [case_names init new_root finish visited discover]:
    assumes init': "\<And>e. I (empty_state e)"
    and new_root': "\<And>s s' v0 e. 
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s = []; v0 \<notin> dom (discovered s); v0 \<in> V0;
        s' = new_root v0 s\<lparr>state.more := e\<rparr>\<rbrakk> 
      \<Longrightarrow> I s'"
    and finish': "\<And>s s' u e. 
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s \<noteq> []; pending s `` {u} = {};
        u = hd (stack s); s' = finish u s\<lparr>state.more := e\<rparr>\<rbrakk> 
       \<Longrightarrow> I s'"
    and visited': "\<And>s s' u v e c b.
       \<lbrakk>I s; cond s; DFS_invar G param s; DFS_invar G param s';
        stack s \<noteq> []; v \<in> pending s `` {u}; v \<in> dom (discovered s);
        u = hd (stack s);
        cross_edges s \<subseteq> c; back_edges s \<subseteq> b;
        s' = s\<lparr> 
          pending := pending s - {(u,v)},
          state.more := e, 
          cross_edges := c, 
          back_edges := b\<rparr>\<rbrakk>
       \<Longrightarrow> I s'"
    and discover': "\<And>s s' u v e.
       \<lbrakk>I s; cond s;  DFS_invar G param s; DFS_invar G param s';
        stack s \<noteq> []; v \<in> pending s `` {u}; v \<notin> dom (discovered s); 
        u = hd (stack s); 
        s' = (discover u v (s\<lparr>pending := pending s - {(u,v)}\<rparr>))\<lparr>state.more := e\<rparr>\<rbrakk>
       \<Longrightarrow> I s'"
    shows "is_invar I"
  proof (induct rule: is_invarI_full)
    case (cross_edge s s' u v e) thus ?case
      apply -
      apply (rule visited'[of s s' v u "insert (u,v) (cross_edges s)" "back_edges s" e])
      apply clarsimp_all
      done
  next
    case (back_edge s s' u v e) thus ?case
      apply -
      apply (rule visited'[of s s' v u "cross_edges s" "insert (u,v) (back_edges s)" e])
      apply clarsimp_all
      done
  qed fact+

  (* Establish some invariants *)
  definition "basic_invar s \<equiv>
    set (stack s) = dom (discovered s) - dom (finished s) \<and>
    distinct (stack s) \<and>
    (stack s \<noteq> [] \<longrightarrow> last (stack s) \<in> V0) \<and>
    dom (finished s) \<subseteq> dom (discovered s) \<and>
    Domain (pending s) \<subseteq> dom (discovered s) - dom (finished s) \<and>
    pending s \<subseteq> E"

  lemma i_basic_invar: "is_invar basic_invar" 
    unfolding basic_invar_def[abs_def]
    apply (induction rule: is_invarI)
    apply (clarsimp_all simp: neq_Nil_conv last_tl)
    apply blast+
    done
end

context DFS_invar begin
  lemmas basic_invar = make_invar_thm[OF i_basic_invar]

  lemma pending_ssE: "pending s \<subseteq> E" 
    using basic_invar 
    by (auto simp: basic_invar_def)

  lemma pendingD:
    "(u,v)\<in>pending s \<Longrightarrow> (u,v)\<in>E \<and> u\<in>dom (discovered s)"
    using basic_invar 
    by (auto simp: basic_invar_def)

  lemma stack_set_def:
    "set (stack s) = dom (discovered s) - dom (finished s)"
    using basic_invar
    by (simp add: basic_invar_def)

  lemma stack_discovered:
    "set (stack s) \<subseteq> dom (discovered s)"
    using stack_set_def
    by auto

  lemma stack_distinct:
    "distinct (stack s)"
    using basic_invar
    by (simp add: basic_invar_def)

  lemma last_stack_in_V0:
    "stack s \<noteq> [] \<Longrightarrow> last (stack s) \<in> V0"
    using basic_invar
    by (simp add: basic_invar_def)

  lemma stack_not_finished:
    "x \<in> set (stack s) \<Longrightarrow> x \<notin> dom (finished s)"
    using stack_set_def
    by auto

  lemma discovered_not_stack_imp_finished:
    "x \<in> dom (discovered s) \<Longrightarrow> x \<notin> set (stack s) \<Longrightarrow> x \<in> dom (finished s)"
    using stack_set_def
    by auto

  lemma finished_discovered:
    "dom (finished s) \<subseteq> dom (discovered s)"
    using basic_invar
    by (auto simp add: basic_invar_def)

  lemma finished_no_pending:
    "v \<in> dom (finished s) \<Longrightarrow> pending s `` {v} = {}"
    using basic_invar
    by (auto simp add: basic_invar_def)

  lemma discovered_eq_finished_un_stack:
    "dom (discovered s) = dom (finished s) \<union> set (stack s)"
    using stack_set_def finished_discovered by auto


  lemma pending_on_stack:
    "(v,w) \<in> pending s \<Longrightarrow> v \<in> set (stack s)"
    using basic_invar
    by (auto simp add: basic_invar_def)

  lemma empty_stack_imp_empty_pending:
    "stack s = [] \<Longrightarrow> pending s = {}"
    using pending_on_stack
    by auto
end


context param_DFS begin

(* Establish some more invariants *)
  lemma i_discovered_reachable: 
    "is_invar (\<lambda>s. dom (discovered s) \<subseteq> reachable)"
  proof (induct rule: is_invarI)
    case (discover s) then interpret i: DFS_invar where s=s by simp
    from discover show ?case 
      apply (clarsimp dest!: i.pendingD)
      by (metis contra_subsetD list.sel_set(1) rtrancl_image_advance i.stack_discovered)
  qed auto

  definition "discovered_closed s \<equiv>
      E``dom (finished s) \<subseteq> dom (discovered s)
    \<and> (E - pending s) `` set (stack s) \<subseteq> dom (discovered s)"

  lemma i_discovered_closed: "is_invar discovered_closed"
  proof (induct rule: is_invarI)
    case (finish s s') 
    hence "(E - pending s) `` set (stack s) \<subseteq> dom (discovered s)" 
      by (simp add: discovered_closed_def)
    moreover from finish have "set (stack s') \<subseteq> set (stack s)" 
      by (auto simp add: neq_Nil_conv cond_def)
    ultimately have "(E - pending s') `` set (stack s') \<subseteq> dom (discovered s')"
      using finish
      by simp blast

    moreover
    from `stack s \<noteq> []` finish have "E``dom (finished s') \<subseteq> dom (discovered s')" 
      apply (cases "stack s") apply simp
      apply (simp add: discovered_closed_def) 
      apply (blast)
      done
    ultimately show ?case by (simp add: discovered_closed_def)
  qed (auto simp add: discovered_closed_def cond_def)

  lemma i_discovered_finite: "is_invar (\<lambda>s. finite (dom (discovered s)))"
    by (induction rule: is_invarI) auto

end

context DFS_invar
begin

  lemmas discovered_reachable = 
    i_discovered_reachable [THEN make_invar_thm]

  lemma stack_reachable: "set (stack s) \<subseteq> reachable" 
    using stack_discovered discovered_reachable by blast

  lemmas discovered_closed = i_discovered_closed[THEN make_invar_thm]

  lemmas discovered_finite[simp, intro!] = i_discovered_finite[THEN make_invar_thm]
  lemma finished_finite[simp, intro!]: "finite (dom (finished s))"
    using finished_discovered discovered_finite by (rule finite_subset)

  lemma finished_closed:
    "E `` dom (finished s) \<subseteq> dom (discovered s)"
    using discovered_closed[unfolded discovered_closed_def]
    by auto

  lemma finished_imp_succ_discovered:
    "v \<in> dom (finished s) \<Longrightarrow> w \<in> succ v \<Longrightarrow> w \<in> dom (discovered s)"
    using discovered_closed[unfolded discovered_closed_def]
    by auto

  lemma pending_reachable: "pending s \<subseteq> reachable \<times> reachable" 
    using pendingD discovered_reachable
    by (fast intro: rtrancl_image_advance_rtrancl)

  lemma no_pending_imp_succ_discovered:
    assumes "u \<in> dom (discovered s)"
    and "pending s `` {u} = {}"
    and "v \<in> succ u"
    shows "v \<in> dom (discovered s)"
  proof (cases "u \<in> dom (finished s)")
    case True with finished_imp_succ_discovered assms show ?thesis by simp
  next
    case False with stack_set_def assms have "u \<in> set (stack s)" by auto
    with assms discovered_closed[unfolded discovered_closed_def] show ?thesis by blast
  qed

  lemma nc_finished_eq_reachable:
    assumes NC: "\<not>cond s" "\<not>is_break param s" 
    shows "dom (finished s) = reachable"
  proof -
    from NC basic_invar 
    have [simp]: "stack s = []" "dom (discovered s) = dom (finished s)" and SS: "V0 \<subseteq> dom (discovered s)"
      unfolding basic_invar_def cond_alt by auto

    show "dom (finished s) = reachable"
    proof 
      from discovered_reachable show "dom (finished s) \<subseteq> reachable"
        by simp
    next
      from discovered_closed have "E``(dom (finished s)) \<subseteq> dom (finished s)"
        unfolding discovered_closed_def by auto
      with SS show "reachable \<subseteq> dom (finished s)"
        by (simp, metis rtrancl_reachable_induct)
    qed
  qed

  lemma nc_V0_finished:
    assumes NC: "\<not> cond s" "\<not> is_break param s"
    shows "V0 \<subseteq> dom (finished s)"
    using nc_finished_eq_reachable[OF NC]
    by blast 

  lemma nc_discovered_eq_finished:
    assumes NC: "\<not> cond s" "\<not> is_break param s"
    shows "dom (discovered s) = dom (finished s)"
    using finished_discovered
    using nc_finished_eq_reachable[OF NC] discovered_reachable
    by blast

  lemma nc_discovered_eq_reachable:
    assumes NC: "\<not> cond s" "\<not> is_break param s"
    shows "dom (discovered s) = reachable"
    using NC
    using nc_discovered_eq_finished nc_finished_eq_reachable
    by blast

  lemma nc_fin_closed: 
    assumes NC: "\<not>cond s"
    assumes NB: "\<not>is_break param s"
    shows "E``dom (finished s) \<subseteq> dom (finished s)"
    using finished_imp_succ_discovered
    by (auto simp: nc_discovered_eq_finished[OF NC NB])

end

(* Timing *)
abbreviation "the_discovered s v \<equiv> the (discovered s v)"
abbreviation "the_finished s v \<equiv> the (finished s v)"

locale timing_syntax
begin
  (* Timing related syntax shortcuts *) 
  notation the_discovered ("\<delta>")
  notation the_finished ("\<phi>")
end

context param_DFS begin context begin interpretation timing_syntax .

  definition "timing_common_inv s \<equiv> 
  (* \<delta> s v < \<phi> s v *)
   (\<forall>v \<in> dom (finished s). \<delta> s v < \<phi> s v)

  (* v \<noteq> w \<longrightarrow> \<delta> s v \<noteq> \<delta> s w \<and> \<phi> s v \<noteq> \<phi> s w
     Can't use 'card dom = card ran' as the maps may be infinite ... *)
  \<and> (\<forall>v \<in> dom (discovered s). \<forall>w \<in> dom (discovered s). v \<noteq> w \<longrightarrow> \<delta> s v \<noteq> \<delta> s w)
  \<and> (\<forall>v \<in> dom (finished s). \<forall>w \<in> dom (finished s). v \<noteq> w \<longrightarrow> \<phi> s v \<noteq> \<phi> s w)

  (* \<delta> s v < counter \<and> \<phi> s v < counter *)
  \<and> (\<forall>v \<in> dom (discovered s). \<delta> s v < counter s)
  \<and> (\<forall>v \<in> dom (finished s). \<phi> s v < counter s)

  \<and> (\<forall>v \<in> dom (finished s). \<forall>w \<in> succ v. \<delta> s w < \<phi> s v)"

  lemma timing_common_inv: 
    "is_invar timing_common_inv"
  proof (induction rule: is_invarI)
    case (finish s s') then interpret DFS_invar where s=s by simp

    from finish have NE: "stack s \<noteq> []" by (simp add: cond_alt)
    
    have *: "hd (stack s) \<notin> dom (finished s)" "hd (stack s) \<in> dom (discovered s)"
      using stack_not_finished stack_discovered hd_in_set[OF NE]
      by blast+

    from discovered_closed have 
      "(E - pending s) `` {hd (stack s)} \<subseteq> dom (discovered s)"
      using hd_in_set[OF NE]
      by (auto simp add: discovered_closed_def)
    hence succ_hd: "pending s `` {hd (stack s)} = {} 
      \<Longrightarrow> succ (hd (stack s)) \<subseteq> dom (discovered s)"
      by blast
    
    from finish show ?case 
        apply (simp add: timing_common_inv_def)
        apply (intro conjI)
        using * apply simp
        using * apply simp
        apply (metis less_irrefl)
        apply (metis less_irrefl)
        apply (metis less_SucI)
        apply (metis less_SucI)
        apply (blast dest!: succ_hd)
        using * apply simp
        done
  next
    case (discover s) then interpret DFS_invar where s=s by simp
    from discover show ?case
      apply (simp add: timing_common_inv_def)
      apply (intro conjI)
      using finished_discovered apply fastforce
      apply (metis less_irrefl)
      apply (metis less_irrefl)
      apply (metis less_SucI)
      apply (metis less_SucI)
      using finished_imp_succ_discovered apply fastforce
      done
  next
    case (new_root s s' v0) then interpret DFS_invar where s=s by simp
    from new_root show ?case
      apply (simp add: timing_common_inv_def)
      apply (intro conjI)
      using finished_discovered apply fastforce
      apply (metis less_irrefl)
      apply (metis less_irrefl)
      apply (metis less_SucI)
      apply (metis less_SucI)
      using finished_imp_succ_discovered apply fastforce
      done
  qed (simp_all add: timing_common_inv_def)
end end

context DFS_invar begin context begin interpretation timing_syntax .

  lemmas s_timing_common_inv = 
    timing_common_inv[THEN make_invar_thm]

  lemma timing_less_counter:
    "v \<in> dom (discovered s) \<Longrightarrow> \<delta> s v < counter s"
    "v \<in> dom (finished s) \<Longrightarrow> \<phi> s v < counter s"
    using s_timing_common_inv
    by (auto simp add: timing_common_inv_def)

  lemma disc_lt_fin:
    "v \<in> dom (finished s) \<Longrightarrow> \<delta> s v < \<phi> s v"
    using s_timing_common_inv
    by (auto simp add: timing_common_inv_def)

  lemma disc_unequal:
    assumes "v \<in> dom (discovered s)" "w \<in> dom (discovered s)"
    and "v \<noteq> w"
    shows "\<delta> s v \<noteq> \<delta> s w"
    using s_timing_common_inv assms
    by (auto simp add: timing_common_inv_def)

  lemma fin_unequal:
    assumes "v \<in> dom (finished s)" "w \<in> dom (finished s)"
    and "v \<noteq> w"
    shows "\<phi> s v \<noteq> \<phi> s w"
    using s_timing_common_inv assms
    by (auto simp add: timing_common_inv_def)

  lemma finished_succ_fin:
    assumes "v \<in> dom (finished s)"
    and "w \<in> succ v"
    shows "\<delta> s w < \<phi> s v"
    using assms s_timing_common_inv
    by (simp add: timing_common_inv_def)
end end

context param_DFS begin context begin interpretation timing_syntax .

  lemma i_prev_stack_discover_all:
    "is_invar (\<lambda>s. \<forall> n < length (stack s). \<forall> v \<in> set (drop (Suc n) (stack s)). 
                   \<delta> s (stack s ! n) > \<delta> s v)"
  proof (induct rule: is_invarI)
    case (finish s) thus ?case 
      by (cases "stack s") auto
  next
    case (discover s s' u v) 
    hence EQ[simp]: "discovered s' = (discovered s)(v \<mapsto> counter s)"
                    "stack s' = v#stack s"
      by simp_all
    
    from discover interpret DFS_invar where s=s by simp
    from discover stack_discovered have v_ni: "v \<notin> set (stack s)" by auto

    from stack_discovered timing_less_counter have 
      "\<And>w. w \<in> set (stack s) \<Longrightarrow> \<delta> s w < counter s" 
      by blast
    with v_ni have "\<And>w. w \<in> set (stack s) \<Longrightarrow> \<delta> s' w < \<delta> s' v" by auto
    hence "\<And>w. w \<in> set (drop (Suc 0) (stack s')) \<Longrightarrow> \<delta> s' w < \<delta> s' (stack s' ! 0)" 
      by auto
      
    moreover
    from v_ni have 
      "\<And>n. \<lbrakk>n < (length (stack s')); n > 0\<rbrakk>
      \<Longrightarrow> \<delta> s' (stack s' ! n) = \<delta> s (stack s' ! n)" 
      by auto
    with discover(1) v_ni 
    have  "\<And>n. \<lbrakk>n < (length (stack s')) - 1; n > 0\<rbrakk> 
      \<Longrightarrow> \<forall> w \<in> set (drop (Suc n) (stack s')). \<delta> s' (stack s' ! n) > \<delta> s' w"
      by (auto dest: in_set_dropD)
    ultimately show ?case 
      by (metis drop_Suc_Cons length_drop length_pos_if_in_set length_tl 
        list.sel(3) neq0_conv nth_Cons_0 EQ(2) zero_less_diff)
  qed simp_all
end end

context DFS_invar begin context begin interpretation timing_syntax .

  lemmas prev_stack_discover_all 
    = i_prev_stack_discover_all[THEN make_invar_thm]

  lemma prev_stack_discover:
    "\<lbrakk>n < length (stack s); v \<in> set (drop (Suc n) (stack s)) \<rbrakk>
     \<Longrightarrow> \<delta> s (stack s ! n) > \<delta> s v"
    by (metis prev_stack_discover_all)

  lemma Suc_stack_discover:
    assumes n: "n < (length (stack s)) - 1"
    shows "\<delta> s (stack s ! n) > \<delta> s (stack s ! Suc n)"
  proof -
    from prev_stack_discover assms have 
      "\<And> v. v \<in> set (drop (Suc n) (stack s)) \<Longrightarrow> \<delta> s (stack s ! n) > \<delta> s v" 
      by fastforce
    moreover from n have "stack s ! Suc n \<in> set (drop (Suc n) (stack s))" 
      using in_set_conv_nth by fastforce
    ultimately show ?thesis .
  qed

  lemma tl_lt_stack_hd_discover:
    assumes notempty: "stack s \<noteq> []"
    and "x \<in> set (tl (stack s))"
    shows "\<delta> s x < \<delta> s (hd (stack s))"
  proof -
    from notempty obtain y ys where "stack s = y#ys" by (metis list.exhaust)
    with assms show ?thesis
      using prev_stack_discover
      by (cases ys) force+
  qed

  lemma stack_nth_order:
    assumes l: "i < length (stack s)" "j < length (stack s)"
    shows "\<delta> s (stack s ! i) < \<delta> s (stack s ! j) \<longleftrightarrow> i > j" (is "\<delta> s ?i < \<delta> s ?j \<longleftrightarrow> _")
  proof
    assume \<delta>: "\<delta> s ?i < \<delta> s ?j"
    
    from l stack_set_def have 
      disc: "?i \<in> dom (discovered s)" "?j \<in> dom (discovered s)"
      by auto
    with disc_unequal[OF disc] \<delta> have "i \<noteq> j" by auto

    moreover
    {
      assume "i < j"

      with l have "stack s ! j \<in> set (drop (Suc i) (stack s))" 
        using in_set_drop_conv_nth[of "stack s ! j" "Suc i" "stack s"]
        by fastforce
      with prev_stack_discover l have "\<delta> s (stack s ! j) < \<delta> s (stack s ! i)" 
        by simp
      with \<delta> have "False" by simp
    }
    ultimately show "i > j" by force
  next
    assume "i > j"
    with l have "stack s ! i \<in> set (drop (Suc j) (stack s))" 
      using in_set_drop_conv_nth[of "stack s ! i" "Suc j" "stack s"]
      by fastforce
    with prev_stack_discover l show "\<delta> s ?i < \<delta> s ?j" by simp
  qed
end end


(* Parenthesis Thm *)
context param_DFS begin context begin interpretation timing_syntax .

  definition "parenthesis s \<equiv>
    \<forall>v \<in> dom (discovered s). \<forall>w \<in> dom (discovered s).
     \<delta> s v < \<delta> s w \<and> v \<in> dom (finished s) \<longrightarrow> (
             \<phi> s v < \<delta> s w (* disjoint *)
           \<or> (\<phi> s v > \<delta> s w \<and> w \<in> dom (finished s) \<and> \<phi> s w < \<phi> s v))"

  lemma i_parenthesis: "is_invar parenthesis"
  proof (induct rule: is_invarI)
    case (finish s s')
    hence EQ[simp]: "discovered s' = discovered s" 
                    "counter s' = Suc (counter s)"
                    "finished s' = finished s(hd (stack s) \<mapsto> counter s)"
      by simp_all

    from finish interpret DFS_invar where s=s by simp
    from finish have NE[simp]: "stack s \<noteq> []" by (simp add: cond_alt)

    {
      fix x y
      assume  dom: "x \<in> dom (discovered s')" "y \<in> dom (discovered s')"
        and \<delta>: "\<delta> s' x < \<delta> s' y"
        and f: "x \<in> dom (finished s')"
      hence neq: "x \<noteq> y" by force

      note assms = dom \<delta> f EQ

      let ?DISJ = "\<phi> s' x < \<delta> s' y"
      let ?IN = "\<delta> s' y < \<phi> s' x \<and> y \<in> dom (finished s') \<and> \<phi> s' y < \<phi> s' x"

      have "?DISJ \<or> ?IN"
      proof (cases "x = hd (stack s)")
        case True note x_is_hd = this
        hence \<phi>x: "\<phi> s' x = counter s" by simp
        from x_is_hd neq have y_not_hd: "y \<noteq> hd (stack s)" by simp
        
        have "\<delta> s y < \<phi> s' x \<and> y \<in> dom (finished s) \<and> \<phi> s y < \<phi> s' x"
        proof (cases "y \<in> set (stack s)")
          -- \<open>y on stack is not possible: According to @{thm [display] \<delta>} it is discovered after @{text "x (= hd (stack s))"}\<close>
          case True with y_not_hd have "y \<in> set (tl (stack s))" 
            by (cases "stack s") simp_all
          with tl_lt_stack_hd_discover[OF NE] \<delta> x_is_hd have "\<delta> s y < \<delta> s x" 
            by simp
          with \<delta> have False by simp
          thus ?thesis ..
        next
          case False -- \<open>y must be a successor of @{text "x (= (hd (stack s)))"}\<close>
          from dom have "y \<in> dom (discovered s)" by simp
          with False discovered_not_stack_imp_finished  have *: 
            "y \<in> dom (finished s)" 
            by simp
          moreover with timing_less_counter \<phi>x have "\<phi> s y < \<phi> s' x" by simp
          moreover with * disc_lt_fin \<phi>x have "\<delta> s y < \<phi> s' x" 
            by (metis less_trans)
          ultimately show ?thesis by simp
        qed
        with y_not_hd show ?thesis by simp
      next
        case False note [simp] = this
        show ?thesis
        proof (cases "y = hd (stack s)")
          case False with finish assms show ?thesis 
            by (simp add: parenthesis_def)
        next
          case True with stack_not_finished have "y \<notin> dom (finished s)" 
            using hd_in_set[OF NE] 
            by auto
          with finish assms have "\<phi> s x < \<delta> s y" 
            unfolding parenthesis_def
            by auto
          hence ?DISJ by simp
          thus ?thesis ..
        qed
      qed
    }
    thus ?case by (simp add: parenthesis_def)
  next
    case (discover s s' u v)
    hence EQ[simp]: "discovered s' = (discovered s)(v \<mapsto> counter s)" 
                    "finished s' = finished s"
                    "counter s' = Suc (counter s)"
      by simp_all
    
    from discover interpret DFS_invar where s=s by simp
    from discover finished_discovered have 
      V': "v \<notin> dom (discovered s)" "v \<notin> dom (finished s)" 
      by auto

    {
      fix x y
      assume  dom: "x \<in> dom (discovered s')" "y \<in> dom (discovered s')"
        and \<delta>: "\<delta> s' x < \<delta> s' y"
        and f: "x \<in> dom (finished s')"

      let ?DISJ = "\<phi> s' x < \<delta> s' y"
      let ?IN = "\<delta> s' y < \<phi> s' x \<and> y \<in> dom (finished s') \<and> \<phi> s' y < \<phi> s' x"
      
      from dom V' f have x: "x \<in> dom (discovered s)""x \<noteq> v" by auto
      
      have "?DISJ \<or> ?IN"
      proof (cases "y = v")
        case True hence "\<delta> s' y = counter s" by simp
        moreover from timing_less_counter x f have "\<phi> s' x < counter s" by auto
        ultimately have "?DISJ" by simp
        thus ?thesis ..
      next
        case False with dom have "y \<in> dom (discovered s)" by simp
        with discover False \<delta> f x show ?thesis by (simp add: parenthesis_def)
      qed
    }
    thus ?case by (simp add: parenthesis_def)
  next
    case (new_root s s' v0) 
    then interpret DFS_invar where s=s by simp
    
    from finished_discovered new_root have "v0 \<notin> dom (finished s')" by auto
    with new_root timing_less_counter show ?case by (simp add: parenthesis_def)
  qed (simp_all add: parenthesis_def)
end end

context DFS_invar begin context begin interpretation timing_syntax .

  lemma parenthesis:
    assumes "v \<in> dom (finished s)" "w \<in> dom (discovered s)"
    and "\<delta> s v < \<delta> s w"
    shows "\<phi> s v < \<delta> s w (* disjoint *)
           \<or> (\<phi> s v > \<delta> s w \<and> w \<in> dom (finished s) \<and> \<phi> s w < \<phi> s v)"
    using assms
    using i_parenthesis[THEN make_invar_thm]
    using finished_discovered
    unfolding parenthesis_def
    by blast

  lemma parenthesis_contained:
    assumes "v \<in> dom (finished s)" "w \<in> dom (discovered s)"
    and "\<delta> s v < \<delta> s w" "\<phi> s v > \<delta> s w"
    shows "w \<in> dom (finished s) \<and> \<phi> s w < \<phi> s v"
    using parenthesis assms
    by force
  
  lemma parenthesis_disjoint:
    assumes "v \<in> dom (finished s)" "w \<in> dom (discovered s)"
    and "\<delta> s v < \<delta> s w" "\<phi> s w > \<phi> s v"
    shows "\<phi> s v < \<delta> s w"
    using parenthesis assms
    by force

  lemma finished_succ_contained:
    assumes "v \<in> dom (finished s)"
    and "w \<in> succ v"
    and "\<delta> s v < \<delta> s w"
    shows "w \<in> dom (finished s) \<and> \<phi> s w < \<phi> s v"
    using finished_succ_fin finished_imp_succ_discovered parenthesis_contained
    using assms
    by metis

end end


(* Edge Types *)


context param_DFS 
begin
  abbreviation "edges s \<equiv> tree s \<union> cross_edges s \<union> back_edges s"

  (* Demo for simple invariant proof *)
  lemma "is_invar (\<lambda>s. finite (edges s))"
    by (induction rule: establish_invarI) auto

  text {* Sometimes it's useful to just chose between tree and non-tree. *}
  lemma edgesE_CB:
    assumes "x \<in> edges s"
    and "x \<in> tree s \<Longrightarrow> P"
    and "x \<in> cross_edges s \<union> back_edges s \<Longrightarrow> P"
    shows "P"
    using assms by auto

  definition "edges_basic s \<equiv> 
     Field (back_edges s) \<subseteq> dom (discovered s) \<and> back_edges s \<subseteq> E - pending s
   \<and> Field (cross_edges s) \<subseteq> dom (discovered s) \<and> cross_edges s \<subseteq> E - pending s
   \<and> Field (tree s) \<subseteq> dom (discovered s) \<and> tree s \<subseteq> E - pending s
   \<and> back_edges s \<inter> cross_edges s = {}
   \<and> back_edges s \<inter> tree s = {}
   \<and> cross_edges s \<inter> tree s = {}
   "

  lemma i_edges_basic:
    "is_invar edges_basic"
    unfolding edges_basic_def[abs_def]
  proof (induct rule: is_invarI_full)
    case (back_edge s)
    then interpret DFS_invar where s=s by simp
    from back_edge show ?case by (auto dest: pendingD)
  next
    case (cross_edge s)
    then interpret DFS_invar where s=s by simp
    from cross_edge show ?case by (auto dest: pendingD)
  next
    case (discover s)
    then interpret DFS_invar where s=s by simp
    from discover show ?case 
      (* Speed optimized proof, using only auto takes too long *)
      apply (simp add: Field_def Range_def Domain_def)
      apply (drule pendingD) apply simp
      by (blast)
  next
    case (new_root s) 
    thus ?case by (simp add: Field_def) blast
  qed auto

  lemmas (in DFS_invar) edges_basic = i_edges_basic[THEN make_invar_thm]
    
  lemma i_edges_covered: 
    "is_invar (\<lambda>s. (E \<inter> dom (discovered s) \<times> UNIV) - pending s = edges s)"
  proof (induction rule: is_invarI_full)
    case (new_root s s' v0) 
    interpret DFS_invar G param s by fact  

    from new_root empty_stack_imp_empty_pending 
    have [simp]: "pending s = {}" by simp

    from `v0 \<notin> dom (discovered s)` 
    have [simp]: "E \<inter> insert v0 (dom (discovered s)) \<times> UNIV - {v0} \<times> succ v0 
      = E \<inter> dom (discovered s) \<times> UNIV" by auto

    from new_root show ?case by simp
  next
    case (cross_edge s s' u v) 
    interpret DFS_invar G param s by fact  

    from cross_edge stack_discovered have "u \<in> dom (discovered s)" 
      by (cases "stack s") auto

    with cross_edge(2-) pending_ssE have 
      "E \<inter> dom (discovered s) \<times> UNIV - (pending s - {(hd (stack s), v)})
      = insert (hd (stack s), v) (E \<inter> dom (discovered s) \<times> UNIV - pending s)"
      by auto

    thus ?case using cross_edge by simp
  next
    case (back_edge s s' u v)
    interpret DFS_invar G param s by fact  

    from back_edge stack_discovered have "u \<in> dom (discovered s)" 
      by (cases "stack s") auto

    with back_edge(2-) pending_ssE have 
      "E \<inter> dom (discovered s) \<times> UNIV - (pending s - {(hd (stack s), v)})
      = insert (hd (stack s), v) (E \<inter> dom (discovered s) \<times> UNIV - pending s)"
      by auto

    thus ?case using back_edge by simp
  next
    case (discover s s' u v)
    interpret DFS_invar G param s by fact  

    from discover stack_discovered have "u \<in> dom (discovered s)" 
      by (cases "stack s") auto

    with discover(2-) pending_ssE have 
      "E \<inter> insert v (dom (discovered s)) \<times> UNIV 
        - (pending s - {(hd (stack s), v)} \<union> {v} \<times> succ v)
      = insert (hd (stack s), v) (E \<inter> dom (discovered s) \<times> UNIV - pending s)"
      by auto

    thus ?case using discover by simp
  qed simp_all
end

context DFS_invar begin

  lemmas edges_covered = 
    i_edges_covered[THEN make_invar_thm]

  lemma edges_ss_reachable_edges:
    "edges s \<subseteq> E \<inter> reachable \<times> UNIV"
    using edges_covered discovered_reachable
    by (fast intro: rtrancl_image_advance_rtrancl)

  lemma nc_edges_covered:
    assumes "\<not>cond s" "\<not>is_break param s"
    shows "E \<inter> reachable \<times> UNIV = edges s"
  proof -
    from assms have [simp]: "stack s = []" 
      unfolding cond_def by (auto simp: pred_defs)
    hence [simp]: "pending s = {}" by (rule empty_stack_imp_empty_pending)
    
    
    from edges_covered nc_discovered_eq_reachable[OF assms] 
    show ?thesis by simp
  qed

  lemma 
    tree_ssE: "tree s \<subseteq> E" and
    tree_not_pending: "tree s \<subseteq> - pending s" and
    tree_edge_is_succ: "(v,w) \<in> tree s \<Longrightarrow> w \<in> succ v" and
    tree_discovered: "Field (tree s) \<subseteq> dom (discovered s)" and

    cross_edges_ssE: "cross_edges s \<subseteq> E" and
    cross_edges_not_pending: "cross_edges s \<subseteq> - pending s" and
    cross_edge_is_succ: "(v,w) \<in> cross_edges s \<Longrightarrow> w \<in> succ v" and
    cross_edges_discovered: "Field (cross_edges s) \<subseteq> dom (discovered s)" and

    back_edges_ssE: "back_edges s \<subseteq> E" and
    back_edges_not_pending: "back_edges s \<subseteq> - pending s" and
    back_edge_is_succ: "(v,w) \<in> back_edges s \<Longrightarrow> w \<in> succ v" and
    back_edges_discovered: "Field (back_edges s) \<subseteq> dom (discovered s)"
    using edges_basic
    unfolding edges_basic_def 
    by auto

  lemma edges_disjoint: 
   "back_edges s \<inter> cross_edges s = {}"
   "back_edges s \<inter> tree s = {}"
   "cross_edges s \<inter> tree s = {}"
    using edges_basic
    unfolding edges_basic_def 
    by auto

  lemma tree_edge_imp_discovered:
    "(v,w) \<in> tree s \<Longrightarrow> v \<in> dom (discovered s)"
    "(v,w) \<in> tree s \<Longrightarrow> w \<in> dom (discovered s)"
    using tree_discovered
    by (auto simp add: Field_def)

  lemma back_edge_imp_discovered:
    "(v,w) \<in> back_edges s \<Longrightarrow> v \<in> dom (discovered s)"
    "(v,w) \<in> back_edges s \<Longrightarrow> w \<in> dom (discovered s)"
    using back_edges_discovered
    by (auto simp add: Field_def)

  lemma cross_edge_imp_discovered:
    "(v,w) \<in> cross_edges s \<Longrightarrow> v \<in> dom (discovered s)"
    "(v,w) \<in> cross_edges s \<Longrightarrow> w \<in> dom (discovered s)"
    using cross_edges_discovered
    by (auto simp add: Field_def)

  lemma edge_imp_discovered:
    "(v,w) \<in> edges s \<Longrightarrow> v \<in> dom (discovered s)"
    "(v,w) \<in> edges s \<Longrightarrow> w \<in> dom (discovered s)"
    using tree_edge_imp_discovered cross_edge_imp_discovered back_edge_imp_discovered
    by blast+

  lemma tree_finite[simp, intro!]: "finite (tree s)"
    using finite_subset[OF tree_discovered discovered_finite] by simp

  lemma cross_edges_finite[simp, intro!]: "finite (cross_edges s)"
    using finite_subset[OF cross_edges_discovered discovered_finite] by simp

  lemma back_edges_finite[simp, intro!]: "finite (back_edges s)"
    using finite_subset[OF back_edges_discovered discovered_finite] by simp

  lemma edges_finite: "finite (edges s)"
    by auto

  lemma pending_finite[simp, intro!]: "finite (pending s)"
  proof -
    have "pending s \<subseteq> (SIGMA u:dom (discovered s). E``{u})"
      by (auto dest: pendingD)
    also have "finite \<dots>"
      apply rule
      apply (rule discovered_finite)
      using discovered_reachable
      by (blast intro: finitely_branching)
    finally (finite_subset) show ?thesis .
  qed
  
end

(* Tree *)

context DFS_invar begin context begin interpretation timing_syntax .
  lemma tree_edge_disc_lt_fin:
    "(v,w) \<in> tree s \<Longrightarrow> v \<in> dom (finished s) \<Longrightarrow> \<delta> s w < \<phi> s v"
    by (metis finished_succ_fin tree_edge_is_succ)

  lemma back_edge_disc_lt_fin:
    "(v,w) \<in> back_edges s \<Longrightarrow> v \<in> dom (finished s) \<Longrightarrow> \<delta> s w < \<phi> s v"
    by (metis finished_succ_fin back_edge_is_succ)

  lemma cross_edge_disc_lt_fin:
    "(v,w) \<in> cross_edges s \<Longrightarrow> v \<in> dom (finished s) \<Longrightarrow> \<delta> s w < \<phi> s v"
    by (metis finished_succ_fin cross_edge_is_succ)
end end

(* Stack & Tree *)
context param_DFS begin

  lemma i_stack_is_tree_path:
    "is_invar (\<lambda>s. stack s \<noteq> [] \<longrightarrow> (\<exists>v0 \<in> V0. 
        path (tree s) v0 (rev (tl (stack s))) (hd (stack s))))"
  proof (induct rule: is_invarI)
    case (discover s s' u v)
    hence EQ[simp]: "stack s' = v # stack s"
                    "tree s' = insert (hd (stack s), v) (tree s)"
      by simp_all
    from discover have NE[simp]: "stack s \<noteq> []" by simp

    from discover obtain v0 where 
      "v0 \<in> V0"
      "path (tree s) v0 (rev (tl (stack s))) (hd (stack s))"
      by blast
    with path_mono[OF _ this(2)] EQ have 
      "path (tree s') v0 (rev (tl (stack s))) (hd (stack s))" 
      by blast
    with `v0 \<in> V0` show ?case
      by (cases "stack s") (auto simp: path_simps)

  next
    case (finish s s')
    hence EQ[simp]: "stack s' = tl (stack s)"
                    "tree s' = tree s"
      by simp_all

    from finish obtain v0 where
      "v0 \<in> V0"
      "path (tree s) v0 (rev (tl (stack s))) (hd (stack s))"
      by blast
    hence P: "path (tree s') v0 (rev (stack s')) (hd (stack s))" by simp
      
    show ?case
    proof
      assume A: "stack s' \<noteq> []"
      with P have "(hd (stack s'), hd (stack s)) \<in> tree s'"
        by (auto simp: neq_Nil_conv path_simps)
      moreover from P A have 
        "path (tree s') v0 (rev (tl (stack s')) @ [hd (stack s')]) (hd (stack s))"
        by (simp)
      moreover note `v0 \<in> V0`
      ultimately show "\<exists>v0\<in>V0. path (tree s') v0 (rev (tl (stack s'))) (hd (stack s'))"
        by (auto simp add: path_append_conv)
    qed
  qed simp_all
end

context DFS_invar begin

  lemmas stack_is_tree_path = 
    i_stack_is_tree_path[THEN make_invar_thm, rule_format]

  lemma stack_is_path:
    "stack s \<noteq> [] \<Longrightarrow> \<exists>v0\<in>V0. path E v0 (rev (tl (stack s))) (hd (stack s))"
    using stack_is_tree_path path_mono[OF tree_ssE]
    by blast

  lemma hd_succ_stack_is_path:
    assumes ne: "stack s \<noteq> []"
    and succ: "v \<in> succ (hd (stack s))"
    shows "\<exists>v0\<in>V0. path E v0 (rev (stack s)) v"
  proof -
    from stack_is_path[OF ne] succ obtain v0 where
      "v0 \<in> V0"
      "path E v0 (rev (tl (stack s)) @ [hd (stack s)]) v"
      by (auto simp add: path_append_conv)
    thus ?thesis using ne
      by (cases "stack s") auto
  qed

  lemma tl_stack_hd_tree_path:
    assumes "stack s \<noteq> []"
    and "v \<in> set (tl (stack s))"
    shows "(v, hd (stack s)) \<in> (tree s)\<^sup>+"
  proof -
    from stack_is_tree_path assms obtain v0 where 
      "path (tree s) v0 (rev (tl (stack s))) (hd (stack s))" 
      by auto
    from assms path_member_reach_end[OF this] show ?thesis by simp
  qed
end

context param_DFS begin
  definition "tree_discovered_inv s \<equiv>
                       (tree s = {} \<longrightarrow> dom (discovered s) \<subseteq> V0 \<and> (stack s = [] \<or> (\<exists>v0\<in>V0. stack s = [v0])))
                     \<and> (tree s \<noteq> {} \<longrightarrow> (tree s)\<^sup>+ `` V0 \<union> V0 = dom (discovered s) \<union> V0)"


  lemma i_tree_discovered_inv: 
    "is_invar tree_discovered_inv"
  proof (induct rule: is_invarI)
    case (discover s s' u v) 
    hence EQ[simp]: "stack s' = v # stack s"
                    "tree s' = insert (hd (stack s), v) (tree s)"
                    "discovered s' = (discovered s)(v \<mapsto> counter s)"
      by simp_all
    
    from discover interpret DFS_invar where s=s by simp

    from discover have NE[simp]: "stack s \<noteq> []" by simp
    note TDI = `tree_discovered_inv s`[unfolded tree_discovered_inv_def]

    have "tree s' = {} \<longrightarrow> dom (discovered s') \<subseteq> V0 \<and> (stack s' = [] \<or> (\<exists>v0\<in>V0. stack s' = [v0]))"
      by simp -- \<open>@{text "tree s' \<noteq> {}"}\<close>

    moreover {
      fix x
      assume A: "x \<in> (tree s')\<^sup>+ `` V0 \<union> V0" "x \<notin> V0"
      then obtain y where y: "(y,x) \<in> (tree s')\<^sup>+" "y \<in> V0" by auto
     

      have "x \<in> dom (discovered s') \<union> V0"
      proof (cases "tree s = {}")
        case True with discover A have "(tree s')\<^sup>+ = {(hd (stack s), v)}"
          by (simp add: trancl_single)
        with A show ?thesis by auto
      next
        case False note t_ne = this

        show ?thesis
        proof (cases "x = v")
          case True thus ?thesis by simp
        next
          case False with y have "(y,x) \<in> (tree s)\<^sup>+"
          proof (induct rule: trancl_induct)
            case (step a b) hence "(a,b) \<in> tree s" by simp
            with tree_edge_imp_discovered have "a \<in> dom (discovered s)" by simp
            with discover have "a \<noteq> v" by blast
            with step show ?case by auto
          qed simp
          with `y \<in> V0` have "x \<in> (tree s)\<^sup>+ `` V0" by auto
          with t_ne TDI show ?thesis by auto
        qed
      qed
    } note t_d = this

    {
      fix x
      assume "x \<in> dom (discovered s') \<union> V0" "x \<notin> V0"
      hence A: "x \<in> dom (discovered s')" by simp

      have "x \<in> (tree s')\<^sup>+ `` V0 \<union> V0"
      proof (cases "tree s = {}")
        case True with trancl_single have "(tree s')\<^sup>+ = {(hd (stack s), v)}" by simp
        moreover from True TDI have "hd (stack s) \<in> V0" "dom (discovered s) \<subseteq> V0" by auto
        ultimately show ?thesis using A `x\<notin>V0` by auto
      next
        case False note t_ne = this

        show ?thesis
        proof (cases "x=v")
          case False with A have "x \<in> dom (discovered s)" by simp
          with TDI t_ne `x \<notin> V0` have "x \<in> (tree s)\<^sup>+ `` V0" by auto
          with trancl_sub_insert_trancl show ?thesis by simp blast
        next
          case True
          from t_ne TDI have "dom (discovered s) \<union> V0 = (tree s)\<^sup>+ `` V0 \<union> V0"
            by simp

          moreover from stack_is_tree_path[OF NE] obtain v0 where "v0 \<in> V0" and 
            "(v0, hd (stack s)) \<in> (tree s)\<^sup>*"
            by (blast intro!: path_is_rtrancl)
          with EQ have "(v0, hd (stack s)) \<in> (tree s')\<^sup>*" by (auto intro: rtrancl_mono_mp)
          ultimately show ?thesis using `v0 \<in> V0` True by (auto elim: rtrancl_into_trancl1)
        qed
      qed
    } with t_d have "(tree s')\<^sup>+ `` V0 \<union> V0 = dom (discovered s') \<union> V0" by blast

    ultimately show ?case by (simp add: tree_discovered_inv_def)
  qed (auto simp add: tree_discovered_inv_def)
      
  lemmas (in DFS_invar) tree_discovered_inv = 
    i_tree_discovered_inv[THEN make_invar_thm]

  lemma (in DFS_invar) discovered_iff_tree_path:
    "v \<notin> V0 \<Longrightarrow> v \<in> dom (discovered s) \<longleftrightarrow> (\<exists>v0\<in>V0. (v0,v) \<in> (tree s)\<^sup>+)"
    using tree_discovered_inv
    by (auto simp add: tree_discovered_inv_def)

  lemma i_tree_one_predecessor:
    "is_invar (\<lambda>s. \<forall>(v,v') \<in> tree s. \<forall>y. y \<noteq> v \<longrightarrow> (y,v') \<notin> tree s)"
  proof (induct rule: is_invarI)
    case (discover s s' u v) 
    hence EQ[simp]: "tree s' = insert (hd (stack s),v) (tree s)" by simp

    from discover interpret DFS_invar where s=s by simp
    from discover have NE[simp]: "stack s \<noteq> []" by (simp add: cond_alt)

    {
      fix w w' y
      assume *: "(w,w') \<in> tree s'"
        and  "y \<noteq> w"

      from discover stack_discovered have v_hd: "hd (stack s) \<noteq> v" 
        using hd_in_set[OF NE] by blast
      from discover tree_discovered have 
        v_notin_tree: "\<forall>(x,x') \<in> tree s. x \<noteq> v \<and> x' \<noteq> v" 
        by (blast intro!: Field_not_elem)

      have "(y,w') \<notin> tree s'"
      proof (cases "w = hd (stack s)")
        case True
        have "(y,v) \<notin> tree s'"
        proof (rule notI)
          assume "(y,v) \<in> tree s'" 
          with True `y\<noteq>w` have "(y,v) \<in> tree s" by simp
          with v_notin_tree show False by auto
        qed
        with True * `y\<noteq>w` v_hd show ?thesis 
          apply (cases "w = v")
          apply simp 
          using discover apply simp apply blast
          done
      next
        case False with v_notin_tree * `y\<noteq>w` v_hd 
        show ?thesis
          apply (cases "w' = v")
          apply simp apply blast
          using discover apply simp apply blast
          done
      qed
    }
    thus ?case by blast
  qed simp_all
 
  lemma (in DFS_invar) tree_one_predecessor:
    assumes "(v,w) \<in> tree s"
    and "a \<noteq> v"
    shows "(a,w) \<notin> tree s"
    using assms make_invar_thm[OF i_tree_one_predecessor]
    by blast

  lemma (in DFS_invar) tree_eq_rule:
    "\<lbrakk>(v,w) \<in> tree s; (u,w) \<in> tree s\<rbrakk> \<Longrightarrow> v=u"
    using tree_one_predecessor
    by blast

context begin interpretation timing_syntax .

  lemma i_tree_edge_disc:
    "is_invar (\<lambda>s. \<forall>(v,v') \<in> tree s. \<delta> s v < \<delta> s v')"
  proof (induct rule: is_invarI)
    case (discover s s' u v) 
    hence EQ[simp]: "tree s' = insert (hd (stack s), v) (tree s)"
                    "discovered s' = (discovered s)(v \<mapsto> counter s)"
      by simp_all

    from discover interpret DFS_invar where s=s by simp
    from discover have NE[simp]: "stack s \<noteq> []" by (simp add: cond_alt)
    
    from discover tree_discovered have 
      v_notin_tree: "\<forall>(x,x') \<in> tree s. x \<noteq> v \<and> x' \<noteq> v" 
      by (blast intro!: Field_not_elem)
    from discover stack_discovered have 
      v_hd: "hd (stack s) \<noteq> v" 
      using hd_in_set[OF NE] 
      by blast
    
    {
      fix a b
      assume T: "(a,b) \<in> tree s'"
      have "\<delta> s' a < \<delta> s' b"
      proof (cases "b = v")
        case True with T v_notin_tree have [simp]: "a = hd (stack s)" by auto
        with stack_discovered have "a \<in> dom (discovered s)" 
          by (metis hd_in_set NE set_mp)
        with v_hd True timing_less_counter show ?thesis by simp
      next
        case False with v_notin_tree T have "(a,b) \<in> tree s" "a \<noteq> v" by auto
        with discover have "\<delta> s a < \<delta> s b" by auto
        with False `a\<noteq>v` show ?thesis by simp
      qed
    } thus ?case by blast
  next
    case (new_root s s' v0) then interpret DFS_invar where s=s by simp
    from new_root have "tree s' = tree s" by simp
    moreover from tree_edge_imp_discovered new_root have "\<forall>(v,v') \<in> tree s. v \<noteq> v0 \<and> v' \<noteq> v0" by blast
    ultimately show ?case using new_root by auto
  qed simp_all
end end

context DFS_invar begin context begin interpretation timing_syntax .

  lemma tree_edge_disc:
    "(v,w) \<in> tree s \<Longrightarrow> \<delta> s v < \<delta> s w"
    using i_tree_edge_disc[THEN make_invar_thm]
    by blast
  
  lemma tree_path_disc:
    "(v,w) \<in> (tree s)\<^sup>+ \<Longrightarrow> \<delta> s v < \<delta> s w"
    by (auto elim!: trancl_induct dest: tree_edge_disc)

  lemma no_loop_in_tree:
    "(v,v) \<notin> (tree s)\<^sup>+"
    using tree_path_disc by auto

  lemma tree_acyclic:
    "acyclic (tree s)"
    by (metis acyclicI no_loop_in_tree)

  lemma no_self_loop_in_tree:
    "(v,v) \<notin> tree s"
    using tree_edge_disc by auto

  lemma tree_edge_unequal:
    "(v,w) \<in> tree s \<Longrightarrow> v \<noteq> w"
    by (metis no_self_loop_in_tree)

  lemma tree_path_unequal:
    "(v,w) \<in> (tree s)\<^sup>+ \<Longrightarrow> v \<noteq> w"
    by (metis no_loop_in_tree)

  lemma tree_subpath':
    assumes x: "(x,v) \<in> (tree s)\<^sup>+"
    and y: "(y,v) \<in> (tree s)\<^sup>+"
    and "x \<noteq> y"
    shows "(x,y) \<in> (tree s)\<^sup>+ \<or> (y,x) \<in> (tree s)\<^sup>+"
  proof -
    from x obtain px where px: "path (tree s) x px v" and "px \<noteq> []"
      using trancl_is_path by metis
    from y obtain py where py: "path (tree s) y py v" and "py \<noteq> []"
      using trancl_is_path by metis

    from `px \<noteq> []` `py \<noteq> []` px py
    show ?thesis
    proof (induction arbitrary: v rule: rev_nonempty_induct2')
      case (single) hence "(x,v) \<in> tree s" "(y,v) \<in> tree s" 
        by (simp_all add: path_simps)
      with tree_eq_rule have "x=y" by simp
      with `x\<noteq>y` show ?case by contradiction
    next
      case (snocl a as) hence "(y,v) \<in> tree s" by (simp add: path_simps)
      moreover from snocl have "path (tree s) x as a" "(a,v) \<in> tree s" 
        by (simp_all add: path_simps)
      ultimately have "path (tree s) x as y"
        using tree_eq_rule
        by auto
      with path_is_trancl `as \<noteq> []` show ?case by metis
    next
      case (snocr _ a as) hence "(x,v) \<in> tree s" by (simp add: path_simps)
      moreover from snocr have "path (tree s) y as a" "(a,v) \<in> tree s" 
        by (simp_all add: path_simps)
      ultimately have "path (tree s) y as x"
        using tree_eq_rule
        by auto
      with path_is_trancl `as \<noteq> []` show ?case by metis
    next
      case (snoclr a as b bs) hence 
        "path (tree s) x as a" "(a,v) \<in> tree s"
        "path (tree s) y bs b" "(b,v) \<in> tree s"
        by (simp_all add: path_simps)
      moreover hence "a=b"  using tree_eq_rule by simp
      ultimately show ?thesis using snoclr.IH by metis
    qed
  qed

  lemma tree_subpath:
    assumes "(x,v) \<in> (tree s)\<^sup>+"
    and "(y,v) \<in> (tree s)\<^sup>+"
    and \<delta>: "\<delta> s x < \<delta> s y"
    shows "(x,y) \<in> (tree s)\<^sup>+"
  proof -
    from \<delta> have "x \<noteq> y" by auto
    with assms tree_subpath' have "(x,y) \<in> (tree s)\<^sup>+ \<or> (y,x) \<in> (tree s)\<^sup>+" 
      by simp
    moreover from \<delta> tree_path_disc have "(y,x) \<notin> (tree s)\<^sup>+" by force
    ultimately show ?thesis by simp
  qed

  lemma on_stack_is_tree_path:
    assumes x: "x \<in> set (stack s)"
    and y: "y \<in> set (stack s)"
    and \<delta>: "\<delta> s x < \<delta> s y"
    shows "(x,y) \<in> (tree s)\<^sup>+"
  proof -
    from x obtain i where i: "stack s ! i = x" "i < length (stack s)"
      by (metis in_set_conv_nth)

    from y obtain j where j: "stack s ! j = y" "j < length (stack s)"
      by (metis in_set_conv_nth)
    
    with i \<delta> stack_nth_order have "j < i" by force
    
    from x have ne[simp]: "stack s \<noteq> []" by auto

    from `j<i` have "x \<in> set (tl (stack s))"
      using nth_mem nth_tl[OF ne, of "i - 1"] i
      by auto
    with tl_stack_hd_tree_path have 
      x_path: "(x, hd (stack s)) \<in> (tree s)\<^sup>+" 
      by simp

    then show ?thesis
    proof (cases "j=0")
      case True with j have "hd (stack s) = y" by (metis hd_conv_nth ne)
      with x_path show ?thesis by simp
    next
      case False hence "y \<in> set (tl (stack s))"
        using nth_mem nth_tl[OF ne, of "j - 1"] j
        by auto
      with tl_stack_hd_tree_path have "(y, hd (stack s)) \<in> (tree s)\<^sup>+" 
        by simp
      with x_path \<delta> show ?thesis
        using tree_subpath
        by metis
    qed
  qed

  lemma hd_stack_tree_path_finished:
    assumes "stack s \<noteq> []"
    assumes "(hd (stack s), v) \<in> (tree s)\<^sup>+"
    shows "v \<in> dom (finished s)"
  proof (cases "v \<in> set (stack s)")
    case True
    from assms no_loop_in_tree have "hd (stack s) \<noteq> v" by auto
    with True have "v \<in> set (tl (stack s))" by (cases "stack s") auto
    with tl_stack_hd_tree_path assms have "(hd (stack s), hd (stack s)) \<in> (tree s)\<^sup>+" by (metis trancl_trans)
    with no_loop_in_tree show ?thesis by contradiction
  next
    case False
    from assms obtain x where "(x,v) \<in> tree s" by (metis tranclE)
    with tree_edge_imp_discovered have "v \<in> dom (discovered s)" by blast
    with False show ?thesis by (simp add: stack_set_def)
  qed

  lemma tree_edge_impl_parenthesis:
    assumes t: "(v,w) \<in> tree s"
    and f: "v \<in> dom (finished s)"
    shows "w \<in> dom (finished s) 
      \<and> \<delta> s v < \<delta> s w
      \<and> \<phi> s w < \<phi> s v "
  proof -
    from tree_edge_disc_lt_fin assms have "\<delta> s w < \<phi> s v" by simp
    with f tree_edge_imp_discovered[OF t] tree_edge_disc[OF t]
    show ?thesis
      using parenthesis_contained
      by metis
  qed

  lemma tree_path_impl_parenthesis:
    assumes "(v,w) \<in> (tree s)\<^sup>+"
    and "v \<in> dom (finished s)"
    shows "w \<in> dom (finished s) 
      \<and> \<delta> s v < \<delta> s w
      \<and> \<phi> s w < \<phi> s v "
    using assms
    by (auto elim!: trancl_induct dest: tree_edge_impl_parenthesis)

  lemma nc_reachable_v0_parenthesis:
    assumes C: "\<not> cond s" "\<not> is_break param s"
    and v: "v \<in> reachable" "v \<notin> V0"
    obtains v0 where "v0 \<in> V0"
                 and "\<delta> s v0 < \<delta> s v \<and> \<phi> s v < \<phi> s v0 "
  proof -
    from nc_discovered_eq_reachable[OF C] discovered_iff_tree_path v
    obtain v0 where "v0 \<in> V0" and
      "(v0,v) \<in> (tree s)\<^sup>+" 
      by auto
    moreover with nc_V0_finished[OF C] have "v0 \<in> dom (finished s)" 
      by auto
    ultimately show ?thesis
      using tree_path_impl_parenthesis that[OF `v0 \<in> V0`]
      by simp
  qed

end end

context param_DFS begin context begin interpretation timing_syntax .

  definition paren_imp_tree_reach where
    "paren_imp_tree_reach s \<equiv> \<forall>v \<in> dom (discovered s). \<forall>w \<in> dom (finished s). 
        \<delta> s v < \<delta> s w \<and> (v \<notin> dom (finished s) \<or> \<phi> s v > \<phi> s w) 
               \<longrightarrow> (v,w) \<in> (tree s)\<^sup>+"

  lemma paren_imp_tree_reach:
    "is_invar paren_imp_tree_reach"
    unfolding paren_imp_tree_reach_def[abs_def]
  proof (induct rule: is_invarI)
    case (discover s s' u v) 
    hence EQ[simp]: "tree s' = insert (hd (stack s), v) (tree s)"
                    "finished s' = finished s"
                    "discovered s' = (discovered s)(v \<mapsto> counter s)"
      by simp_all

    from discover interpret DFS_invar where s=s by simp
    from discover have NE[simp]: "stack s \<noteq> []" by (simp add: cond_alt)
    
    show ?case
    proof (intro ballI impI)
      fix a b
      assume F:"a \<in> dom (discovered s')" "b \<in> dom (finished s')"
        and D: "\<delta> s' a < \<delta> s' b \<and> (a \<notin> dom (finished s') \<or> \<phi> s' a > \<phi> s' b)"
        
      from F finished_discovered discover have "b \<noteq> v" by auto
      show "(a,b) \<in> (tree s')\<^sup>+"
      proof (cases "a = v")
        case True with D `b\<noteq>v` have "counter s < \<delta> s b" by simp
        also from F have "b \<in> dom (discovered s)" 
          using finished_discovered by auto
        with timing_less_counter have "\<delta> s b < counter s" by simp
        finally have False .
        thus ?thesis ..
      next
        case False with `b\<noteq>v` F D discover have "(a,b) \<in> (tree s)\<^sup>+" by simp
        thus ?thesis by (auto intro: trancl_mono_mp)
      qed
    qed
  next
    case (finish s s' u) 
    hence EQ[simp]: "tree s' = tree s"
                    "finished s' = (finished s)(hd (stack s) \<mapsto> counter s)"
                    "discovered s' = discovered s"
                    "stack s' = tl (stack s)"
      by simp_all

    from finish interpret DFS_invar where s=s by simp
    from finish have NE[simp]: "stack s \<noteq> []" by (simp add: cond_alt)

    show ?case
    proof (intro ballI impI)
      fix a b
      assume F: "a \<in> dom (discovered s')" "b \<in> dom (finished s')"
        and paren: "\<delta> s' a < \<delta> s' b \<and> (a \<notin> dom (finished s') \<or> \<phi> s' a > \<phi> s' b)"
      hence "a \<noteq> b" by auto
      
      show "(a,b) \<in> (tree s')\<^sup>+"
      proof (cases "b = hd (stack s)")
        case True hence \<phi>b: "\<phi> s' b = counter s" by simp
        have "a \<in> set (stack s)"
          unfolding stack_set_def
        proof
          from F show "a \<in> dom (discovered s)" by simp
          from True `a\<noteq>b` \<phi>b paren have "a \<in> dom (finished s) \<longrightarrow> \<phi> s a > counter s" by simp
          with timing_less_counter show "a \<notin> dom (finished s)" by force
        qed 
        with paren True on_stack_is_tree_path have "(a,b) \<in> (tree s)\<^sup>+" by auto
        thus ?thesis by (auto intro: trancl_mono_mp)
      next
        case False note b_not_hd = this
        show ?thesis
        proof (cases "a = hd (stack s)")
          case False with b_not_hd F paren finish show ?thesis by simp
        next
          case True with paren b_not_hd F have 
            "a \<in> dom (discovered s)" "b \<in> dom (finished s)"  "\<delta> s a < \<delta> s b"  
            by simp_all
          moreover from True stack_not_finished have "a \<notin> dom (finished s)" 
            by simp
          ultimately show ?thesis by (simp add: finish)
        qed
      qed
    qed  
  next
    case (new_root s s' v0) then interpret DFS_invar where s=s by simp
    from new_root finished_discovered have "v0 \<notin> dom (finished s)" by auto
    moreover note timing_less_counter finished_discovered
    ultimately show ?case using new_root by clarsimp force
  qed simp_all
end end

context DFS_invar begin context begin interpretation timing_syntax .

  lemmas s_paren_imp_tree_reach =
    paren_imp_tree_reach[THEN make_invar_thm]
  
  lemma parenthesis_impl_tree_path_not_finished:
    assumes "v \<in> dom (discovered s)"
    and "w \<in> dom (finished s)"
    and "\<delta> s v < \<delta> s w"
    and "v \<notin> dom (finished s)"
    shows "(v,w) \<in> (tree s)\<^sup>+"
    using s_paren_imp_tree_reach assms
    by (auto simp add: paren_imp_tree_reach_def)

  lemma parenthesis_impl_tree_path:
    assumes "v \<in> dom (finished s)" "w \<in> dom (finished s)"
    and "\<delta> s v < \<delta> s w" "\<phi> s v > \<phi> s w"
    shows "(v,w) \<in> (tree s)\<^sup>+"
  proof -
    from assms(1) have "v \<in> dom (discovered s)"
      using finished_discovered by blast
    with assms show ?thesis
      using s_paren_imp_tree_reach assms
      by (auto simp add: paren_imp_tree_reach_def)
  qed

  lemma tree_path_iff_parenthesis:
    assumes "v \<in> dom (finished s)" "w \<in> dom (finished s)"
    shows "(v,w) \<in> (tree s)\<^sup>+ \<longleftrightarrow> \<delta> s v < \<delta> s w \<and> \<phi> s v > \<phi> s w"
    using assms
    by (metis parenthesis_impl_tree_path tree_path_impl_parenthesis) 

  lemma no_pending_succ_impl_path_in_tree:
    assumes v: "v \<in> dom (discovered s)" "pending s `` {v} = {}"
    and w: "w \<in> succ v"
    and \<delta>: "\<delta> s v < \<delta> s w"
    shows "(v,w) \<in> (tree s)\<^sup>+"
  proof (cases "v \<in> dom (finished s)")
    case True 
    with assms assms have "\<delta> s w < \<phi> s v" "w \<in> dom (discovered s)"
      using finished_succ_fin finished_imp_succ_discovered 
      by simp_all
    with True \<delta> show ?thesis
      using parenthesis_contained parenthesis_impl_tree_path
      by blast
  next
    case False
    show ?thesis
    proof (cases "w \<in> dom (finished s)")
      case True with False v \<delta> show ?thesis by (simp add: parenthesis_impl_tree_path_not_finished)
    next
      case False with `v \<notin> dom (finished s)` no_pending_imp_succ_discovered v w have 
        "v \<in> set (stack s)" "w \<in> set (stack s)" 
        by (simp_all add: stack_set_def)
      with on_stack_is_tree_path \<delta> show ?thesis by simp
    qed
  qed

  lemma finished_succ_impl_path_in_tree:
    assumes f:  "v \<in> dom (finished s)"
    and s: "w \<in> succ v"
    and \<delta>: "\<delta> s v < \<delta> s w"
    shows "(v,w) \<in> (tree s)\<^sup>+"
    using no_pending_succ_impl_path_in_tree finished_no_pending finished_discovered
    using assms
    by blast
end end

(* Cross Edges *)
context param_DFS begin context begin interpretation timing_syntax .

  lemma i_cross_edges_finished: "is_invar (\<lambda>s. \<forall>(u,v)\<in>cross_edges s. 
    v \<in> dom (finished s) \<and> (u \<in> dom (finished s) \<longrightarrow> \<phi> s v < \<phi> s u))"
  proof (induction rule: is_invarI_full)
    case (finish s s' u e)
    interpret DFS_invar G param s by fact
    from finish stack_not_finished have "u \<notin> dom (finished s)" by auto
    with finish show ?case by (auto intro: timing_less_counter)
  next
    case (cross_edge s s' u v e)
    interpret DFS_invar G param s by fact
    from cross_edge stack_not_finished have "u \<notin> dom (finished s)" by auto
    with cross_edge show ?case by (auto intro: timing_less_counter)
  qed simp_all

end end

context DFS_invar begin context begin interpretation timing_syntax .
  lemmas cross_edges_finished 
    = i_cross_edges_finished[THEN make_invar_thm]

  lemma cross_edges_target_finished: 
    "(u,v)\<in>cross_edges s \<Longrightarrow> v \<in> dom (finished s)" 
    using cross_edges_finished by auto

  lemma cross_edges_finished_decr: 
    "\<lbrakk>(u,v)\<in>cross_edges s; u\<in>dom (finished s)\<rbrakk> \<Longrightarrow> \<phi> s v < \<phi> s u" 
    using cross_edges_finished by auto

  lemma cross_edge_unequal:
    assumes cross: "(v,w) \<in> cross_edges s"
    shows "v \<noteq> w"
  proof -
    from cross_edges_target_finished[OF cross] have 
      w_fin: "w \<in> dom (finished s)" .

    show ?thesis
    proof (cases "v \<in> dom (finished s)")
      case True with cross_edges_finished_decr[OF cross] 
      show ?thesis by force
    next
      case False with w_fin show ?thesis by force
    qed
  qed 
end end



(* Back Edges *)
context param_DFS begin context begin interpretation timing_syntax .
  
  lemma i_back_edge_impl_tree_path:
    "is_invar (\<lambda>s. \<forall>(v,w) \<in> back_edges s. (w,v) \<in> (tree s)\<^sup>+ \<or> w = v)"
  proof (induct rule: is_invarI_full)
    case (back_edge s s' u v) then interpret DFS_invar where s=s by simp

    from back_edge have st: "v \<in> set (stack s)" "u \<in> set (stack s)"
      using stack_set_def
      by auto
   
    have "(v,u) \<in> (tree s)\<^sup>+ \<or> u = v"
    proof (rule disjCI)
      assume "u \<noteq> v"
      with st back_edge have "v \<in> set (tl (stack s))" by (metis not_hd_in_tl)
      with tl_lt_stack_hd_discover st back_edge have "\<delta> s v < \<delta> s u" by simp
      with on_stack_is_tree_path st show "(v,u) \<in> (tree s)\<^sup>+" by simp
    qed
    with back_edge show ?case by auto
  next
    case discover thus ?case using trancl_sub_insert_trancl by force
  qed simp_all

end end

context DFS_invar begin context begin interpretation timing_syntax .

  lemma back_edge_impl_tree_path:
    "\<lbrakk>(v,w) \<in> back_edges s; v \<noteq> w\<rbrakk> \<Longrightarrow> (w,v) \<in> (tree s)\<^sup>+"
    using i_back_edge_impl_tree_path[THEN make_invar_thm]
    by blast

  lemma back_edge_disc:
    assumes "(v,w) \<in> back_edges s"
    shows "\<delta> s w \<le> \<delta> s v"
  proof cases
    assume "v\<noteq>w" 
    with assms back_edge_impl_tree_path have "(w,v) \<in> (tree s)\<^sup>+" by simp
    with tree_path_disc show ?thesis by force
  qed simp

  lemma back_edges_tree_disjoint:
    "back_edges s \<inter> tree s = {}"
    using back_edge_disc tree_edge_disc
    by force

  lemma back_edges_tree_pathes_disjoint:
    "back_edges s \<inter> (tree s)\<^sup>+ = {}"
    using back_edge_disc tree_path_disc
    by force

  lemma back_edge_finished:
    assumes "(v,w) \<in> back_edges s"
    and "w \<in> dom (finished s)"
    shows "v \<in> dom (finished s) \<and> \<phi> s v \<le> \<phi> s w"
  proof (cases "v=w")
    case True with assms show ?thesis by simp
  next
    case False with back_edge_impl_tree_path assms have "(w,v) \<in> (tree s)\<^sup>+" by simp
    with tree_path_impl_parenthesis assms show ?thesis by fastforce
  qed

end end

context param_DFS begin context begin interpretation timing_syntax .
  (* The lemma should probably replaced by:
     is_invar (\<lambda>s. \<forall>(v,w) \<in> (tree s)\<^sup>+. v \<in> succ w \<longrightarrow> (w,v) \<in> back_edges s \<or> (w,v) \<in> pending s)
  *)
  lemma i_disc_imp_back_edge_or_pending:
    "is_invar (\<lambda>s. \<forall>(v,w) \<in> E. 
        v \<in> dom (discovered s) \<and> w \<in> dom (discovered s)
      \<and> \<delta> s v \<ge> \<delta> s w
      \<and> (w \<in> dom (finished s) \<longrightarrow> v \<in> dom (finished s) \<and> \<phi> s w \<ge> \<phi> s v)
      \<longrightarrow> (v,w) \<in> back_edges s \<or> (v,w) \<in> pending s)"
  proof (induct rule: is_invarI_full)
    case (cross_edge s s' u v) then interpret DFS_invar where s=s by simp
    from cross_edge stack_not_finished[of u] have "u \<notin> dom (finished s)"
      using hd_in_set
      by (auto simp add: cond_alt)
    with cross_edge show ?case by auto
  next
    case (finish s s' u v) then interpret DFS_invar where s=s by simp

    from finish have 
      IH: "\<And>v w. \<lbrakk>w \<in> succ v; v \<in> dom (discovered s); w \<in> dom (discovered s); 
                  \<delta> s w \<le> \<delta> s v;
                  (w \<in> dom (finished s) \<Longrightarrow> v \<in> dom (finished s) \<and> \<phi> s v \<le> \<phi> s w)\<rbrakk>
                 \<Longrightarrow> (v, w) \<in> back_edges s \<or> (v, w) \<in> pending s" 
      by blast

    from finish have ne[simp]: "stack s \<noteq> []"
                 and p[simp]: "pending s `` {hd (stack s)} = {}"
      by (simp_all)

    from hd_in_set[OF ne] have disc: "hd (stack s) \<in> dom (discovered s)" 
                            and not_fin: "hd (stack s) \<notin> dom (finished s)"
      using stack_discovered stack_not_finished
      by blast+

    {
      fix w
      assume w: "w \<in> succ (hd (stack s))" "w \<noteq> hd (stack s)" "w \<in> dom (discovered s)"
        and f: "w \<in> dom (finished s) \<longrightarrow> counter s \<le> \<phi> s w"
        and \<delta>: "\<delta> s w \<le> \<delta> s (hd (stack s))"

      with timing_less_counter have "w \<notin> dom (finished s)" by force
      with finish w \<delta> disc have "(hd (stack s), w) \<in> back_edges s" by blast
    }

    moreover
    {
      fix w
      assume "hd (stack s) \<in> succ w" "w \<noteq> hd (stack s)"
      and "w \<in> dom (finished s)" "\<delta> s (hd (stack s)) \<le> \<delta> s w"
      with IH[of "hd (stack s)" w] disc not_fin have
        "(w, hd (stack s)) \<in> back_edges s"
        using finished_discovered finished_no_pending[of w]
        by blast
    }

    ultimately show ?case
      using finish
      by clarsimp auto
  next
    case (discover s s' u v) then interpret DFS_invar where s=s by simp

    from discover show ?case 
      using timing_less_counter
      by clarsimp fastforce
  next
    case (new_root s s' v0) then interpret DFS_invar where s=s by simp

    from new_root empty_stack_imp_empty_pending have "pending s = {}" by simp
    with new_root show ?case 
      using timing_less_counter
      by clarsimp fastforce
  qed auto
end end

context DFS_invar begin context begin interpretation timing_syntax .

  lemma disc_imp_back_edge_or_pending:
    "\<lbrakk>w \<in> succ v; v \<in> dom (discovered s); w \<in> dom (discovered s); \<delta> s w \<le> \<delta> s v;
      (w \<in> dom (finished s) \<Longrightarrow> v \<in> dom (finished s) \<and> \<phi> s v \<le> \<phi> s w)\<rbrakk>
     \<Longrightarrow> (v, w) \<in> back_edges s \<or> (v, w) \<in> pending s"
    using i_disc_imp_back_edge_or_pending[THEN make_invar_thm]
    by blast

  lemma finished_imp_back_edge:
    "\<lbrakk>w \<in> succ v; v \<in> dom (finished s); w \<in> dom (finished s); 
      \<delta> s w \<le> \<delta> s v; \<phi> s v \<le> \<phi> s w\<rbrakk>
     \<Longrightarrow> (v, w) \<in> back_edges s"
    using disc_imp_back_edge_or_pending finished_discovered finished_no_pending
    by fast

  lemma finished_not_finished_imp_back_edge:
    "\<lbrakk>w \<in> succ v; v \<in> dom (finished s); w \<in> dom (discovered s);
      w \<notin> dom (finished s);
      \<delta> s w \<le> \<delta> s v\<rbrakk>
     \<Longrightarrow> (v, w) \<in> back_edges s"
    using disc_imp_back_edge_or_pending finished_discovered finished_no_pending
    by fast

  lemma finished_self_loop_in_back_edges:
    assumes "v \<in> dom (finished s)"
    and "(v,v) \<in> E"
    shows "(v,v) \<in> back_edges s"
    using assms
    using finished_imp_back_edge
    by blast
end end

(* Back edges and Cycles *)

context DFS_invar begin 

  context begin interpretation timing_syntax .
    (* Cross and tree edges are always acyclic *)
    lemma tree_cross_acyclic:
      "acyclic (tree s \<union> cross_edges s)" (is "acyclic ?E")
    proof (rule ccontr)
      {
        fix u v
        assume *: "u \<in> dom (finished s)" and "(u,v) \<in> ?E\<^sup>+"
        from this(2) have "\<phi> s v < \<phi> s u \<and> v \<in> dom (finished s)"
        proof induct
          case base thus ?case 
            by (metis Un_iff * cross_edges_finished_decr cross_edges_target_finished tree_edge_impl_parenthesis)
        next
          case (step v w)
          hence "\<phi> s w < \<phi> s v \<and> w \<in> dom (finished s)"
            by (metis Un_iff cross_edges_finished_decr cross_edges_target_finished tree_edge_impl_parenthesis)
          with step show ?case by auto
        qed
      } note aux = this

      assume "\<not> acyclic ?E"
      then obtain u where path: "(u,u) \<in> ?E\<^sup>+" by (auto simp add: acyclic_def)
      show False
      proof cases
        assume "u \<in> dom (finished s)" 
        with aux path show False by blast
      next
        assume *: "u \<notin> dom (finished s)"
        moreover 
        from no_loop_in_tree have "(u,u) \<notin> (tree s)\<^sup>+" .
        with trancl_union_outside[OF path] obtain x y where "(u,x) \<in> ?E\<^sup>*" "(x,y) \<in> cross_edges s" "(y,u) \<in> ?E\<^sup>*" by auto
        with cross_edges_target_finished have "y \<in> dom (finished s)" by simp
        moreover with * `(y,u) \<in> ?E\<^sup>*` have "(y,u) \<in> ?E\<^sup>+" by (auto simp add: rtrancl_eq_or_trancl)
        ultimately show False by (metis aux)
      qed
    qed
  end

  lemma cycle_contains_back_edge:
    assumes cycle: "(u,u) \<in> (edges s)\<^sup>+"
    shows "\<exists>v w. (u,v) \<in> (edges s)\<^sup>* \<and> (v,w) \<in> back_edges s \<and> (w,u) \<in> (edges s)\<^sup>*"
  proof -
    from tree_cross_acyclic have "(u,u) \<notin> (tree s \<union> cross_edges s)\<^sup>+" by (simp add: acyclic_def)
    with trancl_union_outside[OF cycle] show ?thesis .
  qed

  lemma cycle_needs_back_edge:
    assumes "back_edges s = {}"
    shows "acyclic (edges s)"
  proof (rule ccontr)
    assume "\<not> acyclic (edges s)"
    then obtain u where "(u,u) \<in> (edges s)\<^sup>+" by (auto simp: acyclic_def)
    with assms have "(u,u) \<in> (tree s \<union> cross_edges s)\<^sup>+" by auto
    with tree_cross_acyclic show False by (simp add: acyclic_def)
  qed

  lemma back_edge_closes_cycle:
    assumes "back_edges s \<noteq> {}"
    shows "\<not> acyclic (edges s)"
  proof -
    from assms obtain v w where be: "(v,w) \<in> back_edges s" by auto
    hence "(w,w) \<in> (edges s)\<^sup>+"
    proof (cases "v=w")
      case False 
      with be back_edge_impl_tree_path have "(w,v) \<in> (tree s)\<^sup>+" by simp
      hence "(w,v) \<in> (edges s)\<^sup>+" by (blast intro: trancl_mono_mp)
      also from be have "(v,w) \<in> edges s" by simp
      finally show ?thesis .
    qed auto
    thus ?thesis by (auto simp add: acyclic_def)
  qed

  lemma back_edge_closes_reachable_cycle:
    "back_edges s \<noteq> {} \<Longrightarrow> \<not> acyclic (E \<inter> reachable \<times> UNIV)"
    by (metis back_edge_closes_cycle edges_ss_reachable_edges cyclic_subset)

  lemma cycle_iff_back_edges:
    "acyclic (edges s) \<longleftrightarrow> back_edges s = {}"
  by (metis back_edge_closes_cycle cycle_needs_back_edge)
end 


context param_DFS begin

  definition "param_dfs_variant \<equiv> inv_image 
    (finite_psupset reachable <*lex*> finite_psubset <*lex*> less_than)
    (\<lambda>s. (dom (discovered s), pending s, length (stack s)))"

  lemma param_dfs_variant_wf[simp, intro!]:
    assumes [simp, intro!]: "finite reachable"
    shows "wf param_dfs_variant"  
    unfolding param_dfs_variant_def
    by auto

  lemma param_dfs_variant_step:   
    assumes A: "DFS_invar G param s" "cond s" "nofail it_dfs"
    shows "step s \<le> SPEC (\<lambda>s'. (s',s)\<in>param_dfs_variant)"
  proof -
    interpret DFS_invar G param s by fact

    from A show ?thesis
      unfolding rwof_eq_DFS_invar[symmetric] it_dfs_def
      apply -
      apply (drule (2) WHILE_nofail_imp_rwof_nofail)
      unfolding step_def get_new_root_def do_defs get_pending_def
      unfolding param_dfs_variant_def
      apply refine_vcg
      using discovered_reachable (* TODO: Clean, up. *) 
        (* FIXME: auto-steps take loooong *)
      apply (auto 
        split: option.splits 
        simp: refine_pw_simps pw_le_iff is_discovered_def finite_psupset_def
      ) [1]
      apply (auto simp: refine_pw_simps pw_le_iff is_empty_stack_def) []
      apply simp_all
      apply (auto 
        simp: refine_pw_simps pw_le_iff is_discovered_def
        split: split_if_asm
        ) [2]
      apply (clarsimp simp: refine_pw_simps pw_le_iff is_discovered_def)
      using discovered_reachable pending_reachable
      apply (auto
        simp: is_discovered_def
        simp: refine_pw_simps pw_le_iff finite_psupset_def
        split: split_if_asm)
      done
  qed

  lemma it_dfsT_eq_it_dfs:
    assumes [simp, intro!]: "finite reachable"
    shows "it_dfsT = it_dfs"
  proof -
    have "it_dfs \<le> it_dfsT" 
      unfolding it_dfs_def it_dfsT_def WHILE_def WHILET_def
      apply (rule bind_mono)
      apply simp
      apply (rule WHILEI_le_WHILEIT)
      done
    also have "it_dfsT \<le> it_dfs"
    proof (cases "nofail it_dfs")
      case False thus ?thesis by (simp add: not_nofail_iff)
    next
      case True

      show ?thesis
        unfolding it_dfsT_def it_dfs_def
        apply (rule bind_mono)
        apply simp
        apply (subst WHILET_eq_WHILE_tproof[
          where I="DFS_invar G param"
          and V="param_dfs_variant"
          ])
        apply auto []

        apply (subst rwof_eq_DFS_invar[symmetric])
        using rwof_init[OF True[unfolded it_dfs_def]]
        apply (fastforce dest: order_trans) []
        
        apply (rule SPEC_rule_conjI)
          apply (rule DFS_invar_step[OF True], assumption+) []
          apply (rule param_dfs_variant_step, (assumption|rule True)+) []
        
        apply simp
        done
    qed
    finally show ?thesis by simp
  qed
  
end

(* Locale that assumes that parameterization does not fail *)
locale DFS =
  param_DFS G param
  for G :: "('v, 'more) graph_rec_scheme"
  and param :: "('v,'es) parameterization"
  +
  assumes nofail_on_init: 
    "nofail (on_init param)"

  assumes nofail_on_new_root:
    "pre_on_new_root v0 s \<Longrightarrow> nofail (on_new_root param v0 s)"

  assumes nofail_on_finish: 
    "pre_on_finish u s \<Longrightarrow> nofail (on_finish param u s)"

  assumes nofail_on_cross_edge: 
    "pre_on_cross_edge u v s \<Longrightarrow> nofail (on_cross_edge param u v s)"

  assumes nofail_on_back_edge: 
    "pre_on_back_edge u v s \<Longrightarrow> nofail (on_back_edge param u v s)"

  assumes nofail_on_discover: 
    "pre_on_discover u v s \<Longrightarrow> nofail (on_discover param u v s)"

begin
  lemmas nofails = nofail_on_init nofail_on_new_root nofail_on_finish 
    nofail_on_cross_edge nofail_on_back_edge nofail_on_discover


  lemma init_leof_invar: "init \<le>\<^sub>n SPEC (DFS_invar G param)"
    unfolding rwof_eq_DFS_invar[symmetric]
    by (rule rwof_leof_init)

  lemma it_dfs_eq_spec: "it_dfs = SPEC (\<lambda>s. DFS_invar G param s \<and> \<not>cond s)"
    unfolding rwof_eq_DFS_invar[symmetric] it_dfs_def
    apply (rule nofail_WHILE_eq_rwof)
    apply (subst WHILE_eq_I_rwof)
    unfolding rwof_eq_DFS_invar
    apply (rule SPEC_nofail[where \<Phi>="\<lambda>_. True"])
    apply (refine_vcg leofD[OF _ init_leof_invar, THEN weaken_SPEC])
    apply (simp add: init_def refine_pw_simps nofail_on_init)
    apply (rule DFS_invar_step')
    apply (simp add: step_def refine_pw_simps nofail_on_init do_defs 
      get_pending_def get_new_root_def pred_defs
      split: option.split)
    apply (intro allI conjI impI nofails)
    apply (auto simp add: pre_on_defs)
    done

  lemma it_dfs_correct: "it_dfs \<le> SPEC (\<lambda>s. DFS_invar G param s \<and> \<not>cond s)"
    by (simp add: it_dfs_eq_spec)

  lemma it_dfs_SPEC:
    assumes "\<And>s. \<lbrakk>DFS_invar G param s; \<not>cond s\<rbrakk> \<Longrightarrow> P s"
    shows "it_dfs \<le> SPEC P"
    using weaken_SPEC[OF it_dfs_correct]
    using assms
    by blast

  lemma it_dfsT_correct: 
    assumes "finite reachable"
    shows "it_dfsT \<le> SPEC (\<lambda>s. DFS_invar G param s \<and> \<not>cond s)"
    apply (subst it_dfsT_eq_it_dfs[OF assms])
    by (rule it_dfs_correct)

  lemma it_dfsT_SPEC:
    assumes "finite reachable"
    assumes "\<And>s. \<lbrakk>DFS_invar G param s; \<not>cond s\<rbrakk> \<Longrightarrow> P s"
    shows "it_dfsT \<le> SPEC P"
    apply (subst it_dfsT_eq_it_dfs[OF assms(1)])
    using assms(2)
    by (rule it_dfs_SPEC)

context begin interpretation timing_syntax .

  definition white_path where
    "white_path s x y \<equiv> x\<noteq>y 
       \<longrightarrow> (\<exists>p. path E x p y \<and> 
               (\<delta> s x < \<delta> s y \<and> (\<forall> v \<in> set (tl p). \<delta> s x < \<delta> s v)))"

  lemma white_path:
    "it_dfs \<le> SPEC(\<lambda>s. \<forall>x \<in> reachable. \<forall>y \<in> reachable. \<not> is_break param s \<longrightarrow> 
            white_path s x y \<longleftrightarrow> (x,y) \<in> (tree s)\<^sup>*)"
  proof (rule it_dfs_SPEC, intro ballI impI)
    fix s x y
    assume DI: "DFS_invar G param s"
      and C: "\<not> cond s" "\<not> is_break param s"
      and reach: "x \<in> reachable" "y \<in> reachable"

    from DI interpret DFS_invar where s=s .

    note fin_eq_reach = nc_finished_eq_reachable[OF C]

    show "white_path s x y \<longleftrightarrow> (x,y) \<in> (tree s)\<^sup>*"
    proof (cases "x=y")
      case True thus ?thesis by (simp add: white_path_def)
    next
      case False
      
      show ?thesis
      proof
        assume "(x,y) \<in> (tree s)\<^sup>*"
        with `x\<noteq>y` have T: "(x,y) \<in> (tree s)\<^sup>+" by (metis rtranclD)
        then obtain p where P: "path (tree s) x p y" by (metis trancl_is_path)
        with tree_ssE have "path E x p y" using path_mono[where E="tree s"] 
          by simp
        moreover 
        from P have "\<delta> s x < \<delta> s y \<and> (\<forall> v \<in> set (tl p). \<delta> s x < \<delta> s v)"
          using `x\<noteq>y`
        proof (induct rule: path_tl_induct)
          case (single u) thus ?case by (fact tree_edge_disc)
        next
          case (step u v) note `\<delta> s x < \<delta> s u`
          also from step have "\<delta> s u < \<delta> s v" by (metis tree_edge_disc)
          finally show ?case .
        qed
        ultimately show "white_path s x y"
          by (auto simp add: `x\<noteq>y` white_path_def)
      next
        assume "white_path s x y"
        with `x\<noteq>y` obtain p where 
            P:"path E x p y" and
            white: "\<delta> s x < \<delta> s y \<and> (\<forall> v \<in> set (tl p). \<delta> s x < \<delta> s v)"
          unfolding white_path_def
          by blast
        hence "p \<noteq> []" by auto
        thus "(x,y) \<in> (tree s)\<^sup>*" using P white reach(2)
        proof (induction p arbitrary: y rule: rev_nonempty_induct)
          case single hence "y \<in> succ x" by (simp add: path_cons_conv)
          with reach single show ?case 
            using fin_eq_reach finished_succ_impl_path_in_tree[of x y] 
            by simp
        next
          case (snoc u us) hence "path E x us u" by (simp add: path_append_conv)
          moreover hence "(x,u) \<in> E\<^sup>*" by (simp add: path_is_rtrancl)
          with reach have ureach: "u \<in> reachable" 
            by (metis rtrancl_image_advance_rtrancl) 
          moreover from snoc have "\<delta> s x < \<delta> s u" "(\<forall>v\<in>set (tl us). \<delta> s x < \<delta> s v)" 
            by simp_all
          ultimately have x_u: "(x,u) \<in> (tree s)\<^sup>*" by (metis snoc.IH)
            
          from snoc have "y \<in> succ u" by (simp add: path_append_conv)
          from snoc(5) fin_eq_reach finished_discovered have 
            y_f_d: "y \<in> dom (finished s)" "y \<in> dom (discovered s)"
            by auto
            
          from `y \<in> succ u` ureach fin_eq_reach have "\<delta> s y < \<phi> s u" 
            using finished_succ_fin by simp
          also from `\<delta> s x < \<delta> s u` have "x \<noteq> u" by auto
          with x_u have "(x,u) \<in> (tree s)\<^sup>+" by (metis rtrancl_eq_or_trancl)
          with fin_eq_reach reach have "\<phi> s u < \<phi> s x" 
            using tree_path_impl_parenthesis
            by simp
          finally have "\<phi> s y < \<phi> s x" 
            using reach fin_eq_reach y_f_d snoc
            using parenthesis_contained
            by blast
          hence "(x,y) \<in> (tree s)\<^sup>+"
            using reach fin_eq_reach y_f_d snoc
            using parenthesis_impl_tree_path
            by blast
          thus ?case by auto
        qed
      qed
    qed
  qed
end end
end
