theory Stress_Test
imports 
  "../../Refine_Dflt" 
  "../../GenCF/Impl/Impl_Uv_Set"
  "Lib/Code_Target_ICF"
begin
  -- "Simple linear congruence generator for (low-quality) random numbers"

  definition "lcg_next s r = (s mod r, ((22695477::nat)*s + 1) mod 268435456)"

  definition
    stress_test1 :: "nat \<Rightarrow> nat \<Rightarrow> (nat \<times> nat \<times> nat set) nres"
    where
    "stress_test1 m n \<equiv> 
    WHILET (\<lambda>(i,_). i<n) (\<lambda>(i,s,X). do {
      let (x,s) = lcg_next s m;
      let X=insert x X;
      let (x,s) = lcg_next s m;
      (*let X=X-{x};*)
      let (x,s) = lcg_next s m;
      (*let b=x\<in>X;*)
      RETURN (i+1,s,X)
    }) (0::nat,1,{})"

  lemma [autoref_rules,param]: 
    "(lcg_next,lcg_next)\<in>nat_rel \<rightarrow> nat_rel \<rightarrow> nat_rel \<times>\<^sub>r nat_rel"
    unfolding lcg_next_def[abs_def]
    by parametricity

  thm TYRELI

  term "iam_map_rel"

  schematic_lemma st_rs:
    fixes m n
    notes [autoref_rules] = IdI[of m] IdI[of n]
    notes [autoref_tyrel] = TYRELI[where R="\<langle>nat_rel\<rangle>dflt_rs_rel"]
    shows "(RETURN (?c::?'c),stress_test1 m n)\<in>?R"
    unfolding stress_test1_def
    apply (autoref_monadic)
    done
  concrete_definition st_rs for m n uses st_rs
  export_code st_rs checking SML

  schematic_lemma st_ias:
    fixes m n
    notes [autoref_rules] = IdI[of m] IdI[of n]
    notes [autoref_tyrel] = TYRELI[where R="\<langle>nat_rel\<rangle>map2set_rel iam_map_rel"]
    shows "(RETURN (?c::?'c),stress_test1 m n)\<in>?R"
    unfolding stress_test1_def
    apply (autoref_monadic)
    done
  concrete_definition st_ias for m n uses st_ias
  export_code st_ias checking SML

  schematic_lemma st_bs:
    fixes m n
    notes [autoref_rules] = IdI[of m] IdI[of n]
    notes [autoref_tyrel] = TYRELI[where R="\<langle>nat_rel\<rangle>bs_set_rel"]
    shows "(RETURN (?c::?'c),stress_test1 m n)\<in>?R"
    unfolding stress_test1_def
    apply (autoref_monadic)
    done
  concrete_definition st_bs for m n uses st_bs
  export_code st_bs checking SML


  definition "rs_empty \<equiv> \<lambda>_::unit. rbt.Empty::(nat,unit)rbt"
  thm st_rs_def
  definition rs_insert :: "nat \<Rightarrow> _" 
    where "rs_insert \<equiv> map2set_insert rbt_insert"

  schematic_lemma rs_union: 
    "(?f,op \<union>) 
      \<in> \<langle>nat_rel\<rangle>dflt_rs_rel \<rightarrow> \<langle>nat_rel\<rangle>dflt_rs_rel \<rightarrow> \<langle>nat_rel\<rangle>dflt_rs_rel"
    by autoref
  concrete_definition rs_union uses rs_union

  schematic_lemma rs_eq: 
    "(?f,op =) 
      \<in> \<langle>nat_rel\<rangle>dflt_rs_rel \<rightarrow> \<langle>nat_rel\<rangle>dflt_rs_rel \<rightarrow> bool_rel"
    by autoref
  concrete_definition rs_eq uses rs_eq

  thm st_ias
  definition "ias_empty \<equiv> (iam_empty :: _ \<Rightarrow> unit option array)"

  definition ias_insert :: "nat \<Rightarrow> _" 
    where "ias_insert \<equiv> map2set_insert iam_update"

  schematic_lemma ias_union: 
    "(?f,op \<union>) 
      \<in> \<langle>nat_rel\<rangle>map2set_rel iam_map_rel
      \<rightarrow> \<langle>nat_rel\<rangle>map2set_rel iam_map_rel 
      \<rightarrow> \<langle>nat_rel\<rangle>map2set_rel iam_map_rel"
    by autoref
  concrete_definition ias_union uses ias_union

  schematic_lemma ias_eq: 
    "(?f,op =) 
      \<in> \<langle>nat_rel\<rangle>map2set_rel iam_map_rel 
    \<rightarrow> \<langle>nat_rel\<rangle>map2set_rel iam_map_rel \<rightarrow> bool_rel"
    by autoref
  concrete_definition ias_eq uses ias_eq

  thm st_rs_def

  export_code 
    uv_union uv_equal uv_set_bit uv_empty in SML

  export_code st_bs st_rs st_ias 
    bs_union bs_eq bs_insert bs_empty
    rs_union rs_eq rs_insert rs_empty
    ias_union ias_eq ias_insert ias_empty
    uv_union uv_equal uv_set_bit uv_empty
    nat_of_integer integer_of_nat
    in SML
    module_name Stress_Test
    file "code/Stress_Test.sml"


  ML {*
    fun test f m n = let 
      val m' = (@{code nat_of_integer} m);
      val n' = (@{code nat_of_integer} n);
      val t = Time.now ()
      val _ = f m' n'
      val t = Time.- (Time.now (),t)
    in 
      writeln ((IntInf.toString m) ^ " " ^ (IntInf.toString n) 
        ^ " " ^ (IntInf.toString (Time.toMilliseconds t)));
      (m,n,t) 
    end

    val width = 31
    val stepsize = 10000
    val stepnum = 10
    val f = @{code st_bs}

    fun rpt n = 
      if n <= stepnum then (
        test f width (n*stepsize);
        rpt (n+1)
      ) else ()

    fun rpt2 0 = ()
      | rpt2 n = let 
          val a = IntInf.<<(1,Word.fromInt 20)
          val b = IntInf.orb(37547632,a)
        in rpt2 (n-1) end


    (*val _ = PolyML.profiling 1*)
      val t = Time.now ()
      val _ = rpt2 10000000
      val t = Time.- (Time.now (),t)
    (*val m' = (@{code nat_of_integer} 20);
    val n' = (@{code nat_of_integer} 100000);
    val _ = test @{code st_rs} 20 10000*)
    (*val _ = PolyML.profiling 0*)
      
  *}


end
