open Stress_Test

val lcg = ref 63763

fun lcg_next r = (
  lcg := (22695477*(!lcg) + 1) mod 268435456;
  !lcg mod r
)

fun repeat name f s0 (n:int) = let
  fun r s 0 = ()
    | r s n = (r (f s n) (n-1))

  val t = Time.now ()
  val _ = r s0 n
  val t = Time.- (Time.now (),t)
in
  print (
    name ^ ": " ^ (IntInf.toString n) ^ " " ^ (IntInf.toString (Time.toMilliseconds t)) ^ "\n"
    )
end

fun orepeat f n0 w nn =
  if n0>nn then ()
  else (f n0; orepeat f (n0+w) w nn)


type ('s) ops = {
  empty : unit -> 's,
  insert : nat -> 's -> 's,
  union : 's -> 's -> 's,
  eq : 's -> 's -> bool
}

val uv_ops = {
  empty = fn () => uv_empty,
  insert = uv_set_bit,
  union = uv_union,
  eq = uv_equal
}

val bs_ops = {
  empty = bs_empty,
  insert = bs_insert,
  union = bs_union,
  eq = bs_eq
}


fun do_test name (ops:'s ops) w n itn  = let

  fun build_set 0 = #empty ops ()
    | build_set n = let
        val s = build_set (n-1)
        val x = lcg_next w
        val x = nat_of_integer x
        val s = #insert ops x s
      in
        s
      end

  val set1 = build_set n
  val set2 = build_set n
  val set3 = build_set n

  fun basic_bit_ops s n = let
    val s = #union ops s set1
    val s = if #eq ops s set2 then s else #union ops s set3
  in
    s
  end

  val name = name ^ " w=" ^ IntInf.toString w ^ ";"
  
in
  repeat name basic_bit_ops (#empty ops ()) itn
end

fun do_tests name (ops:'s ops) w n it0 its itn = orepeat (do_test name ops w n) it0 its itn

fun main () = let
(*
  val _ = do_tests "bs 10" bs_ops 10 10 100000 100000 1000000
  val _ = do_tests "bs 50" bs_ops 50 50 100000 100000 1000000
  val _ = do_tests "bs 100" bs_ops 100 100 100000 100000 1000000
  val _ = do_tests "bs 500" bs_ops 500 500 100000 100000 1000000
  val _ = do_tests "uv 10" uv_ops 10 10 100000 100000 1000000
  val _ = do_tests "uv 50" uv_ops 50 50 100000 100000 1000000
  val _ = do_tests "uv 100" uv_ops 100 100 100000 100000 1000000
  val _ = do_tests "uv 500" uv_ops 500 500 100000 100000 1000000
*)

  val _ = orepeat (fn w => do_test "bs_w" bs_ops w w 100000) 100 100 5000
  val _ = orepeat (fn w => do_test "uv_w" uv_ops w w 100000) 100 100 5000

in
  ()
end

val _ = if MLton.isMLton then main() else ()



(*



fun test fname f m n = let
  val m' = nat_of_integer m;
  val n' = nat_of_integer n;
  val t = Time.now ()
  val _ = f m' n'
  val t = Time.- (Time.now (),t)
in
  print (fname ^ ": " ^ (IntInf.toString m) ^ " " ^ (IntInf.toString n)
    ^ " " ^ (IntInf.toString (Time.toMilliseconds t)) ^ "\n");
  (m,n,t)
end

local
  val step_size = 30
  val step_num = 20

  fun rpt fname f m n =
    if m>step_num then ()
    else (
      test fname f (m*step_size) n;
      rpt fname f (m+1) n
    )
  
in
  fun rpt_test fname f n = rpt fname f 1 n
end

fun repeat name f s0 (n:int) = let
  fun r s 0 = ()
    | r s n = (r (f s n) (n-1))

  val t = Time.now ()
  val _ = r s0 n
  val t = Time.- (Time.now (),t)
in
  print (
    name ^ ": " ^ (IntInf.toString n) ^ " " ^ (IntInf.toString (Time.toMilliseconds t)) ^ "\n"
    )
end

    val maxWord = IntInf.pow (2, Word.wordSize);

fun foo_t x n =
    if n < maxWord then
      IntInf.orb (x, IntInf.<< (1, Word.fromInt n))
    else raise (Fail ("Bit index too large: " ^ IntInf.toString n))
    
fun foo_f x n =
    if n < maxWord then
      IntInf.andb (x, IntInf.notb (IntInf.<< (1, Word.fromInt n)))
    else raise (Fail ("Bit index too large: " ^ IntInf.toString n));

(*  
fun foo x n = if n < maxWord then
    IntInf.orb (x, IntInf.<< (1, Word.fromInt n))
    else raise (Fail ("Bit index too large: " ^ IntInf.toString n))
    *)

(*fun bs_insert x s = foo_t s (integer_of_nat x)*)
    
structure T = IntInf
val fromInt = T.fromInt

val lcg = ref 63763

fun lcg_next r = (
  lcg := (22695477*(!lcg) + 1) mod 268435456;
  !lcg mod r
)

val uv_ops = {
  empty = fn () => uv_empty,
  insert = uv_set_bit,
  union = uv_union,
  eq = uv_equal
}

val bs_ops = {
  empty = bs_empty,
  insert = bs_insert,
  union = bs_union,
  eq = bs_eq
}


val ops = bs_ops (* Operations *)
val w = 7000       (* Width of sets *)
val n = w        (* Number of random elements inserted *)
val its = 500000 (* Number of iterations *)


val empty = #empty ops
val insert = #insert ops
val union = #union ops
val eq = #eq ops

fun build_set w 0 = empty ()
  | build_set w n = let
      val s = build_set w (n-1)
      val x = lcg_next w
      val x = nat_of_integer x
      val s = insert x s
    in
      s
    end

    
val set1 = build_set w n
val set2 = build_set w n
val set3 = build_set w n
    
   
val basic_init = set1
fun basic_bit_ops s n = let
  val s = union s set1

  val s = if eq s set2 then s else union s set3

  (*val s = bs_insert (nat_of_integer 12) s*)
  (*val s = map2set_insert iam_update (nat_of_integer 12) s*)
  (*val x = IntInf.orb (10,IntInf.<<(1,Word.fromInt 21))*)
  (*val x = set_bit_integer 0 (nat_of_integer 12) true*)
  (*val x = Bits_Integer.set_bit 0 12 true*)

  (*val s = T.orb (s, T.fromInt n)*)
  (*val s = Word.orb (s, Word.<< (Word.fromInt 1, Word.fromInt 12))*)
  
  (*val x = 0
  val n = 12
  val b = true
  
  val x = foo_t x n*)
in
  s
end  


fun main () = let
  (*val _ = rpt_test "bitvector" st_bs 100000
  val _ = rpt_test "array" st_ias 100000
  val _ = rpt_test "rbt" st_rs 100000*)
  
  (*val _ = print ("Word size: " ^ IntInf.toString (Word.wordSize) ^ "\n");*)
  
  
  val _ = repeat "bbops" basic_bit_ops basic_init its
in
  ()
end

val _ = if MLton.isMLton then main() else ()

(*val _ = PolyML.profiling 1;
val _ = main ()
val _ = PolyML.profiling 0;
*)
*)