theory Scratch
imports "../Refine_Dflt"
begin

  (* TODO: Move *)
  lemma rtrancl_over_node:
    assumes A: "B\<subseteq>A"
    assumes P: "(u,v)\<in>(E \<inter> (A\<times>UNIV))\<^sup>*"
    assumes NP: "(u,v)\<notin>(E \<inter> (B\<times>UNIV))\<^sup>*"
    obtains vh 
    where "vh\<in>A-B"
    and "(u,vh)\<in>(E \<inter> (A\<times>UNIV))\<^sup>*"
    and "(vh,v)\<in>(E \<inter> (A\<times>UNIV))\<^sup>*"
    using P NP
  proof induct
    case base hence False by simp thus ?case ..
  next
    case (step v w)
    
    show ?case 
    proof (cases "v\<in>B")
      case True with step.hyps(1,2) step.prems(2)
      have "(u,v)\<notin>(E \<inter> B \<times> UNIV)\<^sup>*"
        by (metis (mono_tags) IntD1 IntI SigmaI UNIV_I rtrancl.simps)
      thus thesis
        apply (rule step.hyps(3)[rotated])
        apply (erule (1) step.prems(1))
        using step.hyps(2)
        by (metis rtrancl.rtrancl_into_rtrancl) 
    next
      case False
      with step.hyps(2) have "v\<in>A-B" by auto
      thus thesis
        apply (rule step.prems(1))
        apply (rule step.hyps(1))
        by (metis r_into_rtrancl step.hyps(2))
    qed
  qed



  typedecl node


  record state = 
    onstack :: "node set"
    visited :: "node set"
    idx :: int
    ndidx :: "node \<Rightarrow> int"
    low_node :: "node \<Rightarrow> node"

locale dfs = 
  fixes E :: "(node\<times>node) set"
  fixes A :: "node set"
  fixes v0 :: "node"
  assumes fin_branching[simp]: "\<forall>v. finite (E``{v})"
begin

  abbreviation reach (infix "\<rightarrow>*" 60) where "reach u v \<equiv> (u,v)\<in>E\<^sup>*"
  abbreviation reachp (infix "\<rightarrow>+" 60) where "reachp u v \<equiv> (u,v)\<in>E\<^sup>+"

  definition non_empty where "non_empty \<equiv> \<exists>v\<in>A. v0\<rightarrow>*v \<and> v\<rightarrow>+v"

  definition s_invar :: "state \<Rightarrow> node \<Rightarrow> bool"
    where "s_invar s v \<equiv> 
      (\<forall>w\<in>visited s. v0 \<rightarrow>* w)
    \<and> v0 \<rightarrow>* v
    \<and> (\<forall>w\<in>visited s. w \<rightarrow>* low_node s w)
    \<and> (\<forall>w\<in>visited s. low_node s w\<in>visited s)
    \<and> (\<forall>w\<in>onstack s. w \<rightarrow>* v)
    \<and> onstack s \<subseteq> visited s
    \<and> (\<forall>w\<in>visited s. 0 \<le> ndidx s w \<and> ndidx s w < idx s)
    \<and> (0 \<le> idx s)
    \<and> (\<forall>u\<in>onstack s. \<forall>w\<in>onstack s. ndidx s u \<le> ndidx s w \<longrightarrow> u\<rightarrow>*w)
    \<and> (E``(visited s - onstack s) \<subseteq> visited s)
    \<and> (\<forall>u\<in>visited s - onstack s. \<forall>w\<in>onstack s. 
        (u,w)\<in>( E \<inter> (visited s - onstack s)\<times>UNIV)\<^sup>*
       \<longrightarrow> low_node s u \<in> onstack s \<and> ndidx s (low_node s u) \<le> ndidx s w)
    "

  definition pre :: "state \<times> node \<times> int \<Rightarrow> bool" where 
    "pre svi \<equiv> case svi of (s,v,acc_idx) \<Rightarrow> 
      s_invar s v
    \<and> v\<notin>visited s
    \<and> (acc_idx\<noteq>-1 \<longrightarrow> (\<exists>w\<in>onstack s\<inter>A. acc_idx = ndidx s w))
    "
  definition post :: "state \<times> node \<times> int \<Rightarrow> state \<times> bool \<Rightarrow> bool" where
    "post svi sb \<equiv> let (s0,v,acc_idx)=svi; (s,brk) = sb in 
      (\<not>brk \<longrightarrow>
        s_invar s v 
      \<and> v\<in>visited s - onstack s
      \<and> onstack s = onstack s0 
      \<and> visited s \<supseteq> visited s0 
      \<and> (\<forall>v\<in>visited s0. ndidx s v = ndidx s0 v \<and> low_node s v = low_node s0 v)
      \<and> (\<forall>u\<in>visited s - visited s0. 
           (v,u) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*)
      ) 
      \<and>
      (brk \<longrightarrow> non_empty)
    "
  definition fe_invar 
    :: "state \<Rightarrow> node \<Rightarrow> int \<Rightarrow> node set \<Rightarrow> (state \<times> bool) \<Rightarrow> bool"
    where "fe_invar s0 v acc_idx it sxbrk \<equiv> let (s,brk)=sxbrk in 
      (\<not>brk \<longrightarrow>
          s_invar s v
        \<and> v\<in>visited s
        \<and> v\<notin>visited s0
        \<and> onstack s = insert v (onstack s0)
        \<and> (acc_idx\<noteq>-1 \<longrightarrow> (\<exists>w\<in>onstack s\<inter>A. acc_idx = ndidx s w))
        \<and> visited s \<supseteq> visited s0
        \<and> (\<forall>u\<in>visited s0. ndidx s u = ndidx s0 u 
           \<and> low_node s u = low_node s0 u)
        \<and> (\<forall>u\<in>visited s - visited s0. 
            (v,u) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*)
        \<and> (E``{v} - it) \<subseteq> visited s
        \<and> (low_node s v \<in> onstack s)
        \<and> (\<forall>w\<in>onstack s. \<forall>u\<in>E``{v}-it-onstack s. 
            (u,w)\<in>( E \<inter> (visited s - onstack s)\<times>UNIV)\<^sup>*
            \<longrightarrow> ndidx s (low_node s v) \<le> ndidx s w
          )
    )
    \<and> (brk \<longrightarrow> non_empty)
    "

  definition initial_state :: state where
    "initial_state \<equiv> \<lparr> 
        onstack={}, 
        visited={}, 
        idx=0, 
        ndidx = (\<lambda>x. undefined), 
        low_node=(\<lambda>x. undefined)
      \<rparr>"

  definition enter_node :: "state \<Rightarrow> node \<Rightarrow> int \<Rightarrow> state \<times> int"
    where
    "enter_node s v acc_idx \<equiv> let
      acc_idx = (if v\<in>A then idx s else acc_idx);
      s = s\<lparr>
        onstack := insert v (onstack s),
        visited := insert v (visited s),
        ndidx := (ndidx s) (v := idx s),
        low_node := (low_node s) (v := v),
        idx := idx s + 1
      \<rparr>
    in (s,acc_idx)"

  definition k_edge :: "state \<Rightarrow> int \<Rightarrow> node \<Rightarrow> node \<Rightarrow> state \<times> bool"
  where
  "k_edge s acc_idx v w \<equiv> 
    if low_node s w \<in> onstack s then
      if ndidx s (low_node s w) \<le> acc_idx then
        (s,True)
      else if ndidx s (low_node s w) < ndidx s (low_node s v) then
        (s\<lparr> low_node := (low_node s) (v := low_node s w)\<rparr>, False)
      else
        (s,False)
    else
      (s,False)"

  definition upd_after_rec :: "state \<Rightarrow> bool \<Rightarrow> node \<Rightarrow> node \<Rightarrow> state\<times>bool" 
    where
    "upd_after_rec s brk v w \<equiv> 
    if \<not>brk 
      \<and> low_node s w \<in> onstack s 
      \<and> ndidx s (low_node s w) < ndidx s (low_node s v) 
    then
      (s\<lparr> low_node := (low_node s) (v := low_node s w)\<rparr>, False)
    else 
      (s,brk)"

  definition result :: "state \<times> bool \<Rightarrow> bool"
    where "result s \<equiv> let (_,brk) = s in brk"

  definition tjs where
    "tjs \<equiv> do {
      s \<leftarrow> REC (\<lambda>D (s0,v,acc_idx0). do {

        let (s,acc_idx) = enter_node s0 v acc_idx0;

        (s,brk) \<leftarrow> FOREACHci (fe_invar s0 v acc_idx) (E``{v}) 
        (\<lambda>(_,brk). \<not> brk) (\<lambda>w (s,_). do {
          if w \<in> visited s then
            RETURN (k_edge s acc_idx v w)
          else do {
            (s,brk) \<leftarrow> D (s,w,acc_idx);
            RETURN (upd_after_rec s brk v w)
          }
        }) (s,False);

        RETURN (s\<lparr> onstack := onstack s - {v} \<rparr>,brk)

      }) (initial_state,v0,-1);

      RETURN (result s)
    }"

  lemma dfs_rule:
    assumes vc0: (* Initial state *)
      "pre (initial_state,v0,-1)"
    assumes vc1: (* Enter new node *)
    "\<And>s0 v acc_idx0 s acc_idx. \<lbrakk> 
        pre (s0,v,acc_idx0); 
        enter_node s0 v acc_idx0 = (s,acc_idx)   
      \<rbrakk> \<Longrightarrow> fe_invar s0 v acc_idx (E``{v}) (s,False)"
    assumes vc2: (* Edge to known node *)
    "\<And>w it v s acc_idx s' brk s0 acc_idx0. \<lbrakk> 
       w\<in>it; it\<subseteq>E``{v}; w\<in>visited s; 
       pre (s0,v,acc_idx0);
       fe_invar s0 v acc_idx it (s,False);
       k_edge s acc_idx v w = (s',brk)
    \<rbrakk>
     \<Longrightarrow> fe_invar s0 v acc_idx (it-{w}) (s',brk)"

    assumes vc3: (* Recursive call for unkown node *)
    "\<And>w it v s acc_idx s0 acc_idx0. \<lbrakk> 
       w\<in>it; it\<subseteq>E``{v}; w\<notin>visited s; 
       pre (s0,v,acc_idx0);
       fe_invar s0 v acc_idx it (s,False)
    \<rbrakk> \<Longrightarrow> pre (s,w,acc_idx)"

    assumes vc4: (* Return from recursive call *)
    "\<And>w it v s acc_idx s' brk s0 acc_idx0. \<lbrakk>
       w\<in>it; it\<subseteq>E``{v}; 
       pre (s0,v,acc_idx0);
       fe_invar s0 v acc_idx it (s,False); 
       pre (s,w,acc_idx); 
       post (s,w,acc_idx) (s',brk)
     \<rbrakk> \<Longrightarrow> fe_invar s0 v acc_idx (it - {w}) (upd_after_rec s' brk v w)"

    assumes vc5: (* Pop node from stack (no break) *)
      "\<And>v acc_idx s s0 acc_idx0.
        \<lbrakk> pre (s0,v,acc_idx0);
          fe_invar s0 v acc_idx ({}::node set) (s,False) \<rbrakk> 
       \<Longrightarrow> post (s0, v, acc_idx0) 
             (s\<lparr> onstack := onstack s - {v} \<rparr>,False)"

    assumes vc6: (* Pop node from stack (break) *)
    "\<And>it v acc_idx s s0 acc_idx0.
      \<lbrakk> it \<subseteq> E``{v}; 
        pre (s0,v,acc_idx0);
        fe_invar s0 v acc_idx it (s, True) 
      \<rbrakk> 
    \<Longrightarrow> post (s0, v, acc_idx0) (s\<lparr> onstack := onstack s - {v}\<rparr>, True)"

    assumes vc7: (* Compute result *)
      "\<And>s brk. \<lbrakk> post (initial_state,v0,-1) (s,brk) \<rbrakk> \<Longrightarrow> \<Psi> (result (s,brk))"

    shows "tjs \<le> SPEC \<Psi>"
    unfolding tjs_def
    apply (refine_rcg refine_vcg)
    apply (rule order_trans)
    apply (rule REC_rule[where \<Phi>=pre and M="\<lambda>svi. SPEC (post svi)"])
    apply refine_mono
    apply (rule vc0)
    apply (refine_rcg refine_vcg FOREACHci_rule')
    apply simp

    apply simp
    apply (rule vc1, simp_all) []

    apply simp
    apply (case_tac "k_edge ac bb aa xa", simp)
    apply (rule vc2, simp_all) []

    apply simp
    apply (frule vc3, simp_all) []
    apply (rule order_trans)
    apply (rprems)
    apply assumption
    apply (rule SPEC_rule)
    apply (case_tac xb, simp) []
    apply (rule vc4, simp_all) []

    apply simp
    apply (erule (1) vc5)

    apply simp
    apply (erule (2) vc6)

    apply (intro refine_vcg)
    apply (case_tac x, simp)
    apply (erule vc7)
    done


   
  lemma "tjs \<le> SPEC (\<lambda>r. r \<longrightarrow> non_empty)"
  proof (rule dfs_rule)
    case goal1 -- "Initialization"
    show ?case
      by (simp add: pre_def s_invar_def initial_state_def)
  next
    case goal2 -- "Entering recursion for new node"
    assume PRE: "pre (s0,v,acc_idx0)"
    assume ENTER: "enter_node s0 v acc_idx0 = (s,acc_idx)"

    thm enter_node_def
    from ENTER have E_STEP[simp]:
      "onstack s = insert v (onstack s0)"
      "visited s = insert v (visited s0)"
      "ndidx s = (ndidx s0) (v := idx s0)"
      "low_node s = (low_node s0) (v := v)"
      "idx s = idx s0 + 1"
      "acc_idx = (if v\<in>A then idx s0 else acc_idx0)"
      unfolding enter_node_def by auto

    from PRE have 1: 
      "s_invar s0 v" 
      "v\<notin>visited s0"
      "(acc_idx0\<noteq>-1 \<longrightarrow> (\<exists>w\<in>onstack s0\<inter>A. acc_idx0 = ndidx s0 w))"
      by (simp_all add: pre_def)
    hence 2: "v\<notin>onstack s0" unfolding s_invar_def by blast

    have FIN_EQ: "visited s - onstack s = visited s0 - onstack s0" 
      using 1 by auto

    have I_low_node: "\<forall>u\<in>visited s - onstack s. \<forall>w\<in>onstack s.
         (u, w) \<in> (E \<inter> (visited s - onstack s) \<times> UNIV)\<^sup>* \<longrightarrow>
         low_node s u \<in> onstack s \<and> ndidx s (low_node s u) \<le> ndidx s w"
        (is "\<forall>u\<in>_. \<forall>w\<in>_. _ \<longrightarrow> ?G u w")
    proof (intro ballI impI)
      fix u w
      assume "u\<in>visited s - onstack s"
      also have VMO_EQ: "visited s - onstack s = visited s0 - onstack s0" 
        using `v\<notin>visited s0` by auto
      finally have A: "u\<in>visited s0 - onstack s0" .
      hence U_ne_V[simp]: "v\<noteq>u" using `v\<notin>visited s0` by auto
      hence [simp]: "u\<noteq>v" ..

      assume B: "w\<in>onstack s"

      assume "(u, w) \<in> (E \<inter> (visited s - onstack s) \<times> UNIV)\<^sup>*"
      hence UW_PATH: "(u, w) \<in> (E \<inter> (visited s0 - onstack s0) \<times> UNIV)\<^sup>*"
        by (simp only: VMO_EQ)

      have [simp]: "low_node s0 u \<noteq> v"
      proof -
        from 1[unfolded s_invar_def] A have "low_node s0 u \<in> visited s0"
          by auto
        with 1 show ?thesis by auto
      qed

      show "?G u w"
      proof (cases "w=v")
        case False[simp] -- "Original path still valid"
        from B have "w\<in>onstack s0" by auto
        with 1[unfolded s_invar_def] UW_PATH A have C:
          "low_node s0 u \<in> onstack s0 \<and> ndidx s0 (low_node s0 u) \<le> ndidx s0 w"
          by simp
        thus "?G u w" by simp
      next
        case True[simp] -- {*Path is not empty, thus there is a last edge from
          a finished node to @{text "v"} already in @{text "s0"}. 
          However, this contradicts the fact that @{text "v"} has just been
          discovered *}
        from UW_PATH obtain vf where 
          "vf\<in>visited s0 - onstack s0"
          and "(vf,v)\<in>E" 
          by (cases rule: rtrancl.cases) auto
        with 1[unfolded s_invar_def] have False by blast
        thus ?thesis ..
      qed
    qed

    have "s_invar s v" 
      using 1 2 unfolding s_invar_def
      apply (intro conjI)
      apply auto []
      apply auto []
      apply simp
      apply simp
      apply auto []
      apply (simp, blast)
      apply fastforce []
      apply simp
      apply (simp, fastforce)
      apply fastforce
      apply (rule I_low_node)
      done
    moreover have 
      "onstack s = insert v (onstack s0)"
      "(acc_idx\<noteq>-1 \<longrightarrow> (\<exists>w\<in>onstack s\<inter>A. acc_idx = ndidx s w))"
      "visited s \<supseteq> visited s0"
      using 1 2
      by (auto simp add: split: split_if_asm)
    ultimately show 
      "fe_invar s0 v acc_idx (E``{v}) (s,False)"
      by (auto simp add: fe_invar_def `v\<notin>visited s0`)

  next
    case goal3 -- "Discover edge to already known node"
    assume C: "w\<in>it" "it \<subseteq> E``{v}" "w\<in>visited s" 
    assume FE: "fe_invar s0 v acc_idx it (s,False)"
    assume K_EDGE: "k_edge s acc_idx v w = (s',brk)"

    from K_EDGE have [simp]:
      "visited s' = visited s"
      "onstack s' = onstack s"
      "idx s' = idx s"
      "ndidx s' = ndidx s"
      unfolding k_edge_def by (auto split: split_if_asm)

    from FE have 
      S_INV: "s_invar s v" and
      VVIS: "v\<in>visited s" and
      VNVIS: "v\<notin>visited s0" and
      OS: "onstack s = insert v (onstack s0)" and
      ACC_IDX: "(acc_idx\<noteq>-1 \<longrightarrow> (\<exists>w\<in>onstack s\<inter>A. acc_idx = ndidx s w))" and
      VIS: "visited s \<supseteq> visited s0" and
      NDIDX: "(\<forall>u\<in>visited s0. ndidx s u = ndidx s0 u 
        \<and> low_node s u = low_node s0 u )" and
      NNR: "\<forall>u\<in>visited s - visited s0. 
        (v,u) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*" and
      SV: "E``{v} - it \<subseteq> visited s" and
      LN_OS: "low_node s v \<in> onstack s" and
      LN: "(\<forall>w\<in>onstack s. \<forall>u\<in>E``{v}-it-onstack s. 
            (u,w)\<in>( E \<inter> (visited s - onstack s)\<times>UNIV)\<^sup>*
            \<longrightarrow> ndidx s (low_node s v) \<le> ndidx s w
          )"
      unfolding fe_invar_def
      apply -
      apply auto [10]
      unfolding Let_def split
      apply (simp (no_asm_use) )
      apply (elim conjE)
      apply assumption
      done

    from C have "(v,w)\<in>E" by auto

    from S_INV have 
      "v\<in>visited s" 
      unfolding s_invar_def by (auto simp: OS)


    have INV': "s_invar s' v"
      using S_INV
      unfolding s_invar_def
    proof (intro conjI)
      (* Low-Node reachable *)
      from S_INV have "\<forall>v\<in>visited s. v \<rightarrow>* low_node s v" 
        unfolding s_invar_def by auto
      moreover from S_INV have "v \<rightarrow>* low_node s w"
        unfolding s_invar_def using `w\<in>visited s` `(v,w)\<in>E`
        by auto
      ultimately show "\<forall>w\<in>visited s'. w \<rightarrow>* low_node s' w"
        using K_EDGE[symmetric]
        unfolding k_edge_def
        by (auto split: split_if_asm)
    next
      (* Low-Node points to lowest node on stack --- actually,
        relevant low-node pointers are not changed in this step! *)
      show "\<forall>u\<in>visited s' - onstack s'. \<forall>w\<in>onstack s'.
          (u, w) \<in> (E \<inter> (visited s' - onstack s') \<times> UNIV)\<^sup>* \<longrightarrow>
          low_node s' u \<in> onstack s' \<and> ndidx s' (low_node s' u) \<le> ndidx s' w"
      proof
        case goal1
        assume A: "u\<in>visited s' - onstack s'"
        hence "u \<noteq> v" by (simp add: OS)
        hence "low_node s' u = low_node s u"
          using K_EDGE[symmetric] unfolding k_edge_def
          by (simp split: split_if_asm)
        with A S_INV show ?case
          unfolding s_invar_def by auto
      qed
    next
      show "\<forall>u\<in>visited s'. low_node s' u \<in> visited s'"
      proof
        fix u
        from K_EDGE[symmetric] have 
          "low_node s' u = low_node s u \<or> low_node s' u = low_node s w"
          unfolding k_edge_def
          by (auto split: split_if_asm)
        moreover
        assume "u\<in>visited s'"
        with S_INV `w\<in>visited s` 
        have "low_node s u \<in> visited s \<and> low_node s w \<in> visited s"
          unfolding s_invar_def by auto
        ultimately show "low_node s' u \<in> visited s'" by auto
      qed
    qed simp_all

    have LN_OS': "low_node s' v \<in> onstack s'"
      using K_EDGE[symmetric] LN_OS
      unfolding k_edge_def
      by (auto split: split_if_asm)

    have LN': "\<not> brk \<Longrightarrow>
      \<forall>wa\<in>onstack s'.
         \<forall>u\<in>E `` {v} - (it - {w}) - onstack s'.
          (u, wa) \<in> (E \<inter> (visited s' - onstack s') \<times> UNIV)\<^sup>* \<longrightarrow>
          ndidx s' (low_node s' v) \<le> ndidx s' wa"
    proof (intro ballI impI)
      fix wa u
      assume [simp]: "\<not>brk"
      assume LN1: "wa\<in>onstack s'"
      assume LN2: "u\<in>E `` {v} - (it - {w}) - onstack s'"
      assume LN3: "(u, wa) \<in> (E \<inter> (visited s' - onstack s') \<times> UNIV)\<^sup>*"
      
      show "ndidx s' (low_node s' v) \<le> ndidx s' wa"
      proof (cases "u=w")
        case False with LN2 have "u\<in>E``{v} - it - onstack s'"
          by blast
        with LN LN1 LN3 have 
          SV_IDX: "ndidx s (low_node s v) \<le> ndidx s wa"
          by auto
        
        show ?thesis proof (cases "low_node s w \<in> onstack s 
          \<and> ndidx s (low_node s w) < ndidx s (low_node s v)")
          case True hence [simp]: "low_node s' v = low_node s w"
            using K_EDGE[symmetric] unfolding k_edge_def 
            by (auto split: split_if_asm)
          from True SV_IDX show ?thesis by simp
        next
          case False hence [simp]: "low_node s' v = low_node s v"
            using K_EDGE[symmetric] unfolding k_edge_def 
            by (auto split: split_if_asm)
          show ?thesis
            by (simp add: SV_IDX)
        qed
      next
        case True[simp]
        with LN2 have "w\<notin>onstack s" by simp
        with C have "w\<in>visited s - onstack s" by auto
        with S_INV LN1 LN3 have ILN: "low_node s w \<in> onstack s" 
          "ndidx s (low_node s w) \<le> ndidx s wa"
          unfolding s_invar_def by auto

        show ?thesis proof (cases "low_node s w \<in> onstack s 
          \<and> ndidx s (low_node s w) < ndidx s (low_node s v)")
          case True hence [simp]: "low_node s' v = low_node s w"
            using K_EDGE[symmetric] unfolding k_edge_def 
            by (auto split: split_if_asm)
          show ?thesis by (simp add: ILN)
        next
          case False hence [simp]: "low_node s' v = low_node s v"
            using K_EDGE[symmetric] unfolding k_edge_def 
            by (auto split: split_if_asm)

          from False ILN have "ndidx s (low_node s w) \<ge> ndidx s (low_node s v)"
            by auto
          with ILN have "ndidx s (low_node s v) \<le> ndidx s wa" by auto

          thus ?thesis by (simp)
        qed
      qed
    qed

    (* Argument that breaking here is sound *)
    {
      assume brk
      with K_EDGE[symmetric] have 
        1: "low_node s w \<in> onstack s" and
        2: "ndidx s (low_node s w) \<le> acc_idx"
        unfolding k_edge_def
        by (auto split: split_if_asm)
      with S_INV have "low_node s w \<in> visited s" unfolding s_invar_def by auto
      with S_INV have "0 \<le> ndidx s (low_node s w)"
         unfolding s_invar_def by auto
       with 2 ACC_IDX obtain u where 
         "acc_idx = ndidx s u" "u\<in>onstack s" "u\<in>A"
         by auto
       with S_INV 1 2 have P1: "low_node s w \<rightarrow>* u"
         unfolding s_invar_def by auto
       from S_INV `w\<in>visited s` have P2: "w\<rightarrow>*low_node s w"
         unfolding s_invar_def by auto
       from `u\<in>onstack s` S_INV have P3: "u\<rightarrow>*v" 
         unfolding s_invar_def by auto
       note P3 also note `(v,w)\<in>E` also (rtrancl_into_trancl1) note P2 
       also note P1 finally have "u\<rightarrow>+u" .
       moreover from `u\<in>onstack s` S_INV have "u\<in>visited s"
         unfolding s_invar_def by auto
       with S_INV have "v0\<rightarrow>*u"
         unfolding s_invar_def by auto
       ultimately have "non_empty" 
         using `u\<in>A`
         unfolding non_empty_def by blast

    } note BRK=this

    from K_EDGE[symmetric] have NDIDX': 
      "\<forall>u\<in>visited s0. low_node s' u = low_node s0 u"
      unfolding k_edge_def
      using NDIDX VNVIS
      by (auto split: split_if_asm)

    show "fe_invar s0 v acc_idx (it - {w}) (s', brk)" 
      unfolding fe_invar_def
      apply (simp only: Let_def split, intro conjI impI)
      apply (rule INV')
      apply (simp, rule VVIS)
      apply (rule VNVIS)
      apply (simp, rule OS)
      apply simp using ACC_IDX apply blast
      apply (simp, rule VIS)
      using NDIDX NDIDX' apply simp
      using NNR apply auto []
      using `w\<in>visited s` SV apply auto []
      apply (rule LN_OS')
      apply (erule LN')
      apply (erule BRK)
      done
  next
    case goal4 -- "Discover edge to unknown node: 
      Establish precondition for recursion"
    assume C: "w\<in>it" "it \<subseteq> E``{v}" "w\<notin>visited s" 
    assume FE: "fe_invar s0 v acc_idx it (s,False)"
    
    from FE have 
      S_INV: "s_invar s v" and
      ACC_IDX: "(acc_idx\<noteq>-1 \<longrightarrow> (\<exists>w\<in>onstack s\<inter>A. acc_idx = ndidx s w))"
      unfolding fe_invar_def by auto
    from S_INV have "s_invar s w" unfolding s_invar_def
      using subsetD[OF `it \<subseteq> _` `w\<in>it`] 
      by auto
    thus "pre (s, w, acc_idx)" unfolding pre_def using `w\<notin>visited s` ACC_IDX
      by simp

  next
    case goal5 -- "Discover edge to unknown node: 
      Establish foreach-invariant from post-condition"
    
    assume C: "w\<in>it" "it \<subseteq> E``{v}"
    assume FE: "fe_invar s0 v acc_idx it (s, False)"
    assume PRE0: "pre (s0,v,acc_idx0)"
    assume PRE: "pre (s, w, acc_idx)"
    assume POST: "post (s, w, acc_idx) (s', brk)"

    from C have E: "(v,w)\<in>E" by auto

    from FE have 
      INV: "s_invar s v" and
      VVIS: "v\<in>visited s" and
      VNVIS: "v\<notin>visited s0" and
      OS: "onstack s = insert v (onstack s0)" and
      ACC_IDX: "(acc_idx\<noteq>-1 \<longrightarrow> (\<exists>w\<in>onstack s\<inter>A. acc_idx = ndidx s w))" and
      VIS: "visited s \<supseteq> visited s0" and
      NDIDX: "(\<forall>u\<in>visited s0. ndidx s u = ndidx s0 u 
        \<and> low_node s u = low_node s0 u)" and
      NNR: "\<forall>u\<in>visited s - visited s0. 
        (v,u) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*" and
      SV: "E``{v} - it \<subseteq> visited s" and
      LN_OS: "low_node s v \<in> onstack s" and
      LN: "(\<forall>w\<in>onstack s. \<forall>u\<in>E``{v}-it-onstack s. 
            (u,w)\<in>( E \<inter> (visited s - onstack s)\<times>UNIV)\<^sup>*
            \<longrightarrow> ndidx s (low_node s v) \<le> ndidx s w
          )"
      unfolding fe_invar_def by auto


    from PRE have "w\<notin>visited s" unfolding pre_def by auto

    from INV have "onstack s \<subseteq> visited s" unfolding s_invar_def by auto

    let ?upd_cond = "low_node s' w \<in> onstack s' 
      \<and> ndidx s' (low_node s' w) < ndidx s' (low_node s' v)"

    obtain s'' where 
      UPD_FMT: "upd_after_rec s' False v w = (s'',False)"
      "s'' = s'\<lparr> 
        low_node := 
          if ?upd_cond
          then (low_node s')(v := low_node s' w)
          else low_node s'
      \<rparr>"
      unfolding upd_after_rec_def
      by fastforce

    hence [simp]: 
      "onstack s'' = onstack s'"
      "ndidx s'' = ndidx s'"
      "visited s'' = visited s'"
      "idx s'' = idx s'"
      by auto


    show "fe_invar s0 v acc_idx (it - {w}) (upd_after_rec s' brk v w)" 
    proof (cases brk)
      case False[simp]
      from POST have 
        "s_invar s' w" and 
        [simp]: "onstack s' = onstack s" and
        NDIDX': "\<forall>u\<in>visited s. ndidx s' u = ndidx s u 
          \<and> low_node s' u = low_node s u" and
        VIS': "visited s' \<supseteq> visited s" and
        WVIS: "w\<in>visited s' - onstack s'" and
        NNR': "\<forall>u\<in>visited s' - visited s. 
          (w,u) \<in> (E \<inter> (visited s' - onstack s) \<times> UNIV)\<^sup>*"
        unfolding post_def by auto
      with INV have 
        INV': "s_invar s' v"
        unfolding s_invar_def
        by auto
  
      have INV'': "s_invar s'' v"
      proof (cases ?upd_cond)
        case False hence [simp]: "low_node s'' = low_node s'"
          by (auto simp add: UPD_FMT(2))
        from INV' show ?thesis
          unfolding s_invar_def
          by simp_all
      next
        case True
        hence [simp]: "low_node s'' = (low_node s')(v := low_node s' w)"
          by (simp add: UPD_FMT(2))
          
        from INV' show ?thesis
          unfolding s_invar_def
        proof (elim conjE, intro conjI)
          assume A: "\<forall>w\<in>visited s'. w \<rightarrow>* low_node s' w"
          show "\<forall>w\<in>visited s''. w \<rightarrow>* low_node s'' w"
          proof -
            note E
            also from INV' WVIS have "w \<rightarrow>* low_node s' w"
              unfolding s_invar_def by auto
            finally show ?thesis
              by (auto intro: A[rule_format])
          qed
        next
          assume A: "\<forall>w\<in>visited s'. low_node s' w \<in> visited s'"
          show "\<forall>w\<in>visited s''. low_node s'' w \<in> visited s''"
          proof -
            from INV' WVIS have "low_node s' w \<in> visited s'"
              unfolding s_invar_def by auto
            thus ?thesis
              by (auto intro: A[rule_format])
          qed
        next
          assume A3: "(\<forall>u\<in>visited s' - onstack s'.
            \<forall>w\<in>onstack s'.
              (u, w) \<in> (E \<inter> (visited s' - onstack s') \<times> UNIV)\<^sup>* \<longrightarrow>
              low_node s' u \<in> onstack s' \<and>
              ndidx s' (low_node s' u) \<le> ndidx s' w)"
  
            (* Low-Node points to lowest node on stack --- actually,
            relevant low-node pointers are not changed in this step! *)
          show "\<forall>u\<in>visited s'' - onstack s''. \<forall>w\<in>onstack s''.
              (u, w) \<in> (E \<inter> (visited s'' - onstack s'') \<times> UNIV)\<^sup>* \<longrightarrow>
              low_node s'' u \<in> onstack s'' 
              \<and> ndidx s'' (low_node s'' u) \<le> ndidx s'' w"
          proof
            case goal1
            assume A: "u\<in>visited s'' - onstack s''"
            hence "u \<noteq> v" by (simp add: OS)
            hence "low_node s'' u = low_node s' u" by simp
            with A INV' show ?case
              unfolding s_invar_def by auto
          qed
        qed simp_all
      qed
        
      have LN_OS'': "low_node s'' v \<in> onstack s"
      proof (cases ?upd_cond)
        case True thus ?thesis
          by (simp add: UPD_FMT(2))
      next
        case False hence [simp]: "low_node s'' = low_node s'"
          by (auto simp add: UPD_FMT(2))

        from POST have "low_node s' v = low_node s v"
          using VVIS unfolding post_def
          by auto
        with LN_OS show ?thesis by simp
      qed

      
      have LN': "
        \<forall>wa\<in>onstack s.
          \<forall>u\<in>E `` {v} - (it - {w}) - onstack s.
             (u, wa) \<in> (E \<inter> (visited s' - onstack s) \<times> UNIV)\<^sup>* \<longrightarrow>
             ndidx s' (low_node s'' v) \<le> ndidx s wa"
      proof (intro ballI impI)
        fix wa u
        assume LN1: "wa\<in>onstack s"
        assume LN2: "u\<in>E `` {v} - (it - {w}) - onstack s"
        assume LN3: "(u, wa) \<in> (E \<inter> (visited s' - onstack s) \<times> UNIV)\<^sup>*"

        show "ndidx s' (low_node s'' v) \<le> ndidx s wa"
        proof (cases "(u, wa) \<in> (E \<inter> (visited s - onstack s) \<times> UNIV)\<^sup>*")
          case True -- "Path goes only over nodes that had been 
            discovered before recursive call"
  
          have "ndidx s' (low_node s'' v) \<le> ndidx s' (low_node s' v)"
            by (auto simp: UPD_FMT(2))
          also from VVIS NDIDX' have "low_node s' v = low_node s v" by auto
          also from LN_OS `onstack s \<subseteq> visited s` NDIDX' have 
            "ndidx s' (low_node s v) = ndidx s (low_node s v)" by auto
          also have "ndidx s (low_node s v) \<le> ndidx s wa" 
          proof -
            from LN1 LN2 have "u\<noteq>wa" by auto
            with True have "u\<in>visited s - onstack s" 
              by (cases rule: converse_rtranclE) auto
            with `w\<notin>visited s` have "u\<noteq>w" by auto
            with LN2 have "u\<in>E``{v} - it - onstack s"
              by blast
            with LN LN1 True show ?thesis
              by auto
          qed
          finally show ?thesis .
        next
          case False -- "Path goes over node that has been newly discovered
            in recursive call"

          have VO_SS: "visited s - onstack s \<subseteq> visited s' - onstack s"
            using VIS' by auto

          from rtrancl_over_node[OF VO_SS LN3 False] obtain va where
            "va\<in>visited s' - visited s - onstack s"
            "(va,wa)\<in>(E \<inter> (visited s' - onstack s) \<times> UNIV)\<^sup>*"
            by blast

          -- {* Thus, there is also a path over @{text "w"} *}
          with NNR' have 
            W_PATH: "(w,wa)\<in>(E \<inter> (visited s' - onstack s) \<times> UNIV)\<^sup>*"
            by (auto dest: rtrancl_trans)
          with W_PATH INV'[unfolded s_invar_def] WVIS LN1 have 
            LN_OS: "low_node s' w \<in> onstack s'" and
            LN_LE: "ndidx s' (low_node s' w) \<le> ndidx s' wa" 
            by auto

          
          from LN_OS have 
            "ndidx s' (low_node s'' v) \<le> ndidx s' (low_node s' w)"
            by (auto simp: UPD_FMT(2))
          also note LN_LE
          also from INV have "onstack s \<subseteq> visited s"
            unfolding s_invar_def by auto
          with NDIDX' LN1 VIS' have "ndidx s' wa = ndidx s wa" by auto
          finally show ?thesis .
        qed
      qed

      have NNR'': "\<forall>u\<in>visited s' - visited s0.
          (v, u) \<in> (E \<inter> (visited s' - onstack s0) \<times> UNIV)\<^sup>*"
      proof 
        fix u
        assume A: "u \<in> visited s' - visited s0"
        show "(v, u) \<in> (E \<inter> (visited s' - onstack s0) \<times> UNIV)\<^sup>*" 
        proof (cases "u\<in>visited s")
          case True with A have "u\<in>visited s - visited s0" by auto
          with NNR have "(v, u) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*" ..
          thus ?thesis using VIS'
            apply -
            apply (erule set_rev_mp[OF _ rtrancl_mono])
            apply auto
            done
        next
          case False with A have "u\<in>visited s' - visited s" by auto
          with NNR' have "(w, u) \<in> (E \<inter> (visited s' - onstack s) \<times> UNIV)\<^sup>*" ..
          hence "(w, u) \<in> (E \<inter> (visited s' - onstack s0) \<times> UNIV)\<^sup>*"
            apply (rule set_rev_mp[OF _ rtrancl_mono])
            apply (auto simp: OS)
            done
          also have "(v,w)\<in>(E \<inter> (visited s' - onstack s0) \<times> UNIV)"
          proof -
            from PRE0 have "v\<notin>onstack s0"
              unfolding pre_def s_invar_def by auto
            with E VIS' VVIS show ?thesis by auto
          qed
          finally (converse_rtrancl_into_rtrancl[rotated])
          show ?thesis .
        qed
      qed

      note INV''
      moreover from NDIDX' have "\<forall>u\<in>onstack s. ndidx s' u = ndidx s u"
        using INV unfolding s_invar_def by auto 
      ultimately 
      show "fe_invar s0 v acc_idx (it - {w}) (upd_after_rec s' brk v w)" 
        unfolding fe_invar_def 
        apply (simp add: UPD_FMT(1) del: ball_simps)
        apply (intro conjI)

        using OS ACC_IDX set_mp[OF VIS' VVIS] VNVIS
        apply (simp_all) [4]

        using order_trans[OF VIS VIS'] apply simp

        using NDIDX VIS VIS' NDIDX' VNVIS apply (auto simp: UPD_FMT(2)) []
        
        apply (rule NNR'')
  
        using SV WVIS VIS' apply auto []
        
        apply (rule LN_OS'')
  
        apply (rule LN')
        done
    next
      case True[simp]
      from POST have "non_empty"
        unfolding post_def by auto
      thus "fe_invar s0 v acc_idx (it - {w}) (upd_after_rec s' brk v w)" 
        unfolding fe_invar_def 
        by (simp add: upd_after_rec_def)
    qed
 
  next
    case goal6 -- "Pop finished node from stack, no break"
    let ?s'= "s\<lparr>onstack := onstack s - {v}\<rparr>"

    assume PRE: "pre (s0, v, acc_idx0)"
    assume FE: "fe_invar s0 v acc_idx {} (s, False)"
    from FE have 
      INV: "s_invar s v" and
      VVIS: "v\<in>visited s" and
      OS: "onstack s = insert v (onstack s0)" and
      NDIDX: "\<forall>u\<in>visited s0. ndidx s u = ndidx s0 u 
        \<and> low_node s u = low_node s0 u" and
      NNR: "(\<forall>u\<in>visited s - visited s0. 
            (v,u) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*)" and
      VIS': "visited s \<supseteq> visited s0" and
      SV: "E``{v} \<subseteq> visited s" and
      LN: "(\<forall>w\<in>onstack s.
          \<forall>u\<in>E `` {v} - onstack s.
             (u, w) \<in> (E \<inter> (visited s - onstack s) \<times> UNIV)\<^sup>* \<longrightarrow>
             ndidx s (low_node s v) \<le> ndidx s w)"
      unfolding fe_invar_def by auto

    moreover

    from PRE have "v\<notin>onstack s0" and "v\<notin>visited s0" 
      and "onstack s0 \<subseteq> visited s0"
      unfolding pre_def s_invar_def by auto
    hence OS0V: "onstack s0 - {v} = onstack s0" by simp

    have ILN': "\<forall>u\<in>visited s - onstack s0.
          \<forall>w\<in>onstack s0.
             (u, w) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>* \<longrightarrow>
             low_node s u \<in> onstack s0 \<and>
             low_node s u \<noteq> v \<and> ndidx s (low_node s u) \<le> ndidx s w"
    proof (intro ballI impI)
      fix u w
      assume 
        LN1: "u \<in> visited s - onstack s0" and
        LN2: "w \<in> onstack s0" and
        LN3: "(u, w) \<in> (E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*"

      show "low_node s u \<in> onstack s0 
            \<and> low_node s u \<noteq> v \<and> ndidx s (low_node s u) \<le> ndidx s w"
      proof (cases "u=v")
        case False with LN1 LN2 have 
          "u \<in> visited s - onstack s"
          "w\<in>onstack s"
          by (auto simp: OS)
        with LN2 LN3 INV
        \<longrightarrow> Argh: This is a real counter-example.
        When we pop a node from stack, we must consider paths over this
          node for low-node correctness, but low-node pointers of already 
          visited nodes are not updated.

        Can perhaps be fixed by transitively following low-node pointers.
        Complexity can perhaps be saved by path-compression.



      proof (cases "(u, w) \<in> (E \<inter> (visited s0 - onstack s0) \<times> UNIV)\<^sup>*")
        case True -- "Path goes only over nodes that had been 
          discovered before recursive call"
  
        from LN1 LN2 have "u\<noteq>w" by auto
        with True have "u\<in>visited s0 - onstack s0"
          by (cases rule: converse_rtranclE) auto

        with True LN2 PRE have 
          G1: "low_node s0 u \<in> onstack s0" and
          G2: "ndidx s0 (low_node s0 u) \<le> ndidx s0 w"
          unfolding pre_def s_invar_def
          by auto
        with `v\<notin>onstack s0` have G3: "low_node s0 u \<noteq> v" by auto
        
        from 
          `u\<in>visited s0 - onstack s0` 
          set_rev_mp[OF LN2 `onstack s0 \<subseteq> visited s0`] 
          set_rev_mp[OF `low_node s0 u \<in> onstack s0` `onstack s0 \<subseteq> visited s0`]
        have
          "low_node s0 u = low_node s u" 
          "ndidx s0 w = ndidx s w"
          "ndidx s0 (low_node s u) = ndidx s (low_node s u)"
          using NDIDX by auto
        
        with G1 G2 G3 show ?thesis by simp
      next
        case False -- "Path goes over node that has been newly discovered
          in recursive call"

        have VO_SS: "visited s0 - onstack s0 \<subseteq> visited s - onstack s0"
          using VIS' by auto

        from rtrancl_over_node[OF VO_SS LN3 False] obtain va where
          "va\<in>visited s - visited s0 - onstack s0"
          "(va,w)\<in>(E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*"
          by blast

        txt {* Thus, there is also a path over @{text "v"} *}
        with NNR have 
          W_PATH: "(v,w)\<in>(E \<inter> (visited s - onstack s0) \<times> UNIV)\<^sup>*"
          by (auto dest: rtrancl_trans)
        

        with INV[unfolded s_invar_def] VVIS LN1 have 
          LN_OS: "low_node s' w \<in> onstack s'" and
          LN_LE: "ndidx s' (low_node s' w) \<le> ndidx s' wa" 
          by auto

          
        from LN_OS have 
          "ndidx s' (low_node s'' v) \<le> ndidx s' (low_node s' w)"
          by (auto simp: UPD_FMT(2))
        also note LN_LE
        also from INV have "onstack s \<subseteq> visited s"
          unfolding s_invar_def by auto
        with NDIDX' LN1 VIS' have "ndidx s' wa = ndidx s wa" by auto
        finally show ?thesis .
      qed
    qed



      
      show "low_node s u \<in> onstack s0 
            \<and> low_node s u \<noteq> v \<and> ndidx s (low_node s u) \<le> ndidx s w"
      proof (cases "u=v")
        case False with UVIS have "u\<in>visited s - (onstack s0 - {v})"
        
      


    qed
*)

    ultimately have INV': "s_invar ?s' v" unfolding s_invar_def 
      apply (intro conjI)
      apply auto [9]
      apply (simp,fast)
      apply (simp add: OS0V)
      done
;
\<longrightarrow> need to encode the low-node-invar to also consider paths starting
  at current node, going over already finished successor

      apply (simp, fast)
      apply auto []

      done

    from PRE have 
      "v \<notin> onstack s0"
      unfolding pre_def s_invar_def by auto

    with INV' show "post (s0, v, acc_idx0) (?s', False)"
      unfolding post_def 
      by (simp add: OS VIS' NDIDX VVIS)

  next
    case goal7
    let ?s'= "s\<lparr>onstack := onstack s - {v}\<rparr>"

    assume C: "it \<subseteq> E``{v}"
    assume PRE: "pre (s0, v, acc_idx0)"
    assume FE: "fe_invar s0 v acc_idx it (s, True)"

    from FE have "non_empty"
      unfolding fe_invar_def by auto
    thus "post (s0, v, acc_idx0) (?s', True)"
      unfolding post_def by simp
  next
    case goal8
    assume "post (initial_state, v0, -1) (s, brk)"
    hence "brk \<longrightarrow> non_empty"
      unfolding post_def by auto

    thus "result (s,brk) \<longrightarrow> non_empty"
      unfolding result_def by auto
  qed




          
; ---------------  
  
          show "\<forall>u\<in>visited s'' - onstack s''.
            \<forall>w\<in>onstack s''.
             (u, w) \<in> (E \<inter> (visited s'' - onstack s'') \<times> UNIV)\<^sup>* \<longrightarrow>
             low_node s'' u \<in> onstack s'' \<and>
             ndidx s'' (low_node s'' u) \<le> ndidx s'' w"
          proof (intro ballI impI)
            fix u wa
            assume LN1: "u\<in>visited s'' - onstack s''"
            assume LN2: "wa\<in>onstack s''"
            assume LN3: "(u, wa) \<in> (E \<inter> (visited s'' - onstack s'') \<times> UNIV)\<^sup>*"
            show "low_node s'' u \<in> onstack s'' \<and>  
              ndidx s'' (low_node s'' u) \<le> ndidx s'' wa"
            proof (cases "v=u")
              case False
              thus ?thesis using A3 LN1 LN2 LN3
                by simp
            next
              case True[simp]
  
              from bspec[OF A3 WVIS, THEN bspec, of wa] LN2 LN3
              have 
                "low_node s' w \<in> onstack s' \<and> 
                ndidx s' (low_node s' w) \<le> ndidx s' wa"
                apply auto
  
              show ?thesis using A3 WVIS LN2 LN3 apply simp apply auto
                apply force
  


          apply_end simp_all
          \<longrightarrow> ctd here
            

          proof (cases ?upd_cond)
            case False with A1 A2 show ?thesis1 
            by (auto simp add: UPD_FMT(2))
        next
          case True 
          hence [simp]: "low_node s'' = (low_node s')(v := low_node s' w)"
            by (simp add: UPD_FMT(2))

          note E
          also from INV' WVIS have "w \<rightarrow>* low_node s' w"
            unfolding s_invar_def by auto
          finally show ?thesis 
            by (auto intro: A[rule_format])


;

\<longrightarrow> That's soundness.
  Now towards completeness!

Try next:
* If there is a path to node on stack, low_node points to smallest such node
  (post, i.e. for finished nodes)

* On return from node, there is no path to the stack over non-stack nodes
  that leads upon an accepting node
* Low-Node points to the lowest node on stack reachable from that node over
  non-stack nodes


The informal argument goes as follows:

  Assume there is an accepting lasso, and the algorithm first visits node u
  of the lasso at time i. Afterwards, it will visit the accepting node v
  at time j\<ge>i. 
  Case 1: At time j, the remaining nodes of the lasso have not yet been visited
    The nodes will be visited while exploring v, with an acc_idx\<ge>j. 
    When exploring a node with a back-edge to u, we will find that 
    low_node u is on stack [after discovery of u, it may only be set to nodes on stack above u] and 
    its index is smaller than i [This is a trivial invariant].

    Thus, we will report a lasso

  Case 2: At time j, some remaining node of the lasso has already been visited.
    The low-node of the visited nodes will point to a node on stack above u. 
    Thus, when reaching the already visited node, we report the lasso



    
;
Try that:
  The nodes on the stack form a path with ascending node indices,
      and the current node at the end. Or: Each higher indexed node on stack 
      is reachable from lower-indexed node


\<longrightarrow> ctd here: For a lasso to exists,
  we have to argue that accepting state is below low_node on stack.
  That's correctness of acc_idx, i.e., acc_idx points to the index of 
  an accepting node on stack.
  And increasing of stack.
    

    


  record state =
    idx :: nat
    ndidx :: "node \<rightharpoonup> nat"
    lowlink :: "node \<rightharpoonup> nat"
    S :: "node list"
  

  primrec extract_scc where
    "extract_scc [] _ = undefined"
  | "\<And>S. extract_scc (v#S) u = (
       if u=v then ({v},S) else let
         (c,S) = extract_scc S u
       in (insert v c,S)
    )"

  definition tj where
    "tj E P u \<equiv> RECT (\<lambda>D (u,s). do {
        let i = idx s;
        let s = s \<lparr>
          ndidx := ndidx s(u\<mapsto>i),        (* Set index of current node *)
          lowlink := lowlink s(u\<mapsto>i),    (* Set lowlink of current node *)
          idx := i+1,                    (* Increment current index *)
          S := u#S s                     (* Push current node on stack *)
        \<rparr>;

        (* Iterate over successors *)
        (s,brk) \<leftarrow>
        FOREACHc (E``{u}) (\<lambda>(_,brk). \<not>brk) (\<lambda>v (s,_).
          if v \<notin> dom (ndidx s) then do { (* If node has not yet been visited *)
            (s,brk) \<leftarrow> D (v,s); (* Recurse *)
            (* Propagate lowlink *)
            let s = s\<lparr>lowlink := lowlink s( 
              u\<mapsto>min (the (lowlink s u)) (the (lowlink s v))
            )\<rparr>;
            RETURN (s,brk)
          } else if (v \<in> set (S s)) then do { (* If node is on stack *)
            (*Adjust lowlink *)
            let s=s\<lparr> lowlink := lowlink s( 
              u\<mapsto>min (the (lowlink s u)) (the (ndidx s v))
            )\<rparr>;
            RETURN (s,False)
          } else RETURN (s,False)
        )
        (s,False);
  
        if brk then RETURN (s,brk)
        else if lowlink s u = ndidx s u then do { (* If root node *)
          let (c,S') = extract_scc (S s) u; (* Extract SCC *)
          let s = s\<lparr> S := S'\<rparr>;
          RETURN (s, P c ) (* Check SCC *)
        } else 
        RETURN (s,brk)
      })
      (u,\<lparr> idx = 0, ndidx = empty, lowlink = empty, S =[]\<rparr>)
      "


end
