theory Edge_Walks
imports
  Digraphs
  Vertex_Walks
  "~~/src/HOL/Library/Multiset"
begin


(* FIXME: Get rid of the Vertex_Walks import ... *)

type_synonym 'b ewalk = "('b list)"

context pre_graph begin

definition ewalk_verts :: "'a \<Rightarrow> 'b ewalk \<Rightarrow> 'a list" where
  "ewalk_verts u p \<equiv> if p = [] then [u] else start G (hd p) # map (target G) p"

abbreviation ewhd :: "'a \<Rightarrow> 'b ewalk \<Rightarrow> 'a" where
  "ewhd u p \<equiv> hd (ewalk_verts u p)"

abbreviation ewlast:: "'a \<Rightarrow> 'b ewalk \<Rightarrow> 'a" where
  "ewlast u p \<equiv> last (ewalk_verts u p)"

fun vert_edge_seq :: "'b \<Rightarrow> 'b list \<Rightarrow> bool" where
    "vert_edge_seq _ [] = True"
  | "vert_edge_seq e1 (e2 # es) = (target G e1 = start G e2 \<and> vert_edge_seq e2 es)"

abbreviation ves :: "'b list \<Rightarrow> bool" where
  "ves p \<equiv> vert_edge_seq (hd p) (tl p)"

(* FIXME: simplify ewalk proofs*)
definition ewalk :: "'a \<Rightarrow> 'b ewalk \<Rightarrow> 'a \<Rightarrow> bool" where
  "ewalk u p v \<equiv> u \<in> verts G
    \<and> set p \<subseteq> edges G
    \<and> ewhd u p = u \<and> ewlast u p = v \<and> ves p"

definition epath :: "'a \<Rightarrow>'b ewalk \<Rightarrow> 'a \<Rightarrow> bool" where
  "epath u p v \<equiv> ewalk u p v \<and> distinct (ewalk_verts u p)"

end

section {* Basic Lemmas *}

context wellformed_graph begin

lemma ewalk_conv:
  "ewalk u p v = (set (ewalk_verts u p) \<subseteq> verts G
    \<and> set p \<subseteq> edges G
    \<and> ewhd u p = u \<and> ewlast u p = v \<and> ves p)"
using wellformed by (cases p) (auto simp add: ewalk_def ewalk_verts_def)

(*XXX move*)
lemma ewalkE[elim]:
  assumes "ewalk u p v"
  assumes "\<lbrakk>set (ewalk_verts u p) \<subseteq> verts G;
    set p \<subseteq> edges G; ves p;
    ewhd u p = u; ewlast u p = v\<rbrakk> \<Longrightarrow> P"
  shows "P"
using assms by (auto simp add: ewalk_conv)

lemma ewalk_verts_non_Nil[simp]:
  "ewalk_verts u p \<noteq> []"
by (simp add: ewalk_verts_def)

lemma ewalk_verts_conv:
  assumes "ewalk u p v"
  shows "ewalk_verts u p = u # map (target G) p"
using assms by (auto simp: ewalk_verts_def)


(* XXX -- necessary? *)
lemma ewalk_imp_ves:
  assumes "ewalk u p v"
  shows "ves p"
using assms by auto

lemma ewalk_induct_raw[case_names Base Cons, induct pred: ewalk]:
  assumes "ewalk u p v"
  assumes "\<And>w1. w1 \<in> verts G \<Longrightarrow> P w1 [] w1"
  assumes "\<And>w1 w2 e es. e \<in> edges G \<Longrightarrow> edge_to_ends G e = (w1, w2)
    \<Longrightarrow> P w2 es v \<Longrightarrow> P w1 (e # es) v"
  shows "P u p v"
  using `ewalk u p v`[THEN ewalk_imp_ves] and assms
proof (induct p arbitrary: u v)
  case Nil then show ?case by (auto simp: ewalk_verts_def)
next
  case (Cons e es)
  then show ?case
    by (cases es) (auto simp: ewalk_verts_def edge_to_ends_def ewalk_conv)
qed

(*XXX move,rename*)
lemma in_edges_ends_imp_in_edges:
  assumes "uv \<in> edges_ends G"
  shows "\<exists>e \<in> edges G. edge_to_ends G e = uv"
using assms by (auto simp: edges_ends_def image_def)

lemma ewalk_verts_Cons:
  assumes "ves (e # es)"
  shows "ewalk_verts u (e # es) = start G e # ewalk_verts (target G e) es"
using assms by (cases es) (auto simp: ewalk_verts_def)

(*XXX move*)
lemma ewalk_verts_Cons_iff:
  "ewalk_verts u (e # es) = start G e # map (target G) (e # es)"
by (simp add: ewalk_verts_def)

lemma ewalk_verts_Nil_iff[simp]:
  "ewalk_verts u [] = [u]"
by (simp add: ewalk_verts_def)

lemma ewhd_Cons[simp]:
  "ewhd u (e # es) = start G e"
by (auto simp: ewalk_verts_def)

lemma ewlast_Cons[simp]:
  "ewlast u (e # es) = ewlast (target G e) es"
by (auto simp: ewalk_verts_def)

lemma edge_to_ends_st:
  assumes "edge_to_ends G e = (u,v)"
  shows "start G e = u" "target G e = v"
  using assms by (simp_all add: edge_to_ends_def)

lemma ewalk_empty_iff:
  "ewalk u [] v \<longleftrightarrow> u = v \<and> u \<in> verts G"
unfolding ewalk_def by auto

lemma epath_empty_iff: "epath u [] v \<longleftrightarrow> u = v \<and> u \<in> verts G"
  by (auto simp: epath_def ewalk_empty_iff)

lemma ewalk_hd_in_verts: "ewalk u p v \<Longrightarrow> u \<in> verts G"
  by (cases p) (auto simp: ewalk_verts_def)

lemma ewalk_last_in_verts: "ewalk u p v \<Longrightarrow> v \<in> verts G"
  unfolding ewalk_conv by auto

lemma ewalk_Cons_iff:
  "ewalk u (e # es) w \<longleftrightarrow> e \<in> edges G \<and> u = start G e \<and> ewalk (target G e) es w" (is "?L \<longleftrightarrow> ?R")
proof
  assume "?L"
  then show "?R"
  proof (cases es)
    case Nil then have "target G e = w"
      using assms `?L` by (auto simp: edge_to_ends_def)
    then show ?thesis
      using `?L` Nil by (auto simp: ewalk_def intro: target_in_verts)
  next
    case (Cons e' es') then show ?thesis
      using assms `?L` by (auto simp: ewalk_def intro: start_in_verts)
  qed
next
  assume "?R" then have "ves (e # es)" using assms
    by (cases es) (auto simp: edge_to_ends_def ewalk_verts_Cons_iff)
  then show "?L"
    using `?R` assms wellformed by (auto simp: ewalk_def)
qed

(*XXX move?*)
lemma epath_Cons_iff:
  "epath u (e # es) w \<longleftrightarrow> e \<in> edges G \<and> start G e = u \<and> epath (target G e) es w
    \<and> start G e \<notin> set (ewalk_verts (target G e) es)" (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L
  from `?L` have "ves (e # es)" by (auto simp: epath_def)
  then show ?R using `?L`
    unfolding epath_def by (auto simp: ewalk_Cons_iff ewalk_verts_Cons)
next
  assume ?R
  then have "ves es" by (auto simp: epath_def)
  with `?R` have "ves (e # es)" by (cases es) (auto simp: epath_def)
  with `?R` show ?L
    by (auto simp: epath_def ewalk_Cons_iff ewalk_verts_Cons)
qed

lemma epath_nonempty_ends:
  assumes "epath u p v"
  assumes "p \<noteq> []"
  shows "u \<noteq> v"
using assms
proof (induct p arbitrary: u)
  case (Cons e es)
  then have "epath (target G e) es v" "u \<notin> set (ewalk_verts (target G e) es)"
    by (auto simp: epath_Cons_iff)
  moreover then have "v \<in> set (ewalk_verts (target G e) es)" by (auto simp: epath_def)
  ultimately show "u \<noteq> v" by auto
qed simp



(* replace by ewalk_Cons_iff?*)
lemma ewalk_ConsI:
  assumes "ewalk v es w"
  assumes "e \<in> edges G" and "edge_to_ends G e = (u,v)"
  shows "ewalk u (e # es) w"
proof -
  from assms have "ves (e # es)"
  by (cases es) (auto simp: edge_to_ends_def)
  then show ?thesis
    using assms wellformed
    by (auto simp: ewalk_def edge_to_ends_def)
qed

lemma (in pre_graph) ewalkI_epath:
  assumes "epath u p v" shows "ewalk u p v"
using assms by (simp add: epath_def)

lemma edgeE:
  assumes "edge e (u,v)"
  assumes "\<lbrakk>e \<in> edges G; start G e = u; target G e = v\<rbrakk> \<Longrightarrow> P"
  shows "P"
  using assms by (auto simp: edge_def)

lemma in_edges_imp_in_edges_ends:
  assumes "e \<in> edges G"
  shows "(start G e, target G e) \<in> edges_ends G"
using assms by (auto simp: edges_ends_def image_def edge_to_ends_def)


section {* Correspondence between @{term ewalk} and @{term vwalk} *}

(* Rename to vwalk_edges sometime ... *)
definition vwalk_to_ewalk :: "'a vwalk \<Rightarrow> 'b ewalk" where
  "vwalk_to_ewalk vp = map (\<lambda>ee. SOME e. e \<in> edges G \<and> edge_to_ends G e = ee) (vwalk_edges vp)"

lemma ewalk_has_vwalk:
  assumes "ewalk u p v"
  shows "vwalk (ewalk_verts u p) G"
  using assms assms
proof induct
  case (Cons w1 w2 e es)
  have "ewhd (target G e) es = target G e"
    using Cons by (cases es) (auto simp: edge_to_ends_def)
  then have "(start G e, ewhd (target G e) es) \<in> edges_ends G"
    using Cons by (auto intro: in_edges_imp_in_edges_ends)
  moreover
  have "vwalk (ewalk_verts (target G e) es) G"
    using Cons ewalk_Cons_iff by (auto simp: edge_to_ends_def)
  ultimately
  show ?case using Cons
    by (subst ewalk_verts_Cons)
       (auto simp: start_in_verts intro: vwalk_consI)
qed auto

lemma vwalk_has_ewalk:
  assumes "vwalk p G" and "hd p = u" and "last p = v"
  shows "ewalk u (vwalk_to_ewalk p) v"
  using assms
  unfolding vwalk_to_ewalk_def
proof (induct p arbitrary: u)
  case Base then show ?case by (auto simp: ewalk_def)
next
  case (Cons w1 w2 es)
  let "?P e" = "e \<in> edges G \<and> edge_to_ends G e = (w1,w2)"
  have "\<exists>e. ?P e"
    using `(w1, w2) \<in> edges_ends G`
    by (auto simp: edges_ends_def image_def)
  then have "?P (SOME e. ?P e)" ..
  then have "(SOME e. ?P e) \<in> edges G"
    and "edge_to_ends G (SOME e. ?P e) = (w1, hd (w2 # es))"
    by auto
  then show ?case
    using Cons by (auto intro: ewalk_ConsI)
qed

section {* ves *}

(* XXX rename? *)
lemma ves_ConsD:
  assumes "ves (e # es)"
  shows "ves es"
using assms by (cases es) auto


section {* Joining ewalks *}

abbreviation (input) ewalk_join :: "'b ewalk \<Rightarrow> 'b ewalk \<Rightarrow> 'b ewalk"
  (infixr "\<oplus>\<^sub>e" 65) where "p \<oplus>\<^sub>e q \<equiv> p @ q"

(*XXX simp a good choice?*)
lemma ewlast_join_edge[simp]:
  "ewlast x (p \<oplus>\<^sub>e [e]) = target G e"
by (simp add: ewalk_verts_def)

lemma ves_join_prefix:
  assumes "ves (p \<oplus>\<^sub>e q)"
  shows "ves p"
using assms
proof (induct p)
  case (Cons e es) then show ?case
  by (cases es q rule: list.exhaust[case_product list.exhaust]) auto
qed auto

lemma ewalk_decomp:
  assumes "ewalk u p v"
  assumes "w \<in> set (ewalk_verts u p)"
  shows "\<exists>q r. p = q @ r \<and> ewalk u q w \<and> ewalk w r v"
using assms
proof (induct p arbitrary: u)
  case Nil then show ?case by (auto simp: ewalk_def)
next
  case (Cons e es)
  from Cons(2) have ewalk_es: "ewalk (target G e) es v"
      and start_e: "start G e = u"
    using ewalk_Cons_iff by auto
  show ?case
  proof (cases "w \<in> set (ewalk_verts (target G e) es)")
    case True
    with ewalk_es obtain q r
        where "es = q @ r \<and> ewalk (target G e) q w \<and> ewalk w r v" 
      by atomize_elim (rule Cons, auto)
    then have "e # es = (e # q) @ r \<and> ewalk u (e # q) w \<and> ewalk w r v"
      using start_e Cons
      by (auto simp: ewalk_Cons_iff edge_to_ends_def)
    then show ?thesis by blast
  next
    case False
    have "target G e = ewhd (target G e) es"
      using Cons.prems(1)
      by (cases es) (auto simp: ewalk_verts_Cons_iff)
    with False Cons.prems have "start G e = w \<and> u = start G e"
      by (cases es) (auto simp: ewalk_verts_def)
    then have "e # es = [] @ (e # es) \<and> ewalk u [] w \<and> ewalk w (e # es) v"
      using Cons.prems(1)
      by (auto simp: ewalk_verts_Cons_iff ewalk_empty_iff)
    then show ?thesis by blast
  qed
qed

(* XXX hilfreich? *)
lemma tl_ewalk_verts:
  "tl (ewalk_verts u p) = map (target G) p"
by (simp add: ewalk_verts_def)

(*XXX move*)
lemma ewalk_verts_join_l:
  assumes "ewalk u (p @ q) v"
  shows "ewalk_verts u (p @ q) = ewalk_verts u p @ tl (ewalk_verts x q)"
using assms by (auto simp: ewalk_verts_def)

lemma ewhd_in_verts:
  shows "ewhd u p \<in> set (ewalk_verts u p)"
using assms by (auto simp: ewalk_verts_def)

lemma ewlast_in_verts:
  shows "ewlast u p \<in> set (ewalk_verts u p)"
using assms by (auto simp: ewalk_verts_def)

(*XXX move*)
lemma ves_join_suffix:
  assumes "ves (p \<oplus>\<^sub>e q)"
  shows "ves q"
using assms
proof (induct p)
  case Nil then show ?case by simp
next
  case (Cons e es)
  have "ves (es @ q)"
    using Cons.prems by (simp add: ves_ConsD)
  then show ?case by (rule Cons)
qed

(*XXX useful? *)
lemma ewalk_ends:
  assumes "ewalk u p v" "ewalk u' p v'"
  shows "(p \<noteq> [] \<and> u = u' \<and> v = v') \<or> (p = [] \<and> u = v \<and> u' = v')"
using assms by (auto simp: ewalk_verts_def ewalk_def split: split_if_asm)

lemma ewalk_empty_ends:
  assumes "ewalk u [] v"
  shows "u = v"
using assms by (auto simp: ewalk_def)

lemma epath_ends:
 assumes "epath u p v" and "epath u' p v'"
  shows "p \<noteq> [] \<and> u \<noteq> v \<and> u = u' \<and> v = v' \<or> p = [] \<and> u = v \<and> u' = v'" 
using assms unfolding epath_def by (metis assms(2) epath_nonempty_ends  ewalk_ends)

(*XXX move, rename? *)
(* -- can we use this to have a nicer proof of the decomp lemma? *)
lemma ewalk_join_ewalk_l:
  assumes "ewalk u (p @ q) v"
  assumes "w = ewlast u p"
  shows "ewalk u p w"
using assms
unfolding ewalk_def
by (cases q) (auto intro: ves_join_prefix simp: ewalk_verts_def)

lemma ewalk_join_ewalk_l':
  assumes "ewalk u (p @ q) w"
  assumes "ewalk v q w"
  shows "ewalk u p v"
using assms(1)
proof (rule ewalk_join_ewalk_l[of u p q w v])
  from assms show "v = ewlast u p"
  proof (induct p arbitrary: u)
(*    case Nil then show ?case
      apply (cases q) apply auto
Mit welchen simp-Regeln könnte ich "ewalk u [] p ==> ewalk v [] p ==> u = v" beweisen?
*)
    case Nil then show ?case using ewalk_ends[of u q w v w] by auto
  next
    case (Cons e es)
    have "v = ewlast (target G e) es"
      using Cons by (auto simp: ewalk_Cons_iff)
    then show ?case
      by (cases es) auto
  qed
qed

(*XXX move, rename*)
lemma ewalk_join_ewalk_r:
  assumes "ewalk u (p @ q) v"
  assumes "w = ewhd v q"
  shows "ewalk w q v"
using assms unfolding ewalk_conv
by (auto simp: ewalk_verts_def intro: ves_join_suffix hd_in_set start_in_verts)

lemma ewalk_join_inner_ends:
  assumes "ewalk u (p @ q) v"
  shows "ewlast u p = ewhd v q"
using assms proof (induct p arbitrary: u)
  case Nil then show ?case by (cases q) auto
next
  case (Cons e es) then show ?case by (auto simp: ewalk_Cons_iff)
qed

lemma ewalk_join_ewalk_r':
  assumes "ewalk u (p @ q) w"
  assumes "ewalk u p v"
  shows "ewalk v q w"
using assms(1)
proof (rule ewalk_join_ewalk_r)
  from assms show "v = ewhd w q"
  proof (induct p arbitrary: u)
    case Nil
    then have "v = u" by (auto simp: ewalk_empty_ends)
    then show ?case
      using Nil by (cases q) auto
  next
    case (Cons e es)
    have "ewalk (target G e) (es @ q) w" "ewalk (target G e) es v"
      using Cons(2-3) by (auto simp: ewalk_Cons_iff)
    then show ?thesis by (rule Cons)
  qed
qed

(*XXX mv*)
lemmas list_exhaust2 = list.exhaust[case_product list.exhaust]

lemma ewlast_append:
  "ewlast u (p @ q) = ewlast (ewlast u p) q"
by (cases p q rule: list_exhaust2) (auto simp: ewalk_verts_Cons_iff)

lemma ewhd_append:
  "ewhd u (p @ q) = ewhd (ewhd u q) p"
by (cases p q rule: list_exhaust2) auto

(*XXX move, rename *)
lemma FOO1:
  assumes "u \<in> verts G" "set p \<subseteq> edges G"
  shows "set (ewalk_verts u p) \<subseteq> verts G"
using assms
by (auto simp: ewalk_verts_def intro: hd_in_set start_in_verts target_in_verts)

lemma  ewalkE_verts:
  assumes "ewalk u p v"
  obtains x where "u \<in> verts G" "v \<in> verts G"
proof -
  have "u \<in> set (ewalk_verts u p)"
    using ewalk_verts_conv[OF assms] by auto
  moreover
  have "v \<in> set (ewalk_verts u p)"
    using assms by auto
  ultimately
  show ?thesis using assms by (auto intro: that)
qed

lemma ewhd_append2:
  assumes "ewalk u p v" "ewalk v q w"
  shows "ewhd u (p @ q) = u"
using assms ewalk_empty_ends[of u v]
by (cases p q rule: list_exhaust2) auto

lemma ewlast_append2:
  assumes "ewalk u p v" "ewalk v q w"
  shows "ewlast u (p @ q) = w"
using assms ewalk_empty_ends[of v w]
by (cases p q rule: list_exhaust2) (auto simp: ewalk_verts_Cons_iff)

lemma ewalk_ewalk_joinI:
  assumes "ewalk u p v"
  assumes "ewalk v q w"
  shows "ewalk u (p @ q) w"
using assms
proof (induct p arbitrary: u)
  case Nil then show ?case by auto
next
  case (Cons e es)
  then have ee_e: "edge_to_ends G e = (u, target G e)"
    unfolding edge_to_ends_def by auto

  have "ewalk (target G e) es v"
    using ee_e Cons(2) ewalk_Cons_iff by auto
  then show ?case using Cons ee_e by (auto simp: ewalk_Cons_iff)
qed

lemma ewalk_join_decomp:
  "ewalk u (p @ q) v \<longleftrightarrow> ewalk u p (ewlast u p) \<and> ewalk (ewlast u p) q v" (is "?L \<longleftrightarrow> ?R")
by (auto simp: ewalk_join_inner_ends
  intro: ewalk_join_ewalk_l ewalk_join_ewalk_r ewalk_ewalk_joinI)

lemma ewalk_verts_append:
  assumes "ewalk u (p @ q) v"
  shows "ewalk_verts u (p @ q) = ewalk_verts u p @ tl (ewalk_verts (ewlast u p) q)"
using assms proof (induct p arbitrary: u)
  case Nil then show ?case by (cases q) (auto simp: ewalk_verts_Cons_iff)
next
  case (Cons e es) then show ?case by (cases q) (auto simp: ewalk_verts_Cons_iff)
qed

lemma ewalk_verts_append2:
  assumes "ewalk u (p @ q) v"
  shows "ewalk_verts u (p @ q) = butlast (ewalk_verts u p) @ ewalk_verts (ewlast u p) q"
proof -
  have "last (ewalk_verts u p) = hd (ewalk_verts (ewlast u p) q)"
    using assms by (auto simp: ewalk_join_decomp)
  then have S1: "ewalk_verts u p = butlast (ewalk_verts u p) @ [hd (ewalk_verts (ewlast u p) q)]"
    by auto
  have S2: "ewalk_verts u (p @ q) = ewalk_verts u p @ tl (ewalk_verts (ewlast u p) q)"
    using assms by (rule ewalk_verts_append)
  show ?thesis by (subst S2, subst S1, simp)
qed

lemma epath_join_decomp:
  "epath u (p @ q) v \<longleftrightarrow> epath u p (ewlast u p) \<and> epath (ewlast u p) q v \<and>
    set (ewalk_verts u p) \<inter> set (tl (ewalk_verts (ewlast u p) q)) = {}" (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L
  then have "distinct (ewalk_verts (ewlast u p) q)" by (auto simp: epath_def ewalk_verts_append2)
  with `?L` show ?R by (auto simp: epath_def ewalk_join_decomp ewalk_verts_append)
next
  assume ?R
  then show ?L by (auto simp: epath_def ewalk_join_decomp ewalk_verts_append dest: distinct_tl)
qed



section {* take'n'drop *}

(*XXX rename? *)
lemma ves_ewalk_take:
  assumes "ves p"
  shows "ves (take n p)"
proof -
  from assms
  have "ves (take n p \<oplus>\<^sub>e drop n p)" by simp
  then show ?thesis by (rule ves_join_prefix)
qed

(*XXX rename? *)
lemma ves_ewalk_drop:
  assumes "ves p"
  shows "ves (drop n p)"
proof -
  from assms
  have "ves (take n p \<oplus>\<^sub>e drop n p)" by simp
  then show ?thesis by (rule ves_join_suffix)
qed

lemma length_ewalk_verts:
  "length (ewalk_verts u p) = Suc (length p)"
by (simp add: ewalk_verts_def)

(*XXX rename? *)
lemma ves_tl:
  assumes "ves p"
  shows "ves (tl p)"
using assms by (auto simp: ves_ConsD)

(*XXX rename*)
lemma ves_nth:
  "ves p \<longleftrightarrow> (\<forall>i. Suc i < length p \<longrightarrow> target G (p ! i) = start G (p ! Suc i))" (is "?L \<longleftrightarrow> ?R")
proof (induct p)
  case Nil then show ?case by simp
next
  case (Cons e es)
  show ?case (is "?L \<longleftrightarrow> ?R")
  proof
    assume "?L"
    { fix i assume A: "Suc i < length (e # es)"
      have "target G ((e # es) ! i) = start G ((e # es) ! Suc i)"
      proof (cases i)
        case 0 then show ?thesis using A `?L` by (cases es) auto
      next
        case (Suc n) then show ?thesis
          using A `?L` Cons by (auto simp: ves_ConsD)
      qed }
    then show ?R by auto
  next
    assume "?R"
    then have "ves es" using Cons by auto
    show "?L"
    proof (cases es)
      case Nil then show ?thesis by simp
    next
      case (Cons e' es')
      with `?R` have "target G e = start G e'"
        by auto
      then show ?thesis using `ves es` Cons by auto
    qed
  qed
qed


section {* Foo *}

(* XXX entfernen, da kein walk? *)
definition closed_w :: "'b ewalk \<Rightarrow> bool" where
  "closed_w p \<equiv> \<exists>u. ewalk u p u \<and> 0 < length p"

definition ecycle :: "'b ewalk \<Rightarrow> bool" where
  "ecycle p \<equiv> \<exists>u. ewalk u p u \<and> distinct (tl (ewalk_verts u p)) \<and> 2 \<le> length p"

(*XXX short closed vwalk
  / LEDA cycle / Volkmann cycle *)
definition loop_cycle :: "'b ewalk \<Rightarrow> bool" where
  "loop_cycle p \<equiv> closed_w p \<and> (\<exists>u. distinct (tl (ewalk_verts u p)))"


lemma ewalk_verts_take_conv:
  "ewalk_verts u (take n p)
    = (if n > 0 \<and> p \<noteq> [] then take (Suc n) (ewalk_verts u p) else [u])"
by (cases n p rule: nat.exhaust[case_product list.exhaust])
   (auto simp: ewalk_verts_Cons_iff take_map)

lemma ewalk_verts_drop_conv:
  assumes "ves p"
  shows "ewalk_verts u (drop n p)
    = (if n < length p then drop n (ewalk_verts u p) else [u])"
using assms
proof (induct n arbitrary: u)
  case 0 then show ?case by simp
next
  case (Suc n)
  then have "ves (take n p @ drop n p)" by simp
  then have "ves (drop n p)" by (rule ves_join_suffix)

  show ?case
  proof (cases "drop n p")
    case Nil then show ?thesis by auto
  next
    case (Cons e es)
    then have dSp: "drop (Suc n) p = es" "Suc n < length p \<Longrightarrow> es \<noteq> []"
      using length_drop[of n p] by (auto simp: drop_Suc drop_tl)
    then have "Suc n < length p \<Longrightarrow> start G (hd (drop (Suc n) p)) = target G e"
      unfolding dSp using Cons `ves (drop n p)` ves_nth[of "e # es"]
      by (auto simp: hd_conv_nth)
    then show ?thesis
      using Cons `drop (Suc n) p = es`
      by (auto simp add: ewalk_join_decomp ewalk_verts_def drop_map
          simp del: append_take_drop_id)
  qed
qed

(*XXX move*)
lemma ewalk_verts_ne_eq:
  assumes "p \<noteq> []"
  shows "ewalk_verts u p = ewalk_verts v p"
using assms by (auto simp: ewalk_verts_def)

lemma ewalk_not_distinct_decomp:
  assumes "ewalk u p v"
  assumes "\<not> distinct (ewalk_verts u p)"
  shows "\<exists>q r s. p = q \<oplus>\<^sub>e r \<oplus>\<^sub>e s \<and> distinct (ewalk_verts u q)
    \<and> 0 < length r
    \<and> (\<exists>w. ewalk u q w \<and> ewalk w r w \<and> ewalk w s v)"
proof -
  from assms
  obtain xs ys zs y where
    pv_decomp: "ewalk_verts u p = xs @ y # ys @ y # zs"
    and xs_y_props: "distinct xs" "y \<notin> set xs" "y \<notin> set ys"
    using not_distinct_decomp_min by blast

  def q \<equiv> "take (length xs) p"
  def r \<equiv> "take (1 + length ys) (drop (length xs) p)"
  def s \<equiv> "(drop (1 + length xs + length ys) p)"

  have "p = (q @ r) @ s"
    unfolding q_def r_def s_def by (simp add: take_add[symmetric])
  then have p_decomp: "p = q @ r @ s" by simp

  have r_len: "0 < length r"
    unfolding r_def using pv_decomp length_ewalk_verts[of u p] by auto

  have dist_q: "distinct (ewalk_verts u q)"
    using q_def pv_decomp xs_y_props
    by (auto simp: ewalk_verts_take_conv)

  have "ves p" using `ewalk u p v` by (simp add: ewalk_def)

  def w \<equiv> "ewlast u q"
  have "ewlast w r = w"
  proof -
    have X:  "w = y" "ves p" "length xs < length p" "p \<noteq> []"
      using pv_decomp length_ewalk_verts[of u p] `ewalk u p v`
      by (auto simp: ewalk_verts_take_conv w_def q_def)
    then show ?thesis
      unfolding r_def using pv_decomp ewalk_verts_ne_eq[of p y u]
      by (simp add: ewalk_verts_take_conv ewalk_verts_drop_conv not_le)
  qed
  then have "ewalk u q w" "ewalk w r w" "ewalk w s v"
    using w_def `ewalk u p v` p_decomp
    by (auto simp: ewalk_join_decomp)
  then show ?thesis using p_decomp dist_q r_len by blast
qed

lemma set_ewalk_verts:
  assumes "ewalk u p v"
  shows "set (ewalk_verts u p) = {u} \<union> set (map (start G) p) \<union> set (map (target G) p)"
using assms
proof (induct p arbitrary: u)
  case Nil then show ?case by simp
next
  case (Cons e es)
  then have "set (ewalk_verts (target G e) es)
      = {target G e} \<union> set (map (start G) es) \<union> set (map (target G) es)"
    by (auto simp: ewalk_Cons_iff)
  with Cons.prems show ?case by (auto simp: ewalk_verts_Cons)
qed

lemma epath_decomp_disjoint:
  assumes "epath u p v"
  assumes "p = q @ r"
  assumes "x \<in> set (ewalk_verts u q)" "x \<in> set (tl (ewalk_verts (ewlast u q) r))"
  shows False
using assms by (auto simp: epath_def ewalk_verts_append) 



section {* Reachability *}

lemma reachable_ewalk:
  "u \<rightarrow>*G v \<longleftrightarrow> (\<exists>p. ewalk u p v)"
proof
  assume "u \<rightarrow>*G v"
  then obtain vp where "vwalk vp G" "hd vp = u" "last vp = v"
    by auto
  then have "ewalk u (vwalk_to_ewalk vp) v"
    by (auto  intro: vwalk_has_ewalk)
  then show "\<exists>p. ewalk u p v" by auto
next
  assume "\<exists>p. ewalk u p v"
  then obtain ep where "ewalk u ep v"
    by auto
  then have "vwalk (ewalk_verts u ep) G"
    "hd (ewalk_verts u ep) = u"
    "last (ewalk_verts u ep) = v"
    by (auto intro: ewalk_has_vwalk)
  then show "u \<rightarrow>*G v" ..
qed

lemma reachable_ewalkI[intro?]:
  assumes "ewalk u p v"
  shows "u \<rightarrow>*G v"
  unfolding reachable_ewalk using assms by auto

(*FIXME: Turn this into an elimination rule?*)
lemma reachable_cons:
  assumes "u \<rightarrow>*G v" "edge e (v,w)" (*FIXME: change to \<exists>v e. ...?*)
  shows "u \<rightarrow>*G w"
proof -
  from `u \<rightarrow>*G v` obtain p where "ewalk u p v"
    by (auto simp: reachable_ewalk)
  moreover have "ewalk v [e] w"
    using `edge e (v,w)`
    by (auto simp: edge_def ewalk_def start_in_verts target_in_verts)
  ultimately have "ewalk u (p @ [e]) w"
    by (rule ewalk_ewalk_joinI)
  then show ?thesis ..
qed

lemma reachable_trans[trans]:
  assumes "u \<rightarrow>*G v" "v \<rightarrow>*G w" shows "u \<rightarrow>*G w"
using assms by (metis reachable_ewalk ewalk_ewalk_joinI)


section {* General Graph stuff -- move *}

lemma ewalk_edges_distinct:
  assumes "distinct (tl (ewalk_verts u p))"
  shows "distinct p"
using assms by (cases p) (auto simp: distinct_map ewalk_verts_def)

lemma (in pseudo_digraph) epaths_finite:
  shows "finite {p. epath u p v}"
proof -
  have "{p. epath u p v}
    \<subseteq> {es. set es \<subseteq> edges G \<and> length es \<le> card (verts G)}"
  proof (clarify, rule conjI)
    fix p assume "epath u p v"
    then show "set p \<subseteq> edges G" by (auto simp: epath_def)

    show "length p \<le> card (verts G)"
    proof -
      have "length p \<le> length (ewalk_verts u p)"
        by (auto simp: ewalk_verts_def)
      also have "\<dots> = card (set (ewalk_verts u p))"
        using `epath u p v` by (auto simp: epath_def distinct_card)
      also have "\<dots> \<le> card (verts G)"
        using `epath u p v`
        by (auto simp: epath_def ewalk_conv intro: card_mono)
      finally show ?thesis .
    qed
  qed
  moreover have "finite ..."
    using finite_verts finite_edges
    by (intro finite_cartesian_product finite_lists_length_le)
  ultimately show ?thesis by (rule finite_subset)
qed

lemma ewalk_omit_closed_w:
  assumes "ewalk u (p @ q @ r) w"
  assumes "ewalk v q v"
  shows "ewalk u (p @ r) w"
proof (cases q)
  case Nil then show ?thesis using assms by simp
next
  case (Cons e es)
  have "start G e = v"
    using Cons assms(2) by auto
  then have *: "ewalk v (q @ r) w"
    using assms(1) Cons
    by (auto simp: ewalk_verts_Cons_iff intro: ewalk_join_ewalk_r)
  with assms(1)  
  have "ewalk u p v" by (rule ewalk_join_ewalk_l')
  moreover have "ewalk v r w"
    using * assms(2) by (rule ewalk_join_ewalk_r')
  ultimately show ?thesis by (rule ewalk_ewalk_joinI)
qed


section {* ewalk to epath *}

fun cycle_decomp_prop :: "'b ewalk =>
  ('b ewalk \<times> 'b ewalk \<times> 'b ewalk) \<Rightarrow> bool" where
  "cycle_decomp_prop p (q,r,s) \<longleftrightarrow> p = q @ r @ s
    \<and> (\<exists>u v w. ewalk u q v \<and> ewalk v r v \<and> ewalk v s w)
    \<and> 0 <length r
    \<and> (\<exists>u. distinct (ewalk_verts u q))"

definition cycle_decomp :: "'b ewalk
    \<Rightarrow> 'b ewalk \<times> 'b ewalk \<times> 'b ewalk" where
  "cycle_decomp p = (SOME qrs. cycle_decomp_prop p qrs)"

function ewalk_to_epath :: "'b ewalk \<Rightarrow> 'b ewalk" where
  "ewalk_to_epath p = (if \<not>(\<exists>u. distinct (ewalk_verts u p)) \<and> (\<exists>u v. ewalk u p v)
     then (let (q,r,s) = cycle_decomp p in ewalk_to_epath (q \<oplus>\<^sub>e s))
     else p)"
by auto

lemma cycle_decomp_has_prop:
  assumes "ewalk u p v" and "\<not>distinct (ewalk_verts u p)"
  shows "cycle_decomp_prop p (cycle_decomp p)"
proof -
  obtain q r s where *: "p = q @ r @ s \<and> distinct (ewalk_verts u q)
      \<and> 0 < length r
      \<and> (\<exists>w. ewalk u q w \<and> ewalk w r w \<and> ewalk w s v)"
    by (atomize_elim) (rule ewalk_not_distinct_decomp[OF assms])
  then have "\<exists>x. cycle_decomp_prop p x"
    by (intro exI[where x="(q,r,s)"]) auto
  then show ?thesis unfolding cycle_decomp_def ..
qed

lemma cycle_decompE:
  assumes dec: "cycle_decomp p = (q,r,s)"
  assumes p_props: "ewalk u p v" "\<not>distinct (ewalk_verts u p)"
  assumes elim: "\<lbrakk>p = q @ r @ s; distinct (ewalk_verts u q); \<exists>w. ewalk u q w \<and> ewalk w r w \<and> ewalk w s v; closed_w r\<rbrakk> \<Longrightarrow> P"
  shows P
proof (rule elim)
  show "p = q @ r @ s" "distinct (ewalk_verts u q)" "closed_w r"
    using cycle_decomp_has_prop[OF p_props] and dec
    by (auto simp: closed_w_def ewalk_verts_def)
  then have "p \<noteq> []" by (auto simp: closed_w_def)

  (* TODO: Can we find some general rules to prove the last property?*)
  obtain u' w' v' where obt_ewalk: "ewalk u' q w'" "ewalk w' r w'" "ewalk w' s v'"
    using cycle_decomp_has_prop[OF p_props] and dec by auto
  then have "ewalk u' p v'"
    using `p = q @ r @ s` by (auto simp: ewalk_ewalk_joinI)
  then have "u = u'" and "v = v'" using `p \<noteq> []` `ewalk u p v` by (metis ewalk_ends)+
  then have "ewalk u q w'" "ewalk w' r w'" "ewalk w' s v"
    using obt_ewalk by auto
  then show "\<exists>w. ewalk u q w \<and> ewalk w r w \<and> ewalk w s v" by auto
qed

termination ewalk_to_epath
proof (relation "measure length")
  fix G p qrs rs q r s

  have X: "\<And>x y. closed_w r \<Longrightarrow> ewalk x r y \<Longrightarrow> x = y"
    unfolding closed_w_def by (blast dest: ewalk_ends)

  assume "\<not>(\<exists>u. distinct (ewalk_verts u p)) \<and>(\<exists>u v. ewalk u p v)"
    and **:"qrs = cycle_decomp p" "(q, rs) = qrs" "(r, s) = rs"
  then obtain u v where *: "ewalk u p v" "\<not>distinct (ewalk_verts u p)"
    by (cases p) auto
  then have "cycle_decomp p = (q,r,s)" using ** by simp 
  then have "cycle_decomp_prop p (q,r,s)"
    apply (rule cycle_decompE[OF _ *])
    using X[of "ewlast u q"  "ewlast (ewlast u q) r"] *(1)
    apply (auto simp: ewalk_join_decomp closed_w_def)
    done
  then show "(q \<oplus>\<^sub>e s, p) \<in> measure length"
    by (auto simp: closed_w_def)
qed simp
declare ewalk_to_epath.simps[simp del]

lemma ewalk_to_epath_induct[consumes 1, case_names path decomp]:
  assumes ewalk: "ewalk u p v"
  assumes dist: "\<And>p. ewalk u p v \<Longrightarrow> distinct (ewalk_verts u p) \<Longrightarrow> P p"
  assumes dec: "\<And>p q r s. \<lbrakk>ewalk u p v; cycle_decomp p = (q,r,s);
    \<not>distinct (ewalk_verts u p); P (q @ s)\<rbrakk> \<Longrightarrow> P p"
  shows "P p"
using ewalk
proof (induct "length p" arbitrary: p rule: less_induct)
  case less
  show ?case
  proof (cases "distinct (ewalk_verts u p)")
    case True then show ?thesis by (auto intro: dist less.prems)
  next
    case False

    obtain q r s where p_cdecomp: "cycle_decomp p = (q,r,s)"
      by (cases "cycle_decomp p") auto
    then have "cycle_decomp_prop p (q,r,s)" "p = q @ r @ s"
      using cycle_decomp_has_prop[OF less.prems(1) False] by auto
    then have "length (q @ s) < length p" "ewalk u (q @ s) v"
      using less.prems by (auto simp add: ewalk_omit_closed_w[of u q r s])
    then have "P (q @ s)" by (auto intro: less)

    with p_cdecomp False show ?thesis by (auto intro: dec less.prems)
  qed
qed

lemma step_ewalk_to_epath:
  assumes ewalk: "ewalk u p v"
    and decomp: "cycle_decomp p = (q, r, s)"
    and dist: "\<not> distinct (ewalk_verts u p)"
  shows "ewalk_to_epath p = ewalk_to_epath (q @ s)"
proof -
  from dist have "\<not>(\<exists>u. distinct (ewalk_verts u p))"
    by (auto simp: ewalk_verts_def)
  with ewalk and decomp show "ewalk_to_epath p = ewalk_to_epath (q @ s)"
    by (auto simp: ewalk_to_epath.simps)
qed

lemma epath_ewalk_to_epath:
  assumes "ewalk u p v"
  shows "epath u (ewalk_to_epath p) v"
using assms
proof (induct rule: ewalk_to_epath_induct)
  case (path p)
  then have "ewalk_to_epath p = p"
    by (auto simp: ewalk_to_epath.simps)
  then show ?case using path by (auto simp: epath_def)
next
  case (decomp p q r s) 
  then show ?case using step_ewalk_to_epath[of _ p _ q r s] by simp
qed

end

end
