theory Shortest_Paths imports
  Edge_Walks
  Weighted_Graphs
  "~~/src/HOL/Library/Extended_Real"
begin

context wellformed_graph begin

section {* minimal path costs *}

(*Rename, as we are outside of any local context?*)
(* XXX: conversion insert map here ... *)
definition \<mu> where
  "\<mu> f u v \<equiv> INF p: {p. ewalk u p v}. ewalk_cost f p"

lemma shortest_path_inf:
  assumes "\<not>(u \<rightarrow>*G v)"
  shows "\<mu> f u v = \<infinity>"
proof -
  have *: "{p. ewalk u p v} = {}"
    using assms by (auto simp: reachable_ewalk)
  show "\<mu> f u v = \<infinity>" unfolding \<mu>_def *
    by (simp add: INF_def top_ereal_def)
qed

lemma min_cost_le_walk_cost:
  assumes "ewalk u p v" 
  shows "\<mu> c u v \<le> ewalk_cost c p"
  using assms unfolding \<mu>_def by (auto intro: INF_lower2)

lemma pos_cost_pos_ewalk_cost:
  assumes "ewalk u p v"
  assumes pos_cost: "\<And>e. e \<in> edges G \<Longrightarrow> c e \<ge> 0"
  shows "ewalk_cost c p \<ge> 0"
using assms by (induct p arbitrary: u) (auto simp: ewalk_Cons_iff)


subsection {* A path consisting of n copies of a cycle *}

fun mk_cycles_path :: "nat
  \<Rightarrow> 'b ewalk \<Rightarrow> 'b ewalk" where
    "mk_cycles_path 0 c = []"
  | "mk_cycles_path (Suc n) c = c \<oplus>\<^sub>e (mk_cycles_path n c)"

(*
lemma mk_cycles_path_hd[simp]:
  shows "ewhd u (mk_cycles_path n p) = ewhd u p"
by (cases n) simp_all

lemma mk_cycles_path_last[simp]:
  assumes "ewalk_joinable p p"
  shows "ewlast (mk_cycles_path n p) = ewlast p"
  using assms
proof (induct n) 
  case 0 then show ?case by (simp add: ewalk_joinable_def)
next
  case (Suc n)
  have "ewalk_joinable p (mk_cycles_path n p)"
    using Suc.prems by (rule mk_cycles_path_ewalk_joinable)
  then show ?case
    using Suc by (simp add: ewlast_ewalk_join)
qed
*)

lemma mk_cycles_path_ewalk:
  assumes "ewalk u c u"
  shows "ewalk u (mk_cycles_path n c) u"
using assms proof (induct n)
  case 0 then show ?case
    (* XXX: should be much simpler? *)
    by (cases c) (auto simp: ewalk_empty_iff ewalk_verts_Cons_iff)
next
  case (Suc n) then show ?case
    using Suc by (auto intro: ewalk_ewalk_joinI)
qed

lemma mk_cycles_ewalk_cost:
  assumes "ewalk u p u"
  shows "ewalk_cost c (mk_cycles_path n p) = n * ewalk_cost c p"
using assms proof (induct rule: mk_cycles_path.induct)
  case 1 show ?case by simp
next
  case (2 n p)
  have "ewalk_cost c (mk_cycles_path (Suc n) p)
      = ewalk_cost c (p \<oplus>\<^sub>e (mk_cycles_path n p))"
    by simp
  also have "\<dots> = ewalk_cost c p + real n * ewalk_cost c p"
  proof (cases n)
    case 0 then show ?thesis by simp
  next
    case (Suc n') then show ?thesis
      using 2 by simp
  qed
  also have "\<dots> = real (Suc n) * ewalk_cost c p"
    by (simp add: algebra_simps real_of_nat_Suc)
  finally show ?case .
qed

lemma inf_over_nats:
  fixes a c :: real
  assumes "c < 0"
  shows "(INF (i :: nat). a + i * c) = - \<infinity>"
proof (rule ereal_INFI)
  fix i :: nat show "- \<infinity> \<le> a + real i * c" by simp
next
  fix y :: ereal
  assume "\<And>(i :: nat). i \<in> UNIV \<Longrightarrow> y \<le> a + real i * c"
  then have l_assm: "\<And>i::nat. y \<le> a + real i * c" by simp

  show "y \<le> - \<infinity>"
  proof (subst ereal_infty_less_eq, rule ereal_bot)
    fix B :: real
    obtain real_x where "a + real_x * c \<le> B" using `c < 0`
      by atomize_elim
         (rule exI[where x="(- abs B -a)/c"], auto simp: field_simps)
    obtain x :: nat where "a + x * c \<le> B"
    proof (atomize_elim, intro exI[where x="natceiling real_x"] conjI)
      have "real (natceiling real_x) * c \<le> real_x * c"
        using `c < 0` by (simp add: real_natceiling_ge)
      then show "a + natceiling real_x * c \<le> B"
        using `a + real_x * c \<le> B` by simp
    qed
    then show "y \<le> ereal B"
    proof -
      have "ereal (a + x * c) \<le> ereal B"
        using `a + x * c \<le> B` by simp
      with l_assm show ?thesis by (rule order_trans)
    qed
  qed
qed

lemma neg_cycle_imp_inf_\<mu>:
  assumes walk_p: "ewalk u p v"
  assumes walk_c: "ewalk w c w"
  assumes w_in_p: "w \<in> set (ewalk_verts u p)"
  assumes "ewalk_cost f c < 0"
  shows "\<mu> f u v = -\<infinity>"
proof -
  from w_in_p obtain xs ys where pv_decomp: "ewalk_verts u p = xs @ w # ys"
    by (auto simp: in_set_conv_decomp)

  def q \<equiv> "take (length xs) p" and r \<equiv> "drop (length xs) p"
  def ext_p \<equiv> "\<lambda>n. q @ mk_cycles_path n c @ r"

  have ext_p_cost: "\<And>n. ewalk_cost f (ext_p n)
      = (ewalk_cost f q + ewalk_cost f r) + n * ewalk_cost f c"
    using `ewalk w c w`
    by (auto simp: ext_p_def intro: mk_cycles_ewalk_cost)

  from q_def r_def have "ewlast u q = w"
    using pv_decomp walk_p by (cases xs) (auto simp: ewalk_verts_take_conv)
  moreover
  from q_def r_def have "ewalk u (q @ r) v"
    using walk_p by simp
  ultimately
  have "ewalk u q w" "ewalk w r v" "\<And>n. ewalk w (mk_cycles_path n c) w"
    using walk_c
    by (auto simp: ewalk_join_decomp intro: mk_cycles_path_ewalk)
  then have "\<And>n. ewalk u (ext_p n) v"
    unfolding ext_p_def by (blast intro: ewalk_ewalk_joinI)
  then have "{ext_p i|i. i \<in> UNIV} \<subseteq> {p. ewalk u p v}"
    by auto
  then have "(INF p:{p. ewalk u p v}. ewalk_cost f p)
      \<le> (INF p: {ext_p i|i. i \<in> UNIV}. ewalk_cost f p)"
    by (auto intro: INF_superset_mono)
  also have "\<dots> = (INF i: UNIV. ewalk_cost f (ext_p i))"
    unfolding INF_def by (rule arg_cong[where f=Inf], auto)
  also have "\<dots> = - \<infinity>" unfolding ext_p_cost 
    by (rule inf_over_nats[OF `ewalk_cost f c < 0`])
  finally show ?thesis unfolding \<mu>_def by simp
qed

(* XXX Move Stuff *)
lemma not_distinct_decomp_distinct:
  assumes "\<not> distinct ws"
  shows "\<exists>xs y ys zs. ws = xs @ y # ys @ y # zs \<and> distinct (ys @ [y])"
using assms
proof (induct ws)
  case (Cons w ws)
  show ?case
  proof (cases "distinct ws")
    case True
    then obtain xs ys where "ws = xs @ w # ys" "w \<notin> set xs"
      using Cons.prems by (fastforce dest: split_list_first)
    then have "distinct (xs @ [w])" "w # ws = [] @ w # xs @ w # ys"
      using `distinct ws` by auto
    then show ?thesis by blast
  next
    case False
    then obtain xs y ys zs where "ws = xs @ y # ys @ y # zs \<and> distinct (ys @ [y])"
      using Cons by auto
    then have "w # ws = (w # xs) @ y # ys @ y # zs \<and> distinct (ys @ [y])"
      by simp
    then show ?thesis by blast
  qed
qed simp

(* XXX move stuff *)
lemma take_drop_take:
   "take n xs @ drop n (take m xs) = take (max n m) xs"
proof cases
  assume "m < n" then show ?thesis by (auto simp: max_def)
next
  assume "\<not>m < n"
  then have "take n xs = take n (take m xs)" by (auto simp: min_def)
  then show ?thesis by (simp del: take_take add: max_def)
qed

lemma drop_take_drop:
  "drop n (take m xs) @ drop m xs = drop (min n m) xs"
proof cases
  assume A: "\<not>m < n"
  then show ?thesis
    using drop_append[of n "take m xs" "drop m xs"]
  by (cases "length xs < n") (auto simp: not_less min_def)
qed (auto simp: min_def)

lemma ewalk_cyclic_imp_ex_scw:
  assumes "ewalk u p u"
  assumes "closed_w p"
  shows "\<exists>q r s. p = q \<oplus>\<^sub>e r \<oplus>\<^sub>e s \<and> loop_cycle r"
proof (cases "distinct (tl (ewalk_verts u p))")
  case True
    have "p = [] \<oplus>\<^sub>e p \<oplus>\<^sub>e  []" by simp
    moreover
    have "loop_cycle p"
      using True and assms unfolding loop_cycle_def by auto
    ultimately
    show ?thesis by fast
next
  case False
  then obtain xs ys zs y where
      "tl (ewalk_verts u p) = xs @ y # ys @ y # zs" "distinct (ys @ [y])"
    using not_distinct_decomp_distinct by blast
  then have pv_decomp: "ewalk_verts u p = u # xs @ y # ys @ y # zs"
    using assms by (cases p) (auto simp: ewalk_verts_Cons)

  def q \<equiv> "take (Suc (length xs)) p"
  def r \<equiv> "take (Suc (length ys)) (drop (Suc (length xs)) p)"
  def s \<equiv> "drop (length xs + length ys + 2) p"

  have p_decomp: "p = q @ r @ s"
    unfolding q_def r_def s_def
    by (auto simp: take_drop take_drop_take drop_take_drop algebra_simps min_def)

  have cycle: "loop_cycle r"
  proof -
    have "ves p" using assms by auto
    have r_verts: "\<And>x. ewalk_verts x r = y # ys @ [y]"
      unfolding r_def using pv_decomp length_ewalk_verts[of u p]
      by (simp add: ewalk_verts_take_conv ewalk_verts_drop_conv[OF `ves p`])
         (auto simp: ewalk_verts_def)
    then have "\<And>x. ewhd x r = y" "\<And>x. ewlast x r = y"
      using r_verts by (auto simp: ewalk_verts_def)
    moreover
    have "ewalk (ewlast u q) r (ewlast (ewlast u q) r)"
      using assms p_decomp by (auto simp: ewalk_join_decomp)
    ultimately
    have "ewalk y r y" by (simp add: ewalk_def)
    moreover
    have "0 < length r" "distinct (tl (ewalk_verts y r))"
      using r_verts and length_ewalk_verts[of _ r] `distinct (ys @ [y])` by auto
    ultimately
    show "loop_cycle r"
      unfolding loop_cycle_def closed_w_def by auto
  qed

  from p_decomp and cycle show ?thesis by blast
qed

lemma walk_cheaper_path_imp_neg_cyc:
  assumes p_props: "ewalk u p v"
  assumes less_path_\<mu>: "ewalk_cost f p < (INF p: {p. epath u p v}. ewalk_cost f p)"
  shows "\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u p) \<and> ewalk_cost f c < 0"
proof -
  def path_\<mu>  \<equiv> "(INF p: {p. epath u p v}. ewalk_cost f p)"
  then have "ewalk u p v" and "ewalk_cost f p < path_\<mu>"
    using p_props less_path_\<mu> by simp_all
  then show ?thesis
  proof (induct rule: ewalk_to_epath_induct)
    case (path p) then have "epath u p v" by (auto simp: epath_def)
    then show ?case using path.prems by (auto simp: path_\<mu>_def dest: not_mem_less_INF)
  next
    case (decomp p q r s)
    then obtain w where p_props: "p = q @ r @ s" "ewalk u q w" "ewalk w r w" "ewalk w s v"
      by (auto elim: cycle_decompE)
    then have "ewalk u (q @ s) v"
      using `ewalk u p v` by (auto simp: ewalk_ewalk_joinI)
    then have verts_ss: "set (ewalk_verts u (q @ s)) \<subseteq> set (ewalk_verts u p)"
      using `ewalk u p v` `p = q @ r @ s`
      by (auto simp add: ewalk_verts_conv)

    show ?case
    proof (cases "ereal (ewalk_cost f (q @ s)) < path_\<mu>")
      case True then have "\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u (q @ s)) \<and> ewalk_cost f c < 0"
        by (rule decomp)
      then show ?thesis using verts_ss by auto
    next
      case False
      note `ewalk_cost f p < path_\<mu>`
      also have "path_\<mu> \<le> ewalk_cost f (q @ s)"
        using False by simp
      finally have "ewalk_cost f r < 0" using `p = q @ r @ s` by simp
      moreover
      have "w \<in> set (ewalk_verts u q)" using `ewalk u q w` by auto
      then have "w \<in> set (ewalk_verts u p)"
        using `ewalk u p v` `ewalk u q w` `p = q @ r @ s`
        by (auto simp: ewalk_verts_conv)    
      ultimately
      show ?thesis using `ewalk w r w` by auto
    qed
  qed
qed

(*XXX rename*)
lemma (in pseudo_digraph) neg_inf_imp_neg_cyc:
  assumes inf_mu: "\<mu> f u v = - \<infinity>"
  shows "\<exists>p. ewalk u p v \<and> (\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u p) \<and> ewalk_cost f c < 0)"
proof -
  def path_\<mu> \<equiv> "INF s:{p. epath u p v}. ereal (ewalk_cost f s)"

  have ewalks_ne: "{p. ewalk u p v} \<noteq> {}"
    using inf_mu unfolding \<mu>_def INF_def by safe (simp add: top_ereal_def)
  then have paths_ne: "{p. epath u p v} ~= {}"
    by (auto intro: epath_ewalk_to_epath)

  obtain p where "epath u p v" "ewalk_cost f p = path_\<mu>"
  proof -
    obtain p where "p \<in> {p. epath u p v}" "ewalk_cost f p = path_\<mu>"
    using finite_INF_in[OF epaths_finite paths_ne, of "ewalk_cost f"]
      by (auto simp: path_\<mu>_def)
    then show ?thesis using that by auto
  qed
  then have "path_\<mu> \<noteq> -\<infinity>" by auto
  then have "\<mu> f u v < path_\<mu>" using inf_mu by simp
  then obtain pw where p_def: "ewalk u pw v" "ewalk_cost f pw < path_\<mu>"
    by atomize_elim (auto simp: \<mu>_def INF_less_iff)
  then have "\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u pw) \<and> ewalk_cost f c < 0"
    by (intro walk_cheaper_path_imp_neg_cyc) (auto simp: path_\<mu>_def)
  with `ewalk u pw v` show ?thesis by auto
qed

lemma (in pseudo_digraph) no_neg_cyc_imp_no_neg_inf:
  assumes no_neg_cyc: "\<And>p. ewalk u p v
    \<Longrightarrow> \<not>(\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u p) \<and> ewalk_cost f c < 0)"
  shows "- \<infinity> < \<mu> f u v"
proof (intro ereal_MInfty_lessI notI)
  assume "\<mu> f u v = - \<infinity>"
  then obtain p where p_props: "ewalk u p v"
    and ex_cyc: "\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u p) \<and> ewalk_cost f c < 0"
    by atomize_elim (rule neg_inf_imp_neg_cyc)
  then show False using no_neg_cyc by blast
qed

lemma \<mu>_reach_conv:
  "\<mu> f u v < \<infinity> \<longleftrightarrow> u \<rightarrow>*G v"
proof
  assume "\<mu> f u v < \<infinity>"
  then have "{p. ewalk u p v} \<noteq> {}"
    unfolding \<mu>_def INF_def by safe (simp add: top_ereal_def)
  then show "u \<rightarrow>*G v" by (simp add: reachable_ewalk)
next
  assume "u \<rightarrow>*G v"
  then obtain p where p_props: "epath u p v"
    by (metis reachable_ewalk epath_ewalk_to_epath)
  then have "{p} \<subseteq> {p. epath u p v}" by simp
  then have "\<mu> f u v \<le> (INF p: {p}. ewalk_cost f p)"
    unfolding \<mu>_def by (intro INF_superset_mono) (auto simp: epath_def)
  also have "\<dots> < \<infinity>" unfolding INF_def by (simp add: min_def)
  finally show "\<mu> f u v < \<infinity>" .
qed

lemma ewalk_to_path_no_neg_cyc_cost:
  assumes p_props:"ewalk u p v"
  assumes no_neg_cyc: "\<not> (\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u p) \<and> ewalk_cost f c < 0)"
  shows "ewalk_cost f (ewalk_to_epath p) \<le> ewalk_cost f p"
using assms
proof (induct rule: ewalk_to_epath_induct)
  case path then show ?case by (auto simp: ewalk_to_epath.simps)
next
  case (decomp p q r s)
  from decomp(2,3) have "cycle_decomp_prop p (q,r,s)"
    using cycle_decomp_has_prop[OF decomp(1)] by auto
  then have decomp_props: "p = q @ r @ s" "\<exists>w. ewalk w r w" by auto

  have "ewalk_cost f (ewalk_to_epath p) = ewalk_cost f (ewalk_to_epath (q @ s))"
    using decomp by (auto simp: step_ewalk_to_epath[of _ p _ q r s])
  also have "\<dots> \<le> ewalk_cost f (q @ s)"
  proof -
    have "ewalk u (q @ s) v"
      using `ewalk u p v` decomp_props by (auto intro: ewalk_omit_closed_w)
    then have "set (ewalk_verts u (q @ s)) \<subseteq> set (ewalk_verts u p)"
      using `ewalk u p v` `p = q @ r @ s`
      by (auto simp add: ewalk_verts_conv)
    then show ?thesis using decomp.prems by (intro decomp.hyps) auto
  qed
  also have "\<dots> \<le> ewalk_cost f p"
  proof -
    obtain w where "ewalk u q w" "ewalk w r w" "ewalk w s v"
      using decomp by (auto elim: cycle_decompE)
    then have "w \<in> set (ewalk_verts u q)" by auto
    then have "w \<in> set (ewalk_verts u p)"
      using `p = q @ r @ s` `ewalk u p v` `ewalk u q w`
      by (auto simp add: ewalk_verts_conv)
    then have "0 \<le> ewalk_cost f r" using `ewalk w r w`
      using decomp.prems by (auto simp: not_less)
    then show ?thesis using `p = q @ r @ s` by simp
  qed
  finally show ?case .
qed

lemma (in pseudo_digraph) no_neg_cyc_reach_imp_path:
  assumes reach: "u \<rightarrow>*G v"
(* XXX: "ewhd x c" ist unschön *)
  assumes no_neg_cyc: "\<And>p. ewalk u p v
    \<Longrightarrow> \<not>(\<exists>w c. ewalk w c w \<and> w \<in> set (ewalk_verts u p) \<and> ewalk_cost f c < 0)"
  shows "\<exists>p. epath u p v \<and> \<mu> f u v = ewalk_cost f p"
proof -
  def set_walks \<equiv> "{p. ewalk u p v}"
  def set_paths \<equiv> "{p. epath u p v}"

  have "set_paths \<noteq> {}"
  proof -
    obtain p where "epath u p v"
      using reach by (metis epath_ewalk_to_epath reachable_ewalk)
    then show ?thesis unfolding set_paths_def by blast
  qed

  have "\<mu> f u v = (INF p: set_walks. ereal (ewalk_cost f p))"
    unfolding \<mu>_def set_walks_def by simp
  also have "\<dots> = (INF p: set_paths. ereal (ewalk_cost f p))"
  proof (rule antisym)
    have "ewalk_to_epath ` set_walks \<subseteq> set_paths"
      unfolding set_walks_def set_paths_def
      by (intro subsetI) (auto elim: epath_ewalk_to_epath)
    then have "(INF p: set_paths. ewalk_cost f p)
      \<le> (INF p: ewalk_to_epath ` set_walks. ewalk_cost f p)"
      by (rule INF_superset_mono) simp
    also have "\<dots> = (INF p: set_walks. ewalk_cost f (ewalk_to_epath p))"
      by (auto simp: INF_def intro!: arg_cong[where f="Inf"])
    also have "\<dots> \<le> (INF p: set_walks. ewalk_cost f p)"
    proof -
      { fix p assume "p \<in> set_walks"
        then have "ewalk u p v" by (auto simp: set_walks_def)
        then have "ewalk_cost f (ewalk_to_epath p) \<le> ewalk_cost f p"
          using no_neg_cyc
          using no_neg_cyc and ewalk_to_path_no_neg_cyc_cost
          by auto }
      then show ?thesis by (intro INF_mono) auto
    qed
    finally show
      "(INF p: set_paths. ereal (ewalk_cost f p))
      \<le> (INF p: set_walks. ereal (ewalk_cost f p))" by simp

    have "set_paths \<subseteq> set_walks"
      unfolding set_paths_def set_walks_def by (auto simp: epath_def)
    then show "(INF p: set_walks. ereal (ewalk_cost f p))
      \<le> (INF p: set_paths. ereal (ewalk_cost f p))"
      by (rule INF_superset_mono) simp
  qed
  also have "\<dots> \<in> (\<lambda>p. ereal (ewalk_cost f p)) ` set_paths"
    using epaths_finite `set_paths \<noteq> {}`
    by (intro finite_INF_in) (auto simp: set_paths_def)
  finally show ?thesis
    by (simp add: set_paths_def image_def)
qed

(*XXX: move shortest_paths*)
lemma (in pseudo_digraph) min_cost_ewalk:
  assumes reach: "u \<rightarrow>*G v"
  assumes pos_cost: "\<And>e. e \<in> edges G \<Longrightarrow> c e \<ge> 0"
  shows "\<exists>p. epath u p v \<and> \<mu> c u v = ewalk_cost c p"
proof -
  have pc: "\<And>u p v. ewalk u p v \<Longrightarrow> 0 \<le> ewalk_cost c p"
    using pos_cost_pos_ewalk_cost pos_cost by auto 

  from reach show ?thesis
    by (rule no_neg_cyc_reach_imp_path) (auto simp: not_less intro: pc)
qed

lemma (in pseudo_digraph) pos_cost_mu_triangle:
  assumes pos_cost: "\<And>e. e \<in> edges G \<Longrightarrow> c e \<ge> 0" (* introduce predicate? *)
  assumes e_props: "edge_to_ends G e = (u,v)" "e \<in> edges G"
  shows "\<mu> c s v \<le> \<mu> c s u + c e"
proof cases
  assume "\<mu> c s u = \<infinity>" then show ?thesis by simp
next
  assume "\<mu> c s u \<noteq> \<infinity>"
  then have "{p. ewalk s p u} \<noteq> {}"
    unfolding \<mu>_def INF_def by safe (simp add: top_ereal_def)
  then have "s \<rightarrow>*G u" by (simp add: reachable_ewalk)
  with pos_cost
  obtain p where p_props: "epath s p u" 
      and p_cost: "\<mu> c s u = ewalk_cost c p"
    by (metis min_cost_ewalk)

  have "ewalk u [e] v"
    using e_props by (auto simp: edge_to_ends_def ewalk_Cons_iff ewalk_empty_iff target_in_verts)
  with `epath s p u`
  have "ewalk s (p @ [e]) v"
    by (auto simp: epath_def ewalk_ewalk_joinI)
  then have "\<mu> c s v \<le> ewalk_cost c (p @ [e])"
    by (rule min_cost_le_walk_cost)
  also have "\<dots> \<le> ewalk_cost c p + c e" by simp
  also have "\<dots> \<le> \<mu> c s u + c e" using p_cost by simp
  finally show ?thesis .
qed


lemma (in pseudo_digraph) mu_exact_triangle:
  assumes "v \<noteq> s"
  assumes "s \<rightarrow>*G v"
  assumes nonneg_edges: "\<And>e. e \<in> edges G \<Longrightarrow> 0 \<le> c e"
  obtains u e where "\<mu> c s v = \<mu> c s u + c e" and "edge e (u,v)"
proof -
  obtain p where p_path: "epath s p v"
    and p_cost: "\<mu> c s v = ewalk_cost c p"
    using assms by (metis min_cost_ewalk)
  then obtain e p' where p'_props: "p = p' @ [e]" using `v \<noteq> s`
    by (cases p rule: rev_cases) (auto simp: epath_def)
  then obtain u where "ewalk s p' u" "ewalk u [e] v"
    using `epath s p v` by (auto simp: epath_def ewalk_join_decomp)
  then have mu_le: "\<mu> c s v \<le> \<mu> c s u + c e" and edge: "edge e (u,v)"
    using nonneg_edges by (auto intro!: pos_cost_mu_triangle simp: edge_to_ends_def ewalk_verts_Cons edge_def)

  have "\<mu> c s u + c e \<le> ereal (ewalk_cost c p') + ereal (c e)"
    using `ewalk s p' u`
    by (fast intro: add_right_mono min_cost_le_walk_cost)
  also have "\<dots> = ewalk_cost c p" using p'_props by simp
  also have "\<dots> = \<mu> c s v" using p_cost by simp
  finally
  have "\<mu> c s v = \<mu> c s u + c e" using mu_le by auto
  then show ?thesis using edge ..
qed

lemma (in pseudo_digraph) mu_exact_triangle_Ex:
  assumes "v \<noteq> s"
  assumes "s \<rightarrow>*G v"
  assumes "\<And>e. e \<in> edges G \<Longrightarrow> 0 \<le> c e"
  shows "\<exists>u e. \<mu> c s v = \<mu> c s u + c e \<and> edge e (u,v)"
using assms by (metis mu_exact_triangle)

lemma (in pseudo_digraph) mu_Inf_triangle:
  assumes "v \<noteq> s"
  assumes "\<And>e. e \<in> edges G \<Longrightarrow> 0 \<le> c e"
  shows "\<mu> c s v = Inf {\<mu> c s u + c e | u e. edge e (u, v)}" (is "_ = Inf ?S")
proof cases
  assume "s \<rightarrow>*G v"
  then obtain u e where "\<mu> c s v = \<mu> c s u + c e" "edge e (u,v)"
    using assms by (metis mu_exact_triangle)
  then have "Inf ?S \<le> \<mu> c s v" by (auto intro: Complete_Lattices.Inf_lower)
  also have "\<dots> \<le> Inf ?S" using assms(2)
    by (auto intro!: Complete_Lattices.Inf_greatest pos_cost_mu_triangle
      simp: edge_def edge_to_ends_def)
  finally show ?thesis by simp
next
  assume "\<not>s \<rightarrow>*G v"
  then have "\<mu> c s v = \<infinity>" by (metis shortest_path_inf)
  def S \<equiv> "?S"
  show "\<mu> c s v = Inf S"
  proof cases
    assume "S = {}" then show ?thesis unfolding `\<mu> c s v = \<infinity>`
      by (simp add: top_ereal_def)
  next
    assume "S \<noteq> {}"
    { fix x assume "x \<in> S"
      then obtain u e where "edge e (u,v)" and x_val: "x = \<mu> c s u + c e"
        unfolding S_def by auto
      then have "\<not>s \<rightarrow>*G u" using `\<not> s \<rightarrow>*G v` by (metis reachable_cons)
      then have "\<mu> c s u + c e= \<infinity>" by (simp add: shortest_path_inf)
      then have "x = \<infinity>" using x_val by simp }
    then have "S = {\<infinity>}" using `S \<noteq> {}` by auto
    then show ?thesis using `\<mu> c s v = \<infinity>` by (simp add: min_def)
  qed
qed

end


end