theory Stuff
imports
  "~~/src/HOL/Main"
  "~~/src/HOL/Library/Extended_Real"
  "~~/src/HOL/Number_Theory/Binomial"
begin

section {* Possible additions to ereal *}

lemma ereal_divide_right_mono:
  fixes a b c :: ereal
  assumes "a \<le> b" "0 < c"
  shows "a / c \<le> b / c"
using assms by (cases a b c rule: ereal3_cases) (auto intro: divide_right_mono)

lemma ereal_divide_left_mono:
  fixes a b c :: ereal
  assumes "b \<le> a" "0 < c" "0 < a * b"
  shows "c / a \<le> c / b"
using assms by (cases a b c rule: ereal3_cases)
  (auto intro: divide_left_mono simp: field_simps sign_simps split: split_if_asm)

lemma ereal_fin_div_eq_inf_iff:
  fixes a b :: ereal
  assumes "0 < a" "a < \<infinity>"
  shows "a / b = \<infinity> \<longleftrightarrow> b = 0"
  using assms by (auto simp: divide_ereal_def)

lemma ereal_add_le_cancel_right:
  assumes "\<bar>c\<bar> < (\<infinity> :: ereal)"
  shows "a + c \<le> b + c = (a \<le> b)"
  using assms
by (cases a b c rule: ereal3_cases) auto

lemma ereal_Sup_finite_Max:
  fixes S :: "ereal set"
  assumes "finite S" and "S \<noteq> {}"
  shows "Sup S = Max S"
using assms
by (induct S rule: finite_ne_induct) (auto simp: max_absorb1)

lemma ereal_Inf_finite_Min:
  fixes S :: "ereal set"
  assumes "finite S" and "S \<noteq> {}"
  shows "Inf S = Min S"
using assms
by (induct S rule: finite_ne_induct) (auto simp: min_absorb1)

lemma finite_INF_in:
  fixes f :: "'a \<Rightarrow> ereal"
  assumes "finite S"
  assumes "S \<noteq> {}"
  shows "(INF s: S. f s) \<in> f ` S"
proof (unfold INF_def)
  from assms
  have "finite (f ` S)" "f ` S \<noteq> {}" by auto
  then show "Inf (f ` S) \<in> f ` S" using ereal_Inf_finite_Min Min_in by metis
qed


section {* Possible additions to enat *}

(*XXX move / keep?*)
lemma enat_setsum: "enat (setsum f A) = (\<Sum>x\<in>A. enat (f x))"
  by (simp add: of_nat_eq_enat[symmetric] of_nat_setsum)

(*XXX rename / move *)
lemma ereal_of_enat_inject: "ereal_of_enat a = ereal_of_enat b \<longleftrightarrow> a = b"
  by (cases a b rule: enat2_cases) auto

(*XXX rename / move *)
lemma ereal_of_enat_inf: "ereal_of_enat a = \<infinity> \<longleftrightarrow> a = \<infinity>"
  by (cases a) auto
(*XXX add: ereal_of_enat rule for numerals? *)

(*XXX rename / move *)
lemma ereal_of_enat_less_iff: "ereal_of_enat a < ereal_of_enat b \<longleftrightarrow> a < b"
  by (cases a b rule: enat2_cases) auto

lemma enat_in_Inf:
  fixes S :: "enat set"
  assumes "Inf S \<noteq> top"
  shows "Inf S \<in> S"
proof (rule ccontr)
  assume A: "~?thesis"

  obtain n where Inf_conv: "Inf S = enat n" using assms by (auto simp: top_enat_def)
  { fix s assume "s \<in> S"
    then have "Inf S \<le> s" by (rule complete_lattice_class.Inf_lower)
    moreover have "Inf S \<noteq> s" using A `s \<in> S` by auto
    ultimately have "Inf S < s" by simp
    with Inf_conv have "enat (Suc n) \<le> s" by (cases s) auto
  }
  then have "enat (Suc n) \<le> Inf S" by (simp add: le_Inf_iff)
  with Inf_conv show False by auto
qed

lemma enat_in_INF:
  fixes f :: "'a \<Rightarrow> enat"
  assumes "(INF x: S. f x) \<noteq> top"
  obtains x where "x \<in> S" and "(INF x: S. f x) = f x"
proof -
  from assms have "(INF x: S. f x) \<in> f ` S"
    unfolding INF_def using enat_in_Inf by blast
  then obtain x where "x \<in> S" "(INF x: S. f x) = f x" by auto
  then show ?thesis ..
qed

lemma enat_less_INF_I:
  fixes f :: "'a \<Rightarrow> enat"
  assumes not_inf: "x \<noteq> \<infinity>" and less: "\<And>y. y \<in> S \<Longrightarrow> x < f y"
  shows "x < (INF y:S. f y)"
proof (rule ccontr)
  assume A: "\<not>?thesis"
  then show False
  proof (cases "x = (INF y:S. f y)")
    case True
    with assms have "(INF y:S. f y) \<noteq> top" by (simp add: top_enat_def)
    then obtain z where "z \<in> S" "f z = x"
      by (rule enat_in_INF) (auto simp: True)
    then show False using less by auto
  next
    case False
    with A have "(INF y:S. f y) < x" by simp
    with less show False by (auto simp: INF_less_iff intro: order_less_asym)
  qed
qed

(*XXX: Move to Nat_Infinity *)
lemma enat_le_Sup_iff:
  "enat k \<le> Sup M \<longleftrightarrow> k = 0 \<or> (\<exists>m \<in> M. enat k \<le> m)" (is "?L \<longleftrightarrow> ?R")
proof cases
  assume "k = 0" then show ?thesis by (auto simp: enat_0)
next
  assume "k \<noteq> 0"
  show ?thesis
  proof
    assume ?L
    then have "\<lbrakk>enat k \<le> (if finite M then Max M else \<infinity>); M \<noteq> {}\<rbrakk> \<Longrightarrow> \<exists>m\<in>M. enat k \<le> m"
      by (metis Max_in Sup_enat_def finite_enat_bounded linorder_linear)
    with `k \<noteq> 0` and `?L` show ?R
      unfolding Sup_enat_def
      by (cases "M={}") (auto simp add: enat_0[symmetric])
  next
    assume ?R then show ?L
      by (auto simp: enat_0 intro: complete_lattice_class.Sup_upper2)
  qed
qed

lemma enat_neq_zero_cancel_iff[simp]:
  "0 \<noteq> enat n \<longleftrightarrow> 0 \<noteq> n"
  "enat n \<noteq> 0 \<longleftrightarrow> n \<noteq> 0"
by (auto simp: enat_0[symmetric])


section {* Possible additions to Fact? *}

lemma fact_div_fact_le_pow:
  assumes "r \<le> n" shows "fact n div fact (n - r) \<le> n ^ r"
proof -
  have "\<And>r. r \<le> n \<Longrightarrow> \<Prod>{n - r..n} = (n - r) * \<Prod>{Suc (n - r)..n}"
    by (subst setprod.insert[symmetric]) (auto simp: atLeastAtMost_insertL)
  with assms show ?thesis
    by (induct r rule: nat.induct) (auto simp add: fact_div_fact Suc_diff_Suc mult_le_mono)
qed

(* XXX: replace proof of choose_le_pow in library *)
lemma choose_le_pow:
  assumes "r \<le> n" shows "n choose r \<le> n ^ r"
proof -
  have "n choose r \<le> fact n div fact (n - r)"
    using `r \<le> n` by (simp add: binomial_altdef_nat div_le_mono2)
  with fact_div_fact_le_pow[OF assms] show ?thesis by auto
qed


section {* Possible additions to Finite Set? *}

(*XXX move *)
lemma list_set_tl: "x \<in> set (tl xs) \<Longrightarrow> x \<in> set xs"
by (cases xs) auto

lemma lists_length_Suc_eq:
  "{xs. set xs \<subseteq> A \<and> length xs = Suc n} =
    (\<lambda>(xs, n). n#xs) ` ({xs. set xs \<subseteq> A \<and> length xs = n} \<times> A)"
  by (auto simp: length_Suc_conv)



section {* Possible additions to RComplete *}

lemma natceiling_lessD:
  assumes "natceiling x < n" shows "x < real n"
by (smt assms nat_less_real_le natceiling_le_eq)

(* XXX - move? RealDef.natfloor_le_eq *)
lemma natfloor_le_iff:
  fixes n :: nat and r :: real
  assumes "r \<le> n"
  shows "natfloor r \<le> n"
proof -
  from assms have "floor r \<le> n" by (simp add: floor_le_eq)
  then show ?thesis by (simp add: natfloor_def)
qed

(* XXX - move? RealDef.le_natceiling_eq *)
lemma le_natceiling_iff:
  fixes n :: nat and r :: real
  assumes "n \<le> r"
  shows "n \<le> natceiling r"
proof -
  from assms have "n \<le> ceiling r" by (simp add: le_ceiling_eq)
  then show ?thesis by (simp add: natceiling_def)
qed

(* XXX - move? RealDef.le_natceiling_eq *)
lemma natceiling_le_iff:
  fixes n :: nat and r :: real
  assumes "r \<le> n"
  shows "natceiling r \<le> n"
proof -
  from assms have "ceiling r \<le> n" by (simp add: ceiling_le_eq)
  then show ?thesis by (simp add: natceiling_def)
qed

lemma dist_real_noabs_less:
  fixes a b c :: real assumes "dist a b < c" shows "a - b < c"
using assms by (simp add: dist_real_def)


section {* Possible additions at a later time? *}

subsection {* Lim / Limits *}

lemma eventually_le_le:
  fixes P :: "'a => ('b :: preorder)"
  assumes "eventually (\<lambda>x. P x \<le> Q x) net"
  assumes "eventually (\<lambda>x. Q x \<le> R  x) net"
  shows "eventually (\<lambda>x. P x \<le> R x) net"
using assms by (rule eventually_elim2) (rule order_trans)

lemma eventually_le_less:
  fixes P :: "'a => ('b :: preorder)"
  assumes "eventually (\<lambda>x. P x \<le> Q x) net"
  assumes "eventually (\<lambda>x. Q x < R  x) net"
  shows "eventually (\<lambda>x. P x < R x) net"
using assms by (rule eventually_elim2) (rule le_less_trans)

lemma eventually_less_le:
  fixes P :: "'a => ('b :: preorder)"
  assumes "eventually (\<lambda>x. P x < Q x) net"
  assumes "eventually (\<lambda>x. Q x \<le> R  x) net"
  shows "eventually (\<lambda>x. P x < R x) net"
using assms by (rule eventually_elim2) (rule less_le_trans)

lemma eventually_less_less:
  fixes P :: "'a => ('b :: preorder)"
  assumes "eventually (\<lambda>x. P x < Q x) net"
  assumes "eventually (\<lambda>x. Q x < R  x) net"
  shows "eventually (\<lambda>x. P x < R x) net"
using assms by (rule eventually_elim2) (rule less_trans)

(*XXX - do not make trans?*)
lemmas eventually_trans[trans] =
  eventually_le_le eventually_le_less eventually_less_le eventually_less_less

subsection {* List *}

lemma not_distinct_decomp_min:
  assumes "\<not> distinct ws"
  shows "\<exists> xs ys zs y. ws = xs @ y # ys @ y # zs \<and> distinct xs \<and> y \<notin> set xs \<and> y \<notin> set ys "
proof -
  obtain xs y ys where "y \<in> set xs" "distinct xs" "ws = xs @ y # ys"
    using assms by (auto simp: not_distinct_conv_prefix)
  moreover then obtain xs' ys' where "xs = xs' @ y # ys'" by (auto simp: in_set_conv_decomp)
  ultimately show ?thesis by auto
qed

lemma list_exhaust3:
  obtains "xs = []" | x where "xs = [x]" | x y ys where "xs = x # y # ys"
by (metis list.exhaust)

lemma card_Ex_subset:
  "k \<le> card M \<Longrightarrow> \<exists>N. N \<subseteq> M \<and> card N = k"
by (induct rule: inc_induct) (auto simp: card_Suc_eq)


subsection {* Log *}

lemma eventually_sequentially_lessI:
  assumes "\<And>x. c < x \<Longrightarrow> P x"
  shows "eventually P sequentially" 
unfolding eventually_sequentially
by (rule exI[where x="Suc c"]) (auto intro: assms)

lemma LIMSEQ_neg_powr:
  assumes s: "s < 0"
  shows "(%x. (real x) powr s) ----> 0"
by (rule tendsto_neg_powr[OF assms filterlim_real_sequentially])


lemma LIMSEQ_inv_powr:
  assumes "0 < c" "0 < d"
  shows "(\<lambda>n :: nat. (c / n) powr d) ----> 0"
proof (rule tendsto_zero_powrI)
  from `0 < c` have "\<And>x. 0 < x \<Longrightarrow> 0 < c / x" by (rule divide_pos_pos) 
  then show "eventually (\<lambda>x. 0 < c / real x) sequentially"
     by (rule eventually_sequentiallyI[of 1]) simp

  show "(\<lambda>x. c / real x) ----> 0"
  proof (rule tendstoI)
    fix e :: real assume "0 < e"
    then have "\<And>x. 0 < x \<Longrightarrow> c / x < e \<longleftrightarrow> c / e < x"
      by (auto simp: field_simps)
    then show "eventually (\<lambda>x. dist (c / real x) 0 < e) sequentially"
      using `0 < c` `0 < e`
      by (intro eventually_sequentially_lessI[of "natceiling (c/e)"])
         (auto simp: dist_real_def natceiling_lessD)
  qed
  show "0 < d" by (rule assms)
qed


end
