header {* Walks based on vertices *}

theory Vertex_Walks
imports Digraphs
begin

text {*
  These definitions are here mainly for historical purposes, as they
  do not really work with multigraphs. Consider using "ewalk" instead.
*}


section {* Basic lemmas *}
lemma vwalkI[intro]:
  assumes "set p \<subseteq> verts G"
  assumes "set (vwalk_edges p) \<subseteq> edges_ends G"
  assumes "p \<noteq> []"
  shows "vwalk p G"
using assms by (auto simp add: vwalk_def)

lemma vwalkE[elim]:
  assumes "vwalk p G"
  assumes "set p \<subseteq> verts G \<Longrightarrow>
    set (vwalk_edges p) \<subseteq> edges_ends G \<and> p \<noteq> [] \<Longrightarrow> P"
  shows "P"
using assms by (simp add: vwalk_def)

lemma pathI[intro]:
  assumes "vwalk p G"
  assumes "distinct p"
  shows "path p G"
using assms by (simp add: path_def)

lemma pathE[elim]:
  assumes "path p G"
  assumes "vwalk p G \<Longrightarrow> distinct p \<Longrightarrow> P"
  shows "P"
using assms by (simp add: path_def)

(* XXX: move *)
lemma cycle_nonempty:
  assumes "cycle c G"
  shows "c \<noteq> []"
using assms by (auto simp add: cycle_def)

lemma cycle_hd_last:
  assumes "cycle c G"
  shows "last c = hd c"
using assms by (simp add: cycle_def)

lemma cycle_vwalk[simp]:
  assumes "cycle c G"
  shows "vwalk c G"
using assms by (auto simp add: cycle_def)

section {* lemmas *}

subsection {* lemmas about paths and vwalks *}

lemma vwalk_consI:
  assumes "vwalk p G"
  assumes "a \<in> verts G"
  assumes "(a, hd p) \<in> edges_ends G"
  shows "vwalk (a # p) G"
using assms by (cases p) (auto simp add: vwalk_def)

lemma vwalk_consE:
  assumes "vwalk (a # p) G"
  assumes "p \<noteq> []"
  assumes "(a, hd p) \<in> edges_ends G \<Longrightarrow> vwalk p G \<Longrightarrow> P"
  shows "P"
using assms by (cases p) (auto simp add: vwalk_def)

(*XXX move*)
lemma vwalk_induct[case_names Base Cons, induct pred: vwalk]:
  assumes "vwalk p G"
  assumes "\<And>u. u \<in> verts G \<Longrightarrow> P [u]"
  assumes "\<And>u v es. (u,v) \<in> edges_ends G \<Longrightarrow> P (v # es) \<Longrightarrow> P (u # v # es)"
  shows "P p"
  using assms
proof (induct p)
  case (Cons u es)
  then show ?case
  proof (cases es)
    fix v es' assume "es = v # es'"
    then have "(u,v) \<in> edges_ends G" and "P (v # es')"
      using Cons by (auto elim: vwalk_consE)
    then show ?thesis using `es = v # es'` Cons.prems by auto
  qed auto
qed auto

lemma vwalk_edges_Cons[simp]:
  assumes "p \<noteq> []"
  shows "vwalk_edges (u # p) = (u, hd p) # vwalk_edges p"
using assms by (cases p) simp+

(* FIXME: change to use vwalk_join? *)
lemma vwalk_edges_append:
  assumes "p \<noteq> []" and "q \<noteq> []"
  shows "vwalk_edges (p @ q) = vwalk_edges p @ (last p, hd q) # vwalk_edges q" 
proof -
  from assms obtain a b p' q' where "p = a # p'" and "q = b # q'"
    by (auto simp add: neq_Nil_conv)
  moreover
  have "vwalk_edges ((a # p') @ (b # q'))
    = vwalk_edges (a # p') @ (last (a # p'), b) # vwalk_edges (b # q')" 
  proof (induct p')
    case Nil show ?case by simp
  next
    case (Cons a' p') then show ?case by (auto simp add: neq_Nil_conv)
  qed
  ultimately
  show ?thesis by auto
qed

lemma set_vwalk_edges_append1:
  "set (vwalk_edges p) \<subseteq> set (vwalk_edges (p @ q))" 
proof (cases p)
  case (Cons a p') note p_Cons = Cons then show ?thesis
  proof (cases q)
    case (Cons b q')
    with p_Cons  have "p \<noteq> []" and "q \<noteq> []" by auto
    then show ?thesis  by (auto simp add: vwalk_edges_append)
  qed simp
qed simp

lemma set_vwalk_edges_append2:
  "set (vwalk_edges q) \<subseteq> set (vwalk_edges (p @ q))" 
proof (cases p)
  case (Cons a p') note p_Cons = Cons then show ?thesis
  proof (cases q)
    case (Cons b q')
    with p_Cons have "p \<noteq> []" and "q \<noteq> []" by auto
    then show ?thesis by (auto simp add: vwalk_edges_append)
  qed simp
qed simp

lemma set_vwalk_edges_cons:
  "set (vwalk_edges p) \<subseteq> set (vwalk_edges (u # p))"
using set_vwalk_edges_append2[of p "[u]"] by simp

lemma set_vwalk_edges_snoc:
  assumes "p \<noteq> []"
  shows "set (vwalk_edges (p @ [a]))
    = insert (last p, a) (set (vwalk_edges p))"
using assms proof (induct p)
  case Nil then show ?case by auto
next
  case (Cons x xs)
  then show ?case
  proof (cases "xs = []")
    case True then show ?thesis by auto
  next
    case False
    have "set (vwalk_edges ((x # xs) @ [a]))
      = set (vwalk_edges (x # (xs @ [a])))"
      by auto
    then show ?thesis using Cons and False
      by (auto simp add: set_vwalk_edges_cons)
  qed
qed

(*XXX change to wellformed*)
lemma (in pseudo_digraph) vwalk_pseudo_digraph_consI:
  assumes "vwalk p G"
  assumes "(a, hd p) \<in> edges_ends G"
  shows "vwalk (a # p) G"
proof
  show "a # p \<noteq> []" by simp

  from assms have "a \<in> verts G" and "set p \<subseteq> verts G"
    using wellformed by auto
  then show "set (a # p) \<subseteq> verts G" by auto

  from `vwalk p G` have "p \<noteq> []" by auto
  then show "set (vwalk_edges (a # p)) \<subseteq> edges_ends G"
    using `vwalk p G` and `(a, hd p) \<in> edges_ends G`
    by (auto simp add: set_vwalk_edges_cons)
qed

(* XXX: this lemma seems hacky, but i need it to avoid "elim!: pathE vwalkE" in various proofs.
   XXX: check how th. automation can be improved instead ...  *)
lemma vwalk_imp_last_in_verts[intro]:
  assumes "vwalk p G"
  shows "last p \<in> verts G"
using assms by (auto elim!: pathE vwalkE)


lemma vwalk_append1:
  assumes "p \<noteq> []"
  assumes "vwalk (p @ q) G"
  shows "vwalk p G"
proof
  from assms show "p \<noteq> []" and "set p \<subseteq> verts G"
    by (auto elim!: vwalkE)
  have "set (vwalk_edges p) \<subseteq> set (vwalk_edges (p @ q))"
    by (auto simp add: set_vwalk_edges_append1)
  then show "set (vwalk_edges p) \<subseteq> edges_ends G"
    using assms by blast
qed

lemma vwalk_append2:
  assumes "q \<noteq> []"
  assumes "vwalk (p @ q) G"
  shows "vwalk q G"
proof
  from `vwalk (p @ q) G` have "set (p @ q) \<subseteq> verts G" by blast
  then show "set q \<subseteq> verts G" by simp

  from `vwalk (p @ q) G` have "set (vwalk_edges (p @ q)) \<subseteq> edges_ends G"
    by blast
  then show "set (vwalk_edges q) \<subseteq> edges_ends G"
    by (metis set_vwalk_edges_append2 subset_trans)

  from `q \<noteq> []` show "q \<noteq> []" by assumption
qed

lemma vwalk_to_path_hd: "hd (vwalk_to_path xs) = hd xs"
proof (induct xs rule: vwalk_to_path.induct)
  case (2 x xs) then show ?case
  proof (cases "x \<in> set xs")
    case True
    then have "hd (dropWhile (\<lambda>y. y \<noteq> x) xs) = x"
      using hd_dropWhile[where P="\<lambda>y. y \<noteq> x"] by auto
    then show ?thesis using True and 2 by auto
  qed auto
qed auto

lemma vwalk_to_path_induct3[consumes 0, case_names base in_set not_in_set]:
  assumes "P []"
  assumes "\<And>x xs. x \<in> set xs \<Longrightarrow> P (dropWhile (\<lambda>y. y \<noteq> x) xs)
    \<Longrightarrow> P (x # xs)"
  assumes "\<And>x xs. x \<notin> set xs \<Longrightarrow> P xs \<Longrightarrow> P (x # xs)"
  shows "P xs"
using assms by (induct xs rule: vwalk_to_path.induct) auto

lemma vwalk_to_path_nonempty:
  assumes "p \<noteq> []"
  shows "vwalk_to_path p \<noteq> []"
using assms by (induct p rule: vwalk_to_path_induct3) auto

lemma vwalk_to_path_last:
  "last (vwalk_to_path xs) = last xs"
by (induct xs rule: vwalk_to_path_induct3)
   (auto simp add: dropWhile_last vwalk_to_path_nonempty)

lemma vwalk_to_path_subset:
  assumes "x \<in> set (vwalk_to_path xs)"
  shows "x \<in> set xs"
using assms proof (induct xs rule: vwalk_to_path.induct)
  case (2 x xs) then show ?case
    by (cases "x \<in> set xs") (auto dest: set_dropWhileD)
qed simp_all

lemma vwalk_to_path_cons:
  assumes "x \<notin> set xs"
  shows "vwalk_to_path (x # xs) = x # vwalk_to_path xs"
using assms by auto

lemma vwalk_to_path_path:
  assumes "vwalk p G"
  shows "path (vwalk_to_path p) G"
using assms proof (induct p rule: vwalk_to_path_induct3)
  case base then show ?case by auto
next
  case (in_set x xs)
  have set_neq: "\<And>x xs. x \<notin> set xs \<Longrightarrow> \<forall>x' \<in> set xs. x' \<noteq> x" by metis
  from `x \<in> set xs` obtain ys zs where "xs = ys @ x # zs" and "x \<notin> set ys"
    by (metis in_set_conv_decomp_first)
  then have vwalk_dW: "vwalk (dropWhile (\<lambda>y. y \<noteq> x) xs) G"
    using in_set and `xs = ys @ x # zs`
    by (auto simp add: dropWhile_append3 set_neq intro: vwalk_append2[where p="x # ys"])
  then show ?case using in_set
    by (auto simp add: vwalk_dW)
next
  case (not_in_set x xs)
  then have "x \<in> verts G" and x_notin: "x \<notin> set (vwalk_to_path xs)"
    by (auto simp add: vwalk_to_path_subset)

  from not_in_set show ?case
  proof (cases "xs")
    case Nil then show ?thesis using not_in_set.prems by auto
  next
    case (Cons x' xs')
    have "path (vwalk_to_path xs) G"
      apply (rule not_in_set)
      apply (rule vwalk_append2[where p="[x]"])
      using Cons and not_in_set
      by auto
    then have "vwalk (x # vwalk_to_path xs) G"
      apply (auto intro!: vwalk_consI simp add: vwalk_to_path_hd)
      using not_in_set
      apply -
      apply (erule vwalk_consE)
      using Cons
      apply (auto intro: `x \<in> verts G`)
      done
    then have "path (x # vwalk_to_path xs) G"
      apply (rule pathI)
      using `path (vwalk_to_path xs) G`
      using x_notin
      by auto
    then show ?thesis using not_in_set
      by (auto simp add: vwalk_to_path_cons)
  qed
qed

lemma vwalk_imp_ex_path:
  assumes "vwalk p G"
  assumes "hd p = u"
  assumes "last p = v"
  shows "\<exists>q. path q G \<and> hd q = u \<and> last q = v"
by (metis assms vwalk_to_path_hd vwalk_to_path_last vwalk_to_path_path)


lemma vwalk_edges_set_nil:
  assumes "x \<in> set (vwalk_edges p)"
  shows "p \<noteq> []"
using assms by fastforce

lemma vwalk_edges_appendI1:
  assumes "x \<in> set (vwalk_edges p) \<or> x \<in> set (vwalk_edges q)"
  shows "x \<in> set (vwalk_edges (p @ q))"
using assms proof
  assume "x \<in> set (vwalk_edges p)"
  then show "x \<in> set (vwalk_edges (p @ q))"
    by (cases "q = []")
       (auto simp add: vwalk_edges_append vwalk_edges_set_nil)
next
  assume "x \<in> set (vwalk_edges q)"
  then show "x \<in> set (vwalk_edges (p @ q))"
    by (cases "p = []")
       (auto simp add: vwalk_edges_append vwalk_edges_set_nil)
qed

lemma vwalk_edges_appendI2:
  assumes nonempty: "p \<noteq> []" "q \<noteq> []"
  assumes disj: "x \<in> set (vwalk_edges p) \<or> x = (last p, hd q)
    \<or> x \<in> set (vwalk_edges q)"
  shows "x \<in> set (vwalk_edges (p @ q))"
using disj proof (elim disjE)
  assume "x = (last p, hd q)"
  then show "x \<in> set (vwalk_edges (p @ q))"
    by (metis nonempty in_set_conv_decomp vwalk_edges_append)
qed (auto intro: vwalk_edges_appendI1)

lemma edges_in_vwalk_edges:
  assumes "u \<in> set (vwalk_edges p)"
  shows "u \<in> set p \<times> set p"
using assms by (induct p rule: vwalk_edges.induct) auto

lemma set_vwalk_edges_rev:
  "set (vwalk_edges (rev p)) = {(v, u). (u,v) \<in> set (vwalk_edges p)}"
proof (induct p)
  case Nil then show ?case by auto
next
  case (Cons x xs)
  then show ?case
  proof (cases "xs = []")
    case True then show ?thesis by auto
  next
    case False
    then have "set (vwalk_edges (rev (x # xs))) = {(hd xs, x)}
      \<union> {a. case a of (v, u) \<Rightarrow> (u, v) \<in> set (vwalk_edges xs)}"
      by (simp add: set_vwalk_edges_snoc last_rev Cons)
    also have "\<dots> = {a. case a of (v, u) \<Rightarrow> (u, v) \<in> set (vwalk_edges (x # xs))}"
      using False by (auto simp add: set_vwalk_edges_cons)
    finally show ?thesis by assumption
  qed
qed

lemma path_self:
  assumes "u \<in> verts G"
  shows "path [u] G"
using assms by (intro pathI vwalkI, auto)

lemma vwalk_verts_in_verts:
  assumes "vwalk p G"
  assumes "u \<in> set p"
  shows "u \<in> verts G"
using assms by auto

lemma vwalk_edges_tl:
  "vwalk_edges (tl xs) = tl (vwalk_edges xs)"
by (induct xs rule: vwalk_edges.induct) simp_all

lemma vwalk_edges_butlast:
  "vwalk_edges (butlast xs) = butlast (vwalk_edges xs)"
proof (induct xs rule: rev_induct)
  case (snoc x xs) thus ?case
  proof (cases "xs = []")
    case True with snoc show ?thesis by simp
  next
    case False
    hence "vwalk_edges (xs @ [x]) = vwalk_edges xs @ [(last xs, x)]" using vwalk_edges_append by force
    with snoc show ?thesis by simp
  qed
qed simp

lemma vwalk_edges_tl_empty:
  "vwalk_edges xs = [] \<Longrightarrow> vwalk_edges (tl xs) = []"
by (induct xs rule: vwalk_edges.induct) simp_all

lemma vwalk_edges_butlast_empty:
  "xs \<noteq> [] \<Longrightarrow> vwalk_edges xs = [] \<Longrightarrow> vwalk_edges (butlast xs) = []"
by (induct xs rule: vwalk_edges.induct) simp_all

section {* Joining paths *}

definition joinable :: "'a vwalk \<Rightarrow> 'a vwalk \<Rightarrow> bool" where
  "joinable p q \<equiv> last p = hd q \<and> p \<noteq> [] \<and> q \<noteq> []"

definition vwalk_join :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
  (infixr "\<oplus>" 65) where
  "p \<oplus> q \<equiv> p @ tl q"

lemma joinableI[intro,simp]:
  assumes "last p = hd q" "p \<noteq> []" "q \<noteq> []"
  shows "joinable p q"
using assms by (simp add: joinable_def)

lemma vwalk_join_non_Nil[simp]:
  assumes "p \<noteq> []"
  shows "p \<oplus> q \<noteq> []"
unfolding vwalk_join_def using assms by simp

(*XXX rename \<rightarrow> Cons*)
lemma vwalk_join_Cons[simp]:
  assumes "p \<noteq> []"
  shows "(u # p) \<oplus> q = u # p \<oplus> q"
unfolding vwalk_join_def using assms by simp

lemma vwalk_join_def2:
  assumes "joinable p q"
  shows "p \<oplus> q = butlast p @ q"
proof -
  from assms have "p \<noteq> []" and "q \<noteq> []" by (simp add: joinable_def)+
  then have "vwalk_join p q = butlast p @ last p # tl q"
    unfolding vwalk_join_def by simp
  then show ?thesis using assms by (simp add: joinable_def)
qed

lemma vwalk_join_hd':
  assumes "p \<noteq> []"
  shows "hd (p \<oplus> q) = hd p"
using assms by (auto simp add: vwalk_join_def)

lemma vwalk_join_hd:
  assumes "joinable p q"
  shows "hd (p \<oplus> q) = hd p"
using assms by (auto simp add: vwalk_join_def joinable_def)

lemma vwalk_join_last:
  assumes "joinable p q"
  shows "last (p \<oplus> q) = last q"
using assms by (auto simp add: vwalk_join_def2 joinable_def)

lemma vwalk_join_Nil[simp]:
  "p \<oplus> [] = p"
by (simp add: vwalk_join_def)

lemma vwalk_join_vwalk':
  assumes "vwalk p G"
  assumes "vwalk q G"
  assumes "last p = hd q"
  shows "vwalk (p \<oplus> q) G"
proof (unfold vwalk_join_def, rule vwalkI)
  have "set p \<subseteq> verts G" and "set q \<subseteq> verts G"
    using `vwalk p G` and `vwalk q G` by blast+
  then show "set (p @ tl q) \<subseteq> verts G"
    by (cases q) auto
next
  show "p @ tl q \<noteq> []" using `vwalk p G` by auto
next
  have pe_p: "set (vwalk_edges p) \<subseteq> edges_ends G"
    using `vwalk p G` by blast
  have pe_q': "set (vwalk_edges (tl q)) \<subseteq> edges_ends G"
  proof -
    have "set (vwalk_edges (tl q)) \<subseteq> set (vwalk_edges q)"
      by (cases q) (simp_all add: set_vwalk_edges_cons)
    then show ?thesis using `vwalk q G` by blast
  qed

  show "set (vwalk_edges (p @ tl q)) \<subseteq> edges_ends G"
  proof (cases "tl q")
    case Nil then show ?thesis using pe_p by auto
  next
    case (Cons x xs)
    then have nonempty: "p \<noteq> []" "tl q \<noteq> []"
      using `vwalk p G` by auto
    moreover
    have "(hd q, hd (tl q)) \<in> set (vwalk_edges q)"
      using `vwalk q G` Cons by  (cases q) auto
    ultimately show ?thesis
      using `vwalk q G`
      by (auto simp: pe_p pe_q' `last p = hd q` vwalk_edges_append)
  qed
qed

lemma vwalk_join_vwalk:
  assumes "vwalk p G"
  assumes "vwalk q G"
  assumes "joinable p q"
  shows "vwalk (p \<oplus> q) G"
using assms vwalk_join_vwalk' by (auto simp: joinable_def)

lemma vwalk_join_split:
  assumes "u \<in> set p"
  shows "\<exists>q r. p = q \<oplus> r
  \<and> last q = u \<and> hd r = u \<and> q \<noteq> [] \<and> r \<noteq> []"
proof -
  from `u \<in> set p`
  obtain pre_p post_p where "p = pre_p @ u # post_p"
    by atomize_elim (auto simp add: split_list)
  then have "p = (pre_p @ [u]) \<oplus> (u # post_p)"
    unfolding vwalk_join_def by simp
  then show ?thesis by fastforce
qed

(*XXX rename*)
lemma vwalk_join_vwalk1:
  assumes "p \<noteq> []"
  assumes "vwalk (p \<oplus> q) G"
  shows "vwalk p G"
using assms unfolding vwalk_join_def
by (auto intro: vwalk_append1)


(*XXX rename*)
lemma vwalk_join_vwalk2:
  assumes "joinable p q"
  assumes "vwalk (p \<oplus> q) G"
  shows "vwalk q G"
using assms
by (auto simp add: vwalk_join_def2 joinable_def intro: vwalk_append2)

lemma vwalk_join_assoc:
  assumes "p \<noteq> []" "q \<noteq> []"
  shows "(p \<oplus> q) \<oplus> r = p \<oplus> q \<oplus> r"
using assms by (simp add: vwalk_join_def)

lemma vwalk_join_assoc':
  assumes "joinable p q" "joinable q r"
  shows "(p \<oplus> q) \<oplus> r = p \<oplus> q \<oplus> r"
using assms by (simp add: vwalk_join_def joinable_def)

lemma joinable_vwalk_join_iff2:
  "joinable p (q \<oplus> r) \<longleftrightarrow> joinable p q \<or> (q = [] \<and> joinable p (tl r))"
by (cases q) (auto simp add: vwalk_join_def joinable_def)

lemma joinable_vwalk_join_iff1:
  assumes "joinable p q"
  shows "joinable (p \<oplus> q) r \<longleftrightarrow> joinable q r"
by (metis Nil_is_append_conv assms joinable_def vwalk_join_def vwalk_join_last)

lemmas joinable_simps =
  joinable_vwalk_join_iff1
  joinable_vwalk_join_iff2

lemma joinable_cyclic_omit:
  assumes "joinable p q" "joinable q r" "joinable q q"
  shows "joinable p r"
using assms by (metis joinable_def)

lemma joinable_non_Nil:
  assumes "joinable p q"
  shows "p \<noteq> []" "q \<noteq> []"
using assms by (simp_all add: joinable_def)

lemma vwalk_join_vwalk_length[simp]:
  assumes "joinable p q"
  shows "vwalk_length (p \<oplus> q) = vwalk_length p + vwalk_length q"
using assms unfolding vwalk_join_def
by (simp add: less_eq_Suc_le[symmetric] joinable_non_Nil)

lemma vwalk_join_edges1:
  assumes "set (vwalk_edges p) \<subseteq> E"
  assumes "p = q \<oplus> r"
  shows "set (vwalk_edges q) \<subseteq> E"
by (metis assms vwalk_join_def set_vwalk_edges_append1 subset_trans)

lemma vwalk_join_edges2:
  assumes "set (vwalk_edges p) \<subseteq> E"
  assumes "joinable q r"
  assumes "p = q \<oplus> r"
  shows "set (vwalk_edges r) \<subseteq> E"
using assms
by (metis vwalk_join_def2 set_vwalk_edges_append2 subset_trans)
(* XXX: nutzen um Beweis von vwalk_append2 zu vereinfachen? *)


subsubsection {* Concatenating paths *}

definition concat_path :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "concat_path p q \<equiv> vwalk_to_path (p \<oplus> q)"

lemma concat_path_is_path:
  assumes p_props: "vwalk p G" "hd p = u" "last p = v"
  assumes q_props: "vwalk q G" "hd q = v" "last q = w"
  shows "path (concat_path p q) G \<and> hd (concat_path p q) = u
    \<and> last (concat_path p q) = w"
proof (intro conjI)
  have joinable: "joinable p q" using assms by auto

  show "path (concat_path p q) G"
    unfolding concat_path_def using assms and joinable
    by (auto intro: vwalk_to_path_path vwalk_join_vwalk)

  show "hd (concat_path p q) = u" "last (concat_path p q) = w"
    unfolding concat_path_def using assms and joinable
    by (auto simp: vwalk_to_path_hd vwalk_to_path_last
      vwalk_join_hd vwalk_join_last)
qed

lemma concat_path_exists:
  assumes p_props: "vwalk p G" "hd p = u" "last p = v"
  assumes q_props: "vwalk q G" "hd q = v" "last q = w"
  shows "\<exists>r. path r G \<and> hd r = u \<and> last r = w"
using assms
by (metis concat_path_is_path)


section {* Unsorted *}

lemma (in pseudo_digraph) paths_finite:
  shows "finite {p. path p G}"
proof -
  have "{p. path p G}
    \<subseteq> {xs. set xs \<subseteq> verts G \<and> length xs \<le> card (verts G)}"
  proof (clarify, rule conjI)
    fix p assume "path p G"
    then show "set p \<subseteq> verts G" by blast

    from `path p G` have "length p = card (set p)"
      by (auto simp add: distinct_card)
    also have "\<dots> \<le> card (verts G)"
      using `path p G` and assms
      by (auto intro!: card_mono elim!: pathE)
    finally show "length p \<le> card (verts G)" .
  qed
  moreover
  have "finite {xs. set xs \<subseteq> verts G \<and> length xs \<le> card (verts G)}"
    by (fastforce intro: finite_lists_length_le)
  ultimately show ?thesis by (rule finite_subset)
qed

lemma reachableE[elim]:
  assumes "u \<rightarrow>*G v"
  assumes "\<exists>p. path p G \<and> hd p = u \<and> last p = v \<Longrightarrow> P"
  shows "P"
by (metis assms reachable_def vwalk_to_path_hd vwalk_to_path_last vwalk_to_path_path)

lemma vwalk_edges_verts:
  assumes "(u,v) \<in> set (vwalk_edges p)"
  shows "u \<in> set p" "v \<in> set p"
using assms
by (induct p rule: vwalk_edges.induct) auto

lemma vwalk_rev_ex:
  assumes "symmetric G"
  assumes "vwalk p G"
  shows "vwalk (rev p) G"
using assms
proof (induct p)
  case Nil then show ?case by simp
next
  case (Cons x xs)
  then show ?case proof (cases "xs = []")
    case True then show ?thesis using Cons by auto
  next
    case False
    then have "vwalk xs G" using `vwalk (x # xs) G`
      by (metis vwalk_consE)
    then have "vwalk (rev xs) G" using Cons by blast
    have "vwalk (rev (x # xs)) G"
    proof (rule vwalkI)
      have "set (x # xs) \<subseteq> verts G" using `vwalk (x # xs) G` by blast
      then show "set (rev (x # xs)) \<subseteq> verts G" by auto
    next
      have "set (vwalk_edges (x # xs)) \<subseteq> edges_ends G"
        using `vwalk (x # xs) G` by auto
      then show "set (vwalk_edges (rev (x # xs))) \<subseteq> edges_ends G"
        using `symmetric G`
        by (simp only: set_vwalk_edges_rev)
           (auto intro: edges_ends_symmetric)
    next
      show "rev (x # xs) \<noteq> []" by auto
    qed
    then show "vwalk (rev (x # xs)) G" by auto
  qed
qed

lemma reachable_sym:
  assumes "v \<rightarrow>* G w" "symmetric G" shows "w \<rightarrow>* G v"
  by (metis reachable_def assms hd_rev assms
    last_rev rev_rev_ident vwalkE vwalk_rev_ex)


end
