theory GraphGA
imports GraphSpec
begin

  definition gga_from_list :: 
    "('V,'W,'G) graph_empty \<Rightarrow> ('V,'W,'G) graph_add_node 
      \<Rightarrow> ('V,'W,'G) graph_add_edge 
    \<Rightarrow> ('V,'W,'G) graph_from_list"
    where 
    "gga_from_list e a u l \<equiv> 
      let (nl,el) = l;
        g1 = foldl (\<lambda>g v. a v g) e nl
      in foldl (\<lambda>g (v,e,v'). u v e v' g) g1 el"
  
  lemma gga_from_list_correct:
    fixes \<alpha> :: "'G \<Rightarrow> ('V,'W) graph"
    assumes "graph_empty \<alpha> invar e"
    assumes "graph_add_node \<alpha> invar a"
    assumes "graph_add_edge \<alpha> invar u"
    shows "graph_from_list \<alpha> invar (gga_from_list e a u)"
  proof -
    interpret 
      graph_empty \<alpha> invar e + 
      graph_add_node \<alpha> invar a + 
      graph_add_edge \<alpha> invar u
      by fact+

    {
      fix nl el
      def g1\<equiv>"(foldl (\<lambda>g v. a v g) e nl)"
      def g2\<equiv>"(foldl (\<lambda>g (v,e,v'). u v e v' g) g1 el)"
      have "invar g1 \<and> \<alpha> g1 = \<lparr> nodes = set nl, edges = {} \<rparr>"
        unfolding g1_def
        by (induct nl rule: rev_induct)
           (auto simp: empty_correct add_node_correct empty_def add_node_def)
      hence "invar g2 
        \<and> \<alpha> g2 = \<lparr> nodes = set nl \<union> fst`set el \<union> snd`snd`set el,
                    edges = set el \<rparr>"
        unfolding g2_def
        by (induct el rule: rev_induct) (auto simp: add_edge_correct add_edge_def)
      hence "invar g2 \<and> adjl_\<alpha> (nl,el) = \<alpha> g2"
        unfolding adjl_\<alpha>_def by auto
    }
    thus ?thesis
      unfolding gga_from_list_def_raw
      apply unfold_locales
      apply auto
      done
  qed
      
  definition set_iteratori_combine :: 
      "('a,'\<sigma>) iteratori_tmpl \<Rightarrow> ('a \<Rightarrow> ('b,'\<sigma>) iteratori_tmpl) \<Rightarrow> 
       ('a \<times> 'b ,'\<sigma>) iteratori_tmpl" where
    "set_iteratori_combine it_a it_b \<equiv> \<lambda>c f \<sigma>0.
      it_a c (
        \<lambda>a \<sigma>. (it_b a) c (\<lambda>b \<sigma>. f (a,b) \<sigma>) \<sigma>
      ) \<sigma>0"

  lemma set_iteratori_combine :
    fixes it_a :: "('a,'\<sigma>) iteratori_tmpl"
    fixes it_b :: "'a \<Rightarrow> ('b,'\<sigma>) iteratori_tmpl" 
    assumes it_a: "set_iteratori it_a S_a"
    assumes it_b: "\<And>a b. a \<in> S_a \<Longrightarrow> set_iteratori (it_b a) (S_b a)"
    shows "set_iteratori (set_iteratori_combine it_a it_b) {(a,b) | a b. a \<in> S_a \<and> b \<in> S_b a}"
  unfolding set_iteratori_def
  proof (clarify)
    def S \<equiv> "{(a,b) | a b. a \<in> S_a \<and> b \<in> S_b a}"

    fix I c f and \<sigma>0 :: '\<sigma>
    assume I_S_\<sigma>0 : "I {(a,b) | a b. a \<in> S_a \<and> b \<in> S_b a} \<sigma>0"
    assume "\<forall>S' \<sigma> x. c \<sigma> \<longrightarrow> x \<in> S' \<longrightarrow> I S' \<sigma> \<longrightarrow> 
                S' \<subseteq> {(a,b) | a b. a \<in> S_a \<and> b \<in> S_b a} \<longrightarrow> I (S' - {x}) (f x \<sigma>)"
    hence step: "\<And>S' \<sigma> a b. \<lbrakk>c \<sigma>; (a, b) \<in> S'; S' \<subseteq> S; I S' \<sigma>\<rbrakk> \<Longrightarrow> I (S' - {(a,b)}) (f (a,b) \<sigma>)"
      unfolding S_def by auto

    def step_c_prop \<equiv> "\<lambda>\<sigma>. \<forall>S'. (S' \<subseteq> S \<and> S' \<noteq> {} \<and> I S' \<sigma>) \<longrightarrow> c \<sigma>"
    assume "\<not> (\<exists>S'\<subseteq>{(a,b) | a b. a \<in> S_a \<and> b \<in> S_b a}. S' \<noteq> {} \<and>
             \<not> c (set_iteratori_combine it_a it_b c f \<sigma>0) \<and>
             I S' (set_iteratori_combine it_a it_b c f \<sigma>0))"
    hence step_c_final: "step_c_prop (set_iteratori_combine it_a it_b c f \<sigma>0)" 
      unfolding step_c_prop_def S_def by auto

    have "step_c_prop (set_iteratori_combine it_a it_b c f \<sigma>0) \<longrightarrow> 
          I {} (set_iteratori_combine it_a it_b c f \<sigma>0)"
    unfolding set_iteratori_combine_def
    proof (induct rule: set_iteratori.iterate_rule_P [OF it_a, where I = 
             "\<lambda>S' \<sigma>. step_c_prop \<sigma> \<longrightarrow> I { (a, b) | a b. (a, b) \<in> S \<and> a \<in> S'} \<sigma>"])
      case 1 
      have "{(a, b) |a b. (a, b) \<in> S \<and> a \<in> S_a} = S" unfolding S_def by auto
      with 1 I_S_\<sigma>0[folded S_def] show ?case by simp
    next
      case 3 thus ?case by simp
    next
      case (4 \<sigma> S_a') 
      def S' \<equiv> "{(a, b) |a b. (a, b) \<in> S \<and> a \<in> S_a'}"
      note pre = 4 [folded S'_def]

      show ?case
      proof (cases "S' = {}")
        case True with pre(4) show ?thesis by simp
      next
        case False note S'_not_emp = this

        from pre(1) have S''_subset: "S' \<subseteq> S" unfolding S'_def by auto
        with pre S'_not_emp have "\<not> step_c_prop \<sigma>" unfolding step_c_prop_def by blast
        thus ?thesis by simp
      qed
    next
      case (2 S_a' \<sigma> a) 
      def S' \<equiv> "{(a, b) |a b. (a, b) \<in> S \<and> a \<in> S_a'}"      
      note a_in_S_a' = 2(2)
      note I_S' = 2(3)[folded S'_def]
      note S_a'_subset = 2(4)

      from it_b[of a] a_in_S_a' S_a'_subset
      have it_b': "set_iteratori (it_b a) (S_b a)" by blast

      show ?case  
      proof (induct rule: set_iteratori.iterate_rule_P [OF it_b', where I = 
               "\<lambda>S_b' \<sigma>. step_c_prop \<sigma> \<longrightarrow> I { (a', b) | a' b. (a', b) \<in> S' \<and> 
                        ((a' = a) \<longrightarrow> b \<in> S_b')} \<sigma>"])
        case 1 
        have "{(a', b) |a' b. (a', b) \<in> S' \<and> (a' = a \<longrightarrow> b \<in> (S_b a))} = S'"
          using S_a'_subset unfolding S'_def S_def by auto
        with I_S' show ?case by simp      
      next
        case 3 thus ?case unfolding S'_def by simp
      next
        case (4 \<sigma> S_b') 
        def S'' \<equiv> "{(a', b) |a' b. (a', b) \<in> S' \<and> (a' = a \<longrightarrow> b \<in> S_b')}"
        note pre = 4 [folded S''_def]

        show ?case
        proof (cases "S'' = {}")
          case True note S''_eq_emp = this
          hence "{(aa, b) |aa b. (aa, b) \<in> S \<and> aa \<in> S_a' - {a}} = {}"
            unfolding S''_def S'_def by auto
          with S''_eq_emp pre(4) show ?thesis by (simp only:)
        next
          case False note S''_not_emp = this
          have S''_subset: "S'' \<subseteq> S" unfolding S''_def S'_def by auto
          with pre S''_not_emp have "\<not> step_c_prop \<sigma>" unfolding step_c_prop_def by blast
          thus ?thesis by simp
        qed
      next
        case (2 S_b' \<sigma> b)   
        def S'' \<equiv> "{(a', b) |a' b. (a', b) \<in> S' \<and> (a' = a \<longrightarrow> b \<in> S_b')}"

        note cond_\<sigma> = 2(1)
        note b_in_S_b' = 2(2)
        from 2(1) 2(3) have I_S'' : "I S'' \<sigma>"
          unfolding S''_def step_c_prop_def by simp
        note S_b'_subset = 2(4)
 
        have S''_delete: 
           "{(a', ba) |a' ba. (a', ba) \<in> S' \<and> (a' = a \<longrightarrow> ba \<in> S_b' - {b})} = S'' - {(a, b)}"
           unfolding S''_def by auto

        have ab_in_S'': "(a,b) \<in> S''"
          using S_a'_subset S_b'_subset b_in_S_b' a_in_S_a'
          unfolding S''_def S'_def S_def by auto

        have S''_subset: "S'' \<subseteq> S"
          unfolding S''_def S'_def by auto

        from step [OF cond_\<sigma> ab_in_S'' S''_subset I_S'']
        have "I (S'' - {(a, b)}) (f (a, b) \<sigma>)" .
        with S''_delete show ?case by simp
      qed
    qed
    with step_c_final show "I {} (set_iteratori_combine it_a it_b c f \<sigma>0)" by simp
  qed

  definition gga_edges_it ::
    "('V,'W,'\<sigma>,'G) graph_nodes_it \<Rightarrow> ('V,'W,'\<sigma>,'G) graph_succ_it  \<Rightarrow> ('V,'W,'\<sigma>,'G) graph_edges_it"
    where
    "gga_edges_it ni si G \<equiv> set_iteratori_combine (ni G) (si G)"

  lemma gga_edges_it_correct:
    fixes ni::"('V,'W,'\<sigma>,'G) graph_nodes_it"
    fixes \<alpha> :: "'G \<Rightarrow> ('V,'W) graph"
    assumes ni_it: "graph_nodes_it \<alpha> invar ni"
    assumes si_it: "graph_succ_it \<alpha> invar si"
    shows "graph_edges_it \<alpha> invar (gga_edges_it ni si)"
  unfolding graph_edges_it_def graph_edges_it_axioms_def
  proof (intro conjI allI impI)
    from ni_it show "graph \<alpha> invar"
      unfolding graph_nodes_it_def by fast
  next
    fix g
    assume invar: "invar g"
    show "set_iteratori (gga_edges_it ni si g) (edges (\<alpha> g))" 
      unfolding gga_edges_it_def
    proof (rule set_iteratori_combine)
      have ni_correct: "set_iteratori (ni g) (nodes (\<alpha> g))"
        using graph_nodes_it.nodes_it_correct [OF ni_it, OF invar] . 

      have "{a'. \<exists>a b. a' = a \<and> (a, b) \<in> edges (\<alpha> g)} = nodes (\<alpha> g)"
        unfolding nodes_def


    have si_correct: "\<And>v. set_iteratori (si g v) (succ (\<alpha> g) v)"
      using graph_succ_it.succ_it_correct [OF si_it, OF invar] .

    show ?thesis
    proof 
      fix g I and \<sigma>0::'\<sigma> and c f
      assume GI[simp, intro!]: "invar g"
      and I0: "I (edges (\<alpha> g)) \<sigma>0"
      and RULE: "\<And>S \<sigma> e. \<lbrakk>c \<sigma>; e\<in>S; I S \<sigma>; S\<subseteq>edges (\<alpha> g)\<rbrakk> 
                    \<Longrightarrow> I (S-{e}) (f e \<sigma>)"

      from valid[OF GI] interpret valid_graph "\<alpha> g" .

      show " I {} (gga_edges_it ni si g c f \<sigma>0) \<or>
        (\<exists>S\<subseteq>E. S \<noteq> {} \<and> \<not> c (gga_edges_it ni si g c f \<sigma>0) 
          \<and> I S (gga_edges_it ni si g c f \<sigma>0))"
        unfolding gga_edges_it_def
      proof (rule_tac 
          I="\<lambda>it \<sigma>. I (E \<inter> it\<times>UNIV\<times>UNIV) \<sigma> \<or>
            (\<not>c \<sigma> \<and> (\<exists>S\<subseteq>E. S\<noteq>{} \<and> I S \<sigma>))
          " 
          in
          set_iteratori.iterate_rule_P[OF nodes_it_correct[OF GI]])
        
        from E_valid have 
          "E \<inter> V\<times>UNIV\<times>UNIV
           = E" by auto
        with I0 show 
          "I (E \<inter> V\<times>UNIV\<times>UNIV) \<sigma>0 \<or> (\<not> c \<sigma>0 \<and> (\<exists>S\<subseteq>E. S\<noteq>{} \<and> I S \<sigma>0))"
          by simp
      next
        fix itV \<sigma> vt
        assume [simp, intro!]: "c \<sigma>" 
          and [simp, intro!]: "vt\<in>itV" 
          and IS0: "I (E \<inter> itV\<times>UNIV\<times>UNIV) \<sigma> 
            \<or> (\<not>c \<sigma> \<and> (\<exists>S\<subseteq>E. S\<noteq>{} \<and> I S \<sigma>))"
          and itV_SS: "itV \<subseteq> V"
        let ?inner = "(si g vt c (\<lambda>(w, v'). f (vt, w, v')) \<sigma>)"
        show "I (E \<inter> (itV-{vt})\<times>UNIV\<times>UNIV) ?inner \<or>
          (\<not> c ?inner \<and> (\<exists>S\<subseteq>E. S\<noteq>{} \<and> I S ?inner))
          "
        proof (rule_tac I="\<lambda>itS \<sigma>. I 
            ((E \<inter> (itV-{vt})\<times>UNIV\<times>UNIV) \<union> {vt}\<times>itS)
            \<sigma>" in 
          set_iteratori.iterate_rule_P[OF succ_it_correct[OF GI]])
          have "((E \<inter> (itV-{vt})\<times>UNIV\<times>UNIV) \<union> {vt}\<times>succ (\<alpha> g) vt)
            = (E \<inter> itV\<times>UNIV\<times>UNIV)"
            unfolding succ_def by auto
          with IS0 show 
            "I ((E \<inter> (itV-{vt})\<times>UNIV\<times>UNIV) \<union> {vt}\<times>succ (\<alpha> g) vt) \<sigma>" 
            by simp
        next
          fix itS \<sigma> wv
          assume [simp, intro!]: "c \<sigma>" 
            and [simp, intro!]: "wv\<in>itS" 
            and "I (E \<inter> (itV - {vt}) \<times> UNIV \<times> UNIV \<union> {vt} \<times> itS) \<sigma>"
          (is "I ?S _")
            and itS_SS: "itS \<subseteq> succ (\<alpha> g) vt"

          have "(vt,wv) \<in> ?S"
            using itS_SS unfolding succ_def by auto
          moreover have "?S\<subseteq>E"
            using itS_SS unfolding succ_def by auto
          moreover note `I ?S \<sigma>`
          ultimately have "I (?S-{(vt,wv)}) (f (vt,wv) \<sigma>)"
            by (auto intro!: RULE)
          also have "(?S-{(vt,wv)}) 
            = (E \<inter> (itV-{vt}) \<times> UNIV \<times> UNIV \<union> {vt} \<times> (itS-{wv}))"
            by auto
          also have "(f (vt,wv) \<sigma>) = (\<lambda>(w, v'). f (vt, w, v')) wv \<sigma>"
            by auto
          finally show 
            "I (E \<inter> (itV-{vt}) \<times> UNIV \<times> UNIV \<union> {vt} \<times> (itS-{wv})) 
               ((\<lambda>(w, v'). f (vt, w, v')) wv \<sigma>)" .
        next
          fix \<sigma>
          assume "I (E \<inter> (itV - {vt}) \<times> UNIV \<times> UNIV \<union> {vt} \<times> {}) \<sigma>"
          thus "I (E \<inter> (itV - {vt}) \<times> UNIV \<times> UNIV) \<sigma> \<or> 
            (\<not>c \<sigma> \<and> (\<exists>S\<subseteq>E. S\<noteq>{} \<and> I S \<sigma>))"
            by auto
        next
          fix \<sigma> S
          assume "S\<subseteq>succ (\<alpha> g) vt"
          and "S\<noteq>{}"
          and "\<not> c \<sigma>"
          and "I (E \<inter> (itV - {vt}) \<times> UNIV \<times> UNIV \<union> {vt} \<times> S) \<sigma>"
          thus "I (E \<inter> (itV - {vt}) \<times> UNIV \<times> UNIV) \<sigma> \<or> 
            (\<not>c \<sigma> \<and> (\<exists>S\<subseteq>E. S\<noteq>{} \<and> I S \<sigma>))"
            by (auto intro!: exI[where 
              x="(E \<inter> (itV - {vt}) \<times> UNIV \<times> UNIV \<union> {vt} \<times> S)" ]
              simp: succ_def)
        qed
      next
        fix \<sigma>
        assume "I (edges (\<alpha> g) \<inter> {} \<times> UNIV \<times> UNIV) \<sigma> 
          \<or> \<not> c \<sigma> \<and> (\<exists>S\<subseteq>edges (\<alpha> g). S \<noteq> {} \<and> I S \<sigma>)"
        thus "I {} \<sigma> \<or> (\<exists>S. S \<subseteq> edges (\<alpha> g) \<and> S \<noteq> {} \<and> \<not> c \<sigma> \<and> I S \<sigma>)"
          by auto
      next
        fix \<sigma> S
        assume "S \<subseteq> nodes (\<alpha> g)" 
          and [simp]: "S \<noteq> {}" 
          and [simp]: "\<not> c \<sigma>"
          and C: "I (edges (\<alpha> g) \<inter> S \<times> UNIV \<times> UNIV) \<sigma> 
               \<or> \<not> c \<sigma> \<and> (\<exists>S\<subseteq>edges (\<alpha> g). S \<noteq> {} \<and> I S \<sigma>)"
        from C show "I {} \<sigma> \<or> (\<exists>S. S \<subseteq> edges (\<alpha> g) \<and> S \<noteq> {} \<and> \<not> c \<sigma> \<and> I S \<sigma>)"
        proof 
          txt {* The case that the inner iterator was interrupted is trivial.
            Thus, we regard the case that the outer iterator was interrupted.
            In this case, the remaining nodes may or may not have outgoing edges.
            In the former case, we get a proper interrupt result. In the latter
            case, we are already finished.
            *}
          assume "I (edges (\<alpha> g) \<inter> S \<times> UNIV \<times> UNIV) \<sigma>"
          thus "I {} \<sigma> \<or> (\<exists>S\<subseteq>edges (\<alpha> g). S \<noteq> {} \<and> \<not> c \<sigma> \<and> I S \<sigma>)"
            apply (cases "(edges (\<alpha> g) \<inter> S \<times> UNIV \<times> UNIV) = {}")
            apply simp
            apply (rule disjI2)
            apply (rule_tac x="(edges (\<alpha> g) \<inter> S \<times> UNIV \<times> UNIV)" in exI)
            apply auto
            done
        qed simp
      qed
    qed
  qed

  definition gga_to_list :: 
    "('V,'W,_,'G) graph_nodes_it \<Rightarrow> ('V,'W,_,'G) graph_edges_it \<Rightarrow> 
     ('V,'W,'G) graph_to_list"
    where 
    "gga_to_list ni ei g \<equiv> 
      (ni g (\<lambda>_. True) (op #) [], ei g (\<lambda>_. True) (op #) [])
    "

  lemma gga_to_list_correct:
    fixes \<alpha> :: "'G \<Rightarrow> ('V,'W) graph"
    assumes "graph_nodes_it \<alpha> invar ni"
    assumes "graph_edges_it \<alpha> invar ei"
    shows "graph_to_list \<alpha> invar (gga_to_list ni ei)"
  proof -
    interpret graph_nodes_it \<alpha> invar ni + 
      graph_edges_it \<alpha> invar ei
      by fact+

    show ?thesis
    proof 
      fix g
      assume [simp, intro!]: "invar g"
      then interpret valid_graph "\<alpha> g" by (rule valid)

      have "set (ni g (\<lambda>_. True) (op #) []) = V"
        apply (rule_tac I="\<lambda>it \<sigma>. set \<sigma> = V - it" 
          in set_iteratori.iterate_rule_P[OF nodes_it_correct])
        by auto
      moreover have "set (ei g (\<lambda>_. True) (op #) []) = E"
        apply (rule_tac I="\<lambda>it \<sigma>. set \<sigma> = E - it" 
          in set_iteratori.iterate_rule_P[OF edges_it_correct])
        by auto
      ultimately show "adjl_\<alpha> (gga_to_list ni ei g) = \<alpha> g"
        unfolding adjl_\<alpha>_def gga_to_list_def
        apply simp
        apply (rule graph.equality)
        apply (auto intro: E_validD)
        done
    qed
  qed

end
