theory Preorder_Equiv_Classes
imports "../../Collections/Refine_Dflt_ICF"
begin

definition "rel_\<alpha> R \<equiv> {(x,y). \<exists>Rx. R x = Some Rx \<and> y \<in> Rx}"

definition "preord_eqclasses_map_invar S R it m \<equiv> 
  S - it \<subseteq> dom m \<and> dom m \<subseteq> S \<and> ran m \<subseteq> S - it \<and>
  (\<forall>s\<in>dom m. \<forall>t\<in>S. m s = m t \<longleftrightarrow> ((s,t) \<in> R \<and> (t,s) \<in> R))"

lemma preord_eqclasses_map_invarI[intro]:
  assumes "S - it \<subseteq> dom m" "dom m \<subseteq> S" "ran m \<subseteq> S - it"
  assumes "\<And>s t. s\<in>dom m \<Longrightarrow> t\<in>S \<Longrightarrow> m s = m t \<longleftrightarrow> 
               ((s,t) \<in> R \<and> (t,s) \<in> R)"
  shows "preord_eqclasses_map_invar S R it m"
  using assms unfolding preord_eqclasses_map_invar_def by simp

lemma preord_eqclasses_map_invarD[dest]:
  assumes "preord_eqclasses_map_invar S R it m"
  shows "S - it \<subseteq> dom m" and "dom m \<subseteq> S" "ran m \<subseteq> S - it"
    and "\<And>s t. s\<in>dom m \<Longrightarrow> t\<in>S \<Longrightarrow> m s = m t \<longleftrightarrow> 
             ((s,t) \<in> R \<and> (t,s) \<in> R)"
  using assms unfolding preord_eqclasses_map_invar_def by simp_all



definition preord_eqclasses_map where
"preord_eqclasses_map S R \<equiv> do {
  ASSUME (finite S);
  ASSUME (preorder_on S R);
  FOREACH\<^bsup>preord_eqclasses_map_invar S R\<^esup> S (\<lambda>s m.
    case m s of
      Some _ \<Rightarrow> RETURN m |
      None \<Rightarrow> RETURN (\<lambda>x. if (s,x)\<in>R \<and> (x,s)\<in>R then Some s else m x)
  ) Map.empty
}"

definition "is_preord_eqclasses_map S R m \<equiv> dom m = S \<and> 
    (\<forall>s\<in>S. \<forall>t\<in>S. m s = m t \<longleftrightarrow> ((s,t)\<in>R \<and> (t,s)\<in>R))"

lemma is_preord_eqclasses_mapI[intro]:
  assumes "dom m = S"
  assumes "\<And>s t. s\<in>S \<Longrightarrow> t\<in>S \<Longrightarrow> m s = m t \<longleftrightarrow> 
               (s,t)\<in>R \<and> (t,s)\<in>R"
  shows "is_preord_eqclasses_map S R m"
  using assms unfolding is_preord_eqclasses_map_def by simp

lemma is_preord_is_eqclasses_mapD[dest]:
  assumes "is_preord_eqclasses_map S R m"
  shows "dom m = S"
    and "\<And>s t. s\<in>S \<Longrightarrow> t\<in>S \<Longrightarrow> m s = m t \<longleftrightarrow> 
             (s,t)\<in>R \<and> (t,s)\<in>R"
  using assms unfolding is_preord_eqclasses_map_def by simp_all


lemma preord_eqclasses_map_correct:
  "preord_eqclasses_map S R \<le> SPEC (is_preord_eqclasses_map S R)"
  unfolding preord_eqclasses_map_def
proof (intro refine_vcg FOREACH_rule)
  assume "finite S" thus "finite S" .
next
  show "preord_eqclasses_map_invar S R S Map.empty"
    by (intro preord_eqclasses_map_invarI, simp_all)
next
  case (goal3 s it m)
    hence "s \<in> S" by blast
    note inv = preord_eqclasses_map_invarD[OF goal3(5)]
    from inv(3) and `s \<in> it` have [dest!]: 
        "\<And>x. m x = Some s \<Longrightarrow> False" by (blast intro: ranI)
    hence [dest!]: "\<And>x. Some s = m x \<Longrightarrow> False" by force

    from `preorder_on S R` have R_in_S: "R \<subseteq> S\<times>S"
        unfolding preorder_on_def refl_on_def by simp
    from `preorder_on S R` have
         refl: "\<And>x. x\<in>S \<Longrightarrow> (x,x) \<in> R" and
         trans: "\<And>x y z. (x,y)\<in>R \<Longrightarrow> (y,z)\<in>R \<Longrightarrow> (x,z)\<in>R"
        unfolding preorder_on_def by (blast dest: refl_onD transD)+

    let ?m' = "\<lambda>x. if (s, x) \<in> R \<and> (x, s) \<in> R then Some s else m x"
    have new_dom: "dom ?m' = dom m \<union> {x. (s,x) \<in> R \<and> (x,s) \<in> R}" by auto
    show ?case
    proof (intro preord_eqclasses_map_invarI)
      have "s \<in> {x. (s,x) \<in> R \<and> (x,s) \<in> R}"
          using `s \<in> S` refl by blast
      thus "S - (it - {s}) \<subseteq> dom ?m'"
          using inv(1) by (subst new_dom, blast)
    next
      show "dom ?m' \<subseteq> S" using inv(2) R_in_S by (subst new_dom, blast)
    next
      show "ran ?m' \<subseteq> S - (it - {s})" using inv(3) `s \<in> S`
          by (auto simp: ran_def)
    next
      fix s' t assume s'_in_dom: "s' \<in> dom ?m'" and t: "t \<in> S"
      hence s': "s' \<in> S" using R_in_S and inv(2) by (auto simp: new_dom)

      show "?m' s' = ?m' t \<longleftrightarrow> (s',t) \<in> R \<and> (t,s') \<in> R"
      proof (cases "(s,s') \<in> R \<and> (s',s) \<in> R")
        case True
          thus ?thesis by (force intro: trans)
      next
        case False
          hence "s' \<in> dom m" using s'_in_dom
              by (simp only: new_dom, blast)
          from inv(4)[OF this t]
               show ?thesis by (force intro: trans)
      qed
  qed
next
  case (goal4 s it m)
     thus ?case by (intro preord_eqclasses_map_invarI, auto)
next
  case (goal5 m)
    thus ?case by blast
qed



definition "preord_eqclasses_map_impl1_loop_invar S R s m it m' \<equiv>
    (\<forall>x. m' x = (if (s,x) \<in> R \<and> (x,s) \<in> R \<and> x \<notin> it 
                 then Some s else m x))"

definition "preord_eqclasses_map_impl1_loop S R s m \<equiv> 
    FOREACH\<^bsup>preord_eqclasses_map_impl1_loop_invar S R s m\<^esup> 
        {x. (s,x) \<in> R} (\<lambda>t m. if (t,s) \<in> R
            then RETURN (m(t \<mapsto> s))
            else RETURN m) m"

definition preord_eqclasses_map_impl1 where
"preord_eqclasses_map_impl1 S R \<equiv> do {
  ASSUME (finite S);
  ASSUME (preorder_on S R);
  FOREACH S (\<lambda>s m.
    case m s of
      Some _ \<Rightarrow> RETURN m |
      None \<Rightarrow> preord_eqclasses_map_impl1_loop S R s m
  ) Map.empty
}"

lemma preord_eqclasses_map_impl1_loop_correct:
  assumes fin: "finite S" and preord: "preorder_on S R"
      and inv: "preord_eqclasses_map_invar S R it m"
  shows "preord_eqclasses_map_impl1_loop S R s m \<le> 
      SPEC (\<lambda>m'. m' = (\<lambda>x. if (s, x) \<in> R \<and> (x, s) \<in> R 
                           then Some s else m x))"
unfolding preord_eqclasses_map_impl1_loop_def
proof (intro refine_vcg FOREACH_rule)
  from preord have "R \<subseteq> S \<times> S"
      by (simp add: preorder_on_def refl_on_def)
  hence "{x. (s, x) \<in> R} \<subseteq> S" by blast
  thus "finite {x. (s, x) \<in> R}" using fin finite_subset by blast
next
  show "preord_eqclasses_map_impl1_loop_invar S R s m {x. (s, x) \<in> R} m"
      unfolding preord_eqclasses_map_impl1_loop_invar_def by force
qed (unfold preord_eqclasses_map_impl1_loop_invar_def, auto)

lemma preord_eqclasses_map_impl1_loop_correct':
  assumes fin: "finite S" and preord: "preorder_on S R"
      and inv: "preord_eqclasses_map_invar S R it' m'"
      and "s' = id s" "(m,m') \<in> Id"
  shows "preord_eqclasses_map_impl1_loop S R s m \<le> 
      SPEC (\<lambda>m''. (m'',(\<lambda>x. if (s', x) \<in> R \<and> (x, s') \<in> R 
                            then Some s' else m' x)) \<in> Id)"
proof-
  from assms have A: "s' = s" "m' = m" by simp_all
  with preord_eqclasses_map_impl1_loop_correct[OF fin preord inv]
    show ?thesis by (simp only: A, simp)
qed
  

lemma preord_eqclasses_map_impl1_refine:
  shows "preord_eqclasses_map_impl1 S R \<le> \<Down>Id (preord_eqclasses_map S R)"
unfolding preord_eqclasses_map_impl1_def preord_eqclasses_map_def
by (refine_rcg inj_on_id, simp, simp, simp,
        erule (4) preord_eqclasses_map_impl1_loop_correct')


definition "preord_eqclasses_map_impl2_loop S R s m \<equiv>
  case R s of
    None \<Rightarrow> RETURN m |
    Some Rs \<Rightarrow> FOREACH Rs (\<lambda>t m. RETURN (
        let ts = case R t of None \<Rightarrow> False | 
                             Some Rt \<Rightarrow> s \<in> Rt
        in if ts then m(t \<mapsto> s) else m)
      ) m"

definition preord_eqclasses_map_impl2 where
"preord_eqclasses_map_impl2 S R \<equiv> FOREACH S (\<lambda>s m.
    case m s of
      Some _ \<Rightarrow> RETURN m |
      None \<Rightarrow> preord_eqclasses_map_impl2_loop S R s m
      ) Map.empty"

lemma preord_eqclasses_map_impl2_loop_refine:
  fixes s::'a and R'::"('a\<times>'a) set"
  assumes "s \<in> S" "preorder_on S R'" "R' = rel_\<alpha> R"
  shows "preord_eqclasses_map_impl2_loop S R s m \<le> 
             \<Down>Id (preord_eqclasses_map_impl1_loop S R' s m)"
proof-
  let ?cond = "\<lambda>t. case R t of None \<Rightarrow> False |
                               Some x \<Rightarrow> s \<in> x"
  have cond_simp: 
      "\<And>t it. t \<in> it \<Longrightarrow> it \<subseteq> {x. (s,x) \<in> R'} \<Longrightarrow> 
           ?cond t = ((t,s) \<in> R')"
      using assms unfolding rel_\<alpha>_def 
      by (force split: option.split_asm)
  from assms have "(s,s) \<in> R'"
      unfolding preorder_on_def refl_on_def by simp
  hence "R s = Some {x. (s,x) \<in> R'}"
      using assms unfolding rel_\<alpha>_def by force
  hence "preord_eqclasses_map_impl2_loop S R s m = 
             FOREACH {x. (s,x) \<in> R'} (\<lambda>t m. 
                 RETURN (if ?cond t then m(t \<mapsto> s) else m)) m"
      unfolding preord_eqclasses_map_impl2_loop_def by simp
  also have "... \<le> \<Down>Id (preord_eqclasses_map_impl1_loop S R' s m)"
      unfolding preord_eqclasses_map_impl1_loop_def
      by (refine_rcg inj_on_id, simp_all add: cond_simp)
  finally show ?thesis .
qed

lemma preord_eqclasses_map_impl2_loop_refine':
  fixes s::'a and R'::"('a\<times>'a) set"
  assumes "preorder_on S R'" "R' = rel_\<alpha> R"
      and "s \<in> it" "it \<subseteq> S" "s' = id s" "(m,m') \<in> Id"
  shows "preord_eqclasses_map_impl2_loop S R s m \<le> 
             \<Down>Id (preord_eqclasses_map_impl1_loop S R' s' m')"
proof-
  from assms(3,4) have "s \<in> S" by blast
  from preord_eqclasses_map_impl2_loop_refine[OF this assms(1,2)] assms
      show ?thesis by simp
qed

lemma preord_eqclasses_map_impl2_refine:
  assumes fin: "finite S" and preord: "preorder_on S R'" 
      and R: "R' = rel_\<alpha> R"
  shows "preord_eqclasses_map_impl2 S R \<le> 
             \<Down>Id (preord_eqclasses_map_impl1 S R')"
unfolding preord_eqclasses_map_impl2_def 
          preord_eqclasses_map_impl1_def
using assms by (refine_rcg inj_on_id, simp, simp, simp)
   (erule (3) preord_eqclasses_map_impl2_loop_refine'[OF preord R])


abbreviation "preord_eqclasses_map_impl \<equiv> preord_eqclasses_map_impl2"
lemmas preord_eqclasses_map_impl_def = preord_eqclasses_map_impl2_def

lemma preord_eqclasses_map_impl_correct:
  assumes "finite S" and "preorder_on S R'"
  assumes "(R,R') \<in> br rel_\<alpha> (\<lambda>_. True)"
  shows "preord_eqclasses_map_impl S R \<le> SPEC (is_preord_eqclasses_map S R')"
proof-
  from assms(3) have "R' = rel_\<alpha> R" unfolding br_def by simp
  from preord_eqclasses_map_impl2_refine[OF assms(1,2) this]
      have "preord_eqclasses_map_impl2 S R \<le> 
                (preord_eqclasses_map_impl1 S R')" by simp
  also from preord_eqclasses_map_impl1_refine
      have "preord_eqclasses_map_impl1 S R' \<le> 
                (preord_eqclasses_map S R')" by simp
  also from preord_eqclasses_map_correct
      have "preord_eqclasses_map S R' \<le> 
                SPEC (is_preord_eqclasses_map S R')" .
  finally show ?thesis .
qed


schematic_lemma preord_eqclasses_map_code_refine:
  assumes [autoref_rules]: "(S,S') \<in> \<langle>nat_rel\<rangle>dflt_rs_rel"
  assumes [autoref_rules]: "(R,R') \<in> \<langle>nat_rel,\<langle>nat_rel\<rangle>dflt_rs_rel\<rangle>dflt_rm_rel"
  shows "(?f::?'c, preord_eqclasses_map_impl S' R') \<in> ?R"
  unfolding preord_eqclasses_map_impl_def 
            preord_eqclasses_map_impl2_loop_def
using assms by autoref_monadic

concrete_definition preord_eqclasses_map_code uses preord_eqclasses_map_code_refine

end
