(*  Title:      HOL/Library/Prefix_Order.thy
    Author:     Tobias Nipkow and Markus Wenzel, TU Muenchen
*)

header {* Prefix order on lists as order class instance *}

theory Prefix_Order
imports "~~/src/HOL/Library/Sublist"
begin

instantiation list :: (type) order
begin

definition "(xs::'a list) \<le> ys \<equiv> prefixeq xs ys"
definition "(xs::'a list) < ys \<equiv> prefix xs ys"

instance by (default, unfold prefix_def less_eq_list_def less_list_def) auto

end

lemmas prefixeqI [intro?] = prefixeqI [folded less_eq_list_def]
lemmas prefixeqE [elim?] = prefixeqE [folded less_eq_list_def]
lemmas prefixI' [intro?] = prefixI' [folded less_list_def]
lemmas prefixE' [elim?] = prefixE' [folded less_list_def]
lemmas prefixI [intro?] = prefixI [folded less_eq_list_def less_list_def]
lemmas prefixE [elim?] = prefixE [folded less_eq_list_def less_list_def]
theorems Nil_prefixeq [iff] = Nil_prefixeq [folded less_eq_list_def]
theorems prefixeq_Nil [simp] = prefixeq_Nil [folded less_eq_list_def]
lemmas prefixeq_snoc [simp] = prefixeq_snoc [folded less_eq_list_def]
lemmas Cons_prefixeq_Cons [simp] = Cons_prefixeq_Cons [folded less_eq_list_def]
lemmas same_prefixeq_prefixeq [simp] = same_prefixeq_prefixeq [folded less_eq_list_def]
lemmas same_prefixeq_nil [iff] = same_prefixeq_nil [folded less_eq_list_def]
lemmas prefixeq_prefixeq [simp] = prefixeq_prefixeq [folded less_eq_list_def]
theorems prefixeq_Cons = prefixeq_Cons [folded less_eq_list_def]
theorems prefixeq_length_le[dest] = prefixeq_length_le [folded less_eq_list_def]
lemmas prefix_simps [simp, code] = prefix_simps [folded less_list_def]
lemmas not_prefixeq_induct [consumes 1, case_names Nil Neq Eq] =
  not_prefixeq_induct [folded less_eq_list_def]
lemmas take_is_prefixeq [intro] = take_is_prefixeq [folded less_eq_list_def]
lemmas map_prefixeqI [intro] = map_prefixeqI [folded less_eq_list_def]
lemmas set_mono_prefixeq [dest] = set_mono_prefixeq [folded less_eq_list_def]

end
