header {* Simple DFS Algorithm *}
theory Sepref_DFS
imports 
  "CAVA_Base" 
  "../../../DFS_Framework/Misc/DFS_Framework_Refine_Aux"
  "../Sepref"
  Sepref_Graph
begin

(*<*)

(* TODO: Move! *)
(* TODO: We do not really need the lower-bound for D in most cases.
  So let the default rules be without lower bound *)
lemma REC_rule_arb':
  fixes x::"'x" and arb::'arb
  assumes M: "trimono body"
  assumes I0: "pre arb x"
  assumes IS: "\<And>f arb x. \<lbrakk>
    \<And>arb' x. pre arb' x \<Longrightarrow> f x \<le> M arb' x; pre arb x
  \<rbrakk> \<Longrightarrow> body f x \<le> M arb x"
  shows "REC body x \<le> M arb x"
  apply (rule REC_rule_arb[where M=M and pre=pre])
  apply fact
  apply fact
  apply (rule IS)
  apply assumption+
  done

lemma RECT_rule_arb':
  fixes x::"'x" and arb::'arb
  assumes "trimono body"
    and "wf V"
    and "pre arb x"
    and
    IS: "\<And>f arb x.
        \<lbrakk>\<And>arb' x'. \<lbrakk>pre arb' x'; (x', x) \<in> V\<rbrakk> \<Longrightarrow> f x' \<le> M arb' x'; pre arb x;
         f \<le> REC\<^sub>T body\<rbrakk>
        \<Longrightarrow> body f x \<le> M arb x"
  shows "REC\<^sub>T body x \<le> M arb x"
  apply (rule RECT_rule_arb[where M=M and pre=pre])
  apply fact
  apply fact
  apply fact
  apply (rule IS)
  apply assumption+
  apply simp
  done

(*>*)

text {*
  We define a simple DFS-algorithm, prove a simple correctness
  property, and do data refinement to an efficient implementation.
*}

subsection {* Definition *}

text {* Recursive DFS-Algorithm. 
  @{text "E"} is the edge relation of the graph, @{text "vd"} the node to 
  search for, and @{text "v0"} the start node.
  Already explored nodes are stored in @{text "V"}.*}

context 
  fixes E :: "'v rel" and v0 :: 'v and tgt :: 'v
begin
  definition dfs :: "bool nres" where
    "dfs \<equiv> do {
      (_,r) \<leftarrow> RECT (\<lambda>dfs (V,v). 
        if v\<in>V then RETURN (V,False)
        else do {
          let V=insert v V;
          if v = tgt then
            RETURN (V,True)
          else
            FOREACH\<^sub>C (E``{v}) (\<lambda>(_,b). \<not>b) (\<lambda>v' (V,_). dfs (V,v')) (V,False)
        }
      ) ({},v0);
      RETURN r
    }"

  definition "reachable \<equiv> {v. (v0,v)\<in>E\<^sup>*}"
  
  lemma dfs_correct:
    assumes fr: "finite reachable"
    shows "dfs \<le> SPEC (\<lambda>r. r \<longleftrightarrow> (v0,tgt)\<in>E\<^sup>*)"
  proof -
    have F: "\<And>v. v\<in>reachable \<Longrightarrow> finite (E``{v})"
      using fr
      apply (auto simp: reachable_def)
      by (metis (mono_tags) Image_singleton Image_singleton_iff
        finite_subset rtrancl.rtrancl_into_rtrancl subsetI)

  
    def rpre \<equiv> "\<lambda>S (V,v). 
        v\<in>reachable 
      \<and> V\<subseteq>reachable
      \<and> S\<subseteq>V
      \<and> tgt\<notin>V
      \<and> E``(V-S) \<subseteq> V"

    def rpost \<equiv> "\<lambda>S (V,v) (V',r). 
          (r\<longleftrightarrow>tgt\<in>V') 
        \<and> V\<subseteq>V' 
        \<and> v\<in>V'
        \<and> V'\<subseteq>reachable
        \<and> (\<not>r \<longrightarrow> (E``(V'-S) \<subseteq> V'))
      "

    def fe_inv \<equiv> "\<lambda>S V v it (V',r).
        (r\<longleftrightarrow>tgt\<in>V')
      \<and> insert v V\<subseteq>V'
      \<and> E``{v} - it \<subseteq> V'
      \<and> V'\<subseteq>reachable
      \<and> S\<subseteq>insert v V
      \<and> (\<not>r \<longrightarrow> E``(V'-S) \<subseteq> V' \<union> it \<and> E``(V'-insert v S) \<subseteq> V')"

    have vc_pre_initial: "rpre {} ({}, v0)"
      by (auto simp: rpre_def reachable_def)

    {
      (* Case: Node already visited *)
      fix S V v
      assume "rpre S (V,v)"
        and "v\<in>V"
      hence "rpost S (V,v) (V,False)"
        unfolding rpre_def rpost_def
        by auto
    } note vc_node_visited = this

    {
      (* Case: Found node *)
      fix S V
      assume "rpre S (V,tgt)"
      hence "rpost S (V,tgt) (insert tgt V, True)"
        unfolding rpre_def rpost_def
        by auto
    } note vc_node_found = this

    { 
      fix S V v
      assume "rpre S (V, v)"
      hence "finite (E``{v})"
        unfolding rpre_def using F by (auto)
    } note vc_foreach_finite = this
  
    {
      (* fe_inv initial *)
      fix S V v
      assume A: "v \<notin> V" "v \<noteq> tgt"
        and PRE: "rpre S (V, v)"
      hence "fe_inv S V v (E``{v}) (insert v V, False)"
        unfolding fe_inv_def rpre_def by (auto)
    } note vc_enter_foreach = this

    {
      (* fe_inv ensures rpre *)
      fix S V v v' it V'
      assume A: "v \<notin> V" "v \<noteq> tgt" "v' \<in> it" "it \<subseteq> E``{v}"
        and FEI: "fe_inv S V v it (V', False)"
        and PRE: "rpre S (V, v)"

      from A have "v' \<in> E``{v}" by auto
      moreover from PRE have "v \<in> reachable" by (auto simp: rpre_def)
      hence "E``{v} \<subseteq> reachable" by (auto simp: reachable_def)
      ultimately have [simp]: "v'\<in>reachable" by blast

      have "rpre (insert v S) (V', v')"
        unfolding rpre_def
        using FEI PRE by (auto simp: fe_inv_def rpre_def) []
    } note vc_rec_pre = this

    {
      (* rpost implies fe_inv *)
      fix S V V' v v' it Vr''
      assume "fe_inv S V v it (V', False)"
        and "rpost (insert v S) (V', v') Vr''"
      hence "fe_inv S V v (it - {v'}) Vr''"
        unfolding rpre_def rpost_def fe_inv_def
        by clarsimp blast
    } note vc_iterate_foreach = this

    {
      (* fe_inv (completed) implies rpost *)
      fix S V v V'
      assume PRE: "rpre S (V, v)" 
      assume A: "v \<notin> V" "v \<noteq> tgt"
      assume FEI: "fe_inv S V v {} (V', False)"
      have "rpost S (V, v) (V', False)"
        unfolding rpost_def
        using FEI by (auto simp: fe_inv_def) []
    } note vc_foreach_completed_imp_post = this

    {
      (* fe_inv (interrupted) implies rpost *)
      fix S V v V' it
      assume PRE: "rpre S (V, v)" 
        and A: "v \<notin> V" "v \<noteq> tgt" "it \<subseteq> E``{v}"
        and FEI: "fe_inv S V v it (V', True)"
      hence "rpost S (V, v) (V', True)"
        by (auto simp add: rpre_def rpost_def fe_inv_def) []
    } note vc_foreach_interrupted_imp_post = this

    {
      fix V r
      assume "rpost {} ({}, v0) (V, r)"
      hence "r = ((v0, tgt) \<in> E\<^sup>*)"
        by (auto 
          simp: rpost_def reachable_def 
          dest: Image_closed_trancl 
          intro: rev_ImageI)
    } note vc_rpost_imp_spec = this

    show ?thesis
      unfolding dfs_def
      apply (refine_rcg refine_vcg)
      apply (rule order_trans)
      
      apply(rule RECT_rule_arb'[where 
          pre=rpre 
          and M="\<lambda>a x. SPEC (rpost a x)"
          and V="finite_psupset reachable <*lex*> {}"
          ])
      apply refine_mono
      apply (blast intro: fr)
      apply (rule vc_pre_initial)
      
      apply (refine_rcg refine_vcg 
        FOREACHc_rule'[where I="fe_inv S v s" for S v s]
        )
      apply (simp_all add: vc_node_visited vc_node_found)

      apply (simp add: vc_foreach_finite)

      apply (auto intro: vc_enter_foreach) []

      apply (rule order_trans)
      apply (rprems)
      apply (erule (5) vc_rec_pre)
        apply (auto simp add: fe_inv_def finite_psupset_def) []
        apply (refine_rcg refine_vcg)
        apply (simp add: vc_iterate_foreach)

      apply (auto simp add: vc_foreach_completed_imp_post) []

      apply (auto simp add: vc_foreach_interrupted_imp_post) []

      apply (auto simp add: vc_rpost_imp_spec) []
      done
    qed
end

subsection \<open>Refinement to Imperative/HOL\<close>

text \<open>We set up a schematic proof goal,
  and use the sepref-tool to synthesize the implementation.
\<close>

schematic_lemma dfs_impl:
  fixes s t :: nat
  notes [sepref_opt_simps del] = imp_nfoldli_def 
    -- \<open>Prevent the foreach-loop to be unfolded to a fixed-point, 
        to produce more readable code for presentation purposes.\<close>
  notes [id_rules] = 
    itypeI[of s "TYPE(nat)"]
    itypeI[of t "TYPE(nat)"]
    itypeI[of E "TYPE(nat i_slg)"]
    -- \<open>Declare parameters to operation identification\<close>
  shows "hn_refine (
    hn_ctxt (is_graph nat_rel) E Ei 
  * hn_val nat_rel s si 
  * hn_val nat_rel t ti) (?c::?'c Heap) ?\<Gamma>' ?R (dfs E s t)"
  unfolding dfs_def -- \<open>Unfold definition of DFS\<close>
  apply sepref -- \<open>Invoke sepref-tool\<close>
  done

concrete_definition dfs_impl uses dfs_impl
  -- \<open>Extract generated implementation into constant\<close>
prepare_code_thms dfs_impl_def
  -- \<open>Set up code equations for recursion combinators\<close>
export_code dfs_impl checking SML_imp
  -- \<open>Generate SML code with Imperative/HOL\<close>

text \<open>Finally, correctness is shown by combining the 
  generated refinement theorem with the abstract correctness theorem.\<close>
corollary dfs_impl_correct:
  "finite (reachable E s) \<Longrightarrow> 
  <is_graph nat_rel E Ei> 
    dfs_impl Ei s t 
  < \<lambda>r. is_graph nat_rel E Ei * \<up>(r \<longleftrightarrow> (s,t)\<in>E\<^sup>* ) >\<^sub>t"
proof -
  assume FIN: "finite (reachable E s)"
  note aux = hn_refine_ref[OF dfs_correct dfs_impl.refine, 
    unfolded hn_refine_def, simplified, OF FIN]
  show ?thesis 
    apply (rule cons_rule[OF _ _ aux])
    apply (sep_auto simp: hn_ctxt_def pure_def)+
    done
qed

end

