header \<open>Imperative Implementation of Dijkstra's Shortest Paths Algorithm\<close>
theory Sepref_Dijkstra
imports 
  "../Sepref"
  "../../Dijkstra_Shortest_Path/Dijkstra"
  "../../Dijkstra_Shortest_Path/Test"
  "~~/src/HOL/Library/Code_Target_Numeral"
  "../../../DFS_Framework/Misc/DFS_Framework_Refine_Aux"
  "Sepref_WGraph"
begin

(* Setup for Infty *)

instantiation infty :: (heap) heap
begin
  instance 
    apply default
    apply (rule_tac x="\<lambda>Infty \<Rightarrow> 0 | Num a \<Rightarrow> to_nat a + 1" in exI)
    apply (rule injI)
    apply (auto split: infty.splits)
    done
end

lemma [sepref_import_param]:
  "(Infty,Infty)\<in>\<langle>R\<rangle>infty_rel"
  "(Num,Num) \<in> R \<rightarrow> \<langle>R\<rangle>infty_rel"
  "(Weight.val,Weight.val) \<in> \<langle>Id\<rangle>infty_rel \<rightarrow> Id" (* TODO/FIXME: The restriction to Id may cause trouble later *)
  by (auto simp: infty_rel_def)

lemma [constraint_rules]: "REL_IS_ID R \<Longrightarrow> REL_IS_ID (\<langle>R\<rangle>infty_rel)"
  apply (auto simp: infty_rel_def)
  apply (rename_tac x, case_tac x, auto)+
  done

definition "infty_plusi plusi a b \<equiv>
  case (a,b) of 
    (Infty, _) \<Rightarrow> Infty
  | (_, Infty) \<Rightarrow> Infty
  | (Num a, Num b) \<Rightarrow> Num (plusi a b)"

lemma [sepref_import_param]:
  assumes "GEN_OP plusi (op +) (R \<rightarrow> R \<rightarrow> R)"
  shows "(infty_plusi plusi, op +) \<in> \<langle>R\<rangle>infty_rel \<rightarrow> \<langle>R\<rangle>infty_rel \<rightarrow> \<langle>R\<rangle>infty_rel"
  using assms
  by (fastforce simp: infty_plusi_def infty_rel_def split: infty.splits dest: fun_relD)

definition "infty_lessi lessi a b \<equiv>
  case (a,b) of 
    (Infty, _) \<Rightarrow> False
  | (Num _, Infty) \<Rightarrow> True
  | (Num a, Num b) \<Rightarrow> lessi a b"

lemma [sepref_import_param]:
  assumes "GEN_OP lessi (op <) (R \<rightarrow> R \<rightarrow> bool_rel)"
  shows "(infty_lessi lessi, op <) \<in> \<langle>R\<rangle>infty_rel \<rightarrow> \<langle>R\<rangle>infty_rel \<rightarrow> bool_rel"
  using assms
  by (fastforce simp: infty_lessi_def infty_rel_def split: infty.splits dest: fun_relD)


(* TODO/FIXME: Currently, we need some fine-tuning to handle 
  these nested product cases. Fix this!
*)
lemma aux_prod: "(\<lambda>(a,b) (c,d). f a b c d) 
  = (\<lambda>p1 p2. let (a,b) = p1; (c,d) = p2 in f a b c d)"
  by auto


lemma param_mpath': "(mpath',mpath')
  \<in> \<langle>\<langle>A\<times>\<^sub>r B \<times>\<^sub>r A\<rangle>list_rel \<times>\<^sub>r B\<rangle>option_rel \<rightarrow> \<langle>\<langle>A\<times>\<^sub>r B \<times>\<^sub>r A\<rangle>list_rel\<rangle>option_rel"
proof -
  have 1: "mpath' = map_option fst"
    apply (intro ext, rename_tac x)
    apply (case_tac x)
    apply simp
    apply (rename_tac a)
    apply (case_tac a)
    apply simp
    done
  show ?thesis  
    unfolding 1
    by parametricity
qed
lemmas (in -) [sepref_import_param] = param_mpath'

lemma [sepref_import_param]: 
  "(mpath_weight', mpath_weight') \<in> \<langle>\<langle>A\<times>\<^sub>rB\<times>\<^sub>rA\<rangle>list_rel \<times>\<^sub>r B\<rangle>option_rel \<rightarrow> \<langle>B\<rangle>infty_rel"
  by (auto elim!: option_relE simp: infty_rel_def top_infty_def)


locale Dijkstra_Impl = Dijkstra G 
  for G :: "(nat,'W::{weight,heap}) graph"
begin

  lemmas [id_rules] = 
    itypeI[of G "TYPE('W i_wgraph)"]
    itypeI[of v0 "TYPE(nat)"]
  
  lemmas [sepref_import_param] = 
    IdI[of v0]
    IdI[of "0::'W"]

  lemma w_plus_param[autoref_rules]: "(op +, op +::'W\<Rightarrow>_) \<in> Id \<rightarrow> Id \<rightarrow> Id" by simp
  lemma w_less_param[autoref_rules]: "(op <, op <::'W\<Rightarrow>_) \<in> Id \<rightarrow> Id \<rightarrow> Id" by simp
  lemmas [sepref_import_param] = w_plus_param w_less_param

  schematic_lemma dijkstra_imp:
    shows "hn_refine (hn_ctxt (is_graph Id) G Gi) (?c1::?'c1 Heap) ?\<Gamma>1' ?R1 mdijkstra"
    unfolding mdijkstra_def mdinit_def mpop_min_def mupdate_def
    unfolding aux_prod
    by sepref

end

concrete_definition dijkstra_imp uses Dijkstra_Impl.dijkstra_imp  
prepare_code_thms dijkstra_imp_def
export_code dijkstra_imp checking SML_imp

end

