header \<open>Imperative Graph Representation\<close>
theory Sepref_Graph
imports 
  "../Sepref"
  "../../../DFS_Framework/Misc/DFS_Framework_Refine_Aux"
begin
(*<*)
(* TODO: Move to Misc *)
lemma succs_empty_dom_conv: "R``{x} = {} \<longleftrightarrow> x\<notin>Domain R"
  by auto

lemma ne_dom_imp_succs_empty: "x\<notin>Domain R \<Longrightarrow> R``{x} = {}"
  by auto
(*>*)

text \<open>Nodes are identified by numbers from ${0..<n}$, graphs are
  represented as adjacency lists\<close>
type_synonym graph_impl = "(nat list) Heap.array"

definition [to_relAPP]: "is_graph R G Gi \<equiv> 
    \<exists>\<^sub>Al. Gi \<mapsto>\<^sub>a l * \<up>(
      R = nat_rel \<and>
      Domain G \<subseteq> {0..<length l} \<and>
      (\<forall>v<length l. (l!v,G``{v})\<in>\<langle>nat_rel\<rangle>list_set_rel)
    )"

definition succi :: "graph_impl \<Rightarrow> nat \<Rightarrow> nat list Heap"
where "succi G v = do {
  l \<leftarrow> Array.len G;
  if v<l then do { 
    r \<leftarrow> Array.nth G v;
    return r
  } else return []
}"

lemma succi_rule[sep_heap_rules]: "
  < is_graph R G Gi > 
    succi Gi v 
  < \<lambda>r. is_graph R G Gi * \<up>((r,G``{v})\<in>\<langle>nat_rel\<rangle>list_set_rel) >"
  unfolding is_graph_def succi_def
  apply (sep_auto)
  apply (subst ne_dom_imp_succs_empty)
  apply (auto simp: list_set_autoref_empty)
  done

term slg_succs

lemma hnr_op_succi[sepref_fr_rules]: " 
  hn_refine 
    (hn_ctxt (is_graph R) G Gi * hn_val nat_rel v vi) 
    (succi Gi vi) 
    (hn_ctxt (is_graph R) G Gi * hn_val nat_rel v vi) 
    (pure (\<langle>nat_rel\<rangle>list_set_rel)) 
    (RETURN$(slg_succs$G$v))"
  apply rule
  unfolding hn_ctxt_def pure_def 
  by (sep_auto simp: list_set_autoref_empty)

definition cr_graph 
  :: "nat \<Rightarrow> (nat \<times> nat) list \<Rightarrow> graph_impl Heap"
where
  "cr_graph numV Es \<equiv> do {
    a \<leftarrow> Array.new numV [];
    a \<leftarrow> imp_nfoldli Es (\<lambda>_. return True) (\<lambda>(u,v) a. do {
      l \<leftarrow> Array.nth a u;
      let l = v#l;
      a \<leftarrow> Array.upd u l a;
      return a
    }) a;
    return a
  }"

(* TODO: Show correctness property for cr_graph *)

export_code cr_graph checking SML_imp


end

