header \<open>Imperative Weighted Graphs\<close>
theory Sepref_WGraph
imports 
  "../Sepref"
  "../../Dijkstra_Shortest_Path/Graph"
begin
  
  type_synonym 'w graph_impl = "(('w\<times>nat) list) Heap.array"
  
  definition "is_graph R G Gi \<equiv> 
      \<exists>\<^sub>Al. Gi \<mapsto>\<^sub>a l * \<up>(
        valid_graph G \<and>
        nodes G = {0..<length l} \<and>
        (\<forall>v\<in>nodes G. (l!v, succ G v) \<in> \<langle>R \<times>\<^sub>r nat_rel\<rangle>list_set_rel)
      )"
  
  definition succi :: "'w::heap graph_impl \<Rightarrow> nat \<Rightarrow> ('w\<times>nat) list Heap"
  where "succi G v = do {
    l \<leftarrow> Array.len G;
    if v<l then do { (* TODO: Alternatively, require v to be a valid node as precondition! *)
      r \<leftarrow> Array.nth G v;
      return r
    } else return []
  }"
  
  lemma "
    < is_graph R G Gi * \<up>(v\<in>nodes G) > 
      succi Gi v 
    < \<lambda>r. is_graph R G Gi * \<up>((r,succ G v)\<in>\<langle>R \<times>\<^sub>r nat_rel\<rangle>list_set_rel) >"
    unfolding is_graph_def succi_def
    by sep_auto
  
  lemma (in valid_graph) succ_no_node_empty: "v\<notin>V \<Longrightarrow> succ G v = {}"
    unfolding succ_def using E_valid
    by auto
  
  lemma [sepref_fr_rules]: " 
    hn_refine 
      (hn_ctxt (is_graph R) G Gi * hn_val nat_rel v vi) 
      (succi Gi vi) 
      (hn_ctxt (is_graph R) G Gi * hn_val nat_rel v vi) 
      (pure (\<langle>R \<times>\<^sub>r nat_rel\<rangle>list_set_rel)) 
      (RETURN$(succ$G$v))"
    apply rule
    unfolding hn_ctxt_def pure_def is_graph_def succi_def
    by (sep_auto simp: valid_graph.succ_no_node_empty list_set_autoref_empty)
  
  definition nodes_impl :: "'w::heap graph_impl \<Rightarrow> nat list Heap"
    where "nodes_impl Gi \<equiv> do {
    l \<leftarrow> Array.len Gi;
    return [0..<l]
  }"
  
  lemma [sepref_fr_rules]: "hn_refine 
    (hn_ctxt (is_graph R) G Gi) 
    (nodes_impl Gi) 
    (hn_ctxt (is_graph R) G Gi) 
    (pure (\<langle>nat_rel\<rangle>list_set_rel)) 
    (RETURN$(nodes$G))"
    apply rule
    unfolding hn_ctxt_def pure_def is_graph_def nodes_impl_def
    by (sep_auto simp: list_set_rel_def br_def)
  
  typedecl ('w)i_wgraph
  typedecl i_weight
  
  
  lemmas [id_rules] =
    itypeI[of succ "TYPE('w i_wgraph \<Rightarrow> nat \<Rightarrow> ('w\<times>nat)set)"]
    itypeI[of nodes "TYPE('w i_wgraph \<Rightarrow> (nat)set)"]
  
  
  definition cr_graph 
    :: "nat \<Rightarrow> (nat \<times> nat \<times> 'w) list \<Rightarrow> 'w::heap graph_impl Heap"
  where
    "cr_graph numV Es \<equiv> do {
      a \<leftarrow> Array.new numV [];
      a \<leftarrow> imp_nfoldli Es (\<lambda>_. return True) (\<lambda>(u,v,w) a. do {
        l \<leftarrow> Array.nth a u;
        let l = (w,v)#l;
        a \<leftarrow> Array.upd u l a;
        return a
      }) a;
      return a
    }"
  
  export_code cr_graph checking SML_imp


end

