theory Abs_Heap
imports 
  "~~/src/HOL/Library/Multiset"
  "../../Automatic_Refinement/Lib/Refine_Util"
  "../../../Modelchecker/CAVA_Automata/CAVA_Base/CAVA_Base"
  "../../../DFS_Framework/Misc/WHILE_aux"
  "../../List-Index/List_Index"
begin
  (* TODO: Do not add to simpset. (Done in Misc) *)
  declare Misc.map_upd_eq_restrict[simp del]

  (* TODO: Move, analogous to List.length_greater_0_conv *) 
  lemma length_ge_1_conv[iff]: "Suc 0 \<le> length l \<longleftrightarrow> l\<noteq>[]"
    by (cases l) auto

  (* TODO: Move to Misc *)
  lemma (in -) min_less_self_conv[simp]: 
    "min a b < a \<longleftrightarrow> b < (a::_::linorder)" 
    "min a b < b \<longleftrightarrow> a < (b::_::linorder)" 
    by (auto simp: min_def)


  (* TODO: Move to Misc *)
  lemma ran_nth_set_encoding_conv[simp]: 
    "ran (\<lambda>i. if i<length l then Some (l!i) else None) = set l"
    apply safe
    apply (auto simp: ran_def split: split_if_asm) []
    apply (auto simp: in_set_conv_nth intro: ranI) []
    done

  lemma nth_image_indices[simp]: "op ! l ` {0..<length l} = set l"
    by (auto simp: in_set_conv_nth)

  (* TODO: Move *)  
  lemma nth_update_invalid[simp]:"\<not>i<length l \<Longrightarrow> l[j:=x]!i = l!i"  
    apply (induction l arbitrary: i j)
    apply (auto split: nat.splits)
    done

  lemma nth_list_update': "l[i:=x]!j = (if i=j \<and> i<length l then x else l!j)"  
    by auto

  lemma distinct_butlast_swap[simp]: 
    "distinct pq \<Longrightarrow> distinct (butlast (pq[i := last pq]))"
    apply (cases pq rule: rev_cases)
    apply (auto simp: list_update_append distinct_list_update split: nat.split)
    done

  (* TODO: Move *)  
  lemma last_take_nth_conv: "n \<le> length l \<Longrightarrow> n\<noteq>0 \<Longrightarrow> last (take n l) = l!(n - 1)"
    apply (induction l arbitrary: n)
    apply (auto simp: take_Cons split: nat.split)
    done



  (* TODO: Move: Additions for List_Index *)  
  lemma index_of_last_distinct[simp]: 
    "distinct l \<Longrightarrow> index l (last l) = length l - 1"  
    apply (cases l rule: rev_cases)
    apply (auto simp: index_append)
    done

  lemma index_eqlen_conv[simp]: "index l x = length l \<longleftrightarrow> x\<notin>set l"
    by (auto simp: index_size_conv)
  


  definition "swap l i j \<equiv> l[i := l!j, j:=l!i]"
  lemma swap_nth[simp]: "\<lbrakk>i < length l; j<length l; k<length l\<rbrakk> \<Longrightarrow>
    swap l i j!k = (
      if k=i then l!j
      else if k=j then l!i
      else l!k
    )"
    unfolding swap_def
    by auto

  lemma swap_set[simp]: "\<lbrakk> i < length l; j<length l \<rbrakk> \<Longrightarrow> set (swap l i j) = set l"  
    unfolding swap_def
    by auto

  lemma swap_multiset[simp]: "\<lbrakk> i < length l; j<length l \<rbrakk> \<Longrightarrow> multiset_of (swap l i j) = multiset_of l"  
    unfolding swap_def
    by (auto simp: multiset_of_swap)


  lemma swap_length[simp]: "length (swap l i j) = length l"  
    unfolding swap_def
    by auto

  lemma swap_same[simp]: "swap l i i = l"
    unfolding swap_def by auto

  lemma distinct_swap[simp]: 
    "\<lbrakk>i<length l; j<length l\<rbrakk> \<Longrightarrow> distinct (swap l i j) = distinct l"
    unfolding swap_def
    by auto




subsection \<open>Min-Heaps\<close>
text \<open>
  We define Min-Heaps, which implement multisets of prioritized values.
  The operations are: 
    empty heap, emptiness check, insert an element, 
    remove a minimum priority element.
  
\<close>

  subsection \<open>Basic Definitions\<close>

  type_synonym 'a heap = "'a list"

  locale heapstruct =
    fixes prio :: "'a \<Rightarrow> 'b::linorder"
  begin

    definition val_of :: "'a heap \<Rightarrow> nat \<Rightarrow> 'a" where "val_of l i \<equiv> l!(i - 1)"
    abbreviation prio_of :: "'a heap \<Rightarrow> nat \<Rightarrow> 'b" where "prio_of h i \<equiv> prio (val_of h i)"

    abbreviation \<alpha> :: "'a heap \<Rightarrow> 'a multiset" where "\<alpha> \<equiv> multiset_of"
  
    definition "valid h i \<equiv> i>0 \<and> i\<le>length h"
    
    lemma valid_empty[simp]: "\<not>valid [] i" by (auto simp: valid_def)
    lemma valid0[simp]: "\<not>valid h 0" by (auto simp: valid_def)
    lemma valid_glen[simp]: "i>length h \<Longrightarrow> \<not>valid h i" by (auto simp: valid_def)

    lemma valid_len[simp]: "h\<noteq>[] \<Longrightarrow> valid h (length h)" by (auto simp: valid_def)

    lemma validI: "0<i \<Longrightarrow> i\<le>length h \<Longrightarrow> valid h i"  
      by (auto simp: valid_def)

    subsubsection \<open>Navigating the tree\<close>

    definition parent :: "nat \<Rightarrow> nat" where "parent i \<equiv> i div 2"
    definition left :: "nat \<Rightarrow> nat" where "left i \<equiv> 2*i"
    definition right :: "nat \<Rightarrow> nat" where "right i \<equiv> 2*i + 1"

    abbreviation "has_parent h i \<equiv> valid h (parent i)"
    abbreviation "has_left h i \<equiv> valid h (left i)"
    abbreviation "has_right h i \<equiv> valid h (right i)"

    abbreviation "vparent h i == val_of h (parent i)"
    abbreviation "vleft h i == val_of h (left i)"
    abbreviation "vright h i == val_of h (right i)"

    abbreviation "pparent h i == prio_of h (parent i)"
    abbreviation "pleft h i == prio_of h (left i)"
    abbreviation "pright h i == prio_of h (right i)"

    lemma parent_left_id[simp]: "parent (left i) = i"
      unfolding parent_def left_def
      by auto

    lemma parent_right_id[simp]: "parent (right i) = i"
      unfolding parent_def right_def
      by auto

    lemma child_of_parentD:
      "has_parent l i \<Longrightarrow> left (parent i) = i \<or> right (parent i) = i"
      unfolding parent_def left_def right_def valid_def
      by auto

    lemma rc_imp_lc: "\<lbrakk>valid h i; has_right h i\<rbrakk> \<Longrightarrow> has_left h i"
      by (auto simp: valid_def left_def right_def)

    lemma plr_corner_cases[simp]: 
      assumes "0<i"
      shows 
      "i\<noteq>parent i"
      "i\<noteq>left i"
      "i\<noteq>right i"
      "parent i \<noteq> i"
      "left i \<noteq> i"
      "right i \<noteq> i"
      using assms
      by (auto simp: parent_def left_def right_def)

    subsubsection \<open>Heap Property\<close>
    text \<open>The heap property states, that every node's priority is greater 
      or equal to its parent's priority \<close>
    definition is_heap :: "'a heap \<Rightarrow> bool"
      where "is_heap l 
      \<equiv> \<forall>i. valid l i \<longrightarrow> has_parent l i \<longrightarrow> pparent l i \<le> prio_of l i"
    
    lemma is_heap_empty[simp]: "is_heap []"
      by (auto simp: is_heap_def)

    function heap_induction_scheme :: "nat \<Rightarrow> unit" where
      "heap_induction_scheme i = (
        if i>1 then heap_induction_scheme (parent i) else ())"
      by pat_completeness auto  

    termination
      apply (relation "less_than")
      apply (auto simp: parent_def)
      done

    lemma 
      heap_parent_le: "\<lbrakk>is_heap l; valid l i; has_parent l i\<rbrakk> 
        \<Longrightarrow> pparent l i \<le> prio_of l i"
      unfolding is_heap_def
      by auto

    lemma heap_min_prop_aux:
      assumes H: "is_heap h"
      assumes V: "valid h i"
      shows "prio_of h 1 \<le> prio_of h i"
    proof (cases "i>1")
      case False with V show ?thesis
        by (auto simp: valid_def intro: Suc_lessI)
    next
      case True
      from V have "i\<le>length h" "valid h (Suc 0)" by (auto simp: valid_def)
      with True show ?thesis
        apply (induction i rule: heap_induction_scheme.induct)  
        apply (rename_tac i)
        apply (case_tac "parent i = Suc 0")
        apply (rule order_trans[rotated])
        apply (rule heap_parent_le[OF H])
        apply (auto simp: valid_def) [3]

        apply (rule order_trans)  
        apply (rprems)
        apply (auto simp: parent_def) [4]
        apply (rule heap_parent_le[OF H])
        apply (auto simp: valid_def parent_def)
        done
    qed

    lemma heap_min_prop:
      assumes H: "is_heap h" 
      assumes NE: "h\<noteq>[]"  
      shows "prio_of h 1 = Min (prio`set h)"
    proof -
      have "prio_of h 1 \<in> prio`set h"
        using NE by (auto simp: neq_Nil_conv val_of_def)
      moreover from heap_min_prop_aux[OF H] have "\<forall>p\<in>prio`set h. prio_of h 1 \<le> p"
        apply (auto simp: valid_def val_of_def in_set_conv_nth)
        by (metis One_nat_def Suc_leI diff_Suc_1 zero_less_Suc)
      ultimately show ?thesis 
        apply - 
        by (rule Min_eqI[THEN sym]) auto  
    qed    

    text \<open> Obviously, the heap property can also be stated in terms of children,
      i.e., each node's priority is smaller or equal to it's children's priority.\<close>

    definition "children_ge h p i \<equiv> 
      (has_left h i \<longrightarrow> p \<le> pleft h i)
    \<and> (has_right h i \<longrightarrow> p \<le> pright h i)"
    
    definition "is_heap' h \<equiv> \<forall>i. valid h i \<longrightarrow> children_ge h (prio_of h i) i"

    lemma heap_eq_heap':
      shows "is_heap h \<longleftrightarrow> is_heap' h"
      using assms 
      unfolding is_heap_def 
      unfolding is_heap'_def children_ge_def
      apply rule
      apply auto []
      apply clarsimp
      apply (frule child_of_parentD)
      apply auto []
      done

    subsection \<open>Auxiliary operations\<close>  
    subsubsection \<open>Exchange\<close>
    text \<open> Exchange two elements \<close>

    definition exch :: "'a heap \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a heap" where
      "exch l i j \<equiv> swap l (i - 1) (j - 1)"
        
    (*    
    definition exch :: "'a heap \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a heap" where
      "exch l i j \<equiv> let
        vi = l!(i - 1);
        vj = l!(j - 1);
        l = l[i - 1 := vj];
        l = l[j - 1 := vi]
      in 
        l"
    *)    
    
    lemma valid_exch[simp]: "valid (exch l i j) k = valid l k"
      unfolding exch_def by (auto simp: valid_def)
    
    lemma val_of_exch[simp]: "\<lbrakk>valid l i; valid l j; valid l k\<rbrakk> \<Longrightarrow> 
      val_of (exch l i j) k = (
        if k=i then val_of l j
        else if k=j then val_of l i
        else val_of l k
      )"
      unfolding exch_def val_of_def valid_def
      by (auto)

    lemma exch_eq[simp]: "exch h i i = h" 
      by (auto simp: exch_def)

    lemma \<alpha>_exch[simp]: "\<lbrakk>valid l i; valid l j\<rbrakk> 
      \<Longrightarrow> \<alpha> (exch l i j) = \<alpha> l"
      unfolding exch_def valid_def 
      by (auto)

    lemma length_exch[simp]: "length (exch l i j) = length l"
      by (auto simp: exch_def)

    definition "exch_op h i j \<equiv> do {
      ASSERT (valid h i \<and> valid h j);
      RETURN (exch h i j)
    }"


    subsubsection \<open>Butlast\<close>
    text \<open>Remove last element\<close>

    lemma valid_butlast_conv: "valid (butlast h) i \<longleftrightarrow> valid h i \<and> i < length h"
      by (auto simp: valid_def)

    lemma valid_butlast: "valid (butlast h) i \<Longrightarrow> valid h i"  
      by (cases h rule: rev_cases) (auto simp: valid_def)
    
    lemma val_of_butlast[simp]: "\<lbrakk>valid h i; i<length h\<rbrakk> 
      \<Longrightarrow> val_of (butlast h) i = val_of h i"
      by (auto simp: valid_def val_of_def nth_butlast)

    lemma val_of_butlast'[simp]: 
      "valid (butlast h) i \<Longrightarrow> val_of (butlast h) i = val_of h i"
      by (cases h rule: rev_cases) (auto simp: valid_def val_of_def nth_append)

    lemma \<alpha>_butlast[simp]: "\<lbrakk> length h \<noteq> 0 \<rbrakk> 
      \<Longrightarrow> \<alpha> (butlast h) = \<alpha> h - {# val_of h (length h)#}"
      apply (cases h rule: rev_cases)
      apply (auto simp: val_of_def)   
      done

    lemma is_heap_butlast[simp]: "is_heap h \<Longrightarrow> is_heap (butlast h)"
      apply (cases "h = []")
      apply simp
      apply (auto simp: is_heap_def dest: valid_butlast)
      done

    definition "butlast_op h \<equiv> do {ASSERT (h\<noteq>[]); RETURN (butlast h)}"

    subsubsection \<open>Update\<close>
    lemma valid_update[simp]: "valid (h[i:=v]) = valid h"
      by (auto simp: valid_def)

    definition "update_op h i v \<equiv> do {
      ASSERT (valid h i); 
      RETURN (h[i - Suc 0 := v])
    }"


    subsubsection \<open>Swim\<close>
    text \<open>Move up element that is too small, until it fits in\<close>
    function swim :: "'a heap \<Rightarrow> nat \<Rightarrow> 'a heap" where
      "swim h i = (
        if i>1 \<and> \<not> (prio_of h (parent i) \<le> prio_of h i) then 
          swim (exch h i (parent i)) (parent i)
        else
          h
      )"
      by pat_completeness auto  

    termination 
      apply (relation "inv_image less_than snd")
      apply (auto simp: valid_def parent_def)
      done

    declare swim.simps[simp del]

    definition "swim_op h i \<equiv> do {
      ASSERT (valid h i);
      RETURN (swim h i)
    }"

    lemma swim_op_unfold: "swim_op h i = do {
      ASSERT (valid h i);
      if i>1 \<and> \<not> (prio_of h (parent i) \<le> prio_of h i) then do {
        h \<leftarrow> exch_op h i (parent i);
        swim_op h (parent i)
      } else
        RETURN h
    }"
    unfolding swim_op_def
    apply (subst swim.simps)
    apply (auto simp: pw_eq_iff refine_pw_simps exch_op_def)
    apply (auto simp: valid_def parent_def)
    done

    lemma valid_has_parent_fold: "valid h i \<Longrightarrow> i>1 \<longleftrightarrow> has_parent h i"
      unfolding valid_def parent_def by auto

    lemma \<alpha>_swim: "valid h i \<Longrightarrow> \<alpha> (swim h i) = \<alpha> h"
      apply (induction h i rule: swim.induct)
      apply (subst swim.simps)
      unfolding exch_op_def
      apply (simp only: valid_has_parent_fold)
      apply auto
      done

    lemma valid_swim[simp]: "valid (swim h i) j = valid h j"  
      apply (induction h i rule: swim.induct)
      apply (subst swim.simps)
      apply auto
      done

    text \<open>This invariant expresses that the heap has a single defect,
      which can be repaired by swimming up\<close>  
    definition swim_invar :: "'a heap \<Rightarrow> nat \<Rightarrow> bool"
      where "swim_invar h i \<equiv> 
        valid h i
      \<and> (\<forall>j. valid h j \<and> has_parent h j \<and> j\<noteq>i \<longrightarrow> pparent h j \<le> prio_of h j)
      \<and> (has_parent h i \<longrightarrow> 
        (\<forall>j. valid h j \<and> has_parent h j \<and> parent j = i 
          \<longrightarrow> pparent h i \<le> prio_of h j))"

    lemma swim_invar_valid: "swim_invar h i \<Longrightarrow> valid h i"
      unfolding swim_invar_def by simp
    
    lemma swim_invar_exit1: "\<not>has_parent h i \<Longrightarrow> swim_invar h i \<Longrightarrow> is_heap h"
      unfolding is_heap_def swim_invar_def by auto

    lemma swim_invar_exit2: "pparent h i \<le> prio_of h i \<Longrightarrow> swim_invar h i \<Longrightarrow> is_heap h"
      unfolding is_heap_def swim_invar_def by auto

    lemma swim_invar_pres:
      assumes HPI: "has_parent h i" 
      assumes VIOLATED: "pparent h i > prio_of h i" 
      and INV: "swim_invar h i"
      defines "h' \<equiv> exch h i (parent i)"
      shows "swim_invar h' (parent i)"
      unfolding swim_invar_def
      apply safe
      apply (simp add: h'_def HPI)

      using HPI VIOLATED INV
      unfolding swim_invar_def h'_def
      apply auto []

      using HPI VIOLATED INV
      unfolding swim_invar_def h'_def
      apply auto
      by (metis order_trans)
    
    lemma swim_restores:
      assumes A: "swim_invar h i"
      shows "is_heap (swim h i)"
      using assms
      apply (induction h i rule: swim.induct)
      apply (subst swim.simps)
      apply (simp only: valid_has_parent_fold[OF swim_invar_valid])
      apply (auto simp: swim_invar_exit1 swim_invar_exit2 swim_invar_pres)
      done

    lemma swim_invar_decr:
      assumes INV: "is_heap h"
      assumes V: "valid h i"
      assumes DECR: "prio v \<le> prio_of h i"
      shows "swim_invar (h[i - Suc 0:=v]) i"
      using INV V

      apply (auto simp add: swim_invar_def)
      apply (auto
        intro!: order_trans[OF DECR]
        simp: val_of_def is_heap_def valid_def nth_list_update) [1]

      apply (clarsimp 
        simp: val_of_def is_heap_def valid_def nth_list_update eq_diff_iff)
      apply (auto intro: dual_order.trans) []
      done

    lemma swim_invar_decr':
      assumes INV: "is_heap h"
      assumes V: "i < length h"
      assumes DECR: "prio v \<le> prio (h!i)"
      shows "swim_invar (h[i := v]) (Suc i)"
      using swim_invar_decr[of h "Suc i" v] assms 
      by (auto simp: valid_def val_of_def)

    lemma swim_pres_heap[simp]: "\<lbrakk>is_heap h; valid h i\<rbrakk> \<Longrightarrow> swim h i = h"
      unfolding is_heap_def
      apply (subst swim.simps)
      apply (auto simp: valid_def parent_def)
      done

    subsubsection \<open>Sink\<close>
    text \<open>Move down an element that is too big, until it fits in\<close>

    function sink :: "'a heap \<Rightarrow> nat \<Rightarrow> 'a heap" where
    "sink h i = (
      if valid h i \<and> has_left h i then
        if has_right h i then
          if (pleft h i < prio_of h i \<and> pright h i \<ge> pleft h i) then
            sink (exch h i (left i)) (left i)
          else if (pright h i < pleft h i \<and> pright h i < prio_of h i) then
            sink (exch h i (right i)) (right i)
          else h
        else if (pleft h i < prio_of h i) then
          sink (exch h i (left i)) (left i)
        else h  
      else h
    )"
      by pat_completeness auto
    
    termination
      by (relation "measure (\<lambda>(l,i). length l - i)")  
         (auto simp add: valid_def left_def right_def Let_def)
    declare sink.simps[simp del]

    definition "sink_op h i \<equiv> do {
      ASSERT (i>0);
      RETURN (sink h i)
    }"

    lemma sink_op_unfold: "
      sink_op h i = do {
        ASSERT (i>0);
        if has_left h i then
          if has_right h i then
            if (pleft h i < prio_of h i \<and> pright h i \<ge> pleft h i) then do {
              h \<leftarrow> exch_op h i (left i);
              sink_op h (left i)
            } else if (pright h i < pleft h i \<and> pright h i < prio_of h i) then do {
              h \<leftarrow> exch_op h i (right i);
              sink_op h (right i)
            } else RETURN h
          else if (pleft h i < prio_of h i) then do {
            h \<leftarrow> exch_op h i (left i);
            sink_op h (left i)
          } else RETURN h  
        else RETURN h
      }"
      unfolding sink_op_def exch_op_def
      apply (subst sink.simps)
      by (auto simp: valid_def left_def)

    lemma sink_op_opt_unfold: 
      -- \<open>Sink operation as presented in Sedgewick et al. Algs4 reference implementation\<close>
      "sink_op h k = do {
        ASSERT (k>0);
        if (2*k \<le> length h) then do {
          let j = 2*k;
          let j = (if j<length h \<and> prio_of h j > prio_of h (Suc j) 
                   then j+1 else j);
          if (prio_of h k > prio_of h j) then do {
            h \<leftarrow> exch_op h k j;
            sink_op h j
          } else
            RETURN h
        } else RETURN h    
      }"  
      apply (subst sink_op_unfold)
      apply (auto simp: valid_def left_def right_def)
      done


    lemma sink_empty[simp]: "sink [] i = []" 
      by (subst sink.simps) auto
    
    lemma \<alpha>_sink[simp]: "\<alpha> (sink h i) = \<alpha> h"
      apply (induction h i rule: sink.induct)
      apply (subst sink.simps)
      apply auto
      done
    
    lemma valid_sink[simp]: "valid (sink h i) j = valid h j"  
      apply (induction h i rule: sink.induct)
      apply (subst sink.simps)
      apply auto
      done

    text \<open>This invariant expresses that the heap has a single defect, 
      which can be repaired by sinking\<close>
    definition "sink_invar l i \<equiv> 
      valid l i
    \<and> (\<forall>j. valid l j \<and> j\<noteq>i \<longrightarrow> children_ge l (prio_of l j) j)
    \<and> (has_parent l i \<longrightarrow> children_ge l (pparent l i) i)"
    
    lemma sink_invar_valid: "sink_invar l i \<Longrightarrow> valid l i"
      unfolding sink_invar_def by auto
    
    lemma sink_invar_exit: "\<lbrakk>sink_invar l i; children_ge l (prio_of l i) i\<rbrakk> 
      \<Longrightarrow> is_heap' l"
      unfolding is_heap'_def sink_invar_def
      by auto
    
    lemma sink_aux1: "\<not> (2*i \<le> length h) \<Longrightarrow> \<not>has_left h i \<and> \<not>has_right h i"
      unfolding valid_def left_def right_def by auto
    
    lemma sink_invar_pres1:
      assumes "sink_invar h i"
      assumes "has_left h i" "has_right h i"
      assumes "prio_of h i \<ge> pleft h i"
      assumes "pleft h i \<ge> pright h i"
      shows "sink_invar (exch h i (right i)) (right i)"
      using assms  
      unfolding sink_invar_def
      apply auto
      apply (auto simp: children_ge_def)
      done
    
    lemma sink_invar_pres2:
      assumes "sink_invar h i"
      assumes "has_left h i" "has_right h i"
      assumes "prio_of h i \<ge> pleft h i"
      assumes "pleft h i \<le> pright h i"
      shows "sink_invar (exch h i (left i)) (left i)"
      using assms  
      unfolding sink_invar_def
      apply auto
      apply (auto simp: children_ge_def)
      done
    
    lemma sink_invar_pres3:
      assumes "sink_invar h i"
      assumes "has_left h i" "has_right h i"
      assumes "prio_of h i \<ge> pright h i"
      assumes "pleft h i \<le> pright h i"
      shows "sink_invar (exch h i (left i)) (left i)"
      using assms  
      unfolding sink_invar_def
      apply auto
      apply (auto simp: children_ge_def)
      done
    
    lemma sink_invar_pres4:
      assumes "sink_invar h i"
      assumes "has_left h i" "has_right h i"
      assumes "prio_of h i \<ge> pright h i"
      assumes "pleft h i \<ge> pright h i"
      shows "sink_invar (exch h i (right i)) (right i)"
      using assms  
      unfolding sink_invar_def
      apply auto
      apply (auto simp: children_ge_def)
      done
    
    lemma sink_invar_pres5:
      assumes "sink_invar h i"
      assumes "has_left h i" "\<not>has_right h i"
      assumes "prio_of h i \<ge> pleft h i"
      shows "sink_invar (exch h i (left i)) (left i)"
      using assms  
      unfolding sink_invar_def
      apply auto
      apply (auto simp: children_ge_def)
      done
    
    lemmas sink_invar_pres = 
      sink_invar_pres1 
      sink_invar_pres2 
      sink_invar_pres3 
      sink_invar_pres4 
      sink_invar_pres5
    
    lemma sink_restores:
      assumes "sink_invar h i" 
      shows "is_heap (sink h i)"
      using assms
      unfolding heap_eq_heap'
    proof (induction h i rule: sink.induct)
      case (1 h i)
      note IH="1.IH"
      note SI=`sink_invar h i`
      hence V: "valid h i" by (blast intro: sink_invar_valid)
    
      show "is_heap' (sink h i)"
        apply (subst sink.simps)
        apply (split split_if, intro impI conjI)
        apply (split split_if, intro impI conjI)
        apply (split split_if, intro impI conjI)
        apply (rule IH(1))
          apply simp_all
          apply (rule sink_invar_pres2, auto simp: SI) []
        apply (intro impI conjI)  
        apply (rule IH(2))
          apply simp_all
          apply (rule sink_invar_pres4, auto simp: SI) []
        apply (rule sink_invar_exit[OF SI])  
          apply (auto simp: children_ge_def) []
        apply (intro impI conjI)  
        apply (rule IH(3))
          apply simp_all
          apply (rule sink_invar_pres5, auto simp: SI) []
        apply (rule sink_invar_exit[OF SI])  
          apply (auto simp: children_ge_def) []
        apply (rule sink_invar_exit[OF SI])  
          apply (auto simp: children_ge_def V
              intro: rc_imp_lc[OF V]) []
        done       
    qed

    lemma sink_invar_incr:
      assumes INV: "is_heap h"
      assumes V: "valid h i"
      assumes INCR: "prio v \<ge> prio_of h i"
      shows "sink_invar (h[i - Suc 0:=v]) i"
      using INV V INCR
      apply (auto simp: valid_def is_heap_def val_of_def sink_invar_def
        children_ge_def nth_list_update' 
        iff: eq_diff_iff
        intro: order_trans) (* Takes long ... Laptop ~2012: 5sec*)
      apply (frule spec[where x="left i"])
      apply auto []
      apply (frule spec[where x="right i"])
      apply auto []
      done

    lemma sink_invar_incr':
      assumes INV: "is_heap h"
      assumes V: "i < length h"
      assumes INCR: "prio v \<ge> prio (h!i)"
      shows "sink_invar (h[i := v]) (Suc i)"
      using sink_invar_incr[of h "Suc i" v] assms 
      by (auto simp: valid_def val_of_def)

    subsubsection \<open>Repair\<close>  
    text \<open>Repair a local defect in the heap. This can be done 
      by swimming and sinking. Note that, depending on the defect, only one
      of the operations will change the heap\<close>

    definition "repair h i \<equiv> if valid h i then swim (sink h i) i else h"
    
    definition "repair_op h i \<equiv> do {
      if (valid h i) then RETURN (repair h i) else RETURN h
    }"

    lemma repair_op_unfold: "repair_op h i = do {
      if valid h i then do {
        h \<leftarrow> sink_op h i;
        h \<leftarrow> swim_op h i;
        RETURN h
      } else RETURN h
    }"
    unfolding repair_def repair_op_def swim_op_def sink_op_def
    apply (auto)
    apply (auto simp: valid_def)
    done


    lemma is_heap_repair:
      fixes h i v
      assumes INV: "is_heap h"
      assumes V: "valid h i"
      defines "h' \<equiv> h[i - Suc 0 := v]"
      shows "is_heap (repair h' i)"
    proof -  
      have [simp]: "valid h' = valid h"
        by (auto simp: valid_def h'_def)

      {
        fix j assume "valid h j"
        hence "val_of h' j = (if i=j then v else val_of h j)"
          using V unfolding h'_def valid_def val_of_def
          by (auto simp: nth_list_update)
      } note val_of'=this

      {
        assume A: "prio v \<le> prio_of h i"
        with INV[unfolded heap_eq_heap'] V have 
          "has_left h' i \<longrightarrow> prio_of h' i \<le> pleft h' i"
          "has_right h' i \<longrightarrow> prio_of h' i \<le> pright h' i"
          unfolding is_heap'_def children_ge_def 
          by (auto simp: val_of')
        hence [simp]: "sink h' i = h'"  
          by (subst sink.simps) auto

        have ?thesis  
          apply (simp add: repair_def V)
          apply (rule swim_restores)
          using swim_invar_decr[OF INV V A]
          by (simp add: h'_def)
      } moreover {
        assume A: "prio v \<ge> prio_of h i"

        have ?thesis
          apply (simp add: repair_def V)
          apply (subst swim_pres_heap)
          apply (rule sink_restores)
          apply (rule sink_invar_incr[OF INV V A, folded h'_def])
          apply (simp add: V)
          done
      } ultimately show ?thesis by (metis linear) 
    qed  

    lemma is_heap_repair':
      fixes h i v
      assumes INV: "is_heap h"
      assumes V: "valid h (Suc i)"
      defines "h' \<equiv> h[i := v]"
      shows "is_heap (repair h' (Suc i))"
      using assms is_heap_repair[OF INV, of "Suc i"]
      by auto


    subsection \<open>Operations\<close>

    subsubsection \<open>Empty\<close>
    definition empty :: "'a heap" -- \<open>The empty heap\<close>
      where "empty \<equiv> []"
    lemma empty_invar[simp]: "is_heap empty"
      by (auto simp: empty_def)
    lemma empty_\<alpha>[simp]: "\<alpha> empty = {#}"  
      by (auto simp: empty_def)

    subsubsection \<open>Emptiness check\<close>  
    definition is_empty :: "'a heap \<Rightarrow> bool" -- \<open>Check for emptiness\<close>
      where "is_empty h \<equiv> h=[]"
    lemma is_empty_correct[simp]: "is_empty h \<longleftrightarrow> \<alpha> h = {#}"  
      by (auto simp: is_empty_def)

    subsubsection \<open>Insert\<close>
    definition ins :: "'a heap \<Rightarrow> 'a \<Rightarrow> 'a heap" -- \<open>Insert element\<close>
      where "ins h v \<equiv> swim (h@[v]) (length h + 1)"
    
    lemma swim_invar_insert: "is_heap l \<Longrightarrow> swim_invar (l@[x]) (Suc (length l))"
      unfolding swim_invar_def
      by (fastforce simp add: is_heap_def swim_invar_def valid_def parent_def val_of_def
          nth_append)

    lemma ins_invar[simp]: "is_heap h \<Longrightarrow> is_heap (ins h v)"
      unfolding ins_def
      apply (rule swim_restores)
      apply (simp add: swim_invar_insert)
      done
    
    lemma ins_\<alpha>[simp]: "is_heap l \<Longrightarrow> \<alpha> (ins l v) = (\<alpha> l) + {#v#}"
      unfolding ins_def
      apply (clarsimp)
      apply (subst \<alpha>_swim)
      apply (auto simp: valid_def)
      done

    definition "ins_op h v == do {
      ASSERT (is_heap h);
      RETURN (ins h v)
    }"

    lemma ins_op_unfold: "ins_op h v = do {
      ASSERT (is_heap h);
      let h = h@[v];
      swim_op h (length h)
    }"
      unfolding ins_op_def ins_def swim_op_def
      by (auto simp: pw_eq_iff refine_pw_simps valid_def)

    lemma ins_op_alt: "is_heap h \<Longrightarrow> ins_op h v = RETURN (ins h v)"
      by (auto simp: ins_op_def)


    subsubsection \<open>Pop minimum element\<close>  
    definition pop_min :: "'a heap \<Rightarrow> 'a heap \<times> 'a" -- \<open>Pop minimum element\<close>
      where
      "pop_min h \<equiv> let
        m = val_of h 1;
        h' = sink (butlast (exch h 1 (length h))) 1
      in (h',m)"
    
    lemma valid_child_lbound:
      assumes "valid h i" 
      shows "left i > Suc 0"
      and "right i > Suc 1"
      using assms
      unfolding valid_def left_def right_def
      by auto
    
    lemma left_ne1[simp]: "left j \<noteq> Suc 0" 
      unfolding left_def by auto
    
    lemma right_eq1_conv[simp]: "right j = Suc 0 \<longleftrightarrow> j=0" 
      unfolding right_def by auto
    
    lemma iparent1_conv[simp]: "parent (Suc 0) = 0"
      by (auto simp: parent_def)
    
    lemma sink_invar_init:
      assumes I: "is_heap h" 
      assumes NE: "length h > 1" 
      shows "sink_invar (butlast (exch h (Suc 0) (length h))) (Suc 0)"
    proof -
      from NE have V: "valid h (Suc 0)" "valid h (length h)" 
        apply -
        apply (auto simp: valid_def neq_Nil_conv) []
        by (cases h) (auto simp: valid_def)
    
      show ?thesis using I
        unfolding heap_eq_heap' is_heap'_def sink_invar_def
        apply (intro impI conjI allI)
        using NE apply (auto simp: V valid_butlast_conv) []
        apply (auto simp add: children_ge_def V NE valid_butlast_conv) []
        apply (auto simp add: children_ge_def V NE valid_butlast_conv) []
        done
    qed
    
    lemma mset_butlast_exch[simp]:
      assumes "valid h (Suc 0)"
      shows "multiset_of (butlast (exch h (Suc 0) (length h))) = multiset_of (tl h)"
      using assms
      apply (cases h, simp)
      apply (rename_tac x h')
      apply (case_tac h' rule: rev_cases)
      apply (auto 
        simp: exch_def valid_def Let_def swap_def
        split: nat.splits)
      done
    
    lemma pop_min_correct_aux:
      assumes I: "is_heap h" and NE[simp]: "h\<noteq>[]"
      assumes PMV: "pop_min h = (h',v)"
      shows 
        "is_heap h'" 
        "multiset_of h' = multiset_of h - {#v#}" 
        "prio v=Min (prio`set h)"
    proof -
      from NE have "length h \<noteq> 0" by auto
      moreover {
        assume "length h = 1"
        then obtain v where [simp]: "h=[v]" by (cases h) auto
        have "is_heap h'" using PMV unfolding pop_min_def by auto
      } moreover {
        assume A: "length h > Suc 0"
        have "is_heap h'" using PMV unfolding pop_min_def
          by (auto simp: sink_restores[OF sink_invar_init] I A)
      } ultimately show "is_heap h'"
        by (metis One_nat_def Suc_lessI neq0_conv)
        
      from NE have [simp]: "valid h (Suc 0)" by (cases h) (auto simp: valid_def) 
    
      show "multiset_of h' = multiset_of h - {#v#}"  
        using PMV unfolding pop_min_def
        by auto
    
      from NE PMV have "v = val_of h (Suc 0)" unfolding pop_min_def
        by auto
      thus "prio v=Min (prio`set h)"
        using heap_min_prop[OF I NE] by simp
    qed

    definition "pop_min_h \<equiv> fst o pop_min"
    definition "pop_min_v \<equiv> snd o pop_min"
    lemma pop_min_parts_conv[simp]: "pop_min h = (pop_min_h h, pop_min_v h)"
      by (auto simp: pop_min_h_def pop_min_v_def)

    lemma 
      assumes "is_heap h" "\<alpha> h \<noteq> {#}"
      shows
        pop_min_invar[simp]: "is_heap (pop_min_h h)"
      and
        pop_min_\<alpha>[simp]: "\<alpha> (pop_min_h h) = \<alpha> h - {# pop_min_v h #}"
      and
        pop_min_prio[simp]: "prio (pop_min_v h) = Min (prio`set_of (\<alpha> h))"
      using assms pop_min_correct_aux[of h]  
      by auto


    lemma butlast_exch_last_conv: "h\<noteq>[] 
      \<Longrightarrow> butlast (exch h (Suc 0) (length h)) = butlast (h[0 := last h])"
      apply (cases h rule: rev_cases)
      apply (auto simp: exch_def swap_def butlast_update'[symmetric])
      done

    definition "pop_min_op h \<equiv> do {
      ASSERT (is_heap h); ASSERT (\<alpha> h \<noteq> {#});
      RETURN (pop_min h)
    }"

    lemma pop_min_unfold: "pop_min_op h = do {
        ASSERT (is_heap h \<and> h \<noteq> []);
        let m = val_of h 1;
        let h = butlast (h[0 := last h]);
        h \<leftarrow> sink_op h 1;
        RETURN (h,m)
      }"
      unfolding pop_min_op_def pop_min_def sink_op_def
      by (auto simp: butlast_exch_last_conv 
        pw_eq_iff refine_pw_simps valid_def val_of_def)

    lemma pop_min_op_alt: "\<lbrakk>is_heap h; \<alpha> h \<noteq> {#}\<rbrakk> 
      \<Longrightarrow> pop_min_op h = RETURN (pop_min h)"  
      by (auto simp: pop_min_op_def)


    definition pop_min_spec :: "'a multiset \<Rightarrow> ('a multiset \<times> 'a) nres" where
      "pop_min_spec h \<equiv> SPEC (\<lambda>(h',v). h'=h - {#v#} \<and> prio v = Min (prio`set_of h))"

    definition "heap_rel \<equiv> br \<alpha> is_heap"  

    lemma pop_min_spec: "\<lbrakk>(h,m)\<in>heap_rel; m\<noteq>{#}\<rbrakk> 
      \<Longrightarrow> (pop_min_op h, pop_min_spec m) \<in> \<langle>heap_rel \<times>\<^sub>r Id\<rangle>nres_rel"  
      unfolding pop_min_op_def pop_min_spec_def heap_rel_def br_def nres_rel_def
      by (auto simp: pw_le_iff refine_pw_simps)

  end  

end

