theory Impl_Heap
imports Abs_Heap "../Sepref"
begin
  (* TODO: Move to Misc. Even to HOL/Product_Type? *)
  definition uncurry :: "('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'a \<times> 'b \<Rightarrow> 'c" where
    "uncurry f \<equiv> \<lambda>(a,b). f a b"
  
  lemma uncurry_apply[simp]: "uncurry f (a,b) = f a b"
    unfolding uncurry_def
    by simp
  
  lemma curry_uncurry_id[simp]: "curry (uncurry f) = f"
    unfolding uncurry_def
    by simp
  
  lemma uncurry_curry_id[simp]: "uncurry (curry f) = f"
    unfolding uncurry_def
    by simp

  lemma do_curry: "f (a,b) = curry f a b" by simp
  lemma do_uncurry: "f a b = uncurry f (a,b)" by simp


  (* TODO: Move to Hoare_Triple *)  
  lemma post_exI_rule: "<P> c <\<lambda>r. Q r x> \<Longrightarrow> <P> c <\<lambda>r. \<exists>\<^sub>Ax. Q r x>"
    by (blast intro: cons_post_rule ent_ex_postI ent_refl)
        
  (* TODO: Move to Refine_Basic:Convenience lemmas *)  
  lemma (in -) ret_le_down_conv: 
    "nofail m \<Longrightarrow> RETURN c \<le> \<Down>R m \<longleftrightarrow> (\<exists>a. (c,a)\<in>R \<and> RETURN a \<le> m)"
    by (auto simp: pw_le_iff refine_pw_simps)
    

  (* TODO: Move to Sepref_Basic *)  
  lemma hnr_pre_ex_conv: 
    shows "hn_refine (\<exists>\<^sub>Ax. \<Gamma> x) c \<Gamma>' R a \<longleftrightarrow> (\<forall>x. hn_refine (\<Gamma> x) c \<Gamma>' R a)"
    unfolding hn_refine_def
    apply safe
    apply (erule cons_pre_rule[rotated])
    apply (rule ent_ex_postI)
    apply (rule ent_refl)
    apply sep_auto
    done

  lemma hnr_pre_pure_conv:  
    shows "hn_refine (\<Gamma> * \<up>P) c \<Gamma>' R a \<longleftrightarrow> (P \<longrightarrow> hn_refine \<Gamma> c \<Gamma>' R a)"
    unfolding hn_refine_def
    by auto



  (* TODO: Move *)

  definition (in -) hr_comp :: "('b \<Rightarrow> 'c \<Rightarrow> assn) \<Rightarrow> ('b \<times> 'a) set \<Rightarrow> 'a \<Rightarrow> 'c \<Rightarrow> assn"
    where "hr_comp R1 R2 a c \<equiv> \<exists>\<^sub>Ab. R1 b c * \<up>((b,a)\<in>R2)"

  lemma hr_comp_Id2[simp]: "hr_comp R Id = R"  
    unfolding hr_comp_def[abs_def]
    apply (intro ext ent_iffI)
    by sep_auto+
    
  lemma hr_comp_Id1[simp]: "hr_comp (pure Id) R = pure R"  
    unfolding hr_comp_def[abs_def] pure_def
    apply (intro ext ent_iffI)
    by sep_auto+

  lemma hr_comp_prod_conv[simp]:
    "hr_comp (hn_prod_aux Ra Rb) (Ra' \<times>\<^sub>r Rb') 
    = hn_prod_aux (hr_comp Ra Ra') (hr_comp Rb Rb')"  
    unfolding hr_comp_def[abs_def] hn_prod_aux_def[abs_def]
    apply (intro ext ent_iffI)
    apply solve_entails apply clarsimp apply sep_auto
    apply clarsimp apply (intro ent_ex_preI)
    apply (rule ent_ex_postI) apply (sep_auto split: prod.splits)
    done

  lemma hr_comp_invalid[simp]: "hr_comp (\<lambda>a c. true) R a c = true * \<up>(\<exists>b. (b,a)\<in>R)"
    unfolding hr_comp_def[abs_def]
    apply (intro ext ent_iffI)
    apply sep_auto+
    done
    
  lemma hr_comp_emp[simp]: "hr_comp (\<lambda>a c. emp) R a c = \<up>(\<exists>b. (b,a)\<in>R)"
    unfolding hr_comp_def[abs_def]
    apply (intro ext ent_iffI)
    apply sep_auto+
    done

  lemma hr_compI: "(b,a)\<in>R2 \<Longrightarrow> R1 b c \<Longrightarrow>\<^sub>A hr_comp R1 R2 a c"  
    unfolding hr_comp_def
    by sep_auto

  lemma hn_rel_compI: 
    "\<lbrakk>nofail a; (b,a)\<in>\<langle>R2\<rangle>nres_rel\<rbrakk> \<Longrightarrow> hn_rel R1 b c \<Longrightarrow>\<^sub>A hn_rel (hr_comp R1 R2) a c"
    unfolding hr_comp_def hn_rel_def nres_rel_def
    apply (clarsimp intro!: ent_ex_preI)
    apply (drule (1) order_trans)
    apply (simp add: ret_le_down_conv)
    by sep_auto

  lemma hnr_comp:
    assumes R: "\<And>b1 c1. P b1 \<Longrightarrow> hn_refine (R1 b1 c1 * \<Gamma>) (c c1) (R1p b1 c1 * \<Gamma>') R (b b1)"
    assumes S: "\<And>a1 b1. \<lbrakk>Q a1; (b1,a1)\<in>R1'\<rbrakk> \<Longrightarrow> (b b1,a a1)\<in>\<langle>R'\<rangle>nres_rel"
    assumes PQ: "\<And>a1 b1. \<lbrakk>Q a1; (b1,a1)\<in>R1'\<rbrakk> \<Longrightarrow> P b1"
    assumes Q: "Q a1"
    shows "hn_refine 
      (hr_comp R1 R1' a1 c1 * \<Gamma>) 
      (c c1)
      (hr_comp R1p R1' a1 c1 * \<Gamma>') 
      (hr_comp R R') 
      (a a1)"
    unfolding hn_refine_alt
  proof clarsimp
    assume NF: "nofail (a a1)"
    show "
      <hr_comp R1 R1' a1 c1 * \<Gamma>> 
        c c1 
      <\<lambda>r. hn_rel (hr_comp R R') (a a1) r * (hr_comp R1p R1' a1 c1 * \<Gamma>')>\<^sub>t"
      apply (subst hr_comp_def)
      apply (clarsimp intro!: norm_pre_ex_rule)
    proof -
      fix b1
      assume R1: "(b1, a1) \<in> R1'"

      from S R1 Q have R': "(b b1, a a1) \<in> \<langle>R'\<rangle>nres_rel" by blast
      with NF have NFB: "nofail (b b1)" 
        by (simp add: nres_rel_def pw_le_iff refine_pw_simps)
      
      from PQ R1 Q have P: "P b1" by blast
      with NFB R have "<R1 b1 c1 * \<Gamma>> c c1 <\<lambda>r. hn_rel R (b b1) r * (R1p b1 c1 * \<Gamma>')>\<^sub>t"
        unfolding hn_refine_alt by auto
      thus "<R1 b1 c1 * \<Gamma>> 
        c c1 
        <\<lambda>r. hn_rel (hr_comp R R') (a a1) r * (hr_comp R1p R1' a1 c1 * \<Gamma>')>\<^sub>t"
        apply (rule cons_post_rule)
        apply (solve_entails)
        by (intro ent_star_mono hn_rel_compI[OF NF R'] hr_compI[OF R1] ent_refl)
    qed
  qed    

  lemma hnr_comp0:
    assumes "hn_refine \<Gamma> c \<Gamma>' R b"
    assumes "(b,a)\<in>\<langle>R'\<rangle>nres_rel"
    shows "hn_refine \<Gamma> c \<Gamma>' (hr_comp R R') a"
    using assms
    using hnr_comp[where P="\<lambda>_. True" and ?R1.0="\<lambda>_ _. emp" and \<Gamma>=\<Gamma> and \<Gamma>'=\<Gamma>'
      and ?R1p.0="\<lambda>_ _. emp" and Q="\<lambda>_. True" and R1'=UNIV 
      and a="\<lambda>_. a" and b="\<lambda>_. b" and c="\<lambda>_. c" and R=R and R'=R']
    by auto

  lemma hnr_comp1_aux:
    assumes R: "\<And>b1 c1. P b1 \<Longrightarrow> hn_refine (hn_ctxt R1 b1 c1) (c c1) (hn_ctxt R1p b1 c1) R (b$b1)"
    assumes S: "\<And>a1 b1. \<lbrakk>Q a1; (b1,a1)\<in>R1'\<rbrakk> \<Longrightarrow> (b$b1,a$a1)\<in>\<langle>R'\<rangle>nres_rel"
    assumes PQ: "\<And>a1 b1. \<lbrakk>Q a1; (b1,a1)\<in>R1'\<rbrakk> \<Longrightarrow> P b1"
    assumes Q: "Q a1"
    shows "hn_refine 
      (hr_comp R1 R1' a1 c1) 
      (c c1)
      (hr_comp R1p R1' a1 c1) 
      (hr_comp R R') 
      (a a1)"
    using assms hnr_comp[where \<Gamma>=emp and \<Gamma>'=emp and a=a and b=b and c=c and P=P and Q=Q]  
    unfolding hn_ctxt_def
    by auto

  lemma hnr_comp2_aux:
    assumes R: "\<And>b1 c1 b2 c2. P b1 b2 \<Longrightarrow> 
      hn_refine (R1 b1 c1 * R2 b2 c2) (c c1 c2) (R1p b1 c1 * R2p b2 c2) R (b b1 b2)"
    assumes S: "\<And>a1 b1 a2 b2. \<lbrakk>Q a1 a2; (b1,a1)\<in>R1'; (b2,a2)\<in>R2'\<rbrakk> \<Longrightarrow> (b b1 b2,a a1 a2)\<in>\<langle>R'\<rangle>nres_rel"
    assumes PQ: "\<And>a1 b1 a2 b2. \<lbrakk>Q a1 a2; (b1,a1)\<in>R1'; (b2,a2)\<in>R2'\<rbrakk> \<Longrightarrow> P b1 b2"
    assumes Q: "Q a1 a2"
    notes [simp] = hn_ctxt_def
    shows "hn_refine 
      (hr_comp R1 R1' a1 c1 * hr_comp R2 R2' a2 c2) 
      (c c1 c2)
      (hr_comp R1p R1' a1 c1 * hr_comp R2p R2' a2 c2) 
      (hr_comp R R') 
      (a a1 a2)"
    apply (rule hn_refine_cons)
    prefer 2
    apply (rule hnr_comp1_aux[where P = "uncurry P" and ?R1.0="hn_prod_aux R1 R2" 
      and c="uncurry c" and a="uncurry a"
      and R1p = "hn_prod_aux R1p R2p" and R=R and b="uncurry b" 
      and Q="uncurry Q" and R1'="R1'\<times>\<^sub>rR2'" and R'=R'
      and ?a1.0="(a1,a2)" and ?c1.0="(c1,c2)", simplified])
    using R apply (auto simp add: hn_prod_aux_def) []
    using S apply (auto) []
    using PQ apply auto []
    apply fact
    apply (simp add: hn_prod_aux_def)
    apply (simp add: hn_prod_aux_def)
    done

  definition "hnr1 P R1 R1' R a b \<equiv> \<forall>a1 b1. P a1 \<longrightarrow> hn_refine (R1 a1 b1) (b b1) (R1' a1 b1) R (a a1)"
  definition "nrr1 P R1 R a b \<equiv> \<forall>a1 b1. P a1 \<longrightarrow> (b1,a1)\<in>R1 \<longrightarrow> (b b1, a a1)\<in>\<langle>R\<rangle>nres_rel"

  lemma hnr_comp1:
    assumes "nrr1 P R1 R a b"
    assumes "hnr1 P' R1' R1p' R' b c"
    assumes "\<And>a b. \<lbrakk>P a; (b,a)\<in>R1\<rbrakk> \<Longrightarrow> P' b"
    shows "hnr1 P (hr_comp R1' R1) (hr_comp R1p' R1) (hr_comp R' R) a c"
    using hnr_comp1_aux[of P' R1' c R1p' R' b P R1 a R] assms
    unfolding hnr1_def nrr1_def hn_ctxt_def
    by auto

  lemma hnr1D:
    assumes "hnr1 P R1 R1p R a b"  
    assumes "bb1 = b b1"
    assumes "aa1 = a a1"
    assumes "\<Gamma> \<Longrightarrow>\<^sub>A R1 a1 b1"
    assumes "R1p a1 b1 \<Longrightarrow>\<^sub>A \<Gamma>'"
    assumes "P a1"
    assumes "R'=R"
    shows "hn_refine \<Gamma> bb1 \<Gamma>' R' aa1"
    using assms unfolding hnr1_def 
    apply simp
    apply (rule hn_refine_cons)
    by auto



  section \<open>Heaps by fixed-size arrays\<close>

  type_synonym 'v fsa_heap = "('v Heap.array \<times> nat)"

  definition fsa_repr :: "nat \<Rightarrow> 'v::heap Abs_Heap.heap \<Rightarrow> 'v fsa_heap \<Rightarrow> assn" where
    "fsa_repr maxsize \<equiv> \<lambda>h (p,n). \<exists>\<^sub>Al. 
        p\<mapsto>\<^sub>al 
      * \<up>((h=take n l \<and> n \<le> maxsize \<and> maxsize = length l) 
        \<or> (h=[] \<and> l=[] \<and> n=maxsize))"

  definition is_fsa_heap :: "('v \<Rightarrow> 'b::linorder) \<Rightarrow> nat \<Rightarrow> 'v::heap multiset \<Rightarrow> 'v fsa_heap \<Rightarrow> assn" where
    "is_fsa_heap prio maxsize \<equiv> \<lambda>s hi. \<exists>\<^sub>Ah. 
      fsa_repr maxsize h hi * \<up>(s = heapstruct.\<alpha> h \<and> heapstruct.is_heap prio h)"

  definition fsa_new :: "nat \<Rightarrow> 'v::heap fsa_heap Heap" where
    "fsa_new maxsize \<equiv> do {
      p \<leftarrow> Array.of_list [];  (* Note: Due to lack of witness elements, we 
        allocate an empty array first *)
      return (p,maxsize)
    }"

  definition fsa_alloc :: "'v::heap fsa_heap \<Rightarrow> 'v \<Rightarrow> 'v::heap fsa_heap Heap" 
    -- \<open> Ensure that array is allocated, once we have some default element \<close>
  where
    "fsa_alloc \<equiv> \<lambda>(p,n) v. do {
      l \<leftarrow> Array.len p;
      if l=0 \<and> n>0 then do {
        p \<leftarrow> Array.new n v;
        let n=0;
        return (p,n)
      } else
        return (p,n)
    }"

locale heapstruct_impl = heapstruct prio for prio :: "'a::heap \<Rightarrow> 'b::linorder"
begin

  lemma hnr_empty: (* Dependent relator, must be instantiated with maxsize *)
    "hn_refine emp (fsa_new maxsize) emp ((fsa_repr maxsize)) (RETURN empty)"
    apply rule
    apply (sep_auto simp: fsa_new_def hn_ctxt_def fsa_repr_def empty_def)
    done

  definition val_of_impl :: "'a::heap fsa_heap \<Rightarrow> nat \<Rightarrow> 'a Heap" 
    where "val_of_impl \<equiv> \<lambda>(p,n) i. do {
      let i = i - 1;
      Array.nth p i
    }"

  lemma hnr_val_of[sepref_fr_rules]: "SIDE_PRECOND (valid h i) \<Longrightarrow> hn_refine 
    (hn_ctxt (fsa_repr maxsize) h p * hn_val nat_rel i ii) 
    (val_of_impl p ii) 
    (hn_ctxt (fsa_repr maxsize) h p * hn_val nat_rel i ii)
    (pure Id)
    (RETURN$(val_of$h$i))"
    apply (cases p)
    apply rule
    apply (sep_auto simp add: hn_ctxt_def val_of_impl_def val_of_def valid_def pure_def fsa_repr_def)
    done

  definition exch_impl :: "'a::heap fsa_heap \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a fsa_heap Heap"
  where "exch_impl \<equiv> \<lambda>(p,n) i j. do {
    let i = i - 1; let j = j - 1;
    vi \<leftarrow> Array.nth p i;
    vj \<leftarrow> Array.nth p j;
    p\<leftarrow>Array.upd i vj p;
    p\<leftarrow>Array.upd j vi p;
    return (p,n)
  }"

  lemma hnr_exch[sepref_fr_rules]: "hn_refine 
    (hn_ctxt (fsa_repr maxsize) h p * hn_val nat_rel i ii * hn_val nat_rel j ji) 
    (exch_impl p ii ji) 
    (hn_invalid h p * hn_val nat_rel i ii * hn_val nat_rel j ji)
    (fsa_repr maxsize) 
    (exch_op$h$i$j)"
    unfolding APP_def
    apply (rule hn_refineI)
    unfolding exch_op_def exch_impl_def fsa_repr_def exch_def 
      Abs_Heap.swap_def hn_ctxt_def pure_def
    apply (cases p, simp)
    by (sep_auto simp: refine_pw_simps valid_def)

  definition butlast_impl :: "'a::heap fsa_heap \<Rightarrow> 'a fsa_heap Heap" where
    "butlast_impl \<equiv> \<lambda>(p,n). return (p,n - 1)"

  lemma hnr_butlast[sepref_fr_rules]: 
    "hn_refine 
      (hn_ctxt (fsa_repr maxsize) h p) (butlast_impl p) (hn_invalid h p) (fsa_repr maxsize) 
        (butlast_op$h)"
    apply (rule hn_refineI)
    unfolding butlast_op_def butlast_impl_def fsa_repr_def hn_ctxt_def APP_def
    by (sep_auto simp: refine_pw_simps butlast_take)
    
  definition update_impl :: "'a::heap fsa_heap \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a fsa_heap Heap"  
  where "update_impl \<equiv> \<lambda>(p,n) i v. do {
    p\<leftarrow>Array.upd (i - 1) v p;
    return (p,n)
  }"

  lemma hnr_update[sepref_fr_rules]: "hn_refine 
    (hn_ctxt (fsa_repr maxsize) h p * hn_val nat_rel i ii * hn_val Id v vi) 
    (update_impl p i v) 
    (hn_invalid h p * hn_val nat_rel i ii * hn_val Id v vi) 
    (fsa_repr maxsize) 
    (update_op$h$i$v)"  
    apply (rule hn_refineI)
    unfolding update_op_def update_impl_def fsa_repr_def hn_ctxt_def APP_def
    by (sep_auto simp: refine_pw_simps valid_def split: prod.split)


  definition "swim_while h i \<equiv> do {
    ASSERT (valid h i);
    (h,_) \<leftarrow> WHILEIT (\<lambda>(h,i). valid h i) (\<lambda>(h,i). i>1 \<and> \<not> (prio_of h (parent i) \<le> prio_of h i)) 
      (\<lambda>(h,i). do {
        h \<leftarrow> exch_op h i (parent i);
        let i = parent i;
        ASSERT (valid h i);
        RETURN (h,i)
      }) (h,i);
    RETURN h
    }"

  lemma swim_while_unfold: "swim_while h i = do {
      ASSERT (valid h i);
      if i>1 \<and> \<not> (prio_of h (parent i) \<le> prio_of h i) then do {
        h \<leftarrow> exch_op h i (parent i);
        swim_while h (parent i)
      } else
        RETURN h
    }"
    unfolding swim_while_def
    apply (subst WHILEIT_unfold)
    by auto

  lemma swim_op_as_while: "swim_op h i = swim_while h i"  
    apply (induction i arbitrary: h rule: heap_induction_scheme.induct)
    apply (subst swim_op_unfold)
    apply (subst swim_while_unfold)
    apply auto
    done  

  typedecl i_heap  
  typedecl i_elem  

  term val_of
  lemma [id_rules]:
    "prio ::\<^sub>i TYPE(i_elem\<Rightarrow>'b)"
    "val_of ::\<^sub>i TYPE(i_heap \<Rightarrow> nat \<Rightarrow> i_elem)"
    "valid ::\<^sub>i TYPE(i_heap \<Rightarrow> nat \<Rightarrow> bool)"
    "exch_op ::\<^sub>i TYPE(i_heap \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> i_heap nres)"
    by auto

  lemmas [sepref_la_skel] = mk_skel3[of exch_op]  
  lemmas [sepref_monadify_comb] = mk_mcomb3[of exch_op]

  lemma fsa_repr_precise[constraint_rules]: "precise (fsa_repr maxsize)"
    unfolding fsa_repr_def
    apply rule
    apply (auto dest!: preciseD[OF snga_prec])
    done

  lemma [sepref_import_param]:
    "(parent, parent) \<in> nat_rel \<rightarrow> nat_rel"  
    "(left, left) \<in> nat_rel \<rightarrow> nat_rel"  
    "(right, right) \<in> nat_rel \<rightarrow> nat_rel"  
    by auto

  abbreviation "prio_rel \<equiv> Id :: ('b \<times> 'b) set"
  

  lemma [sepref_import_param]:
    "(prio, prio) \<in> Id \<rightarrow> prio_rel"  
    "(op \<le>, op \<le>) \<in> prio_rel \<rightarrow> prio_rel \<rightarrow> bool_rel"
    "(op <, op <) \<in> prio_rel \<rightarrow> prio_rel \<rightarrow> bool_rel"
    by auto

  lemma valid_parent_gt_1: "valid h i \<Longrightarrow> Suc 0 < i \<Longrightarrow> valid h (parent i)"
    by (auto simp: valid_def parent_def)  

  schematic_lemma swim_impl:
    notes [id_rules] = itypeI[of h "TYPE(i_heap)"] itypeI[of i "TYPE(nat)"]
    notes [simp] = valid_parent_gt_1
    shows "hn_refine 
      (hn_ctxt (fsa_repr maxsize) h p * hn_val nat_rel i ii) 
      (?c::'a fsa_heap Heap) 
      ?\<Gamma> 
      ?R 
      (swim_op h i)"
    unfolding swim_op_as_while swim_while_def short_circuit_conv
    using [[id_debug, goals_limit=1]]
    apply sepref
    done
  concrete_definition swim_impl uses swim_impl

  lemma hnr_swim_impl[sepref_fr_rules]: 
  "hn_refine
    (hn_ctxt (fsa_repr maxsize) h p * hn_val nat_rel i ii)
    (heapstruct_impl.swim_impl prio p ii)
    (hn_invalid i ii * hn_invalid h p) (fsa_repr maxsize)
    (PR_CONST swim_op$h$i)"
    by (simp add: swim_impl.refine)


  lemma [def_pat_rules]: "heapstruct.swim_op$prio \<equiv> PR_CONST swim_op"
    by simp

  lemmas [sepref_monadify_comb] = mk_mcomb2[where c="PR_CONST swim_op"]
  lemmas [sepref_la_skel] = mk_skel2[where c="PR_CONST swim_op"]
  lemmas [id_rules] =
    itypeI[of "PR_CONST swim_op" "TYPE(i_heap \<Rightarrow> nat \<Rightarrow> i_heap nres)"]

  definition is_empty_impl :: "'a::heap fsa_heap \<Rightarrow> bool Heap" where
    "is_empty_impl \<equiv> \<lambda>(p,n). do {
      l \<leftarrow> Array.len p;
      return (l=0 \<or> n=0)
    }"

  lemma hnr_is_empty[sepref_fr_rules]:  
    "hn_refine
      (hn_ctxt (fsa_repr maxsize) h p)
      (is_empty_impl p)
      (hn_ctxt (fsa_repr maxsize) h p)
      (pure bool_rel)
      (RETURN$(is_empty$h))"
    apply rule  
    apply (sep_auto simp: hn_ctxt_def is_empty_impl_def fsa_repr_def pure_def)
    done


  definition heap_size_op :: "'a Abs_Heap.heap \<Rightarrow> nat" where
    [simp]: "heap_size_op \<equiv> length"

  lemmas [id_rules] =
    itypeI[of heap_size_op "TYPE(i_heap \<Rightarrow> nat)"]

  lemma [pat_rules]: 
    "length$h \<equiv> heap_size_op$h"
    by (auto simp: heap_size_op_def)
    
  definition heap_size_impl :: "'a::heap fsa_heap \<Rightarrow> nat Heap" where
    "heap_size_impl \<equiv> \<lambda>(p,n). do {
      l \<leftarrow> Array.len p;
      return (min l n)
    }"

  lemma hnr_heap_size_op[sepref_fr_rules]:  
    "hn_refine 
      (hn_ctxt (fsa_repr maxsize) h p) 
      (heap_size_impl p) 
      (hn_ctxt (fsa_repr maxsize) h p)
      (pure nat_rel)
      (RETURN$(heap_size_op$h))"
    apply rule  
    unfolding heap_size_op_def heap_size_impl_def hn_ctxt_def pure_def fsa_repr_def
    by sep_auto

    
  definition heap_append_op :: "'a Abs_Heap.heap \<Rightarrow> 'a \<Rightarrow> 'a Abs_Heap.heap" 
    where [simp]: "heap_append_op \<equiv> \<lambda>h a. h@[a]"

  lemmas [id_rules] =
    itypeI[of heap_append_op "TYPE(i_heap \<Rightarrow> i_elem \<Rightarrow> i_heap)"]

  lemma [pat_rules]: 
    "op @$h$(op #$a$[]) \<equiv> heap_append_op$h$a"
    by (auto simp: heap_append_op_def)


  definition heap_append_impl :: "'a::heap fsa_heap \<Rightarrow> 'a \<Rightarrow> 'a::heap fsa_heap Heap"
    where "heap_append_impl \<equiv> \<lambda>h a. do {
      (p,n) \<leftarrow> fsa_alloc h a;
      p \<leftarrow> Array.upd n a p;
      return (p, n + 1)
    }"

  lemma fsa_alloc_rule[sep_heap_rules]: 
    "<fsa_repr maxsize h p> fsa_alloc p a <\<lambda>(p,n). fsa_repr maxsize h (p,n) * \<up>(n = length h)>\<^sub>t"
    unfolding fsa_alloc_def fsa_repr_def
    apply (cases p) 
    apply sep_auto
    done

  thm if_distrib[where f=Heap_Monad.bind]


  lemma hnr_heap_append[sepref_fr_rules]:
    assumes "SIDE_PRECOND (length h < maxsize)"
    shows "hn_refine 
      (hn_ctxt (fsa_repr maxsize) h p * hn_val Id v vi)
      (heap_append_impl p vi)
      (hn_invalid h p * hn_val Id v vi)
      (fsa_repr maxsize)
      (RETURN$(heap_append_op$h$v))"
    apply rule  
    using assms
    unfolding hn_ctxt_def pure_def heap_append_impl_def heap_append_op_def
    unfolding fsa_repr_def fsa_alloc_def
    apply (cases p, clarsimp simp: if_distrib[where f=Heap_Monad.bind])
    apply (sep_auto)
    apply (cases maxsize, simp_all) []
    apply (sep_auto simp: take_update_last)
    done

  schematic_lemma ins_impl:
    notes [id_rules] = itypeI[of h "TYPE(i_heap)"] itypeI[of v "TYPE(i_elem)"]
    assumes "length h < maxsize"
    shows "hn_refine 
      (hn_ctxt (fsa_repr maxsize) h p * hn_val Id v vi) 
      (?c::'a fsa_heap Heap) 
      ?\<Gamma> 
      ?R 
      (PR_CONST ins_op$h$v)"
    unfolding ins_op_unfold PR_CONST_def APP_def
    using assms
    apply sepref
    done
  concrete_definition ins_impl uses ins_impl

  lemma [def_pat_rules]: "heapstruct.ins_op$prio \<equiv> PR_CONST ins_op"
    by simp

  lemmas [sepref_monadify_comb] = mk_mcomb2[where c="PR_CONST ins_op"]
  lemmas [sepref_la_skel] = mk_skel2[where c="PR_CONST ins_op"]
  lemmas [id_rules] =
    itypeI[of "PR_CONST ins_op" "TYPE(i_heap \<Rightarrow> i_elem \<Rightarrow> i_heap nres)"]



  definition "sink_rect h k \<equiv> RECT (\<lambda>sink (h,k). do {
    ASSERT (k>0);
    if (2*k \<le> length h) then do {
      let j = 2*k;
      let j = (if j<length h \<and> prio_of h j > prio_of h (Suc j) 
               then j+1 else j);
      if (prio_of h k > prio_of h j) then do {
        h \<leftarrow> exch_op h k j;
        sink (h,j)
      } else
        RETURN h
    } else RETURN h    
  }) (h,k)"  

  lemma sink_rect_unfold: "sink_rect h k = do {
      ASSERT (k>0);
      if (2*k \<le> length h) then do {
        let j = 2*k;
        let j = (if j<length h \<and> prio_of h j > prio_of h (Suc j) 
                 then j+1 else j);
        if (prio_of h k > prio_of h j) then do {
          h \<leftarrow> exch_op h k j;
          sink_rect h j
        } else
          RETURN h
      } else RETURN h    
    }"
    unfolding sink_rect_def
    apply (subst RECT_unfold, tagged_solver)
    apply simp
    done

  function sink_induction_scheme where  
    "sink_induction_scheme h k = ( 
      if k>0 \<and> 2*k \<le> length h then let
        x = sink_induction_scheme (exch h k (2*k)) (2*k)
      in (
        if 2*k<length h then 
          sink_induction_scheme (exch h k (Suc (2*k))) (Suc (2*k)) 
        else ()
      ) else ())"
    by pat_completeness auto

  termination  
    by (relation "measure (\<lambda>(l,i). length l - i)")  
       (auto simp add: valid_def left_def right_def Let_def)

  lemma sink_op_as_rect: "sink_op h i = sink_rect h i"
    apply (induction h i rule: sink_induction_scheme.induct)
    apply (subst sink_op_opt_unfold)
    apply (subst sink_rect_unfold)
    apply (subst pw_eq_iff)
    apply (auto 
      simp: refine_pw_simps exch_op_def left_def right_def valid_def
      simp: not_le)
    done 
    
  schematic_lemma sink_impl:
    notes [id_rules] = itypeI[of h "TYPE(i_heap)"] itypeI[of i "TYPE(nat)"]
    notes [simp] = validI
    shows "hn_refine 
      (hn_ctxt (fsa_repr maxsize) h p * hn_val nat_rel i ii) 
      (?c::'a fsa_heap Heap) 
      ?\<Gamma> 
      ?R 
      (PR_CONST sink_op$h$i)"
    unfolding PR_CONST_def APP_def  
    unfolding sink_op_as_rect sink_rect_def short_circuit_conv
    using [[id_debug, goals_limit=1]]
    by sepref
    
  concrete_definition sink_impl uses sink_impl

  lemmas [sepref_fr_rules] = sink_impl.refine

  lemmas [id_rules] =
    itypeI[of "PR_CONST sink_op" "TYPE(i_heap \<Rightarrow> nat \<Rightarrow> i_heap nres)"]
  lemma [def_pat_rules]: "heapstruct.sink_op$prio \<equiv> PR_CONST sink_op"
    by simp

  lemmas [sepref_monadify_comb] = mk_mcomb2[where c="PR_CONST sink_op"]
  lemmas [sepref_la_skel] = mk_skel2[where c="PR_CONST sink_op"]


  definition move_top_last_op :: "'a Abs_Heap.heap \<Rightarrow> 'a Abs_Heap.heap"
    where [simp]: "move_top_last_op h \<equiv> butlast (h[0 := last h])"

  lemmas [id_rules] =
    itypeI[of "move_top_last_op" "TYPE(i_heap \<Rightarrow> i_heap)"]
  lemma [pat_rules]: "butlast$(list_update$h$0$(last$h)) \<equiv> move_top_last_op$h"
    by simp


  definition move_top_last_impl :: "'a::heap fsa_heap \<Rightarrow> 'a::heap fsa_heap Heap"
    where "move_top_last_impl \<equiv> \<lambda>(p,n). do {
      v \<leftarrow> Array.nth p (n - 1);
      p \<leftarrow> Array.upd 0 v p;
      return (p, n - 1)
    }"

  lemma hnr_move_top_last_op[sepref_fr_rules]: 
  "SIDE_PRECOND (h \<noteq> []) \<Longrightarrow> hn_refine 
    (hn_ctxt (fsa_repr maxsize) h p)
    (move_top_last_impl p)
    (hn_invalid h p)
    (fsa_repr maxsize)
    (RETURN$(move_top_last_op$h))"
    apply rule
    unfolding move_top_last_impl_def hn_ctxt_def fsa_repr_def
    by (sep_auto simp: butlast_take butlast_update'[symmetric] last_take_nth_conv)

  lemma valid_firstI: "h\<noteq>[] \<Longrightarrow> valid h (Suc 0)"  
    by (auto simp: valid_def)
    
  schematic_lemma pop_min_impl: 
    notes [id_rules] = itypeI[of h "TYPE(i_heap)"]
    notes [simp] = valid_firstI
    shows "hn_refine 
      (hn_ctxt (fsa_repr maxsize) h p) 
      (?c::('a fsa_heap \<times> 'a) Heap) 
      ?\<Gamma> 
      ?R 
      (pop_min_op h)"
    unfolding pop_min_unfold
    by sepref
  concrete_definition pop_min_impl uses pop_min_impl
      
  lemma pop_min_impl_refine:
    shows
      "hn_refine (hn_ctxt (fsa_repr maxsize) h p)
        (heapstruct_impl.pop_min_impl prio p)
        (hn_invalid h p)
        (hn_prod_aux (fsa_repr maxsize) (pure Id))
        (pop_min_op $ h)"
    using pop_min_impl.refine by auto    

  lemmas [id_rules] =
    itypeI[of "PR_CONST pop_min_op" "TYPE(i_heap \<Rightarrow> (i_heap \<times> i_elem) nres)"]
  lemma [def_pat_rules]: "heapstruct.pop_min_op$prio \<equiv> PR_CONST pop_min_op"
    by simp

  lemmas [sepref_monadify_comb] = mk_mcomb1[where c="PR_CONST pop_min_op"]
  lemmas [sepref_la_skel] = mk_skel1[where c="PR_CONST pop_min_op"]



  lemma is_fsa_heap_alt: 
      "is_fsa_heap prio maxsize = hr_comp (fsa_repr maxsize) (heap_rel)"
    unfolding is_fsa_heap_def hr_comp_def[abs_def] heap_rel_def br_def
    by auto
    


  typedecl i_absheap
  lemmas [id_rules] = itypeI[of "{#}" "TYPE(i_absheap)"]

  (* TODO: Define hnr0 *)  
  lemma abs_empty_correct_rel: "(RETURN empty, RETURN {#})\<in>\<langle>heap_rel\<rangle>nres_rel"
    by (simp add: heap_rel_def br_def nres_rel_def)

  lemma hnr_empty_heap[sepref_fr_rules]: "hn_refine 
    emp
    (fsa_new maxsize)
    (emp)
    (is_fsa_heap prio maxsize)
    (RETURN${#})"
    using hnr_comp0[OF hnr_empty abs_empty_correct_rel]
    by (simp add: is_fsa_heap_alt)


  declare (in -) [[unify_trace_failure]]
  schematic_lemma "hn_refine emp (?c::?'a Heap) ?\<Gamma>' ?R (RETURN ({#}::'a multiset))"
    apply sepref_keep
    done    

xxx, ctd here ...  



  lemma pop_min_spec_nrr: "nrr1 (\<lambda>m. m\<noteq>{#}) heap_rel (heap_rel \<times>\<^sub>r Id) pop_min_spec pop_min_op"
    using pop_min_spec unfolding nrr1_def by auto

  lemma pop_min_op_hrr: "hnr1 
    (\<lambda>_. True) (fsa_repr maxsize) (\<lambda>_ _. true) (hn_prod_aux (fsa_repr maxsize) (pure Id))
    pop_min_op pop_min_impl"  
    using pop_min_impl.refine unfolding hnr1_def
    by (auto simp: hn_ctxt_def)

  lemma 
    assumes NE: "h \<noteq> {#}"
    shows "hn_refine
      (hn_ctxt (is_fsa_heap prio maxsize) h p)
      (pop_min_impl p)
      (hn_invalid h p)
      (hn_prod_aux (is_fsa_heap prio maxsize) (pure Id))
      (PR_CONST pop_min_spec$h)"
    apply (rule hnr1D[OF hnr_comp1[OF pop_min_spec_nrr pop_min_op_hrr]])
    apply (simp add: is_fsa_heap_alt hn_ctxt_def NE | rule ent_refl)+
    done





.xxx, ctd this way ... and set up sepref for abs<->heap


.xx
    xxx, ctd here ... try with a 1-parameter operation (pop_min), then
      do general lemmas for 2,3 params.





end

