header \<open>Sepref Tool\<close>
theory Sepref_Tool
imports Sepref_Translate
begin

text \<open>In this theory, we set up the sepref tool.\<close>

definition [simp]: "CNV x y \<equiv> x=y"

lemma ID_init: "\<lbrakk>ID a a' TYPE('T); hn_refine \<Gamma> c \<Gamma>' R a'\<rbrakk> 
  \<Longrightarrow> hn_refine \<Gamma> c \<Gamma>' R a" by simp

lemma TRANS_init: "\<lbrakk> hn_refine \<Gamma> c \<Gamma>' R a; CNV c c' \<rbrakk> 
  \<Longrightarrow> hn_refine \<Gamma> c' \<Gamma>' R a"
  by simp

lemma CNV_I: "CNV x x" by simp

ML {*
  structure Sepref = struct
    structure sepref_opt_simps = Named_Thms (
      val name = @{binding sepref_opt_simps}
      val description = "Sepref: Post-Translation optimizations, phase 1"
    )

    structure sepref_opt_simps2 = Named_Thms (
      val name = @{binding sepref_opt_simps2}
      val description = "Sepref: Post-Translation optimizations, phase 2"
    )

    fun id_tac ctxt = 
      rtac @{thm ID_init} 
      THEN' CONVERSION Thm.eta_conversion
      THEN' Id_Op.id_tac Id_Op.Normal ctxt

    fun id_init_tac ctxt = 
      rtac @{thm ID_init} 
      THEN' CONVERSION Thm.eta_conversion
      THEN' Id_Op.id_tac Id_Op.Init ctxt

    fun id_step_tac ctxt = 
      Id_Op.id_tac Id_Op.Step ctxt

    fun id_solve_tac ctxt = 
      Id_Op.id_tac Id_Op.Solve ctxt

    fun id_param_tac ctxt = CONVERSION (Refine_Util.HOL_concl_conv 
      (K (Sepref_Param.id_param_conv ctxt)) ctxt)

    fun monadify_tac ctxt = Sepref_Monadify.monadify_tac ctxt

    fun lin_ana_tac ctxt = Sepref_Lin_Ana.lin_ana_tac ctxt

    fun trans_tac ctxt = 
      rtac @{thm TRANS_init} 
      THEN' Sepref_Translate.trans_tac ctxt

    fun opt_tac ctxt = let 
      val opt1_ss = put_simpset HOL_basic_ss ctxt
        addsimps sepref_opt_simps.get ctxt
        addsimprocs [@{simproc "HOL.let_simp"}]
      |> Simplifier.add_cong @{thm SP_cong}
      |> Simplifier.add_cong @{thm PR_CONST_cong}

      val unsp_ss = put_simpset HOL_basic_ss ctxt addsimps @{thms SP_def}

      val opt2_ss = put_simpset HOL_basic_ss ctxt
        addsimps sepref_opt_simps2.get ctxt
        addsimprocs [@{simproc "HOL.let_simp"}]

    in 
      simp_tac opt1_ss THEN' simp_tac unsp_ss THEN'
      simp_tac opt2_ss THEN' simp_tac unsp_ss
    end

    fun PHASES _ [] _ = K all_tac
      | PHASES n (tac::tacs) ctxt = 
          IF_EXGOAL (tac ctxt)
          THEN_ELSE' (
            fn i => fn st => 
              if nprems_of st <= n then   (* Bail out if a phase leaves subgoals *)
                (PHASES n tacs ctxt i st)
              else
                Seq.single st
          , 
            K all_tac)

    fun sepref_tac ctxt i st = let
      val n = nprems_of st
    in
      PHASES n
        [ id_param_tac, 
          id_tac,
          monadify_tac,
          lin_ana_tac,
          trans_tac,
          opt_tac, 
          K (CONVERSION Thm.eta_conversion),
          K (rtac @{thm CNV_I})
        ]
        ctxt i st
    end
    
    val setup = sepref_opt_simps.setup #> sepref_opt_simps2.setup
  end
*}

setup Sepref.setup

method_setup sepref = \<open>Scan.succeed (fn ctxt =>
  SIMPLE_METHOD (DETERM (SOLVED' (IF_EXGOAL (
      Sepref.sepref_tac ctxt  
    )) 1)))\<close>
  \<open>Automatic refinement to Imperative/HOL\<close>

method_setup sepref_keep = \<open>Scan.succeed (fn ctxt =>
  SIMPLE_METHOD (IF_EXGOAL (Sepref.sepref_tac ctxt) 1))\<close>
  \<open>Automatic refinement to Imperative/HOL\<close>


subsubsection \<open>Debugging Methods\<close>
ML \<open>
  fun SIMPLE_METHOD_NOPARAM' tac = Scan.succeed (fn ctxt => SIMPLE_METHOD' (tac ctxt))
  fun SIMPLE_METHOD_NOPARAM tac = Scan.succeed (fn ctxt => SIMPLE_METHOD (tac ctxt))
\<close>
method_setup sepref_dbg_id_param = \<open>SIMPLE_METHOD_NOPARAM' Sepref.id_param_tac\<close>
  \<open>Sepref debug: Identify parameters phase\<close>
method_setup sepref_dbg_id = \<open>SIMPLE_METHOD_NOPARAM' Sepref.id_tac\<close>
  \<open>Sepref debug: Identify operations phase\<close>
method_setup sepref_dbg_id_init = \<open>SIMPLE_METHOD_NOPARAM' Sepref.id_init_tac\<close>
  \<open>Sepref debug: Initialize operation identification phase\<close>
method_setup sepref_dbg_id_step = \<open>SIMPLE_METHOD_NOPARAM' Sepref.id_step_tac\<close>
  \<open>Sepref debug: Single step operation identification phase\<close>
method_setup sepref_dbg_id_solve = \<open>SIMPLE_METHOD_NOPARAM' Sepref.id_solve_tac\<close>
  \<open>Sepref debug: Complete current operation identification goal\<close>

method_setup sepref_dbg_monadify = \<open>SIMPLE_METHOD_NOPARAM' Sepref.monadify_tac\<close>
  \<open>Sepref debug: Monadify phase\<close>
method_setup sepref_dbg_lin_ana = \<open>SIMPLE_METHOD_NOPARAM' Sepref.lin_ana_tac\<close>
  \<open>Sepref debug: Linearity analysis phase\<close>
method_setup sepref_dbg_trans = \<open>SIMPLE_METHOD_NOPARAM' Sepref.trans_tac\<close>
  \<open>Sepref debug: Translation phase\<close>
method_setup sepref_dbg_opt = \<open>SIMPLE_METHOD_NOPARAM' Sepref.opt_tac\<close>
  \<open>Sepref debug: Optimization phase\<close>

method_setup sepref_dbg_trans_step = \<open>SIMPLE_METHOD_NOPARAM' ( 
  Config.put Sepref_Translate.sepref_cfg_keep_side true #>
  Sepref_Translate.cstep_tac)\<close>
  \<open>Sepref debug: Translation phase single step\<close>

method_setup sepref_dbg_prepare_frame = \<open>SIMPLE_METHOD_NOPARAM' Sepref_Frame.prepare_frame_tac\<close>
  \<open>Sepref debug: Prepare frame inference\<close>

method_setup sepref_dbg_frame = \<open>SIMPLE_METHOD_NOPARAM' Sepref_Frame.frame_tac\<close>
  \<open>Sepref debug: Frame inference\<close>


lemmas [sepref_opt_simps] = return_bind bind_return bind_bind id_def

text \<open>We allow the synthesized function to contain tagged function applications.
  This is important to avoid higher-order unification problems when synthesizing
  generic algorithms, for example the to-list algorithm for foreach-loops.\<close>
lemmas [sepref_opt_simps] = Autoref_Tagging.APP_def


text {* Revert case-pulling done by monadify *}
lemma case_prod_return_opt[sepref_opt_simps]:
  "case_prod (\<lambda>a b. return (f a b)) p = return (case_prod f p)"
  by (simp split: prod.split)

lemma case_option_return_opt[sepref_opt_simps]:
  "case_option (return fn) (\<lambda>s. return (fs s)) v = return (case_option fn fs v)"
  by (simp split: option.split)

lemma case_list_return[sepref_opt_simps]:
  "case_list (return fn) (\<lambda>x xs. return (fc x xs)) l = return (case_list fn fc l)"
  by (simp split: list.split)

lemma if_return[sepref_opt_simps]:
  "If b (return t) (return e) = return (If b t e)" by simp

text {* In some cases, pushing in the returns is more convenient *}
lemma case_prod_opt2[sepref_opt_simps2]:
  "(\<lambda>x. return (case x of (a,b) \<Rightarrow> f a b)) 
  = (\<lambda>(a,b). return (f a b))"
  by auto

subsection {* Setup of Extraction-Tools *}
  declare [[cd_patterns "hn_refine _ ?f _ _ _"]]

  (* TODO: Move *)
  definition [simp, code_unfold]: "TRIV_EXTRACTION x \<equiv> x"

  lemma TRIV_EXTRACTION_codegen:
    assumes DEF: "f \<equiv> TRIV_EXTRACTION B"
    shows "f = B"
    using assms by simp

  lemma TRIV_extraction_cong: 
    "TRIV_EXTRACTION x \<equiv> TRIV_EXTRACTION x" 
    by simp

  setup {*
    Refine_Automation.add_extraction "trivial" {
      pattern = term_of @{cpat "TRIV_EXTRACTION _"},
      gen_thm = @{thm TRIV_EXTRACTION_codegen},
      gen_tac = (K (K all_tac))
    }
  *}

  lemma heap_fixp_codegen:
    assumes DEF: "f \<equiv> heap.fixp_fun cB"
    assumes M: "(\<And>x. mono_Heap (\<lambda>f. cB f x))"
    shows "f x = cB f x"
    unfolding DEF
    apply (rule fun_cong[of _ _ x])
    apply (rule heap.mono_body_fixp)
    apply fact
    done

  setup {*
    Refine_Automation.add_extraction "heap" {
      pattern = term_of @{cpat "heap.fixp_fun _"},
      gen_thm = @{thm heap_fixp_codegen},
      gen_tac = (fn ctxt => 
        (* TODO: This is a bit hacky: We should better handle those stuff in
          the mono-prover *)
        simp_tac 
          (put_simpset HOL_basic_ss ctxt addsimps @{thms TRIV_EXTRACTION_def}) 
        THEN'
        Pf_Mono_Prover.mono_tac ctxt
      )
    }
  *}


subsection \<open>Utility Lemmas\<close>
text \<open>To be replaced by more automatic methods\<close>

lemma 
      mk_skel1: "\<And>c. SKEL (c$x1) = la_op x1"
  and mk_skel2: "\<And>c. SKEL (c$x1$x2) = la_op (x1,x2)"   
  and mk_skel3: "\<And>c. SKEL (c$x1$x2$x3) = la_op (x1,x2,x3)"
    by auto

lemma 
      mk_mcomb1: "\<And>c. c$x1 \<equiv> op \<guillemotright>=$(EVAL$x1)$(\<lambda>\<^sub>2x1. SP (c$x1))"
  and mk_mcomb2: "\<And>c. c$x1$x2 \<equiv> op \<guillemotright>=$(EVAL$x1)$(\<lambda>\<^sub>2x1. op \<guillemotright>=$(EVAL$x2)$(\<lambda>\<^sub>2x2. SP (c$x1$x2)))"
  and mk_mcomb3: "\<And>c. c$x1$x2$x3 \<equiv> op \<guillemotright>=$(EVAL$x1)$(\<lambda>\<^sub>2x1. op \<guillemotright>=$(EVAL$x2)$(\<lambda>\<^sub>2x2. op \<guillemotright>=$(EVAL$x3)$(\<lambda>\<^sub>2x3. SP (c$x1$x2$x3))))"
    by auto

text \<open>Convert boolean operators to short-circuiting. 
  When applied before monadify, this will generate a short-circuit execution.\<close>
lemma short_circuit_conv:  
  "(a \<and> b) \<longleftrightarrow> (if a then b else False)"
  "(a \<or> b) \<longleftrightarrow> (if a then True else b)"
  "(a\<longrightarrow>b) \<longleftrightarrow> (if a then b else True)"
  by auto


end

