theory DFS
imports
  Main
  Graph_Ext
  "../CAVA_Refine"
begin

text {* The representation of a working state. *}
record ('S, 'n) dfs_sws =
  stack :: "'n list"   -- "the search stack of the current run"
  state :: "'S" -- "the state of the implementation"
  start :: "'n" -- "the node we started from"
(*  discover :: "('n, nat) map" -- "mapping a node to the timestamp of their discovery"
  finish :: "('n, nat) map" -- "mapping a node to the timestamp of the finishing, i.e. the moment the search backtracks"
  counter :: "nat" -- "the counter for discover and finish"*)
  discovered :: "'n set"

(*abbreviation "discovered s \<equiv> dom (discover s)"
abbreviation "finished s \<equiv> dom (finish s)"
abbreviation disc ("\<delta>") where "disc s x \<equiv> the (discover s x)"
abbreviation fin ("\<phi>") where "fin s x \<equiv> the (finish s x)"*)

definition "finished s \<equiv> discovered s - set (stack s)"

record ('S, 'n) dfs_algorithm =
  dfs_cond :: "'S \<Rightarrow> bool"  -- "the condition to be satisified by the state S to continue searching from here."
  dfs_visit :: "('S,'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'S nres"  -- "modifies the state for the current node BEFORE visiting the successors."
  dfs_post :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'S nres"  -- "modifies the state for the current node AFTER having visited the successors (i.e. during backtracking)."
  dfs_remove :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'S nres" -- "modifies the state if a node has already been visited and is removed from the stack"
  dfs_start :: "'n \<Rightarrow> 'S" -- "the starting state"
  dfs_pre_cond :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool"
  dfs_post_cond :: "('S, 'n) dfs_sws \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool"
  dfs_fe_inv :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'n set \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> bool"

(* TODO -- convert them sometime into something meaningful *)

lemma (in succs_digraph) FIXME_Y:
  "x \<in> V \<Longrightarrow> {z. \<exists>y \<in> (succs x - A). y \<rightarrow>\<setminus>A+ z} \<union> (succs x - A) \<union> {x} = 
   {z. \<exists>y \<in> (succs x - insert x A). y \<rightarrow>\<setminus>(insert x A)+ z} \<union> (succs x - insert x A) \<union> {x}"
apply (cases "x \<in> succs x")
apply rule
apply clarsimp
apply rule
apply clarsimp
apply (case_tac "y=x")
apply (metis  Diff_iff restr_reachable1_ne_split restricted_start restricted_target)
apply (metis (full_types) Diff_iff equalityI insert_subset restr_reachable1_ne_split restr_reachable1_trans restricted_start restricted_target subset_refl succs_restricted)
apply clarsimp
apply rule
apply clarsimp
apply (metis Diff_iff insertCI restr_reachable_restrict restricted_target subset_insertI)
apply clarsimp
apply rule
apply clarsimp
apply rule
apply clarsimp
apply (metis (full_types) Diff_iff equalityI insert_subset restr_reachable1_ne_split restr_reachable1_trans restricted_target subset_insertI subset_refl succs_restricted)
apply clarsimp
apply clarsimp
apply rule
apply clarsimp
apply (metis Diff_iff restr_reachable_restrict subset_insertI)
apply clarsimp
done

lemma (in succs_digraph) FIXME_Z:
  assumes "it \<subseteq> succs x"
  and "t \<in> it"
  and "t \<notin> set (stack s)"
  shows "{z. \<exists>y \<in> (succs x - it - set (stack s)). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - it - set (stack s))
          \<union> {t} \<union>  {z. t \<rightarrow>\<setminus>set (stack s)+ z} =
         {z. \<exists>y \<in> (succs x - (it - {t}) - set (stack s)). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - (it - {t}) - set (stack s))"
using assms by auto

lemma (in succs_digraph) FIXME_A:
  assumes "t \<notin> set (stack s)"
  and "t \<in> V"
  shows "{z. \<exists>y\<in>succs t - set (stack s). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs t - set (stack s)) =  {z. t  \<rightarrow>\<setminus>set (stack s)+ z}"
apply rule
apply rule
apply clarsimp
apply (metis Diff_iff assms restr_reachable1_trans succs_restricted)
apply rule
apply clarsimp
apply (case_tac "x \<in> succs t")
apply (clarsimp simp add: restricted_target)
by (metis restr_reachable1_split)

locale DFS_def = succs_digraph succs v\<^sub>0
  for succs :: "'n \<Rightarrow> 'n set" and v\<^sub>0 :: "'n" +
  fixes dfs :: "('S,'n) dfs_algorithm"
begin

definition dfs_alg_start :: "'n \<Rightarrow> ('S, 'n) dfs_sws"
(*where "dfs_alg_start x\<^sub>0 \<equiv> \<lparr> stack = [], discover = Map.empty, finish = Map.empty, counter = 0, state = dfs_start dfs x\<^sub>0, start = x\<^sub>0\<rparr>"*)
where "dfs_alg_start x\<^sub>0 \<equiv> \<lparr> stack = [], state = dfs_start dfs x\<^sub>0, start = x\<^sub>0, discovered = {} \<rparr>"

definition disc_step_upd :: "['n, ('S, 'n) dfs_sws, 'S] \<Rightarrow> ('S, 'n) dfs_sws" where
 (* "disc_step_upd x s S \<equiv> s\<lparr>stack := x # stack s, discover := discover s(x \<mapsto> counter s), counter := Suc (counter s), state := S\<rparr>"*)
  "disc_step_upd x s S \<equiv> s\<lparr>stack := x # stack s, discovered := insert x (discovered s), state := S\<rparr>"

definition fin_step_upd :: "['n, ('S, 'n) dfs_sws, 'S] \<Rightarrow> ('S, 'n) dfs_sws" where
(*  "fin_step_upd x s S \<equiv> s\<lparr>stack := tl (stack s), finish := finish s(x\<mapsto> counter s), counter := Suc (counter s), state := S\<rparr>"  *)
  "fin_step_upd x s S \<equiv> s\<lparr>stack := tl (stack s), state := S\<rparr>"

definition dfs_disc_step :: "'n \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> ('S, 'n) dfs_sws nres" where
  "dfs_disc_step x s \<equiv> do {
       S \<leftarrow> dfs_visit dfs s x;
       RETURN (disc_step_upd x s S)
  }"

definition dfs_fin_step :: "'n \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> ('S, 'n) dfs_sws nres" where
  "dfs_fin_step x s \<equiv> if dfs_cond dfs (state s) then do {
      S \<leftarrow> dfs_post dfs s x;
      RETURN (fin_step_upd x s S)
    } else RETURN s"

definition dfs_rem_step :: "'n \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> ('S, 'n) dfs_sws nres" where
  "dfs_rem_step x s \<equiv> do {
      S' \<leftarrow> dfs_remove dfs s x;
      RETURN (s\<lparr> state := S' \<rparr>)
  }"

definition sws_rel :: "('S, 'n) dfs_sws rel" where
  "sws_rel \<equiv> inv_image (finite_psupset V) discovered"

subsubsection "Invariants"

(*definition timing_inv :: "('S, 'n) dfs_sws \<Rightarrow> bool"
where "timing_inv s \<longleftrightarrow> 
  (\<forall>v \<in> discovered s. \<delta> s v < counter s)
  \<and> (\<forall>v \<in> finished s. \<phi> s v < counter s)
  \<and> (\<forall>v \<in> finished s. \<phi> s v > \<delta> s v)"*)

definition common_inv :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool"
where "common_inv s x \<longleftrightarrow>
  x \<in> V
  \<and> distinct (stack s)
  \<and> (\<forall> n < (length (stack s)) - 1. (stack s ! n) \<in> succs (stack s ! Suc n))
  \<and> (stack s \<noteq> [] \<longrightarrow> last (stack s) = start s)
  \<and> (stack s = [] \<longrightarrow> start s = x)
  \<and> (stack s \<noteq> [] \<longrightarrow> vwalk (rev (stack s)) \<G>)
  \<and> discovered s \<subseteq> V 
  \<and> finished s \<union> set (stack s) = discovered s
  \<and> set (stack s) \<inter> finished s = {}"

definition pre_dfs' :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool"
where
"pre_dfs' s x \<longleftrightarrow>
  (stack s = [] \<longleftrightarrow> discovered s = {})
  \<and> (stack s \<noteq> [] \<longrightarrow> x \<in> succs (hd (stack s)))
  \<and> (\<forall>y \<in> finished s. (\<forall>z. y \<rightarrow>\<setminus>set (stack s)+ z \<longrightarrow> z \<in> finished s))"

definition pre_dfs :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool"
where
"pre_dfs s x \<longleftrightarrow>
  pre_dfs' s x
(*  \<and> timing_inv s*)
  \<and> common_inv s x"

definition post_dfs' :: "[('S, 'n) dfs_sws, ('S, 'n) dfs_sws, 'n] \<Rightarrow> bool"
where
"post_dfs' s s' x \<longleftrightarrow>
  discovered s \<subseteq> discovered s'
  \<and> finished s \<subseteq> finished s'
  \<and> start s = start s'
  \<and> (\<exists>xs. stack s' = xs@stack s)
  \<and> (\<not> dfs_cond dfs (state s) \<longrightarrow> \<not> dfs_cond dfs (state s'))

  (* conditionals -- non-aborts *)
  \<and> (dfs_cond dfs (state s') \<longrightarrow> 
      stack s' = stack s
      \<and> x \<in> discovered s'
      \<and> (x \<notin> discovered s \<longrightarrow> discovered s \<union> {z. \<exists>y \<in> succs x - set (stack s). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - set (stack s)) \<union> {x} = discovered s')
      \<and> (x \<in> discovered s \<longrightarrow> discovered s = discovered s')
      \<and> (x \<notin> discovered s \<longrightarrow> finished s \<union> {z. \<exists>y \<in> succs x - set (stack s). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - set (stack s)) \<union> {x} = finished s')
      \<and> (x \<in> discovered s \<longrightarrow> finished s = finished s')
      \<and> (\<forall>y \<in> finished s'. (\<forall>z. y \<rightarrow>\<setminus>set (stack s)+ z \<longrightarrow> z \<in> finished s'))
    )"

definition post_dfs :: "[('S, 'n) dfs_sws, ('S, 'n) dfs_sws, 'n] \<Rightarrow> bool"
where
"post_dfs s s' x \<longleftrightarrow>
  (* for convencience *)
  pre_dfs s x

  (* general *)
  \<and> post_dfs' s s' x
(*  \<and> timing_inv s'*)
  \<and> common_inv s' x"

definition fe_inv_dfs' :: "[('S, 'n) dfs_sws, 'n, 'n set, ('S, 'n) dfs_sws] \<Rightarrow> bool"
where
"fe_inv_dfs' s x it s' \<longleftrightarrow>
  discovered s \<subseteq> discovered s'
  \<and> finished s \<subseteq> finished s'
  \<and> start s = start s'
  \<and> (\<exists>xs. stack s' = xs@stack s)
  \<and> (\<not> dfs_cond dfs (state s) \<longrightarrow> \<not> dfs_cond dfs (state s'))
   
  (* non-aborting *)
  \<and> (dfs_cond dfs (state s') \<longrightarrow>
      stack s' = stack s
      \<and> discovered s \<union> {z . \<exists>y \<in> succs x - it - set (stack s). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - it - set (stack s)) = discovered s'
      \<and> finished s  \<union> {z . \<exists>y \<in> succs x - it - set (stack s). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - it - set (stack s)) = finished s'
      \<and> hd (stack s) = x
      \<and> (\<forall>y \<in> finished s'. (\<forall>z. y \<rightarrow>\<setminus>set (stack s)+ z \<longrightarrow> z \<in> finished s'))
    )"

definition fe_inv_dfs :: "[('S, 'n) dfs_sws, 'n, 'n set, ('S, 'n) dfs_sws] \<Rightarrow> bool"
where
"fe_inv_dfs s x it s' \<longleftrightarrow>
  (* general *)
  fe_inv_dfs' s x it s'
(*  \<and> timing_inv s'*)
  \<and> common_inv s' x"

lemmas pre_dfs_unfold = pre_dfs_def pre_dfs'_def common_inv_def
lemmas post_dfs_unfold = post_dfs_def post_dfs'_def common_inv_def 
lemmas fe_inv_dfs_unfold = fe_inv_dfs_def fe_inv_dfs'_def common_inv_def

definition dfs_algo_body :: "('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> ('S, 'n) dfs_sws nres" where
  "dfs_algo_body s\<^sub>0 x\<^sub>0 \<equiv> REC\<^sub>T (\<lambda>D (s, x). do {
           ASSERT (pre_dfs s x \<and> dfs_pre_cond dfs s x \<and> dfs_cond dfs (state s));
           if x \<in> discovered s then dfs_rem_step x s
           else
              (dfs_disc_step x s \<guillemotright>=
              FOREACH\<^sub>C (succs x) (\<lambda>s. dfs_cond dfs (state s)) (\<lambda> x s. D (s, x))) \<guillemotright>=
              (\<lambda> s'. ASSERT (\<exists>S. dfs_cond dfs (state s') \<longrightarrow> 
                                 inres (dfs_visit dfs s x) S 
                               \<and> fe_inv_dfs (disc_step_upd x s S) x {} s' 
                               \<and> dfs_fe_inv dfs (disc_step_upd x s S) x {} s' ) \<guillemotright>
              dfs_fin_step x s')
        }) (s\<^sub>0, x\<^sub>0)"

definition dfs_algo :: "'n \<Rightarrow> ('S, 'n) dfs_sws nres"  where
  "dfs_algo x\<^sub>0 \<equiv> do {
     let s\<^sub>0 = dfs_alg_start x\<^sub>0;
     if \<not> dfs_cond dfs (state s\<^sub>0) then RETURN s\<^sub>0
     else dfs_algo_body s\<^sub>0 x\<^sub>0
   }"

(* to show
lemma post_imp_pre:
  "dfs_cond dfs (state s') \<Longrightarrow> post_dfs s s' x \<Longrightarrow> pre_dfs s' x" 
*)

(*lemma disc_step_timing_inv:
  "x \<notin> discovered s \<Longrightarrow> x \<notin> finished s \<Longrightarrow> timing_inv s \<Longrightarrow> timing_inv (disc_step_upd x s S)"
unfolding disc_step_upd_def timing_inv_def by simp smt

lemma fin_step_timing_inv:
  "x \<in> discovered s \<Longrightarrow> timing_inv s \<Longrightarrow> timing_inv (fin_step_upd x s S)"
unfolding fin_step_upd_def timing_inv_def by simp smt*)

(* lifting to another DFS *)
lemma post_lift:
  assumes C: "dfs_cond dfs = dfs_cond dfs'"
  shows "post_dfs s s' x = DFS_def.post_dfs succs v\<^sub>0 dfs' s s' x"
unfolding post_dfs_def DFS_def.post_dfs_def
unfolding post_dfs'_def DFS_def.post_dfs'_def
by (auto simp: C)

lemma fe_inv_lift:
  assumes C: "dfs_cond dfs = dfs_cond dfs'"
  shows "fe_inv_dfs s x it s' = DFS_def.fe_inv_dfs succs v\<^sub>0 dfs' s x it s'"
unfolding fe_inv_dfs_def DFS_def.fe_inv_dfs_def
unfolding fe_inv_dfs'_def DFS_def.fe_inv_dfs'_def
by (auto simp: C)

(* common_inv rules *)
lemma common_inv_x_in_V[intro]:
  "common_inv s x \<Longrightarrow> x \<in> V"
by (simp add: common_inv_def)

lemma common_inv_stack_distinct:
  "common_inv s x \<Longrightarrow> distinct (stack s)"
by (simp add: common_inv_def)

lemma discovered_subset_verts:
  "common_inv s x \<Longrightarrow> discovered s \<subseteq> V"
by (simp add: common_inv_def)

lemma finished_subset_discovered:
  "common_inv s x \<Longrightarrow> finished s \<subseteq> discovered s"
by (auto simp add: common_inv_def)

lemma finished_subset_verts:
  "common_inv s x \<Longrightarrow> finished s \<subseteq> V"
by (metis finished_subset_discovered discovered_subset_verts subset_trans)

lemma stack_subset_discovered:
  "common_inv s x \<Longrightarrow> set (stack s) \<subseteq> discovered s"
by (auto simp add: common_inv_def)

lemma stack_subset_verts:
  "common_inv s x \<Longrightarrow> set (stack s) \<subseteq> V"
by (metis stack_subset_discovered discovered_subset_verts subset_trans)

(*
lemma pre_ran_discover_finite:
  "common_inv s x \<Longrightarrow> finite (ran (discover s))" 
by (metis discovered_finite map_dom_ran_finite)

lemma ran_finished_finite:
  "common_inv s x \<Longrightarrow> finite (ran (finish s))"

by (metis finished_finite map_dom_ran_finite)
*)
lemma finished_implies_discovered:
  "common_inv s x \<Longrightarrow> v \<in> finished s \<Longrightarrow> v \<in> discovered s"
by (auto dest!: finished_subset_discovered)

lemma not_discovered_implies_not_finished:
  "common_inv s x \<Longrightarrow> v \<notin> discovered s \<Longrightarrow> v \<notin> finished s"
by (metis finished_implies_discovered)

lemma stack_implies_discovered:
  "common_inv s x \<Longrightarrow> v \<in> set (stack s) \<Longrightarrow> v \<in> discovered s"
by (auto dest!: stack_subset_discovered)

lemma not_discovered_implies_non_stack:
  "common_inv s x \<Longrightarrow> v \<notin> discovered s \<Longrightarrow> v \<notin> set (stack s)"
by (metis stack_implies_discovered)

lemma stack_implies_not_finished:
  "common_inv s x \<Longrightarrow> v \<in> set(stack s) \<Longrightarrow> v \<notin> finished s"
by (auto simp add: common_inv_def)

lemma finished_implies_not_stack:
  "common_inv s x \<Longrightarrow> v \<in> finished s \<Longrightarrow> v \<notin> set (stack s)"
by (metis stack_implies_not_finished)

lemma discovered_non_stack_implies_finished:
  "common_inv s x \<Longrightarrow> v \<in> discovered s \<Longrightarrow> v \<notin> set (stack s) \<Longrightarrow> v \<in> finished s"
by (auto simp add: common_inv_def)

lemma disc_upd_finished [intro, simp]:
  "x \<notin> discovered s \<Longrightarrow> finished (disc_step_upd x s S) = finished s"
unfolding disc_step_upd_def finished_def
by auto

lemma dfs_alg_start_finished [simp]:
  "finished (dfs_alg_start x) = {}"
unfolding dfs_alg_start_def finished_def
by simp

lemma fin_upd_finished:
  assumes CI: "common_inv s x"
  and "x = hd (stack s)"
  and "stack s \<noteq> []"
  shows "finished (fin_step_upd x s S) = insert x (finished s)"
proof -
  from CI have "distinct (stack s)" unfolding common_inv_def by metis
  with `x = hd (stack s)` have "x \<notin> set (tl (stack s))" by (metis distinct_hd_tl)
  moreover from assms hd_in_set stack_implies_discovered have "x \<in> discovered s" by auto
  ultimately show ?thesis
    using assms
    using in_set_tlD[of _ "stack s"] not_hd_in_tl[of _ "stack s"]
    unfolding finished_def fin_step_upd_def
    by auto
qed

lemma start_discovered:
  assumes "common_inv s x"
  and "stack s \<noteq> []"
  shows "start s \<in> discovered s"
proof -
  from assms have "start s = last (stack s)" by (auto simp: common_inv_def)
  with assms last_in_set have "start s \<in> set (stack s)" by auto
  with assms show ?thesis by (auto simp: common_inv_def)
qed

lemma not_start_imp_stack:
  "common_inv s x \<Longrightarrow> x \<noteq> start s \<Longrightarrow> stack s \<noteq> []"
by (simp add: common_inv_def)

lemma stack_in_Suc_succs:"
  common_inv s x \<Longrightarrow> n < length (stack s) - 1 \<Longrightarrow> stack s ! n \<in> succs (stack s ! Suc n)"
unfolding common_inv_def
by metis

lemma Suc_stack_stack_in_edges:
  "common_inv s x \<Longrightarrow> n < length (stack s) - 1 \<Longrightarrow> (stack s ! Suc n, stack s ! n) \<in> E"
by (smt nth_mem set_rev_mp stack_in_Suc_succs stack_subset_verts succs_implies_edge)

lemma stack_Suc_reachable_stack:
  "common_inv s x \<Longrightarrow> n < (length (stack s)) - 1 \<Longrightarrow> stack s ! Suc n \<rightarrow>+ stack s ! n"
by (metis Suc_stack_stack_in_edges edge_in_nodes stack_in_Suc_succs succs_reachable)

lemma stack_prev_reachable_stack:
  "common_inv s x \<Longrightarrow> n < length (stack s) \<Longrightarrow> m < n \<Longrightarrow> stack s ! n \<rightarrow>+ stack s ! m"
by (metis reachable1_trancl nth_step_trancl Suc_stack_stack_in_edges)

lemma stack_tl_reachable_stack_hd:
  assumes sinv: "common_inv s x"
  and "x \<in> set (tl (stack s))"
  shows "x \<rightarrow>+ hd (stack s)"
proof -
  from assms have ne: "stack s \<noteq> []" by auto
  from assms obtain n where "n < length (tl (stack s))" "tl (stack s) ! n = x" unfolding in_set_conv_nth by auto
  with nth_tl[OF ne] have "Suc n < length (stack s)" "stack s ! (Suc n) = x"  by simp_all
  moreover with stack_prev_reachable_stack[OF sinv, of "Suc n" 0] have "stack s ! Suc n \<rightarrow>+ stack s ! 0" by simp
  ultimately show ?thesis using hd_conv_nth[OF ne] by simp
qed

lemma start_reachable_stack:
  assumes sinv: "common_inv s x"
  and "x \<in> set (stack s)"
  and "set (stack s) \<subseteq> V"
  shows "start s \<rightarrow>\<star> x"
proof (cases "x = start s")
  case True with self_reachable assms show ?thesis by auto
next
  case False with assms have "stack s \<noteq> []" by auto
  with sinv last_conv_nth have s_nth: "start s = stack s ! (length (stack s) - 1)" unfolding common_inv_def by metis
  
  from `x \<in> set (stack s)` obtain n where n: "n < length (stack s)" "stack s ! n = x" unfolding in_set_conv_nth by auto
  
  have "n < length (stack s) - 1"
  proof (rule ccontr)
    assume "\<not> n < length (stack s) - 1"
    with n have "n = length (stack s) - 1" by auto
    with s_nth n False show "False" by simp
  qed
  with stack_prev_reachable_stack[OF sinv] have "stack s ! (length (stack s) - 1) \<rightarrow>+ stack s ! n" by auto
  with s_nth n have "start s \<rightarrow>+ x" by auto
  thus ?thesis by blast
qed

lemma stack_suc_vwalk:
  assumes INV: "common_inv s x"
  and NE: "stack s \<noteq> []"
  and SUC: "x \<in> succs (hd (stack s))"
  shows "vwalk (rev (x#stack s)) \<G>"
proof -
  from NE INV[unfolded common_inv_def] have "vwalk (rev (stack s)) \<G>" by simp
  moreover from SUC stack_subset_verts[OF INV] hd_in_set NE have "hd (stack s) \<in> V" by auto
  with SUC succs_implies_edge  have "(hd (stack s), x) \<in> E" by simp
  hence "vwalk [hd (stack s),x] \<G>" unfolding vwalk_def by (simp add: edge_in_nodes)
  moreover from NE have "joinable (rev (stack s)) [hd (stack s), x]" by (simp add: last_rev)
  ultimately have "vwalk (rev (stack s) \<oplus> [hd (stack s), x]) \<G>" by (rule vwalk_join_vwalk)
  thus ?thesis unfolding vwalk_join_def by simp
qed

lemma common_inv_change:
  "stack s \<noteq> [] \<Longrightarrow> y \<in> V \<Longrightarrow> common_inv s x \<Longrightarrow> common_inv s y"
unfolding common_inv_def by auto

lemma pre_common:
  "pre_dfs s x \<Longrightarrow> common_inv s x"
by (simp add: pre_dfs_def)

lemma post_common:
  "post_dfs s s' x \<Longrightarrow> common_inv s' x"
by (simp add: post_dfs_def)

lemma post_common2:
  "post_dfs s s' x \<Longrightarrow> common_inv s x"
by (simp add: post_dfs_def pre_common)

lemma fe_inv_common:
  "fe_inv_dfs s t it s' \<Longrightarrow> common_inv s' t"
by (simp add: fe_inv_dfs_def)

lemmas pre_commonE[elim] = pre_common[elim_format]
lemmas post_commonE[elim] = post_common[elim_format]
lemmas fe_inv_commonE[elim] = fe_inv_common[elim_format]

(* other rules *)
lemma fe_inv_start_eq[intro]:
  "fe_inv_dfs s x it s' \<Longrightarrow> start s' = start s"
unfolding fe_inv_dfs_unfold by simp

lemma post_start_eq[intro]:
  "post_dfs s s' x \<Longrightarrow> start s' = start s"
unfolding post_dfs_unfold by simp

lemma fe_inv_stack_eq[intro]:
  "fe_inv_dfs s x it s' \<Longrightarrow> dfs_cond dfs (state s') \<Longrightarrow> stack s' = stack s"
unfolding fe_inv_dfs_unfold by simp

lemma fe_inv_stack:
  assumes "fe_inv_dfs (disc_step_upd x s S) x it s'" (is "fe_inv_dfs ?sd x it s'")
  shows "x \<in> set (stack s')"
proof -
  from assms[unfolded fe_inv_dfs_unfold] have "\<exists>xs. stack s' = xs @ stack ?sd" by simp
  hence "set (stack ?sd) \<subseteq> set (stack s')" by auto
  thus ?thesis by (auto simp: disc_step_upd_def)
qed

lemma fe_inv_hd_stack:
  assumes "fe_inv_dfs (disc_step_upd x s S) x it s'" (is "fe_inv_dfs ?sd x it s'")
  and "dfs_cond dfs (state s')"
  shows "x = hd (stack s')"
proof -
  from assms have "stack s' = stack ?sd" by blast
  thus ?thesis by (simp add: disc_step_upd_def)
qed

lemma fe_inv_discovered:
  assumes "fe_inv_dfs (disc_step_upd x s S) x it s'" (is "fe_inv_dfs ?sd x it s'")
  shows "x \<in> discovered s'"
proof -
  from assms[unfolded fe_inv_dfs_unfold] have "discovered ?sd \<subseteq> discovered s'" by simp
  thus ?thesis by (auto simp: disc_step_upd_def)
qed

lemma post_stack_eq[intro]:
  "post_dfs s s' x \<Longrightarrow> dfs_cond dfs (state s') \<Longrightarrow> stack s' = stack s"
unfolding post_dfs_unfold by simp

lemma post_cond_stack_empty:
  "post_dfs (dfs_alg_start x) s x \<Longrightarrow> dfs_cond dfs (state s) \<Longrightarrow> stack s = []"
using post_stack_eq
by (simp add: dfs_alg_start_def)

lemma fe_inv_state_preserve:
  "fe_inv_dfs s x it s' \<Longrightarrow> \<not> dfs_cond dfs (state s) \<Longrightarrow> \<not> dfs_cond dfs (state s')"
unfolding fe_inv_dfs_unfold
by metis

lemma fe_inv_state_backtrack:
  "fe_inv_dfs s x it s' \<Longrightarrow> dfs_cond dfs (state s') \<Longrightarrow> dfs_cond dfs (state s)"
by (metis fe_inv_state_preserve)

lemma post_state_preserve:
  "post_dfs s s' x \<Longrightarrow> \<not> dfs_cond dfs (state s) \<Longrightarrow> \<not> dfs_cond dfs (state s')"
unfolding post_dfs_unfold
by metis

lemma post_state_backtrack:
  "post_dfs s s' x \<Longrightarrow> dfs_cond dfs (state s') \<Longrightarrow> dfs_cond dfs (state s)"
by (metis post_state_preserve)

lemma pre_finished_reach_finished:
  "pre_dfs s x \<Longrightarrow> y \<in> finished s \<Longrightarrow> y \<rightarrow>\<setminus>set (stack s)+ z \<Longrightarrow> z \<in> finished s"
by (simp add: pre_dfs_unfold)

lemma pre_start_reach:
  assumes PRE:"pre_dfs s x"
  and "stack s \<noteq> []"
  shows "start s \<rightarrow>+ x"
proof -
  from assms[unfolded pre_dfs_unfold] have "x \<in> succs (hd (stack s))" "hd (stack s) \<in> V" by auto
  moreover 
  {
    from PRE stack_subset_verts have "set (stack s) \<subseteq> V" by blast
    moreover with assms have "common_inv s (hd (stack s))" 
      apply (elim pre_commonE) 
      apply (rule common_inv_change) 
      by auto
    moreover from assms hd_in_set have "hd (stack s) \<in> set (stack s)" by simp
    ultimately have "start s \<rightarrow>\<star> hd(stack s)" using start_reachable_stack by simp
  }
  ultimately show ?thesis by (metis reachable_reachable1_trans succs_reachable)
qed

lemma post_reach_finished:
  assumes "post_dfs s s' x"
  and "x \<notin> discovered s"
  and "dfs_cond dfs (state s')"
  and "x \<rightarrow>\<setminus>set (stack s)+ v"
  shows "v \<in> finished s'"
proof -
  from assms have "x \<in> V" by auto
  from assms restricted_start have "x \<notin> set (stack s)" by auto
  from assms[unfolded post_dfs_unfold] have "finished s' = finished s \<union> {z. \<exists>y\<in>succs x - set (stack s). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - set (stack s)) \<union> {x}" by simp
  with FIXME_A `x \<notin> set (stack s)` `x \<in> V` have "finished s' = finished s' \<union> {z. x \<rightarrow>\<setminus>set (stack s)+ z} \<union> {x}" by auto
  with assms show ?thesis by auto
qed

lemma post_reach_finished_empty_stack:
  assumes "post_dfs s s' x"
  and "stack s' = []"
  and "dfs_cond dfs (state s')"
  and "x \<rightarrow>+ v"
  shows "v \<in> finished s'"
proof-
  note A = assms[unfolded post_dfs_unfold pre_dfs_unfold]
  
  from A have "stack s = []" by simp
  with A have "x \<notin> discovered s" and "x \<rightarrow>\<setminus>set (stack s)+ v" by simp_all
  with assms post_reach_finished show ?thesis by simp
qed

lemma post_empty_stack_finished_eq:
  assumes "post_dfs s s' x"
  and "stack s' = []"
  and "dfs_cond dfs (state s')"
  shows "finished s' = {z. x = z \<or> x \<rightarrow>+ z}"
proof -
  note A = assms[unfolded post_dfs_unfold]
  
  from A have "stack s = []" by simp
  moreover with assms(1)[unfolded post_dfs_def pre_dfs_unfold] have "x \<notin> discovered s" and "finished s = {}" by auto
  with A FIXME_A[of x s] have "finished s' = insert x ({z. x\<rightarrow>\<setminus>set(stack s)+z})" by simp force
  ultimately show ?thesis by auto
qed

lemma CI_finished_implies_succs_discovered:
  assumes CI: "common_inv s x"
  and F: "\<forall>y \<in> finished s. (\<forall>z. y \<rightarrow>\<setminus>set (stack s)+ z \<longrightarrow> z \<in> finished s)"
  and xF: "x \<in> finished s"
  shows "succs x \<subseteq> discovered s"
proof
  fix y
  assume A: "y \<in> succs x"
  thus "y \<in> discovered s"
  proof (cases "x \<rightarrow>\<setminus>set (stack s)+ y")
    case True with F xF have "y \<in> finished s" by blast
    with finished_implies_discovered[OF CI] show ?thesis .
  next
    case False note not_reach = this
    show ?thesis
    proof (cases "y \<in> set (stack s)")
      case True with stack_implies_discovered[OF CI] show ?thesis .
    next
      case False from xF finished_subset_verts CI have "x \<in> V" by auto
      from xF finished_implies_not_stack[OF CI] have "x \<notin> set (stack s)" by blast
      with False A succs_restricted `x \<in> V` have "x \<rightarrow>\<setminus>set (stack s)+ y" by blast
      with not_reach show ?thesis by contradiction
    qed
  qed
qed

lemma post_finished_implies_succs_discovered:
  assumes "post_dfs s s' x"
  and "dfs_cond dfs (state s')"
  and "x \<in> finished s'"
  shows "succs x \<subseteq> discovered s'"
using assms[unfolded post_dfs_def post_dfs'_def] CI_finished_implies_succs_discovered
by auto

lemma fe_inv_finished_implies_succs_discovered:
  assumes "fe_inv_dfs s t it s'"
  and "dfs_cond dfs (state s')"
  and "t \<in> finished s'"
  shows "succs t \<subseteq> discovered s'"
using assms[unfolded fe_inv_dfs_def fe_inv_dfs'_def] CI_finished_implies_succs_discovered
by auto

lemma pre_finished_implies_succs_discovered:
  assumes "pre_dfs s x"
  and "x \<in> finished s"
  shows "succs x \<subseteq> discovered s"
using assms[unfolded pre_dfs_def pre_dfs'_def] CI_finished_implies_succs_discovered
by auto
end

(* I have to split up the DFS-locale into two parts, else I trigger a PG?-Bug that results in parsing abortion. As it seems, a locale definition (including comments) may only have a certain length. *)

locale DFS_A = finite_succs_digraph succs v\<^sub>0 +
               DFS_def succs v\<^sub>0 dfs
  for dfs :: "('S,'n) dfs_algorithm"
  and succs and v\<^sub>0 :: "'n" +
  assumes visit_NF: "\<lbrakk>x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s)\<rbrakk> \<Longrightarrow> nofail (dfs_visit dfs s x)"
  and     remove_NF: "\<lbrakk>x \<in> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s)\<rbrakk> \<Longrightarrow> nofail (dfs_remove dfs s x)"
  and     post_NF: "\<lbrakk>x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s); inres (dfs_visit dfs s x) S;
                      fe_inv_dfs (disc_step_upd x s S) x {} s'; dfs_fe_inv dfs (disc_step_upd x s S) x {} s'; dfs_cond dfs (state s')\<rbrakk> 
                     \<Longrightarrow> nofail (dfs_post dfs s' x)"
  and start_pre: "x \<in> V \<Longrightarrow> pre_dfs (dfs_alg_start x) x \<Longrightarrow> dfs_pre_cond dfs (dfs_alg_start x) x"
  and start_abort_post: "\<lbrakk>x \<in> V; \<not> dfs_cond dfs (dfs_start dfs x); post_dfs (dfs_alg_start x) (dfs_alg_start x) x\<rbrakk>
                        \<Longrightarrow> dfs_post_cond dfs (dfs_alg_start x) (dfs_alg_start x) x"
  and rem_post: "\<lbrakk>x \<in> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s); inres (dfs_remove dfs s x) S;
                  post_dfs s (s\<lparr>state:=S\<rparr>) x\<rbrakk>  
                 \<Longrightarrow> dfs_post_cond dfs s (s\<lparr>state:=S\<rparr>) x"
  and fin_post: "\<lbrakk>x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s); inres (dfs_visit dfs s x) S;
                  fe_inv_dfs (disc_step_upd x s S) x {} s'; dfs_fe_inv dfs (disc_step_upd x s S) x {} s'; 
                  inres (dfs_post dfs s' x) S'; dfs_cond dfs (state s'); post_dfs s (fin_step_upd x s' S') x\<rbrakk> 
                 \<Longrightarrow> dfs_post_cond dfs s (fin_step_upd x s' S') x"  
  and fin_post_abort: "\<lbrakk>x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s); inres (dfs_visit dfs s x) S; 
                        it \<subseteq> succs x; fe_inv_dfs (disc_step_upd x s S) x it s'; dfs_fe_inv dfs (disc_step_upd x s S) x it s';
                        \<not> dfs_cond dfs (state s'); post_dfs s s' x\<rbrakk>
                       \<Longrightarrow> dfs_post_cond dfs s s' x"

locale DFS = DFS_A dfs succs v\<^sub>0
  for dfs :: "('S,'n) dfs_algorithm" and succs v\<^sub>0 +
  assumes fe_inv_start: "\<lbrakk>x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s); inres (dfs_visit dfs s x) S;
                          fe_inv_dfs (disc_step_upd x s S) x (succs x) (disc_step_upd x s S)\<rbrakk>
                         \<Longrightarrow> dfs_fe_inv dfs (disc_step_upd x s S) x (succs x) (disc_step_upd x s S)"
  and pre_step_pre: "\<lbrakk>x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s); inres (dfs_visit dfs s x) S; 
                      it \<subseteq> succs x; t \<in> it; dfs_cond dfs (state s'); fe_inv_dfs (disc_step_upd x s S) x it s';
                      dfs_fe_inv dfs (disc_step_upd x s S) x it s'; pre_dfs s' t\<rbrakk>
                     \<Longrightarrow> dfs_pre_cond dfs s' t"
  and fe_inv_step: "\<lbrakk>x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s); inres (dfs_visit dfs s x) S; 
                     it \<subseteq> succs x; t \<in> it; dfs_cond dfs (state s'); post_dfs s' s'' t; dfs_post_cond dfs s' s'' t;
                     fe_inv_dfs (disc_step_upd x s S) x it s'; dfs_fe_inv dfs (disc_step_upd x s S) x it s'; fe_inv_dfs (disc_step_upd x s S) x (it - {t}) s''\<rbrakk> 
                    \<Longrightarrow> dfs_fe_inv dfs (disc_step_upd x s S) x (it - {t}) s''"

begin

(* these require finite V *)
theorem wf_sws_rel:
  "wf sws_rel"
unfolding sws_rel_def
by (auto simp: sws_rel_def)

lemma discovered_finite:
  "common_inv s x \<Longrightarrow> finite (discovered s)"
by (metis finite_V finite_subset discovered_subset_verts) 

lemma finished_finite:
  "common_inv s x \<Longrightarrow> finite (finished s)"
by (metis finite_V finite_subset finished_subset_verts)

lemma start_pre_dfs:
  "x \<in> V \<Longrightarrow> pre_dfs (dfs_alg_start x) x"
using start_pre
unfolding pre_dfs_unfold (*timing_inv_def*) dfs_alg_start_def
by (simp add: finished_def)

lemma pre_disc_step_fe:
  assumes ND: "x \<notin> discovered s"
  and PRE: "pre_dfs s x"
  and INR: "inres (dfs_visit dfs s x) S"
  and COND: "dfs_cond dfs (state s)"
  shows "fe_inv_dfs (disc_step_upd x s S) x (succs x) (disc_step_upd x s S)"
unfolding fe_inv_dfs_def
proof (intro conjI)
  let ?sd = "disc_step_upd x s S"
  note disc_step_upd_def[simp]

  from assms[unfolded pre_dfs_unfold] have xF: "x \<notin> finished s" and "x \<notin> set(stack s)" by blast+
  (*with assms disc_step_timing_inv show "timing_inv ?sd" by (simp add: pre_dfs_def)*)

  {
    fix y z
    assume "dfs_cond dfs S" and yF: "y \<in> finished s" and "y \<rightarrow>\<setminus>insert x (set (stack s))+ z"
    hence "y \<rightarrow>\<setminus>set (stack s)+ z" by (metis restr_reachable_restrict subset_insertI) 
    with yF PRE pre_finished_reach_finished have "z \<in> finished s" by metis
  } note fin_reach = this

  from assms show "fe_inv_dfs' ?sd x (succs x) ?sd"
    unfolding fe_inv_dfs'_def
  proof (intro conjI)
    case goal6 hence "finished ?sd = finished s" by (simp del: disc_step_upd_def)
    with goal6 show ?case by simp (metis fin_reach)
  qed simp_all

  {
    fix  n
    assume A: "n < length (stack s)"
    have "stack ?sd ! n \<in> succs (stack ?sd ! Suc n)"
    proof (cases "stack s = []")
      case True with A have "False" by simp
      thus ?thesis ..
    next
      case False
      from assms have "\<And> n. n < length (stack s) - 1 \<Longrightarrow> stack s ! n \<in> succs (stack s ! Suc n)"
        unfolding pre_dfs_unfold by simp
      with A have "n > 0 \<Longrightarrow> (x#stack s)!n \<in> succs ((x#stack s) ! Suc n)" by (smt nth_Cons_Suc nth_Cons_pos)
      hence "n > 0 \<Longrightarrow> (stack ?sd)!n \<in> succs (stack ?sd ! Suc n)" by simp

      moreover have "stack ?sd ! 0 = x" by simp
      moreover from False have "stack ?sd ! 1 = hd (stack s)" by (simp add: hd_conv_nth)
      moreover from PRE ND not_discovered_implies_non_stack have "x \<notin> set (stack s)" by blast
      hence "x \<noteq> hd (stack s)" using hd_in_set False by blast
      with PRE False COND have "x \<in> succs (hd (stack s))" unfolding pre_dfs_unfold by simp

      ultimately show ?thesis by fastforce
    qed
  }
  moreover
  have "vwalk (rev (x#stack s)) \<G>"
  proof (cases "stack s = []")
    case True thus ?thesis using PRE unfolding pre_dfs_unfold vwalk_def by simp
  next
    case False with PRE[unfolded pre_dfs_unfold] have "x \<in> succs (hd (stack s))" by simp
    with False PRE[unfolded pre_dfs_def] stack_suc_vwalk show ?thesis by simp
  qed
  moreover note assms `x \<notin> set (stack s)`
  moreover from ND not_discovered_implies_not_finished PRE have "x \<notin> finished s" by blast
  moreover from ND have "finished ?sd = finished s" by (simp del: disc_step_upd_def)
  ultimately show "common_inv ?sd x" unfolding pre_dfs_unfold by simp
qed

lemma fe_fin_step_post_abort:
  assumes ND: "x \<notin> discovered s"
  and PRE: "pre_dfs s x"
  and COND: "dfs_cond dfs (state s)"
  and INR: "inres (dfs_visit dfs s x) S"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x it s'"
  and ABORT: "\<not> dfs_cond dfs (state s')"
  and "it \<subseteq> succs x"
  shows "post_dfs s s' x"
proof -
  note disc_step_upd_def[simp]
  note FE' = FE[unfolded fe_inv_dfs_unfold]

  from FE have "common_inv s' x" by blast

  moreover
  from FE' ND have "finished (disc_step_upd x s S) = finished s"
    by (metis disc_upd_finished)
  with FE' have "finished s \<subseteq> finished s'" by metis
    
  moreover
  from FE' have 
    "discovered s \<subseteq> discovered s'"
    "start s = start s'"
    by auto
  
  moreover
  from ABORT have "\<not> dfs_cond dfs (state s')" by simp

  moreover
  from FE' have "\<exists>xs. stack s' = xs @ stack (disc_step_upd x s S)" by simp
  hence "\<exists>xs. stack s' = xs @ stack s" by auto

  ultimately show ?thesis unfolding post_dfs_unfold
    by (simp add: PRE FE[unfolded fe_inv_dfs_def])
qed

lemma fe_fin_step_post:
  assumes ND: "x \<notin> discovered s"
  and PRE: "pre_dfs s x"
  and COND: "dfs_cond dfs (state s)"
  and INR: "inres (dfs_visit dfs s x) S"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x {} s'"
  and COND': "dfs_cond dfs (state s')"
  and INR': "inres (dfs_post dfs s' x) S'"
  shows "post_dfs s (fin_step_upd x s' S') x"
proof -
  let ?sd = "disc_step_upd x s S"
  let ?sf = "fin_step_upd x s' S'"
  
  note disc_step_upd_def[simp] fin_step_upd_def[simp]
  note FE' = FE[unfolded fe_inv_dfs_unfold]
  note PRE' = PRE[unfolded pre_dfs_unfold]

  from FE' COND' have "discovered s' \<subseteq> V" and xD: "x \<in> discovered ?sf" by simp_all
  with PRE' have "discovered ?sf \<subseteq> V" by simp

  moreover
  from FE COND' have s'_eq_d: "stack s' = stack ?sd" by blast
  hence f_eq_old: "stack ?sf = stack s" by simp

  moreover
  from PRE ND have not_X: "x \<notin> set (stack s)" using stack_subset_discovered by blast

  from FE' have "finished s' \<union> set (stack s') = discovered s'" "set (stack s') \<inter> finished s' = {}" by simp_all
  with s'_eq_d have "finished ?sf \<union> set (stack ?sf) = discovered ?sf" "set (stack ?sf) \<inter> finished ?sf = {}" using tl_subset xD not_X by (auto simp: finished_def)

  moreover
  from FE fe_inv_start_eq have start: "start s = start ?sf" by auto
  
  moreover from PRE have "common_inv s x" by blast
  ultimately have com: "common_inv ?sf x"
    unfolding common_inv_def
    by simp
  hence "x \<in> V" by blast

  from FE' ND have "finished s \<subseteq> finished s'" by (simp del: disc_step_upd_def)
  with f_eq_old have fsub: "finished s \<subseteq> finished ?sf" unfolding finished_def by auto

  from FE' have dsub: "discovered s \<subseteq> discovered ?sf" by simp
  
  (*from FE' xD have timing: "timing_inv ?sf" using fin_step_timing_inv by simp*)

  show ?thesis
  proof (cases "dfs_cond dfs (state ?sf)")
    case False thus ?thesis
      using PRE fsub dsub start f_eq_old (*timing*) com by (simp add: post_dfs_unfold)
  next
    case True

    note xD fsub dsub start f_eq_old (*timing*) com
    moreover {
      have "discovered ?sf = discovered s'" by simp
      also from FE' COND' have "... = discovered s \<union> {z. \<exists>y\<in>succs x - set (stack ?sd).  y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - set (stack ?sd)) \<union> {x}" by auto
      also have " ... = discovered s \<union> {z. \<exists>y\<in>succs x - insert x (set (stack s)).  y \<rightarrow>\<setminus>insert x (set (stack s))+ z} \<union> (succs x - insert x (set (stack s))) \<union> {x}" by simp
      also from FIXME_Y[OF `x \<in> V`, of "set (stack s)"] have " ... = discovered s \<union> {z. \<exists>y\<in>succs x - set (stack s).  y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - set (stack s)) \<union> {x}" by blast
      finally have "... = discovered ?sf" ..
    }
    moreover

    with not_X f_eq_old restricted_target have fin_eqn:
      "finished s \<union> {z. \<exists>y\<in>succs x - set (stack s). y \<rightarrow>\<setminus>set (stack s)+ z} \<union> (succs x - set (stack s)) \<union> {x} = finished (fin_step_upd x s' S')"
      unfolding finished_def
      by auto

    moreover have "\<forall>y \<in> finished ?sf. (\<forall>z. y \<rightarrow>\<setminus>set (stack s)+ z \<longrightarrow> z \<in> finished ?sf)"
    proof (rule ballI, rule allI, rule impI)
      fix y z
      assume yF: "y \<in> finished ?sf" and y_z: "y \<rightarrow>\<setminus>set (stack s)+ z"
      
      from FIXME_A[OF not_X] fin_eqn[symmetric] `x \<in> V` have F: "finished ?sf = finished s \<union> {z. x \<rightarrow>\<setminus>set (stack s)+ z} \<union> {x}" by auto
      
      show "z \<in> finished ?sf"
      proof (cases " x \<rightarrow>\<setminus>set (stack s)+ y")
        case True with y_z restr_reachable1_trans have "x \<rightarrow>\<setminus>set (stack s)+ z" by blast
        with F show ?thesis by auto
      next
        case False
        show ?thesis
        proof (cases "y \<noteq> x")
          case False with y_z F show ?thesis by auto
        next
          case True with False yF F have "y \<in> finished s" by auto
          with y_z have "z \<in> finished s" using PRE' by blast
          thus ?thesis using F by simp
        qed
      qed
    qed

    ultimately show ?thesis using True COND ND
      unfolding post_dfs_unfold by (simp add: PRE)
  qed
qed

lemma pre_rem_step_post:
  assumes "pre_dfs s x"
  and "x \<in> discovered s"
  and "inres (dfs_remove dfs s x) S"
  and "dfs_cond dfs (state s)"
  shows "post_dfs s (s\<lparr>state := S\<rparr>) x"
proof -
  have "finished ((s\<lparr>state := S\<rparr>)) = finished s"
    unfolding finished_def
    by simp
  
  with discovered_subset_verts finished_subset_verts assms(1)
  show ?thesis
    using assms assms[unfolded pre_dfs_unfold (*timing_inv_def*)]
    by (simp add: post_dfs_def[unfolded post_dfs'_def (*timing_inv_def*) common_inv_def])
qed

lemma in_sws_rel:
  assumes "pre_dfs s x"
  and "x \<notin> discovered s"
  and "fe_inv_dfs (disc_step_upd x s S) x it s'"
  shows "(s',s) \<in> sws_rel"
proof -
  from assms have *: "discovered s' \<subseteq> V" by (simp add: fe_inv_dfs_unfold)

  from assms have "discovered s \<subset> discovered (disc_step_upd x s S)" by (auto simp add: disc_step_upd_def)
  also from assms have " ... \<subseteq> discovered s'"  by (auto simp add: fe_inv_dfs_unfold)
  finally have "discovered s \<subset> discovered s'" .
  thus ?thesis by (simp add: finite_psupset_def sws_rel_def *)
qed

lemma pre_inv_step_pre:
  assumes PRE: "pre_dfs s x"
  and X: "x \<notin> discovered s"
  and ITS: "it \<subseteq> succs x"
  and T: "t \<in> it"
  and INR: "inres (dfs_visit dfs s x) S"
  and COND: "dfs_cond dfs (state s)"
  and COND': "dfs_cond dfs (state s')"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x it s'"
  shows "pre_dfs s' t"
unfolding pre_dfs_def
proof (intro conjI)
  let ?sd = "disc_step_upd x s S"

  note PRE' = PRE[unfolded pre_dfs_unfold]
  note FE'= FE[unfolded fe_inv_dfs_unfold]
  note disc_step_upd_def[simp]
  
(*  from FE' show "timing_inv s'" by simp_all*)

  from FE COND' have s'_eq_d: "stack s' = stack ?sd" by blast
  with FE' show "common_inv s' t"
    unfolding common_inv_def
  proof (elim conjE, intro conjI)
    case goal1
    from  T ITS succs_in_V PRE' show "t \<in> V" by auto
  next
    case goal5 show ?case using s'_eq_d by simp
  qed assumption+

  show "pre_dfs' s' t"
    unfolding pre_dfs'_def
  proof (intro conjI)
    from FE' COND' have "x = hd (stack s')" by simp
    with T ITS show "stack s' \<noteq> [] \<longrightarrow> t \<in> succs (hd (stack s'))" by auto

    from FE have "set (stack s') \<subseteq> discovered s'" using stack_subset_discovered by blast
    thus "stack s' = [] \<longleftrightarrow> discovered s' = {}" using s'_eq_d by auto
    
    show " \<forall>y\<in>finished s'. \<forall>z. y \<rightarrow>\<setminus>set (stack s')+ z \<longrightarrow> z \<in> finished s'"
    proof (rule ballI, rule allI, rule impI)
      fix y z
      assume yF: "y \<in> finished s'" and "y \<rightarrow>\<setminus>set (stack s')+ z"

      with `stack s' = stack ?sd` have "y \<rightarrow>\<setminus>set (stack ?sd)+ z" by simp
      with FE' COND' yF show "z \<in> finished s'" by blast
    qed
  qed
qed

lemma fe_inv_dfs_step:
  assumes PRE: "pre_dfs s x"
  and POST: "post_dfs s' s'' t"
  and T: "it \<subseteq> succs x" "t \<in> it"
  and X: "x \<notin> discovered s"
  and COND: "dfs_cond dfs (state s)"
  and COND': "dfs_cond dfs (state s')"
  and INR: "inres (dfs_visit dfs s x) S"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x it s'"
  shows "fe_inv_dfs (disc_step_upd x s S) x (it - {t}) s''"
unfolding fe_inv_dfs_def
proof (intro conjI)
  let ?sd = "disc_step_upd x s S"

  note POST' = POST[unfolded post_dfs_unfold]
  note FE'= FE[unfolded fe_inv_dfs_unfold]
  note disc_step_upd_def[simp]

(*  from POST' show "timing_inv s''" by simp*)

  have "stack s'' \<noteq> []"
  proof (rule notI)
    assume "stack s'' = []"
    with POST' have "stack s' = []" by simp
    with FE' show False by simp
  qed
  moreover from PRE have "x \<in> V" by blast
  moreover have "common_inv s'' t" using POST by blast
  ultimately show "common_inv s'' x" by (rule common_inv_change)

  from FE' have x_disc: "x \<in> discovered s'" by auto

  have "start ?sd = start s''" by (metis fe_inv_start_eq post_start_eq FE POST) 

  moreover
  from FE COND' have "stack s' = stack ?sd" by blast
  with POST have "dfs_cond dfs (state s'') \<Longrightarrow> stack s'' = stack ?sd" by (metis post_stack_eq)

  moreover
  from COND' POST' have "dfs_cond dfs (state s')" by simp
  with FE' have "dfs_cond dfs (state ?sd)" by simp

  moreover
  {
    from FE' have "\<exists>xs. stack s' = xs@stack ?sd" by simp
    moreover from POST' have "\<exists>xs. stack s'' = xs @ stack s'" by simp
    ultimately have "\<exists>xs. stack s'' = xs@stack ?sd" by force
  }

  moreover { 
    have "discovered ?sd \<subseteq> discovered s'" using FE' by simp
    also have "... \<subseteq> discovered s''" by (simp add: POST')
    finally have "discovered ?sd \<subseteq> ..." by blast 
  }
  
  moreover {
    have "finished ?sd \<subseteq> finished s'" using FE' by (simp del: disc_step_upd_def)
    also have "... \<subseteq> finished s''" using POST' by simp
    finally have "finished ?sd \<subseteq> ..." by blast
  }

  moreover
  have "dfs_cond dfs (state s'') \<Longrightarrow> discovered ?sd \<union> {z. \<exists>y\<in>succs x - (it - {t}) - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - (it - {t}) - set (stack ?sd)) = discovered s''" (is "?C \<Longrightarrow> ?F")
  proof -
    assume C: "dfs_cond dfs (state s'')"

    {
      assume nis: "t \<notin> set (stack ?sd)"
      and "discovered s'' = discovered s' \<union> {z. t  \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}"
      
      note this(2)
      also from COND' FE' have " ... = discovered ?sd \<union> {z. \<exists>y\<in>succs x - it - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - it - set (stack ?sd)) \<union> {z. t  \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}" by auto
      also have "... = discovered ?sd \<union> {z. \<exists>y\<in>succs x - (it - {t}) - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - (it - {t}) - set (stack ?sd))" using FIXME_Z[OF T nis] by auto
      finally have "... = discovered s''" ..
    } note no_stack = this

    show ?F
    proof (cases "t \<in> discovered s'")
      case False
      from FE' have "discovered s \<subseteq> discovered s'" by simp
      with False x_disc have "t \<notin> discovered s" "x \<noteq> t" by auto
      with PRE[unfolded pre_dfs_unfold] have nis: "t \<notin> set (stack ?sd)" by auto
      from False C POST' `stack s' = stack ?sd` have "discovered s'' = discovered s' \<union> {z. \<exists>y\<in>succs t - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs t - set (stack ?sd)) \<union> {t}" by simp
      with FIXME_A[OF nis] POST have "discovered s'' = discovered s' \<union> {z. t  \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}" by auto
      with no_stack[OF nis this] show ?thesis by blast
    next
      case True note disc = this
      from pre_inv_step_pre[OF PRE X T INR COND COND' FE] have s'PRE[unfolded pre_dfs_unfold]: "pre_dfs s' t" .
      thus ?thesis
      proof (cases "t \<in> finished s'")
        case True with s'PRE `stack s' = stack ?sd` have nis: "t \<notin> set (stack ?sd)" by auto
        
        from True s'PRE have "{z. t \<rightarrow>\<setminus>set (stack s')+ z} \<subseteq> finished s'" by auto
        with s'PRE `stack s' = stack ?sd` have "{z. t \<rightarrow>\<setminus>set (stack ?sd)+ z} \<subseteq> discovered s'" by auto
        with POST' C disc have "discovered s'' = discovered s' \<union> {z. t \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}" by auto
        with no_stack[OF nis this] show ?thesis by blast
      next
        case False with disc s'PRE `stack s' = stack ?sd` have "t \<in> set (stack ?sd)"by auto
        from C disc POST' have "discovered s'' = discovered s'" by auto
        also from FE' COND' have "... =  discovered ?sd \<union> {z. \<exists>y\<in>succs x - it - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - it - set (stack ?sd))" by auto
        also from `t \<in> set (stack ?sd)` have "... = discovered ?sd \<union> {z. \<exists>y\<in>succs x - (it - {t}) - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - (it - {t}) - set (stack ?sd))" by auto
        finally show ?thesis ..
      qed
    qed
  qed

  moreover
  have "dfs_cond dfs (state s'') \<Longrightarrow> finished ?sd \<union> {z. \<exists>y\<in>succs x - (it - {t}) - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - (it - {t}) - set (stack ?sd)) = finished s''" (is "?C \<Longrightarrow> ?F")
  proof -
    assume C: "dfs_cond dfs (state s'')"

    {
      assume nis: "t \<notin> set (stack ?sd)"
      and "finished s'' = finished s' \<union> {z. t \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}"
      
      note this(2)
      also from COND' FE' have " ... = finished ?sd \<union> {z. \<exists>y\<in>succs x - it - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - it - set (stack ?sd)) \<union> {z. t  \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}" by auto
      also have "... = finished ?sd \<union> {z. \<exists>y\<in>succs x - (it - {t}) - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - (it - {t}) - set (stack ?sd))" using FIXME_Z[OF T nis] by auto
      finally have "... = finished s''" ..
    } note no_stack = this

    show ?F
    proof (cases "t \<in> discovered s'")
      case False
      from FE' have "discovered s \<subseteq> discovered s'" by simp
      with False x_disc have "t \<notin> discovered s" "x \<noteq> t" by auto
      with PRE[unfolded pre_dfs_unfold] have nis: "t \<notin> set (stack ?sd)" by auto
      from False C POST' `stack s' = stack ?sd` have "finished s'' = finished s' \<union> {z. \<exists>y\<in>succs t - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs t - set (stack ?sd)) \<union> {t}" by simp
      with FIXME_A[OF nis] POST have "finished s'' = finished s' \<union> {z. t  \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}" by auto
      with no_stack[OF nis this] show ?thesis by blast
    next
      case True note disc = this
      from pre_inv_step_pre[OF PRE X T INR COND COND' FE] have s'PRE[unfolded pre_dfs_unfold]: "pre_dfs s' t" .
      thus ?thesis
      proof (cases "t \<in> finished s'")
        case True with s'PRE `stack s' = stack ?sd` have nis: "t \<notin> set (stack ?sd)" by auto
        
        from True s'PRE have "{z. t \<rightarrow>\<setminus>set (stack s')+ z} \<subseteq> finished s'" by auto
        with POST' C disc True `stack s' = stack ?sd` have "finished s'' = finished s' \<union> {z. t \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}" by auto
        with no_stack[OF nis this] show ?thesis by blast
      next
        case False with disc s'PRE `stack s' = stack ?sd` have "t \<in> set (stack ?sd)"by auto
        from C disc POST' have "finished s'' = finished s'" by simp
        also from FE' COND' have "... =  finished ?sd \<union> {z. \<exists>y\<in>succs x - it - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - it - set (stack ?sd))" by auto
        also from `t \<in> set (stack ?sd)` have "... = finished ?sd \<union> {z. \<exists>y\<in>succs x - (it - {t}) - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs x - (it - {t}) - set (stack ?sd))" by auto
        finally show ?thesis ..
      qed
    qed
  qed

  moreover
  have "dfs_cond dfs (state s'') \<Longrightarrow> \<forall>y \<in> finished s''. (\<forall> z. y \<rightarrow>\<setminus>set (stack ?sd)+ z \<longrightarrow> z \<in> finished s'')"
  proof (rule ballI, rule allI, rule impI)
    fix y z
    assume C: "dfs_cond dfs (state s'')" and *: "y \<in> finished s''" "y \<rightarrow>\<setminus>set (stack ?sd)+ z"
    show "z \<in> finished s''"
    proof (cases "t \<in> discovered s'")
      case True with C POST' have "finished s'' = finished s'" by simp
      with FE' COND' *  show ?thesis by simp
    next
      case False 
      from FE' have "discovered s \<subseteq> discovered s'" by simp
      with False x_disc have "t \<notin> discovered s" "x \<noteq> t" by auto
      with PRE[unfolded pre_dfs_unfold] have nis: "t \<notin> set (stack ?sd)" by auto
      
      from False C POST' `stack s' = stack ?sd` have "finished s'' = finished s' \<union> {z. \<exists>y\<in>succs t - set (stack ?sd). y \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> (succs t - set (stack ?sd)) \<union> {t}" by simp
      with FIXME_A[OF nis] POST have F: "finished s'' = finished s' \<union> {z. t  \<rightarrow>\<setminus>set (stack ?sd)+ z} \<union> {t}" by auto
      
      show ?thesis
      proof (cases " t \<rightarrow>\<setminus>set (stack ?sd)+ y")
        case True with * restr_reachable1_trans have "t \<rightarrow>\<setminus>set (stack ?sd)+ z" by blast
        with F show ?thesis by auto
      next
        case False
        show ?thesis
        proof (cases "y \<noteq> t")
          case False with * F show ?thesis by auto
        next
          case True with False * F have "y \<in> finished s'" by auto
          with * have "z \<in> finished s'" using FE' COND' by blast
          thus ?thesis using F by simp
        qed
      qed
    qed
  qed

  ultimately show "fe_inv_dfs' ?sd x (it - {t}) s''" by (simp add: fe_inv_dfs'_def)
qed

theorem body_is_post:
  assumes "pre_dfs s x"
  and "dfs_cond dfs (state s)"
  and "dfs_pre_cond dfs s x"
  shows "dfs_algo_body s x \<le> SPEC (\<lambda>s'. post_dfs s s' x \<and> dfs_post_cond dfs s s' x)"
unfolding dfs_algo_body_def
  apply (rule RECT_rule_arb2[where \<Phi>="\<lambda>s' x (s, x'). x = x' \<and> s' = s \<and> pre_dfs s x \<and> dfs_pre_cond dfs s x\<and> dfs_cond dfs (state s)" and V="sws_rel <*lex*> {}" and arb="s" and arc="x"])
  apply refine_mono
  apply (auto simp add: wf_sws_rel)[]
  apply (simp add: assms)
proof (intro refine_vcg)
  case goal1 thus ?case by blast
next
  case goal2 thus ?case
    apply (auto simp: dfs_rem_step_def)
    apply (intro refine_vcg)
    apply (rule pw_leI)
    apply (auto intro: remove_NF pre_rem_step_post rem_post)
    done
next
  case (goal3 D _ _ sx s x)
  thus ?case
    unfolding dfs_disc_step_def
    apply (intro refine_vcg)
    apply (rule pw_leI)
    apply (auto intro: visit_NF)
    apply (rename_tac S)
  proof (rule_tac I="\<lambda>it s'. fe_inv_dfs (disc_step_upd x s S) x it s'\<and> dfs_fe_inv dfs (disc_step_upd x s S) x it s'" in FOREACHc_rule)
    case goal1 thus ?case by (auto intro!: succs_finite)
  next
    case goal2 thus ?case by (auto intro: pre_disc_step_fe fe_inv_start)
  next
    case (goal3 S t it s')
    with in_sws_rel have "(s', s) \<in> sws_rel" by blast
    moreover from goal3 have "pre_dfs s' t" by (blast intro: pre_inv_step_pre)
    moreover with goal3 have "dfs_pre_cond dfs s' t" by (blast intro: pre_step_pre)
    ultimately have *: "D (s', t) \<le> SPEC (\<lambda>s''. post_dfs s' s'' t \<and> dfs_post_cond dfs s' s'' t)" 
      using goal3(1)[where a=s' and b=t] and `dfs_cond dfs (state s')` 
      by blast
    show ?case 
      apply (rule order_trans[OF *]) 
      apply (rule pw_leI)
      using goal3
      apply (auto intro: fe_inv_dfs_step fe_inv_step)
      done
  next
    case (goal4 S s') thus ?case
      unfolding dfs_fin_step_def
      apply (intro refine_vcg)
      apply blast
      apply (rule pw_leI)
      apply (auto intro: post_NF fe_fin_step_post fin_post)[]
      apply (blast intro: fe_fin_step_post_abort fin_post_abort)
      done
  next
    case (goal5 S it s') thus ?case by (auto simp: dfs_fin_step_def intro: fe_fin_step_post_abort fin_post_abort)
  qed
qed

lemma dfs_post:
  assumes "x\<^sub>0 \<in> V"
  shows "dfs_algo x\<^sub>0 \<le> SPEC (\<lambda>s. post_dfs (dfs_alg_start x\<^sub>0) s x\<^sub>0 \<and> dfs_post_cond dfs (dfs_alg_start x\<^sub>0) s x\<^sub>0)"
unfolding dfs_algo_def
using assms
  apply (intro refine_vcg)
     apply (insert start_pre_dfs[OF assms])
     apply (insert start_pre[OF assms])
     apply (rule context_conjI)
       apply (simp add: post_dfs_unfold dfs_alg_start_def (*timing_inv_def*) finished_def)
       apply (fastforce simp add: dfs_alg_start_def intro!: start_abort_post)
     apply (blast intro: start_pre_dfs order_trans[OF body_is_post])
done

lemma dfs_algo_NF:
  "x \<in> V \<Longrightarrow> nofail (dfs_algo x)"
by (metis (lifting) dfs_post SPEC_nofail)

lemma dfs_algo_SPEC:
  assumes "x \<in> V"
  and S: "\<And>s. \<lbrakk>post_dfs (dfs_alg_start x) s x; dfs_post_cond dfs (dfs_alg_start x) s x; inres (dfs_algo x) s\<rbrakk> \<Longrightarrow> \<Phi> s"
  shows "dfs_algo x \<le> SPEC \<Phi>"
apply (intro pw_leI)
apply (simp add: dfs_algo_NF[OF `x \<in> V`])
proof (intro impI allI)
  fix s
  assume inr: "inres (dfs_algo x) s"
  hence  "post_dfs (dfs_alg_start x) s x \<and> dfs_post_cond dfs (dfs_alg_start x) s x"
    by (rule inres_SPEC) (rule dfs_post[OF `x \<in> V`])
  with inr S show "\<Phi> s" by metis
qed
end

section {* Simple DFS *}
text {* Introduce a simple dfs algorithm, that ignores all post and pre functions and just returns the filled discover and finish sets. *}

definition simple_dfs :: "(unit, 'n) dfs_algorithm" where
  "simple_dfs = \<lparr> dfs_cond = \<lambda>_. True, 
                   dfs_visit = \<lambda>s _. RETURN (state s), 
                   dfs_post = \<lambda>s _. RETURN (state s), 
                   dfs_remove = \<lambda>s _. RETURN (state s), 
                   dfs_start = \<lambda>_. (), 
                   dfs_pre_cond = \<lambda>_ _. True,
                   dfs_post_cond = \<lambda>_ _ _. True,
                   dfs_fe_inv = \<lambda>_ _ _ _. True
                   \<rparr>"

lemma simple_dfs_simps [simp]:
  "dfs_cond simple_dfs S = True"
  "dfs_visit simple_dfs s x = RETURN (state s)"
  "dfs_post simple_dfs s x = RETURN (state s)"
  "dfs_remove simple_dfs s x = RETURN (state s)"
  "dfs_start simple_dfs x = ()"
  "dfs_pre_cond simple_dfs s x"
  "dfs_post_cond simple_dfs s s' x"
  "dfs_fe_inv simple_dfs s t it s'"
by (simp_all add: simple_dfs_def)

locale SimpleDFS = DFS_def succs v\<^sub>0 simple_dfs +
                   finite_succs_digraph succs v\<^sub>0
  for succs v\<^sub>0

sublocale SimpleDFS \<subseteq> DFS simple_dfs by unfold_locales simp_all

end