theory DFS_Code
imports 
  DFS_Impl
  Nested_DFS_Impl
  SE_DFS_Impl
begin

text {* This theory includes several examples for code generation of DFS. *}

subsection {* Simple example using lists for representing the graph. *}

definition list_succs :: "('n \<times> 'n) list \<Rightarrow> 'n \<Rightarrow> 'n ls" where
  "list_succs E x = list_to_ls (map snd (filter (\<lambda> p. fst p = x) E))"

fun list_field :: "('n \<times> 'n) list \<Rightarrow> 'n list" where 
  "list_field [] = []"
 | "list_field ((a,b)#xs) = a#b#list_field xs"

lemma list_field_Field:
  "set (list_field xs) = Field (set xs)"
by (induct xs) auto

interpretation ex: SimpleDFS_Impl rs_ops "set (list_field E)" "set E" "list_succs E"
  for E :: "(nat \<times> nat) list"
proof -
  interpret G: finite_digraph "set (list_field E)" "set E" by (unfold_locales) (simp | simp add: list_field_Field R_subset_Field)+

  have succs_s: "\<And>y. lsr.\<alpha> (list_succs E y) = G.succs y" 
    unfolding ls_ops_def list_succs_def G.succs_def list_to_ls_def
    by (auto simp add: image_iff ls_\<alpha>_def)
  thus "SimpleDFS_Impl rs_ops (set (list_field E)) (set E) (list_succs E)"
    by unfold_locales (simp_all add: ls_ops_def)
qed

definition "ex_dfs_code E = ex.dfs_code E"
prepare_code_thms ex.simple_dfs_code_def[folded ex_dfs_code_def]

(* export_code ex_dfs_code in SML  *)

(*export_code ex_dfs_code ex_to_fd in Haskell module_name DFS_Impl file Impl*)

subsection {* Simple example now using RB-Trees for representing the graph. *}

locale graph = g: StdMap mops
  for mops :: "(nat, nat ls, 'm, 'X) map_ops_scheme"
begin

definition hm_list_succs :: "'m \<Rightarrow> nat \<Rightarrow> nat ls" where
  "hm_list_succs G x \<equiv> case g.lookup x G of None \<Rightarrow> lsr.empty () | Some e \<Rightarrow> e"

fun hm_to_edges' :: "(nat \<times> nat ls) list \<Rightarrow> (nat \<times> nat) set" where
  "hm_to_edges' [] = {}"
  |" hm_to_edges' ((k, succs)#xs) = set (map (\<lambda>x. (k,x)) (lsr.to_list succs)) \<union> hm_to_edges' xs"

definition hm_to_edges where
  "hm_to_edges G = hm_to_edges' (g.to_list G)"

lemma hm_to_edges_finite:
  "finite (hm_to_edges G)"
unfolding hm_to_edges_def
by (induct ("g.to_list G")) auto

lemma in_succs_in_edges:
  assumes [simp]:"g.invar G"
  shows "x \<in> lsr.\<alpha> (hm_list_succs G y) \<longleftrightarrow> (y,x) \<in> hm_to_edges G"
proof -
  note g.correct[simp]
  note lsr.correct[simp]
  have inv[simp]: "\<And>s. lsr.invar s" by (simp add: ls_ops_def)
  
  show ?thesis
  proof
    assume A: "x \<in> lsr.\<alpha> (hm_list_succs G y)"
    show "(y,x) \<in> hm_to_edges G"
    proof (cases "g.lookup y G")
      case None with A have "hm_list_succs G y = lsr.empty ()" by (simp add: hm_list_succs_def)
      with A show ?thesis by simp
    next
      case (Some e) with A have "x \<in> lsr.\<alpha> e" unfolding hm_list_succs_def by simp
      
      from Some inv have "g.\<alpha> G y = Some e" by simp
      with inv have "(y, e) \<in> set (g.to_list G)" using map_of_SomeD by force
      thus ?thesis unfolding hm_to_edges_def
      proof (induct rule: hm_to_edges'.induct)
        case goal1 thus ?case by simp
      next
        case goal2 thus ?case
        proof (cases "(y,e) = (k, succs)")
          case False with goal2 show ?thesis by auto
        next
          case True with `x \<in> lsr.\<alpha> e` show ?thesis by simp
        qed
      qed
    qed
  next
    assume A: "(y, x) \<in> hm_to_edges G"
    then obtain xs where xs: "xs = g.to_list G" by blast
    with hm_to_edges_def have hme': "hm_to_edges G = hm_to_edges' xs" by simp
    
    from xs inv have "distinct (map fst xs)" by (simp add: hmr.correct)
    with A have "\<exists>s. x \<in> set (lsr.to_list s) \<and> (y, s) \<in> set xs"
      unfolding hme'
    proof (induct xs)
      case Nil thus ?case by simp
    next
      case (Cons a as) thus ?case
      proof (cases "y = fst a")
        case True with Cons have "y \<notin> set (map fst as)" by simp
        hence *: "(y,x) \<notin> hm_to_edges' as" by (induct as) auto
        
        from True obtain s where "snd a = s" by simp
        with True have **: "a = (y ,s)" by (cases a) simp
        with Cons(2) * have "(y,x) \<in> hm_to_edges' [(y, s)]" by simp
        hence "x \<in> set (lsr.to_list s)" by auto
        with ** show ?thesis by force
      next
        case False with Cons have "(y,x) \<in> hm_to_edges' as" by (cases a) auto
        moreover
        from Cons have "distinct (map fst as)" by simp
        ultimately show ?thesis using Cons(1) by auto
      qed
    qed
    then guess s .. note s = this
    with rmr.to_list_correct xs have "map_of (g.to_list G) y = Some s" by (fastforce intro: map_of_is_SomeI)
    with inv have "g.lookup y G = Some s" by auto
    hence "hm_list_succs G y = s" unfolding hm_list_succs_def by simp
    with s show "x \<in> lsr.\<alpha> (hm_list_succs G y)" by simp
  qed
qed

lemma to_list_empty:
  "g.to_list (g.empty ()) = []"
proof -
  note g.correct[simp]
  have "g.invar (g.empty ())" by simp
  hence "map_of (g.to_list (g.empty ())) = Map.empty" by simp
  thus ?thesis using map_of_eq_empty_iff[of "g.to_list (g.empty ())"] by simp
qed

primrec edges_to_graph :: "(nat \<times> nat) list \<Rightarrow> 'm" where
  "edges_to_graph [] = g.empty ()"
| "edges_to_graph (x#xs) = (let G = edges_to_graph xs in
                            case g.lookup (fst x) G of
                            None \<Rightarrow> g.update (fst x) (dlist_of_list [snd x]) G |
                            Some s \<Rightarrow> g.update (fst x) (Dlist.insert (snd x) s) G)"
end

definition list_to_set where "list_to_set (xs :: nat list) = list_to_rs xs"

interpretation G!: graph rm_ops by unfold_locales

definition "edges_to_graph \<equiv> G.edges_to_graph"
lemmas[code] = G.edges_to_graph.simps[folded edges_to_graph_def]

definition "hm_list_succs \<equiv> G.hm_list_succs"
lemmas[code] = G.hm_list_succs_def[folded hm_list_succs_def]

interpretation fg!: finite_digraph "Field (G.hm_to_edges g)" "G.hm_to_edges g" 
  for g
by (unfold_locales) (simp_all add: G.hm_to_edges_finite finite_Field R_subset_Field)

interpretation cour!: Courcoubetis_Impl rs_ops rs_ops ias_ops ias_ops "Field (G.hm_to_edges g)" "G.hm_to_edges g" A "hm_list_succs g"
  for g A
proof -
  have "rmr.invar g" by (simp add: rm_ops_def)
  hence "\<And> x y. x \<in> lsr.\<alpha> (G.hm_list_succs g y) \<longleftrightarrow> (y,x) \<in> G.hm_to_edges g" by (rule G.in_succs_in_edges)
  also have "\<And> x y. (y,x) \<in> G.hm_to_edges g \<longleftrightarrow> x \<in> fg.succs g y" unfolding fg.succs_def by simp
  finally have "\<And>y. lsr.\<alpha> (hm_list_succs g y) = fg.succs g y" by (auto simp add: hm_list_succs_def)
  thus "Courcoubetis_Impl rs_ops rs_ops ias_ops ias_ops (Field (G.hm_to_edges g)) (G.hm_to_edges g) A (hm_list_succs g)"
    apply unfold_locales
    apply (simp_all add: rs_ops_def ias_ops_def ls_ops_def)
    done
qed

definition "cour_restr_succs g = cour.restr_\<gamma>succs g"
prepare_code_thms cour.restr_\<gamma>succs_def[folded cour_restr_succs_def]

interpretation courR!: RedDFS_Impl_def R "cour_restr_succs g R" rs_ops ias_ops "Field (G.hm_to_edges g) - iasr.\<alpha> R" "rel_restrict (G.hm_to_edges g) (iasr.\<alpha> R)" ss
  for g and ss
by (rule cour.interpret_red_dfs_def)

definition "red_visit_impl ss = courR.red_visit_impl ss"
prepare_code_thms courR.red_visit_impl_def[folded red_visit_impl_def, unfolded courR.mk_ret_impl_def]

definition "red_remove_impl ss = courR.red_remove_impl ss"
prepare_code_thms courR.red_remove_impl_def[folded red_remove_impl_def, unfolded courR.mk_ret_impl_def]

definition "cour_code R g ss x = courR.red_dfs_code R g ss x"
prepare_code_thms  courR.dfs_code_def[folded cour_code_def, unfolded courR.red_dfs_impl_simps courR.red_post_impl_def, folded red_visit_impl_def red_remove_impl_def]

definition "mk_cycle = cour.mk_cycle"
lemmas[code] = cour.mk_cycle.simps[folded mk_cycle_def]

definition "blue_visit = cour.blue_visit_impl"
prepare_code_thms cour.blue_visit_impl_def[folded blue_visit_def]

definition "cour_post A g = cour.blue_post_impl A g"
prepare_code_thms cour.blue_post_impl_def[folded cour_post_def cour_restr_succs_def mk_cycle_def, folded cour_code_def, unfolded cour.courcoubetisM_impl_def]

definition "cour_dfs_code g A = cour.courcoubetis_code g A"
prepare_code_thms cour.dfs_code_def[folded cour_dfs_code_def cour.courcoubetisM_impl_def, unfolded cour.blue_dfs_impl_simps, folded cour_post_def blue_visit_def]

interpretation hpy!: HPY_Impl rs_ops rs_ops ias_ops ias_ops "Field (G.hm_to_edges g)" "G.hm_to_edges g" A "hm_list_succs g"
  for g A
proof -
  have "rmr.invar g" by (simp add: rm_ops_def)
  hence "\<And> x y. x \<in> lsr.\<alpha> (G.hm_list_succs g y) \<longleftrightarrow> (y,x) \<in> G.hm_to_edges g" by (rule G.in_succs_in_edges)
  also have "\<And> x y. (y,x) \<in> G.hm_to_edges g \<longleftrightarrow> x \<in> fg.succs g y" unfolding fg.succs_def by simp
  finally have "\<And>y. lsr.\<alpha> (hm_list_succs g y) = fg.succs g y" by (auto simp add: hm_list_succs_def)
  thus "HPY_Impl rs_ops rs_ops ias_ops ias_ops (Field (G.hm_to_edges g)) (G.hm_to_edges g) A (hm_list_succs g)"
    apply unfold_locales
    apply (simp_all add: rs_ops_def ias_ops_def ls_ops_def)
    done
qed

definition "hpy_restr_succs g = hpy.restr_\<gamma>succs g"
prepare_code_thms hpy.restr_\<gamma>succs_def[folded hpy_restr_succs_def]

interpretation hpyR!: RedDFS_Impl_def R "hpy_restr_succs g R" rs_ops ias_ops "Field (G.hm_to_edges g) - iasr.\<alpha> R" "rel_restrict (G.hm_to_edges g) (iasr.\<alpha> R)" ss
  for g and ss
by (rule hpy.interpret_red_dfs_def)

definition "hpyr_code R g ss x = hpyR.red_dfs_code R g ss x"
prepare_code_thms  hpyR.dfs_code_def[folded hpyr_code_def, unfolded courR.red_dfs_impl_simps courR.red_post_impl_def, folded red_visit_impl_def red_remove_impl_def]

definition "hpy_post A g = hpy.blue_post_impl A g"
prepare_code_thms hpy.blue_post_impl_def[folded hpy_post_def hpy_restr_succs_def mk_cycle_def, folded hpyr_code_def, unfolded hpy.hpyM_impl_def]

definition "hpy_dfs_code g A = hpy.hpy_code g A"
prepare_code_thms hpy.dfs_code_def[folded hpy_dfs_code_def, unfolded hpy.blue_dfs_impl_simps, folded hpy_post_def blue_visit_def]

definition "to_visited = hpy.to_visited"
prepare_code_thms hpy.to_visited_def[folded to_visited_def]

definition "is_aborted g A = hpy.is_aborted A g"
prepare_code_thms hpy.is_aborted_def[folded is_aborted_def, unfolded hpy.blue_dfs_impl_simps]

definition "counter_example = hpy.counter_example"
prepare_code_thms hpy.counter_example_def[folded counter_example_def]

interpretation IAse!:  HPY_SE_Impl rs_ops rs_ops ias_ops ias_ops "Field (G.hm_to_edges g)" "G.hm_to_edges g" A "hm_list_succs g"
  for g A
by unfold_locales (simp add: rs_ops_def)

definition "iase_remove A = IAse.se_remove_impl A"
prepare_code_thms IAse.se_remove_impl_def[folded iase_remove_def mk_cycle_def]

definition "iase_dfs_code g A = IAse.hpyse_code g A"
prepare_code_thms IAse.dfs_code_def[folded iase_dfs_code_def, unfolded IAse.se_dfs_impl_simps, folded iase_remove_def hpy_post_def blue_visit_def]

interpretation RBTse!: HPY_SE_Impl rs_ops rs_ops ias_ops rs_ops "Field (G.hm_to_edges g)" "G.hm_to_edges g" A "hm_list_succs g"
  for g A
by unfold_locales (simp add: rs_ops_def)

definition "se_remove A = RBTse.se_remove_impl A"
prepare_code_thms RBTse.se_remove_impl_def[folded se_remove_def mk_cycle_def]

definition "se_visit = RBTse.hpyblue.blue_visit_impl"
prepare_code_thms RBTse.hpyblue.blue_visit_impl_def[folded se_visit_def]

definition "se_post A g = RBTse.hpyblue.blue_post_impl A g"
prepare_code_thms RBTse.hpyblue.blue_post_impl_def[folded se_post_def hpy_restr_succs_def mk_cycle_def, folded hpyr_code_def, unfolded RBTse.hpyblue.hpyM_impl_def]

definition "se_dfs_code g A = RBTse.hpyse_code g A"
prepare_code_thms RBTse.dfs_code_def[folded se_dfs_code_def, unfolded RBTse.se_dfs_impl_simps RBTse.hpyblue.blue_dfs_impl_simps, folded se_remove_def se_post_def se_visit_def]

definition "se_to_visited = RBTse.hpyblue.to_visited"
prepare_code_thms RBTse.hpyblue.to_visited_def[folded se_to_visited_def]

definition "se_is_aborted g A = RBTse.hpyblue.is_aborted A g"
prepare_code_thms RBTse.hpyblue.is_aborted_def[folded se_is_aborted_def, unfolded RBTse.hpyblue.blue_dfs_impl_simps]

definition "se_counter_example = RBTse.hpyblue.counter_example"
prepare_code_thms RBTse.hpyblue.counter_example_def[folded se_counter_example_def]

export_code hpy_dfs_code cour_dfs_code iase_dfs_code se_dfs_code to_visited edges_to_graph list_to_set is_aborted counter_example se_to_visited se_is_aborted se_counter_example in SML file "code/Nested_DFS_Export.sml"
(*export_code hpy_dfs_code cour_dfs_code se_dfs_code to_visited edges_to_graph list_to_set is_aborted counter_example in OCaml module_name "Nested_DFS_Export" file "code/Nested_DFS_Export.ml"*)

(*export_code basic_dfs_code se_dfs_code hpy_dfs_code nested_to_visited edges_to_graph list_to_set in Haskell module_name "Nested_DFS_Export" file "code/haskell/"*)

end
