theory DFS_Impl
imports 
  DFS
  Graph_Impl
  "../CAVA_Refine"
  "../CAVA_Code_Target"
begin

type_synonym ('a,'b,'c,'d) impl_sws = "('a \<times> 'b \<times> 'c \<times> 'b list \<times> 'd)"

definition "idisc x \<equiv> let (S, n, d, st, ss) = x in d"
definition "istack x \<equiv> let (S, n, d, st, ss) = x in st"
definition "istate x \<equiv> let (S, n, d, st, ss) = x in S"
definition "istart x \<equiv> let (S, n, d, st, ss) = x in n"
definition "iss x \<equiv> let (S, n, d, st, ss) = x in ss"

lemma iaccess_simps [simp, code]:
  "idisc (S, n, d, st, ss) = d"
  "istack (S, n, d, st, ss) = st"
  "istate (S, n, d, st, ss) = S"
  "istart (S, n, d, st, ss) = n"
  "iss (S, n, d, st, ss) = ss"
by (simp_all add: idisc_def istack_def istate_def istart_def iss_def)

record ('S, 'n, 's, '\<gamma>) dfs_algorithm_impl =
  dfs_impl_cond :: "'S \<Rightarrow> bool"  -- "the condition to be satisified by the state S to continue searching from here."
  dfs_impl_visit :: "('S,'n,'s,'\<gamma>) impl_sws \<Rightarrow> 'n \<Rightarrow> 'S \<times> '\<gamma>"  -- "modifies the state for the current node BEFORE visiting the successors."
  dfs_impl_post :: "('S,'n,'s,'\<gamma>) impl_sws \<Rightarrow> 'n \<Rightarrow> 'S \<times> '\<gamma>"  -- "modifies the state for the current node AFTER having visited the successors (i.e. during backtracking)."
  dfs_impl_remove :: "('S,'n,'s,'\<gamma>) impl_sws  \<Rightarrow> 'n \<Rightarrow> 'S \<times> '\<gamma>" -- "modifies the state if a node has already been visited and is removed from the stack"
  dfs_impl_start :: "'n \<Rightarrow> 'S" -- "the starting state"

locale sws_impl = set!: StdSetDefs sops
  for sops :: "('n, 's, 'X) set_ops_scheme" +
  fixes \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> 'Sa"
begin
definition impl_sws_\<alpha> :: "('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> ('Sa, 'n) dfs_sws" where
  "impl_sws_\<alpha> x = (let (S, n, d, st, ss) = x in dfs_sws.make st (\<alpha>\<^sub>\<sigma> S) n (set.\<alpha> d))"

lemma impl_sws_\<alpha>_simps [simp]:
  "stack (impl_sws_\<alpha> (S, n, d, st, ss)) = st"
  "start (impl_sws_\<alpha> (S, n, d, st, ss)) = n"
  "discovered (impl_sws_\<alpha> (S, n, d, st, ss)) = set.\<alpha> d"
  "state (impl_sws_\<alpha> (S, n, d, st, ss)) = \<alpha>\<^sub>\<sigma> S"
 by (simp_all add: impl_sws_\<alpha>_def dfs_sws.defs)

lemma impl_sws_\<alpha>_conv [simp]:
  "stack (impl_sws_\<alpha> x) = istack x"
  "start (impl_sws_\<alpha> x) = istart x"
  "discovered (impl_sws_\<alpha> x) = set.\<alpha> (idisc x)"
  "state (impl_sws_\<alpha> x) = \<alpha>\<^sub>\<sigma> (istate x)"
by (cases x, simp)+
end

locale DFS_Impl_def = sws_impl sops \<alpha>\<^sub>\<sigma> +
                      Graph_succ_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 +
                      DFS_def \<alpha>succs v\<^sub>0 dfs
  for \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma>"
  and \<gamma>\<^sub>0 :: "'\<gamma>"
  and v\<^sub>0 :: "'n"
  and dfs :: "('Sa, 'n) dfs_algorithm"
  and sops :: "('n, 's, 'X) set_ops_scheme"
  and \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> 'Sa" +
  fixes impl_dfs :: "('S, 'n, 's, '\<gamma>, 'more) dfs_algorithm_impl_scheme"
  and dfs_impl_inv :: "('S,'n,'s,'\<gamma>) impl_sws \<Rightarrow> bool"
  and \<gamma>_inv :: "'\<gamma> \<Rightarrow> bool"
begin
abbreviation DI_to_graph ("\<G>") where
  "DI_to_graph \<equiv> \<lparr> verts = V, edges = E, start = fst, target = snd \<rparr>"

abbreviation DI_reachable ("_ \<rightarrow>\<star> _" [100,100] 40) where
  "DI_reachable v w \<equiv> Digraphs.reachable v \<G> w"

abbreviation DI_reachable1 ("_ \<rightarrow>+ _" [100,100] 40) where
  "DI_reachable1 \<equiv> reachable1"

definition statebr where
  "statebr = br (\<alpha>\<^sub>\<sigma> \<circ> fst) (\<lambda>(S, ss). \<gamma>_inv ss)"

lemma sv_statebr[refine]:
  "single_valued statebr"
unfolding statebr_def
by blast

lemma [refine_dref_RELATES]: "RELATES statebr" 
  by (simp add: RELATES_def)

lemma statebr_simp[simp]:
  "\<gamma>_inv ss \<Longrightarrow> ((Si, ss), S) \<in> statebr \<longleftrightarrow> \<alpha>\<^sub>\<sigma> Si = S"
unfolding statebr_def by (auto simp: refine_rel_defs)

definition impl_sws_invar where
  "impl_sws_invar s \<equiv> set.invar (idisc s) \<and> dfs_impl_inv s \<and> \<gamma>_inv (iss s)"

definition dfsbr where
  "dfsbr \<equiv> br impl_sws_\<alpha> impl_sws_invar"

lemma [refine_dref_RELATES]: "RELATES dfsbr"
  by (simp add: RELATES_def)

lemma sv_dfsbr[refine]:
  "single_valued dfsbr"
unfolding dfsbr_def
by blast

lemma dfsbr_statebr:
  "(si, s) \<in> dfsbr \<Longrightarrow> ((istate si, iss si), state s) \<in> statebr"
by (simp add: dfsbr_def impl_sws_invar_def statebr_def refine_rel_defs)

lemma dfsbrI:
  "\<lbrakk>set.invar (idisc si); dfs_impl_inv si; \<gamma>_inv (iss si); impl_sws_\<alpha> si = s\<rbrakk> 
    \<Longrightarrow> (si, s) \<in> dfsbr"
by (simp add: dfsbr_def impl_sws_invar_def refine_rel_defs)

lemma dfsbrRETURNI:
  "\<lbrakk>set.invar (idisc si); dfs_impl_inv si; \<gamma>_inv (iss si); impl_sws_\<alpha> si = s\<rbrakk> 
    \<Longrightarrow> RETURN si \<le>\<Down>dfsbr (RETURN s)"
by (refine_rcg dfsbrI)

lemma dfsbr_inv[intro]:
  "(si, s) \<in> dfsbr \<Longrightarrow> impl_sws_invar si"
unfolding dfsbr_def by (auto simp: refine_rel_defs)

lemma impl_sws_inv_impl_inv [intro]:
  "impl_sws_invar s \<Longrightarrow> dfs_impl_inv s"
by (simp add: impl_sws_invar_def)

lemma dfsbr_impl_inv [intro]:
  "(si, s) \<in> dfsbr \<Longrightarrow> dfs_impl_inv si"
by blast

lemma dfsbr_\<gamma>_inv [intro]:
  "(si, s) \<in> dfsbr \<Longrightarrow> \<gamma>_inv (iss si)"
unfolding dfsbr_def
by (simp add: impl_sws_invar_def refine_rel_defs)

lemma impl_sws_invar_set_invar:
  assumes "impl_sws_invar si"
  shows "set.invar (idisc si)"
using assms unfolding impl_sws_invar_def
by simp_all

lemma dfsbr_impl_sws_\<alpha>_simps:
  assumes "(si, s) \<in> dfsbr"
  shows "stack s = istack si"
  and "start s = istart si"
  and "discovered s = set.\<alpha> (idisc si)"
  and "state s = \<alpha>\<^sub>\<sigma> (istate si)"
using assms 
unfolding dfsbr_def 
by (simp_all add: refine_rel_defs)

definition dfs_start_impl :: "'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws" where
  "dfs_start_impl x = (dfs_impl_start impl_dfs x, x, set.empty (), [], \<gamma>\<^sub>0)"

definition succ_step_upd_impl :: "('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> '\<gamma> \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws"
where
  "succ_step_upd_impl s ss' = (let (S, n, d, st, ss) = s in (S, n, d, st, ss'))"

definition disc_step_upd_impl :: "'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'S \<Rightarrow> '\<gamma> \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws" 
where
  "disc_step_upd_impl x s S' ss' = (let (S, n, d, st, ss) = s in (S', n, set.ins_dj x d, x#st, ss'))"

definition fin_step_upd_impl  :: "'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'S \<Rightarrow> '\<gamma> \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws"
where
  "fin_step_upd_impl x s S' ss' = (let (S, n, d, st, ss) = s in (S', n, d, tl st, ss'))"

definition rem_step_upd_impl :: "('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'S \<Rightarrow> '\<gamma> \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws" 
where
  "rem_step_upd_impl s S' ss' = (let (S, n, d, st, ss) = s in (S', n, d, st, ss'))"

definition dfs_disc_step_impl :: "'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws"
where
  "dfs_disc_step_impl x s \<equiv> let (S', ss') = (dfs_impl_visit impl_dfs s x) in 
                            disc_step_upd_impl x s S' ss'"

definition dfs_fin_step_impl :: "'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws" 
where
  "dfs_fin_step_impl x s \<equiv> if dfs_impl_cond impl_dfs (istate s) 
        then let (S, ss) = dfs_impl_post impl_dfs s x in fin_step_upd_impl x s S ss
        else s"

definition dfs_rem_step_impl :: "'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws"
where
  "dfs_rem_step_impl x s \<equiv> let (S, ss) = dfs_impl_remove impl_dfs s x in rem_step_upd_impl s S ss"

definition dfs_algo_body_impl :: "('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws nres " 
where
  "dfs_algo_body_impl s x \<equiv>
        REC\<^sub>T (\<lambda>D (s, x). do {
           ASSERT (pre_dfs (impl_sws_\<alpha> s) x \<and> dfs_pre_cond dfs (impl_sws_\<alpha> s) x \<and> dfs_impl_cond impl_dfs (istate s) 
                   \<and> impl_sws_invar s \<and> \<gamma>_inv (iss s));
           if set.memb x (idisc s) then RETURN (dfs_rem_step_impl x s)
           else 
              (RETURN (dfs_disc_step_impl x s) \<guillemotright>=
              (\<lambda>s. do { let (suc, ss') = \<gamma>succs (iss s) x;
                        let s' = succ_step_upd_impl s ss';
                        ASSERT (\<gamma>_inv (iss s')); ASSERT (lsi.invar suc);
                        FOREACH\<^sub>C (lsi.\<alpha> suc) (\<lambda>s. dfs_impl_cond impl_dfs (istate s)) (\<lambda> x s. D (s, x)) s' }) ) \<guillemotright>=
              (\<lambda>s'. (* ASSERT ((\<exists>S. dfs_impl_cond impl_dfs (istate s') \<longrightarrow> 
                                 S = dfs_impl_visit impl_dfs s x
                               \<and> fe_inv_dfs (disc_step_upd x (impl_sws_\<alpha> s) (\<alpha>\<^sub>\<sigma> S)) x {} (impl_sws_\<alpha> s') 
                               \<and> dfs_fe_inv dfs (disc_step_upd x (impl_sws_\<alpha> s) (\<alpha>\<^sub>\<sigma> S)) x {} (impl_sws_\<alpha> s') )
                             \<and> impl_sws_invar s') \<guillemotright> *)
              RETURN (dfs_fin_step_impl x s'))
        }) (s, x)"

definition dfs_algo_impl :: "'n \<Rightarrow> ('S, 'n, 's, '\<gamma>) impl_sws nres" where
  "dfs_algo_impl x\<^sub>0 \<equiv> do {
     let s\<^sub>0 = dfs_start_impl x\<^sub>0;
     if \<not> dfs_impl_cond impl_dfs (istate s\<^sub>0) then RETURN s\<^sub>0
     else dfs_algo_body_impl s\<^sub>0 x\<^sub>0
   }"

schematic_lemma dfs_algo_impl_aux: 
  "RETURN (?foo x) \<le> dfs_algo_impl x"
  unfolding dfs_algo_impl_def dfs_start_impl_def
  unfolding dfs_algo_body_impl_def idisc_def istate_def
  unfolding dfs_disc_step_impl_def disc_step_upd_impl_def
  unfolding dfs_rem_step_impl_def rem_step_upd_impl_def
  unfolding dfs_fin_step_impl_def fin_step_upd_impl_def Let_def
  apply (refine_transfer the_resI)
  done

concrete_definition dfs_code for x uses dfs_algo_impl_aux

(* Convenience functions *)
definition to_visited :: "('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'n list"
  where "to_visited s \<equiv> set.to_list (idisc s)"

definition num_visited :: "('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> nat"
  where "num_visited s \<equiv> set.size (idisc s)"

definition is_aborted :: "('S, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> bool"
  where "is_aborted s \<equiv> \<not> dfs_impl_cond impl_dfs (istate s)"
end

locale DFS_Impl = set!: StdSet sops +
                  Graph_state_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv +
                  DFS_Impl_def \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 dfs sops \<alpha>\<^sub>\<sigma> impl_dfs dfs_impl_inv \<gamma>_inv +
                  DFS dfs \<alpha>succs v\<^sub>0
  for dfs :: "('Sa, 'n) dfs_algorithm"
  and impl_dfs :: "('S, 'n, 's, '\<gamma>, 'more) dfs_algorithm_impl_scheme"
  and dfs_impl_inv
  and \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> 'Sa"
  and sops :: "('n, 's, 'X) set_ops_scheme"
  and v\<^sub>0 :: "'n"
  and \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma>"
  and \<gamma>\<^sub>0 :: "'\<gamma>"
  and \<gamma>_inv :: "'\<gamma> \<Rightarrow> bool" +
  assumes cond_correct: "(si, s) \<in> dfsbr \<Longrightarrow> dfs_impl_cond impl_dfs (istate si) = dfs_cond dfs (state s)"
  and start_correct: "x \<in> V \<Longrightarrow> \<alpha>\<^sub>\<sigma>(dfs_impl_start impl_dfs x) = dfs_start dfs x"
  and start_inv: "x \<in> V \<Longrightarrow> dfs_impl_inv (dfs_start_impl x)"
  and visit_correct: "\<lbrakk>(si, s) \<in> dfsbr; (xi, x) \<in> Id; x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s) \<rbrakk>
                       \<Longrightarrow> RETURN (dfs_impl_visit impl_dfs si xi) \<le>\<Down>statebr (dfs_visit dfs s x)"
  and visit_inv: "\<lbrakk>(si, s) \<in> dfsbr; (xi, x) \<in> Id; x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s);
                   (Si, ssi) = dfs_impl_visit impl_dfs si xi \<rbrakk>
                       \<Longrightarrow> dfs_impl_inv (disc_step_upd_impl xi si Si ssi)"
  and remove_correct: "\<lbrakk>(si, s) \<in> dfsbr; (xi, x) \<in> Id; x \<in> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s) \<rbrakk> 
                       \<Longrightarrow> RETURN (dfs_impl_remove impl_dfs si xi) \<le>\<Down>statebr (dfs_remove dfs s x)"
  and remove_inv: "\<lbrakk>(si, s) \<in> dfsbr; (xi, x) \<in> Id; x \<in> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; dfs_cond dfs (state s);
                    (Si, ssi) = dfs_impl_remove impl_dfs si xi\<rbrakk>
                       \<Longrightarrow> dfs_impl_inv (rem_step_upd_impl si Si ssi)"
  and post_correct: "\<lbrakk>(si, s) \<in> dfsbr; (xi, x) \<in> Id; (si', s') \<in> dfsbr; x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; 
                      dfs_cond dfs (state s); inres (dfs_visit dfs s x) S; fe_inv_dfs (disc_step_upd x s S) x {} s'; 
                       dfs_fe_inv dfs (disc_step_upd x s S) x {} s'; dfs_cond dfs (state s'); x \<notin> finished s'\<rbrakk> 
                     \<Longrightarrow> RETURN (dfs_impl_post impl_dfs si' xi) \<le>\<Down>statebr (dfs_post dfs s' x)"
  and post_inv: "\<lbrakk>(si, s) \<in> dfsbr; (xi, x) \<in> Id; (si', s') \<in> dfsbr; x \<notin> discovered s; pre_dfs s x; dfs_pre_cond dfs s x; 
                   dfs_cond dfs (state s); inres (dfs_visit dfs s x) S; fe_inv_dfs (disc_step_upd x s S) x {} s'; 
                   dfs_fe_inv dfs (disc_step_upd x s S) x {} s'; dfs_cond dfs (state s'); x \<notin> finished s';
                   (Si, ssi) = dfs_impl_post impl_dfs si' xi\<rbrakk>
                  \<Longrightarrow> dfs_impl_inv (fin_step_upd_impl xi si' Si ssi)"
  and impl_inv_succ_inv: "dfs_impl_inv si \<Longrightarrow> dfs_impl_inv (succ_step_upd_impl si ssi)"
begin

lemma dfsbr_disc:
  "(si, s) \<in> dfsbr \<Longrightarrow> set.memb x (idisc si) \<longleftrightarrow> x \<in> discovered s"
unfolding dfsbr_def
by (simp add: impl_sws_invar_def set.correct refine_rel_defs)

lemma dfs_start_impl_correct:
  "x \<in> V \<Longrightarrow> (dfs_start_impl x, dfs_alg_start x) \<in> dfsbr"
using start_correct start_inv
unfolding dfs_start_impl_def dfs_alg_start_def
by (intro dfsbrI) (simp_all add: set.correct refine_rel_defs)

lemma disc_step_upd_impl_correct:
  assumes "(si, s) \<in> dfsbr" "(xi, x) \<in> Id" "((Si, ssi), S) \<in> statebr"
  and "x \<notin> discovered s"
  and "pre_dfs s x" "dfs_pre_cond dfs s x" "dfs_cond dfs (state s)"
  and "(Si, ssi) = dfs_impl_visit impl_dfs si xi"
  shows "RETURN (disc_step_upd_impl xi si Si ssi) \<le>\<Down>dfsbr (RETURN (disc_step_upd x s S))"
  using assms
  apply (refine_rcg)
  apply (cases si)
  using visit_inv[OF assms(1,2,4-)]
  apply (simp_all 
    add: impl_sws_\<alpha>_def disc_step_upd_def disc_step_upd_impl_def dfs_sws.defs 
    set.correct statebr_def dfsbr_def impl_sws_invar_def refine_rel_defs
    split: prod.splits)
done

lemma fin_step_upd_impl_correct:
  assumes "(si, s) \<in> dfsbr" "(xi, x) \<in> Id" "(si', s') \<in> dfsbr" "((Si', ssi'), S') \<in> statebr"
  and "x \<notin> discovered s"
  and "pre_dfs s x"
  and "dfs_pre_cond dfs s x"
  and "dfs_cond dfs (state s)"
  and "inres (dfs_visit dfs s x) S"
  and "fe_inv_dfs (disc_step_upd x s S) x {} s'"
  and "dfs_fe_inv dfs (disc_step_upd x s S) x {} s'"
  and "dfs_cond dfs (state s')"
  and "x \<notin> finished s'"
  and "(Si', ssi') = dfs_impl_post impl_dfs si' xi"
  shows "RETURN (fin_step_upd_impl xi si' Si' ssi') \<le>\<Down>dfsbr (RETURN (fin_step_upd x s' S'))" (is "?P (RETURN ?L) (RETURN ?R)")
proof -
  from assms have "set.invar (idisc si')" "s' = impl_sws_\<alpha> si'" 
    unfolding dfsbr_def refine_rel_defs using impl_sws_invar_set_invar by auto
  then show ?thesis
  proof (intro dfsbrRETURNI)
    case goal1 thus ?case by (cases si') (simp add: fin_step_upd_impl_def)
  next
    case goal2 show ?case using assms post_inv by blast
  next
    case goal3 thus ?case using assms(4) 
      by (simp add: statebr_def refine_rel_defs fin_step_upd_impl_def split: prod.split)
  next
    case goal4 thus ?case using post_correct[OF assms(1-3,5-13)] assms
      by (cases si') (simp 
        add: fin_step_upd_impl_def fin_step_upd_def impl_sws_\<alpha>_def 
             dfs_sws.defs set.correct statebr_def refine_rel_defs
        split: prod.splits)
  qed
qed

lemma rem_step_upd_impl_correct:
  assumes "(si, s) \<in> dfsbr" "(xi, x) \<in> Id" "((Si, ssi), S) \<in> statebr"
  and "x \<in> discovered s"
  and "pre_dfs s x" "dfs_pre_cond dfs s x" "dfs_cond dfs (state s)"
  and "(Si, ssi) = dfs_impl_remove impl_dfs si xi"
  shows "RETURN (rem_step_upd_impl si Si ssi) \<le> \<Down>dfsbr (RETURN (s\<lparr> state := S\<rparr>))"
  using assms
  apply (refine_rcg)
  apply (cases si)
  using remove_inv[OF assms(1,2,4-)]
  apply (simp 
    add: impl_sws_\<alpha>_def rem_step_upd_impl_def impl_sws_invar_def dfs_sws.defs dfsbr_def 
         statebr_def refine_rel_defs
    split: prod.splits)
done

lemma dfs_disc_step_impl_correct:
  assumes "(si, s) \<in> dfsbr" "(xi, x) \<in> Id"
  and "x \<notin> discovered s"
  and "pre_dfs s x" "dfs_pre_cond dfs s x" "dfs_cond dfs (state s)"
  shows "RETURN (dfs_disc_step_impl xi si) \<le>\<Down>dfsbr (dfs_disc_step x s)"
proof -
  from assms have "\<not> set.memb xi (idisc si)" using dfsbr_disc by blast
  with assms show ?thesis
    unfolding dfs_disc_step_impl_def dfs_disc_step_def
    apply (intro bind2letRETURN_refine)
    apply (blast intro: visit_correct)
    apply (auto intro: disc_step_upd_impl_correct split: prod.split)
    done
qed

lemma dfs_fin_step_impl_correct:
  assumes INV: "(si, s) \<in> dfsbr" "(si', s') \<in> dfsbr" "(xi, x) \<in> Id"
  and "x \<notin> discovered s"
  and "pre_dfs s x"
  and "dfs_pre_cond dfs s x"
  and "dfs_cond dfs (state s)"
  and A: "dfs_cond dfs (state s') \<longrightarrow> inres (dfs_visit dfs s x) S
                                  \<and> fe_inv_dfs (disc_step_upd x s S) x {} s'
                                  \<and> dfs_fe_inv dfs (disc_step_upd x s S) x {} s'"
  shows "RETURN (dfs_fin_step_impl xi si') \<le>\<Down>dfsbr (dfs_fin_step x s')"
proof -
  {
    assume C: "dfs_cond dfs (state s')"
    with A have FE: "fe_inv_dfs (disc_step_upd x s S) x {} s'" (is "fe_inv_dfs ?sd x {} s'") by blast
    
    have "x \<in> set (stack ?sd)" by (simp add: disc_step_upd_def)
    with FE[unfolded fe_inv_dfs_unfold] C have "x \<in> set (stack s')" by simp
    with stack_implies_not_finished FE have "x \<notin> finished s'" by blast
  } note C_F = this

  show ?thesis
    unfolding dfs_fin_step_impl_def dfs_fin_step_def
  proof (refine_rcg, refine_dref_type)
    case goal1 from cond_correct INV(2) show ?case .
  next
    case goal2 with assms C_F show ?case by (blast intro: post_correct)
  next
    case goal3 with assms C_F show ?case 
      apply (intro RETURN_ref_RETURND)
      apply (intro fin_step_upd_impl_correct)
      apply auto
      done
  next
    case goal4 with INV show ?case by simp
  qed
qed

lemma dfs_rem_step_impl_correct:
  assumes "(si, s) \<in> dfsbr" "(xi, x) \<in> Id"
  and "x \<in> discovered s"
  and "pre_dfs s x" "dfs_pre_cond dfs s x" "dfs_cond dfs (state s)"
  shows "RETURN (dfs_rem_step_impl xi si) \<le>\<Down>dfsbr (dfs_rem_step x s)"
unfolding dfs_rem_step_impl_def dfs_rem_step_def
using assms
  apply (intro bind2letRETURN_refine)
    apply (blast intro: remove_correct)
    apply (auto intro!: rem_step_upd_impl_correct split: prod.split)
done

lemma \<gamma>succs_\<gamma>_inv:
  assumes "(si, s) \<in> dfsbr"
  and "\<gamma>succs (iss si) n = (ls, ss')"
  shows "\<gamma>_inv ss'"
proof -
  from assms have "\<gamma>_inv (iss si)" by blast
  moreover from assms have "ss' = snd (\<gamma>succs (iss si) n)" by simp
  ultimately show ?thesis by (simp add: \<gamma>_inv_preserve)
qed

lemma \<gamma>succs_lsi_inv:
  assumes "(si, s) \<in> dfsbr"
  and "n \<in> V"
  and "\<gamma>succs (iss si) n = (ls, ss')"
  shows "lsi.invar ls"
proof -
  from assms have "\<gamma>_inv (iss si)" by blast
  moreover from assms have "ls = fst (\<gamma>succs (iss si) n)" by simp
  moreover note `n \<in> V`
  ultimately show ?thesis by (simp add: ls_invar)
qed

lemma succ_step_upd_impl_\<gamma>_inv:
  "\<gamma>_inv ss = \<gamma>_inv (iss (succ_step_upd_impl s ss))"
by (simp add: succ_step_upd_impl_def split: prod.split)

lemma succ_step_upd_impl_correct:
  assumes "(si, s) \<in> dfsbr"
  and "\<gamma>_inv ss"
  shows "(succ_step_upd_impl si ss, s) \<in> dfsbr"
using assms impl_inv_succ_inv[of si]
unfolding succ_step_upd_impl_def
by (auto simp add: dfsbr_def impl_sws_invar_def refine_rel_defs split: prod.split)

lemma dfs_algo_body_impl_correct:
  assumes "(si, s) \<in> dfsbr" "(xi, x) \<in> Id"
  shows "dfs_algo_body_impl si xi \<le>\<Down>dfsbr (dfs_algo_body s x)"
unfolding dfs_algo_body_impl_def dfs_algo_body_def
using assms
  apply (refine_rcg inj_on_id)
  apply (refine_dref_type)
    apply (simp add: refine_rel_defs)
    apply (simp add: dfsbr_def impl_sws_invar_def set.correct refine_rel_defs)
    apply (simp add: dfsbr_def impl_sws_invar_def set.correct refine_rel_defs)
    apply (simp add: refine_rel_defs)
      apply (blast intro!: cond_correct[THEN iffD2])
    apply (simp add: refine_rel_defs)
      apply (blast)
    apply (simp add: refine_rel_defs dfsbr_def impl_sws_invar_def)
    apply (simp add: dfsbr_def impl_sws_invar_def set.correct refine_rel_defs)
    apply (auto intro!: dfs_rem_step_impl_correct simp: refine_rel_defs)[]
    apply (auto intro!: dfs_disc_step_impl_correct simp: refine_rel_defs)[]
    apply (blast intro!: succ_step_upd_impl_\<gamma>_inv[THEN iffD1] \<gamma>succs_\<gamma>_inv)
    apply (blast intro: \<gamma>succs_lsi_inv)
    apply (simp add: refine_rel_defs)
      apply (rule_tac \<gamma> = "iss xaa" in \<alpha>succs_\<gamma>succs)
      apply (simp add: dfsbr_def impl_sws_invar_def refine_rel_defs)
      apply simp
    apply (blast intro: succ_step_upd_impl_correct dest: succ_step_upd_impl_\<gamma>_inv[THEN iffD2])
    apply (blast intro!: cond_correct)
    apply (simp add: refine_rel_defs)
    apply (erule exE)
      apply (rule_tac s = a and si = aa in dfs_fin_step_impl_correct)
      apply (simp_all add: refine_rel_defs)
done

lemma dfs_algo_impl_refine:
  assumes "x \<in> V"
  shows "dfs_algo_impl x \<le>\<Down>dfsbr (dfs_algo x)"
unfolding dfs_algo_def dfs_algo_impl_def
using assms(1) dfs_start_impl_correct[OF `x \<in> V`]
by refine_rcg (auto intro!: cond_correct dfs_algo_body_impl_correct)

lemma dfs_algo_impl_correct:
  assumes "x \<in> V"
  shows "dfs_algo_impl x \<le>\<Down>dfsbr (SPEC (\<lambda>s. post_dfs (dfs_alg_start x) s x \<and> dfs_post_cond dfs (dfs_alg_start x) s x))"
proof -
  note dfs_algo_impl_refine[OF assms]
  also note dfs_post[OF assms]
  finally show ?thesis .
qed

lemma dfs_code_correct:
  assumes "x \<in> V"
  shows "RETURN (dfs_code x) \<le>\<Down>dfsbr (SPEC (\<lambda>s. post_dfs (dfs_alg_start x) s x \<and> dfs_post_cond dfs (dfs_alg_start x) s x))"
proof -
  note dfs_code.refine
  also note dfs_algo_impl_correct[OF assms]
  finally show ?thesis .
qed

lemma dfs_codeD:
  "\<lbrakk> x \<in> V; s = dfs_code x \<rbrakk> \<Longrightarrow> post_dfs (dfs_alg_start x) (impl_sws_\<alpha> s) x \<and> dfs_post_cond dfs (dfs_alg_start x) (impl_sws_\<alpha> s) x \<and> impl_sws_invar s"
using dfs_code_correct[of x]
unfolding dfsbr_def refine_rel_defs
by (auto elim!: RETURN_ref_SPECD)
end

declare (in DFS_Impl_def) dfsbr_impl_sws_\<alpha>_simps[simp]

definition simple_impl_dfs :: "(unit, 'n, 's, '\<gamma>) dfs_algorithm_impl" where
  "simple_impl_dfs = \<lparr> dfs_impl_cond = \<lambda>_. True, 
                        dfs_impl_visit = \<lambda>s _. (istate s, iss s), 
                        dfs_impl_post = \<lambda>s _. (istate s, iss s), 
                        dfs_impl_remove = \<lambda>s _. (istate s, iss s), 
                        dfs_impl_start = \<lambda>_. ()\<rparr>"

lemma simple_impl_dfs_simps [simp]:
  "dfs_impl_cond simple_impl_dfs S = True"
  "dfs_impl_visit simple_impl_dfs s x = (istate s, iss s)"
  "dfs_impl_post simple_impl_dfs s x = (istate s, iss s)"
  "dfs_impl_remove simple_impl_dfs s x = (istate s, iss s)"
  "dfs_impl_start simple_impl_dfs x = ()"
by (simp_all add: simple_impl_dfs_def)

locale SimpleDFS_Impl_def = DFS_Impl_def _ _ _ simple_dfs _ id simple_impl_dfs "\<lambda>_. True"

locale SimpleDFS_Impl = StdSet sops +
                        Graph_state_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv +
                        SimpleDFS_Impl_def \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 sops \<gamma>_inv
  for sops :: "('n, 's, 'X) set_ops_scheme" 
  and \<gamma>succs
  and v\<^sub>0 :: "'n" and \<gamma>\<^sub>0 :: "'\<gamma>"
  and \<gamma>_inv

sublocale SimpleDFS_Impl \<subseteq> DFS_Impl simple_dfs simple_impl_dfs "\<lambda>_.True" id
 apply unfold_locales
 apply simp_all
 apply (refine_rcg, fastforce dest: dfsbr_\<gamma>_inv)+
 done

context SimpleDFS_Impl
begin
lemmas simple_dfs_code_def = dfs_code_def[simplified simple_impl_dfs_simps istate_def Let_def]
end
end
