theory Graph_Impl
imports
  Graph_Ext
  "../CAVA_Refine"
begin

locale Graph_succ_Impl = succ!: StdSetDefs succ_ops
  for succ_ops :: "('n, '\<Gamma>, _) set_ops_scheme" +
  fixes \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> '\<Gamma> \<times> '\<gamma>"
  and v\<^sub>0 :: "'n"
  and \<gamma>\<^sub>0 :: "'\<gamma>"
begin
definition "\<alpha>succs x = succ.\<alpha> (fst (\<gamma>succs \<gamma>\<^sub>0 x))"
end

sublocale Graph_succ_Impl \<subseteq> succs_digraph \<alpha>succs v\<^sub>0 .

context Graph_succ_Impl
begin

abbreviation fsd_to_graph ("\<G>") where
  "fsd_to_graph \<equiv> \<lparr> verts = V, edges = E, start = fst, target = snd \<rparr>"

abbreviation fsd_reachable ("_ \<rightarrow>\<star> _" [100,100] 40) where
  "fsd_reachable v w \<equiv> Digraphs.reachable v \<G> w"

abbreviation fsd_reachable1 ("_ \<rightarrow>+ _" [100,100] 40) where
  "fsd_reachable1 \<equiv> reachable1"
end

locale Graph_fin_succ_Impl = Graph_succ_Impl succ_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 +
                             finite_succs_digraph \<alpha>succs v\<^sub>0
  for succ_ops \<gamma>succs and v\<^sub>0 :: "'n" and \<gamma>\<^sub>0 :: "'\<gamma>"


locale Graph_state_Impl = Graph_fin_succ_Impl succ_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0
  for succ_ops :: "('n, '\<Gamma>, _) set_ops_scheme"
  and \<gamma>succs v\<^sub>0
  and \<gamma>\<^sub>0 :: "'\<gamma>" +
  fixes \<gamma>_inv :: "'\<gamma> \<Rightarrow> bool"
  assumes \<gamma>succs_indep: "\<And>\<gamma> \<gamma>'. \<lbrakk> \<gamma>_inv \<gamma>; \<gamma>_inv \<gamma>'\<rbrakk> \<Longrightarrow> fst (\<gamma>succs \<gamma> v) = fst (\<gamma>succs \<gamma>' v)"
  and \<gamma>0_inv[simp]: "\<gamma>_inv \<gamma>\<^sub>0"
  and \<gamma>_inv_preserve: "\<gamma>_inv \<gamma> \<Longrightarrow> \<gamma>_inv (snd (\<gamma>succs \<gamma> v))"
  and ls_invar: "\<gamma>_inv \<gamma> \<Longrightarrow> succ.invar (fst (\<gamma>succs \<gamma> v))"
begin

lemma \<alpha>succs_\<gamma>succs:
  assumes "\<gamma>_inv \<gamma>"
  and "\<gamma>succs \<gamma> s = (ls, \<gamma>')"
  shows "\<alpha>succs s = succ.\<alpha> ls"
proof -
  from assms have "\<alpha>succs s = succ.\<alpha> (fst (\<gamma>succs \<gamma> s))"
    using \<gamma>succs_indep[OF \<gamma>0_inv assms(1)] by (simp add: \<alpha>succs_def)
  with assms show ?thesis by simp
qed

lemma \<gamma>succs_fin_subgraphI:
  fixes \<gamma>succs' :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> '\<Gamma> \<times> '\<gamma>"
  assumes "{(v, v'). v' \<in> succ.\<alpha> (fst (\<gamma>succs' \<gamma> v))} \<subseteq> {(v, v'). v' \<in> succ.\<alpha> (fst (\<gamma>succs \<gamma> v))}"
  and "w \<in> V" "\<gamma>_inv \<gamma>"
  shows "Graph_fin_succ_Impl succ_ops \<gamma>succs' w \<gamma>"
proof -
  interpret sub: Graph_succ_Impl succ_ops \<gamma>succs' w \<gamma> .
  
  from \<gamma>succs_indep[OF assms(3) \<gamma>0_inv] assms(1) have
    "{(v, v'). v' \<in> succ.\<alpha> (fst (\<gamma>succs' \<gamma> v))} \<subseteq> {(v, v'). v' \<in> succ.\<alpha> (fst (\<gamma>succs \<gamma>\<^sub>0 v))}"
    by simp

  then interpret sub': finite_succs_digraph sub.\<alpha>succs w
    using finite_succs_subgraphI[OF _ `w \<in> V`, of sub.\<alpha>succs]
    unfolding \<alpha>succs_def sub.\<alpha>succs_def by auto

  show ?thesis by unfold_locales
qed
end
end