theory Nested_DFS
imports DFS
begin

(* General Nested DFS *)
datatype 'n nested_res = NO_LASSO | LASSO 'n "'n list" "'n list"
lemma nested_res_refine_transfer [refine_transfer]:
  assumes "\<And>r. \<alpha> fnci \<le> fnc"
  assumes "\<And>n nl l. \<alpha> (fcpi n nl l) \<le> fcp n nl l"
  shows "\<alpha> (nested_res_case fnci fcpi x) \<le> nested_res_case fnc fcp x"
  using assms
  by (auto split: nested_res.split)

datatype ('n, 's) blue_res = CYCLE_NO "'s" | CYCLE_PRE 'n "'n list" | CYCLE_FIN 'n 'n "'n list" "'n list" "'n list"
lemma blue_res_refine_transfer [refine_transfer]:
  assumes "\<And>r. \<alpha> (fnci r) \<le> fnc r"
  assumes "\<And>n nl. \<alpha> (fcpi n nl) \<le> fcp n nl"
  assumes "\<And>y v ys vs cyc. \<alpha> (fcfi y v ys vs cyc) \<le> fcf y v ys vs cyc"
  shows "\<alpha> (blue_res_case fnci fcpi fcfi x) \<le> blue_res_case fnc fcp fcf x"
  using assms
  by (auto split: blue_res.split)

locale NestedDFS_pre = succs_digraph succs v\<^sub>0
  for succs :: "'n \<Rightarrow> 'n set"
  and v\<^sub>0 :: "'n" +
  fixes A :: "'n set"
begin

definition is_lasso_full where
  "is_lasso_full x y v ys vs cyc \<longleftrightarrow> (y \<in> A \<or> v \<in> A) \<and> vwalk ys \<G> \<and> hd ys = x \<and> last ys = y
                                   \<and> vwalk vs \<G> \<and> 1 < length vs \<and> hd vs = y \<and> last vs = v 
                                   \<and> vwalk cyc \<G> \<and> hd cyc = v \<and> last cyc = y"

definition is_lasso where
  "is_lasso x y to_y cyc \<longleftrightarrow> y \<in> A \<and> vwalk to_y \<G> \<and> hd to_y = x \<and> last to_y = y
                             \<and> vwalk cyc \<G> \<and> length cyc > 1 \<and> hd cyc = y \<and> last cyc = y"
end

locale NestedDFS_def = NestedDFS_pre succs v\<^sub>0 A +
                       DFS_def succs v\<^sub>0 dfs
  for dfs :: "(('n, 'n set) blue_res, 'n) dfs_algorithm"
  and succs :: "'n \<Rightarrow> 'n set"
  and v\<^sub>0 :: "'n"
  and A :: "'n set"
begin

definition nested_dfs :: "'n \<Rightarrow> 'n nested_res nres" where
  "nested_dfs x = do {
                s \<leftarrow> dfs_algo x;
                case (state s) of
                CYCLE_FIN y v ys vs cyc \<Rightarrow> if y \<in> A then RETURN (LASSO y ys (vs \<oplus> cyc))
                                          else RETURN (LASSO v (ys \<oplus> vs) (cyc \<oplus> vs))
               | _ \<Rightarrow> RETURN NO_LASSO }"
end

locale NestedDFS = NestedDFS_def dfs succs v\<^sub>0 A  +
                   DFS dfs succs v\<^sub>0
  for dfs succs v\<^sub>0 A +
  assumes dfs_sound: "x \<in> V \<Longrightarrow> dfs_algo x \<le> SPEC (\<lambda>s. \<forall>y v ys vs cyc. (state s = CYCLE_FIN y v ys vs cyc \<longrightarrow> is_lasso_full x y v ys vs cyc))"
  and dfs_complete: "x \<in> V \<Longrightarrow> dfs_algo x \<le> SPEC (\<lambda>s. (\<exists>R. state s = CYCLE_NO R) \<longrightarrow> \<not>(\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> A \<and> v \<rightarrow>+ v))"
  and dfs_not_CYCLE_PRE: "x \<in> V \<Longrightarrow> dfs_algo x \<le> SPEC (\<lambda>s. \<not>(\<exists>a b. state s = CYCLE_PRE a b))"
begin

lemma nested_dfs_sound:
  assumes "x \<in> V"
  shows "nested_dfs x \<le> SPEC (\<lambda>res. \<forall>y to_y cyc. (res = LASSO y to_y cyc \<longrightarrow> is_lasso x y to_y cyc))"
  unfolding nested_dfs_def
  apply (intro refine_vcg)
  apply (intro pw_leI)
  apply (simp add: dfs_algo_NF[OF assms])
proof (intro allI impI)
  case (goal1 s) thus ?case
  proof (cases "state s")
    case (CYCLE_FIN y v ys vs cyc) with goal1 dfs_sound[OF assms] have cyc: "is_lasso_full x y v ys vs cyc"
      by (blast dest: inres_SPEC)
    note is_lasso_full_def[simp]
    note is_lasso_full_def[THEN iffD1,dest]
    
    show ?thesis
    proof (cases "y \<in> A")
      case True with cyc have "vwalk ys \<G>" "hd ys = x" "last ys = y" by blast+
      moreover from cyc have join: "joinable vs cyc" by auto
      with cyc have "vwalk (vs \<oplus> cyc) \<G>" by (blast intro: vwalk_join_vwalk)
      moreover from join cyc have "hd (vs \<oplus> cyc) = y" and "last (vs \<oplus> cyc) = y" by (simp_all add: vwalk_join_hd vwalk_join_last)
      moreover from cyc have "length (vs \<oplus> cyc) > 1" by (simp add: vwalk_join_def)
      ultimately show ?thesis using True CYCLE_FIN by (simp add: is_lasso_def)
    next
      case False with cyc have "v \<in> A" and ne: "v \<noteq> y" by auto
      moreover from cyc have join_yv: "joinable ys vs" and join_cv: "joinable cyc vs" by auto
      with cyc have "vwalk (ys \<oplus> vs) \<G>" by (blast intro: vwalk_join_vwalk)
      moreover from join_yv cyc have "hd (ys \<oplus> vs) = x" "last (ys \<oplus> vs) = v" by (simp_all add: vwalk_join_hd vwalk_join_last)
      moreover from join_cv cyc have "vwalk (cyc \<oplus> vs) \<G>" by (blast intro: vwalk_join_vwalk)
      moreover from join_cv cyc have "hd (cyc \<oplus> vs) = v" and "last (cyc \<oplus> vs) = v" by (simp_all add: vwalk_join_hd vwalk_join_last)
      moreover from ne cyc have "length cyc > 1" by (induct cyc arbitrary: v y) auto
      with cyc have "length (cyc \<oplus> vs) > 1" by (fastforce simp add: vwalk_join_def)
      ultimately show ?thesis using False CYCLE_FIN by (simp add: is_lasso_def)
    qed
  qed simp_all
qed

lemma nested_dfs_sound_simple:
  assumes "x \<in> V"
  shows "nested_dfs x \<le> SPEC (\<lambda>res. \<forall>y to_y cyc. (res = LASSO y to_y cyc \<longrightarrow> y \<in> A \<and> x \<rightarrow>\<star> y \<and> y \<rightarrow>+ y))"
apply (rule order_trans[OF nested_dfs_sound[OF assms]])
apply (intro refine_vcg)
proof (clarsimp simp add: is_lasso_def)
  case goal1 hence "last to_y \<rightarrow>+ last to_y" using vwalk_start_last_reachable[of cyc] by auto
  moreover have "hd to_y \<rightarrow>\<star> last to_y"
  proof (cases "hd to_y = last to_y")
    case True from goal1 have "hd to_y \<in> V" unfolding vwalk_def by auto
    with True self_reachable show ?thesis by simp
  next
    case False with goal1 have "to_y \<noteq> []" unfolding vwalk_def by auto
    with False have "length to_y > 1" by (induct to_y) auto
    with goal1 vwalk_start_last_reachable[of to_y] show ?thesis by auto
  qed
  ultimately show ?case using goal1 by simp
qed

lemma nested_dfs_sound_alt:
  assumes "x \<in> V"
  and "\<not>(\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> A \<and> v \<rightarrow>+ v)"
  shows "nested_dfs x \<le> SPEC (\<lambda>res. res = NO_LASSO)"
apply (rule order_trans[OF nested_dfs_sound_simple[OF assms(1)]])
apply (intro refine_vcg)
proof clarsimp
  case (goal1 res) show ?case
  proof (cases res)
    case (LASSO y to_y cyc) with goal1 have "y \<in> A \<and> x \<rightarrow>\<star> y \<and> y \<rightarrow>+ y" by auto
    with assms have "False" by auto
    thus ?thesis ..
  qed simp
qed

lemma nested_dfs_complete:
  assumes "x \<in> V"
  shows "nested_dfs x \<le> SPEC (\<lambda>res. res = NO_LASSO \<longrightarrow> \<not>(\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> A \<and> v \<rightarrow>+ v))"
unfolding nested_dfs_def
apply (intro refine_vcg)
apply (intro pw_leI)
apply (simp add: dfs_algo_NF[OF assms])
proof (intro impI allI)
  case (goal1 s) with dfs_not_CYCLE_PRE[OF assms] have NP: "\<And>a b. state s \<noteq> CYCLE_PRE a b"
    by (blast dest: inres_SPEC)
  with goal1 dfs_complete[OF assms] show ?case by (auto dest: inres_SPEC split: blue_res.split)
qed

lemma nested_dfs_NF:
  "x \<in> V \<Longrightarrow> nofail (nested_dfs x)"
by (rule SPEC_nofail) (rule nested_dfs_sound)

lemma nested_dfs_NO_LASSO_iff:
  assumes "x \<in> V"
  shows "nested_dfs x \<le> SPEC (\<lambda>res. res = NO_LASSO \<longleftrightarrow> \<not>(\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> A \<and> v \<rightarrow>+ v))"
apply (rule SPEC_iff)
  apply (fastforce intro: order_trans[OF nested_dfs_complete[OF assms]])
  apply (cases "\<not> (\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> A \<and> v \<rightarrow>+ v)")
    apply (rule order_trans[OF nested_dfs_sound_alt[OF assms]])
    apply (auto simp add: nested_dfs_NF[OF assms])
done
end

(* Now the Red/Blue variant *)

locale RedDFS_pre = succs_digraph succs v\<^sub>0
  for succs and v\<^sub>0 :: "'n"+
  fixes ss :: "'n set"
begin

type_synonym 'x red_cycle = "('x \<times> 'x list) option"
type_synonym 'x red_sws = "('x red_cycle, 'x) dfs_sws"

abbreviation mk_ret :: "'n \<Rightarrow> 'n red_sws \<Rightarrow> 'n red_cycle" where
  "mk_ret x s \<equiv> Some (x, rev (x#stack s))"

definition red_dfs_visit :: "'n red_sws \<Rightarrow> 'n \<Rightarrow> 'n red_cycle nres" where
  "red_dfs_visit s x \<equiv> case state s of 
                      Some x \<Rightarrow> RETURN (Some x) 
                    | None \<Rightarrow> if (x \<noteq> start s \<and> x \<in> ss) then RETURN (mk_ret x s) else RETURN None"

definition red_dfs_remove :: "'n red_sws \<Rightarrow> 'n \<Rightarrow> 'n red_cycle nres" where
  "red_dfs_remove s x \<equiv> case state s of 
                        None \<Rightarrow> if (x = start s \<and> x \<in> ss) then RETURN (mk_ret x s) else RETURN None 
                      | Some x \<Rightarrow> RETURN (Some x)"

definition red_dfs_invar :: "'n red_sws \<Rightarrow> bool" where
  "red_dfs_invar s \<equiv> state s = None \<longrightarrow> (finished s \<inter> ss \<subseteq> {start s} \<and> 
                                        set (stack s) \<inter> ss \<subseteq> {start s} \<and>
                                        (start s \<in> ss \<longrightarrow> ( (\<forall> x \<in> finished s. start s \<notin> succs x))))"

definition red_cycle_inv :: "'n red_sws \<Rightarrow> bool"where 
  "red_cycle_inv s \<equiv> (case state s of Some (v, vs) \<Rightarrow> (stack s \<noteq> [] \<and> length vs > 1 \<and> v \<in> ss \<and> vwalk vs \<G> \<and> hd vs = start s \<and> last vs = v) 
                                    | None \<Rightarrow> True)"

definition red_dfs :: "('n red_cycle, 'n) dfs_algorithm" where
  "red_dfs = \<lparr> dfs_cond = \<lambda>s. s = None,
                dfs_visit = red_dfs_visit,
                dfs_post = \<lambda> s _. RETURN (state s), 
                dfs_remove = red_dfs_remove,
                dfs_start = \<lambda>_. None,
                dfs_pre_cond = \<lambda>s _. red_dfs_invar s,
                dfs_post_cond = \<lambda>s s' x. (state s' = None \<longrightarrow> stack s \<noteq> [] \<longrightarrow> x \<notin> ss) \<and> 
                                          red_dfs_invar s' \<and> red_cycle_inv s',
                dfs_fe_inv = \<lambda>s x it s'. (state s = None \<longrightarrow> x \<noteq> start s' \<longrightarrow> x \<notin> ss) \<and> 
                                         ((succs x - it) \<inter> ss \<noteq> {} \<longrightarrow> state s' \<noteq> None) \<and>
                                          red_dfs_invar s' \<and> red_cycle_inv s' \<rparr>"

lemma red_dfs_simps[simp]:
  "dfs_cond red_dfs S \<longleftrightarrow> S = None"
  "dfs_visit red_dfs = red_dfs_visit"
  "dfs_post red_dfs s x = RETURN (state s)"
  "dfs_remove red_dfs = red_dfs_remove"
  "dfs_start red_dfs x = None"
  "dfs_pre_cond red_dfs s x \<longleftrightarrow> red_dfs_invar s"
  "dfs_post_cond red_dfs s s' x \<longleftrightarrow> (state s' = None \<longrightarrow> stack s \<noteq> [] \<longrightarrow> x \<notin> ss) \<and> 
                                    red_dfs_invar s' \<and> red_cycle_inv s'"
  "dfs_fe_inv red_dfs s x it s' \<longleftrightarrow> (state s = None \<longrightarrow> x \<noteq> start s' \<longrightarrow> x \<notin> ss) \<and> 
                                    ((succs x - it) \<inter> ss \<noteq> {} \<longrightarrow> state s' \<noteq> None) \<and>
                                    red_dfs_invar s' \<and> red_cycle_inv s'"
unfolding red_dfs_def
by simp_all
end

locale RedDFS_def = RedDFS_pre succs v\<^sub>0 ss +
                    DFS_def succs v\<^sub>0 "red_dfs"
  for succs v\<^sub>0 ss

locale RedDFS = RedDFS_def succs v\<^sub>0 ss +
                finite_succs_digraph succs v\<^sub>0
  for succs v\<^sub>0 and ss :: "'n set"
begin

lemma pre_ss_red_cycle_inv:
  assumes "pre_dfs s x"
  and "stack s \<noteq> []"
  and "x \<in> ss"
  and "S = mk_ret x s"
  shows "red_cycle_inv (s\<lparr>state := S\<rparr>)"
proof -
  from assms[unfolded pre_dfs_unfold] hd_rev have "hd (rev (stack s)) = start s" by force
  moreover from assms have "x \<in> succs (hd (stack s))" by (simp add: pre_dfs_unfold)
  moreover note assms[unfolded pre_dfs_def]
  ultimately show ?thesis using stack_suc_vwalk by (simp add: red_cycle_inv_def)
qed

lemma red_cycle_inv_lift:
  assumes "red_cycle_inv s"
  and "stack s' \<noteq> []"
  and "state s' = state s" "start s' = start s"
  shows "red_cycle_inv s'"
using assms
unfolding red_cycle_inv_def
by (cases "state s") simp_all
  
lemma start_pre_cond:
  "x \<in> V \<Longrightarrow> pre_dfs (dfs_alg_start x) x \<Longrightarrow> dfs_pre_cond red_dfs (dfs_alg_start x) x"
by (simp add: red_dfs_invar_def dfs_alg_start_def finished_def)

lemma remove_post_cond:
  assumes X: "x \<in> discovered s"
  and PRE: "pre_dfs s x"
  and PREC: "dfs_pre_cond red_dfs s x"
  and COND: "dfs_cond red_dfs (state s)"
  and INR: "inres (dfs_remove red_dfs s x) S"
  shows "dfs_post_cond red_dfs s (s\<lparr>state := S\<rparr>) x"
proof (cases "x = start s")
  case False
  from PRE[unfolded pre_dfs_unfold] X have "x \<in> finished s \<or> x \<in> set (stack s)" by blast
  with COND False PREC have "x \<notin> ss" by (auto simp add: red_dfs_invar_def)
  with False assms show ?thesis by (auto simp add: red_cycle_inv_def red_dfs_invar_def red_dfs_remove_def finished_def)
next
  case True  
  from X PRE have "stack s \<noteq> []" by (auto simp add: pre_dfs_unfold)
  with True INR COND pre_ss_red_cycle_inv[OF PRE] PREC
    show ?thesis by (cases "x \<in> ss") 
      (auto simp add: red_cycle_inv_def red_dfs_remove_def red_dfs_invar_def finished_def)
qed

lemma fin_post_cond:
  assumes X: "x \<notin> discovered s"
  and PRE: "pre_dfs s x"
  and PREC: "dfs_pre_cond red_dfs s x"
  and COND: "dfs_cond red_dfs (state s)"
  and INR: "inres (dfs_visit red_dfs s x) S"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x {} s'"
  and FEC: "dfs_fe_inv red_dfs (disc_step_upd x s S) x {} s'"
  and INR': "inres (dfs_post red_dfs s' x) S'"
  and COND': "dfs_cond red_dfs (state s')"
  and POST: "post_dfs s (fin_step_upd x s' S') x"
  shows "dfs_post_cond red_dfs s (fin_step_upd x s' S') x" (is "dfs_post_cond red_dfs s ?sf x")
proof -
  note fin_step_upd_def[simp] disc_step_upd_def[simp]
  have "red_cycle_inv ?sf" using INR' COND' by (simp add: red_cycle_inv_def)
  moreover 
  from FE COND' have start_eq: "start s' = start s" using fe_inv_start_eq by auto
  from FE COND' have "S = None" using fe_inv_state_backtrack by fastforce
  with start_eq have x: "x = start s \<or> x \<notin> ss" using FEC by auto

  from fe_inv_stack_eq[OF FE COND'] have "x = hd (stack s')" "stack s' \<noteq> []" by simp_all
  with FE fin_upd_finished have F': "finished ?sf = insert x (finished s')" by blast

  from X PRE not_discovered_implies_not_finished have "x \<notin> finished s" by blast
  from COND' INR' POST have "stack ?sf = stack s" by fastforce
  hence "red_dfs_invar ?sf"
    apply (simp add: red_dfs_invar_def del: fin_step_upd_def)
  proof (rule impI, intro conjI)
    case goal1 with FEC COND' have "finished s' \<inter> ss \<subseteq> {start s'}" by (simp add: red_dfs_invar_def)
    with F' x start_eq show ?case by auto
  next
    case goal2 show ?case using PREC COND start_eq by (simp add: red_dfs_invar_def)
  next
    case goal3 with F' start_eq show ?case
      apply simp
    proof (rule impI, rule conjI)
      case goal1 thus ?case using FEC COND' by auto
    next
      case goal2 thus ?case using FEC COND' start_eq by (simp add: red_dfs_invar_def)
    qed
  qed
  moreover {
    assume "stack s \<noteq> []"
    hence "x \<noteq> start s" using start_discovered PRE X by auto
    with x have "x \<notin> ss" by simp
  }
  ultimately show ?thesis by simp
qed

lemma abort_post_cond:
  assumes "dfs_fe_inv red_dfs (disc_step_upd x s S) x it s'"
  and "\<not>dfs_cond red_dfs (state s')"
  shows "dfs_post_cond red_dfs s s' x"
using assms by auto

lemma fe_start_fe_inv:
  assumes X:"x \<notin> discovered s"
  and PRE: "pre_dfs s x"
  and PREC: "dfs_pre_cond red_dfs s x"
  and COND: "dfs_cond red_dfs (state s)"
  and INR: "inres (dfs_visit red_dfs s x) S"
  shows "dfs_fe_inv red_dfs (disc_step_upd x s S) x (succs x) (disc_step_upd x s S)"
proof (cases "x = start s")
  case True hence "stack s = []" using X start_discovered PRE by auto
  moreover with PRE[unfolded pre_dfs_unfold] have "finished s = {}" by auto
  moreover from True have "S = None" using INR COND by (auto simp add: red_dfs_visit_def)
  ultimately show ?thesis using True 
    by (simp add: disc_step_upd_def red_dfs_invar_def red_cycle_inv_def finished_def)
next
  case False with PRE not_start_imp_stack have "stack s \<noteq> []" by auto
  show ?thesis
  proof (cases "x \<in> ss")
    case False with PREC COND INR show ?thesis by (auto simp add: disc_step_upd_def red_dfs_visit_def red_dfs_invar_def red_cycle_inv_def finished_def)
  next
    case True with `stack s \<noteq> []` INR COND False have "red_cycle_inv (disc_step_upd x s S)" 
      apply (rule_tac red_cycle_inv_lift)
      apply (rule pre_ss_red_cycle_inv[OF PRE, of S])
      apply (auto simp add: disc_step_upd_def red_dfs_visit_def)
      done
    with True False COND INR show ?thesis 
      by (auto simp add: disc_step_upd_def red_dfs_visit_def red_dfs_invar_def)
  qed
qed

lemma fe_step_pre_cond:
  assumes "dfs_fe_inv red_dfs (disc_step_upd x s S) x it s'"
  shows "dfs_pre_cond red_dfs s' t"
using assms by simp

lemma fe_step_fe_inv:
  notes disc_step_upd_def[simp]
  assumes X: "x \<notin> discovered s"
  and PRE: "pre_dfs s x"
  and PREC: "dfs_pre_cond red_dfs s x"
  and COND: "dfs_cond red_dfs (state s)"
  and INR: "inres (dfs_visit red_dfs s x) S"
  and IT: "it \<subseteq> succs x" "t \<in> it"
  and COND': "dfs_cond red_dfs (state s')"
  and POST: "post_dfs s' s'' t"
  and POSTC: "dfs_post_cond red_dfs s' s'' t"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x it s'"
  and FEC: "dfs_fe_inv red_dfs (disc_step_upd x s S) x it s'"
  and FE': "fe_inv_dfs (disc_step_upd x s S) x (it - {t}) s''" 
  shows "dfs_fe_inv red_dfs (disc_step_upd x s S) x (it - {t}) s''"
proof (cases "state s''")
  case None with FE' have "S = None" using fe_inv_state_backtrack by fastforce
  moreover from FE'[unfolded fe_inv_dfs_unfold] have "start s'' = start s" by simp
  moreover
  from POSTC None have "stack s' = [] \<or> t \<notin> ss" by auto
  with COND' FE[unfolded fe_inv_dfs_unfold] have "t \<notin> ss" by simp
  with FEC COND' have "(succs x - (it - {t})) \<inter> ss = {}" by (auto simp add: red_dfs_invar_def)
  moreover note None INR COND POSTC
  ultimately show ?thesis by (auto simp add: red_cycle_inv_def red_dfs_invar_def red_dfs_visit_def)
next
  case (Some CE) from FE'[unfolded fe_inv_dfs_unfold] have *: "start s'' = start s" by simp
  from Some POSTC have "red_cycle_inv s''" by simp
  with * Some INR COND show ?thesis by (auto simp add: red_dfs_invar_def red_dfs_visit_def)
qed
end

sublocale RedDFS \<subseteq> DFS red_dfs
  apply unfold_locales
    apply (simp add: red_dfs_visit_def)
    apply simp
    apply (simp add: red_dfs_remove_def)
    apply simp
    apply (metis start_pre_cond)
    apply simp
    apply (metis remove_post_cond)
    apply (metis fin_post_cond)
    apply (metis abort_post_cond)
    apply (metis fe_start_fe_inv)
    apply (metis fe_step_pre_cond)
    apply (metis fe_step_fe_inv)
    done

context RedDFS
begin

lemma red_post_complete:
  assumes POST: "post_dfs s s' x"
  and INV: "red_dfs_invar s'"
  and  "stack s' = []"
  and S: "dfs_cond red_dfs (state s')"
  shows "\<not> (\<exists>v. start s \<rightarrow>+ v \<and> v \<in> ss)"
proof (rule notI, erule exE, erule conjE)
  fix v
  assume sv: "start s \<rightarrow>+ v" and "v \<in> ss"
  
  note A = assms[unfolded post_dfs_unfold]

  hence start: "x = start s" by simp
  with sv have xv: "x \<rightarrow>+ v" by simp

  from A INV S have nr: "finished s' \<inter> ss \<subseteq> {start s}" "start s \<in> ss \<Longrightarrow> \<forall>x\<in>finished s'. start s \<notin> succs x" unfolding red_dfs_invar_def by simp_all

  from post_reach_finished_empty_stack assms xv have "v \<in> finished s'" by simp
  thus "False"
  proof (cases "v = x")
    case True with xv `v \<in> ss` `v \<in> finished s'` show "False"
    proof induction
      case (base y) with base nr start show "False" by blast
    next
      case (trans y z) with post_reach_finished_empty_stack[OF assms(1,3-4)] have "y \<in> finished s'" by blast
      with trans nr start show "False" by blast
    qed
  next
    case False 
    from `v \<in> finished s'` `v \<in> ss ` have "v \<in> finished s' \<inter> ss" by auto
    with False start nr show "False" by auto
  qed
qed

theorem red_dfs_sound:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. state s = Some (v,vs) \<longrightarrow> x \<rightarrow>+ v \<and> v \<in> ss)"
proof (rule dfs_algo_SPEC[OF assms], rule impI)
  case (goal1 s) have "x = start (dfs_alg_start x)" unfolding dfs_alg_start_def by simp
  with goal1 have "x = start s" using post_start_eq by auto
  with goal1 show ?case by (auto simp add: red_cycle_inv_def dest: vwalk_start_last_reachable)
qed

corollary red_dfs_sound_Ex:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. state s \<noteq> None \<longrightarrow> (\<exists>v. x \<rightarrow>+v \<and> v \<in> ss))"
proof (rule dfs_algo_SPEC[OF assms], rule impI)
  case (goal1 s) then obtain v vs where *: "state s = Some (v, vs)" by auto
  from goal1 have "x = start (dfs_alg_start x)" unfolding dfs_alg_start_def by simp
  with goal1 have "x = start s" using post_start_eq by auto
  with goal1 * have "x \<rightarrow>+ v \<and> v \<in> ss" by (auto simp add: red_cycle_inv_def dest: vwalk_start_last_reachable)
  thus ?case by blast
qed

theorem red_dfs_complete:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. state s = None \<longrightarrow> \<not> (\<exists>v. x \<rightarrow>+ v \<and> v \<in> ss))"
proof (rule dfs_algo_SPEC[OF assms], rule impI)
  case (goal1 s) hence "post_dfs (dfs_alg_start x) s x" by simp
  moreover from goal1 have "red_dfs_invar s" by simp
  moreover from goal1 have "stack s = stack (dfs_alg_start x)" and "start s = start (dfs_alg_start x)" by auto
  hence "stack s = []" and "x = start s" unfolding dfs_alg_start_def by simp_all
  moreover note goal1(4)
  ultimately show ?case using red_post_complete by (simp add: dfs_alg_start_def)
qed

lemma red_dfs_complete_sound:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. state s \<noteq> None \<longleftrightarrow> (\<exists>v. x\<rightarrow>+v \<and> v \<in> ss))"
  apply (rule SPEC_iff)
  apply (fact red_dfs_sound_Ex[OF assms])
  apply (simp only: not_not)
  apply (fact red_dfs_complete[OF assms])
  done

theorem red_dfs_counter_example_correct:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. state s = Some (v,vs) \<longrightarrow> hd vs = x \<and> v = last vs \<and> vwalk vs \<G> \<and> v \<in> ss \<and> length vs > 1)"
proof (rule dfs_algo_SPEC[OF assms], rule impI)
  case (goal1 s) hence "x = start s" unfolding dfs_alg_start_def post_dfs_unfold by simp
  moreover from goal1 have "red_cycle_inv s" by simp
  ultimately show ?case by (simp add: red_cycle_inv_def goal1(4))
qed
end

locale BlueDFS_pre = NestedDFS_pre succs v\<^sub>0 A
  for succs and v\<^sub>0 :: "'n" and A +
  fixes M :: "(('n, 'n set) blue_res, 'n) dfs_sws \<Rightarrow> 'n set \<Rightarrow> 'n \<Rightarrow> bool"
begin

type_synonym 'x blue_sws = "(('x, 'x set) blue_res, 'x) dfs_sws"

definition succs_restrict :: "'n set \<Rightarrow> 'n \<Rightarrow> 'n set" where
  "succs_restrict R n \<equiv> succs n - R"

lemma succs_restrict_sub:
  "{(v, v'). v' \<in> succs_restrict R v} \<subseteq> {(v, v'). v' \<in> succs v}"
unfolding succs_restrict_def
by auto

abbreviation "red_dfs x R ss \<equiv> RedDFS_pre.red_dfs (succs_restrict R) x ss"
abbreviation "red_dfs_algo R ss x \<equiv> DFS_def.dfs_algo (succs_restrict R) x (red_dfs x R ss) x"

definition run_red_dfs :: "'n set \<Rightarrow> 'n set \<Rightarrow> 'n \<Rightarrow> ('n, 'n set) blue_res nres"
where 
  "run_red_dfs R ss x \<equiv> do { 
               red \<leftarrow> red_dfs_algo R ss x;
               case state red of
                 None \<Rightarrow> RETURN (CYCLE_NO (finished red \<union> R))
               | Some (v,vs) \<Rightarrow> RETURN (CYCLE_PRE v vs)}"

primrec mk_cycle :: "'n list \<Rightarrow> 'n \<Rightarrow> 'n list" where
  "mk_cycle (x#xs) v = (if x = v then x#xs else mk_cycle xs v)"

definition blue_dfs_post :: "'n blue_sws \<Rightarrow> 'n \<Rightarrow> ('n, 'n set) blue_res nres" 
where
  "blue_dfs_post s x \<equiv> if x \<in> A then (case state s of CYCLE_NO R \<Rightarrow>
                         do {
                            S' \<leftarrow> run_red_dfs R {y. M s R y} x;
                            case S' of
                               CYCLE_PRE v vs \<Rightarrow> do {
                                 let rs = rev (stack s);
                                 let cyc = mk_cycle rs v;
                                 RETURN (CYCLE_FIN x v rs vs cyc) }
                             | _ \<Rightarrow> RETURN S'
                         }
                       | _ \<Rightarrow> RETURN (state s))
                       else RETURN (state s)"

definition blue_dfs_invar :: "'n blue_sws \<Rightarrow> bool" where
  "blue_dfs_invar s \<equiv> case state s of CYCLE_NO R \<Rightarrow> 
                           (\<forall>x \<in> A \<inter> finished s. \<not> x \<rightarrow>+ x) \<and> 
                           set (stack s) \<inter> R = {} \<and> 
                           (\<forall>x. x \<notin> discovered s \<longrightarrow> x \<notin> R) \<and>
                           (\<forall>x y. x \<rightarrow>+ y \<longrightarrow> x \<in> R \<longrightarrow> y \<in> R)
                        | _ \<Rightarrow> True"

definition blue_dfs_lasso_inv :: "'n blue_sws \<Rightarrow> bool" where
  "blue_dfs_lasso_inv s \<equiv> case state s of CYCLE_FIN y v ys vs cyc \<Rightarrow> is_lasso_full (start s) y v ys vs cyc
                            | _ \<Rightarrow> True"

definition blue_dfs_pre_cond :: "'n blue_sws \<Rightarrow> 'n \<Rightarrow> bool" where
  "blue_dfs_pre_cond s x \<equiv> case state s of CYCLE_NO R \<Rightarrow> blue_dfs_invar s | _ \<Rightarrow> False"

definition blue_dfs_fe_inv :: "'n blue_sws \<Rightarrow> 'n \<Rightarrow> 'n set \<Rightarrow> 'n blue_sws \<Rightarrow> bool" 
where
  "blue_dfs_fe_inv s x it s' \<equiv> case state s' of
                                CYCLE_NO R \<Rightarrow> blue_dfs_invar s' 
                              | CYCLE_FIN _ _ _ _ _ \<Rightarrow> blue_dfs_lasso_inv s'
                              | CYCLE_PRE _ _  \<Rightarrow> False"

definition blue_dfs_post_cond :: "'n blue_sws \<Rightarrow> 'n blue_sws \<Rightarrow> 'n \<Rightarrow> bool" 
where
  "blue_dfs_post_cond s s' x \<equiv> case state s' of
                                 CYCLE_NO R \<Rightarrow> blue_dfs_invar s' 
                              | CYCLE_FIN _ _ _ _ _ \<Rightarrow> blue_dfs_lasso_inv s'
                              | CYCLE_PRE _ _  \<Rightarrow> False"

definition blue_dfs :: "(('n, 'n set) blue_res, 'n) dfs_algorithm" where
  "blue_dfs = \<lparr> dfs_cond = \<lambda>s. (case s of CYCLE_NO _ \<Rightarrow> True | _ \<Rightarrow> False),
                dfs_visit = \<lambda> s _. RETURN (state s),
                dfs_post = blue_dfs_post,
                dfs_remove = \<lambda> s _. RETURN (state s),
                dfs_start = \<lambda>_. CYCLE_NO {},
                dfs_pre_cond = blue_dfs_pre_cond,
                dfs_post_cond = blue_dfs_post_cond,
                dfs_fe_inv = blue_dfs_fe_inv \<rparr>"

lemma blue_dfs_simps[simp]:
  "dfs_cond blue_dfs S \<longleftrightarrow> (case S of CYCLE_NO _ \<Rightarrow> True | _ \<Rightarrow> False)"
  "dfs_visit blue_dfs s x = RETURN (state s)"
  "dfs_post blue_dfs = blue_dfs_post"
  "dfs_remove blue_dfs s x = RETURN (state s)"
  "dfs_start blue_dfs x = CYCLE_NO {}"
  "dfs_pre_cond blue_dfs = blue_dfs_pre_cond"
  "dfs_post_cond blue_dfs = blue_dfs_post_cond"
  "dfs_fe_inv blue_dfs = blue_dfs_fe_inv"
unfolding blue_dfs_def
by simp_all

lemma blue_dfs_lasso_invI:
  assumes "state s = CYCLE_FIN y v ys vs cyc"
  and "y \<in> A \<or> v \<in> A"
  and "vwalk ys \<G>"
  and "hd ys = start s"
  and "last ys = y"
  and "vwalk vs \<G>"
  and "1 < length vs"
  and "hd vs = y"
  and "last vs = v"
  and "vwalk cyc \<G>"
  and "hd cyc = v"
  and "last cyc = y"
  shows "blue_dfs_lasso_inv s"
using assms 
by (simp add: blue_dfs_lasso_inv_def is_lasso_full_def)
end

locale BlueDFS_def = BlueDFS_pre succs v\<^sub>0 A M +
                     NestedDFS_def "blue_dfs" succs v\<^sub>0 A
  for M and succs v\<^sub>0 and A :: "'n set"

locale BlueDFS = BlueDFS_def M succs v\<^sub>0 A +
                 finite_succs_digraph succs v\<^sub>0
  for M and succs v\<^sub>0 and A :: "'n set"+
  assumes M_hd: "\<lbrakk>common_inv s x; blue_dfs_invar s; stack s \<noteq> []; state s = CYCLE_NO R\<rbrakk> \<Longrightarrow> M s R (hd (stack s))"
  and M_stack: "\<lbrakk>common_inv s x; blue_dfs_invar s; stack s \<noteq> []; state s = CYCLE_NO R; M s R x\<rbrakk> \<Longrightarrow> x \<in> set (stack s)"
begin

lemma interpret_red_dfs:
  assumes "x \<in> V"
  shows "RedDFS (succs_restrict R) x"
proof -
  from assms succs_restrict_sub interpret g!: finite_succs_digraph "succs_restrict R" x
    using finite_succs_subgraphI by blast

  show ?thesis by unfold_locales
qed

lemma run_red_dfs_complete:
  assumes x: "x \<in> V" "x \<notin> R"
  shows "run_red_dfs R ss x \<le> SPEC (\<lambda>s. (case s of CYCLE_NO _ \<Rightarrow> True | _ \<Rightarrow> False) \<longleftrightarrow> \<not>(\<exists>v. x \<rightarrow>\<setminus>R+ v \<and> v \<in> ss))"
proof -
  interpret red: RedDFS "succs_restrict R" x ss
    using interpret_red_dfs `x \<in> V` by metis

  from assms have "x \<in> red.V" by auto 
  note compl = red.red_dfs_complete_sound[OF this]
  have *: "red_dfs_algo R ss x \<le> SPEC (\<lambda>s. state s \<noteq> None \<longleftrightarrow> (\<exists>v. x \<rightarrow>\<setminus>R+ v \<and> v \<in> ss))"
    apply (rule order_trans[OF compl])
    apply (intro refine_vcg)
    apply (metis sub_reachable1_iff_reachable_restrict succs_restrict_def x)
    done
  show ?thesis
    unfolding run_red_dfs_def 
    apply (intro refine_vcg)
    apply (rule order_trans[OF *])
    apply (intro refine_vcg)
    apply simp_all
    done
qed

lemma run_red_dfs_counter_example_correct:
  assumes "x \<in> V" and "x \<notin> R"
  shows "run_red_dfs R ss x \<le> SPEC (\<lambda>s. s = CYCLE_PRE v vs \<longrightarrow> hd vs = x \<and> v = last vs \<and> vwalk vs \<G> \<and> v \<in> ss \<and> length vs > 1 \<and> set vs \<inter> R = {})"
proof -
  interpret red: RedDFS "succs_restrict R" x ss
    using interpret_red_dfs `x \<in> V` by metis

  from assms restrict_succs_digraph_V_restrict succs_restrict_def have redV: "red.V \<subseteq> V - R" by auto
  from subgraph_E_sub succs_restrict_sub `x \<in> V` have redE: "red.E \<subseteq> E" by auto
  
  have vw: "\<And>vs. vwalk vs red.sd_to_graph \<Longrightarrow> vwalk vs \<G> \<and> set vs \<inter> R = {}"
    unfolding vwalk_def
    using rel_restrict_sub[THEN subset_trans, of _ E]
    using redE redV
    by force

  from assms have "x \<in> red.V" by auto 
  note compl = red.red_dfs_counter_example_correct[OF this]
  have *: "red_dfs_algo R ss x \<le> SPEC (\<lambda>s. state s = Some (v,vs) \<longrightarrow> hd vs = x \<and> v = last vs \<and> vwalk vs \<G> \<and> v \<in> ss \<and> length vs > 1 \<and> set vs \<inter> R = {})"
    apply (rule order_trans[OF compl])
    apply (intro refine_vcg)
    apply (intro impI conjI)
      apply blast
      apply (auto simp add: vw)
  done
  thus ?thesis unfolding run_red_dfs_def
    apply (intro refine_vcg)
    apply (rule order_trans[OF *])
    apply (intro refine_vcg)
    apply simp_all
    done
qed

lemma run_red_dfs_finished:
  assumes x: "x \<in> V" "x \<notin> R"
  shows "run_red_dfs R ss x \<le> SPEC (\<lambda>s. s = CYCLE_NO f \<longrightarrow> f = {z. z = x \<or> x \<rightarrow>\<setminus>R+ z} \<union> R)"
proof -
  interpret red: RedDFS "succs_restrict R" x ss
    using interpret_red_dfs `x \<in> V` by metis
  
  have "x \<in> red.V" by simp
  note post = red.dfs_post[OF this]
  have *: "red_dfs_algo R ss x \<le> SPEC (\<lambda>s. state s = None \<longrightarrow> finished s = {z. z = x \<or> x \<rightarrow>\<setminus>R+ z})"
    apply (rule order_trans[OF post])
    apply (intro refine_vcg)
  proof (erule conjE, rule impI)
    case (goal1 s) hence "stack s = stack (red.dfs_alg_start x)" by (simp add: red.post_dfs_unfold)
    hence "stack s = []" by (simp add: red.dfs_alg_start_def)
    with red.post_empty_stack_finished_eq goal1 have "finished s = {z. x = z \<or> red.sd_reachable1 x z}" by simp
    thus ?case using sub_reachable1_iff_reachable_restrict succs_restrict_def x by auto
  qed
  thus ?thesis unfolding run_red_dfs_def
    apply (intro refine_vcg)
    apply (rule order_trans[OF *])
    apply (intro refine_vcg)
    apply simp_all
    done
qed

lemma run_red_dfs_NF:
  "x \<in> V \<Longrightarrow> x \<notin> R \<Longrightarrow> nofail (run_red_dfs R ss x)"
by (metis (lifting) run_red_dfs_finished SPEC_nofail)

lemma red_dfs_algo_NF:
  "x \<in> V \<Longrightarrow> x \<notin> R \<Longrightarrow> nofail (red_dfs_algo R ss x)"
using run_red_dfs_NF
by (simp add: refine_pw_simps run_red_dfs_def)

lemma mk_cycle_vwalk':
  assumes "distinct ls"
  and "vwalk ls \<G>"
  and "v \<in> set ls"
  and "last ls = x"
  and "mk_cycle ls v = xs"
  shows "vwalk xs \<G> \<and> hd xs = v \<and> last xs = x"
using assms
proof (induction ls arbitrary: xs)
  case (Cons l ls xs)
  show ?case
  proof (cases "v = l")
    case True with Cons.prems have *: "xs = l#ls" by simp
    show ?thesis
    proof (intro conjI)
      from Cons.prems * show "vwalk xs \<G>" by simp
      from * True show "hd xs = v" by simp
      from * Cons.prems show "last xs = x" by blast
    qed
  next
    case False with Cons.prems have "xs = mk_cycle ls v" by simp
    moreover from Cons.prems False have "v \<in> set ls" by auto
    moreover hence "ls \<noteq> []" by auto
    with Cons.prems False have "distinct ls" "last ls = x" by simp_all
    moreover from Cons.prems `ls \<noteq> []` have "vwalk ls \<G>" unfolding vwalk_def by simp
    ultimately show ?thesis using Cons by blast
  qed
qed simp

lemma mk_cycle_vwalk:
  assumes C: "common_inv s x"
  and xhd: "x = hd (stack s)"
  and "v \<in> set (stack s)"
  and MK: "mk_cycle (rev (stack s)) v = xs"
  shows "vwalk xs \<G> \<and> hd xs = v \<and> last xs = x"
proof -
  from assms have ne: "stack s \<noteq> []" by auto
  with C[unfolded common_inv_def] have "distinct (rev (stack s))" "vwalk (rev (stack s)) \<G>" by auto
  moreover from xhd last_rev[OF ne] have "last (rev (stack s)) = x" by metis
  moreover from assms have "v \<in> set (rev (stack s))" by auto
  ultimately show ?thesis using mk_cycle_vwalk'[where ls = "rev (stack s)"] MK by blast
qed

lemma blue_dfs_fe_x_not_R:
  assumes "state s' = CYCLE_NO R'"
  and  "blue_dfs_pre_cond s x" 
  and "fe_inv_dfs (disc_step_upd x s S) x {} s'" (is "fe_inv_dfs ?sd x {} s'")
  and "blue_dfs_fe_inv (disc_step_upd x s S) x {} s'"
  shows "x \<notin> R'"
proof -
  from assms have "stack s' = stack ?sd" by force
  with assms show "x \<notin> R'" by (simp add: blue_dfs_pre_cond_def blue_dfs_fe_inv_def blue_dfs_invar_def disc_step_upd_def)
qed

lemma blue_dfs_post_NF: 
  assumes "pre_dfs s x" 
  and "blue_dfs_pre_cond s x" 
  and S: "state s = CYCLE_NO R" "state s' = CYCLE_NO R'"
  and "fe_inv_dfs (disc_step_upd x s S) x {} s'"
  and "blue_dfs_fe_inv (disc_step_upd x s S) x {} s'"
  shows "nofail (blue_dfs_post s' x)"
proof -
  from assms have "x \<in> V" by blast
  moreover from assms blue_dfs_fe_x_not_R have "x \<notin> R'" by blast
  ultimately show ?thesis using S run_red_dfs_NF by (auto simp add: refine_pw_simps blue_dfs_post_def Let_def split: blue_res.splits)
qed

lemma pre_start:
  assumes "x \<in> V" and "pre_dfs (dfs_alg_start x) x"
  shows "blue_dfs_pre_cond (dfs_alg_start x) x"
unfolding blue_dfs_pre_cond_def blue_dfs_invar_def dfs_alg_start_def
by (simp add: finished_def)

lemma discovered_pre_post:
  assumes "x \<in> discovered s"
  and "state s = CYCLE_NO R"
  and "blue_dfs_pre_cond s x"
  shows "blue_dfs_post_cond s s x"
using assms
unfolding blue_dfs_pre_cond_def blue_dfs_post_cond_def blue_dfs_lasso_inv_def
by simp_all

lemma CYCLE_NO_inres_run_red_dfs:
  assumes "x \<in> V" "x \<notin> R" "x \<in> A"
  and "inres (blue_dfs_post s x) (CYCLE_NO R')"
  and "state s = CYCLE_NO R"
  shows "inres (run_red_dfs R {y. M s R y} x) (CYCLE_NO R')"
using assms
unfolding blue_dfs_post_def
  apply (simp add: refine_pw_simps run_red_dfs_NF)
  apply (erule exE)
  apply (auto simp add: Let_def split: blue_res.splits)
done

lemma CYCLE_PRE_not_inres_blue_dfs_post:
  assumes "x \<in> V" "x \<notin> R" "x \<in> A"
  and "inres (blue_dfs_post s x) (CYCLE_PRE v vs)"
  and "state s = CYCLE_NO R"
  shows "False"
using assms
unfolding blue_dfs_post_def 
  apply (simp add: refine_pw_simps run_red_dfs_NF Let_def split: blue_res.splits)
  apply (metis blue_res.exhaust)
done

lemma CYCLE_FIN_not_inres_run_red_dfs:
  assumes "x \<in> V" "x \<notin> R" "x \<in> A"
  shows "\<not> inres (run_red_dfs R {y. M s R y} x) (CYCLE_FIN y v ys vs cyc)"
using assms
by (simp add: refine_pw_simps run_red_dfs_def red_dfs_algo_NF split: option.split)

lemma CYCLE_PRE_inres_run_red_dfs:
  assumes "x \<in> V" "x \<notin> R" "x \<in> A"
  and "inres (blue_dfs_post s x) (CYCLE_FIN y v xs vs cyc)"
  and "state s = CYCLE_NO R"
  shows "inres (run_red_dfs R {y. M s R y} x) (CYCLE_PRE v vs)"
using assms
unfolding blue_dfs_post_def
  apply (simp add: refine_pw_simps run_red_dfs_NF)
  apply (erule exE)
  apply (auto simp add: CYCLE_FIN_not_inres_run_red_dfs Let_def split: blue_res.splits)
done

lemma CYCLE_FIN_properties:
  assumes "x \<in> V" "x \<notin> R" "x \<in> A"
  and "inres (blue_dfs_post s x) (CYCLE_FIN y v xs vs cyc)"
  and "state s = CYCLE_NO R"
  shows "xs = rev (stack s) \<and> cyc = mk_cycle (rev (stack s)) v \<and> y = x"
using assms
unfolding blue_dfs_post_def
  apply (simp add: refine_pw_simps run_red_dfs_NF)
  apply (erule exE)
  apply (auto simp add: CYCLE_FIN_not_inres_run_red_dfs Let_def split: blue_res.splits)
done

lemma post_fin_step:
  assumes POST: "post_dfs s (fin_step_upd x s' S') x"
  and ND: "x \<notin> discovered s"
  and PRE: "pre_dfs s x"
  and PREC: "blue_dfs_pre_cond s x"
  and S: "state s = CYCLE_NO R" "state s' = CYCLE_NO R'"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x {} s'" (is "fe_inv_dfs ?sd x {} s'")
  and FEC: "blue_dfs_fe_inv (disc_step_upd x s S) x {} s'"
  and INR: "inres (blue_dfs_post s' x) S'"
  shows "blue_dfs_post_cond s (fin_step_upd x s' S') x" (is "blue_dfs_post_cond s ?sf x")
proof -
  note FEC' = FEC[unfolded blue_dfs_fe_inv_def blue_dfs_invar_def]
  note FE' = FE[unfolded fe_inv_dfs_unfold]
  note [simp] = disc_step_upd_def fin_step_upd_def
  
  from FEC S have inv': "blue_dfs_invar s'" by (simp add: blue_dfs_fe_inv_def)

  from PRE have xV: "x \<in> V" by blast
  from assms blue_dfs_fe_x_not_R have xR': "x \<notin> R'" by simp
  
  from FE' S have stack_eq: "stack s' = stack ?sd" and s'_ne: "stack s' \<noteq> []" by auto
  hence xhd: "x = hd (stack s')" by simp
  with fin_upd_finished FE s'_ne have Ff: "finished ?sf = insert x (finished s')" by auto

  have tl_stack_reach:
    "\<And>y. y \<in> set (stack s') \<Longrightarrow> y \<noteq> x \<Longrightarrow> y \<rightarrow>+ x"
  proof -
    fix y
    assume y_stack: "y \<in> set (stack s')" and "y \<noteq> x"
    with stack_eq have "y \<in> set (tl (stack s'))" by auto
    moreover from y_stack stack_subset_verts FE have "y \<in> V" by auto
    ultimately show "y \<rightarrow>+ x" 
      using FE xhd stack_tl_reachable_stack_hd[OF common_inv_change, OF s'_ne `y \<in> V`] 
      by blast
  qed
  
  show ?thesis
  proof (cases "S'")
    case (CYCLE_NO R'')
    have "blue_dfs_invar ?sf"
    proof (cases "x \<in> A")
      case False 
      moreover with S INR CYCLE_NO have "R'' = R'" unfolding blue_dfs_post_def by simp
      moreover from False Ff have "finished ?sf \<inter> A = finished s' \<inter> A" by auto
      moreover from S FEC' have "set (stack s') \<inter> R' = {}" by simp
      moreover note FEC' S CYCLE_NO s'_ne
      ultimately show ?thesis by (auto dest: tl_subset simp: blue_dfs_invar_def)
    next
      case True with S CYCLE_NO INR xV xR' have inr: "inres (run_red_dfs R' {y. M s' R' y} x) (CYCLE_NO R'')" by (simp add: CYCLE_NO_inres_run_red_dfs)
      with CYCLE_NO have R'': "R'' = {z. x = z \<or> x \<rightarrow>\<setminus>R'+ z} \<union> R'" by (auto dest: order_trans[OF _ run_red_dfs_finished[OF xV xR', of _ R'']] simp: inres_def)
      
      from inr CYCLE_NO have no_cycle: "\<not> (\<exists>v. x \<rightarrow>\<setminus>R'+ v \<and> M s' R' v)" by (auto dest: order_trans[OF _ run_red_dfs_complete[OF xV xR']] simp: inres_def)
      moreover from stack_eq M_hd FE inv' S have "M s' R' x" by force
      ultimately have no_x_self: "\<not> x \<rightarrow>\<setminus>R'+ x" by blast
      
      have no_x_stack: "\<And>y. x \<rightarrow>\<setminus>R'+ y \<Longrightarrow> y \<notin> set (stack s')"
      proof (rule notI)
        fix y
        assume x_y: "x \<rightarrow>\<setminus>R'+ y" and y_stack: "y \<in> set (stack s')"
        with FE have "y \<in> V" using stack_subset_verts by auto

        from x_y no_cycle have "\<not> M s' R' y" by simp
        with `M s' R' x` have "x \<noteq> y" by blast
        with tl_stack_reach y_stack have "y \<rightarrow>+ x" by blast
        moreover from y_stack FEC' S have "y \<notin> R'" by auto
        ultimately have "y \<rightarrow>\<setminus>R'+ x" using reach_impl_restr_reach FEC' S xR' by auto
        with x_y no_x_self restr_reachable1_trans show "False" by blast
      qed
      
      from CYCLE_NO show ?thesis
        unfolding blue_dfs_invar_def
      proof (simp, intro conjI)
        case goal1
        have "\<not> x \<rightarrow>+ x" using xR' no_x_self reach_impl_restr_reach FEC' S by auto
        with Ff CYCLE_NO no_cycle FEC' S True show ?case by simp
      next
        case goal2
        {
          fix y
          assume A: "y \<in> set (tl (stack s'))" "y \<in> R''"
          with R'' have "x = y \<or> x \<rightarrow>\<setminus>R'+ y \<or> y \<in> R'" by auto
          hence "False"
          proof (elim disjE)
            case goal1
            from FE' have "distinct (stack s')" by simp
            with xhd distinct_hd_tl have "x \<notin> set (tl (stack s'))" by metis
            with goal1 A show "False" by blast
          next
            case goal2 with A no_x_stack show "False" using tl_subset s'_ne by auto
          next
            case goal3
            from FEC' S xR' have "set (stack s') \<inter> R' = {}" by auto
            with s'_ne tl_subset have "set (tl (stack s')) \<inter> R' = {}" by auto
            with A goal3 show "False" by auto
          qed
        }
        thus ?case by auto
      next
        case goal3
        show ?case
        proof (rule allI, rule impI)
          fix y
          assume ndisc: "y \<notin> discovered s'"
          thus "y \<notin> R''"
          proof (rule contrapos_nn)
            assume "y \<in> R''"
            with R'' have *: "y = x \<or> x \<rightarrow>\<setminus>R'+ y \<or> y \<in> R'" by auto
            moreover have "y \<notin> R'" using ndisc FEC' S by simp
            moreover have "y \<notin> set (stack s')" using ndisc FE stack_subset_discovered by auto
            with xhd s'_ne hd_in_set have "x \<noteq> y" by auto
            ultimately have " x \<rightarrow>\<setminus>R'+ y" by auto
            thus "y \<in> discovered s'"
            proof induction
              case base with Ff post_finished_implies_succs_discovered[OF POST] CYCLE_NO show ?case by auto
            next
              case (trans y z) with discovered_subset_verts POST have "y \<in> V" by auto
              from FE' S stack_eq have F: "\<forall>y\<in>finished s'. \<forall>z. y \<rightarrow>\<setminus>set (stack s')+ z \<longrightarrow> z \<in> finished s'" by simp
              
              from trans no_x_stack have "y \<notin> set (stack s')" by simp
              with trans discovered_non_stack_implies_finished[OF common_inv_change, OF s'_ne `y \<in> V`] FE have "y \<in> finished s'" by blast
              with CI_finished_implies_succs_discovered[OF common_inv_change, OF s'_ne `y \<in> V` _ F] trans FE show ?case by blast
            qed
          qed
        qed
      next
        case goal4
        show ?case
        proof (intro allI impI)
          fix y z
          assume reach: "y \<rightarrow>+ z" and inr: "y \<in> R''"
          with R'' have "y \<in> {z. x = z \<or> x \<rightarrow>\<setminus>R'+ z} \<or> y \<in> R'" by simp
          thus "z \<in> R''"
          proof (rule disjE)
            assume "y \<in> R'" with reach FEC' S have "z \<in> R'" by simp
            with R'' show ?thesis by simp
          next
            assume A: "y \<in> {z. x = z \<or> x \<rightarrow>\<setminus>R'+ z}"
            with reach no_cycle `x \<notin> R'` have y_notin: "y \<notin> R'" using restricted_target by blast
            thus ?thesis using no_cycle R''
            proof (cases "z \<in> R'")
              case False with reach_impl_restr_reach FEC' S y_notin reach have "y \<rightarrow>\<setminus>R'+ z" by fastforce
              with A no_cycle restr_reachable1_trans have "x \<rightarrow>\<setminus>R'+ z" by blast
              with no_cycle R'' show ?thesis by blast
            qed simp
          qed
        qed
      qed
    qed
    with CYCLE_NO show ?thesis unfolding blue_dfs_post_cond_def by simp
  next
    case CYCLE_PRE with INR S CYCLE_PRE_not_inres_blue_dfs_post[OF xV xR', of s'] have "False" 
      by (cases "x \<in> A") (blast, simp add: blue_dfs_post_def)
    thus ?thesis ..
  next
    case (CYCLE_FIN y v ys vs cyc) with INR S have "blue_dfs_lasso_inv ?sf"
    proof (cases "x \<in> A")
      case True with CYCLE_FIN xV xR' S INR have ys: "ys = rev (stack s')" and cyc: "cyc = mk_cycle (rev (stack s')) v" and "y = x" using CYCLE_FIN_properties[where s=s'] by blast+
      with FE mk_cycle_vwalk xhd have "v \<in> set (stack s') \<Longrightarrow> vwalk cyc \<G> \<and> hd cyc = v \<and> last cyc = x" by blast

      moreover from xhd ys FE' s'_ne have "vwalk ys \<G>" and "last ys = x" using last_rev by auto
      moreover from FE' s'_ne have "last (stack s') = start s'" by simp
      with hd_rev[OF s'_ne] ys have "hd ys = start s'" by simp

      moreover
      from True S CYCLE_FIN INR xV xR' have inr: "inres (run_red_dfs R' {y. M s' R' y} x) (CYCLE_PRE v vs)" by (simp add: CYCLE_PRE_inres_run_red_dfs)
      with CYCLE_FIN  have pre: "x = hd vs \<and> v = last vs \<and> vwalk vs \<G> \<and> M s' R' v \<and> length vs > 1 \<and> set vs \<inter> R' = {}"
        unfolding inres_def
        by (auto dest!: order_trans[OF _ run_red_dfs_counter_example_correct[OF xV xR', of "{y. M s' R' y}" v vs]] simp: inres_def)
      moreover
      hence "v \<in> V" unfolding vwalk_def by auto
      with  M_stack[OF common_inv_change] s'_ne FE pre inv' S have "v \<in> set (stack s')" by blast

      ultimately 
      show ?thesis using CYCLE_FIN True `y = x`
        unfolding blue_dfs_lasso_inv_def
        by (simp add: is_lasso_full_def)
    qed (simp add: blue_dfs_post_def)
    with CYCLE_FIN show ?thesis unfolding blue_dfs_post_cond_def by simp
  qed
qed

lemma fe_inv_post:
  assumes "blue_dfs_fe_inv s'' x it s'"
  shows "blue_dfs_post_cond s s' x"
using assms
by (simp add: blue_dfs_fe_inv_def blue_dfs_post_cond_def)

lemma pre_fe_inv:
  assumes "S = CYCLE_NO R" "state s = CYCLE_NO R"
  and "x \<notin> discovered s"
  and "blue_dfs_pre_cond s x"
  shows "blue_dfs_fe_inv (disc_step_upd x s S) x (succs x) (disc_step_upd x s S)"
using assms
unfolding blue_dfs_pre_cond_def blue_dfs_fe_inv_def disc_step_upd_def blue_dfs_invar_def
by (simp add: finished_def)

lemma fe_inv_pre:
  assumes "blue_dfs_fe_inv s x it s'"
  and "state s' = CYCLE_NO R'"
  shows "blue_dfs_pre_cond s' t"
using assms
unfolding blue_dfs_pre_cond_def blue_dfs_fe_inv_def blue_dfs_invar_def
by simp

lemma post_fe_inv:
  assumes "blue_dfs_post_cond s' s'' t"
  and "state s' = CYCLE_NO R'"
  shows "blue_dfs_fe_inv s x (it - {t}) s''"
using assms
unfolding blue_dfs_fe_inv_def blue_dfs_post_cond_def blue_dfs_invar_def
by (simp split: blue_res.splits )
end

sublocale BlueDFS \<subseteq> DFS blue_dfs
  apply unfold_locales
  apply simp
  apply simp
  apply (simp add: blue_dfs_post_NF split: blue_res.splits)
  apply (simp add: pre_start)
  apply simp
  apply (clarsimp, simp add: discovered_pre_post split: blue_res.splits) []
  apply (simp add: post_fin_step split: blue_res.splits)
  apply (simp add: fe_inv_post)
  apply (clarsimp,simp add: pre_fe_inv split: blue_res.splits) []
  apply (simp add: fe_inv_pre split: blue_res.splits)
  apply (simp add: post_fe_inv split: blue_res.splits)
done

context BlueDFS
begin

lemma blue_dfs_not_CYCLE_PRE:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. \<not>(\<exists>a b. state s = CYCLE_PRE a b))"
using assms
by (auto intro!: dfs_algo_SPEC simp add: blue_dfs_post_cond_def)

lemma blue_dfs_sound:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. \<forall>y v ys vs cyc. (state s = CYCLE_FIN y v ys vs cyc \<longrightarrow> is_lasso_full x y v ys vs cyc))"
proof (rule dfs_algo_SPEC[OF assms], rule impI allI)
  case (goal1 s) have "x = start (dfs_alg_start x)" unfolding dfs_alg_start_def by simp
  with goal1 have "x = start s" using post_start_eq by force
  with goal1 show ?case by (fastforce simp add: blue_dfs_lasso_inv_def blue_dfs_post_cond_def)
qed

lemma blue_dfs_complete:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. (\<exists>R. state s = CYCLE_NO R) \<longrightarrow> \<not>(\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> A \<and> v \<rightarrow>+ v))"
proof (rule dfs_algo_SPEC[OF assms])
  case goal1
  show ?case
    apply (rule ccontr)
    apply clarify
  proof -
    fix v R
    assume reach: "x \<rightarrow>\<star> v" and A: "v \<in> A" and v_reach: "v \<rightarrow>+ v" and s: "state s = CYCLE_NO R"
    from goal1 have "blue_dfs_post_cond (dfs_alg_start x) s x" (is "blue_dfs_post_cond ?s s x") by simp
    with goal1 s A v_reach have "v \<in> finished s \<Longrightarrow> False" unfolding blue_dfs_post_cond_def blue_dfs_invar_def by simp

    moreover
    have "v \<in> finished s"
    proof -
      from s goal1 have "stack s = stack ?s" by auto
      hence "stack s = []" unfolding dfs_alg_start_def by simp
      with s goal1 have "finished s = {v. x = v \<or> x \<rightarrow>+ v}" using post_empty_stack_finished_eq by simp
      thus ?thesis using reach by auto
    qed
    
    ultimately show "False" .
  qed
qed
end

sublocale BlueDFS \<subseteq> NestedDFS blue_dfs
  using blue_dfs_sound blue_dfs_complete blue_dfs_not_CYCLE_PRE
  by unfold_locales

definition "courcoubetisM \<equiv> \<lambda>s R x. x = hd (stack s)"

locale Courcoubetis_def = BlueDFS_def courcoubetisM
locale Courcoubetis = Courcoubetis_def succs v\<^sub>0 +
                      finite_succs_digraph succs v\<^sub>0
  for succs v\<^sub>0
begin

lemma courcoubetisM_correct:
  "stack s \<noteq> [] \<Longrightarrow> courcoubetisM s R (hd (stack s))"
  "stack s \<noteq> [] \<Longrightarrow> \<forall>x. courcoubetisM s R x \<longrightarrow> x \<in> set (stack s)"
unfolding courcoubetisM_def
by simp_all
end

sublocale Courcoubetis \<subseteq> BlueDFS courcoubetisM
by unfold_locales (simp_all add: courcoubetisM_correct)

definition "hpyM \<equiv> \<lambda>s R x. x \<in> set (stack s)"

locale HPY_def = BlueDFS_def hpyM
locale HPY = HPY_def succs v\<^sub>0 +
             finite_succs_digraph succs v\<^sub>0
  for succs v\<^sub>0
begin

lemma hpyM_correct:
  assumes "blue_dfs_invar s"
  and "stack s \<noteq> []"
  and "state s = CYCLE_NO R"
 shows
  "hpyM s R (hd (stack s))"
  "\<And>x. hpyM s R x \<Longrightarrow> x \<in> set (stack s)"
using assms blue_dfs_invar_def
unfolding hpyM_def
by simp_all
end

sublocale HPY \<subseteq> BlueDFS hpyM
by unfold_locales (auto intro: hpyM_correct)

end
