theory Nested_DFS_Impl
imports 
  Nested_DFS DFS_Impl
  "../LTL_MC/Lasso/NDFS_Statistics"
begin

locale Accpt =
  fixes accpt :: "'n \<Rightarrow> bool"

locale NestedDFS_Impl_pre = set!: StdSetDefs sops +
                            Graph_succ_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 +
                            Accpt accpt +
                            NestedDFS_pre \<alpha>succs v\<^sub>0 "Collect accpt"
  for sops :: "('n, 's, _) set_ops_scheme"
  and \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma>"
  and v\<^sub>0 \<gamma>\<^sub>0 accpt

locale NestedDFS_Impl_def = NestedDFS_Impl_pre sops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 accpt +
                            NestedDFS_def dfs \<alpha>succs v\<^sub>0 "Collect accpt" +
                            DFS_Impl_def \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 dfs sops \<alpha>\<^sub>\<sigma> impl_dfs dfs_impl_inv \<gamma>_inv
  for dfs :: "(('n, 'n set) blue_res, 'n) dfs_algorithm"
  and impl_dfs :: "('S, 'n, 's, '\<gamma>) dfs_algorithm_impl"
  and dfs_impl_inv
  and \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> ('n, 'n set) blue_res"
  and sops :: "('n, 's, _) set_ops_scheme"
  and \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma>"
  and v\<^sub>0 \<gamma>\<^sub>0 accpt \<gamma>_inv +
  fixes ext :: "'S \<Rightarrow> ('n, 'B) blue_res"
begin

definition nested_dfs_impl where
  "nested_dfs_impl x = do {
                     NDFS_Statistics.start_nres;
                     s \<leftarrow> dfs_algo_impl x;
                     NDFS_Statistics.stop_nres;
                     case ext (istate s) of
                     CYCLE_FIN y v ys vs cyc \<Rightarrow> if accpt y then RETURN (LASSO y ys (vs \<oplus> cyc))
                                               else RETURN (LASSO v (ys \<oplus> vs) (cyc \<oplus> vs))
                    | _ \<Rightarrow> RETURN NO_LASSO }"

schematic_lemma nested_dfs_impl_aux:
  "RETURN (?code x) \<le> nested_dfs_impl x"
  unfolding nested_dfs_impl_def
  by (refine_transfer dfs_code.refine)

concrete_definition nested_dfs_code for x uses nested_dfs_impl_aux
end

fun blue_res\<alpha> :: "('r \<Rightarrow> 'n set) \<Rightarrow> ('n, 'r) blue_res \<Rightarrow> ('n, 'n set) blue_res" where
  "blue_res\<alpha> \<alpha> (CYCLE_NO R) = CYCLE_NO (\<alpha> R)"
| "blue_res\<alpha> \<alpha> (CYCLE_PRE v vs) = CYCLE_PRE v vs"
| "blue_res\<alpha> \<alpha> (CYCLE_FIN a b c d e) = CYCLE_FIN a b c d e"

locale NestedDFS_Impl = DFS_Impl dfs impl_dfs dfs_impl_inv \<alpha>\<^sub>\<sigma> sops v\<^sub>0 \<gamma>succs \<gamma>\<^sub>0 \<gamma>_inv +
                        NestedDFS_Impl_def dfs impl_dfs dfs_impl_inv \<alpha>\<^sub>\<sigma> sops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 accpt \<gamma>_inv ext +
                        NestedDFS dfs \<alpha>succs v\<^sub>0 "Collect accpt" 
  for dfs :: "(('n, 'n set) blue_res, 'n) dfs_algorithm"
  and impl_dfs :: "('S, 'n, 's, '\<gamma>) dfs_algorithm_impl"
  and dfs_impl_inv
  and \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> ('n, 'n set) blue_res"
  and ext :: "'S \<Rightarrow> ('n, 'B) blue_res"
  and sops :: "('n, 's, _) set_ops_scheme"
  and \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma>"
  and v\<^sub>0 \<gamma>\<^sub>0 accpt \<gamma>_inv +
  fixes B\<alpha> :: "'B \<Rightarrow> 'n set"
  assumes ext_correct: "\<And>si s. (si,s) \<in> dfsbr \<Longrightarrow> blue_res\<alpha> B\<alpha> (ext (istate si)) = state s"
begin

definition resbr where
  "resbr = br (blue_res\<alpha> B\<alpha>) (\<lambda>_. True)"

lemma ext_correct_resbr:
  "(si,s) \<in> dfsbr \<Longrightarrow> (ext (istate si), state s) \<in> resbr"
unfolding resbr_def refine_rel_defs
using ext_correct
by auto

lemma blue_res_split_refine_resbr [refine]:
  assumes "(xi, x) \<in> resbr"
  assumes "\<And> r ri. \<lbrakk>xi = CYCLE_NO ri; x = CYCLE_NO r; r = B\<alpha> ri\<rbrakk> \<Longrightarrow> fnci ri \<le> \<Down>R (fnc r)"
  assumes "\<And> ni nli n nl. \<lbrakk>xi = CYCLE_PRE ni nli; x = CYCLE_PRE n nl; (ni, n) \<in> Id; (nli, nl) \<in> Id\<rbrakk> \<Longrightarrow> fcpi ni nli \<le>\<Down>R (fcp n nl)"
  assumes "\<And> yi vi ysi vsi cyci y v ys vs cyc. \<lbrakk> xi = CYCLE_FIN yi vi ysi vsi cyci; x = CYCLE_FIN y v ys vs cyc;
           (yi, y) \<in> Id; (vi, v) \<in> Id; (ysi, ys) \<in> Id; (vsi, vs) \<in> Id; (cyci, cyc) \<in> Id \<rbrakk> \<Longrightarrow>
           fcfi yi vi ysi vsi cyci \<le> \<Down>R (fcf y v ys vs cyc)"
  shows "blue_res_case fnci fcpi fcfi xi \<le> \<Down>R (blue_res_case fnc fcp fcf x)"
  using assms
  unfolding resbr_def refine_rel_defs
  by (force split: blue_res.split)

context 
  fixes x
  assumes "x \<in> V"
begin

lemma nested_dfs_impl_refine:
  "nested_dfs_impl x \<le>\<Down>Id (nested_dfs x)"
unfolding nested_dfs_impl_def nested_dfs_def
  apply (simp del: Refine_Basic.conc_Id)
  apply (refine_rcg dfs_algo_impl_refine[OF assms] ext_correct_resbr)
  apply (auto simp add: `x \<in> V`)
  done

lemma nested_dfs_code_refine:
  "RETURN (nested_dfs_code x) \<le> nested_dfs x"
proof -
  note nested_dfs_code.refine
  also note nested_dfs_impl_refine
  finally show ?thesis .
qed

lemma nested_dfs_impl_lift:
  assumes "nested_dfs x \<le> SPEC \<Phi>"
  shows "nested_dfs_impl x \<le> SPEC \<Phi>"
proof -
  note nested_dfs_impl_refine
  also note assms
  finally show "nested_dfs_impl x \<le> SPEC \<Phi>" .
qed

(* Correctness lemmas *)
lemmas nested_dfs_impl_sound = nested_dfs_sound[OF `x\<in>V`, THEN nested_dfs_impl_lift]

lemma nested_dfs_code_sound:
  assumes "res = nested_dfs_code x"
  and "res = LASSO y to_y cyc"
  shows "is_lasso x y to_y cyc"
proof -
  note nested_dfs_code.refine
  also note nested_dfs_impl_sound
  finally show ?thesis using assms by (auto elim!: RETURN_ref_SPECD)
qed

lemmas nested_dfs_impl_sound_simple = nested_dfs_sound_simple[OF `x\<in>V`, THEN nested_dfs_impl_lift]
lemma nested_dfs_code_sound_simple:
  assumes "res = nested_dfs_code x"
  and "res = LASSO y to_y cyc"
  shows "accpt y \<and> x \<rightarrow>\<star> y \<and> y \<rightarrow>+ y"
proof -
  note nested_dfs_code.refine
  also note nested_dfs_impl_sound_simple
  finally show ?thesis using assms by (force elim!: RETURN_ref_SPECD)
qed

lemma nested_dfs_impl_sound_alt:
  assumes "\<not>(\<exists>v. x \<rightarrow>\<star>v \<and> accpt v \<and> v \<rightarrow>+ v)"
  shows "nested_dfs_impl x \<le> SPEC (\<lambda>res. res = NO_LASSO)"
proof -
  note nested_dfs_impl_refine
  also from assms have "\<not>(\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> Collect accpt \<and> v \<rightarrow>+ v)" by auto
  note nested_dfs_sound_alt[OF `x \<in> V` this]
  finally show ?thesis .
qed

lemma nested_dfs_code_sound_alt:
  assumes "\<not>(\<exists>v. x \<rightarrow>\<star>v \<and> accpt v \<and> v \<rightarrow>+ v)"
  and "res = nested_dfs_code x"
  shows "res = NO_LASSO"
proof -
  note nested_dfs_code.refine
  also note nested_dfs_impl_sound_alt[OF assms(1)]
  finally show ?thesis using assms by (auto elim!: RETURN_ref_SPECD)
qed

lemmas nested_dfs_impl_complete = nested_dfs_complete[OF `x\<in>V`, THEN nested_dfs_impl_lift]
lemma nested_dfs_code_complete:
  assumes "res = nested_dfs_code x"
  and "res = NO_LASSO"
  shows " \<not>(\<exists>v. x \<rightarrow>\<star>v \<and> accpt v \<and> v \<rightarrow>+ v)"
proof -
  note nested_dfs_code.refine
  also note nested_dfs_impl_complete
  finally show ?thesis using assms by (auto elim!: RETURN_ref_SPECD)
qed

lemmas nested_dfs_impl_NO_LASSO_iff = nested_dfs_NO_LASSO_iff[OF `x\<in>V`, THEN nested_dfs_impl_lift]
lemma nested_dfs_code_NO_LASSO_iff:
  assumes "res = nested_dfs_code x"
  shows "res = NO_LASSO \<longleftrightarrow> \<not>(\<exists>v. x \<rightarrow>\<star>v \<and> accpt v \<and> v \<rightarrow>+ v)"
using nested_dfs_code_complete[OF assms]
using nested_dfs_code_sound_alt[OF _ assms]
by blast
end

end

(*  Red and Blue DFS Implementations *)

locale RedDFS_Impl_pre = set!: StdSetDefs sops +
                         setR!: StdSetDefs sopsR +
                         Graph_succ_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 +
                         RedDFS_pre \<alpha>succs v\<^sub>0 ss
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'XR) set_ops_scheme"
  and \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma>" and v\<^sub>0 \<gamma>\<^sub>0 ss +
  fixes restr :: "'r" and sup_st :: "'n \<Rightarrow> bool"
begin

definition mk_ret_impl where
  "mk_ret_impl x s \<equiv> Some (x, rev (x#istack s))"

definition red_visit_impl :: "('n red_cycle \<times> 'r, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'n \<Rightarrow> ('n red_cycle \<times> 'r) \<times> '\<gamma>" where
  "red_visit_impl s x \<equiv> (let _ = NDFS_Statistics.vis_red () in case istate s of
                   (Some x, R) \<Rightarrow> (Some x, R)
                 | (None, R) \<Rightarrow> (if (x \<noteq> istart s \<and> sup_st x) then mk_ret_impl x s else None, R), iss s)"

definition red_remove_impl :: "('n red_cycle \<times> 'r, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'n \<Rightarrow> ('n red_cycle \<times> 'r) \<times> '\<gamma>" where
  "red_remove_impl s x \<equiv> (case istate s of 
                        (None, R) \<Rightarrow> (if (x = istart s \<and> sup_st x) then mk_ret_impl x s else None, R) 
                      | (Some x, R) \<Rightarrow> (Some x, R), iss s)"

definition red_post_impl :: "('n red_cycle \<times> 'r, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'n \<Rightarrow> ('n red_cycle \<times> 'r) \<times> '\<gamma>" where
  "red_post_impl s x \<equiv> (case istate s of (F, R) \<Rightarrow> (F, setR.ins_dj x R), iss s)"

definition red_impl_invar where
  "red_impl_invar s \<equiv> setR.invar (snd (istate s)) \<and> setR.\<alpha> (snd (istate s)) = setR.\<alpha> restr \<union> (set.\<alpha> (idisc s) - set (istack s))"

definition red_dfs_impl :: "('n red_cycle \<times> 'r, 'n, 's, '\<gamma>) dfs_algorithm_impl" where
  "red_dfs_impl = \<lparr> dfs_impl_cond = \<lambda>S. fst S = None, 
                     dfs_impl_visit = red_visit_impl,
                     dfs_impl_post = red_post_impl,
                     dfs_impl_remove = red_remove_impl,
                     dfs_impl_start = \<lambda>_. (None, restr) \<rparr>"

lemma red_dfs_impl_simps[simp]:
  "dfs_impl_cond red_dfs_impl S \<longleftrightarrow> fst S = None"
  "dfs_impl_visit red_dfs_impl = red_visit_impl"
  "dfs_impl_post red_dfs_impl = red_post_impl"
  "dfs_impl_remove red_dfs_impl = red_remove_impl"
  "dfs_impl_start red_dfs_impl x = (None, restr)"
by (simp_all add: red_dfs_impl_def)
end

locale RedDFS_Impl_def = RedDFS_Impl_pre sops sopsR \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 "Collect sup_st" _ sup_st +
                         DFS_Impl_def \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 red_dfs sops "fst" red_dfs_impl red_impl_invar \<gamma>_inv
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'XR) set_ops_scheme"
  and \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 \<gamma>_inv sup_st
begin
abbreviation "red_dfs_code \<equiv> dfs_code"
abbreviation "red_dfs_algo_impl \<equiv> dfs_algo_impl"
abbreviation "red_dfsbr \<equiv> dfsbr"
abbreviation "red_dfs_algo \<equiv> dfs_algo"
end

locale RedDFS_Impl = set!: StdSet sops +
                     setR!: StdSet sopsR +
                     Graph_state_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv +
                     RedDFS_Impl_def restr sops sopsR \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 \<gamma>_inv sup_st +
                     RedDFS \<alpha>succs v\<^sub>0 "{x. sup_st x}"
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'XR) set_ops_scheme"
  and restr sup_st \<gamma>succs \<gamma>\<^sub>0 
  and \<gamma>_inv :: "'\<gamma> \<Rightarrow> bool"  and v\<^sub>0 :: "'n" +
  assumes rs_invar [simp]: "setR.invar restr"
  and restr_not_verts: "setR.\<alpha> restr \<inter> V = {}"
begin

lemma red_impl_invar_start:
  "x \<in> V \<Longrightarrow> red_impl_invar (dfs_start_impl x)"
unfolding red_impl_invar_def dfs_start_impl_def
by (simp add: set.correct)

lemma visit_correct:
  assumes "(si, s) \<in> red_dfsbr"
  shows "RETURN (red_visit_impl si x) \<le> \<Down>statebr (red_dfs_visit s x)"
proof -
  have "\<And>S b ss. \<gamma>_inv ss \<Longrightarrow> RETURN ((S, b), ss) \<le>\<Down>statebr (RETURN S)" by refine_rcg simp
  with  assms show ?thesis
    unfolding red_visit_impl_def red_dfs_visit_def mk_ret_impl_def
    by (refine_rcg) (simp add: dfsbr_\<gamma>_inv split: option.split)
qed

lemma visit_invar:
  assumes br:"(si, s) \<in> red_dfsbr"
  and PRE: "pre_dfs s x"
  and ND: "x \<notin> discovered s"
  and "(Si, ssi) = red_visit_impl si x"
  shows "red_impl_invar (disc_step_upd_impl x si Si ssi)" (is "red_impl_invar ?sdi")
proof -
  from assms have "red_impl_invar si" unfolding dfsbr_def refine_rel_defs
    by (simp add: impl_sws_invar_def)

  moreover from assms have "snd Si = snd (istate si)"
    unfolding red_visit_impl_def 
    by (simp split: prod.splits option.splits)
  hence "snd (istate ?sdi) = snd (istate si)"
    by (simp add: disc_step_upd_impl_def split: prod.splits)

  moreover
  from PRE ND stack_subset_discovered have "x \<notin> set (stack s)" by auto
  with br ND have "set.invar (idisc si)" "x \<notin> set (istack si)" "x \<notin> set.\<alpha> (idisc si)" 
    unfolding dfsbr_def refine_rel_defs by (simp_all add: impl_sws_invar_def)
  
  hence "set.\<alpha> (idisc ?sdi) - set (istack ?sdi) = set.\<alpha> (idisc si) - set (istack si)" 
    unfolding disc_step_upd_impl_def by (auto simp add: set.correct split: prod.splits)
  
  ultimately show ?thesis
    unfolding red_impl_invar_def
    by simp
qed

lemma remove_correct:
  assumes "(si, s) \<in> red_dfsbr"
  shows "RETURN (red_remove_impl si x) \<le> \<Down>statebr (red_dfs_remove s x)"
proof -
  have "\<And>S b ss. \<gamma>_inv ss \<Longrightarrow> RETURN ((S, b), ss) \<le>\<Down>statebr (RETURN S)" by refine_rcg simp
  with assms show ?thesis
    unfolding red_remove_impl_def red_dfs_remove_def mk_ret_impl_def
    by refine_rcg (auto intro: dfsbr_\<gamma>_inv split: option.split)
qed

lemma remove_invar:
  assumes "(si, s) \<in> red_dfsbr"
  and "(Si, ss) = red_remove_impl si x"
  shows "red_impl_invar (rem_step_upd_impl si Si ss)"
proof -
  from assms have "red_impl_invar si" 
    unfolding dfsbr_def refine_rel_defs by (simp add: impl_sws_invar_def)
  with assms show ?thesis
    unfolding red_impl_invar_def rem_step_upd_impl_def red_remove_impl_def mk_ret_impl_def
    by (auto simp add: set.correct split: option.split prod.split)
qed

lemma post_correct:
  assumes "(si, s) \<in> red_dfsbr"
  and "dfs_cond red_dfs (state s)"
  shows "RETURN (red_post_impl si x) \<le> \<Down>statebr (RETURN None)"
using assms
unfolding red_post_impl_def
by (refine_rcg) (simp add: dfsbr_\<gamma>_inv)

lemma post_invar:
  assumes br: "(si, s) \<in> red_dfsbr" "(si', s') \<in> red_dfsbr"
  and "pre_dfs s x"
  and COND: "dfs_cond red_dfs (state s')"
  and nF: "x \<notin> finished s'"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x {} s'" (is "fe_inv_dfs ?sd x {} s'")
  and "(Si, ssi) = red_post_impl si' x"
  shows "red_impl_invar (fin_step_upd_impl x si' Si ssi)" (is "red_impl_invar ?F")
proof -
  from assms have "x \<in> V" by force

  from assms have inv: "red_impl_invar si'" "set.invar (idisc si')" 
    unfolding dfsbr_def refine_rel_defs by (simp_all add: impl_sws_invar_def)

  from FE COND have "stack s' = stack ?sd" by blast
  hence sts': "x = hd (stack s')" "stack s' \<noteq> []" by (simp_all add: disc_step_upd_def)
  with distinct_hd_tl FE common_inv_stack_distinct have "x \<notin> set (tl (stack s'))" by fastforce
  with sts' have "set (tl (stack s')) = set (stack s') - {x}" by (induct ("stack s'")) simp_all
  with br have xtl: "set (tl (istack si')) = set (istack si') - {x}" by simp
  
  from FE COND fe_inv_discovered have "x \<in> discovered s'" by simp_all
  with inv br have xiD: "x \<in> set.\<alpha> (idisc si')"  unfolding dfsbr_def refine_rel_defs by simp_all
  
  from `x \<in> V` restr_not_verts have "x \<notin> setR.\<alpha> restr" by auto
  moreover from assms have xnD: "x \<notin> set.\<alpha> (idisc si') - set (istack si')" by (simp add: finished_def)
  ultimately have "x \<notin> setR.\<alpha> (snd (istate si'))" using inv unfolding red_impl_invar_def by (auto simp add: set.correct setR.correct)
  
  moreover
  from xtl xiD have "x \<in> set.\<alpha> (idisc si') - set (tl (istack si'))" by auto
  with xnD xtl have "... = insert x (set.\<alpha> (idisc si') - set (istack si'))" by auto

  moreover
  from assms have "snd (istate ?F) = setR.ins_dj x (snd (istate si'))"
    unfolding fin_step_upd_impl_def red_post_impl_def by (auto split: prod.splits)

  ultimately show ?thesis using xnD inv xtl
    by (cases si') (simp add: set.correct setR.correct red_impl_invar_def fin_step_upd_impl_def)
qed

lemma succ_step_inv:
  "red_impl_invar si \<Longrightarrow> red_impl_invar (succ_step_upd_impl si ssi)"
by (auto simp add: succ_step_upd_impl_def red_impl_invar_def split: prod.split)
end

sublocale RedDFS_Impl \<subseteq> DFS_Impl red_dfs red_dfs_impl red_impl_invar fst
  apply unfold_locales
  apply simp+
  apply (simp add: red_impl_invar_start)
  apply (simp add: visit_correct)
  apply (simp add: visit_invar)
  apply (simp add: remove_correct)
  apply (simp add: remove_invar)
  apply (simp add: post_correct)
  apply simp
  apply (simp add: post_invar) 
  apply (simp add: succ_step_inv)
done

context RedDFS_Impl
begin

context 
  fixes x
  assumes "x \<in> V"
begin

lemma red_dfs_code_refine:
  "RETURN (red_dfs_code x) \<le>\<Down>red_dfsbr (red_dfs_algo x)"
proof -
  note dfs_code.refine[of x]
  also note dfs_algo_impl_refine[OF `x\<in>V`]
  finally show ?thesis .
qed

lemma red_dfs_algo_lift:
  assumes "dfs_algo x \<le> SPEC \<Phi>"
  shows "red_dfs_algo_impl x \<le>\<Down>red_dfsbr (SPEC \<Phi>)"
proof -
  note dfs_algo_impl_refine[OF `x\<in>V`]
  also note assms
  finally show "red_dfs_algo_impl x \<le>\<Down>red_dfsbr (SPEC \<Phi>)" .
qed

theorems red_dfs_impl_sound = red_dfs_sound[OF `x\<in>V`, THEN red_dfs_algo_lift]

theorem red_dfs_code_sound:
  assumes "s = red_dfs_code x" and "fst (istate s) = Some (v, vs)"
  shows "x \<rightarrow>+ v \<and> sup_st v"
proof -
  note dfs_code.refine
  also note red_dfs_impl_sound[of v vs]
  finally show ?thesis using assms by (fastforce elim!: RETURN_ref_SPECD)
qed

theorems red_dfs_impl_complete = red_dfs_complete[OF `x\<in>V`, THEN red_dfs_algo_lift]

theorem red_dfs_code_complete:
  assumes "s = red_dfs_code x" and "fst (istate s) = None"
  shows "\<not> (\<exists>v. x \<rightarrow>+ v \<and> sup_st v)"
proof -
  note dfs_code.refine
  also note red_dfs_impl_complete
  finally show ?thesis using assms by (auto elim!: RETURN_ref_SPECD)
qed

theorems red_dfs_impl_counter_example_correct = red_dfs_counter_example_correct[OF `x\<in>V`, THEN red_dfs_algo_lift]

theorem red_dfs_code_counter_example_correct:
  assumes "s = red_dfs_code x" and "fst (istate s) = Some (v, vs)"
  shows "x = hd vs \<and> v = last vs \<and> vwalk vs \<G> \<and> sup_st v \<and> length vs > 1"
proof -
  note dfs_code.refine
  also note red_dfs_impl_counter_example_correct[of v vs]
  finally show ?thesis using assms by (auto elim!: RETURN_ref_SPECD)
qed

lemma red_dfs_code_reachable:
  assumes "s = red_dfs_code x" and "fst (istate s) = None"
  shows "setR.\<alpha> (snd (istate s)) = setR.\<alpha> restr \<union> {v. x = v \<or> x \<rightarrow>+ v}"
proof -
  from dfs_codeD[OF `x\<in>V`] assms have POST: "post_dfs (dfs_alg_start x) (impl_sws_\<alpha> s) x"
                        and POSTC: "red_dfs_invar (impl_sws_\<alpha> s)"
                        and INV: "red_impl_invar s" by (auto simp add: impl_sws_invar_def)

  from assms have "dfs_impl_cond red_dfs_impl (istate s)" by simp
  hence COND: "dfs_cond red_dfs (state (impl_sws_\<alpha> s))" by simp
  with post_cond_stack_empty[OF POST] have "stack (impl_sws_\<alpha> s) = []" by simp
  with post_empty_stack_finished_eq POST COND have "finished (impl_sws_\<alpha> s) = {v. x = v \<or> x \<rightarrow>+ v}" by blast
  hence "set.\<alpha> (idisc s) - set (istack s) = ..." by (simp add: finished_def)
  with INV[unfolded red_impl_invar_def] show ?thesis by blast
qed
end

end

locale BlueDFS_Impl_pre = set!: StdSetDefs sops +
                          setR!: StdSetDefs sopsR +
                          setS!: StdSetDefs sopsS +
                          NestedDFS_Impl_pre sops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 accpt +
                          BlueDFS_pre \<alpha>succs v\<^sub>0 "Collect accpt" M
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'Z) set_ops_scheme"
  and sopsS :: "('n, 'st, 'ZZ) set_ops_scheme"
  and M accpt 
  and \<gamma>succs :: "'\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma> "
  and v\<^sub>0 \<gamma>\<^sub>0 +
  fixes \<gamma>M :: "('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'r \<Rightarrow> 'st \<Rightarrow> 'n \<Rightarrow> bool"
  and \<gamma>_inv :: "'\<gamma> \<Rightarrow> bool"
begin

definition restr_\<gamma>succs :: "'r \<Rightarrow> '\<gamma> \<Rightarrow> 'n \<Rightarrow> 'n lsi \<times> '\<gamma>" where
  "restr_\<gamma>succs R \<gamma> x \<equiv> let (suc, \<gamma>') = \<gamma>succs \<gamma> x in (lsi.filter (\<lambda>x. \<not> setR.memb x R) suc, \<gamma>')"

abbreviation red_dfs_impl :: "'r \<Rightarrow> ('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> _ " where
  "red_dfs_impl R s ss \<equiv> RedDFS_Impl_def.red_dfs_code R sops sopsR (restr_\<gamma>succs R) (iss s) ss"

definition run_red_dfs_impl' :: "'r \<Rightarrow> ('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> ('n \<Rightarrow> bool) \<Rightarrow> 'n \<Rightarrow> (('n, 'r) blue_res \<times> '\<gamma>) nres"
where 
"run_red_dfs_impl' R s ss x \<equiv> do {
    let red = red_dfs_impl R s ss x;
    case (istate red) of
      (None, R') \<Rightarrow> RETURN (CYCLE_NO R', iss red)
    | (Some (v, vs), R') \<Rightarrow> RETURN (CYCLE_PRE v vs, iss red) }"

schematic_lemma run_red_dfs_impl_aux:
  "RETURN ?code \<le> run_red_dfs_impl' R s ss x"
  unfolding run_red_dfs_impl'_def
  by refine_transfer

concrete_definition run_red_dfs_impl for R s ss x uses run_red_dfs_impl_aux

definition blue_post_impl' where
  "blue_post_impl' s x \<equiv> if accpt x then (
                          case istate s of (St, CYCLE_NO R) \<Rightarrow>
                            do {
                                 (S', ss') \<leftarrow> run_red_dfs_impl' R s (\<gamma>M s R St) x;
                                 case S' of
                                   CYCLE_PRE v vs \<Rightarrow> do {
                                     let rs = rev (istack s);
                                     let cyc = mk_cycle rs v;
                                     RETURN ((setS.delete x St, CYCLE_FIN x v rs vs cyc), ss') }
                                   | _ \<Rightarrow> RETURN ((setS.delete x St, S'), ss')
                                }
                          | _ \<Rightarrow> RETURN (istate s, iss s))
                        else (case istate s of
                           (St, CYCLE_NO R) \<Rightarrow> RETURN ((setS.delete x St, CYCLE_NO R), iss s)
                          | _ \<Rightarrow> RETURN (istate s, iss s))"

schematic_lemma blue_post_impl_aux:
  "RETURN (?code s x) \<le> blue_post_impl' s x"
  unfolding blue_post_impl'_def
  by (refine_transfer run_red_dfs_impl.refine)

concrete_definition blue_post_impl for s x uses blue_post_impl_aux

definition blue_visit_impl :: "('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> _ " where
  "blue_visit_impl s x \<equiv> (let _ = NDFS_Statistics.vis_blue() in case istate s of (St, CYCLE_NO R) \<Rightarrow> (setS.ins_dj x St, CYCLE_NO R) | _ \<Rightarrow> istate s, iss s)"

definition blue_impl_invar :: "('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> _" where
  "blue_impl_invar s \<equiv> case istate s of (St, CYCLE_NO R) \<Rightarrow> setR.invar R \<and> setS.invar St \<and> setS.\<alpha> St = set (istack s) | _ \<Rightarrow> True"

definition blue_dfs_impl :: "('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) dfs_algorithm_impl" where
  "blue_dfs_impl = \<lparr> dfs_impl_cond = \<lambda>S. (case S of (_, CYCLE_NO _) \<Rightarrow> True | _ \<Rightarrow> False),
                     dfs_impl_visit = blue_visit_impl,
                     dfs_impl_post = blue_post_impl,
                     dfs_impl_remove = \<lambda>s _. let _ = NDFS_Statistics.match_blue() in (istate s, iss s),
                     dfs_impl_start = \<lambda>_. (setS.empty (), CYCLE_NO (setR.empty ())) \<rparr>"

lemma blue_dfs_impl_simps[simp]:
  "dfs_impl_cond blue_dfs_impl S \<longleftrightarrow> (case S of (_, CYCLE_NO _) \<Rightarrow> True | _ \<Rightarrow> False)"
  "dfs_impl_visit blue_dfs_impl = blue_visit_impl"
  "dfs_impl_post blue_dfs_impl = blue_post_impl"
  "dfs_impl_remove blue_dfs_impl s x = (istate s, iss s)"
  "dfs_impl_start blue_dfs_impl x = (setS.empty (), CYCLE_NO (setR.empty ()))"
by (simp_all add: blue_dfs_impl_def)

abbreviation "blue\<alpha>\<^sub>\<sigma>' \<equiv> blue_res\<alpha> setR.\<alpha>"

definition blue\<alpha>\<^sub>\<sigma> :: "'st \<times> ('n, 'r) blue_res \<Rightarrow> ('n, 'n set) blue_res" 
where
  "blue\<alpha>\<^sub>\<sigma> \<equiv> blue\<alpha>\<^sub>\<sigma>' \<circ> snd"

lemma blue\<alpha>\<^sub>\<sigma>_simps [simp]:
  "blue\<alpha>\<^sub>\<sigma> (St, CYCLE_NO R) = CYCLE_NO (setR.\<alpha> R)"
  "blue\<alpha>\<^sub>\<sigma> (St, CYCLE_PRE v vs) = CYCLE_PRE v vs"
  "blue\<alpha>\<^sub>\<sigma> (St, CYCLE_FIN a b c d e) = CYCLE_FIN a b c d e"
unfolding blue\<alpha>\<^sub>\<sigma>_def
by simp_all
end

locale BlueDFS_Impl_def = BlueDFS_Impl_pre sops sopsR sopsS _ accpt \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 _ \<gamma>_inv +
                          NestedDFS_Impl_def blue_dfs blue_dfs_impl blue_impl_invar blue\<alpha>\<^sub>\<sigma> sops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 accpt \<gamma>_inv snd
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'Y) set_ops_scheme"
  and sopsS :: "('n, 'st, 'Z) set_ops_scheme"
  and \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 \<gamma>_inv accpt
begin
abbreviation "blue_dfs_code \<equiv> dfs_code"
abbreviation "blue_dfs_algo_impl \<equiv> dfs_algo_impl"
abbreviation "blue_dfsbr \<equiv> dfsbr"
abbreviation "blue_dfs_algo \<equiv> dfs_algo"
end

locale BlueDFS_Impl = set!: StdSet sops +
                      setR!: StdSet sopsR +
                      setS!: StdSet sopsS +
                      Graph_state_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv +
                      BlueDFS_Impl_def M \<gamma>M sops sopsR sopsS \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 \<gamma>_inv accpt +
                      BlueDFS M \<alpha>succs v\<^sub>0 "Collect accpt"
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'Z) set_ops_scheme"
  and sopsS :: "('n, 'st, 'ZZ) set_ops_scheme"
  and accpt :: "'n \<Rightarrow> bool" and M \<gamma>M \<gamma>succs 
  and \<gamma>\<^sub>0 :: "'\<gamma>" and \<gamma>_inv and v\<^sub>0 :: "'n" +
  assumes \<gamma>M_correct: "impl_sws_invar s \<Longrightarrow> istate s = (St, CYCLE_NO R) \<Longrightarrow> \<gamma>M s R St = M (impl_sws_\<alpha> s) (setR.\<alpha> R)"
begin

definition statebr' :: "((('n, 'r) blue_res \<times> '\<gamma>) \<times> ('n, 'n set) blue_res) set" where
  "statebr' \<equiv> br (blue\<alpha>\<^sub>\<sigma>' \<circ> fst) (\<lambda>(s,\<gamma>). \<gamma>_inv \<gamma>)"

lemma sv_statebr'[refine]:
  "single_valued statebr'"
  unfolding statebr'_def
  by blast

lemma [refine_dref_RELATES]: "RELATES statebr'"
by (simp add: RELATES_def)

lemma blue_res_split_refine' [refine]:
  assumes "((xi, \<gamma>),x) \<in> statebr'"
  assumes "\<And> r ri. \<lbrakk>xi = CYCLE_NO ri; x = CYCLE_NO r; r = setR.\<alpha> ri\<rbrakk> \<Longrightarrow> fnci ri \<le> \<Down>R (fnc r)"
  assumes "\<And> n nl. xi = CYCLE_PRE n nl \<Longrightarrow> fcpi n nl \<le>\<Down>R (fcp n nl)"
  assumes "\<And> y v ys vs cyc. xi = CYCLE_FIN y v ys vs cyc \<Longrightarrow>
           fcfi y v ys vs cyc \<le> \<Down>R (fcf y v ys vs cyc)"
  shows "blue_res_case fnci fcpi fcfi xi \<le> \<Down>R (blue_res_case fnc fcp fcf x)"
  using assms unfolding statebr'_def refine_rel_defs
  by (auto split: blue_res.split)

lemma blue_impl_invar_start:
  "x \<in> V \<Longrightarrow> blue_impl_invar (dfs_start_impl x)"
unfolding blue_impl_invar_def dfs_start_impl_def
by (simp add: setR.correct setS.correct)

lemma blue\<alpha>\<sigma>_correct:
  "\<gamma>_inv (iss si) \<Longrightarrow> RETURN (istate si, iss si) \<le> \<Down>statebr (RETURN (blue\<alpha>\<^sub>\<sigma> (istate si)))"
by refine_rcg simp

abbreviation "red_dfsbr R \<equiv> RedDFS_Impl_def.red_dfsbr R sops sopsR \<gamma>_inv"

lemma restr_\<gamma>succs_invar:
  "\<gamma>_inv \<gamma> \<Longrightarrow> lsi.invar (fst (restr_\<gamma>succs R \<gamma> x))"
unfolding restr_\<gamma>succs_def
using lsi.filter_correct[where P = "\<lambda>v. \<not> setR.memb v R" and s = "fst (\<gamma>succs \<gamma> x)"]
using ls_invar[of \<gamma>]
by (auto split: prod.split)

lemma restr_\<gamma>succs_\<gamma>_inv:
  "\<gamma>_inv \<gamma> \<Longrightarrow> \<gamma>_inv (snd (restr_\<gamma>succs R \<gamma> x))"
unfolding restr_\<gamma>succs_def
using \<gamma>_inv_preserve[of \<gamma> x]
by (auto split: prod.split)

lemma restr_\<gamma>succs_alt:
  assumes "\<gamma>_inv \<gamma>"
  shows "lsi.\<alpha> (fst (restr_\<gamma>succs R \<gamma> x)) = {v \<in> lsi.\<alpha> (fst (\<gamma>succs \<gamma> x)). \<not> setR.memb v R}"
proof -
  from ls_invar assms have "lsi.invar (fst (\<gamma>succs \<gamma> x))" by simp
  thus ?thesis
    unfolding restr_\<gamma>succs_def
    using lsi.filter_correct[where P = "\<lambda>v. \<not> setR.memb v R"] 
    by (simp split: prod.split) force
qed

lemma restr_\<gamma>succs_succs_restrict:
  assumes "\<gamma>_inv \<gamma>"
  and "setR.invar R"
  shows "lsi.\<alpha> (fst (restr_\<gamma>succs R \<gamma> x)) = succs_restrict (setR.\<alpha> R) x"
proof -
  from succs_restrict_def have "succs_restrict (setR.\<alpha> R) x = \<alpha>succs x - (setR.\<alpha> R)" by simp
  also have "... = {v \<in> \<alpha>succs x. v \<notin> setR.\<alpha> R}" by auto
  also from assms have "... = {v \<in> \<alpha>succs x. \<not> setR.memb v R }" by (simp add: setR.correct)
  also from \<alpha>succs_def have "... = {v \<in> lsi.\<alpha> (fst (\<gamma>succs \<gamma>\<^sub>0 x)). \<not> setR.memb v R}" by simp
  also from \<gamma>succs_indep[OF \<gamma>0_inv assms(1)] have "... =  {v \<in> lsi.\<alpha> (fst (\<gamma>succs \<gamma> x)). \<not> setR.memb v R}" by simp
  also from restr_\<gamma>succs_alt[OF assms(1)] have "... = lsi.\<alpha> (fst (restr_\<gamma>succs R \<gamma> x))" by simp
  finally show ?thesis ..
qed

lemma restr_\<gamma>succs_subset:
  assumes "\<gamma>_inv \<gamma>"
  shows "lsi.\<alpha> (fst (restr_\<gamma>succs R \<gamma> x)) \<subseteq> lsi.\<alpha> (fst (\<gamma>succs \<gamma> x))"
using assms restr_\<gamma>succs_alt 
by auto

lemma restr_\<gamma>succs_state_indep:
  assumes "\<gamma>_inv \<gamma>" and "\<gamma>_inv \<gamma>'"
  shows "fst (restr_\<gamma>succs R \<gamma> x) = fst (restr_\<gamma>succs R \<gamma>' x)"
unfolding restr_\<gamma>succs_def
using \<gamma>succs_indep[OF assms, of x]
by (auto split: prod.split)

lemma interpret_sub_graph:
  assumes "\<gamma>_inv \<gamma>"
  and "x \<in> V"
  shows "Graph_state_Impl lsi_ops (restr_\<gamma>succs R) x \<gamma> \<gamma>_inv"
proof -
  from restr_\<gamma>succs_subset[OF assms(1)] have "{(v,v'). v' \<in> lsi.\<alpha> (fst (restr_\<gamma>succs R \<gamma> v))} \<subseteq> {(v, v'). v' \<in> lsi.\<alpha> (fst (\<gamma>succs \<gamma> v))}" by auto
  with \<gamma>succs_fin_subgraphI assms interpret g': Graph_fin_succ_Impl lsi_ops "restr_\<gamma>succs R" x \<gamma> 
    by blast
  
  from assms show ?thesis
    apply unfold_locales
      apply (metis restr_\<gamma>succs_state_indep)
      apply metis
      apply (blast intro: restr_\<gamma>succs_\<gamma>_inv)
      apply (blast intro: restr_\<gamma>succs_invar)
  done
qed

lemma interpret_red_dfs:
  assumes inv: "setR.invar R" "\<gamma>_inv \<gamma>"
  and x: "x \<in> V" "x \<notin> setR.\<alpha> R"
  shows "RedDFS_Impl sops sopsR R (restr_\<gamma>succs R) \<gamma> \<gamma>_inv x"
proof -
  interpret G: Graph_state_Impl lsi_ops "restr_\<gamma>succs R" x \<gamma> \<gamma>_inv
    using assms interpret_sub_graph by metis
  
  show ?thesis
    by unfold_locales 
       (auto simp: G.V_alt_def G.\<alpha>succs_def inv setR.correct restr_\<gamma>succs_alt x
            elim: tranclE)
qed

lemma red_dfs_impl_refine:
  assumes inv: "setR.invar R" "\<gamma>_inv (iss s)"
  and "x \<in> V" "x \<notin> setR.\<alpha> R"
  shows "RETURN (red_dfs_impl R s ss x) \<le>\<Down>(red_dfsbr R) (red_dfs_algo (setR.\<alpha> R) {y. ss y} x)"
proof -
  interpret red: RedDFS_Impl sops sopsR R ss "restr_\<gamma>succs R" "iss s" \<gamma>_inv x
    using interpret_red_dfs assms by blast

  have transfer: "red.\<alpha>succs = succs_restrict (setR.\<alpha> R)"
    using restr_\<gamma>succs_succs_restrict[OF inv(2,1)]
    by (auto simp add: red.\<alpha>succs_def)
   
  have "x \<in> red.V" by simp
  with red.red_dfs_code_refine transfer show ?thesis by simp
qed

lemma run_red_dfs_impl_refine:
  assumes inv: "setR.invar R" and "\<gamma>_inv (iss s)"
  and x: "x \<in> V" "x \<notin> setR.\<alpha> R"
  shows "run_red_dfs_impl' R s ss x \<le>\<Down>statebr' (run_red_dfs (setR.\<alpha> R) {y. ss y} x)"
proof -
  interpret red: RedDFS_Impl sops sopsR R ss "restr_\<gamma>succs R" "iss s" \<gamma>_inv x
    using interpret_red_dfs assms by blast
  
  show ?thesis
    unfolding run_red_dfs_impl'_def run_red_dfs_def
    apply (refine_rcg red_dfs_impl_refine[OF assms])
    apply (simp add: refine_hsimp)
  proof -
    case (goal1 s' F R') then obtain si where si: "si = red_dfs_impl R s ss x" by blast
    with goal1[unfolded red.dfsbr_def] have "setR.\<alpha> R' = setR.\<alpha> R \<union> (set.\<alpha> (idisc si) - set (istack si))" and "set.invar (idisc si)" and "stack s' = istack si" and "discovered s' = set.\<alpha> (idisc si)" "\<gamma>_inv (iss si)"
      by (simp_all add: red.impl_sws_invar_def red.red_impl_invar_def refine_rel_defs)
    with si show ?case 
      unfolding finished_def
      by (auto simp add: setS.correct statebr'_def refine_rel_defs)
  next
    case (goal2 s' F R' f) then obtain si where si: "si = red_dfs_impl R s ss x" by blast
    with goal2[unfolded red.dfsbr_def] have "\<gamma>_inv (iss si)" 
      by (simp_all add: red.impl_sws_invar_def refine_rel_defs)
    with si goal2 show ?case by (simp add: statebr'_def refine_rel_defs)
  qed
qed
  
lemma red_dfs_impl_invar:
  assumes inv: "setR.invar R" "\<gamma>_inv (iss s)"
  and x: "x \<in> V" "x \<notin> setR.\<alpha> R"
  and S: "istate (red_dfs_impl R s ss x) = (None, R')"
  shows "setR.invar R'"
proof -
  interpret red: RedDFS_Impl sops sopsR  R ss "restr_\<gamma>succs R" "iss s" \<gamma>_inv x
    using interpret_red_dfs assms by metis

  from x red_dfs_impl_refine[OF inv x, of ss]  have "red.red_impl_invar (red_dfs_impl R s ss x)"
    by (metis red.dfs_codeD red.impl_sws_inv_impl_inv red.v\<^sub>0_in_V)

  with S show ?thesis unfolding red.red_impl_invar_def by simp
qed

lemma visit_correct_blue:
  assumes "(si, s) \<in> blue_dfsbr"
  and "blue\<alpha>\<^sub>\<sigma> (istate si) = CYCLE_NO R"
  shows "RETURN (blue_visit_impl si x) \<le> \<Down>statebr (RETURN (CYCLE_NO R))"
using assms
unfolding blue_visit_impl_def
by refine_rcg (auto split: blue_res.split simp: statebr_def refine_rel_defs)

lemma visit_invar:
  assumes "(si, s) \<in> blue_dfsbr"
  and "x \<notin> discovered s"
  and "pre_dfs s x"
  and "blue_dfs_pre_cond s x"
  and "blue\<alpha>\<^sub>\<sigma> (istate si) = CYCLE_NO R"
  and "(Si, \<gamma>) = blue_visit_impl si x"
  shows "blue_impl_invar (disc_step_upd_impl x si Si \<gamma>)"
proof -
  from assms have "blue_impl_invar si" unfolding dfsbr_def 
    by (simp add: impl_sws_invar_def refine_rel_defs)
  moreover from assms have "x \<notin> set (stack s)" using stack_subset_discovered by blast
  ultimately show ?thesis
    using assms
    unfolding blue_impl_invar_def disc_step_upd_impl_def blue_visit_impl_def
    by (cases si) (simp add: setS.correct split: blue_res.splits prod.splits)
qed

lemma remove_invar:
  assumes "(si, s) \<in> blue_dfsbr"
  and "Si = istate si"
  shows "blue_impl_invar (rem_step_upd_impl si Si \<gamma>)"
proof -
  from assms have "blue_impl_invar si" 
    unfolding dfsbr_def refine_rel_defs by (simp add: impl_sws_invar_def)
  with assms show ?thesis
    unfolding blue_impl_invar_def rem_step_upd_impl_def
    by (cases si, simp)
qed

lemma post_correct_blue':
  assumes br: "(si, s) \<in> blue_dfsbr" "(si', s') \<in> blue_dfsbr"
  and C\<gamma>: "blue\<alpha>\<^sub>\<sigma> (istate si) = CYCLE_NO R\<gamma>" "blue\<alpha>\<^sub>\<sigma> (istate si') = CYCLE_NO R\<gamma>'"
  and FE: "fe_inv_dfs (disc_step_upd x s S) x {} s'" (is "fe_inv_dfs ?sd x {} s'")
  and FEC: "blue_dfs_fe_inv (disc_step_upd x s S) x {} s'"
  shows "blue_post_impl' si' x \<le> \<Down>statebr (blue_dfs_post s' x)"
proof -
  from C\<gamma> obtain R St R' St' where C: "istate si = (St, CYCLE_NO R)" "istate si' = (St', CYCLE_NO R')" 
    apply (cases "istate si", cases "snd (istate si)")
      apply auto
    apply (cases "istate si'", cases "snd (istate si')")
      apply auto
    done
  with br have S: "state s = CYCLE_NO (setR.\<alpha> R)" "state s' = CYCLE_NO (setR.\<alpha> R')" by simp_all

  with FEC have inv: "blue_dfs_invar s'" unfolding blue_dfs_fe_inv_def by simp
  moreover

  moreover from br have "\<gamma>_inv (iss si')" by auto

  moreover
  from FE fe_inv_stack have st: "x \<in> set (stack s')" by simp
  with stack_subset_verts FE have "x \<in> V" by auto
  
  moreover
  from st inv S have "x \<notin> setR.\<alpha> R'" unfolding blue_dfs_invar_def by auto

  moreover
  from br have "blue_impl_invar si'" by blast
  with C have "setR.invar R'" and Stinv: "setS.invar St'" by (simp_all add: blue_impl_invar_def)

  moreover have "{y. M s' (setR.\<alpha> R') y} = {y. \<gamma>M si' R' St' y}" (is "?M = {y. ?\<gamma>M y}")
    using \<gamma>M_correct br C unfolding dfsbr_def refine_rel_defs by simp

  ultimately have "run_red_dfs_impl' R' si' ?\<gamma>M x \<le>\<Down>statebr' (run_red_dfs (setR.\<alpha> R') ?M x)"
    using run_red_dfs_impl_refine by metis

  with S br C Stinv show ?thesis
    unfolding blue_post_impl'_def blue_dfs_post_def
    by refine_rcg (auto simp add: statebr'_def statebr_def refine_rel_defs)
qed

lemmas post_correct_blue = post_correct_blue'[THEN order_trans[OF blue_post_impl.refine]]

lemma post_invar:
  assumes br: "(si, s) \<in> blue_dfsbr" "(si', s') \<in> blue_dfsbr"
  and C\<gamma>: "blue\<alpha>\<^sub>\<sigma> (istate si) = CYCLE_NO R\<gamma>" "blue\<alpha>\<^sub>\<sigma> (istate si') = CYCLE_NO R\<gamma>'"
   and FE: "fe_inv_dfs (disc_step_upd x s S) x {} s'" (is "fe_inv_dfs ?sd x {} s'")
  and FEC: "blue_dfs_fe_inv (disc_step_upd x s S) x {} s'"
  and Si: "(Si, \<gamma>) = blue_post_impl si' x"
  shows "blue_impl_invar (fin_step_upd_impl x si' Si \<gamma>)" (is "blue_impl_invar ?F")
proof (cases "snd Si")
  case (CYCLE_NO R'')
  obtain St'' where St'': "St'' = fst Si" by auto

  from C\<gamma> obtain R St R' St' where C: "istate si = (St, CYCLE_NO R)" "istate si' = (St', CYCLE_NO R')" 
    apply (cases "istate si", cases "snd (istate si)")
      apply auto
    apply (cases "istate si'", cases "snd (istate si')")
      apply auto
    done
  with br have S: "state s = CYCLE_NO (setR.\<alpha> R)" "state s' = CYCLE_NO (setR.\<alpha> R')" by simp_all

  with FEC have inv: "blue_dfs_invar s'" unfolding blue_dfs_fe_inv_def by simp

  from FE fe_inv_stack have st: "x \<in> set (stack s')" by simp
  with stack_subset_verts FE inv S have x: "x \<in> V" "x \<notin> setR.\<alpha> R'" unfolding blue_dfs_invar_def by auto
  
  from br have "blue_impl_invar si'" by blast
  with C have R': "setR.invar R'" and St': "setS.invar St'" and St_stack: "setS.\<alpha> St' = set (istack si')" by (simp_all add: blue_impl_invar_def)

  from assms have "setR.invar R'' \<and> setS.invar St''"
  proof (cases "accpt x")
    case False with inv CYCLE_NO C R' St' St'' Si show ?thesis
      unfolding blue_post_impl_def
      by (auto simp: blue_impl_invar_def setS.correct split: blue_res.splits)
  next
    case True with C Si CYCLE_NO have "istate (red_dfs_impl R' si' (\<gamma>M si' R' St') x) = (None, R'')"
      unfolding blue_post_impl_def run_red_dfs_impl_def
      by (simp add: Let_def split: prod.splits option.splits blue_res.splits)
    with br red_dfs_impl_invar R' x have "setR.invar R''" by blast

    moreover
    from C True Si CYCLE_NO St' St'' have "setS.invar St''" unfolding blue_post_impl_def by (simp add: Let_def setS.correct split: blue_res.splits prod.splits)

    ultimately show ?thesis by simp
  qed
  moreover
  have "setS.\<alpha> St'' = set (istack ?F)"
  proof -
    from C Si CYCLE_NO St' St'' have "St'' = setS.delete x St'"
      unfolding blue_post_impl_def
      by (cases "accpt x") (auto simp: Let_def split: blue_res.splits prod.splits )

    moreover
    from FE S fe_inv_hd_stack have xhd: "x = hd (stack s')" by auto
    with FE common_inv_stack_distinct have "x \<notin> set (tl (stack s'))" using distinct_hd_tl by fastforce
    with st have "set (tl (stack s')) = set (stack s') - {x}" by (induct ("stack s'")) simp_all
    with br St_stack have "set (tl (stack s')) = setS.\<alpha> St' - {x}" by simp
    
    ultimately
    have "setS.\<alpha> St'' = set (tl (stack s'))" using St' by (simp add: setS.correct)
    with br show ?thesis by (simp add: fin_step_upd_impl_def split: prod.splits)
  qed

  ultimately show ?thesis
    using CYCLE_NO St''
    by (fastforce simp add: fin_step_upd_impl_def blue_impl_invar_def split: prod.splits)
qed (fastforce simp add: fin_step_upd_impl_def blue_impl_invar_def split: prod.splits)+

lemma succ_step_invar:
  "blue_impl_invar si \<Longrightarrow> blue_impl_invar (succ_step_upd_impl si \<gamma>)"
unfolding succ_step_upd_impl_def blue_impl_invar_def 
by (auto split: blue_res.splits prod.splits)
end

sublocale BlueDFS_Impl \<subseteq> DFS_Impl blue_dfs blue_dfs_impl blue_impl_invar blue\<alpha>\<^sub>\<sigma>
  apply unfold_locales
  apply (simp split: blue_res.splits prod.split)
  apply (simp add: setS.correct setR.correct)
  apply (simp add: blue_impl_invar_start)
  apply (simp add: visit_correct_blue split: blue_res.split_asm)
  apply (simp add: visit_invar split: blue_res.split_asm)
  apply (auto dest: dfsbr_\<gamma>_inv intro: blue\<alpha>\<sigma>_correct)[]
  apply (simp add: remove_invar)
  apply (simp add: post_correct_blue split: blue_res.split_asm)
  apply (simp add: post_invar split: blue_res.split_asm)
  apply (simp add: succ_step_invar)
done

context BlueDFS_Impl
begin

lemma blue_dfs_code_refine:
  assumes "x \<in> V"
  shows "RETURN (blue_dfs_code x) \<le>\<Down>blue_dfsbr (blue_dfs_algo x)"
proof -
  note dfs_code.refine[of x]
  also note dfs_algo_impl_refine[OF assms]
  finally show ?thesis .
qed

lemma blue\<alpha>\<^sub>\<sigma>'_correct:
  "(si, s) \<in> blue_dfsbr \<Longrightarrow> blue\<alpha>\<^sub>\<sigma>' (snd (istate si)) = state s"
by (auto simp add: blue\<alpha>\<^sub>\<sigma>_def statebr_def dest: dfsbr_statebr)

end

sublocale BlueDFS_Impl \<subseteq> NestedDFS_Impl blue_dfs blue_dfs_impl blue_impl_invar blue\<alpha>\<^sub>\<sigma> snd _ _ _ _ _ _ "setR.\<alpha>"
  using blue\<alpha>\<^sub>\<sigma>'_correct
  by unfold_locales

(* Courcoubetis et al *)
locale Courcoubetis_Impl_pre = set!: StdSetDefs sops 
                   + setR!: StdSetDefs sopsR
                   + setS!: StdSetDefs sopsS
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'Z) set_ops_scheme"
  and sopsS :: "('n, 'st, 'ZZ) set_ops_scheme"
begin

definition courcoubetisM_impl :: "('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'r \<Rightarrow> 'st \<Rightarrow> 'n \<Rightarrow> bool"
where 
  "courcoubetisM_impl s R St \<equiv> \<lambda>x. hd (istack s) = x"
end

locale Courcoubetis_Impl_def = Courcoubetis_Impl_pre sops sopsR sopsS
                             + b: BlueDFS_Impl_def courcoubetisM courcoubetisM_impl sops sopsR sopsS \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 \<gamma>_inv accpt
                             + c: Courcoubetis_def \<alpha>succs v\<^sub>0 "Collect accpt"

  for \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv accpt sops sopsR sopsS
begin
abbreviation "courcoubetis_code \<equiv> dfs_code"
abbreviation "courcoubetis_dfs_algo_impl \<equiv> dfs_algo_impl"
abbreviation "courcoubetis_dfsbr \<equiv> dfsbr"
abbreviation "courcoubetis_dfs_algo \<equiv> dfs_algo"
end

locale Courcoubetis_Impl = set!: StdSet sops 
               + setR!: StdSet sopsR
               + setS!: StdSet sopsS
               + Courcoubetis_Impl_def \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv accpt sops sopsR sopsS
               + Courcoubetis "Collect accpt" \<alpha>succs v\<^sub>0
               + Graph_state_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'Z) set_ops_scheme"
  and sopsS :: "('n, 'st, 'ZZ) set_ops_scheme"
  and \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv accpt
begin
lemma courcoubetisM_impl_correct:
  "courcoubetisM_impl s R St = courcoubetisM (impl_sws_\<alpha> s) (setR.\<alpha> R)"
unfolding courcoubetisM_impl_def courcoubetisM_def
by auto
end

sublocale Courcoubetis_Impl \<subseteq> BlueDFS_Impl sops sopsR sopsS _ courcoubetisM courcoubetisM_impl
  by unfold_locales (simp_all add: courcoubetisM_impl_correct)

(* HPY *)
locale HPY_Impl_pre = set!: StdSetDefs sops 
                   + setR!: StdSetDefs sopsR
                   + setS!: StdSetDefs sopsS
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'Z) set_ops_scheme"
  and sopsS :: "('n, 'st, 'ZZ) set_ops_scheme"
begin

definition hpyM_impl :: "('st \<times> ('n, 'r) blue_res, 'n, 's, '\<gamma>) impl_sws \<Rightarrow> 'r \<Rightarrow> 'st \<Rightarrow> 'n \<Rightarrow> bool" 
where
  "hpyM_impl s R St \<equiv> \<lambda>x. setS.memb x St"
end

locale HPY_Impl_def = HPY_Impl_pre sops sopsR sopsS
                     + b: BlueDFS_Impl_def hpyM hpyM_impl sops sopsR sopsS \<gamma>succs \<gamma>\<^sub>0 v\<^sub>0 \<gamma>_inv accpt
                     + h: HPY_def \<alpha>succs v\<^sub>0 "Collect accpt"
  for \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv accpt sops sopsR sopsS
begin
abbreviation "hpy_code \<equiv> dfs_code"
abbreviation "hpy_dfs_algo_impl \<equiv> dfs_algo_impl"
abbreviation "hpy_dfsbr \<equiv> dfsbr"
abbreviation "hpy_dfs_algo \<equiv> dfs_algo"
end

locale HPY_Impl = set!: StdSet sops 
               + setR!: StdSet sopsR
               + setS!: StdSet sopsS
               + HPY_Impl_def \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv accpt sops sopsR sopsS
               + HPY "Collect accpt" \<alpha>succs v\<^sub>0
               + Graph_state_Impl lsi_ops \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv
  for sops :: "('n, 's, 'X) set_ops_scheme"
  and sopsR :: "('n, 'r, 'Z) set_ops_scheme"
  and sopsS :: "('n, 'st, 'ZZ) set_ops_scheme"
  and \<gamma>succs v\<^sub>0 \<gamma>\<^sub>0 \<gamma>_inv accpt
begin

lemma hpyM_impl_correct:
  assumes "impl_sws_invar s"
  and "istate s = (St, CYCLE_NO R)"
  shows "hpyM_impl s R St = hpyM (impl_sws_\<alpha> s) (setR.\<alpha> R)"
proof -
  from assms have "setS.invar St" and "setS.\<alpha> St = set (istack s)" by (simp_all add: impl_sws_invar_def blue_impl_invar_def)
  thus ?thesis unfolding hpyM_impl_def hpyM_def by (auto simp add: setS.correct)
qed
end

sublocale HPY_Impl \<subseteq> BlueDFS_Impl sops sopsR sopsS _ hpyM hpyM_impl
  by unfold_locales (simp_all add: hpyM_impl_correct)

end
