theory SE_DFS
imports Main Nested_DFS
begin

locale SE_DFS_pre = BlueDFS_pre succs v\<^sub>0 A M 
  for succs v\<^sub>0 and A :: "'n set" and M
begin

definition se_dfs_remove where
  "se_dfs_remove s x \<equiv> case state s of
               CYCLE_NO R \<Rightarrow> do { 
                            let hs = hd (stack s);
                            if (x \<in> A \<or> hs \<in> A ) \<and> x \<in> set (stack s) 
                            then do {
                              let rs = rev (stack s);
                              let cyc = mk_cycle rs x;
                              let p = [hs, x];
                              RETURN (CYCLE_FIN hs x rs p cyc) }
                            else RETURN (CYCLE_NO R) }
              | _ \<Rightarrow> RETURN (state s)"

definition se_dfs where
  "se_dfs = blue_dfs\<lparr> dfs_remove := se_dfs_remove \<rparr>"

lemma se_dfs_simps[simp]:
  "dfs_cond se_dfs S \<longleftrightarrow> (case S of CYCLE_NO _ \<Rightarrow> True | _ \<Rightarrow> False)"
  "dfs_visit se_dfs s x = RETURN (state s)"
  "dfs_post se_dfs = blue_dfs_post"
  "dfs_remove se_dfs = se_dfs_remove"
  "dfs_start se_dfs x = CYCLE_NO {}"
  "dfs_pre_cond se_dfs = blue_dfs_pre_cond"
  "dfs_post_cond se_dfs = blue_dfs_post_cond"
  "dfs_fe_inv se_dfs = blue_dfs_fe_inv"
unfolding se_dfs_def
by simp_all
end

locale SE_DFS_def = SE_DFS_pre succs v\<^sub>0 A M
                  + SE: NestedDFS_def se_dfs succs v\<^sub>0 A
  for M and succs v\<^sub>0 and A :: "'n set"

locale SE_DFS = SE: SE_DFS_def M succs v\<^sub>0 A +
                blue!: BlueDFS M succs v\<^sub>0 A
  for M and succs v\<^sub>0 and A :: "'n set"
begin

(* lift some stuff from the BlueDFS world in the SE_DFS world *)
lemma [simp]: "SE.post_dfs s s' x = blue.post_dfs s s' x"
  by (rule post_lift) (intro_locales, auto)

lemma [simp]: "SE.fe_inv_dfs s x it s' = blue.fe_inv_dfs s x it s'"
  by (rule fe_inv_lift) (intro_locales, auto)

lemma [simp]: "SE.dfs_alg_start x = blue.dfs_alg_start x"
  unfolding dfs_alg_start_def blue.dfs_alg_start_def by simp

(* the properties of the new remove version *)
lemma se_remove_NF:
  assumes "state s = CYCLE_NO R"
  shows "nofail (se_dfs_remove s x)"
unfolding se_dfs_remove_def
by (simp add: refine_pw_simps Let_def assms)

lemma remove_post_cond:
  assumes "x \<in> discovered s"
  and PRE: "pre_dfs s x"
  and PREC: "blue_dfs_pre_cond s x"
  and S: "state s = CYCLE_NO R"
  and INR:"inres (se_dfs_remove s x) S"
  and POST: "post_dfs s (s\<lparr>state := S\<rparr>) x"
  shows "blue_dfs_post_cond s (s\<lparr>state := S\<rparr>) x"
proof (cases "(x \<in> A \<or> hd (stack s) \<in> A) \<and> x \<in> set (stack s)")
  case False 
  with INR S have "S = state s" by (auto simp: se_dfs_remove_def inres_if)
  hence "s\<lparr>state := S\<rparr> = s" by simp
  moreover from PREC S have "blue_dfs_invar s" by (simp add: blue_dfs_pre_cond_def)
  ultimately show ?thesis by (simp add: blue_dfs_post_cond_def S)
next
  let ?hs = "hd (stack s)"
  let ?rs = "rev (stack s)"
  let ?cyc = "mk_cycle ?rs x"
  let ?p = "[?hs, x]"
  case True 
  with INR S have "RETURN S \<le> do {
                              let rs = ?rs;
                              let cyc = ?cyc;
                              let p = ?p;
                              RETURN (CYCLE_FIN ?hs x rs p cyc) }"
    by (auto simp add: se_dfs_remove_def refine_pw_simps Let_def)
  hence spec: "S = CYCLE_FIN ?hs x ?rs ?p ?cyc" by simp
  hence "blue_dfs_lasso_inv (s\<lparr>state := S\<rparr>)"
    apply (intro blue_dfs_lasso_invI)
    apply fastforce
    apply simp_all
  proof -
    from True show "?hs \<in> A \<or> x \<in> A" by blast
    
    from True have ne: "stack s \<noteq> []" by auto

    moreover from PRE have "common_inv s x" by blast
    moreover with ne hd_in_set stack_subset_verts have "?hs \<in> V" by blast
    ultimately have Chs: "common_inv s ?hs" using common_inv_change by blast

    with True blue.mk_cycle_vwalk show "vwalk ?cyc \<G>" "hd ?cyc = x" "last ?cyc = ?hs"
      by blast+

    from PRE[unfolded pre_dfs_unfold] ne have "last (stack s) = start s" by simp
    with hd_rev ne show "hd ?rs = start s" by metis

    from PRE[unfolded pre_dfs_unfold] ne have "x \<in> succs ?hs" by simp
    with `?hs \<in> V` have "(?hs, x) \<in> E" by (rule succs_implies_edge)
    thus "vwalk [?hs, x] \<G>" unfolding vwalk_def by (simp add: edge_in_nodes)
    
    from PRE[unfolded pre_dfs_unfold] ne show "vwalk ?rs \<G>" by simp
  qed
  with spec show ?thesis unfolding blue_dfs_post_cond_def by simp
qed
end

sublocale SE_DFS \<subseteq> DFS se_dfs
  apply unfold_locales
  apply simp
  apply (simp add: se_remove_NF split: blue_res.splits)
  apply (simp add: blue.blue_dfs_post_NF split: blue_res.splits)
  apply (simp add: blue.pre_start)
  apply simp
  apply (simp add: remove_post_cond split: blue_res.splits)
  apply (simp add: blue.post_fin_step split: blue_res.splits)
  apply (simp add: blue.fe_inv_post)
  apply (simp add: blue.pre_fe_inv split: blue_res.splits)
  apply (simp add: blue.fe_inv_pre split: blue_res.splits)
  apply (simp add: blue.post_fe_inv split: blue_res.splits)
done

context SE_DFS
begin

(* just copy'n'paste from BlueDFS *)
lemma se_dfs_not_CYCLE_PRE:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. \<not>(\<exists>a b. state s = CYCLE_PRE a b))"
using assms
by (auto intro!: dfs_algo_SPEC simp add: blue_dfs_post_cond_def)

lemma se_dfs_sound:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. \<forall>y v ys vs cyc. (state s = CYCLE_FIN y v ys vs cyc \<longrightarrow> is_lasso_full x y v ys vs cyc))"
proof (rule dfs_algo_SPEC[OF assms], rule impI allI)
  case (goal1 s) have "x = start (dfs_alg_start x)" unfolding dfs_alg_start_def by simp
  with goal1 have "x = start s" using post_start_eq by force
  with goal1 show ?case by (fastforce simp add: blue_dfs_lasso_inv_def blue_dfs_post_cond_def)
qed

lemma se_dfs_complete:
  assumes "x \<in> V"
  shows "dfs_algo x \<le> SPEC (\<lambda>s. (\<exists>R. state s = CYCLE_NO R) \<longrightarrow> \<not>(\<exists>v. x \<rightarrow>\<star>v \<and> v \<in> A \<and> v \<rightarrow>+ v))"
proof (rule dfs_algo_SPEC[OF assms])
  case goal1
  show ?case
    apply (rule ccontr)
    apply clarify
  proof -
    fix v R
    assume reach: "x \<rightarrow>\<star> v" and A: "v \<in> A" and v_reach: "v \<rightarrow>+ v" and s: "state s = CYCLE_NO R"
    from goal1 have "blue_dfs_post_cond (dfs_alg_start x) s x" (is "blue_dfs_post_cond ?s s x") by simp
    with goal1 s A v_reach have "v \<in> finished s \<Longrightarrow> False" unfolding blue_dfs_post_cond_def blue_dfs_invar_def by simp

    moreover
    have "v \<in> finished s"
    proof -
      from s goal1 have "stack s = stack ?s" by auto
      hence "stack s = []" unfolding dfs_alg_start_def by simp
      with s goal1 have "finished s = {v. x = v \<or> x \<rightarrow>+ v}" using post_empty_stack_finished_eq by simp
      thus ?thesis using reach by auto
    qed
    
    ultimately show "False" .
  qed
qed
end

sublocale SE_DFS \<subseteq> NestedDFS se_dfs
  using se_dfs_sound se_dfs_complete se_dfs_not_CYCLE_PRE
  by unfold_locales

locale HPY_SE_def = SE_DFS_def hpyM
locale HPY_SE = HPY_SE_def succs v\<^sub>0 A +
                HPY A succs v\<^sub>0
  for succs v\<^sub>0 A

sublocale HPY_SE \<subseteq> SE_DFS hpyM
by unfold_locales

end