(*
#load "nums.cma";;
#load "unix.cma";;
*)

open Nested_DFS_Export
open Big_int

let measure mf (start,edges,accept) =
  let st = Sys.time() in
  let r' = mf edges accept start in
  let et = Sys.time() in
  let t1 = et -. st in

  let cycle = Nested_DFS_Export.is_aborted edges accept r'
  and no_disc = List.length (Nested_DFS_Export.to_visited r')
  and ce = List.length (Nested_DFS_Export.counter_example r') in
  print_endline ("Finished.");
  print_endline ("Cycle?: " ^ string_of_bool cycle ^ 
                "\n# of states visited: " ^ string_of_int no_disc ^
                "\nLength of counter example: " ^ string_of_int ce ^
                "\nTime: " ^ string_of_float t1 ^ "\n")

let create_graph (s', e', a') = 
    (s', Nested_DFS_Export.edges_to_graph e', Nested_DFS_Export.list_to_set a')

let get_acc () =
    let rec get_acc' acc =
        let a = read_int () in
        if a = -1 then acc
        else get_acc' (big_int_of_int a::acc)
    in get_acc' []

let get_rel () =
    let w = Str.regexp " " in
    let rec get_rel' acc =
        let l = read_line () in
        match Str.split w l with
            [a;b] -> get_rel' ((big_int_of_string a, big_int_of_string b)::acc)|
            x::[] -> assert (x = "-1"); acc
    in get_rel' []

let num = read_int();;
let acc = get_acc();;
let edges = get_rel();;

print_endline "Creating graph representation ... please be patient.";;

let g = create_graph (big_int_of_int 0, edges, acc);;

print_endline ("Vertices: "^string_of_int num);;
print_endline ("Edges: "^string_of_int (List.length edges));;
print_endline ("Accepting: "^string_of_int (List.length acc));;
print_endline ("----------------------");;

print_endline ("Now performing the search... (Courcoubetis)");;
measure Nested_DFS_Export.cour_dfs_code g;;

print_endline ("The second algorithm (HPY)...");;
measure Nested_DFS_Export.hpy_dfs_code g;;

print_endline ("The third algorithm (SE)...");;
measure Nested_DFS_Export.se_dfs_code g;;

