(*
#load "nums.cma";;
#load "unix.cma";;
#use "Nested_DFS_Export.ml";;
*)

open DFS_Code
open DFS_Impl

structure TIO = TextIO
val scanInt = TIO.scanStream (IntInf.scan StringCvt.DEC)

fun read_int () = case scanInt TIO.stdIn of
          NONE => raise Fail "No integer"
        | SOME intVal => intVal

fun measure what mf tov isa cte (start,edges,accept) =
  let 
    val st = Time.now();
    val r' = mf edges accept start;
    val td =  Time.toMilliseconds (Time.- (Time.now(), st))
    val cycle = isa edges accept r';
    val ce = List.length (cte r');
    val no_disc = List.length (tov r');
  in
    print ("Now performing the search using "^what^"...\n");
    print ("Cycle?: " ^ Bool.toString cycle ^ 
    "\n# of states visited: " ^ Int.toString no_disc ^
    "\nLength of counter example: " ^ Int.toString ce ^
    "\nTime: " ^ IntInf.toString td ^ "ms\n\n")
  end

fun get_acc () =
  let fun get_acc' acc =
    let val a = read_int () in
    if a = ~1 then acc
    else get_acc' (a::acc) end
  in get_acc' [] end

fun get_edges () =
  let fun get_edges' acc =
    let val a = read_int () in
    if a = ~1 then acc
    else get_edges' ((a, read_int())::acc) end
  in get_edges' [] end

fun create_graph printg =
  let 
    val num = read_int();
    val acc = get_acc();
    val edges = get_edges();
    val work = fn () => (num,
      List.length edges, 
      List.length acc, 
      (IntInf.fromInt 0, edges_to_graph edges, list_to_set acc))
  in
    case printg of
       "true" =>
          (
            print ("Creating graph representation ... please be patient.\n");
            let
              val (num, e, a, g) = work ()
            in
                print ("\nVertices: "^IntInf.toString num^"\n");
                print ("Edges: "^Int.toString e^"\n");
                print ("Accepting: "^Int.toString a^"\n");
                print ("----------------------\n");
                (num, e, a, g)
            end
          )
      | _ => work ()
end

fun main () = 
  let 
    val [which, print_graph] = CommandLine.arguments ()
    val (num, e, a, g) = create_graph print_graph
  in
    case which of
      "Courcoubetis" =>
          measure which cour_dfs_code to_visited is_aborted counter_example g
    | "HPY" =>
          measure which hpy_dfs_code to_visited is_aborted counter_example g
    | "SE_IA" =>
          measure which iase_dfs_code to_visited is_aborted counter_example g
    | "SE_RBT" =>
          measure which se_dfs_code se_to_visited se_is_aborted se_counter_example g
    |  _ => print ("Unknown type "^which)
  end

val _ = if MLton.isMLton then main() else ()
