module Nested_DFS (main)
where

import Nat
import qualified Nested_DFS_Export_Add as DFS
import qualified Data.Time.Clock.POSIX as POSIX
import qualified Control.Parallel.Strategies as CPS
import Text.ParserCombinators.Parsec
import Control.Monad (liftM, liftM2)
import System (getArgs)

r $| f = (CPS.$|) r CPS.rdeepseq f

getTime = POSIX.getPOSIXTime

readInt :: IO Integer
readInt = readLn

nodes = nodes' `manyTill` minus_one
    where nodes' = digit `manyTill` newline

edges = edges' `manyTill` minus_one
    where edges' = do
              a <- many1 digit
              space
              b <- many1 digit
              newline
              return $| (a, b)

minus_one = string "-1\n"

get_stuff' :: String -> Either ParseError ([String], [(String, String)])
get_stuff' = parse nodes_and_edges "stdin"
    where nodes_and_edges = liftM2 (,) nodes edges

get_stuff :: IO ([Nat], [(Nat, Nat)])
get_stuff = get_stuff' `liftM` getContents >>= \r ->
            case r of 
                 Left error -> fail (show error)
                 Right (n,e) -> return $| (map conv n, map (\(a,b) -> (conv a, conv b)) e)
            where conv = fromInteger . read

measure mf t (start,edges,accept) = do
    st <- getTime
    putStrLn $ "Creation time: " ++ show (st - t)
    let r = mf edges accept start
        cycle = fst (DFS.istate r)
        no_disc = length (DFS.nested_to_visited r)
    putStrLn $ "Cycle?: " ++ show cycle
    et <- getTime -- this needs to be here, s.t. the compiler does not execute the dfs_code AFTER getTime
    putStrLn $ "# of states visited: " ++ show no_disc
    et2 <- getTime
    putStrLn $ "Time: "  ++ show (et - st) ++ " / " ++ show (et2 - st)

create_graph = do
    num <- readInt
    t <- getTime
    (acc, edges) <- get_stuff
    let edges' = DFS.edges_to_graph edges
        acc' = DFS.list_to_set acc
    putStrLn "Creating graph representation ... please be patient."
    r <- return $ (num, length edges, length acc, (Nat 0, edges' , acc'))
    return (t, r)

main = create_graph >>= measure'
    where measure' (t, (num, e, a, g)) = do
              args <- getArgs
              let (alg, algstr) = if null args then (DFS.hpy_dfs_code, "Nested/HPY") else 
                    let s:_ = args in 
                        if s == "SE" then (DFS.se_dfs_code, "Nested/SE")
                        else (DFS.basic_dfs_code, "Nested/Courcoubetis")
              putStrLn $ "\nVertices: " ++ show num
              putStrLn $ "Edges: " ++ show e
              putStrLn $ "Accepting: " ++ show a
              putStrLn "----------------------"
              putStrLn $ "Now performing the search ... (" ++ algstr ++ ")"
              measure alg t g

