#!/bin/zsh

autoload -U colors
colors

log() {
    echo ${fg_bold[white]}">>> $*"${reset_color}
}

dist=tmpdist
dist_final=CAVA-DFS
dist_thy=$dist/CAVA-DFS
heap_name=HOL-CAVA-DFS

libs=(Binomial-Heaps \
      Collections  \
      Finger-Trees  \
      Graphs  \
      Refine_Monadic)

general=(Misc_Additions \
         Code_Additions \
         Accessible \
         Workset)

haskell=(Nested_DFS.hs Nested_DFS_Export_Add.hs)

cd `dirname $(readlink -f $0)`

if [[ -d $dist ]]; then
    log "Dist directory already existing. Deleting it."
    rm -rf $dist
fi

if [[ -d $dist_final ]]; then
    log "Final dist directory already exist. Deleting it."
    rm -rf $dist_final
fi

log "Creating dist directory"
mkdir $dist
cp README $dist

log "Copying libs"
curr=$dist/Libs
mkdir $curr
for l in $libs; do
    cp -a ../../Libs/$l $curr
done
cp README_Libs $curr

log "Patching libs"
for p in lib_fixes/*; do
    log "\t$(basename $p)"
    sh $p $curr
done

log "Copying General"
curr=$dist/General
mkdir $curr
for g in $general; do
    cp -a ../../General/${g}.thy $curr
done

log "Copying theories"
mkdir $dist_thy
for thy in ../*.thy; do
    cp -a $thy $dist_thy
done

log "Copying ROOT.ML etc"
cp -a ../ROOT.ML ../IsaMakefile $dist_thy
mkdir $dist_thy/document
cp -a ../document/root.tex $dist_thy/document

log "Copying non-haskell code"
curr=$dist_thy/code
mkdir $curr
for f in ../code/*(.); do
    if [[ $f != *Export* || $f == *.mlb ]]; then
        cp -a $f $curr
    fi
done

log "Copying haskell code"
curr=$dist_thy/code/haskell
mkdir $curr
for h in $haskell; do
    cp -a ../code/haskell/$h $curr
done

log "Copying examples"
curr=$dist_thy/code
cp -a "ex/" $curr

log "Removing old heaps and docs"
isauser=`isabelle getenv -b ISABELLE_HOME_USER`
rm -rf $isauser/browser_info/HOL/${heap_name/HOL-/}
rm -rf $isauser/heaps/`isabelle getenv -b ML_IDENTIFIER`/log/${heap_name}*

log "Running isabelle make"
pushd -q $dist_thy

    if ! isabelle make; then
        log "isabelle make failed"
        exit
    fi

    log "Copying documentation"
    for docfile in $isauser/browser_info/HOL/${heap_name/HOL-/}/*.pdf; do
        log "\t`basename $docfile`"
        cp $docfile document/
    done

    log "Verifying, that code generation works"
    tmpfile=`mktemp`
    pushd -q code
        if ./build all &> $tmpfile; then
            log "\tDoes work!"
        else
            log "\tDoes not work!"
            cat $tmpfile
        fi
        
        rm bin/*
        rm Nested_DFS
    popd -q

    rm $tmpfile
popd -q

log "Renaming dist directory: '$dist' -> '$dist_final'"
mv $dist $dist_final

log "Creating archive"
tar -czf $dist_final.tar.gz $dist_final
