theory Cycle_DFS_old
imports Main DFS_old
begin

context finite_digraph
begin

fun check_cycle :: "bool \<Rightarrow> (bool, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool" where
  "check_cycle True _ _ = True"
| "check_cycle False s e = (e \<in> set (stack s))" -- "if e is on the stack \<longrightarrow> cycle"

lemma in_stack_implies_cyclic:
  "e \<in> set (stack s) \<Longrightarrow> check_cycle S s e = True"
by (cases S) simp_all

lemma not_cyclic_not_stack[simp]:
  "check_cycle S s e = False \<Longrightarrow> e \<notin> set (stack s)"
by (cases S) simp_all

definition cdfs_invar :: "(bool, 'n) dfs_sws \<Rightarrow> bool" where
  "cdfs_invar s \<equiv> (
      (stack s \<noteq> [] \<longrightarrow> \<not> state s \<longrightarrow> (\<forall>n < length (stack s). (succs (stack s ! n) - wl s ! n) \<inter> set (drop n (stack s)) = {}))
   \<and>  (stack s \<noteq> [] \<longrightarrow> (\<forall>x \<in> finished s. \<forall>y \<in> set (stack s). \<not> x \<rightarrow>+ y))
   \<and>  (\<forall>x \<in> finished s. \<not> x \<rightarrow>+ x))"

definition cdfs :: " (bool, 'n) dfs_algorithm_invar" where
  "cdfs = \<lparr> dfs_cond = Not, dfs_action = \<lambda> S _ _. S, dfs_post = \<lambda> S _ _. S, dfs_remove = check_cycle, dfs_start = \<lambda>_. False, dfs_restrict = {}, dfs_invar = cdfs_invar \<rparr>"

lemma cdfs_simps[simp]:
  "dfs_cond cdfs S \<longleftrightarrow> \<not> S"
  "dfs_action cdfs S s n = S"
  "dfs_post cdfs S s n = S"
  "dfs_remove cdfs = check_cycle"
  "dfs_start cdfs x = False"
  "dfs_invar cdfs = cdfs_invar"
  "dfs_restrict cdfs = {}"
unfolding cdfs_def
by simp_all

lemma state_not_next:
  "dfs_constructable cdfs s \<Longrightarrow> state s \<Longrightarrow> \<not> dfs_next cdfs s s'"
unfolding dfs_next_def by auto

definition cyclic where "cyclic s \<equiv> stack s \<noteq> [] \<and> hd (stack s)\<rightarrow>+hd (stack s) \<and> state s"
definition cycle where "cycle x \<equiv> \<exists>s \<in> dfs_constr_from cdfs x. cyclic s"

lemma state_implies_cyclic:
  "dfs_constructable cdfs s \<Longrightarrow> state s \<Longrightarrow> cyclic s"
proof (induction rule: dfs_constructable_induct)
  case (remove s s' e x) hence "x \<in> V" using stack_subset_verts[OF remove(2)] by auto
  from remove have "\<not> state s" using state_not_next by force
  with remove have stack: "e \<in> set (stack s)" by simp
  hence "e \<rightarrow>\<star> x" using remove.hyps(4) tl_reachable_stack_hd[OF remove(2)] self_reachable `x \<in> V` by (cases "e = x") auto
  moreover
  from remove wl_subset_succs[OF remove(2), of 0] have "e \<in> succs x" by auto
  with succs_reachable have "x \<rightarrow>+ e" .
  ultimately have "x \<rightarrow>+ x" using reachable1_reachable_trans by metis
  with remove show ?case unfolding cyclic_def by simp
qed (simp_all add: state_not_next)

lemma cycle_generates_cyclic:
  assumes reach: "x \<rightarrow>\<star> v"
  and cycle: "v \<rightarrow>+ v"
  obtains s where "s \<in> dfs_constr_from cdfs x" "cyclic s"
proof -
  from reach obtain p where p: "ewalk x p v" using reachable_ewalk by auto
  hence "v \<in> set (ewalk_verts x p)" using ewalk_conv last_in_set[OF ewalk_verts_non_Nil] by auto

  moreover
  from cycle obtain q where q: "ewalk v q v" and "q \<noteq> []" using reachable1_def by auto
  hence "ewlast v q = v" by auto
  with `q \<noteq> []` have "v \<in> set (tl (ewalk_verts v q))" unfolding ewalk_verts_conv[OF `ewalk v q v`] using last_ConsR[of q] last_in_set[of "map snd q"] by simp

  moreover from p q have ewalk: "ewalk x (p @ q) v" by (rule ewalk_ewalk_joinI)

  ultimately have "\<not> distinct (ewalk_verts x (p@q))" using ewalk_verts_join_l[of x p q v v] by simp auto
  then obtain y ys zs where y: "distinct ys" "y \<in> set ys" "ewalk_verts x (p@q) = ys@[y]@zs" using not_distinct_split_distinct by metis
  moreover hence "ys \<noteq> []" by auto
  ultimately obtain q' r' where q': "p@q = q'@r'" "ewalk_verts x q' = ys" "ewalk x q' (last ys)" "ewalk (last ys) r' v" using split_ewalk[OF `ewalk x (p @ q) v`, of ys "[y]@zs"] by metis
  with `distinct ys` have "epath x q' (last ys)" unfolding epath_def by simp

  from `ys \<noteq> []` y(3) ewalk_verts_in_edges[OF `ewalk x (p@q) v`] have "(last ys, y) \<in> E" by (induct ys rule: rev_induct) simp_all

  have "set (ewalk_verts x q') \<inter> dfs_restrict cdfs = {}" by simp
  with `epath x q' (last ys)` show ?thesis
  proof (cases rule: epath_generate_sws[where dfs=cdfs])
    case (step s) with q' y have "y \<in> discovered s" "y \<in> set(stack s)" by auto
    
    moreover
    def s' \<equiv> "s\<lparr> wl := (hd (wl s) - {y})#tl (wl s), state := True \<rparr>"

    from `(last ys, y) \<in> E` step have "y \<in> hd (wl s)" by (simp add: succs_def)
    moreover hence "hd (wl s) \<noteq> {}" by auto
    moreover from `ys \<noteq> []` q' step have "stack s = (last ys)#rev (butlast ys)" by simp
    ultimately have "s\<lparr> wl := (hd (wl s) - {y})#tl (wl s), state := dfs_remove cdfs (state s) s y\<rparr> \<in> dfs_step cdfs s"
      apply (simp add: `wl s \<noteq> []` dfs_step_def)
      apply (intro exI[where x=y])
      by (simp add: dfs_sws.defs)
    hence "s' \<in> dfs_step cdfs s" by (simp add: `y \<in> set (stack s)` s'_def in_stack_implies_cyclic)
    hence "dfs_next cdfs s s'" unfolding dfs_next_def dfs_cond_compl_def using step by simp
    hence "s' \<in> dfs_constr_from cdfs x" using step dfs_constr_from.step by simp
    with that[OF this] state_implies_cyclic show ?thesis unfolding s'_def by force
  next
    case (finished s) hence "dfs_constructable cdfs s" "state s" by (auto intro!: dfs_constr_from_constructable)
    with finished that show ?thesis using state_implies_cyclic by simp
  qed
qed

lemma cdfs_preserves_invar[intro!]:
  "dfs_preserves_invar cdfs"
proof (rule dfs_preserves_invarI)
  fix x s s'
  assume step: "dfs_next cdfs s s'" and "dfs_invar_compl cdfs x s" 
  hence inv: "\<And>n. n < length (stack s) \<Longrightarrow> (succs (stack s ! n) - wl s ! n) \<inter> set (drop n (stack s)) = {}" 
    and inv2: "\<forall>x \<in> finished s. \<forall> y \<in> set(stack s). \<not> x \<rightarrow>+ y"
    and inv3: "\<forall>x \<in> finished s. \<not> x \<rightarrow>+ x"
    unfolding dfs_invar_compl_def
    by (auto simp add: cdfs_invar_def state_not_next)

  from `dfs_invar_compl cdfs x s` have constr: "dfs_constructable cdfs s" by auto
  with step have constr': "dfs_constructable cdfs s'" by (simp add: dfs_constructable.step)
 
  from step have "cdfs_invar s'"
  proof (cases rule: dfs_next_cases_elem)
    case (remove e x xs w ws) hence st': "stack s' \<noteq> []" by simp
    from remove inv2 inv3 have invs': "\<forall>x \<in> finished s'. \<forall> y \<in> set(stack s'). \<not> x \<rightarrow>+ y" "\<forall>x \<in> finished s'. \<not> x \<rightarrow>+ x" by simp_all
    show ?thesis
    proof (cases "state s'")
      case True with st' invs' show ?thesis unfolding cdfs_invar_def by simp
    next
      case False with remove have "e \<notin> set (stack s)" using not_cyclic_not_stack[where s=s] by simp
      moreover from inv[of 0] remove have "(succs x - w) \<inter> set (stack s) = {}" by simp
      ultimately have "(succs x - (w - {e})) \<inter> set (stack s) = {}" by auto
      with remove have "(succs (stack s' ! 0) - wl s' ! 0) \<inter> set (drop 0 (stack s')) = {}" by simp
      
      moreover
      from inv have  "\<And>n. n < length (stack s) \<Longrightarrow> n > 0 \<Longrightarrow> (succs (stack s ! n) - wl s ! n) \<inter> set (drop n (stack s)) = {}" by metis
      with remove have "\<And>n. n < length (stack s') \<Longrightarrow> n > 0 \<Longrightarrow> (succs (stack s' ! n) - wl s' ! n) \<inter> set (drop n (stack s')) = {}" by simp
    
      ultimately have "\<And>n. n < length (stack s') \<Longrightarrow> (succs (stack s' ! n) - wl s' ! n) \<inter> set (drop n (stack s')) = {}" by blast
      with False st' invs' show ?thesis unfolding cdfs_invar_def by simp
    qed
  next
    case (visit e x xs w ws) with inv3 have inv3': "\<forall>x \<in> finished s'. \<not> x \<rightarrow>+ x" by simp

    from finished_no_reach_nondiscovered[OF constr] inv2 visit have "\<forall>x \<in> finished s'. \<not> x \<rightarrow>+ e" by simp
    with inv2 visit have inv2': "\<forall>x \<in> finished s'. \<forall>y \<in> set (stack s'). \<not> x \<rightarrow>+ y" by simp

    from visit have "(succs (stack s' ! 0) - wl s' ! 0) \<inter> set (drop 0 (stack s')) = {}" by auto

    moreover
    from visit(6) have "e \<notin> set (stack s)" using stack_subset_discovered[OF constr] by auto
    with visit inv[of 0] have "(succs x - (w - {e})) \<inter> set (stack s) = {}" by auto
    with visit have "(succs (stack s' ! 1) - wl s' ! 1) \<inter> set (drop 1 (stack s')) = {}" by simp

    moreover
    from visit inv have "\<And>n. n < length (stack s) \<Longrightarrow> n > 0 \<Longrightarrow> (succs (stack s ! n) - wl s ! n) \<inter> set (drop n (stack s)) = {}" by metis
    hence "\<And>n. n < length (e#stack s) \<Longrightarrow> n > 1 \<Longrightarrow> (succs ((e#stack s) ! n) - (succs e#wl s) ! n) \<inter> set (drop n (e#stack s)) = {}" by (smt drop_Cons' list.size(4) nth_Cons_pos)
    with visit have "\<And>n. n < length (stack s') \<Longrightarrow> n > 1 \<Longrightarrow> (succs (stack s' ! n) - wl s' ! n) \<inter> set (drop n (stack s')) = {}" by auto

    ultimately have "\<And>n. n < length (stack s') \<Longrightarrow> (succs (stack s' ! n) - wl s' ! n) \<inter> set (drop n (stack s')) = {}" by smt
    with visit inv2' inv3' show ?thesis unfolding cdfs_invar_def by simp
  next
    case (empty x xs w ws) with inv[of 0] inv2 have x_succs: "succs x \<inter> set (stack s) = {}" by simp
    have fin_set_nr': "\<forall>y \<in> finished s. \<forall>x \<in> set (stack s). \<not> y \<rightarrow>\<star> x"
    proof
      fix y
      assume A: "y \<in> finished s"
      with empty stack_implies_not_finished[OF constr] have "\<forall> x \<in> set (stack s). y \<noteq> x" by auto
      moreover from inv2 A have "\<forall>x \<in> set (stack s). \<not> y \<rightarrow>+ x" by simp
      ultimately show "\<forall>x \<in> set (stack s). \<not> y \<rightarrow>\<star> x" using reachable_neq_reachable1 by metis
    qed

    have x_nr_set: "\<forall>y \<in> set (stack s). \<not> x \<rightarrow>+ y"
    proof
      fix y
      assume A: "y \<in> set (stack s)"
      show "\<not> x \<rightarrow>+ y"
      proof (rule notI)
        assume "x \<rightarrow>+ y"
        then obtain p where "ewalk x p y" "p \<noteq> []" unfolding reachable1_def by auto
        then obtain e es where "ewalk x (e#es) y" by (induct p) auto
        with ewalk_Cons_iff have "e \<in> E" "x = fst e" "ewalk (snd e) es y" by auto
        hence "snd e \<in> succs x" unfolding succs_def by simp
        hence "snd e \<in> finished s"
        proof -
          assume A:"snd e \<in> succs x"
          moreover from empty have "x \<in> finished s'" by simp
          ultimately have "snd e \<in> discovered s'" using finished_implies_succs_discovered[OF constr'] by auto
          moreover from A empty x_succs have "snd e \<notin> set (stack s')" by auto
          ultimately have "snd e \<in> finished s'" using discovered_non_stack_implies_finished[OF constr'] by simp
          with empty A x_succs show "snd e \<in> finished s" by auto
        qed

        moreover from `ewalk (snd e) es y` have "snd e \<rightarrow>\<star> y" using reachable_ewalk by auto
        ultimately show "False" using fin_set_nr' A by simp
      qed
    qed
    with inv3 empty have inv3': "\<forall>y \<in> finished s'. \<not> y \<rightarrow>+ y" by simp
    
    moreover
    from inv2 x_nr_set empty have inv2': "\<forall>y \<in> finished s'. \<forall>x \<in> set (stack s'). \<not> y \<rightarrow>+ x" by simp
   
    moreover
    have inv1': "\<forall>n < length (stack s'). (succs (stack s' ! n) - wl s' ! n) \<inter> set (drop n (stack s')) = {}"
    proof (cases "xs = []")
      case True with empty show ?thesis by simp
    next
      case False with inv empty show ?thesis by fastforce
    qed

    ultimately show ?thesis unfolding cdfs_invar_def by simp
  qed simp
  thus "dfs_invar cdfs s'" by simp
qed (simp add: cdfs_invar_def)

lemma cycle_get_cycle:
  assumes "cycle x"
  obtains v where "x \<rightarrow>\<star> v" "v \<rightarrow>+ v"
proof -
  from assms obtain s where constr: "s \<in> dfs_constr_from cdfs x" and "cyclic s" unfolding cycle_def by auto
  hence reach: "hd (stack s) \<rightarrow>+ hd (stack s)" and ne: "stack s \<noteq> []"  unfolding cyclic_def by simp_all

  from constr have c: "dfs_constructable cdfs s" using dfs_constr_from_constructable by metis
  with reach have in_d: "hd (stack s) \<in> discovered s" using stack_subset_discovered[OF c] hd_in_set[OF ne] by auto

  from constr have "start s = x" using constr_from_implies_start by blast
  hence "x \<rightarrow>\<star> hd (stack s)" using start_reach_discovered[OF c ne in_d] by simp
  with reach that show ?thesis by blast
qed

lemma cycle_is_cycle:
  "cycle x \<longleftrightarrow> (\<exists>v. x \<rightarrow>\<star> v \<and> v \<rightarrow>+ v)"
proof
  assume "cycle x" thus "\<exists>v. x \<rightarrow>\<star> v \<and> v \<rightarrow>+ v" by (metis cycle_get_cycle)
next
  assume "\<exists>v. x \<rightarrow>\<star> v \<and> v \<rightarrow>+ v"
  then guess v ..
  then obtain s where "cyclic s" "s \<in> dfs_constr_from cdfs x" using cycle_generates_cyclic by blast
  thus "cycle x" unfolding cycle_def by blast
qed

lemma dfs_fun_cycle_correct:
  assumes "x \<in> V"
  shows "dfs_fun cdfs x \<le> SPEC (\<lambda>s. state s \<longleftrightarrow> cycle x)"
unfolding dfs_fun_def
using assms
apply (intro refine_vcg)
  apply (rule wf_ws_rel)
  apply (simp add: assms dfs_invar_compl_def dfs_constructable.start cdfs_invar_def)
  apply (simp add: cdfs_preserves_invar dfs_next_in_ws_rel dfs_preserves_invar_compl)
proof -
  fix s
  assume inv: "dfs_invar_compl cdfs x s" and "\<not> dfs_cond_compl cdfs s"
  hence *: "cdfs_invar s" "stack s = [] \<or> wl s = [] \<or> \<not> dfs_cond cdfs (state s)" unfolding dfs_cond_compl_def by auto
  show "state s \<longleftrightarrow> cycle x"
  proof
    assume "state s"
    with inv state_implies_cyclic have "cyclic s" by auto
    with inv show "cycle x" unfolding cycle_def by auto
  next
    assume "cycle x"
    show "state s" using *
    proof (cases "dfs_cond cdfs (state s)")
      case True with * have "stack s = [] \<or> wl s = []" by simp
      with length_wl_eq_stack[OF dfs_invar_compl_constr[OF inv]] have "stack s = []" by auto
      with * have nr: "\<forall>x \<in> finished s. \<not> x\<rightarrow>+x" unfolding cdfs_invar_def by simp
      
      from inv have "s \<in> dfs_constr_from cdfs x" by auto
      hence start: "start s = x" using constr_from_implies_start by blast
      
      from inv `stack s = []` True have compl: "dfs_completed cdfs s" by (auto intro: dfs_completed_stackI)
      
      from `cycle x` obtain v where "x \<rightarrow>\<star> v" and "v \<rightarrow>+ v" using cycle_get_cycle by auto
      hence "x \<rightarrow>\<setminus>dfs_restrict cdfs+ v" by auto
      with compl start completed_start_finished[OF compl] completed_start_reach_finished have "v \<in> finished s" by metis
      with `v \<rightarrow>+ v` nr have "False" by simp
      thus ?thesis ..
    qed simp
  qed
qed

theorem cdfs_correct:
  assumes "x \<in> V"
  and "inres (dfs_fun cdfs x) s"
  shows "state s \<longleftrightarrow> cycle x"
using assms dfs_fun_cycle_correct
by (bestsimp dest!: pwD2)

end
end