theory DFS_Impl_old
imports 
  DFS_old 
  "../Empty_Set"
  "../../Libs/Collections/Collections" 
  "~~/src/HOL/Library/Efficient_Nat"
begin

type_synonym ('a,'b,'c) impl_sws = "('a \<times> 'b \<times> 'c \<times> 'c \<times> nat \<times> 'b list \<times> 'b list list)"

definition "idisc x \<equiv> let (S, n, d, f, c, st, wll) = x in d"
definition "ifinish x \<equiv> let (S, n, d, f, c, st, wll) = x in f"
definition "istack x \<equiv> let (S, n, d, f, c, st, wll) = x in st"
definition "istate x \<equiv> let (S, n, d, f, c, st, wll) = x in S"
definition "icounter x \<equiv> let (S, n, d, f, c, st, wll) = x in c"
definition "iwll x \<equiv> let (S, n, d, f, c, st, wll) = x in wll"
definition "istart x \<equiv> let (S, n, d, f, c, st, wll) = x in n"

lemma iaccess_simps [simp]:
  "idisc (S, n, d, f, c, st, wll) = d"
  "ifinish (S, n, d, f, c, st, wll) = f"
  "istack (S, n, d, f, c, st, wll) = st"
  "istate (S, n, d, f, c, st, wll) = S"
  "icounter (S, n, d, f, c, st, wll) = c"
  "iwll (S, n, d, f, c, st, wll) = wll"
  "istart (S, n, d, f, c, st, wll) = n"
by (simp_all add: idisc_def ifinish_def istack_def istate_def icounter_def iwll_def istart_def)

record ('S, 'n, 'm, 's) dfs_algorithm_impl =
  dfs_impl_cond :: "'S \<Rightarrow> bool"  -- "the condition to be satisified by the state S to continue searching from here."
  dfs_impl_action :: "'S \<Rightarrow> ('S,'n,'m) impl_sws \<Rightarrow> 'n \<Rightarrow> 'S"  -- "modifies the state for the current node BEFORE visiting the successors."
  dfs_impl_post :: "'S \<Rightarrow> ('S,'n,'m) impl_sws \<Rightarrow> 'n \<Rightarrow> 'S"  -- "modifies the state for the current node AFTER having visited the successors (i.e. during backtracking)."
  dfs_impl_remove :: "'S \<Rightarrow> ('S,'n,'m) impl_sws  \<Rightarrow> 'n \<Rightarrow> 'S" -- "modifies the state if a node has already been visited and is removed from the stack"
  dfs_impl_start :: "'n \<Rightarrow> 'S" -- "the starting state"
  dfs_impl_restrict :: "'s" -- "the set of restricted nodes"
  dfs_impl_invar :: "('S,'n,'m) impl_sws \<Rightarrow> bool"

definition dfs_impl_state_invar where
  "dfs_impl_state_invar P \<equiv> \<lambda>s. P (istate s)"

locale DFS_Impl' = 
  fixes  dfs :: "('Sa, 'n, 'morea) dfs_algorithm_invar_scheme"
  and impl_dfs :: "('S, 'n, 'm, 's, 'more) dfs_algorithm_impl_scheme"

locale sws_impl = g: finite_digraph V E +
                  map: StdMap mops
  for mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and V :: "'n set" and E :: "('n \<times> 'n) set" +
  fixes \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> 'Sa"
begin
definition impl_sws_\<alpha> :: "('S, 'n, 'm) impl_sws \<Rightarrow> ('Sa, 'n) dfs_sws" where
  "impl_sws_\<alpha> x = (let (S, n, d, f, c, st, wll) = x in dfs_sws.make n st (map set wll) (map.\<alpha> d) (\<alpha> f) c (\<alpha>\<^sub>\<sigma> S))"

lemma impl_sws_\<alpha>_simps [simp]:
  "stack (impl_sws_\<alpha> (S, n, d, f, c, st, wll)) = st"
  "start (impl_sws_\<alpha> (S, n, d, f, c, st, wll)) = n"
  "discover (impl_sws_\<alpha> (S, n, d, f, c, st, wll)) = map.\<alpha> d"
  "finish (impl_sws_\<alpha> (S, n, d, f, c, st, wll)) = map.\<alpha> f"
  "counter (impl_sws_\<alpha> (S, n, d, f, c, st, wll)) = c"
  "state (impl_sws_\<alpha> (S, n, d, f, c, st, wll)) = \<alpha>\<^sub>\<sigma> S"
  "wl (impl_sws_\<alpha> (S, n, d, f, c, st, wll)) = map set wll"
by (simp_all add: impl_sws_\<alpha>_def dfs_sws.defs)

lemma impl_sws_\<alpha>_conv [simp]:
  "stack (impl_sws_\<alpha> x) = istack x"
  "start (impl_sws_\<alpha> x) = istart x"
  "discover (impl_sws_\<alpha> x) = map.\<alpha> (idisc x)"
  "finish (impl_sws_\<alpha> x) = map.\<alpha> (ifinish x)"
  "state (impl_sws_\<alpha> x) = \<alpha>\<^sub>\<sigma> (istate x)"
  "counter (impl_sws_\<alpha> x) = icounter x"
  "wl (impl_sws_\<alpha> x) = map set (iwll x)"
by (cases x, simp)+
end

locale DFS_Impl = DFS_Impl' dfs impl_dfs +
                  sws_impl mops V E \<alpha>\<^sub>\<sigma> +
                  set: set_memb set\<alpha> setinvar setmemb
  for dfs :: "('Sa, 'n, 'morea) dfs_algorithm_invar_scheme"
  and impl_dfs :: "('S, 'n, 'm, 's, 'more) dfs_algorithm_impl_scheme"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and set\<alpha> :: "'s \<Rightarrow> 'n set" and setinvar :: "'s \<Rightarrow> bool" and setmemb :: "'n \<Rightarrow> 's \<Rightarrow> bool"
  and V :: "'n set" and E :: "('n \<times> 'n) set"
  and \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> 'Sa" +
  fixes \<gamma>succs :: "'n \<Rightarrow> 'n list"
  assumes succs_correct: "set (\<gamma>succs y) = succs y"
  and succs_distinct: "distinct (\<gamma>succs y)"
begin

fun dfs_step_impl' :: "('S,'n,'m) impl_sws \<Rightarrow> ('S,'n,'m) impl_sws \<Rightarrow> ('S,'n,'m) impl_sws" where
 "dfs_step_impl' s (S, n, d, f, c, (x#xs), ([]#ys)) = (dfs_impl_post impl_dfs S s x, n, d, map.update x c f, c + 1, xs, ys)"
| "dfs_step_impl' s (S, n, d, f, c, (x#xs), ((z#zs)#ys)) = (if setmemb z (dfs_impl_restrict impl_dfs) then (S, n, d, f, c, (x#xs), (zs#ys))
                                                            else if map.lookup z d \<noteq> None then (dfs_impl_remove impl_dfs S s z, n, d, f, c, x # xs, zs # ys)
                                                            else (dfs_impl_action impl_dfs S s z, n, map.update z c d, f, c + 1, z # x # xs, (\<gamma>succs z) # zs # ys))"
| "dfs_step_impl' s (S, n, d, f, c, [], []) = s"

abbreviation "dfs_step_impl s \<equiv> dfs_step_impl' s s"

definition dfs_cond_impl :: "('S, 'n, 'm) impl_sws \<Rightarrow> bool" where
  "dfs_cond_impl s \<equiv> istack s \<noteq> [] \<and> iwll s \<noteq> [] \<and> dfs_impl_cond impl_dfs (istate s)"

definition dfs_start_impl :: "'n \<Rightarrow> ('S, 'n, 'm) impl_sws" where
  "dfs_start_impl x = (dfs_impl_start impl_dfs x, x, map.sng x 0, map.empty (), (1::nat), [x], [\<gamma>succs x])"

definition impl_sws_invar :: "('S, 'n, 'm) impl_sws \<Rightarrow> bool" where
  "impl_sws_invar s \<equiv> 
       dfs_constructable dfs (impl_sws_\<alpha> s) \<and>
       map.invar (ifinish s) \<and> map.invar (idisc s) \<and>
       (\<forall> w \<in> set (iwll s). distinct w \<and> set w \<subseteq> V)"
        
lemma impl_sws_invarI:
  "\<lbrakk> dfs_constructable dfs (impl_sws_\<alpha> s);
     map.invar (ifinish s); map.invar (idisc s); 
     \<And>w. w \<in> set (iwll s) \<Longrightarrow> distinct w \<and> set w \<subseteq> V \<rbrakk> \<Longrightarrow> impl_sws_invar s"
unfolding impl_sws_invar_def
by metis

lemma impl_sws_invarE:
  "\<lbrakk> impl_sws_invar s;
      \<lbrakk> dfs_constructable dfs (impl_sws_\<alpha> s);
        map.invar (ifinish s); map.invar (idisc s);
        \<And>w. w \<in> set (iwll s) \<Longrightarrow> distinct w \<and> set w \<subseteq> V \<rbrakk> \<Longrightarrow> P \<rbrakk>
  \<Longrightarrow> P"
unfolding impl_sws_invar_def
by metis

lemma impl_sws_invar_constructable:
  "impl_sws_invar s \<Longrightarrow> dfs_constructable dfs (impl_sws_\<alpha> s)"
by (metis impl_sws_invarE)

lemmas impl_sws_invar_constructableE[elim] = impl_sws_invar_constructable[elim_format]

lemma impl_sws_invar_mapinvars:
  "impl_sws_invar s \<Longrightarrow> map.invar (idisc s)"
  "impl_sws_invar s \<Longrightarrow> map.invar (ifinish s)"
unfolding impl_sws_invar_def
by metis+

lemma impl_sws_invar_dfs_invar:
  "dfs_preserves_invar dfs \<Longrightarrow> impl_sws_invar s \<Longrightarrow> dfs_invar dfs (impl_sws_\<alpha> s)"
by auto

definition impl_preserves_invar where
  "impl_preserves_invar \<equiv> (\<forall>x \<in> V. x \<notin> set\<alpha> (dfs_impl_restrict impl_dfs) \<and> impl_sws_invar (dfs_start_impl x) \<longrightarrow> 
                                   dfs_impl_invar impl_dfs (dfs_start_impl x)) \<and> 
                          (\<forall> s s'. impl_sws_invar s \<and> dfs_cond_impl s \<and> s' = dfs_step_impl s \<and> dfs_impl_invar impl_dfs s \<longrightarrow>
                                   dfs_impl_invar impl_dfs s')"

lemma impl_preserves_invar_start:
  "impl_preserves_invar \<Longrightarrow> x \<in> V \<Longrightarrow> x \<notin> set\<alpha> (dfs_impl_restrict impl_dfs) \<Longrightarrow> impl_sws_invar (dfs_start_impl x) \<Longrightarrow> dfs_impl_invar impl_dfs (dfs_start_impl x)"
by (simp add: impl_preserves_invar_def)

lemma impl_preserves_invar_step:
  "impl_preserves_invar \<Longrightarrow> impl_sws_invar s \<Longrightarrow> dfs_cond_impl s \<Longrightarrow> dfs_impl_invar impl_dfs s \<Longrightarrow> s' = dfs_step_impl s \<Longrightarrow> dfs_impl_invar impl_dfs s'"
unfolding impl_preserves_invar_def
by blast

lemma impl_preserves_invarI:
  "\<lbrakk> \<And>v. \<lbrakk>v \<in> V; v \<notin> set\<alpha> (dfs_impl_restrict impl_dfs); impl_sws_invar (dfs_start_impl v)\<rbrakk> \<Longrightarrow> dfs_impl_invar impl_dfs (dfs_start_impl v);
     \<And>s s'. \<lbrakk> impl_sws_invar s; dfs_cond_impl s; dfs_impl_invar impl_dfs s; s' = dfs_step_impl s \<rbrakk> \<Longrightarrow> dfs_impl_invar impl_dfs s'
   \<rbrakk> \<Longrightarrow> impl_preserves_invar"
unfolding impl_preserves_invar_def
by blast

lemma state_impl_preserves_invarI:
  assumes SI: "dfs_impl_invar impl_dfs = dfs_impl_state_invar P"
  and start: "\<And>x. \<lbrakk> x \<in> V; x \<notin> set\<alpha> (dfs_impl_restrict impl_dfs) \<rbrakk> \<Longrightarrow> P (dfs_impl_start impl_dfs x)"
  and post: "\<And>s x. \<lbrakk>x \<in> set (istack s); P (istate s); impl_sws_invar s; dfs_cond_impl s\<rbrakk> \<Longrightarrow> P (dfs_impl_post impl_dfs (istate s) s x)"
  and remove: "\<And>s x. \<lbrakk>x \<in> set (hd (iwll s)); map.lookup x (idisc s) \<noteq> None; P (istate s); impl_sws_invar s; dfs_cond_impl s\<rbrakk> \<Longrightarrow> P (dfs_impl_remove impl_dfs (istate s) s x)"
  and action: "\<And>s x. \<lbrakk>x \<in> set (hd (iwll s)); map.lookup x (idisc s) = None; P (istate s); impl_sws_invar s; dfs_cond_impl s\<rbrakk> \<Longrightarrow> P (dfs_impl_action impl_dfs (istate s) s x)"
  shows "impl_preserves_invar"
proof (rule impl_preserves_invarI)
  case (goal1 v) with SI start show ?case unfolding dfs_impl_state_invar_def dfs_start_impl_def by simp
next
  note SIS[simp] = SI dfs_impl_state_invar_def
  case (goal2 s s') hence inv: "P (istate s)" and "length (istack s) = length (iwll s)" using length_wl_eq_stack[OF impl_sws_invar_constructable] by simp_all
  with goal2(4,1-3) show ?case
  proof (induction s rule: dfs_step_impl'.induct)
    case (goal1 s') hence "x \<in> set (x#xs)" by simp
    with post[OF _ goal1(5)] goal1 show ?case by simp
  next
    case (goal2 s') hence "z \<in> set (hd ((z#zs)#ys))" by auto
    with goal2 remove[OF _ _ goal2(5)] action[OF _ _ goal2(5)] show ?case by simp
  qed simp_all
qed
end

locale DFS_Impl_correct = DFS_Impl dfs impl_dfs mops set\<alpha> setinvar setmemb V E \<alpha>\<^sub>\<sigma> \<gamma>succs
  for dfs :: "('Sa, 'n, 'morea) dfs_algorithm_invar_scheme"
  and impl_dfs :: "('S, 'n, 'm, 's, 'more) dfs_algorithm_impl_scheme"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and set\<alpha> :: "'s \<Rightarrow> 'n set" and setinvar :: "'s \<Rightarrow> bool" and setmemb :: "'n \<Rightarrow> 's \<Rightarrow> bool"
  and V :: "'n set" and E :: "('n \<times> 'n) set"
  and \<alpha>\<^sub>\<sigma> :: "'S \<Rightarrow> 'Sa"
  and \<gamma>succs :: "'n \<Rightarrow> 'n list"+
  assumes action_correct: "\<lbrakk>x \<in> set (hd (iwll s)); map.lookup x (idisc s) = None; dfs_impl_invar impl_dfs s; impl_sws_invar s; dfs_cond_impl s \<rbrakk> \<Longrightarrow> \<alpha>\<^sub>\<sigma> (dfs_impl_action impl_dfs (istate s) s x) = dfs_action dfs (\<alpha>\<^sub>\<sigma> (istate s)) (impl_sws_\<alpha> s) x"
  and post_correct: "\<lbrakk>x \<in> set (istack s); dfs_impl_invar impl_dfs s; impl_sws_invar s; dfs_cond_impl s \<rbrakk> \<Longrightarrow> \<alpha>\<^sub>\<sigma> (dfs_impl_post impl_dfs (istate s) s x) = dfs_post dfs (\<alpha>\<^sub>\<sigma> (istate s)) (impl_sws_\<alpha> s) x"
  and remove_correct: "\<lbrakk>x \<in> set (hd (iwll s)); map.lookup x (idisc s) \<noteq> None; dfs_impl_invar impl_dfs s; impl_sws_invar s; dfs_cond_impl s \<rbrakk> \<Longrightarrow> \<alpha>\<^sub>\<sigma> (dfs_impl_remove impl_dfs (istate s) s x) = dfs_remove dfs (\<alpha>\<^sub>\<sigma> (istate s)) (impl_sws_\<alpha> s) x"
  and start_correct: "\<lbrakk>x \<in> V; x \<notin> set\<alpha> (dfs_impl_restrict impl_dfs)\<rbrakk> \<Longrightarrow> \<alpha>\<^sub>\<sigma> (dfs_impl_start impl_dfs x) = dfs_start dfs x"
  and cond_correct: "dfs_impl_cond impl_dfs S = dfs_cond dfs (\<alpha>\<^sub>\<sigma> S)"
  and restr_correct: "set\<alpha> (dfs_impl_restrict impl_dfs) = dfs_restrict dfs"
  and restr_invar: "setinvar (dfs_impl_restrict impl_dfs)"
  and impl_preserves_invar: "impl_preserves_invar"
begin

lemma dfs_cond_impl_correct:
  "dfs_cond_impl s = dfs_cond_compl dfs (impl_sws_\<alpha> s)"
unfolding dfs_cond_impl_def dfs_cond_compl_def
by (simp add: cond_correct)

lemma dfs_cond_impl_conv:
  "dfs_cond_impl (S, n, d ,f ,c, st, wll) \<equiv> st \<noteq> [] \<and> wll \<noteq> [] \<and> dfs_impl_cond impl_dfs S"
unfolding dfs_cond_impl_def istate_def istack_def iwll_def
by simp

lemma dfs_start_impl_correct:
  "x \<in> V \<Longrightarrow> x \<notin> set\<alpha> (dfs_impl_restrict impl_dfs) \<Longrightarrow> impl_sws_\<alpha> (dfs_start_impl x) = dfs_constr_start dfs x"
unfolding dfs_start_impl_def dfs_constr_start_def
by (simp add: start_correct map.correct succs_correct impl_preserves_invar_start[OF impl_preserves_invar])

lemma dfs_start_impl_constructable:
  assumes "x \<in> V"
  and "x \<notin> set\<alpha> (dfs_impl_restrict impl_dfs)"
  shows "dfs_constructable dfs (impl_sws_\<alpha> (dfs_start_impl x))"
proof -
  from assms(2) have "x \<notin> dfs_restrict dfs" using restr_correct by simp
  with assms dfs_start_impl_correct dfs_constructable.start show ?thesis by simp
qed

lemma dfs_step_impl_correct:
  assumes invars: "impl_sws_invar s" "dfs_impl_invar impl_dfs s" "dfs_cond_impl s"
  and "s' = dfs_step_impl s"
  shows "impl_sws_\<alpha> s' \<in> dfs_step dfs (impl_sws_\<alpha> s)"
proof -
  from invars length_wl_eq_stack have "length (istack s) = length (iwll s)" by force
  moreover from invars have "istack s \<noteq> []" unfolding dfs_cond_impl_def by simp
  moreover note assms
  ultimately show ?thesis
  proof (induction (s) rule: dfs_step_impl'.induct)
    case (goal1 s') let ?s = "(S, n, d, f, c, x # xs, [] # ys)" let ?R = "impl_sws_\<alpha> ?s"
    from goal1 invars have inf :"invar f" and inv: "dfs_impl_invar impl_dfs ?s" unfolding impl_sws_invar_def by auto
    
    from goal1 have "dfs_step dfs ?R = {dfs_sws.make n xs (map set ys) (map.\<alpha> d) ((map.\<alpha> f)(x \<mapsto> c)) (Suc c) (dfs_post dfs (\<alpha>\<^sub>\<sigma> S) ?R x)}" 
      using dfs_step_simps(2)[of "?R" x xs dfs] by (simp add: dfs_sws.defs impl_sws_\<alpha>_def)
    also from goal1 post_correct[OF _ inv] update_correct[OF inf] have "... = {dfs_sws.make n xs (map set ys) (\<alpha> d) (map.\<alpha> (update x c f)) (Suc c) (\<alpha>\<^sub>\<sigma> (dfs_impl_post impl_dfs S ?s x))}" 
      by (simp add: dfs_sws.defs)
    also from goal1 have "... = { impl_sws_\<alpha> (dfs_step_impl ?s) }" by (simp add: impl_sws_\<alpha>_def)
    finally show ?case by (simp add: goal1)
  next
    case (goal2 s') let ?s = "(S, n, d, f, c, x # xs, (z # zs) # ys)" let ?R = "impl_sws_\<alpha> ?s"
    from goal2 have ind: "invar d" and inv: "dfs_impl_invar impl_dfs ?s" unfolding impl_sws_invar_def by auto
    from goal2(3) wl_subset_succs[where n= 0] succs_correct have z_in: "z \<in> set (\<gamma>succs x)" by fastforce
    
    from goal2 have "\<And>w. w \<in> set (iwll ?s) \<Longrightarrow> distinct w" "(z#zs) \<in> set (iwll ?s)" unfolding impl_sws_invar_def by force+
    hence z_dist: "distinct (z#zs)" .
    
    show ?case
    proof (cases "setmemb z (dfs_impl_restrict impl_dfs)")
      case True with restr_correct restr_invar set.memb_correct have "z \<in> dfs_restrict dfs" by blast
      with goal2 z_dist have "dfs_sws.make n (x#xs) (map set (zs#ys)) (\<alpha> d) (\<alpha> f) c (\<alpha>\<^sub>\<sigma> S) \<in> dfs_step dfs ?R"
        using dfs_step_simps(3)[of "?R" x xs dfs]
        apply simp
        apply (rule_tac x=z in exI)
        by simp
      with goal2 True show ?thesis by (simp add: impl_sws_\<alpha>_def)
    next
      case False with restr_correct restr_invar set.memb_correct have z_not_R: "z \<notin> dfs_restrict dfs" by blast
      show ?thesis
      proof (cases "lookup z d")
        case None with z_in have "set (\<gamma>succs x) - {v. lookup v d \<noteq> None} \<noteq> {}" by auto
        with ind None succs_correct z_in have "z \<notin> dom (\<alpha> d)" and "succs x - dom (\<alpha> d) \<noteq> {}" and "z \<in> succs x" by (auto simp add: correct)
        with goal2 z_not_R None z_dist have "dfs_sws.make n (z#x#xs) (map set (\<gamma>succs z#zs#ys)) ((\<alpha> d)(z \<mapsto> c)) (\<alpha> f) (Suc c) (dfs_action dfs (\<alpha>\<^sub>\<sigma> S) ?R z) \<in> dfs_step dfs ?R"
          using dfs_step_simps(3)[of "?R" x xs dfs]
          apply simp
          apply (rule_tac x=z in exI)
          by (auto simp add: succs_correct)
        with goal2 None action_correct[OF _ _ inv] update_correct[OF ind] have "dfs_sws.make n (z#x#xs) (map set (\<gamma>succs z#zs#ys)) (\<alpha> (map.update z c d)) (\<alpha> f) (Suc c) (\<alpha>\<^sub>\<sigma> (dfs_impl_action impl_dfs S ?s z)) \<in> dfs_step dfs ?R" 
          by (simp add: dfs_sws.defs)
        with None False goal2 show ?thesis by (simp add: impl_sws_\<alpha>_def)
      next
        case (Some e) with ind have "z \<in> dom (\<alpha> d)" by (auto simp add: correct)
        with goal2 Some z_not_R z_dist have "dfs_sws.make n (x#xs) (map set (zs#ys)) (\<alpha> d) (\<alpha> f) c (dfs_remove dfs (\<alpha>\<^sub>\<sigma> S) ?R z) \<in> dfs_step dfs ?R"
          using dfs_step_simps(3)[of "?R" x xs dfs]
          apply simp
          apply (rule_tac x=z in exI)
          by simp
        with remove_correct[OF _ _ inv] Some False goal2 show ?thesis by (auto simp add: impl_sws_\<alpha>_def)
      qed
    qed
  qed simp_all
qed

lemma dfs_step_impl_dfs_next:
  assumes invar: "impl_sws_invar s" "dfs_impl_invar impl_dfs s"
  and cond: "dfs_cond_impl s"
  shows "dfs_next dfs (impl_sws_\<alpha> s) (impl_sws_\<alpha> (dfs_step_impl s))"
proof -
  from cond dfs_cond_impl_correct have "dfs_cond_compl dfs (impl_sws_\<alpha> s)" by simp
  moreover from cond invar have "impl_sws_\<alpha> (dfs_step_impl s) \<in> dfs_step dfs (impl_sws_\<alpha> s)" by (simp add: dfs_step_impl_correct)
  ultimately show ?thesis unfolding dfs_next_def by simp
qed

lemma dfs_impl_constructable:
  assumes "impl_sws_invar s" "dfs_impl_invar impl_dfs s"
  and constr: "dfs_constructable dfs (impl_sws_\<alpha> s)"
  and "dfs_cond_impl s"
  shows "dfs_constructable dfs (impl_sws_\<alpha> (dfs_step_impl s))"
proof -
  from assms have "dfs_next dfs (impl_sws_\<alpha> s) (impl_sws_\<alpha> (dfs_step_impl s))" using dfs_step_impl_dfs_next by blast
  with constr show ?thesis using dfs_constructable.step by metis
qed

lemma dfs_step_impl_invars:
  assumes invars: "impl_sws_invar s" "dfs_impl_invar impl_dfs s" "dfs_cond_impl s"
  and step: "s' = dfs_step_impl s"
  shows "impl_sws_invar s'"
proof -
  from invars step dfs_impl_constructable have "dfs_constructable dfs (impl_sws_\<alpha> s')" by blast
  moreover from invars length_wl_eq_stack have "length (istack s) = length (iwll s)" by force
  moreover from invars have "istack s \<noteq> []" unfolding dfs_cond_impl_def by simp
  moreover note assms
  ultimately show ?thesis
  proof (induction s rule: dfs_step_impl'.induct)
    case (goal1 s') with goal1[unfolded impl_sws_invar_def] show ?case by (intro impl_sws_invarI) (simp_all add: map.correct)
  next
    case (goal2 s')
    note g2 = goal2[unfolded impl_sws_invar_def]
  
    show ?case
    proof (cases "setmemb z (dfs_impl_restrict impl_dfs) \<or> lookup z d \<noteq> None")
      case True with g2 show ?thesis
        apply (intro impl_sws_invarI)
        apply (case_tac [!] "setmemb z (dfs_impl_restrict impl_dfs)")
        apply auto
        done
    next
      case False hence membFalse: "\<not> setmemb z (dfs_impl_restrict impl_dfs)" "map.lookup z d = None" by simp_all
      with g2 show ?thesis
      proof (intro impl_sws_invarI)
        from g2 have *: "\<And> w. w \<in> set ((z#zs)#ys) \<Longrightarrow> distinct w \<and> set w \<subseteq> V" by auto
        moreover
        from *[of "z#zs"] distinct_tl have "distinct zs" by simp
        moreover from succs_correct have "set (\<gamma>succs z) \<subseteq> V" using succs_in_V by auto
        ultimately have "\<And>w. w \<in> set (\<gamma>succs z#zs#ys) \<Longrightarrow> distinct w \<and> set w \<subseteq> V" by (metis in_set_member member_rec(1) order_trans set_subset_Cons succs_distinct)
        thus "\<And>w. w \<in> set (iwll s') \<Longrightarrow> distinct w \<and> set w \<subseteq> V" using g2 membFalse by auto
      qed (simp_all add: map.correct)
    qed
  qed simp_all
qed

definition dfs_fun_impl :: "'n \<Rightarrow> (('S,'n,'m) impl_sws) nres" where
  "dfs_fun_impl x \<equiv> WHILE\<^sub>T dfs_cond_impl (\<lambda>s. RETURN (dfs_step_impl s)) (dfs_start_impl x)"

definition dfs_impl_build_rel where
  "dfs_impl_build_rel \<equiv> br impl_sws_\<alpha> (\<lambda>s. impl_sws_invar s \<and> dfs_impl_invar impl_dfs s)"

lemma dfs_fun_impl_refine:
  assumes "x \<in> V" "x \<notin> dfs_restrict dfs"
  shows "dfs_fun_impl x \<le> \<Down> dfs_impl_build_rel (dfs_fun dfs x)"
unfolding dfs_fun_impl_def dfs_fun_def dfs_impl_build_rel_def
apply (refine_rcg)
  apply (simp add: dfs_start_impl_correct[OF assms(1)] restr_correct assms)
  apply (rule context_conjI)
    apply (intro impl_sws_invarI)
      apply (simp add: dfs_start_impl_constructable restr_correct assms)
      apply (simp_all add: dfs_start_impl_def correct succs_distinct assms succs_correct succs_in_V)[3]
    apply (simp add: impl_preserves_invar_start[OF impl_preserves_invar] assms restr_correct)
  apply (simp add: dfs_cond_impl_correct)
  apply (rule SPEC_refine_sv)
    apply (rule br_single_valued)
    apply (simp add: dfs_step_impl_dfs_next)
    apply (rule conjI)
      apply (blast dest: dfs_step_impl_invars)
      apply (blast dest: impl_preserves_invar_step[OF impl_preserves_invar])
done

theorem dfs_fun_impl_correct:
  assumes "dfs_preserves_invar dfs" "x \<in> V" "x \<notin> dfs_restrict dfs"
  shows "dfs_fun_impl x \<le> \<Down> dfs_impl_build_rel (SPEC (\<lambda>s. s \<in> dfs_constr_from dfs x \<and> dfs_finished dfs s))"
proof -
  note dfs_fun_impl_refine[OF assms(2-)]
  also note dfs_fun_correct[OF assms]
  finally show ?thesis .
qed

lemma dfs_fun_impl_nofail[refine_pw_simps]:
  assumes "dfs_preserves_invar dfs" "x \<in> V" "x \<notin> dfs_restrict dfs"
  shows "nofail (dfs_fun_impl x)"
proof -
  note dfs_fun_nofail[OF assms]
  also note pw_conc_nofail[symmetric]
  finally show ?thesis by (blast intro: pwD dfs_fun_impl_refine[OF assms(2-)])
qed

schematic_lemma dfs_code_aux:
  "RETURN ?dfs_code \<le> dfs_fun_impl x"
unfolding dfs_fun_impl_def
by refine_transfer

definition dfs_code :: "'n \<Rightarrow> ('S,'n,'m) impl_sws" where
  "dfs_code x \<equiv> while dfs_cond_impl dfs_step_impl (dfs_start_impl x)"

lemmas dfs_code_refine = dfs_code_aux[folded dfs_code_def]

theorem dfs_code_correct:
  "\<lbrakk> dfs_preserves_invar dfs; x \<in> V; x \<notin> dfs_restrict dfs; s = dfs_code x \<rbrakk> \<Longrightarrow> impl_sws_\<alpha> s \<in> dfs_constr_from dfs x \<and> dfs_finished dfs (impl_sws_\<alpha> s) \<and> impl_sws_invar s \<and> dfs_impl_invar impl_dfs s"
using order_trans[OF dfs_code_refine dfs_fun_impl_correct, of x]
unfolding dfs_impl_build_rel_def
by (auto elim!: RETURN_ref_SPECD)

corollary dfs_code_constructable:
  "\<lbrakk> dfs_preserves_invar dfs; x \<in> V; x \<notin> dfs_restrict dfs \<rbrakk> \<Longrightarrow> impl_sws_\<alpha> (dfs_code x) \<in> dfs_constr_from dfs x"
by (metis dfs_code_correct)

corollary dfs_code_preserves_invar:
  "\<lbrakk> dfs_preserves_invar dfs; x \<in> V; x \<notin> dfs_restrict dfs \<rbrakk> \<Longrightarrow> impl_sws_invar (dfs_code x)"
by (metis dfs_code_correct)

corollary dfs_code_preserves_dfs_invar:
  "\<lbrakk> dfs_preserves_invar dfs; x \<in> V; x \<notin> dfs_restrict dfs \<rbrakk> \<Longrightarrow> dfs_impl_invar impl_dfs (dfs_code x)"
by (metis dfs_code_correct)

corollary dfs_code_finished:
  "\<lbrakk> dfs_preserves_invar dfs; x \<in> V; x \<notin> dfs_restrict dfs \<rbrakk> \<Longrightarrow> dfs_finished dfs (impl_sws_\<alpha> (dfs_code x))"
by (metis dfs_code_correct)

-- "Helper function to return the list of visited nodes."
definition to_visited :: "('S,'n,'m) impl_sws \<Rightarrow> 'n list" where
  "to_visited s = map fst (to_list (idisc s))"

lemma to_visited_correct:
  assumes invar: "invar (idisc s)" 
  shows "set (to_visited s) = discovered (impl_sws_\<alpha> s)"
proof -
  have "map_of (to_list (idisc s)) = discover (impl_sws_\<alpha> s)" using invar by (simp add: correct)
  hence "dom (map_of (to_list (idisc s))) = discovered (impl_sws_\<alpha> s)" by simp
  with dom_map_of_conv_image_fst[of "to_list (idisc s)"] show ?thesis unfolding to_visited_def by simp
qed

-- "Helper function returning a tuple (node, @{term \<delta>} node, @{term \<phi>} node)" 
definition to_fd :: "('S, 'n, 'm) impl_sws \<Rightarrow> ('n \<times> nat \<times> nat) list" where
  "to_fd s = map (\<lambda> n. let d = the (lookup n (idisc s)) in
                       let f = the (lookup n (ifinish s)) in
                       (n,d,f)) (to_visited s)"

lemma to_fd_correct:
  assumes invars: "invar (ifinish s)" "invar (idisc s)"
  and elem: "(n, d, f) \<in> set (to_fd s)"
  shows "\<phi> (impl_sws_\<alpha> s) n = f \<and> \<delta> (impl_sws_\<alpha> s) n = d"
using assms unfolding to_fd_def
by (auto simp add: to_visited_correct correct)
end

definition simple_impl_dfs :: "(unit, 'n, 'm, 'n es) dfs_algorithm_impl" where
  "simple_impl_dfs = \<lparr> dfs_impl_cond = \<lambda>_. True, 
                        dfs_impl_action = \<lambda>_ _ _. (), 
                        dfs_impl_post = \<lambda>_ _ _. (), 
                        dfs_impl_remove = \<lambda>_ _ _. (), 
                        dfs_impl_start = \<lambda>_. (), 
                        dfs_impl_restrict = es_empty (),
                        dfs_impl_invar = \<lambda>_. True\<rparr>"

lemma simple_impl_dfs_simps [simp]:
  "dfs_impl_cond simple_impl_dfs S"
  "dfs_impl_action simple_impl_dfs S s x = ()"
  "dfs_impl_post simple_impl_dfs S s x = ()"
  "dfs_impl_remove simple_impl_dfs S s x = ()"
  "dfs_impl_start simple_impl_dfs x = ()"
  "dfs_impl_restrict simple_impl_dfs = ()"
  "dfs_impl_invar simple_impl_dfs s"
by (simp_all add: simple_impl_dfs_def)

locale SimpleDFS_Impl = DFS_Impl simple_dfs simple_impl_dfs _ es_\<alpha> es_invar es_memb _ _ id
begin

lemma simple_preserves_invar:
  "impl_preserves_invar"
by (intro impl_preserves_invarI) simp_all

lemma dfs_cond_impl_conv_simple:
  "dfs_cond_impl (S, n, d, f, c, st, wll) \<equiv> st \<noteq> [] \<and> wll \<noteq> []"
unfolding dfs_cond_impl_def
by simp

lemmas dfs_start_impl_conv_simple = dfs_start_impl_def[unfolded simple_impl_dfs_simps]
lemmas dfs_step_impl_conv_simple =  dfs_step_impl'.simps[unfolded simple_impl_dfs_simps]
end

sublocale SimpleDFS_Impl \<subseteq> DFS_Impl_correct simple_dfs simple_impl_dfs _ es_\<alpha> es_invar es_memb _ _ id
by unfold_locales (simp_all add: simple_preserves_invar)

end
