theory Nested_DFS_Impl_old
imports Nested_DFS_old DFS_Impl_old
begin

locale SubDFS_Impl_def = set: StdSet sops + 
                         map: StdMap mops + 
                         finite_digraph V E
  for sops :: "('n, 's, 'Y) set_ops_scheme"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and V :: "'n set" and E :: "('n \<times> 'n) set" +
  fixes sup_st :: "'n \<Rightarrow> bool" and restr :: "'s"
  assumes rs_invar [simp]: "set.invar restr"
begin

fun check_cycle_impl where
  "check_cycle_impl True _ _ = True"
| "check_cycle_impl False _ e = sup_st e" -- "if e is on the parent (=super) stack, we have a cycle"

definition sub_impl_dfs_invar where
  "sub_impl_dfs_invar s \<equiv> set.invar (snd (istate s)) \<and> set.\<alpha> (snd (istate s)) = set.\<alpha> restr \<union> dom (map.\<alpha> (ifinish s))"

definition sub_impl_dfs :: "(bool \<times> 's, 'n, 'm, 's) dfs_algorithm_impl" where
  "sub_impl_dfs = \<lparr> dfs_impl_cond = Not \<circ> fst, 
                     dfs_impl_action = \<lambda>(f,S) s x. (check_cycle_impl f s x, S), 
                     dfs_impl_post = \<lambda>(f,S) _ x. (f, set.ins_dj x S), 
                     dfs_impl_remove = \<lambda>(f,S) s x. (x = istart s \<and> sup_st x, S), 
                     dfs_impl_start = \<lambda>_. (False, restr), 
                     dfs_impl_restrict = restr,
                     dfs_impl_invar = sub_impl_dfs_invar \<rparr>"

lemma sub_impl_dfs_simps[simp]:
  "dfs_impl_cond sub_impl_dfs S \<longleftrightarrow> \<not> (fst S)"
  "dfs_impl_action sub_impl_dfs S s x = (check_cycle_impl (fst S) s x, snd S)"
  "dfs_impl_post sub_impl_dfs S s x = (fst S, set.ins_dj x (snd S))"
  "dfs_impl_remove sub_impl_dfs S s x = (x = istart s \<and> sup_st x, snd S)"
  "dfs_impl_start sub_impl_dfs x = (False, restr)"
  "dfs_impl_restrict sub_impl_dfs = restr"
  "dfs_impl_invar sub_impl_dfs = sub_impl_dfs_invar"
unfolding sub_impl_dfs_def
by (simp_all split: prod.split)

end

locale SubDFS_Impl =  SubDFS_Impl_def sops mops V E ss rs + 
                      DFS_Impl "sub_dfs {x. ss x} (set.\<alpha> rs)" sub_impl_dfs mops set.\<alpha> set.invar set.memb V E fst
  for sops :: "('n, 's, 'Y) set_ops_scheme"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and V :: "'n set" and E :: "('n \<times> 'n) set"
  and ss :: "'n \<Rightarrow> bool" and rs :: "'s"
begin

lemma check_cycle_impl_correct:
  "check_cycle_impl S s e \<longleftrightarrow> check_cycle {x. ss x} S s' e"
by (cases S) simp_all

lemma sub_impl_preserves_invar:
  "impl_preserves_invar"
proof (rule impl_preserves_invarI)
  case goal1 thus ?case by (simp add: sub_impl_dfs_invar_def dfs_start_impl_def map.correct)
next
  case goal2 hence "length (istack s) = length (iwll s)" using length_wl_eq_stack[OF impl_sws_invar_constructable] by simp_all
  with goal2 show ?case
  proof (induction s rule: dfs_step_impl'.induct)
    case (goal1 s') let ?s = "(S, n, d, f, c, x # xs, [] # ys)" let ?\<alpha>s = "impl_sws_\<alpha> ?s"

    from goal1 have constr: "dfs_constructable (sub_dfs {x. ss x} (set.\<alpha> rs)) ?\<alpha>s" using impl_sws_invar_constructable by simp
    
    from stack_implies_not_finished[OF this] stack_not_restricted[OF this] have x_not: "x \<notin> dom (map.\<alpha> f)"  "x \<notin> set.\<alpha> rs" by (simp_all add: set.correct)

    from goal1 have "ifinish s' = map.update x c f" and  "map.invar f" using impl_sws_invar_mapinvars[where s="?s"] by simp_all
    hence "map.\<alpha> (ifinish s') = (map.\<alpha> f)(x \<mapsto> c)" by (simp add: map.correct)
    hence "dom (map.\<alpha> (ifinish s')) = insert x (dom (map.\<alpha> f))" by simp

    moreover from goal1 have "set.\<alpha> (snd S) = set.\<alpha> rs \<union> dom (map.\<alpha> f)" and S_inv: "set.invar (snd S)" by (simp_all add: sub_impl_dfs_invar_def)
    moreover with x_not have x_not_S: "x \<notin> set.\<alpha> (snd S)" by simp

    from goal1 have "snd (istate s') = set.ins_dj x (snd S)" by simp
    with S_inv x_not_S have "set.invar (snd (istate s'))" and "set.\<alpha> (snd (istate s')) = insert x (set.\<alpha> (snd S))"  by (simp_all add: set.correct)
    
    ultimately show ?case by (simp add: sub_impl_dfs_invar_def)
  qed (simp_all add: sub_impl_dfs_invar_def)
qed

end

sublocale SubDFS_Impl \<subseteq> DFS_Impl_correct "sub_dfs {x. ss x} (set.\<alpha> rs)" sub_impl_dfs mops set.\<alpha> set.invar set.memb V E fst
by unfold_locales (simp_all add: check_cycle_impl_correct sub_impl_preserves_invar)

context SubDFS_Impl
begin

abbreviation "sub_dfs_fun_impl \<equiv> dfs_fun_impl"
abbreviation "sub_dfs_code \<equiv> dfs_code"
abbreviation "sub_build_rel \<equiv> dfs_impl_build_rel"

theorem sub_dfs_fun_impl_correct:
  assumes "x \<in> V" and "x \<notin> set.\<alpha> rs"
  shows "sub_dfs_fun_impl x \<le> \<Down> sub_build_rel (SPEC (\<lambda>s. state s \<longleftrightarrow> sub_cycle {x. ss x} (set.\<alpha> rs) x))"
proof -
  from assms(2) have "x \<notin> dfs_restrict (sub_dfs {x. ss x} (set.\<alpha> rs))" by simp
  note dfs_fun_impl_refine[OF assms(1) this]
  also note dfs_fun_sub_dfs_correct[OF assms, of "{x. ss x}"]
  finally show ?thesis .
qed

lemmas sub_dfs_fun_impl_correct_unfolded = sub_dfs_fun_impl_correct[unfolded sub_cycle_iff_in_sup_st]

theorem sub_dfs_code_correct:
  "\<lbrakk> x \<in> V; x \<notin> set.\<alpha> rs; s = sub_dfs_code x \<rbrakk> \<Longrightarrow> fst (istate s) \<longleftrightarrow> sub_cycle {x. ss x} (set.\<alpha> rs) x"
using order_trans[OF dfs_code_refine sub_dfs_fun_impl_correct, of x]
unfolding dfs_impl_build_rel_def
by (auto elim!: RETURN_ref_SPECD)

lemmas sub_dfs_code_correct_unfolded = sub_dfs_code_correct[unfolded sub_cycle_iff_in_sup_st]

lemmas sub_dfs_code_preserves_invar = dfs_code_preserves_invar[OF sub_dfs_preserves_invar]
lemmas sub_dfs_code_preserves_dfs_invar = dfs_code_preserves_dfs_invar[OF sub_dfs_preserves_invar]

lemma sub_dfs_code_finished_unfolded:
  assumes x: "x \<in> V" "x \<notin> set.\<alpha> rs"
  and s_def: "s = sub_dfs_code x"
  and "\<not> fst (istate s)"
  shows "finished (impl_sws_\<alpha> s) = {v. x = v \<or> x \<rightarrow>\<setminus>set.\<alpha> rs+ v}"
proof -
  let ?DFS = "sub_dfs {x. ss x} (set.\<alpha> rs)"
  from `\<not> fst (istate s)` have "\<not> state (impl_sws_\<alpha> s)" by simp
  moreover from dfs_code_finished[OF sub_dfs_preserves_invar] x s_def have "dfs_finished ?DFS (impl_sws_\<alpha> s)" by auto
  ultimately have "dfs_completed ?DFS (impl_sws_\<alpha> s)" unfolding dfs_completed_def by simp
  note completed_finished_eq_reachable[OF this]
  moreover from dfs_code_constructable[OF sub_dfs_preserves_invar] x s_def have "impl_sws_\<alpha> s \<in> dfs_constr_from ?DFS x" by simp
  with constr_from_implies_start have "start (impl_sws_\<alpha> s) = x" by blast
  ultimately
  show "finished (impl_sws_\<alpha> s) = {v. x = v \<or> x \<rightarrow>\<setminus>set.\<alpha> rs+ v}" by auto
qed
end

locale NestedDFS_pre = set: StdSet sops +
                       setA: set_memb setA\<alpha> setAinvar setAmemb +
                       sws_impl mops V E \<sigma>\<alpha> +
                       finite_accepting_digraph V E "setA\<alpha> \<A>"
  for sops :: "('n, 's, 'Y) set_ops_scheme"
  and setA\<alpha> :: "'sA \<Rightarrow> 'n set" and setAinvar :: "'sA \<Rightarrow> bool" and setAmemb :: "'n \<Rightarrow> 'sA \<Rightarrow> bool"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and \<sigma>\<alpha> :: "bool \<times> 's \<Rightarrow> bool \<times> 'n set"
  and V::"'n set" and E :: "('n \<times> 'n) set" and \<A> :: "'sA"+
  fixes \<gamma>succs :: "'n \<Rightarrow> 'n list"
  defines \<sigma>\<alpha>_def: "\<sigma>\<alpha> \<equiv> \<lambda>(b,s). (b, set.\<alpha> s)"
  assumes ssuccs_correct: "set (\<gamma>succs y) = succs y"
  and ssuccs_distinct: "distinct (\<gamma>succs y)"
  and A_invar: "setAinvar \<A>"
begin

lemma \<sigma>\<alpha>_simps[simp]:
  "fst (\<sigma>\<alpha> S) \<longleftrightarrow> fst S"
  "snd (\<sigma>\<alpha> S) = set.\<alpha> (snd S)"
  "\<sigma>\<alpha> (b, F) = (b, set.\<alpha> F)"
unfolding \<sigma>\<alpha>_def
by (cases S, simp_all)+

end

locale NestedDFS_Impl_def = NestedDFS_pre sops _ _ _ mops 
  for sops :: "('n, 's, 'Y) set_ops_scheme"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme" +
  fixes M :: "(bool \<times> 'n set, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool"
  and \<gamma>M :: "(bool \<times> 's, 'n, 'm) impl_sws \<Rightarrow> 'n \<Rightarrow> bool"
  assumes M_correct: "\<gamma>M s = M (impl_sws_\<alpha> s)"
  and M_defs: "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))" "\<And>s. stack s \<noteq> [] \<Longrightarrow> \<forall>x. M s x \<longrightarrow> x \<in> set (stack s)"
begin

declare A_invar[simp]

definition sub_impl :: "('n \<Rightarrow> bool) \<Rightarrow> 's \<Rightarrow> 'n \<Rightarrow> 's option"
  where "sub_impl ss R x \<equiv> let s = SubDFS_Impl.sub_dfs_code \<gamma>succs sops mops ss R x
                           in if fst (istate s) then None else Some (snd (istate s))"

fun run_sub_dfs_impl where
  "run_sub_dfs_impl (True, F) _ _ = (True, F)"
| "run_sub_dfs_impl (False, F) s e = (if setAmemb e \<A> then 
                                       case sub_impl (\<gamma>M s) F e of
                                         None \<Rightarrow> (True, F)
                                       | Some F' \<Rightarrow> (False, F')
                                      else (False, F))"

abbreviation nested_impl_invar' where
  "nested_impl_invar' \<equiv> set.invar \<circ> snd"

definition nested_impl_invar where
  "nested_impl_invar \<equiv> dfs_impl_state_invar nested_impl_invar'"

lemma nested_impl_invar_conv[simp]:
  "nested_impl_invar s \<longleftrightarrow> set.invar (snd (istate s))"
unfolding nested_impl_invar_def dfs_impl_state_invar_def
by simp

definition nested_impl_dfs :: "(bool \<times> 's, 'n, 'm, 'n es) dfs_algorithm_impl" where
  "nested_impl_dfs = \<lparr> dfs_impl_cond = \<lambda>S. \<not> fst S,
                        dfs_impl_action = \<lambda>S _ _. S, 
                        dfs_impl_post = run_sub_dfs_impl, 
                        dfs_impl_remove = \<lambda>S _ _. S, 
                        dfs_impl_start = \<lambda>x. (False, set.empty ()),
                        dfs_impl_restrict = es_empty (),
                        dfs_impl_invar = nested_impl_invar \<rparr>"

lemma nested_impl_dfs_simps[simp]:
  "dfs_impl_cond nested_impl_dfs S \<longleftrightarrow> \<not> fst S"
  "dfs_impl_post nested_impl_dfs = run_sub_dfs_impl"
  "dfs_impl_action nested_impl_dfs S s x = S"
  "dfs_impl_remove nested_impl_dfs S s x = S"
  "dfs_impl_start nested_impl_dfs x = (False, set.empty ())"
  "dfs_impl_restrict nested_impl_dfs = es_empty ()"
  "dfs_impl_invar nested_impl_dfs = nested_impl_invar"
unfolding nested_impl_dfs_def
by simp_all

lemma sub_impl_correct:
  assumes inv: "set.invar R"
  and x: "x \<in> V" "x \<notin> set.\<alpha> R"
  shows "Option.map (set.\<alpha>) (sub_impl ss R x) = Option.map (\<lambda>s. s \<union> set.\<alpha> R) (run_sub_dfs' {x. ss x} (set.\<alpha> R) x)"
proof -
  from inv interpret sub!: SubDFS_Impl \<gamma>succs sops mops V E ss R  by unfold_locales (simp_all add: ssuccs_distinct ssuccs_correct)

  show ?thesis
  proof (cases "run_sub_dfs' {x. ss x} (set.\<alpha> R) x")
    case None with run_sub_dfs'_correct[OF x] have "\<exists>v. x \<rightarrow>\<setminus>set.\<alpha> R+ v \<and> ss v" by simp
    hence "fst (istate (sub.sub_dfs_code x))" using sub.sub_dfs_code_correct_unfolded[OF x] by blast
    with None show ?thesis unfolding sub_impl_def by simp
  next
    case (Some F) def s \<equiv> "sub.sub_dfs_code x"
    hence inv: "sub.impl_sws_invar s" using sub.sub_dfs_code_preserves_invar x by simp

    from s_def have *: "set.\<alpha> (snd (istate s)) = dom (map.\<alpha> (ifinish s)) \<union> set.\<alpha> R" 
      using sub.sub_dfs_code_preserves_dfs_invar x by (auto simp add: sub.sub_impl_dfs_invar_def)

    from Some have "run_sub_dfs' {x. ss x} (set.\<alpha> R) x \<noteq> None" by simp
    with run_sub_dfs'_correct[OF x] have nis: "\<not> fst (istate s)" using sub.sub_dfs_code_correct_unfolded[OF x] s_def by auto
    with sub.sub_dfs_code_finished_unfolded x s_def have "finished (sub.impl_sws_\<alpha> s) = {v. x = v \<or> x \<rightarrow>\<setminus>set.\<alpha> R+ v}" by simp
    with sub.impl_sws_invar_mapinvars[OF inv] have "dom (map.\<alpha> (ifinish s)) = {v. x = v \<or> x \<rightarrow>\<setminus>set.\<alpha> R+ v}" by (simp add: map.correct)
    with nis Some show ?thesis using run_sub_dfs'_finished[OF x Some] * unfolding sub_impl_def s_def by auto
  qed
qed

lemma sub_impl_set_invar:
  assumes inv: "set.invar R"
  and x: "x \<in> V" "x \<notin> set.\<alpha> R"
  and some: "sub_impl ss R x = Some F'"
  shows "set.invar F'"
proof -
  from inv interpret sub!: SubDFS_Impl \<gamma>succs sops mops V E ss R  by unfold_locales (simp_all add: ssuccs_distinct ssuccs_correct)

  def s \<equiv> "sub.sub_dfs_code x"
  hence "set.invar (snd (istate s))" using sub.sub_dfs_code_preserves_dfs_invar x by (auto simp add: sub.sub_impl_dfs_invar_def)
  with s_def some show ?thesis  unfolding sub_impl_def by (cases "fst (istate s)") simp_all
qed

lemma run_sub_dfs_impl_correct':
  assumes "x \<in> set (istack s)" "nested_impl_invar s" "nested_dfs_invar (impl_sws_\<alpha> s)"
  and constr: "dfs_constructable (nested_dfs M) (impl_sws_\<alpha> s)"
  shows "\<sigma>\<alpha> (run_sub_dfs_impl (istate s) s x) = run_sub_dfs M (\<sigma>\<alpha> (istate s)) (impl_sws_\<alpha> s) x"
proof (cases "istate s")
  case (Pair b F) hence "run_sub_dfs M (b, set.\<alpha> F) (impl_sws_\<alpha> s) x = \<sigma>\<alpha> (run_sub_dfs_impl (b, F) s x)" (is "?L = ?R")
  proof (cases b)
    case True hence "?L = (b, set.\<alpha> F)" by simp
    also from True have "... = ?R" by simp
    finally show ?thesis .
  next
    case False with Pair have "\<not> has_cycle (impl_sws_\<alpha> s)" by simp
    with assms have "set (istack s) \<inter> set.\<alpha> F = {}" unfolding nested_dfs_invar_def by (simp add: Pair)
    with assms have "x \<notin> set.\<alpha> F" by auto
    hence xnF:"x \<notin> snd (b, set.\<alpha> F)" by simp

    from False have nb: "\<not> fst (b, set.\<alpha> F)" by simp

    from constr stack_subset_verts have "set (istack s) \<subseteq> V" by fastforce
    with assms have xV: "x \<in> V" by auto

    from assms Pair have F_inv: "set.invar F" by simp
    note sub_impl_correct = sub_impl_correct[OF F_inv xV `x \<notin> set.\<alpha> F`, where ss = "\<gamma>M s"]

    show ?thesis using nb xV xnF
    proof (cases rule: run_sub_dfs_casesE[where s="impl_sws_\<alpha> s" and M=M])
      case same hence "\<not> setAmemb x \<A>" by (simp add: setA.memb_correct)
      hence "?R = (b, set.\<alpha> F)" by (simp add: False)
      with same show ?thesis by simp
    next
      case (cycle v) hence "setAmemb x \<A>" by (simp add: setA.memb_correct)
      
      moreover
      from cycle have "run_sub_dfs' {x. \<gamma>M s x} (set.\<alpha> F) x = None" by (auto simp add: set.correct M_correct)
      hence "sub_impl (\<gamma>M s) F x = None" using sub_impl_correct by simp
      
      ultimately show ?thesis using cycle False by simp
    next
      case (no_cycle F') hence "setAmemb x \<A>" by (simp add: setA.memb_correct)

      moreover
      from no_cycle have "run_sub_dfs' {x. \<gamma>M s x} (set.\<alpha> F) x = Some F'" by (simp add: set.correct M_correct)
      hence "Option.map set.\<alpha> (sub_impl (\<gamma>M s) F x) = Some (F' \<union> set.\<alpha> F)" using sub_impl_correct by simp
      ultimately show ?thesis using no_cycle False by auto
    qed
  qed
  with Pair show ?thesis by simp
qed

lemma run_sub_dfs_preserves_invar:
  assumes "x \<in> set (istack s)" "nested_impl_invar' (istate s)"
  and constr: "dfs_constructable (nested_dfs M) (impl_sws_\<alpha> s)"
  shows "nested_impl_invar' (run_sub_dfs_impl (istate s) s x)"
proof -
  obtain b F where bF: "istate s = (b, F)" by (cases "istate s") auto
  with assms have F_inv: "set.invar F" by simp
  with bF show ?thesis
  proof (cases b)
    case False with F_inv bF show ?thesis
    proof (cases "setAmemb x \<A>")
      case True with F_inv False bF show ?thesis
      proof (cases "sub_impl (\<gamma>M s) F x")
        case (Some F')
        note F_inv
        moreover from constr stack_subset_verts have "set (istack s) \<subseteq> V" by fastforce
        with assms have "x \<in> V" by auto
        moreover from constr have "nested_dfs_invar (impl_sws_\<alpha> s)" using dfs_constructable_invarI[OF _ nested_dfs_preserves_invar] M_defs by simp
        with False bF have "set (istack s) \<inter> set.\<alpha> F = {}" unfolding nested_dfs_invar_def by simp
        with assms have "x \<notin> set.\<alpha> F" by auto
        ultimately have "set.invar F'" using sub_impl_set_invar [OF _ _ _ Some] by metis+
        with Some False True bF show ?thesis by simp
      qed simp
    qed simp
  qed simp
qed

end

locale NestedDFS_Impl = NestedDFS_Impl_def _ _ _ \<sigma>\<alpha> V E _ \<gamma>succs sops mops +
                        DFS_Impl "(nested_dfs M)" nested_impl_dfs mops es_\<alpha> es_invar es_memb V E \<sigma>\<alpha> \<gamma>succs
  for sops :: "('n, 's, 'Y) set_ops_scheme"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and V::"'n set" and E :: "('n \<times> 'n) set"
  and \<gamma>succs :: "'n \<Rightarrow> 'n list"
  and \<sigma>\<alpha> :: "bool \<times> 's \<Rightarrow> bool \<times> 'n set"
begin

lemma run_sub_dfs_impl_correct:
  assumes "x \<in> set (istack s)" "nested_impl_invar s" "impl_sws_invar s"
  shows "\<sigma>\<alpha> (run_sub_dfs_impl (istate s) s x) = run_sub_dfs M (\<sigma>\<alpha> (istate s)) (impl_sws_\<alpha> s) x"
proof -
  from assms have "dfs_constructable (nested_dfs M) (impl_sws_\<alpha> s)" and "nested_dfs_invar (impl_sws_\<alpha> s)" 
    using impl_sws_invar_dfs_invar[OF nested_dfs_preserves_invar] M_defs by auto
  with run_sub_dfs_impl_correct'  assms show ?thesis by blast
qed

lemma nested_impl_preserves:
  "impl_preserves_invar"
using nested_impl_dfs_simps(7)[unfolded nested_impl_invar_def]
proof (rule state_impl_preserves_invarI)
  case goal2 hence "dfs_constructable (nested_dfs M) (impl_sws_\<alpha> s)" by auto
  note run_sub_dfs_preserves_invar[OF goal2(1,2) this]
  thus ?case by simp
qed (simp_all add: set.correct)
end

sublocale NestedDFS_Impl \<subseteq> DFS_Impl_correct "nested_dfs M" nested_impl_dfs mops es_\<alpha> es_invar es_memb V E \<sigma>\<alpha>
by unfold_locales (simp_all add: run_sub_dfs_impl_correct set.correct nested_impl_preserves)

locale BasicDFS_Impl_def = NestedDFS_pre 
begin

definition "basicM_impl s \<equiv> \<lambda>x. x = hd (istack s)"

lemma basicM_impl_correct:
  "basicM_impl s = basicM (impl_sws_\<alpha> s)"
unfolding basicM_impl_def basicM_def
by simp
end

sublocale BasicDFS_Impl_def \<subseteq> NestedDFS_Impl_def _ _ _ _ _ _ _ _ _ _ basicM basicM_impl
by unfold_locales (simp_all add: basicM_impl_correct \<sigma>\<alpha>_def)

locale BasicDFS_Impl = BasicDFS_Impl_def

sublocale BasicDFS_Impl \<subseteq> NestedDFS_Impl _ _ _ _ basicM basicM_impl
by unfold_locales (simp_all add: ssuccs_correct ssuccs_distinct)

context BasicDFS_Impl
begin

abbreviation "basic_impl_dfs \<equiv> nested_impl_dfs"
abbreviation "basic_dfs_fun_impl \<equiv> dfs_fun_impl"
abbreviation "basic_dfs_code \<equiv> dfs_code"
abbreviation "basic_build_rel \<equiv> dfs_impl_build_rel"

theorem basic_dfs_fun_impl_correct:
  assumes "x \<in> V"
  shows "basic_dfs_fun_impl x \<le> \<Down> basic_build_rel (SPEC (\<lambda>s. has_cycle s \<longleftrightarrow> cycle basicM x))"
proof -
  note dfs_fun_impl_refine[OF assms, folded basic_dfs_def]
  also note dfs_fun_basic_dfs_correct[OF assms]
  finally show ?thesis by (simp add: basic_dfs_def)
qed

lemmas basic_dfs_fun_impl_correct_unfolded = basic_dfs_fun_impl_correct[unfolded basic_cycle_iff_cycle]

lemmas basic_dfs_code_preserves_invar = dfs_code_preserves_invar[folded basic_dfs_def, OF basic_dfs_preserves_invar]
lemmas basic_dfs_code_preserves_dfs_invar = dfs_code_preserves_dfs_invar[folded basic_dfs_def, OF basic_dfs_preserves_invar]

theorem basic_dfs_code_correct:
  "\<lbrakk> x \<in> V; s = basic_dfs_code x \<rbrakk> \<Longrightarrow> fst (istate s) \<longleftrightarrow> cycle basicM x"
using order_trans[OF dfs_code_refine basic_dfs_fun_impl_correct, of x]
unfolding dfs_impl_build_rel_def
by (auto elim!: RETURN_ref_SPECD)

lemmas basic_dfs_code_correct_unfolded = basic_dfs_code_correct[unfolded basic_cycle_iff_cycle]
end

locale HPYDFS_Impl_def = NestedDFS_pre
begin

definition "hpyM_impl s \<equiv> \<lambda>x. set.memb x (set.from_list (istack s))"

lemma hpyM_impl_correct:
  "hpyM_impl s = hpyM (impl_sws_\<alpha> s)"
unfolding hpyM_impl_def hpyM_def
by (simp add: set.correct)
end

sublocale HPYDFS_Impl_def \<subseteq> NestedDFS_Impl_def _ _ _ _ _ _ _ _ _ _ hpyM hpyM_impl
by unfold_locales (simp_all add: hpyM_impl_correct \<sigma>\<alpha>_def)

locale HPYDFS_Impl = HPYDFS_Impl_def

sublocale HPYDFS_Impl \<subseteq> NestedDFS_Impl _ _ _ _ hpyM hpyM_impl
by unfold_locales (simp_all add: ssuccs_correct ssuccs_distinct)

context HPYDFS_Impl
begin

abbreviation "hpy_impl_dfs \<equiv> nested_impl_dfs"
abbreviation "hpy_dfs_fun_impl \<equiv> dfs_fun_impl"
abbreviation "hpy_dfs_code \<equiv> dfs_code"
abbreviation "hpy_build_rel \<equiv> dfs_impl_build_rel"

theorem hpy_dfs_fun_impl_correct:
  assumes "x \<in> V"
  shows "hpy_dfs_fun_impl x \<le> \<Down> hpy_build_rel (SPEC (\<lambda>s. has_cycle s \<longleftrightarrow> cycle hpyM x))"
proof -
  note dfs_fun_impl_refine[OF assms, folded hpy_dfs_def]
  also note dfs_fun_hpy_dfs_correct[OF assms]
  finally show ?thesis by (simp add: hpy_dfs_def)
qed

lemmas hpy_dfs_fun_impl_correct_unfolded = hpy_dfs_fun_impl_correct[unfolded hpy_cycle_iff_cycle]

lemmas hpy_dfs_code_preserves_invar = dfs_code_preserves_invar[folded hpy_dfs_def, OF hpy_dfs_preserves_invar]
lemmas hpy_dfs_code_preserves_dfs_invar = dfs_code_preserves_dfs_invar[folded hpy_dfs_def, OF hpy_dfs_preserves_invar]

theorem hpy_dfs_code_correct:
  "\<lbrakk> x \<in> V; s = hpy_dfs_code x \<rbrakk> \<Longrightarrow> fst (istate s) \<longleftrightarrow> cycle hpyM x"
using order_trans[OF dfs_code_refine hpy_dfs_fun_impl_correct, of x]
unfolding dfs_impl_build_rel_def
by (auto elim!: RETURN_ref_SPECD)

lemmas hpy_dfs_code_correct_unfolded = hpy_dfs_code_correct[unfolded hpy_cycle_iff_cycle]
end

locale SEDFS_Impl_def = HPYDFS_Impl_def
begin

fun SE_remove_impl :: "(bool \<times> 'b) \<Rightarrow> (bool \<times> 'b, 'a, 'e) impl_sws \<Rightarrow> 'a \<Rightarrow> (bool \<times> 'b)" where
  "SE_remove_impl (True, F) _ _ = (True, F)"
| "SE_remove_impl (False, F) s x = ((setAmemb x \<A> \<or> setAmemb (hd (istack s)) \<A>) \<and> map.lookup x (ifinish s) = None, F)"

definition SE_impl_dfs :: "(bool \<times> 'b, 'a, 'e, 'a es) dfs_algorithm_impl" where
  "SE_impl_dfs = nested_impl_dfs \<lparr> dfs_impl_remove := SE_remove_impl \<rparr>"

lemma SE_impl_dfs_simps[simp]:
  "dfs_impl_cond SE_impl_dfs S \<longleftrightarrow> \<not> fst S"
  "dfs_impl_post SE_impl_dfs = run_sub_dfs_impl"
  "dfs_impl_action SE_impl_dfs S s x = S"
  "dfs_impl_remove SE_impl_dfs = SE_remove_impl"
  "dfs_impl_start SE_impl_dfs x = (False, set.empty ())"
  "dfs_impl_restrict SE_impl_dfs = es_empty ()"
  "dfs_impl_invar SE_impl_dfs = nested_impl_invar"
unfolding SE_impl_dfs_def
by simp_all
end

locale SEDFS_Impl = SEDFS_Impl_def sops _ _ _ mops \<sigma>\<alpha> V E _ \<gamma>succs +
                    DFS_Impl SE_dfs SE_impl_dfs mops es_\<alpha> es_invar es_memb V E \<sigma>\<alpha> \<gamma>succs
  for sops :: "('n, 's, 'Y) set_ops_scheme"
  and mops :: "('n, nat, 'm, 'X) map_ops_scheme"
  and V::"'n set" and E :: "('n \<times> 'n) set"
  and \<gamma>succs :: "'n \<Rightarrow> 'n list"
  and \<sigma>\<alpha> :: "bool \<times> 's \<Rightarrow> bool \<times> 'n set"
begin

lemma SE_remove_correct:
  assumes x_in_d: "map.lookup x (idisc s) \<noteq> None"
  and inv: "impl_sws_invar s"
  shows "\<sigma>\<alpha> (SE_remove_impl (istate s) s x) = SE_remove (\<sigma>\<alpha> (istate s)) (impl_sws_\<alpha> s) x"
proof (cases "istate s")
  case (Pair b F) hence "SE_remove (b, set.\<alpha> F) (impl_sws_\<alpha> s) x = \<sigma>\<alpha> (SE_remove_impl (b, F) s x)" (is "?L = ?R")
  proof (cases b)
    case True hence "?L = (b, set.\<alpha> F)" by simp
    also from True have "... = ?R" by simp
    finally show ?thesis .
  next
    case False
    from inv have constr: "dfs_constructable SE_dfs (impl_sws_\<alpha> s)" by auto

    from inv have "map.lookup x (ifinish s) = None \<longleftrightarrow> x \<notin> finished (impl_sws_\<alpha> s)" by (auto simp add: map.correct impl_sws_invar_mapinvars)
    also have "... \<longleftrightarrow> x \<in> set (istack s)" 
    proof -
      from inv x_in_d have d: "x \<in> discovered (impl_sws_\<alpha> s)" by (auto simp add: map.correct impl_sws_invar_mapinvars)
      thus ?thesis by (bestsimp dest!: discovered_not_finished_implies_stack[OF constr d] intro!: stack_implies_not_finished[OF constr])
    qed
    finally show ?thesis by (simp add: setA.memb_correct map.correct False)
  qed
  thus ?thesis by (simp add: Pair)
qed

lemma SE_impl_preserves:
  "impl_preserves_invar"
using SE_impl_dfs_simps(7)[unfolded nested_impl_invar_def]
proof (rule state_impl_preserves_invarI)
  case goal2 hence constr: "dfs_constructable SE_dfs (impl_sws_\<alpha> s)" by auto
  show ?case
  proof (cases "has_cycle (impl_sws_\<alpha> s)")
    case False with SE_to_nested' constr have "dfs_constructable hpy_dfs (impl_sws_\<alpha> s)" by blast
    with goal2 run_sub_dfs_preserves_invar show ?thesis by (simp add: hpy_dfs_def)
  next
    case True with goal2 show ?thesis by (cases "istate s") simp
  qed
next
  case goal3 thus ?case by (cases "istate s", cases "fst (istate s)") simp_all
qed (simp_all add: set.correct)

lemma run_sub_dfs_impl_correct:
  assumes "x \<in> set (istack s)" "nested_impl_invar s" "impl_sws_invar s"
  shows "\<sigma>\<alpha> (run_sub_dfs_impl (istate s) s x) = run_sub_dfs hpyM (\<sigma>\<alpha> (istate s)) (impl_sws_\<alpha> s) x"
proof -
  from assms have constr: "dfs_constructable SE_dfs (impl_sws_\<alpha> s)" and inv: "nested_dfs_invar (impl_sws_\<alpha> s)" 
    using impl_sws_invar_dfs_invar[OF SE_dfs_preserves_invar] by auto
  show ?thesis
  proof (cases "has_cycle (impl_sws_\<alpha> s)")
    case False with constr SE_to_nested' have "dfs_constructable hpy_dfs (impl_sws_\<alpha> s)" by blast
    with inv run_sub_dfs_impl_correct' assms show ?thesis by (simp add: hpy_dfs_def)
  next
    case True thus ?thesis by (cases "istate s") simp
  qed
qed
end

sublocale SEDFS_Impl \<subseteq> DFS_Impl_correct SE_dfs SE_impl_dfs mops es_\<alpha> es_invar es_memb V E \<sigma>\<alpha>
by unfold_locales (simp_all add: run_sub_dfs_impl_correct SE_remove_correct set.correct SE_impl_preserves)

context SEDFS_Impl
begin

abbreviation "SE_dfs_fun_impl \<equiv> dfs_fun_impl"
abbreviation "SE_dfs_code \<equiv> dfs_code"
abbreviation "SE_build_rel \<equiv> dfs_impl_build_rel"

theorem SE_dfs_fun_impl_correct:
  assumes "x \<in> V"
  shows "SE_dfs_fun_impl x \<le> \<Down> SE_build_rel (SPEC (\<lambda>s. has_cycle s \<longleftrightarrow> SE_cycle x))"
proof -
  note dfs_fun_impl_refine[OF assms]
  also note dfs_fun_SE_dfs_correct[OF assms]
  finally show ?thesis by simp
qed

lemmas SE_dfs_fun_impl_correct_unfolded = SE_dfs_fun_impl_correct[unfolded SE_cycle_iff_cycle]

lemmas SE_dfs_code_preserves_invar = dfs_code_preserves_invar[OF SE_dfs_preserves_invar]
lemmas SE_dfs_code_preserves_dfs_invar = dfs_code_preserves_dfs_invar[OF SE_dfs_preserves_invar]

theorem SE_dfs_code_correct:
  "\<lbrakk> x \<in> V; s = SE_dfs_code x \<rbrakk> \<Longrightarrow> fst (istate s) \<longleftrightarrow> SE_cycle  x"
using order_trans[OF dfs_code_refine SE_dfs_fun_impl_correct, of x]
unfolding dfs_impl_build_rel_def
by (auto elim!: RETURN_ref_SPECD)

lemmas SE_dfs_code_correct_unfolded = SE_dfs_code_correct[unfolded SE_cycle_iff_cycle]
end
end
