theory Nested_DFS_old
imports DFS_old
begin

context finite_digraph
begin

fun check_cycle :: "'n set \<Rightarrow> bool \<Rightarrow> (bool, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool" where
  "check_cycle _ True _ _ = True"
| "check_cycle sup_st False _ e = (e \<in> sup_st)" -- "if e is on the parent (=super) stack, we have a cycle"

definition sub_dfs_invar where
  "sub_dfs_invar sup_st s \<equiv> \<not> state s \<longrightarrow> (finished s \<inter> sup_st \<subseteq> {start s} \<and> 
                                          set (stack s) \<inter> sup_st \<subseteq> {start s} \<and>
                                          (start s \<in> sup_st \<longrightarrow> ( (\<forall> x \<in> finished s. start s \<notin> succs x) \<and> 
                                          (\<forall> n < length (stack s). start s \<notin> wl s ! n \<longrightarrow> start s \<notin> succs (stack s ! n)))))"

lemma sub_dfs_invarI:
  "\<lbrakk> \<not> state s; finished s \<inter> sup_st \<subseteq> {start s}; set (stack s) \<inter> sup_st \<subseteq> {start s}; start s \<in> sup_st \<Longrightarrow> \<forall> x \<in> finished s. start s \<notin> succs x; start s \<in> sup_st \<Longrightarrow> \<forall> n < length (stack s). start s \<notin> wl s ! n \<longrightarrow> start s \<notin> succs (stack s ! n) \<rbrakk> \<Longrightarrow> sub_dfs_invar sup_st s"
unfolding sub_dfs_invar_def
by blast

definition sub_dfs :: "'n set \<Rightarrow> 'n set \<Rightarrow> (bool, 'n) dfs_algorithm_invar" where
  "sub_dfs sup_st R = \<lparr> dfs_cond = Not, dfs_action = check_cycle sup_st, dfs_post = \<lambda> S _ _. S, dfs_remove = \<lambda> S s x. x = start s \<and> x \<in> sup_st,  dfs_start = \<lambda>_. False, dfs_restrict = R, dfs_invar = sub_dfs_invar sup_st \<rparr>"

lemma sub_dfs_simps[simp]:
  "dfs_cond (sub_dfs ss R) S \<longleftrightarrow> \<not> S"
  "dfs_action (sub_dfs ss R) = check_cycle ss"
  "dfs_post (sub_dfs ss R) S s x = S"
  "dfs_remove (sub_dfs ss R) S s x \<longleftrightarrow> x = start s \<and> x \<in> ss"
  "dfs_start (sub_dfs ss R) x = False"
  "dfs_invar (sub_dfs ss R) = sub_dfs_invar ss"
  "dfs_restrict (sub_dfs ss R) = R"
unfolding sub_dfs_def
by (simp_all)

lemma sub_dfs_state_not_next:
  "dfs_constructable (sub_dfs ss R) s \<Longrightarrow> state s \<Longrightarrow> \<not> dfs_next (sub_dfs ss R) s s'"
unfolding dfs_next_def by auto


lemma sub_dfs_preserves_invar[intro!]:
  "dfs_preserves_invar (sub_dfs ss R)"
proof (rule dfs_preserves_invarI)
  fix x s s'
  assume step: "dfs_next (sub_dfs ss R) s s'" and inv_cl:"dfs_invar_compl (sub_dfs ss R) x s"
  hence inv1: "finished s \<inter> ss \<subseteq> {start s}" and
        inv2: "set (stack s) \<inter> ss \<subseteq> {start s}" and
        inv3: "start s \<in> ss \<longrightarrow> (\<forall> x \<in> finished s. start s \<notin> succs x)" and
        inv4: "start s \<in> ss \<longrightarrow> (\<forall> n < length (stack s). start s \<notin> wl s ! n \<longrightarrow> start s \<notin> succs (stack s ! n))"
    unfolding dfs_invar_compl_def by (auto simp add: sub_dfs_invar_def sub_dfs_state_not_next)
  note invs = inv1 inv2 inv3 inv4

  from step inv_cl have "\<not> state s" unfolding dfs_invar_compl_def by (auto simp add: sub_dfs_state_not_next)

  have "sub_dfs_invar ss s'"
  proof (cases "state s'")
    case True thus ?thesis by (simp add: sub_dfs_invar_def)
  next
    case False with step invs show ?thesis
    proof (cases rule: dfs_next_cases_elem)
      case (remove e x xs w ws) with False have e_neq_s: "start s \<in> ss \<Longrightarrow> e \<noteq> start s" by fastforce

      from remove invs False show ?thesis
      proof (intro sub_dfs_invarI)
        case goal5
        with inv4 have inv4': "\<forall> n < length (stack s). start s \<notin> wl s ! n \<longrightarrow> start s \<notin> succs (stack s ! n)" by force
        
        from remove have * : "stack s' = x#xs" "wl s' = (w - {e}) # ws" by simp_all
        with remove(1,2) show ?case  using inv4'
        proof (induct rule: stack_wl_remove_induct)
          case "0" with e_neq_s goal5 show ?case by auto
        qed (auto simp add: remove)
      qed simp_all
    next
      case (restrict e x xs w ws) with inv_cl start_not_restr have e_not_start: "e \<noteq> start s" by blast

      from restrict invs False show ?thesis
      proof (intro sub_dfs_invarI)
        case goal5
        with inv4 have inv4': "\<forall> n < length (stack s). start s \<notin> wl s ! n \<longrightarrow> start s \<notin> succs (stack s ! n)" by force
        
        from restrict have * : "stack s' = x#xs" "wl s' = (w - {e}) # ws" by simp_all
        with restrict(1,2) show ?case  using inv4'
        proof (induct rule: stack_wl_remove_induct)
          case "0" with e_not_start goal5 show ?case by auto
        qed (auto simp add: restrict)
      qed simp_all
    next
      case empty with invs show ?thesis unfolding sub_dfs_invar_def by auto
    next
      case (visit e x xs w ws) with inv_cl start_discovered have "e \<noteq> start s" by auto
      from visit False `\<not> state s` have "e \<notin> ss" by simp

      with False visit invs show ?thesis
      proof (intro sub_dfs_invarI)
        case goal5
        with inv4 have inv4': "\<forall> n < length (stack s). start s \<notin> wl s ! n \<longrightarrow> start s \<notin> succs (stack s ! n)" by force
        
        from visit have * : "stack s' = e#x#xs" "wl s' = succs e#(w - {e}) # ws" by simp_all
        with visit(1,2) show ?case
          using inv4'
        proof (induct rule: stack_wl_visit_induct)
          case case_1 with `e \<noteq> start s` visit show ?case by auto
        qed (auto simp add: visit)
      qed simp_all
    qed
  qed
  thus "dfs_invar (sub_dfs ss R) s'" by simp
qed (simp add: sub_dfs_invar_def)

definition sub_cyclic where "sub_cyclic ss s \<equiv> 
  stack s \<noteq> [] \<and> 
  ((hd (stack s) \<in> ss \<and> start s \<noteq> hd (stack s)) \<or> (start s \<in> succs (hd (stack s)) \<and> start s \<in> ss)) \<and> 
  state s"
definition sub_cycle where "sub_cycle ss R x \<equiv> \<exists>s \<in> dfs_constr_from (sub_dfs ss R) x. sub_cyclic ss s"

lemma state_implies_sub_cyclic:
  "dfs_constructable (sub_dfs ss R) s \<Longrightarrow> state s \<Longrightarrow> sub_cyclic ss s"
proof (induction rule: dfs_constructable_induct)
  case (visit s s' e x)  hence "\<not> state s" using sub_dfs_state_not_next by force
  moreover from visit have "check_cycle ss (state s) s e" by simp
  ultimately have "e \<in> ss" by simp
  moreover from visit start_discovered have "e \<noteq> start s" by auto
  ultimately show ?case using visit unfolding sub_cyclic_def by force
next
  case (remove s s' e x) hence "\<not> state s" using sub_dfs_state_not_next by force
  with remove have "e \<in> ss" "e = start s" by auto
  moreover from remove have "e \<in> succs (hd (stack s))" using wl_subset_succs[OF remove(2), of 0] by auto
  ultimately show ?case unfolding sub_cyclic_def using remove by simp
qed (simp_all add: sub_dfs_state_not_next sub_cyclic_def)

lemma sub_cycle_not_completed:
  assumes constr: "s \<in> dfs_constr_from (sub_dfs ss R) x"
  and sub: "x \<rightarrow>\<setminus>R+ v" "v \<in> ss"
  shows "\<not> dfs_completed (sub_dfs ss R) s"
proof (rule notI)
  let ?DFS = "sub_dfs ss R"
  assume compl: "dfs_completed ?DFS s" 
  hence cond: "dfs_cond ?DFS (state s)" unfolding dfs_completed_def by simp
  moreover from constr_from_invarI[OF constr] have inv: "sub_dfs_invar ss s" by auto
  ultimately have nr: "finished s \<inter> ss \<subseteq> {start s}" "start s \<in> ss \<Longrightarrow> \<forall>x\<in>finished s. start s \<notin> succs x" unfolding sub_dfs_invar_def by simp_all
  
  from constr have start: "start s = x" using constr_from_implies_start by blast
  
  from sub have x_reach: "x \<rightarrow>\<setminus>dfs_restrict ?DFS+ v" by simp
  with compl start completed_start_finished completed_reach_implies_finished have "v \<in> finished s" by metis
  thus "False"
  proof (cases "v = x")
    case True with x_reach `v \<in> ss` `v \<in> finished s` show "False"
    proof induction
      case (base y) with base nr start show "False" by blast
    next
      case (trans y z) with completed_reach_implies_finished[OF compl] `v \<in> finished s` True have "y \<in> finished s" by blast
      with trans nr start show "False" by blast
    qed
  next
    case False 
    from `v \<in> finished s` `v \<in> ss ` have "v \<in> finished s \<inter> ss" by auto
    with False start nr show "False" by auto
  qed
qed

lemma sub_cycle_generates_sub_cyclic:
  assumes reach: "x \<rightarrow>\<setminus>R+ v"
  and sub_cycle: "v \<in> ss"
  obtains s where "s \<in> dfs_constr_from (sub_dfs ss R) x" "sub_cyclic ss s"
proof -
  let ?DFS = "sub_dfs ss R"
  from reach have "x \<in> V" and "x \<notin> dfs_restrict ?DFS" using restricted_start by auto
  with exists_finished[OF _ _ sub_dfs_preserves_invar] obtain s where s: "s \<in> dfs_constr_from ?DFS x" "dfs_finished ?DFS s" by blast
  moreover with sub_cycle_not_completed assms have "\<not> dfs_completed ?DFS s" by simp
  ultimately have "\<not> dfs_cond ?DFS (state s)" using finished_is_completed_or_not_cond by blast
  hence "state s" by simp
  with s state_implies_sub_cyclic that constr_from_implies_start show ?thesis by blast
qed

lemma sub_cycle_get_in_sup_st:
  assumes "sub_cycle ss R x"
  obtains v where "x \<rightarrow>\<setminus>R+ v" "v \<in> ss"
proof -
  from assms obtain s where constr: "s \<in> dfs_constr_from (sub_dfs ss R) x" and "sub_cyclic ss s" unfolding sub_cycle_def by auto
  with constr_from_implies_start have sne: "stack s \<noteq> []" and *: "(x \<noteq> hd (stack s) \<and> hd (stack s) \<in> ss) \<or> (x \<in> succs (hd (stack s)) \<and> x \<in> ss)" unfolding sub_cyclic_def by auto

  show ?thesis
  proof (cases "x \<noteq> hd (stack s) \<and> hd (stack s) \<in> ss")
    case True with constr_from_restr_reach_stack[OF constr] sne have "x \<rightarrow>\<setminus>R+ hd (stack s)" by auto
    with True that show ?thesis by blast
  next
    case False with * have **:"x \<in> succs (hd (stack s)) \<and> x \<in> ss" by simp
    moreover from constr sne have "x \<notin> dfs_restrict (sub_dfs ss R)" "hd (stack s) \<notin> dfs_restrict (sub_dfs ss R)" 
      using constr_from_implies_start stack_not_restricted start_not_restr hd_in_set 
      by blast+
    ultimately have reach: "hd (stack s) \<rightarrow>\<setminus>R+ x" using succs_restricted by simp
    show ?thesis
    proof (cases "x = hd (stack s)")
      case True with reach that ** show ?thesis by simp
    next
      case False with constr_from_restr_reach_stack[OF constr] sne have "x \<rightarrow>\<setminus>R+ hd (stack s)" by auto
      with ** restr_reachable1_trans[OF _ reach] that show ?thesis by blast
    qed
  qed
qed

lemma sub_cycle_iff_in_sup_st:
  "sub_cycle ss R x \<longleftrightarrow> (\<exists>v. x \<rightarrow>\<setminus>R+ v \<and> v \<in> ss)"
proof
  assume "sub_cycle ss R x" thus "\<exists>v. x \<rightarrow>\<setminus>R+ v \<and> v \<in> ss" by (metis sub_cycle_get_in_sup_st)
next
  assume "\<exists>v. x \<rightarrow>\<setminus>R+ v \<and> v \<in> ss"
  then guess v ..
  then obtain s where "sub_cyclic ss s" "s \<in> dfs_constr_from (sub_dfs ss R) x" using sub_cycle_generates_sub_cyclic by blast
  thus "sub_cycle ss R x" unfolding sub_cycle_def by blast
qed

lemma dfs_fun_sub_dfs_correct:
  assumes "x \<in> V" "x \<notin> R"
  shows "dfs_fun (sub_dfs ss R) x \<le> SPEC (\<lambda>s. state s \<longleftrightarrow> sub_cycle ss R x)"
using assms
proof (intro dfs_fun_pred)
  fix s
  assume "s \<in> dfs_constr_from (sub_dfs ss R) x" "state s"
  moreover hence "sub_cyclic ss s" using state_implies_sub_cyclic by auto
  ultimately show "sub_cycle ss R x" by (auto simp: sub_cycle_def)
next
  fix s
  assume "s \<in> dfs_constr_from (sub_dfs ss R) x" 
    and  "dfs_completed (sub_dfs ss R) s" 
    and  "sub_cycle ss R x"
  hence "False" by (metis sub_cycle_iff_in_sup_st sub_cycle_not_completed)
  thus "state s" ..
qed (simp_all add: sub_dfs_preserves_invar)

theorem sub_dfs_correct:
  assumes "x \<in> V" "x \<notin> R"
  and "inres (dfs_fun (sub_dfs ss R) x) s"
  shows "state s \<longleftrightarrow> sub_cycle ss R x"
using assms dfs_fun_sub_dfs_correct
by (bestsimp dest!: pwD2)

theorems sub_dfs_correct_unfolded = sub_dfs_correct[unfolded sub_cycle_iff_in_sup_st]

corollary sub_dfs_correct_compl:
  assumes "x \<in> V" "x \<notin> R"
  and "inres (dfs_fun (sub_dfs ss R) x) s"
  shows "dfs_completed (sub_dfs ss R) s \<longleftrightarrow> \<not> sub_cycle ss R x"
proof -
  from sub_dfs_correct[OF assms] have "\<not> dfs_cond (sub_dfs ss R) (state s) \<longleftrightarrow> sub_cycle ss R x" by simp
  moreover from dfs_fun_correct assms have "dfs_finished (sub_dfs ss R) s" by (bestsimp dest!: pwD2)
  ultimately show ?thesis by (simp add: dfs_completed_def)
qed

lemmas sub_dfs_correct_compl_unfolded = sub_dfs_correct_compl[unfolded sub_cycle_iff_in_sup_st]

lemma RETURN_sub_cycle_correct:
  assumes "x \<in> V" "x \<notin> R"
  shows "RETURN (sub_cycle ss R x) = do { s \<leftarrow> dfs_fun (sub_dfs ss R) x; RETURN (state s) }"
apply (rule pw_eqI)
  apply (simp_all add: pw_bind_inres pw_bind_nofail dfs_fun_nofail[OF sub_dfs_preserves_invar] assms)
  apply (metis assms dfs_fun_nonempty sub_dfs_correct)
done

lemmas RETURN_sub_cycle_correct_unfolded = RETURN_sub_cycle_correct[unfolded sub_cycle_iff_in_sup_st]

lemma RETURN_sub_cycle_finished_correct:
  assumes "x \<in> V" "x \<notin> R"
  shows "RETURN (if sub_cycle ss R x then None else Some {v. v = x \<or> x \<rightarrow>\<setminus>R+ v}) = 
         do { s \<leftarrow> dfs_fun (sub_dfs ss R) x; RETURN (if state s then None else Some (finished s)) }" (is "?L = ?R")
proof (rule pw_eqI)
  show "nofail ?L \<longleftrightarrow> nofail ?R" by (simp_all add: pw_bind_inres pw_bind_nofail dfs_fun_nofail[OF sub_dfs_preserves_invar] assms)
next
  let ?DFS = "sub_dfs ss R"
  fix s 
  from `x \<notin> R` have restr: "x \<notin> dfs_restrict ?DFS" by simp

  show "inres ?L s \<longleftrightarrow> inres ?R s"
  proof (cases "sub_cycle ss R x")
    case True 
    from dfs_fun_nonempty obtain s' where s': "inres (dfs_fun ?DFS x) s'" by blast
    with s' True sub_dfs_correct[OF assms] show ?thesis by (auto simp add: pw_bind_inres pw_bind_nofail dfs_fun_nofail[OF sub_dfs_preserves_invar] assms)
  next
    case False 
    {
      fix s'
      assume "inres (dfs_fun ?DFS x) s'"
      moreover with dfs_fun_constructable assms restr have "s' \<in> dfs_constr_from ?DFS x" by blast
      moreover hence "start s' = x" using constr_from_implies_start by blast
      moreover note assms False
      ultimately have "dfs_completed ?DFS s'" "finished s' = {v. v = x \<or> x \<rightarrow>\<setminus>R+ v}"
        using completed_finished_eq_reachable[where dfs="?DFS"] sub_dfs_correct_compl by (simp_all)
    } note inres_R_impl = this

    show ?thesis
    proof
      assume "inres ?L s" with False have "s = Some {v. v = x \<or> x \<rightarrow>\<setminus>R+ v}" by simp
      moreover from dfs_fun_nonempty obtain s' where s': "inres (dfs_fun ?DFS x) s'" by blast
      moreover note inres_R_impl[OF this]
      ultimately show "inres ?R s" by (auto simp add: pw_bind_inres pw_bind_nofail dfs_fun_nofail[OF sub_dfs_preserves_invar] assms dfs_completed_def)
    next
      assume inres: "inres ?R s" 
      hence "s = Some {v. v = x \<or> x \<rightarrow>\<setminus>R+ v}" using inres_R_impl[unfolded dfs_completed_def] by (auto simp add: pw_bind_inres pw_bind_nofail dfs_fun_nofail[OF sub_dfs_preserves_invar] assms)
      thus "inres ?L s" using False by simp
    qed
  qed
qed

lemmas RETURN_sub_cycle_finished_correct_unfolded = RETURN_sub_cycle_finished_correct[unfolded sub_cycle_iff_in_sup_st]

end

subsection {* Now the grand loop *}

locale finite_accepting_digraph = finite_digraph V E
  for V :: "'n set" and E :: "('n \<times> 'n) set" +
  fixes \<A> :: "'n set"
begin

definition run_sub_dfs' :: "'n set \<Rightarrow> 'n set \<Rightarrow> 'n \<Rightarrow> 'n set option" where
  "run_sub_dfs' ss R e \<equiv> THE b. RETURN b \<le> do { s \<leftarrow> dfs_fun (sub_dfs ss R) e; RETURN (if state s then None else Some (finished s)) }"

lemma run_sub_dfs'_correct:
  assumes "e \<in> V" and "e \<notin> R"
  shows "run_sub_dfs' ss R e = None \<longleftrightarrow> (\<exists>v. e \<rightarrow>\<setminus>R+ v \<and> v \<in> ss)"
unfolding run_sub_dfs'_def
  apply (rule theI2)
  apply (simp_all only: RETURN_sub_cycle_finished_correct_unfolded[OF assms, symmetric]) 
  apply simp_all
done

lemma run_sub_dfs'_finished:
  assumes e: "e \<in> V" "e \<notin> R"
  and "run_sub_dfs' ss R e = Some f"
  shows "f = {v. v = e \<or> e \<rightarrow>\<setminus>R+ v}"
proof -
  note RETURN_sub_cycle_finished_correct_unfolded[OF e, of ss, symmetric]
  also from assms have "run_sub_dfs' ss R e \<noteq> None" by simp
  with run_sub_dfs'_correct[OF e] have "RETURN (if \<exists>v. e \<rightarrow>\<setminus>R+ v \<and> v \<in> ss then None else Some {v. v = e \<or> e \<rightarrow>\<setminus>R+ v}) = RETURN (Some {v. v = e \<or> e \<rightarrow>\<setminus>R+ v})"  by simp
  finally have "run_sub_dfs' ss R e = (Some {v. v = e \<or> e \<rightarrow>\<setminus>R+ v})" unfolding run_sub_dfs'_def by simp
  with assms show ?thesis by simp
qed

fun run_sub_dfs :: "((bool \<times> 'n set, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool) \<Rightarrow> bool \<times> 'n set \<Rightarrow> (bool \<times> 'n set, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool \<times> 'n set" where
  "run_sub_dfs M (True, F) _ _ = (True, F)"
| "run_sub_dfs M (False, F) s e = (if e \<in> \<A> then 
                                    case run_sub_dfs' {x. M s x} F e of
                                    None \<Rightarrow> (True, F) | Some F' \<Rightarrow> (False, F \<union> F')
                                  else (False, F))"

lemma run_sub_dfs_not_accpt:
  "e \<notin> \<A> \<Longrightarrow> run_sub_dfs M S s e = S"
by (cases  "(M, S, s, e)" rule: run_sub_dfs.cases) simp_all

abbreviation has_cycle :: "(bool \<times> 'n set, 'n) dfs_sws \<Rightarrow> bool" where
  "has_cycle x \<equiv> fst (state x)"

lemma run_sub_dfs_casesE[consumes 3, case_names same cycle no_cycle]:
  assumes "\<not> fst S"
  and "e \<in> V" "e \<notin> snd S"
  and same: "\<lbrakk> e \<notin> \<A>; run_sub_dfs M S s e = S\<rbrakk> \<Longrightarrow> P"
  and cycle: "\<And>v. \<lbrakk> e \<in> \<A>; run_sub_dfs' {x. M s x} (snd S) e = None; run_sub_dfs M S s e = (True, snd S); e \<rightarrow>\<setminus>snd S+ v; M s v\<rbrakk> \<Longrightarrow> P"
  and no_cycle: "\<And>F'. \<lbrakk> e \<in> \<A>; run_sub_dfs' {x. M s x} (snd S) e = Some F'; run_sub_dfs M S s e = (False, snd S \<union> F'); \<forall>v. e \<rightarrow>\<setminus>snd S+ v \<longrightarrow> \<not> M s v;  F' = {v. v = e \<or> e \<rightarrow>\<setminus>snd S+ v} \<rbrakk> \<Longrightarrow> P"
  shows "P"
proof (cases "e \<in> \<A>")
  case False with run_sub_dfs_not_accpt same show ?thesis by simp
next
  case True
  obtain F where F: "F = snd S" by auto
  then obtain subS where subS: "run_sub_dfs' {x. M s x} F e = subS" by auto
  show ?thesis
  proof (cases subS)
    case None with F subS assms True have "run_sub_dfs M S s e = (True, F)" by (cases "(M, S, s, e)" rule: run_sub_dfs.cases) simp_all
    moreover from assms subS None run_sub_dfs'_correct[of e F] F obtain v where "e \<rightarrow>\<setminus>snd S+ v" "M s v" by auto
    ultimately show ?thesis using cycle F True None subS by simp
  next
    case (Some F') with F subS assms True have "run_sub_dfs M S s e = (False, F \<union> F')" by (cases "(M, S, s, e)" rule: run_sub_dfs.cases) simp_all
    moreover from assms subS Some run_sub_dfs'_finished[of e F _ F'] F have  "F' = {v. v = e \<or> e \<rightarrow>\<setminus>F+ v}" by simp
    moreover from Some have "subS \<noteq> None" by simp
    with assms subS run_sub_dfs'_correct[of e F "{x. M s x}"] F have "\<forall>v. e \<rightarrow>\<setminus>F+ v \<longrightarrow> \<not> M s v" by simp
    ultimately show ?thesis using no_cycle F True Some subS by simp
  qed
qed

definition nested_dfs_invar where
  "nested_dfs_invar s \<equiv> (\<not> has_cycle s \<longrightarrow> 
                               (\<forall>x \<in> \<A> \<inter> finished s. \<not> x \<rightarrow>+ x) \<and> 
                               set (stack s) \<inter> snd (state s) = {} \<and> 
                               (\<forall>x. x \<notin> discovered s \<longrightarrow> x \<notin> snd (state s)) \<and>
                               (\<forall>x y. x \<rightarrow>+ y \<longrightarrow> x \<in> snd (state s) \<longrightarrow> y \<in> snd (state s)))"

lemma nested_dfs_invarI:
  "\<lbrakk> \<not> has_cycle s; 
     \<forall>x \<in> \<A> \<inter> finished s. \<not> x \<rightarrow>+ x; 
     set (stack s) \<inter> snd (state s) = {}; 
     \<And>x. x \<notin> discovered s \<Longrightarrow> x \<notin> snd (state s); 
     \<And>x y. x \<rightarrow>+ y \<Longrightarrow> x \<in> snd (state s) \<Longrightarrow> y \<in> snd (state s)\<rbrakk> \<Longrightarrow> nested_dfs_invar s"
unfolding nested_dfs_invar_def
by simp

definition nested_dfs :: "((bool \<times> 'n set, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> bool) \<Rightarrow> (bool \<times> 'n set, 'n) dfs_algorithm_invar" where
  "nested_dfs M = \<lparr> dfs_cond = Not \<circ> fst, dfs_action = \<lambda> S _ _. S, dfs_post = run_sub_dfs M, dfs_remove = \<lambda> S _ _. S, dfs_start = \<lambda>x. (False, {}), dfs_restrict = {}, dfs_invar = nested_dfs_invar \<rparr>"

lemma nested_dfs_simps[simp]:
  "dfs_cond (nested_dfs M) S \<longleftrightarrow> \<not> fst S"
  "dfs_post (nested_dfs M) = run_sub_dfs M"
  "dfs_action (nested_dfs M) S s x = S"
  "dfs_remove (nested_dfs M) S s x = S"
  "dfs_start (nested_dfs M) x = (False, {})"
  "dfs_invar (nested_dfs M) = nested_dfs_invar"
  "dfs_restrict (nested_dfs M) = {}"
unfolding nested_dfs_def
by simp_all

lemma nested_dfs_has_cycle_not_next:
  "dfs_constructable (nested_dfs M) s \<Longrightarrow> has_cycle s \<Longrightarrow> \<not> dfs_next (nested_dfs M) s s'"
unfolding dfs_next_def by auto

lemma nested_dfs_preserves_invar[intro!]:
  assumes M_hd: "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))"
  shows "dfs_preserves_invar (nested_dfs M)"
proof (rule dfs_preserves_invarI)
  fix y s s'
  assume step: "dfs_next (nested_dfs M) s s'" and inv_cl:"dfs_invar_compl (nested_dfs M) y s"
  hence inv: "\<forall>v \<in> \<A> \<inter> finished s. \<not> v \<rightarrow>+ v" "set (stack s) \<inter> snd (state s) = {}" 
             "\<forall>x. x \<notin> discovered s \<longrightarrow> x \<notin> snd (state s)""\<forall>x y. x \<rightarrow>+ y \<longrightarrow> x \<in> snd (state s) \<longrightarrow> y \<in> snd (state s)"
    unfolding dfs_invar_compl_def using nested_dfs_has_cycle_not_next nested_dfs_invar_def by simp_all blast+
  
  from inv_cl have constr: "dfs_constructable (nested_dfs M) s" by auto
  with step have constr': "dfs_constructable (nested_dfs M) s'" using dfs_constructable.step by blast

  have "nested_dfs_invar s'"
  proof (cases "has_cycle s'")
    case True thus ?thesis by (simp add: nested_dfs_invar_def)
  next
    case False with step inv show ?thesis
    proof (cases rule: dfs_next_cases_elem)
      case (empty x xs) hence state': "state s' = run_sub_dfs M (state s) s x" by simp
      from step constr have no_cycl: "\<not> has_cycle s" using nested_dfs_has_cycle_not_next by blast
      moreover from inv(1-3) empty have "x \<in> V" "x \<notin> snd (state s)" using stack_subset_verts[OF constr] by auto
      ultimately show ?thesis
      proof (cases rule: run_sub_dfs_casesE[of "state s" x M s])
        case same with inv no_cycl state' empty show ?thesis by (intro nested_dfs_invarI) simp_all
      next
        case cycle with False state' have "False" by simp
        thus ?thesis ..
      next
        case (no_cycle F') with inv(1-3) empty M_hd[of s] have no_x_self: "\<not> x \<rightarrow>\<setminus>snd (state s)+ x" by auto
        have no_x_stack: "\<And>y. x \<rightarrow>\<setminus>snd (state s)+ y \<Longrightarrow> y \<notin> set (stack s)"
        proof (rule notI)
          fix y
          assume x_y: "x \<rightarrow>\<setminus>snd (state s)+ y" and y_stack: "y \<in> set (stack s)"
          
          from x_y no_cycle have "\<not> M s y" by simp
          with M_hd[of s] empty y_stack have "y \<in> set (tl (stack s))" by auto
          with empty tl_reachable_stack_hd[OF constr] have "y \<rightarrow>+ x" by simp
          moreover from y_stack empty inv(2) have "y \<notin> snd (state s)" "x \<notin> snd (state s)" by auto
          ultimately have "y \<rightarrow>\<setminus>snd (state s)+ x" using reach_impl_restr_reach inv(4) by blast
          with x_y no_x_self restr_reachable1_trans show "False" by blast
        qed

        show ?thesis
        proof (intro nested_dfs_invarI)
          from no_cycle state' show "\<not> has_cycle s'" by simp
        next
          have "\<not> x \<rightarrow>+ x"
          proof (rule notI)
            assume "x \<rightarrow>+ x"
            moreover from empty inv(2) have "x \<notin> snd (state s)" by auto
            ultimately show False using no_x_self reach_impl_restr_reach inv(4) by blast
          qed 
          with no_cycle inv(1-3) empty show "\<forall>x \<in> \<A> \<inter> finished s'. \<not> x\<rightarrow>+x" by auto
        next
          from inv(2) empty have "set (stack s') \<inter> snd (state s) = {}" by simp
          moreover {
            fix y
            assume A: "y \<in> set (stack s')" "y \<in> F'"
            with no_cycle have "x = y \<or> x \<rightarrow>\<setminus>snd (state s)+ y" by auto
            hence "False"
            proof (rule disjE)
              case goal1 with A empty stack_distinct[OF constr] show "False" by simp
            next
              case goal2 with A empty no_x_stack show "False" by simp
            qed
          }
          hence "F' \<inter> set (stack s') = {}" by auto
          ultimately show "set (stack s') \<inter> snd (state s') = {}" using no_cycle state' by auto
        next
          fix y
          assume ndisc: "y \<notin> discovered s'"
          hence "y \<notin> F'"
          proof (rule contrapos_nn)
            assume "y \<in> F'"
            with no_cycle have "y = x \<or> x \<rightarrow>\<setminus>snd (state s)+ y" by simp
            with empty stack_subset_discovered[OF constr] ndisc have "x \<rightarrow>\<setminus>snd (state s)+ y" by auto
            thus "y \<in> discovered s'"
            proof induction
              case base from empty have "x \<in> finished s'" by simp
              with finished_implies_succs_discovered[OF constr']  base show ?case by auto
            next
              case (trans y z) with no_x_stack have "y \<notin> set (stack s)" by simp
              with empty have "y \<notin> set (stack s')" by simp
              with trans discovered_non_stack_implies_finished[OF constr'] have "y \<in> finished s'" by simp
              with finished_implies_succs_discovered[OF constr'] trans show ?case by auto
            qed
          qed
          with ndisc inv(3) empty have "y \<notin> F' \<union> snd (state s)" by simp
          with no_cycle state' show "y \<notin> snd (state s')" by simp
        next
          fix y z
          assume reach: "y \<rightarrow>+ z" and in_s: "y \<in> snd (state s')"
          with no_cycle state' have "y \<in> snd (state s) \<or> y \<in> F'" by simp
          thus "z \<in> snd (state s')"
          proof (rule disjE)
            assume "y \<in> snd (state s)" with reach inv(4) no_cycle state' show ?thesis by simp
          next
            assume A: "y \<in> F'" with reach no_cycle `x \<notin> snd (state s)` have y_notin: "y \<notin> snd (state s)" using restricted_target by blast
            thus ?thesis using no_cycle state'
            proof (cases "z \<in> snd (state s)")
              case False with reach_impl_restr_reach inv(4) y_notin reach have "y \<rightarrow>\<setminus>snd (state s)+ z" by blast
              with A no_cycle restr_reachable1_trans have "x \<rightarrow>\<setminus>snd (state s)+ z" by blast
              with no_cycle have "z \<in> F'" by blast
              with state' no_cycle show ?thesis by simp
            qed simp
          qed
        qed
      qed
    qed (simp_all add: nested_dfs_invar_def)
  qed
  thus "dfs_invar (nested_dfs M) s'" by simp
qed (auto simp add: nested_dfs_invar_def)

definition cyclic where "cyclic x s \<equiv> has_cycle s \<and> (\<exists>v \<in> finished s \<inter> \<A>. v \<rightarrow>+ v \<and> x \<rightarrow>\<star> v)"
definition cycle where "cycle M x \<equiv> \<exists>s \<in> dfs_constr_from (nested_dfs M) x. cyclic x s"

lemma has_cycle_implies_cyclic:
  "dfs_constructable (nested_dfs M) s \<Longrightarrow> (\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))) \<Longrightarrow> (\<And>s. stack s \<noteq> [] \<Longrightarrow> \<forall>x. M s x \<longrightarrow> x \<in> set (stack s)) \<Longrightarrow> has_cycle s \<Longrightarrow> cyclic (start s) s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x) hence state': "state s' = run_sub_dfs M (state s) s x" by simp

  from start_reachable_stack[OF empty(2)] empty have sreach: "\<And>x. x \<in> set (stack s) \<Longrightarrow> start s' \<rightarrow>\<star> x" by auto

  from empty have no_cycle: "\<not> has_cycle s" using nested_dfs_has_cycle_not_next by force
  
  moreover from empty stack_subset_verts[OF empty(2)] have "x \<in> V" by auto
  
  moreover
  have inv: "nested_dfs_invar s" using dfs_constructable_invarI[OF empty(2)] empty by auto
  with no_cycle  have "set (stack s) \<inter> snd (state s) = {}" unfolding nested_dfs_invar_def by simp
  with empty have "x \<notin> snd (state s)" by simp
  
  ultimately show ?case
  proof (cases rule: run_sub_dfs_casesE[of "state s" x M s])
    case same with state' no_cycle empty.prems have "False" by simp
    thus ?thesis ..
  next
    case (cycle v)
    thus ?thesis
    proof (cases "v = x")
      case True with cycle show ?thesis using empty sreach unfolding cyclic_def by auto
    next
      case False from cycle have "x \<rightarrow>+ v" by blast
      also from empty have "stack s \<noteq> []" by simp
      with `M s v` empty.prems(2) have "v \<in> set (stack s)" by auto
      with `v \<noteq> x` empty have"v \<in> set (tl (stack s))" by auto
      with tl_reachable_stack_hd[OF empty(2)] empty have "v \<rightarrow>+ x" by simp
      finally show ?thesis using empty cycle sreach unfolding cyclic_def by auto
    qed
  next
    case no_cycle with state' empty.prems have "False" by simp
    thus ?thesis ..
  qed
qed (simp_all add: nested_dfs_has_cycle_not_next cyclic_def)

lemma cycle_not_completed:
  assumes M_hd: "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))"
  and reach: "x \<rightarrow>\<star> v"
  and cycle: "v \<rightarrow>+ v"
  and fin: "v \<in> \<A>"
  and constr: "s \<in> dfs_constr_from (nested_dfs M) x"
  shows "\<not> dfs_completed (nested_dfs M) s"
proof (rule notI)
  assume compl: "dfs_completed (nested_dfs M) s"
  hence "\<not> has_cycle s" by (simp add: dfs_completed_def)
  
  from compl no_restrict_finished_eq_reachable[of "nested_dfs M" s] constr_from_implies_start[OF constr] have "finished s = {v. x \<rightarrow>\<star> v}" by simp
  hence "v \<in> finished s" using reach by simp
  
  moreover
  from compl M_hd have "dfs_invar (nested_dfs M) s" by blast
  with `\<not> has_cycle s` have "\<forall>v \<in> \<A> \<inter> finished s. \<not> v \<rightarrow>+ v" by (simp add: nested_dfs_invar_def)
  with cycle fin have "v \<notin> finished s" by auto
  
  ultimately show "False" by contradiction
qed

lemma cycle_generates_cyclic:
  assumes M_hd: "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))" and M_sse: "\<And>s. stack s \<noteq> [] \<Longrightarrow> \<forall>x. M s x \<longrightarrow> x \<in> set (stack s)"
  and reach: "x \<rightarrow>\<star> v"
  and cycle: "v \<rightarrow>+ v"
  and fin: "v \<in> \<A>"
  obtains s where "s \<in> dfs_constr_from (nested_dfs M) x" "cyclic x s"
proof -
  from reach have "x \<in> V" by auto
  with exists_finished[OF _ _ nested_dfs_preserves_invar[of M, OF M_hd]] obtain s where s: "s \<in> dfs_constr_from (nested_dfs M) x" "dfs_finished (nested_dfs M) s" by auto
  moreover with cycle_not_completed assms have "\<not> dfs_completed (nested_dfs M) s" by simp
  ultimately have "\<not> dfs_cond (nested_dfs M) (state s)" using finished_is_completed_or_not_cond by blast
  hence "has_cycle s" by simp
  with s has_cycle_implies_cyclic that constr_from_implies_start M_hd M_sse show ?thesis by blast
qed

lemma cycle_get_cycle:
  assumes "cycle M x"
  obtains v where "x \<rightarrow>\<star> v" "v \<in> \<A>" "v \<rightarrow>+ v"
proof -
  from assms obtain s where constr: "s \<in> dfs_constr_from (nested_dfs M) x" and "cyclic x s" unfolding cycle_def by auto
  then obtain v where "v \<in> finished s \<inter> \<A>" "v \<rightarrow>+ v" "x \<rightarrow>\<star> v" unfolding cyclic_def by blast
  with that show ?thesis by auto
qed

lemma cycle_iff_cycle:
  assumes M_correct: "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))" "\<And>s. stack s \<noteq> [] \<Longrightarrow> \<forall>x. M s x \<longrightarrow> x \<in> set (stack s)"
  shows "cycle M x \<longleftrightarrow> (\<exists>v. x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v)"
proof
  assume "cycle M x" thus "\<exists>v. x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v" by (metis cycle_get_cycle)
next
  assume "\<exists>v.  x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v"
  then guess v ..
  then obtain s where "cyclic x s" "s \<in> dfs_constr_from (nested_dfs M) x" using cycle_generates_cyclic[OF M_correct] by blast
  thus "cycle M x" unfolding cycle_def by blast
qed

lemma dfs_fun_nested_dfs_correct:
  assumes "x \<in> V"
  and M_correct: "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))" "\<And>s. stack s \<noteq> [] \<Longrightarrow> \<forall>x. M s x \<longrightarrow> x \<in> set (stack s)"
  shows "dfs_fun (nested_dfs M) x \<le> SPEC (\<lambda>s. has_cycle s \<longleftrightarrow> cycle M x)"
using assms
proof (intro dfs_fun_pred)
  fix s
  assume "s \<in> dfs_constr_from (nested_dfs M) x" "has_cycle s"
  moreover hence "cyclic x s" using has_cycle_implies_cyclic[OF _ M_correct] constr_from_implies_start by force
  ultimately show "cycle M x" by (auto simp: cycle_def)
next
  fix s
  assume "s \<in> dfs_constr_from (nested_dfs M) x" "dfs_completed (nested_dfs M) s" "cycle M x"
  hence "False" by (metis (lifting) cycle_iff_cycle M_correct cycle_not_completed)
  thus "has_cycle s" ..
qed (simp_all add: nested_dfs_preserves_invar)

theorem nested_dfs_correct:
  assumes "x \<in> V"
  and "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))" "\<And>s. stack s \<noteq> [] \<Longrightarrow> \<forall>x. M s x \<longrightarrow> x \<in> set (stack s)"
  and "inres (dfs_fun (nested_dfs M) x) s"
  shows "has_cycle s \<longleftrightarrow> cycle M x"
using assms(4) dfs_fun_nested_dfs_correct[OF assms(1-3)]
by (bestsimp dest!: pwD2)

lemma RETURN_nested_cycle_correct:
  assumes "x \<in> V"
  and "\<And>s. stack s \<noteq> [] \<Longrightarrow> M s (hd (stack s))" "\<And>s. stack s \<noteq> [] \<Longrightarrow> \<forall>x. M s x \<longrightarrow> x \<in> set (stack s)"
  shows "RETURN (cycle M x) = do { s \<leftarrow> dfs_fun (nested_dfs M) x; RETURN (has_cycle s) }"
apply (rule pw_eqI)
  apply (simp_all add: pw_bind_inres pw_bind_nofail dfs_fun_nofail[OF nested_dfs_preserves_invar[of M, OF assms(2)] assms(1)])
  apply (metis assms dfs_fun_nonempty nested_dfs_correct)
done

end

subsection {* Basic Nested DFS *}
context finite_accepting_digraph
begin

definition "basicM \<equiv> \<lambda>s. \<lambda>x. x = hd (stack s)"
definition "basic_dfs \<equiv> nested_dfs basicM"

lemmas basic_dfs_simps[simp] = nested_dfs_simps[where M=basicM, folded basic_dfs_def]

lemma basicM_correct[simp]:
  "stack s \<noteq> [] \<Longrightarrow> basicM s (hd (stack s))"
  "stack s \<noteq> [] \<Longrightarrow> \<forall>x. basicM s x \<longrightarrow> x \<in> set (stack s)"
unfolding basicM_def
by simp_all

lemma basic_dfs_preserves_invar:
  "dfs_preserves_invar basic_dfs"
by (simp add: basic_dfs_def nested_dfs_preserves_invar)

lemma basic_cycle_iff_cycle:
  "cycle basicM x \<longleftrightarrow> (\<exists>v. x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v)"
by (simp add: cycle_iff_cycle)

lemma dfs_fun_basic_dfs_correct:
  "x \<in> V \<Longrightarrow> dfs_fun basic_dfs x \<le> SPEC (\<lambda>s. has_cycle s \<longleftrightarrow> cycle basicM x)"
by (simp add: dfs_fun_nested_dfs_correct basic_dfs_def)

theorem basic_dfs_correct:
  "x \<in> V \<Longrightarrow> inres (dfs_fun basic_dfs x) s \<Longrightarrow> has_cycle s \<longleftrightarrow> cycle basicM x"
by (simp add: nested_dfs_correct basic_dfs_def)

theorems basic_dfs_correct_unfolded = basic_dfs_correct[unfolded basic_cycle_iff_cycle]
end

subsection {* HPY *}
context finite_accepting_digraph
begin

definition "hpyM \<equiv> \<lambda>s. \<lambda>x. x \<in> set (stack s)"
definition "hpy_dfs \<equiv> nested_dfs hpyM"

lemmas hpy_dfs_simps[simp] = nested_dfs_simps[where M=hpyM, folded hpy_dfs_def]

lemma hpyM_correct[simp]:
  "stack s \<noteq> [] \<Longrightarrow> hpyM s (hd (stack s))"
  "stack s \<noteq> [] \<Longrightarrow> \<forall>x. hpyM s x \<longrightarrow> x \<in> set (stack s)"
unfolding hpyM_def
by simp_all

lemma hpy_dfs_preserves_invar:
  "dfs_preserves_invar hpy_dfs"
by (simp add: hpy_dfs_def nested_dfs_preserves_invar)

lemma hpy_cycle_iff_cycle:
  "cycle hpyM x \<longleftrightarrow> (\<exists>v. x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v)"
by (simp add: cycle_iff_cycle)

lemma dfs_fun_hpy_dfs_correct:
  "x \<in> V \<Longrightarrow> dfs_fun hpy_dfs x \<le> SPEC (\<lambda>s. has_cycle s \<longleftrightarrow> cycle hpyM x)"
by (simp add: dfs_fun_nested_dfs_correct hpy_dfs_def)

theorem hpy_dfs_correct:
  "x \<in> V \<Longrightarrow> inres (dfs_fun hpy_dfs x) s \<Longrightarrow> has_cycle s \<longleftrightarrow> cycle hpyM x"
by (simp add: nested_dfs_correct hpy_dfs_def)

theorems hpy_dfs_correct_unfolded = hpy_dfs_correct[unfolded hpy_cycle_iff_cycle]
end

subsection {* Schwoon-Esparza *}

context finite_accepting_digraph
begin

fun SE_remove where 
  "SE_remove (True, F) _ _ = (True, F)"
| "SE_remove (False, F) s x = ((x \<in> \<A> \<or> hd (stack s) \<in> \<A>) \<and> x \<in> set (stack s), F)"

definition SE_dfs :: "(bool \<times> 'n set, 'n) dfs_algorithm_invar" where
  "SE_dfs = hpy_dfs\<lparr> dfs_remove := SE_remove \<rparr>"

lemma SE_dfs_simps[simp]:
  "dfs_cond SE_dfs S \<longleftrightarrow> \<not> fst S"
  "dfs_post SE_dfs = run_sub_dfs hpyM"
  "dfs_action SE_dfs S s x = S"
  "dfs_remove SE_dfs = SE_remove"
  "dfs_start SE_dfs x = (False, {})"
  "dfs_invar SE_dfs = nested_dfs_invar"
  "dfs_restrict SE_dfs = {}"
unfolding SE_dfs_def
by simp_all

lemma SE_dfs_has_cycle_not_next:
  "dfs_constructable SE_dfs s \<Longrightarrow> has_cycle s \<Longrightarrow> \<not> dfs_next SE_dfs s s'"
unfolding dfs_next_def by auto

lemma SE_to_nested:
  "s \<in> dfs_constr_from SE_dfs x \<Longrightarrow> \<not> has_cycle s \<Longrightarrow> s \<in> dfs_constr_from hpy_dfs x"
proof (induction rule: dfs_constr_from.induct)
  case start
  have "dfs_constr_start SE_dfs x = dfs_constr_start hpy_dfs x" by (simp add: dfs_constr_start_def)
  with start show ?case by (simp add: dfs_constructable.start)
next
  case (step s s') with SE_dfs_has_cycle_not_next have no: "\<not> has_cycle s" by blast
  with step.IH have constr: "s \<in> dfs_constr_from hpy_dfs x" .

  from step have "dfs_cond_compl hpy_dfs s" unfolding dfs_next_def dfs_cond_compl_def by simp
  note thms = step(1,4) no constr this
  
  from step(2) have "dfs_next hpy_dfs s s'"
  proof (cases rule: dfs_next_cases_elem)
    case empty thus ?thesis by (simp add: dfs_next_def dfs_step_def dfs_sws.defs thms)
  next
    case restrict thus ?thesis by (simp add: dfs_next_def dfs_step_def dfs_sws.defs thms)
  next
    case (remove e) 
    obtain b F where bf: "state s = (b,F)" by (cases "state s") auto
    with remove `\<not> has_cycle s'` have "state s' = (b, F)" by (cases b) simp_all
    with remove bf show ?thesis by (simp add: dfs_next_def dfs_sws.defs thms dfs_step_def exI[where x=e])
  next
    case (visit e) obtain b F where bf: "state s = (b,F)" by (cases "state s") auto
    with visit have "state s' = (b, F)" by simp
    with visit bf show ?thesis by (simp add: dfs_next_def dfs_sws.defs thms dfs_step_def exI[where x=e])
  qed
  with constr show ?case by (rule dfs_constr_from.step)
qed

corollary SE_to_nested':
  "dfs_constructable SE_dfs s \<Longrightarrow> \<not> has_cycle s \<Longrightarrow> dfs_constructable hpy_dfs s"
using SE_to_nested dfs_constructable_constr_from_start dfs_constr_from_constructable
by metis

lemma SE_dfs_preserves_invar[intro]:
  "dfs_preserves_invar SE_dfs"
proof (rule dfs_preserves_invarI)
  fix y s s'
  assume step: "dfs_next SE_dfs s s'" and inv_cl:"dfs_invar_compl SE_dfs y s"
  hence inv: "\<forall>v \<in> \<A> \<inter> finished s. \<not> v \<rightarrow>+ v" "set (stack s) \<inter> snd (state s) = {}" 
             "\<forall>x. x \<notin> discovered s \<longrightarrow> x \<notin> snd (state s)""\<forall>x y. x \<rightarrow>+ y \<longrightarrow> x \<in> snd (state s) \<longrightarrow> y \<in> snd (state s)"
    unfolding dfs_invar_compl_def using SE_dfs_has_cycle_not_next nested_dfs_invar_def by simp_all blast+
  
  from inv_cl have constr: "s \<in> dfs_constr_from SE_dfs y" by auto
  with step have constr': "s' \<in> dfs_constr_from SE_dfs y" using dfs_constr_from.step by blast

  have "nested_dfs_invar s'"
  proof (cases "has_cycle s'")
    case True thus ?thesis by (simp add: nested_dfs_invar_def)
  next
    case False with step inv show ?thesis
    proof (cases rule: dfs_next_cases_elem)
      case (empty x xs) with step have "dfs_next hpy_dfs s s'" unfolding dfs_next_def dfs_cond_compl_def dfs_step_def by simp
      moreover from constr step SE_to_nested have "s \<in> dfs_constr_from hpy_dfs y" using SE_dfs_has_cycle_not_next by blast
      with inv_cl have "dfs_invar_compl hpy_dfs y s" unfolding dfs_invar_compl_def by simp
      ultimately have "dfs_invar hpy_dfs s'" using hpy_dfs_preserves_invar by (metis dfs_preserves_invarE)
      thus ?thesis by simp
    next
      case visit with step inv show ?thesis by (simp add: nested_dfs_invar_def)
    next
      case restrict with step inv show ?thesis by (simp add: nested_dfs_invar_def)
    next
      case (remove e) hence "snd (SE_remove (state s) s e) = snd (state s)" by (cases "state s", cases "fst (state s)") simp_all
      with remove step inv show ?thesis by (simp add: nested_dfs_invar_def)
    qed
  qed
  thus "dfs_invar SE_dfs s'" by simp
qed (auto simp add: nested_dfs_invar_def)

definition SE_cyclic where "SE_cyclic s \<equiv> has_cycle s \<and> (cyclic (start s) s \<or> (\<exists> x \<in> set (stack s). x \<in> \<A> \<and> x \<rightarrow>+ x))"
definition SE_cycle where "SE_cycle x \<equiv> \<exists>s. s \<in> dfs_constr_from SE_dfs x \<and> SE_cyclic s"

lemma has_cycle_implies_SE_cyclic:
  "dfs_constructable SE_dfs s \<Longrightarrow> has_cycle s \<Longrightarrow> SE_cyclic s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s') with SE_dfs_has_cycle_not_next have "\<not> has_cycle s" by blast
  with empty SE_to_nested' have "dfs_constructable hpy_dfs s" by simp
  moreover from empty have "dfs_next hpy_dfs s s'" by (simp add: dfs_next_def dfs_cond_compl_def dfs_step_def)
  ultimately have "dfs_constructable hpy_dfs s'" using dfs_constructable.step by blast
  with empty.prems have "cyclic (start s') s'" using has_cycle_implies_cyclic[OF _ hpyM_correct, of s', folded hpy_dfs_def] by metis
  with empty.prems show ?case unfolding SE_cyclic_def by simp
next
  case (remove s s' e x) with SE_dfs_has_cycle_not_next have "\<not> has_cycle s" by blast
  with remove have "e \<in> set (stack s) \<and> (e \<in> \<A> \<or> x \<in> \<A>)" by (cases "state s", cases "fst (state s)") simp_all
  hence e_stack: "e \<in> set (stack s)" and A: "e \<in> \<A> \<or> x \<in> \<A>" by simp_all

  from remove have "e \<in> succs x" using wl_subset_succs[OF remove(2), of 0] by auto
  with succs_reachable have xe: "x \<rightarrow>+ e" by auto
  
  have "\<exists>y \<in> set (stack s'). y \<rightarrow>+ y \<and> y \<in> \<A>"
  proof (cases "e = x")
    case True with xe A remove show ?thesis by auto
  next
    case False with remove e_stack have "e \<in> set (tl (stack s))" by auto
    with tl_reachable_stack_hd[OF remove(2)] remove have ex: "e \<rightarrow>+ x" by auto
    from A show ?thesis
    proof
      assume "e \<in> \<A>"
      moreover note xe ex
      moreover from e_stack remove have "e \<in> set (stack s')" by auto
      ultimately show ?thesis using reachable1_trans by blast
    next
      assume "x \<in> \<A>" 
      moreover note xe ex
      moreover from remove have "x \<in> set (stack s')" by auto 
      ultimately show ?thesis using reachable1_trans by blast
    qed
  qed
  with remove.prems show ?case unfolding SE_cyclic_def by blast
qed (simp_all add: SE_dfs_has_cycle_not_next SE_cyclic_def)

lemma SE_cycle_not_completed:
  assumes reach: "x \<rightarrow>\<star> v"
  and cycle: "v \<rightarrow>+ v"
  and fin: "v \<in> \<A>"
  and constr: "s \<in> dfs_constr_from SE_dfs x"
  shows "\<not> dfs_completed SE_dfs s"
proof (rule notI)
  assume compl: "dfs_completed SE_dfs s"
  hence "\<not> has_cycle s" by (simp add: dfs_completed_def)
  
  from compl no_restrict_finished_eq_reachable[of SE_dfs s] constr_from_implies_start[OF constr] have "finished s = {v. x \<rightarrow>\<star> v}" by simp
  hence "v \<in> finished s" using reach by simp
  
  moreover
  from compl have "dfs_invar SE_dfs s" by blast
  with `\<not> has_cycle s` have "\<forall>v \<in> \<A> \<inter> finished s. \<not> v \<rightarrow>+ v" by (simp add: nested_dfs_invar_def)
  with cycle fin have "v \<notin> finished s" by auto
  
  ultimately show "False" by contradiction
qed

lemma cycle_generates_SE_cyclic:
  assumes reach: "x \<rightarrow>\<star> v"
  and cycle: "v \<rightarrow>+ v"
  and fin: "v \<in> \<A>"
  obtains s where "s \<in> dfs_constr_from SE_dfs x" "SE_cyclic s"
proof -
  from reach have "x \<in> V" by auto
  with exists_finished[OF _ _ SE_dfs_preserves_invar] obtain s where s: "s \<in> dfs_constr_from SE_dfs x" "dfs_finished SE_dfs s" by auto
  moreover with SE_cycle_not_completed assms have "\<not> dfs_completed SE_dfs s" by simp
  ultimately have "\<not> dfs_cond SE_dfs (state s)" using finished_is_completed_or_not_cond by blast
  hence "has_cycle s" by simp
  with s has_cycle_implies_SE_cyclic that show ?thesis by blast
qed

lemma SE_cycle_get_cycle:
  assumes "SE_cycle x"
  obtains v where "x \<rightarrow>\<star> v" "v \<in> \<A>" "v \<rightarrow>+ v"
proof -
  from assms obtain s where constr: "s \<in> dfs_constr_from SE_dfs x" and "SE_cyclic s" unfolding SE_cycle_def by auto
  show ?thesis
  proof (cases "cyclic x s")
    case True then obtain v where "v \<in> finished s \<inter> \<A>" "v \<rightarrow>+ v" "x \<rightarrow>\<star> v" unfolding cyclic_def by blast
    with that show ?thesis by blast
  next
    case False 
    from constr constr_from_implies_start have start: "start s = x" by blast
    with False `SE_cyclic s` obtain v where "v \<in> set (stack s)" "v \<in> \<A>" "v \<rightarrow>+v" unfolding SE_cyclic_def by auto
    with start start_reachable_stack constr that show ?thesis by auto
  qed
qed

lemma SE_cycle_iff_cycle:
  "SE_cycle x \<longleftrightarrow> (\<exists>v. x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v)"
proof
  assume "SE_cycle x" thus "\<exists>v. x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v" by (metis SE_cycle_get_cycle)
next
  assume "\<exists>v.  x \<rightarrow>\<star> v \<and> v \<in> \<A> \<and> v \<rightarrow>+ v"
  then guess v ..
  then obtain s where "SE_cyclic s" "s \<in> dfs_constr_from SE_dfs x" using cycle_generates_SE_cyclic by blast
  thus "SE_cycle x" unfolding SE_cycle_def by blast
qed

lemma dfs_fun_SE_dfs_correct:
  assumes "x \<in> V"
  shows "dfs_fun SE_dfs x \<le> SPEC (\<lambda>s. has_cycle s \<longleftrightarrow> SE_cycle x)"
using assms
proof (intro dfs_fun_pred)
  fix s
  assume "s \<in> dfs_constr_from SE_dfs x" "has_cycle s"
  moreover hence "SE_cyclic s" using has_cycle_implies_SE_cyclic by force
  ultimately show "SE_cycle x" by (auto simp: SE_cycle_def)
next
  fix s
  assume "s \<in> dfs_constr_from SE_dfs x" "dfs_completed SE_dfs s" "SE_cycle x"
  hence "False" by (metis SE_cycle_iff_cycle SE_cycle_not_completed)
  thus "has_cycle s" ..
qed (simp_all add: SE_dfs_preserves_invar)

theorem SE_dfs_correct:
  assumes "x \<in> V"
  and "inres (dfs_fun SE_dfs x) s"
  shows "has_cycle s \<longleftrightarrow> SE_cycle x"
using assms dfs_fun_SE_dfs_correct
by (bestsimp dest!: pwD2)

theorems SE_dfs_correct_unfolded = SE_dfs_correct[unfolded SE_cycle_iff_cycle]

lemma RETURN_SE_cycle_correct:
  assumes "x \<in> V"
  shows "RETURN (SE_cycle x) = do { s \<leftarrow> dfs_fun SE_dfs x; RETURN (has_cycle s) }"
apply (rule pw_eqI)
  apply (simp_all add: pw_bind_inres pw_bind_nofail dfs_fun_nofail[OF SE_dfs_preserves_invar assms])
  apply (metis assms dfs_fun_nonempty SE_dfs_correct)
done

lemmas RETURN_SE_cycle_correct_unfolded = RETURN_SE_cycle_correct[unfolded SE_cycle_iff_cycle]
end
end
