theory Tree_DFS_old
imports Main DFS_old
begin

type_synonym ('n) tree = "('n \<times> 'n) set"

definition add_edge :: "'n tree \<Rightarrow> ('n tree, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'n tree" where
  "add_edge e sws n = insert (hd (stack sws), n) e"

definition tdfs :: "('n tree, 'n) dfs_algorithm_invar"  where
  "tdfs = \<lparr> dfs_cond = \<lambda>_. True, dfs_action = add_edge, dfs_post = \<lambda> s _ _. s, dfs_remove = \<lambda> s _ _. s, dfs_start = \<lambda>_. {}, dfs_restrict = {}, dfs_invar = \<lambda>_. True \<rparr>"

definition edges :: "('n tree, 'n, 'X) dfs_sws_scheme \<Rightarrow> 'n tree"
  where  "edges s = state s"

(* add this to the simp-set for the moment. it will be removed again when the basic stuff is done. *)
declare edges_def[simp]

context finite_digraph
begin

lemma tdfs_simps[simp]:
  "dfs_cond tdfs S"
  "dfs_action tdfs = add_edge"
  "dfs_post tdfs S s x = S"
  "dfs_remove tdfs S s x = S"
  "dfs_start tdfs x = {}"
  "dfs_restrict tdfs = {}"
  "dfs_invar tdfs s"
unfolding tdfs_def
by simp_all

lemma dfs_next_edges_subset:
  "dfs_next tdfs s s' \<Longrightarrow> edges s \<subseteq> edges s'"
by (cases rule: dfs_next_cases_elem) (auto simp add: dfs_sws.defs add_edge_def)

lemma no_self_edges:
  "dfs_constructable tdfs s \<Longrightarrow> (v, w) \<in> edges s \<Longrightarrow> v \<noteq> w"
proof (induction rule: dfs_constructable_induct)
  case (visit s s' t x xs) hence
    s': "stack s' = t#x#xs" and
    e': "edges s' = insert (x, t) (edges s)"
    by (simp_all add: add_edge_def)
  
  from s' have "x \<noteq> t" using stack_distinct[OF visit(3)] by auto
  thus ?case
  proof (cases "(v,w) = (x, t)")
    case False with e' visit.prems have "(v,w) \<in> edges s" by auto
    with visit.IH show ?thesis .
  qed simp
qed simp+

lemma stack_tl_subset_edges:
  "dfs_constructable tdfs s \<Longrightarrow> stack s \<noteq> [] \<Longrightarrow> set (tl(stack s)) \<subseteq> Field (edges s)"
by (induct rule: dfs_constructable_induct) (auto simp add: add_edge_def tl_subset)

lemma stack_hd_in_edges:
  "dfs_constructable tdfs s \<Longrightarrow> edges s \<noteq> {} \<Longrightarrow> stack s \<noteq> [] \<Longrightarrow> hd (stack s) \<in> Field (edges s)"
proof (induct rule: dfs_constructable_induct)
  case (empty s s')
  hence "set (stack s') \<subseteq> Field (edges s')" using stack_tl_subset_edges[OF empty(2)] by simp
  with empty.prems show ?case by (cases "stack s'") simp_all
qed (simp add: add_edge_def)+

lemma stack_subset_edges:
  assumes constr: "dfs_constructable tdfs s"
  and         ne: "edges s \<noteq> {}"
  shows "set (stack s) \<subseteq> Field (edges s)"
proof (cases "stack s")
  case (Cons x xs) then have "set xs \<subseteq> Field (edges s)" using stack_tl_subset_edges[OF constr] by simp
  moreover from Cons have "x \<in> Field (edges s)" using stack_hd_in_edges[OF assms] by simp
  ultimately show ?thesis using Cons by simp
qed simp

lemma stack_gt_1_implies_edges:
  assumes "dfs_constructable tdfs s"
  and "length (stack s) > 1"
  shows "edges s \<noteq> {}"
using stack_tl_subset_edges[OF assms(1)] assms(2)
by (cases "stack s") auto

lemma edges_stack_is_discovered:
  assumes "dfs_constructable tdfs s"
  and "stack s \<noteq> []"
  shows "Field (edges s) \<union> set (stack s) = discovered s"
using assms
proof induction
  case (empty s s') hence "stack s' = tl (stack s)" by simp
  with `stack s' \<noteq> []` have "length (stack s) > 1" by (smt length_0_conv length_tl)
  with empty have  "set (stack s) \<subseteq> Field (edges s)" using stack_subset_edges stack_gt_1_implies_edges by metis
  with empty have "Field (edges s') = discovered s'" by auto
  thus ?case using stack_subset_discovered[OF dfs_constructable.step[OF empty(1,2)]] by auto
next
  case (visit s s' t x) hence
    e': "edges s' = insert (x, t) (edges s)" and
    s': "stack s' = t # stack s"
    by (simp_all add: add_edge_def)
 
  hence "Field (edges s') = {x, t} \<union> Field (edges s)" by auto
  with s' have "Field (edges s') \<union> set(stack s') = Field (edges s) \<union> {x, t} \<union> set (stack s)" by auto
  also with visit have "... = insert t (Field (edges s) \<union> set (stack s))" by auto
  also with visit have "... = discovered s'" by simp
  finally show ?case .
qed simp+

lemma completed_edges_eq_discovered:
  assumes "dfs_completed tdfs s"
  and "edges s \<noteq> {}"
  shows "Field (edges s) = discovered s"
using assms
proof induct
  case (prev s r) then have ne: "edges r \<noteq> {}" by simp
  
  from prev have stack_r: "stack r = [start s]" using dfs_completed_prev_stack by blast
  with prev ne have "start s \<in> Field (edges r)" using stack_hd_in_edges[of r] by force
  with edges_stack_is_discovered[of r, OF prev(1)] stack_r  have "Field (edges r) = discovered r" by force
  with prev show ?case by simp
qed

(* this sometimes is enough to know and it has fewer prerequisites than the more stricter lemmas *)
lemma edges_subset_discovered:
  assumes "dfs_constructable tdfs s"
  shows "Field (edges s) \<subseteq> discovered s"
using edges_stack_is_discovered[OF assms(1)]
proof (cases "stack s")
  case Cons then show ?thesis using edges_stack_is_discovered[OF assms(1)] by auto
next
  case Nil with assms have "dfs_completed tdfs s" using dfs_completed_stackI tdfs_simps(1) by force
  from completed_edges_eq_discovered[OF this] show ?thesis by (cases "edges s = {}") simp_all
qed

lemma etrancl_is_discovered:
  assumes constr: "dfs_constructable tdfs s"
  and etrancl: "(v,w) \<in> (edges s)\<^sup>+"
  shows "v \<in> discovered s" and "w \<in> discovered s"
proof -
  from edges_subset_discovered[OF constr] have "(edges s)\<^sup>+ \<subseteq> discovered s \<times> discovered s" using trancl_subset_Field2 by auto
  with etrancl show "v \<in> discovered s" "w \<in> discovered s" by auto
qed

lemma edges_subset_E:
  "dfs_constructable tdfs s \<Longrightarrow> edges s \<subseteq> E"
proof (induction rule: dfs_constructable_induct)
  case (visit s s' t x) hence e': "edges s' = insert (x,t) (edges s)" by (simp add: add_edge_def)
  moreover from wl_subset_succs[OF visit(2), of 0] visit have "t \<in> succs x" by auto
  ultimately have "(x, t) \<in> E" unfolding succs_def by simp
  with e' visit.IH show ?case by simp
qed simp+

lemma edge_is_succs:
  assumes "dfs_constructable tdfs s"
  and "(v, w) \<in> edges s"
  shows "w \<in> succs v"
proof -
  from assms have "(v, w) \<in> E" using edges_subset_E by auto
  thus ?thesis by (auto simp add: succs_def)
qed
  
lemma discover_edge:
  "dfs_constructable tdfs s \<Longrightarrow> (v,w) \<in> edges s \<Longrightarrow> \<delta> s v < \<delta> s w"
proof (induction rule: dfs_constructable_induct)
  case (visit s s' t x) hence
    e': "edges s' = insert (x, t) (edges s)" and
    d': "discover s' = (discover s)(t \<mapsto> counter s)"
    by (simp_all add: add_edge_def)

  show ?case
  proof (cases "(v,w) \<in> edges s")
    case True
    from visit(9) have "t \<notin> Field (edges s)" using visit(2) edges_subset_discovered by auto
    with True have "v \<noteq> t" and "w \<noteq> t" unfolding Field_def by auto
    with d' visit.IH True show ?thesis by auto
  next
    case False 
    with visit e' have "(v,w) =  (x, t)" by auto
    
    moreover from visit have "t \<noteq> x" using stack_distinct[OF visit(3)] by auto
    
    moreover
    from visit have "x \<in> discovered s" using stack_subset_discovered[OF visit(2)] by auto
    hence "\<delta> s x < counter s" using discover_lt_counter[OF visit(2)] by simp

    ultimately show ?thesis using d' by simp
  qed
qed simp+

lemma discover_etrancl:
  assumes  constr: "dfs_constructable tdfs s"
  and reach: "(v,w) \<in> (edges s)\<^sup>+"
  shows "\<delta> s v < \<delta> s w"
using reach discover_edge[OF constr]
by induct force+

lemma no_cycle:
  "dfs_constructable tdfs s \<Longrightarrow> (v,w) \<in> (edges s)\<^sup>+ \<Longrightarrow> v \<noteq> w"
by (blast dest: discover_etrancl)

lemma stack_contains_edges_all:
  "dfs_constructable tdfs s \<Longrightarrow> \<forall> n < length (stack s) - 1. (stack s ! Suc n, stack s ! n) \<in> edges s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x xs) then show ?case by (induct xs) auto
next
  case (visit s s' t x) hence
    e': "edges s' = insert (x, t) (edges s)" and
    s': "stack s' = t # stack s"
    by (simp_all add: add_edge_def)

  with visit have "t = stack s' ! 0" and "x = stack s' ! Suc 0" using nth_Cons_0 hd_conv_nth by simp_all
  with e' have "(stack s' ! Suc 0, stack s' ! 0) \<in> edges s'" by simp

  moreover
  from e' s' visit.IH have  "\<And> n. n < (length (stack s')) - 1 \<Longrightarrow> n > 0 \<Longrightarrow> (stack s' ! Suc n, stack s' ! n) \<in> edges s'"
    by (smt gr0_implies_Suc insertI2 list.size(4) nth_Cons_Suc)

  ultimately show ?case by smt
qed simp+

lemma stack_contains_edges:
  "dfs_constructable tdfs s \<Longrightarrow> n < length (stack s) - 1 \<Longrightarrow> (stack s ! Suc n, stack s ! n) \<in> edges s"
by (metis stack_contains_edges_all)

lemma nth_stack_hd_in_etrancl:
  "dfs_constructable tdfs s \<Longrightarrow> n < length (stack s) \<Longrightarrow> n > 0 \<Longrightarrow> (stack s ! n, stack s ! 0) \<in> (edges s)\<^sup>+"
by (metis nth_step_trancl stack_contains_edges)

lemma hd_stack_no_cycle:
  assumes constr: "dfs_constructable tdfs s"
  and edge: "(hd (stack s), w) \<in> (edges s)\<^sup>+"
  and ne: "stack s \<noteq> []"
  shows "w \<notin> set (stack s)"
proof (rule notI)
  assume A: "w \<in> set (stack s)"
  then obtain n where n: "n < length (stack s)" "stack s ! n = w" unfolding in_set_conv_nth by auto
  
  from A edge no_cycle[OF constr] have "w \<noteq> hd (stack s)" by auto
  with n have "n > 0" by (smt hd_conv_nth ne)
  with n nth_stack_hd_in_etrancl[OF constr] have "(w, hd(stack s)) \<in> (edges s)\<^sup>+" using hd_conv_nth[OF ne] by auto
  with edge have "(w,w) \<in>  (edges s)\<^sup>+" by auto
  with no_cycle[OF constr] show "False" by auto
qed

lemma no_edge_to_stack:
  "dfs_constructable tdfs s \<Longrightarrow> (v, w) \<in> edges s \<Longrightarrow> v \<notin> set (stack s) \<Longrightarrow> w \<notin> set (stack s)"
proof (induction rule: dfs_constructable_induct)
  case (empty s s') hence edge: "(v,w) \<in> edges s" by simp
  from empty have ne: "stack s \<noteq> []" and tl: "stack s' = tl (stack s)" by simp_all
  hence "w \<notin> set (stack s)"
  proof (cases "v = hd (stack s)")
    case True thus ?thesis using hd_stack_no_cycle[OF empty(2) _ ne] edge by auto
  next
    case False with empty.prems ne tl have "v \<notin> set (stack s)" by (fastforce intro: list.exhaust)
    with edge empty.IH show ?thesis by simp
  qed
  with tl ne show ?case by (fastforce intro: list.exhaust)
next
  case (visit s s' t x)  hence "t \<notin> discovered s" by simp
  hence no_edge: "t \<notin> Field (edges s)" using edges_subset_discovered[OF visit(2)] by auto
  
  from visit have "v \<notin> set (stack s)" by simp
  moreover with visit have "(v,w) \<in> edges s" by (auto simp add: add_edge_def)
  ultimately have "w \<notin> set (stack s)" using visit.IH by simp
  with visit show ?case
  proof (cases "w = t")
    case True with `(v,w) \<in> edges s` have "t \<in> Field (edges s)" unfolding Field_def by auto
    with no_edge show ?thesis by contradiction
  qed simp
qed simp+

lemma no_edge_to_stack_trancl:
  assumes constr: "dfs_constructable tdfs s"
  and "(v,w) \<in> (edges s)\<^sup>+"
  and "v \<notin> set (stack s)"
  shows "w \<notin> set (stack s)"
using assms(2,3) no_edge_to_stack[OF constr]
by (induct rule: trancl_induct) simp_all

lemma finished_etrancl_finished:
  assumes constr: "dfs_constructable tdfs s"
  and edge: "(v,w) \<in> (edges s)\<^sup>+"
  and finished: "v \<in> finished s"
  shows "w \<in> finished s"
proof -
  from constr finished have "v \<notin> set (stack s)" by (rule finished_implies_not_stack)
  with no_edge_to_stack_trancl[OF constr edge] have "w \<notin> set (stack s)" .

  moreover
  from edge have "w \<in> discovered s" using etrancl_is_discovered[OF constr] by simp
  ultimately show ?thesis using discovered_non_stack_implies_finished[OF constr] by auto
qed

lemma finish_edge:
  assumes constr: "dfs_constructable tdfs s"
  and edge: "(v,w) \<in> edges s"
  and fin: "v \<in> finished s"
  shows "\<phi> s v > \<phi> s w"
proof -
  from assms have "w \<in> finished s" using finished_etrancl_finished[OF constr] by force
  moreover from fin have "v \<in> discovered s" using finished_implies_discovered[OF constr] by simp
  moreover from discover_edge constr edge have "\<delta> s v < \<delta> s w" .
  moreover from edge have "w \<in> succs v" using edge_is_succs[OF constr] by simp
  ultimately show ?thesis using finished_succs_finish assms by blast
qed

lemma finish_etrancl:
 assumes  constr: "dfs_constructable tdfs s"
  and reach: "(v,w) \<in> (edges s)\<^sup>+"
  and fin: "v \<in> finished s"
  shows "\<phi> s v > \<phi> s w"
using reach finish_edge[OF constr] fin finished_etrancl_finished[OF constr]
by induct fastforce+
  
lemma completed_finish_etrancl:
  assumes fin: "dfs_completed tdfs s"
  and edge: "(v,w) \<in> (edges s)\<^sup>+"
  shows "\<phi> s v > \<phi> s w"
proof -
  from fin have "Field (edges s) \<subseteq> finished s" using edges_subset_discovered completed_finished_eq_discovered by fastforce
  with edge have "v \<in> finished s" using trancl_subset_Field2 by auto
  with fin edge show ?thesis using finish_etrancl by auto
qed

(* just slightly modified copies from DFS.thy *)
lemma discover_finish_implies_in_etrancl:
  assumes "dfs_constructable tdfs s"
  and "v \<in> discovered s" and "w \<in> discovered s"
  and "\<delta> s v < \<delta> s w" and "v \<notin> finished s \<or> (w \<in> finished s \<and> \<phi> s w < \<phi> s v)"
  shows "(v,w) \<in> (edges s)\<^sup>+"
  using assms
proof induction
  case (empty s s' x) hence
    s': "discover s = discover s'" and
    f': "finish s' = finish s (x \<mapsto> counter s)" and
    e': "(edges s')\<^sup>+ = (edges s)\<^sup>+"
    by simp_all
  
  show ?case
  proof (cases "v = x")
    case True with empty f' have "v \<notin> finished s" and "finish s' w = finish s w" using stack_implies_not_finished[OF empty(2)] by auto
    with empty.prems empty.IH s' e' show ?thesis by simp
  next
    case False  with f' have fv: "finish s v = finish s' v" by simp
    show ?thesis
    proof (cases "w \<in> finished s'")
      case False with empty.IH s' e' empty.prems fv show ?thesis by fastforce
    next
      case True note wf' = this
      thus ?thesis
      proof (cases "v \<in> finished s'")
        case True with empty(12) wf' have *: "\<phi> s' w < \<phi> s' v" by simp
        have "w \<noteq> x"
        proof (rule notI)
          assume "w = x"
          with f' have "\<phi> s' w = counter s" by simp
          
          moreover
          from True fv have "v \<in> finished s" by auto        
          hence "\<phi> s v < counter s" using finish_lt_counter[OF empty(2)] by simp
          with fv have "\<phi> s' v < counter s" by simp
          ultimately show "False" using * by simp
        qed
        with * fv f' wf' have "w \<in> finished s \<and> \<phi> s w < \<phi> s v" by simp
        with empty.prems s' e' empty.IH show ?thesis by simp
      next
        case False with empty.prems empty.IH s' f' e' show ?thesis by simp
      qed
    qed
  qed
next
  case (visit s s' e x) 
  hence d': "discover s' = discover s (e \<mapsto> counter s)" 
    and s': "stack s' = e # stack s"
    and f': "finish s' = finish s"
    and ne: "v \<noteq> w"
    and e': "edges s' = insert (x, e) (edges s)"
    by (auto simp add: add_edge_def)

  show ?case
  proof (cases "v = e")
    case True with d' ne have "\<delta> s' v = counter s" "discover s w = discover s' w" by auto

    moreover
    then have "w \<in> discovered s" using visit.prems by auto
    hence "\<delta> s w < counter s" using discover_lt_counter[OF visit(2)] by simp

    ultimately have "False" using visit.prems by simp
    thus ?thesis ..
  next
    case False with d' have dv: "discover s v = discover s' v" by simp
    show ?thesis
    proof (cases "w = e")
      case True with s' have "w \<notin> finished s'" using stack_implies_not_finished[OF visit(3)] by simp
      with visit.prems have "v \<in> set (stack s')" using discovered_not_finished_implies_stack[OF visit(3)] by simp
      then obtain n where n: "n < length (stack s')" "stack s' ! n = v" unfolding in_set_conv_nth by auto
      with False s' have "n > 0" by (smt nth_Cons')
      with True e' s' n show ?thesis using nth_stack_hd_in_etrancl[OF visit(3), of n] by simp
    next
      case False with d' dv f' visit.prems visit.IH have "(v, w) \<in> (Tree_DFS_old.edges s)\<^sup>+" by fastforce
      with e' trancl_insert show ?thesis by force
    qed
  qed
qed simp+

lemma not_in_etrancl_discover_finish:
  assumes constr: "dfs_constructable tdfs s"
  and ne: "v \<noteq> w"
  and not_in_etrancl: "(v,w) \<notin> (edges s)\<^sup>+"
  shows "v \<notin> discovered s \<or> w \<notin> discovered s \<or> w \<notin> finished s \<or> \<delta> s v > \<delta> s w \<or> (v \<in> finished s \<and> \<phi> s v < \<phi> s w)"
proof (cases "v \<in> V \<and> w \<in> V")
  case False hence "w \<notin> discovered s \<or> v \<notin> discovered s" using discovered_subset_verts[OF constr] by auto
  thus ?thesis by simp
next
  case True hence verts: "v \<in> V" "w \<in> V" by simp_all
  from not_in_etrancl show ?thesis
    apply (rule contrapos_np)
    apply (rule discover_finish_implies_in_etrancl[OF constr, of v w])
    apply (insert discover_neq_discover[OF constr verts ne] finish_neq_finish[OF constr verts ne])
    by force+
qed

lemma both_not_in_etrancl_discover_finish:
  assumes constr: "dfs_constructable tdfs s"
  and ne: "v \<noteq> w"
  and not_in_etrancl: "(v,w) \<notin> (edges s)\<^sup>+" "(w,v) \<notin> (edges s)\<^sup>+"
  and fin: "v \<in> finished s" "w \<in> finished s"
  shows "\<delta> s w > \<phi> s v \<or> \<phi> s w < \<delta> s v"
proof -
  from constr fin have disc: "v \<in> discovered s" "w \<in> discovered s" using finished_implies_discovered by metis+
  with constr ne not_in_etrancl not_in_etrancl_discover_finish fin have "\<delta> s v > \<delta> s w \<or> \<phi> s v < \<phi> s w" "\<delta> s w > \<delta> s v \<or> \<phi> s w < \<phi> s v" by metis+
  thus ?thesis using correct_order constr fin by smt
qed

end
(* I promised :) *)
declare edges_def[simp del]

subsection {* The DFS search tree *}
text {* We now are interested in the search-tree that is built by a DFS. For the sake of easier proofs, we assume the graph is non-empty, because this case is trivial and uninteresting. *}
locale dfs_run =  dfs: finite_digraph V E + G: finite_digraph V\<^sub>G E\<^sub>G
  for V V\<^sub>G :: "'n set" and  E E\<^sub>G :: "('n \<times> 'n) set" +
  fixes s :: "('n tree, 'n) dfs_sws"
  assumes E_edges: "E = edges s"
  and E_ne: "E \<noteq> {}"
  and V_discovered: "V = discovered s"
  and completed: "dfs_completed tdfs s"
begin

abbreviation G_fd_to_graph ("\<G>\<^sub>G") where
  "G_fd_to_graph \<equiv> G.fd_to_graph"
abbreviation G_fd_reachable ("_ \<rightarrow>G\<star> _" [100,100] 40) where
  "G_fd_reachable \<equiv> G.fd_reachable"
abbreviation G_fd_reachable1 ("_ \<rightarrow>G+ _" [100,100] 40) where
  "G_fd_reachable1 \<equiv> G.fd_reachable1"

abbreviation dfs_fd_to_graph ("\<G>\<^sub>d") where
  "dfs_fd_to_graph \<equiv> dfs.fd_to_graph"
abbreviation dfs_fd_reachable ("_ \<rightarrow>d\<star> _" [100,100] 40) where
  "dfs_fd_reachable \<equiv> dfs.fd_reachable"
abbreviation dfs_fd_reachable1 ("_ \<rightarrow>d+ _" [100,100] 40) where
  "dfs_fd_reachable1 \<equiv> dfs.fd_reachable1"

lemma dfs_reach_etrancl:
  "v \<rightarrow>d+ w \<longleftrightarrow> (v,w) \<in> (edges s)\<^sup>+"
by (metis E_edges dfs.reachable1_trancl)

lemma constructable:
  "dfs_constructable tdfs s"
by (rule dfs_completed_constructable[OF completed])

lemma dfs_edges_sub:
  "E \<subseteq> E\<^sub>G"
by (metis constructable edges_subset_E E_edges)

lemma dfs_vertices_sub:
  "V \<subseteq> V\<^sub>G"
by (metis constructable discovered_subset_verts V_discovered)

lemma verts_eq_edges:
  "V = Field E"
by (metis completed V_discovered completed_edges_eq_discovered E_ne E_edges)

lemma s_discovered[simp, intro]:
  "x \<in> V \<Longrightarrow> x \<in> discovered s"
using V_discovered
by simp

lemma s_finished[simp,intro]:
  "x \<in> V \<Longrightarrow> x \<in> finished s"
using completed_finished_eq_discovered[OF completed] s_discovered
by auto

lemma vert_reach_implies_vert:
  assumes "x \<in> V"
  and reach: "x \<rightarrow>G+ y"
  shows "y \<in> V"
proof -
  have fin: "V = finished s" using V_discovered completed_finished_eq_discovered[OF completed] by simp
  with `x \<in> V` have "x \<in> finished s" by simp
  with reach completed_reach_implies_finished[OF completed] have "y \<in> finished s" by auto
  with fin show ?thesis by simp
qed

lemma vert_ewalk_ewalk_verts:
  assumes v: "x \<in> V"
  and walk: "ewalk x p y"
  shows "set (ewalk_verts x p) \<subseteq> V"
using walk v walk
proof induction
  case Base thus ?case by simp
next
  case (Cons w1 w2 e es) with ewalk_Cons_iff have "ewalk w2 es y" by simp
  moreover from edge_implies_reach1[of e] vert_reach_implies_vert Cons have "w2 \<in> V" by simp
  ultimately have "set (ewalk_verts w2 es) \<subseteq> V" using Cons.IH by simp
  with ewalk_verts_conv[OF `ewalk w2 es y`] have "snd ` set es \<subseteq> V" by simp
  with Cons ewalk_verts_conv[OF Cons(5)] `w2 \<in> V` show ?case by simp
qed

lemma vert_ewalk_edge_verts:
  assumes v: "x \<in> V"
  and walk: "ewalk x p y"
  and edge: "e \<in> set p"
  shows "{fst e, snd e} \<subseteq> V"
proof -
  from v walk have "set (ewalk_verts x p) \<subseteq> V" using vert_ewalk_ewalk_verts by simp
  with G.ewalk_edge_in_ewalk_verts[OF walk edge] show ?thesis by auto
qed
end

sublocale dfs_run \<subseteq> finite_subgraph V\<^sub>G V E\<^sub>G E
using dfs_edges_sub verts_eq_edges by unfold_locales

context dfs_run
begin

no_notation disc ("\<delta>")
no_notation fin ("\<phi>")

abbreviation "\<delta> \<equiv> disc s"
abbreviation "\<phi> \<equiv> fin s"

lemma times_relation:
  "v \<in> V \<Longrightarrow> \<delta> v < \<phi> v"
using finish_gt_discover[OF constructable] s_finished
by force

lemma no_self_reach: 
  "v \<rightarrow>d+ w \<Longrightarrow> v \<noteq> w"
using dfs_reach_etrancl no_cycle[OF constructable]
by simp

lemma correct_order:
  "v \<in> V \<Longrightarrow> w \<in> V \<Longrightarrow> \<delta> v < \<delta> w \<Longrightarrow>  \<phi> v < \<delta> w \<or> \<phi> v > \<phi> w"
by (metis G.correct_order[OF constructable] s_finished)

lemma reachE [case_names reach_vw reach_wv no_reach, consumes 2]:
  assumes verts: "v \<in> V" "w \<in> V"
  obtains 
    (reach_vw) "v \<rightarrow>d\<star> w" 
  | (reach_wv) "w \<rightarrow>d\<star> v" 
  | (no_reach) "\<not> v \<rightarrow>d\<star> w" "\<not> w \<rightarrow>d\<star> v"
proof (cases "v \<rightarrow>d\<star> w")
  case True with reach_vw show ?thesis .
next
  case False note no_vw = this
  show ?thesis
  proof (cases "w \<rightarrow>d\<star> v")
    case True with reach_wv show ?thesis .
  next
    case False with no_vw no_reach show ?thesis by simp
  qed
qed

lemma discover_neq_finish:
  "v \<in> V \<Longrightarrow> w \<in> V \<Longrightarrow> \<delta> v \<noteq> \<phi> w"
proof -
  assume v: "v \<in> V" and w: "w \<in> V"
  with dfs_vertices_sub have "v \<in> V\<^sub>G" "w \<in> V\<^sub>G" by auto
  hence "discover s v \<noteq> finish s w" using discover_neq_finish[OF constructable] s_discovered[OF v] by simp
  with s_discovered[OF v] s_finished[OF w] show "\<delta> v \<noteq> \<phi> w" by auto
qed

lemma discover_neq_discover:
  "v \<in> V \<Longrightarrow> w \<in> V \<Longrightarrow> v \<noteq> w \<Longrightarrow> \<delta> v \<noteq> \<delta> w"
proof -
  assume v: "v \<in> V" and w: "w \<in> V" and ne: "v \<noteq> w"
  with dfs_vertices_sub have "v \<in> V\<^sub>G" "w \<in> V\<^sub>G" by auto
  hence "discover s v \<noteq> discover s w" using discover_neq_discover[OF constructable] s_discovered[OF v] ne by simp
  with s_discovered[OF v] s_discovered[OF w] show "\<delta> v \<noteq> \<delta> w" by auto
qed

lemma finish_neq_finish:
  "v \<in> V \<Longrightarrow> w \<in> V \<Longrightarrow> v \<noteq> w \<Longrightarrow> \<phi> v \<noteq> \<phi> w"
proof -
  assume v: "v \<in> V" and w: "w \<in> V" and ne: "v \<noteq> w"
  with dfs_vertices_sub have "v \<in> V\<^sub>G" "w \<in> V\<^sub>G" by auto
  hence "finish s v \<noteq> finish s w" using finish_neq_finish[OF constructable] s_finished[OF v] ne by simp
  with s_finished[OF v] s_finished[OF w] show "\<phi> v \<noteq> \<phi> w" by auto
qed

lemma reach_discover:
  "v \<rightarrow>d+ w \<Longrightarrow> \<delta> v < \<delta> w"
by (metis discover_etrancl[OF constructable] dfs_reach_etrancl)

lemma reach_finish:
  "v \<rightarrow>d+ w \<Longrightarrow> \<phi> v > \<phi> w"
by (metis completed_finish_etrancl[OF completed] dfs_reach_etrancl)

lemma no_reach_obtain_split:
  assumes nr: "\<not> x \<rightarrow>d\<star> y"
  and vert: "x \<in> V"
  and ewalk: "G.ewalk x p y"
  obtains e where "e \<in> set p" and "x \<rightarrow>d\<star> fst e" and "\<not> x \<rightarrow>d\<star> snd e"
proof -
  note vert_ewalk_edge_verts[OF vert ewalk]
  with ewalk nr vert that show ?thesis
  proof (induction rule: G.ewalk_induct_rev)
    case Base hence "False" using dfs.self_reachable by simp
    thus ?case ..
  next
    case (Snoc v1 v2 e es)
    from Snoc.prems(4)[of e] Snoc.hyps have "v1 \<in> V" "v2 \<in> V" by simp_all
    show ?case
    proof (cases "v1 \<rightarrow>d\<star> v2")
      case True with Snoc have "\<not> x \<rightarrow>d\<star> v1" using dfs.reachable_trans by metis
      with Snoc.IH Snoc.prems show ?thesis by auto
    next
      case False with Snoc show ?thesis
      proof (cases "x \<rightarrow>d\<star> v1")
        case True with Snoc Snoc.prems(3)[of e] show ?thesis by auto
      qed auto
    qed
  qed
qed

lemma discover_finish_implies_reach:
  "v \<in> V \<Longrightarrow> w \<in> V \<Longrightarrow> \<delta> v < \<delta> w \<Longrightarrow> \<phi> v > \<phi> w \<Longrightarrow> v \<rightarrow>d+ w"
by (metis discover_finish_implies_in_etrancl[OF constructable] s_discovered s_finished dfs_reach_etrancl)

lemma no_reach_discover_finish:
  assumes verts: "v \<in> V" "w \<in> V"
  and no_reach: "\<not> v \<rightarrow>d\<star> w"
  shows "\<delta> v > \<delta> w \<or> \<phi> v < \<phi> w"
proof -
  from assms have ne: "v \<noteq> w" using dfs.reachable_ewalk[of v w] dfs.ewalk_empty_iff[of v w] by simp
  from no_reach show ?thesis
  proof (rule contrapos_np)
    assume  "\<not> (\<delta> w < \<delta> v \<or> \<phi> v < \<phi> w)"
    with verts ne have "v \<rightarrow>d+ w" using discover_finish_implies_reach[of v w] discover_neq_discover finish_neq_finish by smt
    thus "v \<rightarrow>d\<star> w" by blast
  qed
qed

lemma both_no_reach_discover_finish:
"v \<in> V \<Longrightarrow> w \<in> V \<Longrightarrow> \<not> v \<rightarrow>d\<star> w \<Longrightarrow> \<not> w \<rightarrow>d\<star> v \<Longrightarrow> \<delta> w > \<phi> v \<or> \<phi> w < \<delta> v"
using no_reach_discover_finish correct_order
by smt

lemma treeish:
  assumes reach: "v \<rightarrow>d+ w"
  shows "\<not> w \<rightarrow>d\<star> v"
proof (rule notI)
  assume "w \<rightarrow>d\<star> v"
  with no_self_reach[OF reach] dfs.reachable_neq_reachable1 have "w \<rightarrow>d+ v" by simp
  hence "\<phi> w > \<phi> v" using reach_finish by simp
  moreover from reach have "\<phi> v > \<phi> w" using reach_finish by simp
  ultimately show "False" by simp
qed

lemma discover_finish_impl_disj:
  assumes "v \<in> V" "w \<in> V"
  and "\<delta> v < \<delta> w" "\<phi> v < \<phi> w"
  shows "\<phi> v < \<delta> w"
using assms
by (smt correct_order)

lemma disjointI:
  assumes verts: "v \<in> V" "w \<in> V"
  and fini: "\<phi> v < \<delta> w"
  shows "\<delta> v < \<delta> w" and "\<delta> v < \<phi> w" and "\<phi> v < \<phi> w"
proof -
  from verts have "\<delta> v < \<phi> v" and "\<delta> w < \<phi> w" by (metis times_relation)+
  with fini show  "\<delta> v < \<delta> w" and "\<delta> v < \<phi> w" and "\<phi> v < \<phi> w" by auto
qed

text {* We now introduce the well-known notions of white, grey and black nodes. *}

definition "WHITE u x \<equiv> \<delta> x \<ge> u"
definition "GREY u x \<equiv> \<delta> x < u \<and> \<phi> x \<ge> u"
definition "BLACK u x \<equiv> \<phi> x < u"

lemma single_color:
  assumes "x \<in> V"
  shows "WHITE u x \<noteq> (GREY u x \<or> BLACK u x)"
  and "GREY u x \<noteq> (WHITE u x \<or> BLACK u x)"
  and "BLACK u x \<noteq> (WHITE u x \<or> GREY u x)"
proof -
  from assms have "\<delta> x < \<phi> x" using finish_gt_discover[OF constructable] s_finished s_discovered by force
  thus "WHITE u x \<noteq> (GREY u x \<or> BLACK u x)"
    and "GREY u x \<noteq> (WHITE u x \<or> BLACK u x)"
    and "BLACK u x \<noteq> (WHITE u x \<or> GREY u x)"
  unfolding WHITE_def GREY_def BLACK_def by auto
qed

lemma grey_impl_reach:
  assumes verts: "v \<in> V" "w \<in> V"
  and grey: "GREY u v" "GREY u w"
  shows "v \<rightarrow>d\<star> w \<or> w \<rightarrow>d\<star> v"
using verts
proof (cases rule: reachE)
  case no_reach with verts have "\<delta> w > \<phi> v \<or> \<delta> v > \<phi> w" using both_no_reach_discover_finish by simp
  with grey show ?thesis unfolding GREY_def by auto
qed simp_all

text {* The following two notions for disjunction and inclusion are just for expressing the parenthesis theorem in better-known ways. *}
definition "DISJ v w \<equiv> ( \<phi> v < \<delta> w \<or> \<phi> w < \<delta> v) \<and> \<not> v \<rightarrow>d\<star> w \<and> \<not> w \<rightarrow>d\<star> v"
definition "IN v w \<equiv> ( \<delta> v < \<delta> w \<and> \<phi> v > \<phi> w) \<and> v \<rightarrow>d+ w"

lemma reach_eq_IN:
  "v \<rightarrow>d+ w \<longleftrightarrow> IN v w"
unfolding IN_def
by (force dest: reach_finish reach_discover)

lemma not_reach_eq_DISJ:
  "v \<in> V \<Longrightarrow> w \<in> V \<Longrightarrow> \<not> v \<rightarrow>d\<star> w \<and> \<not> w \<rightarrow>d\<star> v \<longleftrightarrow> DISJ v w"
unfolding DISJ_def
by (blast dest: both_no_reach_discover_finish)

theorem parenthesis:
  assumes verts: "v \<in> V" "w \<in> V"
  and "v \<noteq> w"
  shows "DISJ v w \<noteq> (IN v w \<or> IN w v)"
  and "IN v w \<noteq> (DISJ v w \<or> IN w v)"
using assms
by (metis not_reach_eq_DISJ reach_eq_IN treeish dfs.reachable_neq_reachable1 dfs.reachable1_reachable)+

subsubsection {* White Path Theorem *}

definition p_white_path where
  "p_white_path p x y \<equiv> G.ewalk x p y \<and> (\<forall> v \<in> set (G.ewalk_verts x p). WHITE (\<delta> x) v)"

definition white_path where
  "white_path x y \<equiv> \<exists>p. p_white_path p x y"

lemma white_path_iff:
  "white_path x y \<longleftrightarrow> (\<exists>p. G.ewalk x p y \<and> (\<forall> v \<in> set (ewalk_verts x p). WHITE (\<delta> x) v))"
unfolding p_white_path_def white_path_def
by (simp add: ewalk_verts_eq)

lemma reach_impl_white_path:
  assumes "x \<rightarrow>d\<star> y" (* this could be x \<rightarrow>d+ y, but I stick with this, b/c it seems to be what the literature says *)
  shows "white_path x y"
proof (cases "x = y")
  case True with assms have  "G.ewalk x [] y" using G.ewalk_empty_iff dfs_vertices_sub dfs.reach_implies_vert by auto
  moreover
  hence "G.ewalk_verts x [] = [x]" by simp
  hence "\<forall> v \<in> set (G.ewalk_verts x []). WHITE (\<delta> x) v" unfolding WHITE_def by simp
  ultimately show ?thesis unfolding white_path_iff by force
next
  case False with assms have "x \<rightarrow>d+ y" by auto
  with reach_discover have disc: "\<delta> x < \<delta> y" .

  from assms obtain p where ewalk: "dfs.ewalk x p y" using dfs.reachable_ewalk by blast
  hence "dfs.ves p" unfolding dfs.ewalk_def by simp

  have "\<And>v. v \<in> set (dfs.ewalk_verts x p) \<Longrightarrow> WHITE (\<delta> x) v"
  proof -
    fix v
    assume "v \<in> set (dfs.ewalk_verts x p)"
    with ewalk `dfs.ves p` have "x \<rightarrow>d\<star> v"
    proof induction
      case (Base w) thus ?case using dfs.self_reachable by simp
    next
      case (Cons w1 w2 e es)
      show ?case
      proof (cases "v = w1")
        case True 
        from Cons have "w1 = fst e" by simp
        hence "w1 \<in> V" using Cons(1) verts_eq_edges fst_in_Field by force
        with True show ?thesis by (simp add: dfs.self_reachable)
      next
        case False with Cons have "w2 \<rightarrow>d\<star> v" using dfs.ewalk_verts_Cons dfs.ves_ConsD by simp
        moreover
        from Cons.hyps have "w1 \<rightarrow>d\<star> w2" using dfs.succs_reachable[unfolded dfs.succs_def] by auto
        ultimately show ?thesis using dfs.reachable_trans by metis
      qed
    qed
    thus "WHITE (\<delta> x) v" unfolding WHITE_def
    proof (cases "x=v")
      assume "x \<noteq> v" "x \<rightarrow>d\<star> v" hence "x \<rightarrow>d+ v" by auto
      with reach_discover show "\<delta> x \<le> \<delta> v" by force
    qed simp
  qed
  then show ?thesis unfolding white_path_iff ewalk_verts_eq by (blast intro: ewalk2_implies_ewalk ewalk)
qed

lemma white_path_impl_reach:
  assumes wp: "white_path x y"
  and "x \<in> V"
  shows "x \<rightarrow>d\<star> y"
proof (rule ccontr)
  assume no_reach: "\<not> x \<rightarrow>d\<star> y"

  from wp obtain p where p: "p_white_path p x y" unfolding white_path_def by blast
  hence ewalk: "ewalk x p y" unfolding p_white_path_def by simp
  with no_reach `x \<in> V` obtain e where 
    e: "e \<in> set p" "x \<rightarrow>d\<star> fst e" "\<not> x \<rightarrow>d\<star> snd e"
    using no_reach_obtain_split by blast
  then obtain u v where u: "u = fst e" and v: "v = snd e" by blast
  with e have x_u: "x \<rightarrow>d\<star> u" and x_v: "\<not> x \<rightarrow>d\<star> v" by simp_all
  hence u_ne_v: "u \<noteq> v" and x_ne_v: "x \<noteq> v" using dfs.self_reachable `x \<in> V` by force+

  from e ewalk have "e \<in> E\<^sub>G" unfolding ewalk_conv by auto
  with u v have succs: "v \<in> G.succs u" unfolding succs_def by simp

  from e u v ewalk have "u \<in> V" "v \<in> V" using vert_ewalk_edge_verts[OF `x \<in> V`] by simp_all
  note verts = this `x \<in> V`

  from x_u have xu_rel: "\<delta> x \<le> \<delta> u" "\<phi> x \<ge> \<phi> u" using reach_discover[of x u] reach_finish[of x u] dfs.reachable_neq_reachable1 by (case_tac [!] "x = u") fastforce+

  from e v ewalk have "v \<in> set (G.ewalk_verts x p)" using G.ewalk_edge_in_ewalk_verts by auto
  with p have "WHITE (\<delta> x) v" unfolding p_white_path_def by simp
  with x_ne_v have xv_rel: "\<delta> x < \<delta> v" unfolding WHITE_def using discover_neq_discover verts by force
  with correct_order[OF _ _ this] verts have xv_relf: "\<phi> x < \<delta> v \<or> \<phi> x > \<phi> v" by simp

  from verts show "False"
  proof (cases rule: reachE)
    case reach_vw thus ?thesis by (metis x_u x_v dfs.reachable_trans)
  next
    case reach_wv with u_ne_v have "v \<rightarrow>d+ u" by auto
    hence "\<delta> v < \<delta> u" and "\<phi> v > \<phi> u" using reach_discover reach_finish by simp_all
    
    from xv_relf show ?thesis
    proof (rule disjE)
      assume "\<phi> x < \<delta> v" with `\<delta> v < \<delta> u` times_relation[of u] verts have "\<phi> x < \<phi> u" by simp
      with xu_rel show "False" by simp
    next
      assume "\<phi> x > \<phi> v" with xv_rel have "x \<rightarrow>d+ v" using discover_finish_implies_reach verts by simp
      with x_v show "False" by auto
    qed
  next
    case no_reach with both_no_reach_discover_finish verts have "\<phi> v < \<delta> u \<or> \<phi> u < \<delta> v" by simp
    thus ?thesis
    proof (rule disjE)
      assume "\<phi> v < \<delta> u" with xu_rel xv_rel times_relation verts have "\<delta> x < \<delta> v" and "\<phi> x > \<phi> v" by force+
      with discover_finish_implies_reach verts x_v show "False" by auto
    next
      assume "\<phi> u < \<delta> v" with times_relation[of u] verts have "\<delta> u < \<delta> v" by simp
      moreover
      with G.completed_finished_succs_finish[OF completed _ _ succs] s_finished verts have "\<phi> v < \<phi> u" by simp
      ultimately have "u \<rightarrow>\<star> v" using discover_finish_implies_reach verts by auto
      with x_u x_v dfs.reachable_trans show "False" by metis
    qed
  qed
qed

theorem white_path:
  "x \<in> V \<Longrightarrow> white_path x y \<longleftrightarrow> x \<rightarrow>d\<star> y"
by (metis reach_impl_white_path white_path_impl_reach)

end
end