theory DFS_Test
imports 
  "../../CAVA_Automata/Digraph"
  "../../Collections/Refine_Dflt"
begin

lemma RECT_eq_REC_arb:
  fixes arb :: 'a
  assumes WF: "wf V"
  assumes I0: "I arb x"
  assumes IS: "\<And>f arb x. \<lbrakk>
    mono body;
    I arb x; 
    \<And>arb' x'. \<lbrakk>I arb' x'; (x',x)\<in>V\<rbrakk>  \<Longrightarrow> f x' \<le> M arb' x';
    f \<le> REC body
  \<rbrakk> \<Longrightarrow> body (\<lambda>x'. if ((\<exists>arb. I arb x') \<and> (x',x)\<in>V) then f x' else top) x \<le> body f x"
  assumes IS': "\<And>f arb x. \<lbrakk>
    mono body;
    I arb x; 
    \<And>arb' x'. \<lbrakk>I arb' x'; (x',x)\<in>V\<rbrakk>  \<Longrightarrow> f x' \<le> M arb' x';
    f \<le> REC body
  \<rbrakk> \<Longrightarrow> body f x \<le> M arb x"
  shows "RECT body x = REC body x"
proof (cases "mono body")
  assume "\<not>mono body"
  thus ?thesis unfolding REC_def RECT_def by simp
next
  assume MONO: "mono body"

  from lfp_le_gfp' MONO have "lfp body x \<le> gfp body x" .
  moreover have "\<forall>arb. I arb x \<longrightarrow> gfp body x \<le> lfp body x \<and> lfp body x \<le> M arb x"
    using WF
  proof (induction rule: wf_induct_rule)
    case (less x)

    from less.IH have 
      IH1: "\<And>arb y. \<lbrakk>(y,x)\<in>V; I arb y\<rbrakk> \<Longrightarrow> gfp body y \<le> lfp body y" 
      by blast
    from less.IH have 
      IH2: "\<And>arb y. \<lbrakk>(y,x)\<in>V; I arb y\<rbrakk> \<Longrightarrow> lfp body y \<le> M arb y" 
      by blast

    show ?case 
    proof (intro allI impI conjI)
      fix arb
      assume "I arb x"

      show "gfp body x \<le> lfp body x"
        apply (subst lfp_unfold[OF MONO])
        apply (subst gfp_unfold[OF MONO])
        apply (rule order_trans[OF _ IS])
        apply (rule monoD[OF MONO,THEN le_funD], rule le_funI)
        using IH1 apply auto [] 
        apply fact
        apply fact
        apply (blast intro: IH2)

        unfolding REC_def[abs_def] by (auto intro: le_funI)

      show "lfp body x \<le> M arb x"
        apply (subst lfp_unfold[OF MONO])
        apply (rule order_trans[OF _ IS'])
        apply (rule monoD[OF MONO,THEN le_funD], rule le_funI)
        apply (rule order_refl)
        apply fact
        apply fact
        apply (blast intro: IH2)
        unfolding REC_def[abs_def] by (auto intro: le_funI)
    qed
  qed
  ultimately show ?thesis
    unfolding REC_def RECT_def
    apply (rule_tac antisym)
    using I0 MONO by auto
qed

(* m fails or is smaller than m' *)
definition "LE_OR_FAIL m m' == m\<noteq>FAIL \<longrightarrow> m\<le>m'"

(* Restrict the function parameter for the recursive call to values in P.
  for all other values, it fails.*)
definition "RESTRICT_RCALL P D x \<equiv> if P x then D x else FAIL"

(*
  If the recursion body cannot catch and handle exceptions, it is enough to
  do an arbitrary correctness proof for a restricted body, in order to show 
  that RECT equals REC.

  TODO: Can we replace the "restricted body" by body in STEP-assumption?
*)
lemma RECT_eq_REC_test:
  fixes arb :: 'a
  assumes 
        NO_CATCH: "\<And>P f x. mono body 
          \<Longrightarrow> LE_OR_FAIL (body (RESTRICT_RCALL P f) x) (body f x)"
    and WF: "wf V'"
    and I0: "\<Phi> arb x"
    and STEP: "\<And>f arb x. \<lbrakk>
      mono body;
      \<Phi> arb x; 
      \<And>arb' x'. \<lbrakk>\<Phi> arb' x'; (x',x)\<in>V'\<rbrakk> \<Longrightarrow> f x' \<le> SPEC (post arb' x');
      f \<le> REC body
    \<rbrakk> \<Longrightarrow> body (\<lambda>x'. if \<exists>arb. (\<Phi> arb x') \<and> (x', x) \<in> V' then f x' else top) x \<le> SPEC (post arb x)"
  shows "REC\<^sub>T body x = REC body x"
proof (rule RECT_eq_REC_arb[of V' \<Phi> arb x _ "\<lambda>arb x. SPEC (post arb x)",OF WF I0])
  fix f arb x
  assume I: "\<Phi> arb x" and B: "f \<le> REC body"
  assume H: "\<And>arb' x'. \<lbrakk>\<Phi> arb' x'; (x', x) \<in> V'\<rbrakk> \<Longrightarrow> f x' \<le> SPEC (post arb' x')"
  assume MONO: "mono body"

  let ?D' = "(\<lambda>x'. if (\<exists>arb. \<Phi> arb x') \<and> (x', x) \<in> V' then f x' else top)"

  from STEP[OF MONO I H B] have 
    1: "body ?D' x \<le> SPEC (post arb x)" by metis
  with NO_CATCH[OF MONO, of "\<lambda>x'. (\<exists>arb. \<Phi> arb x') \<and> (x',x)\<in>V'" f x] show
    2: "body ?D' x \<le> body f x" 
    by (auto simp: LE_OR_FAIL_def RESTRICT_RCALL_def[abs_def])
  also have "body f x \<le> body ?D' x" 
    apply (rule monoD[OF MONO, THEN le_funD])
    apply (auto intro: le_funI)
    done
  finally have [simp]: "body ?D' x = body f x" .

  from 1 show "body f x \<le> SPEC (post arb x)"  by simp
qed


lemma LEOF_case_prod[refine_mono]: "(\<And>a b. LE_OR_FAIL (f a b) (g a b)) 
  \<Longrightarrow> LE_OR_FAIL (case_prod f x) (case_prod g x)"
  by (auto split: prod.split)

lemma LEOF_if[refine_mono]: 
  "\<lbrakk> LE_OR_FAIL t' t; LE_OR_FAIL e' e\<rbrakk> \<Longrightarrow> 
    LE_OR_FAIL (If b t' e') (If b t e)" by auto

lemma LEOF_let[refine_mono]: "(\<And>x. LE_OR_FAIL (f x) (f' x)) 
  \<Longrightarrow> LE_OR_FAIL (Let x f) (Let x f')"
  by simp

lemma LEOF_bind[refine_mono]: "\<lbrakk>LE_OR_FAIL m' m; (\<And>x. LE_OR_FAIL (f' x) (f x))\<rbrakk> 
  \<Longrightarrow> LE_OR_FAIL (bind m' f') (bind m f)"
  unfolding LE_OR_FAIL_def
  by (auto simp: pw_le_iff refine_pw_simps) 

lemma LEOF_refl[simp, intro!,refine_mono]: "LE_OR_FAIL m m"
  by (simp add: LE_OR_FAIL_def)

lemma LEOF_triv[simp, intro!,refine_mono]: 
  "LE_OR_FAIL FAIL m"
  "LE_OR_FAIL m FAIL"
  "LE_OR_FAIL SUCCEED m"
  by (auto simp add: LE_OR_FAIL_def)

lemma LEOF_Inf_distrib: "\<forall>x\<in>A. LE_OR_FAIL x m \<Longrightarrow> LE_OR_FAIL (Inf A) m"
  by (auto simp: LE_OR_FAIL_def intro: Inf_lower2)

lemma LEOF_RECT[refine_mono]:
  assumes LEOF: "\<And>D D' x. \<lbrakk> \<And>x. LE_OR_FAIL (D' x) (D x) \<rbrakk> \<Longrightarrow> LE_OR_FAIL (B' D' x) (B D x)"
  shows "LE_OR_FAIL (RECT B' x) (RECT B x)"
  unfolding RECT_def
  apply (cases "mono B'", simp_all)
  apply (cases "mono B", simp_all)
proof -
  assume [simp]: "mono B" "mono B'"

  have "\<forall>x. gfp B' x \<noteq> FAIL \<longrightarrow> gfp B' x \<le> gfp B x"
    apply (rule gfp_cadm_induct[where f=B], simp_all)

    apply rule 
    apply (auto intro: INF_lower2 INF_greatest) []
    
    apply auto
    apply (subst gfp_unfold, simp)
  proof -
    fix x D
    assume "\<forall>x. gfp B' x \<noteq> FAIL \<longrightarrow> gfp B' x \<le> D x"
    hence "\<forall>x. LE_OR_FAIL (gfp B' x) (D x)" by (auto simp: LE_OR_FAIL_def)
    with LEOF have "LE_OR_FAIL (B' (gfp B') x) (B D x)" by simp
    moreover assume "gfp B' x \<noteq> FAIL" hence "B' (gfp B') x \<noteq> FAIL"
      by (subst (asm) gfp_unfold, simp_all)
    ultimately show "B' (gfp B') x \<le> B D x" by (auto simp: LE_OR_FAIL_def)
  qed
  thus "LE_OR_FAIL (gfp B' x) (gfp B x)"
    by (auto simp: LE_OR_FAIL_def)
qed

lemma LEOF_RESTRICT_RCALL[simp, intro!, refine_mono]: 
  "LE_OR_FAIL (RESTRICT_RCALL P f x) (f x)"
  by (auto simp: RESTRICT_RCALL_def)

lemma LEOF_FOREACHc[refine_mono]: 
  assumes [refine_mono]: "\<And>x \<sigma>. LE_OR_FAIL (f' x \<sigma>) (f x \<sigma>)"
  shows "LE_OR_FAIL (FOREACHc s c f' \<sigma>0) (FOREACHc s c f \<sigma>0)"
  unfolding FOREACHc_def FOREACHci_def FOREACHoci_def WHILEIT_def
  unfolding WHILEI_body_def FOREACH_body_def
  apply (refine_mono)
  done


(*record 'v dfs_state =
  visited :: "'v set"
*)

(* 
 * REMARK
 *
 * Ich unterscheide zwischen 's und 's0. Dadurch ist ein Vererben möglich, ohne
 * für jede Zwischenstation einen neuen _param-Record definieren zu müssen.
 *
 * Einziger Nachteil: Typsignaturen werden ekliger, da type_synonyms nicht zusammengefaltet werden.
 *)

record ('v,'s) dfs_params = 
  enter_node :: "'v \<Rightarrow> 's \<Rightarrow> 's nres"
  leave_node :: "'v \<Rightarrow> 's \<Rightarrow> 's nres"
  known_node :: "'v \<Rightarrow> 's \<Rightarrow> 's nres"
  init_state :: "'v \<Rightarrow> 's nres"
  is_brk :: "'s \<Rightarrow> bool"

(*
 * REMARK
 *
 * Die Namen sind vorläufig.
 *
 * post_inv und fe_inv bekommen als zweites Argument immer den Zustand zum Beginn des Aufrufs.
 * Damit ist es möglich, eventuelle Zusammenhänge zwischen pre_ und post_/fe_ herzustellen.
 *
 * Die Signature von fe_inv ist recht hässlich, durch FOREACH aber so festgelegt. Kann bei Bedarf
 * aber noch geändert werden.
 *
 * Nachtrag (von der Stack-Implementierung): Graph und Startknoten als Parameter zu den Invarianten?
 *)
record ('v,'s) dfs_invs =
  pre_cond :: "'v set \<Rightarrow> 'v \<Rightarrow> 's \<Rightarrow> bool"
  post_cond :: "'v set \<Rightarrow> 'v \<Rightarrow> 's \<Rightarrow> 's \<Rightarrow>  bool"
  fe_inv :: "'v set \<Rightarrow> 'v \<Rightarrow> 's \<Rightarrow> 'v set \<Rightarrow> 's \<Rightarrow> bool"

abbreviation (input) "NOOP \<equiv> \<lambda> _ s. RETURN s"

locale fb_graph = digraph E for E :: "'v digraph" +
  fixes v0 :: 'v
  assumes finitely_branching': "\<And>v. v\<in>(E\<^sup>*)``{v0} \<Longrightarrow> finite (E``{v})"
begin
  definition "reachable \<equiv> (E\<^sup>*)``{v0}"

  abbreviation "succ v \<equiv> E``{v}"

  lemma finitely_branching[simp, intro]: "v\<in>reachable \<Longrightarrow> finite (succ v)"
    using finitely_branching' unfolding reachable_def by simp

  lemma v0_reachable[simp, intro!]: "v0\<in>reachable" unfolding reachable_def by auto

  lemma succ_reachable:
    "v \<in> reachable \<Longrightarrow> succ v \<subseteq> reachable"
    by (auto simp add: reachable_def)

  lemma reachable_trans[trans]:
    "v \<in> reachable \<Longrightarrow> v' \<in> succ v \<Longrightarrow> v' \<in> reachable"
    using succ_reachable by auto
end

(*
 * REMARK
 *
 * Die DFS-Locale ist in zwei Teile gespalten: Definition und Invarianten.
 * Damit ist es möglich, direkt von der _def zu erben ohne Beweisverpflichtungen
 * und kann so später beim Beweis der Invarianten eventuelle Definitionen mitbenutzen.
 * Desweiteren erlaubt dies, die Verpflichtungen abzuändern, anstatt nur anzuhäufen.
 *
 * Die Namen kann man auch noch ändern.
 *)
locale DFS_raw = fb_graph E v0 for E and v0 :: "'v" +
  fixes visited :: "'s \<Rightarrow> 'v set"
  and update_visited :: "('v set \<Rightarrow> 'v set) \<Rightarrow> 's \<Rightarrow> 's"
  assumes update_visited_correct[simp]:
    "visited (update_visited f s) = f (visited s)"

locale DFS_def = DFS_raw E v0 visited update_visited
     for E and v0 :: "'v"
     and visited :: "'s \<Rightarrow> 'v set" and update_visited +
  fixes params :: "('v,'s) dfs_params"
  fixes invs   :: "('v,'s) dfs_invs"
begin
  definition "rpre \<equiv> \<lambda>S v s. 
        v\<in>reachable 
      \<and> visited s\<subseteq>reachable
      \<and> S\<subseteq>visited s
      \<and> E``(visited s - S) \<subseteq> visited s
      \<and> (S = {} \<longrightarrow> v = v0)"

  abbreviation "rpre_full \<equiv> \<lambda>S v s. rpre S v s \<and> pre_cond invs S v s"

  definition "rpost \<equiv> \<lambda>S v s s'. 
          visited s \<subseteq> visited s' 
        \<and> v\<in>visited s'
        \<and> visited s'\<subseteq>reachable
        \<and> (S = {} \<longrightarrow> v = v0)
        \<and> (\<not>is_brk params s' \<longrightarrow> (E``(visited s' - S) \<subseteq> visited s'))"

  abbreviation "rpost_full \<equiv> \<lambda>S v s s'. rpost S v s s' \<and> post_cond invs S v s s'"

  definition "rfe_inv \<equiv> \<lambda>S v s it s'.
        insert v (visited s) \<subseteq> visited s'
      \<and> succ v - it \<subseteq> visited s'
      \<and> visited s'\<subseteq>reachable
      \<and> S \<subseteq> insert v (visited s)
      \<and> (S = {} \<longrightarrow> v = v0)
      \<and> (\<not>is_brk params s' 
        \<longrightarrow> E``(visited s' - S) \<subseteq> visited s' \<union> it 
          \<and> E``(visited s' - insert v S) \<subseteq> visited s')"

  abbreviation "rfe_inv_full \<equiv> \<lambda>S v s it s'. rfe_inv S v s it s' \<and> fe_inv invs S v s it s'"
  
  lemma rpre_finite_reachable:
    assumes "rpre S v s"
    shows "finite (succ v)"
    using assms[unfolded rpre_def] finitely_branching
    by blast

  definition "dfs_aux_body \<equiv> \<lambda>D (s,v).
    if v\<in>visited s then
       known_node params v s
    else do {
       let s' = update_visited (insert v) s;
       s'\<leftarrow>enter_node params v s';
       s'\<leftarrow>FOREACHc
          (succ v) 
          (Not \<circ> is_brk params) 
          (\<lambda>v' s. D (s,v')) s';
            
          if is_brk params s' then RETURN s'
          else leave_node params v s'
    }"

  definition "dfs_aux s v \<equiv> REC dfs_aux_body (s,v)"

  definition "dfs_init \<equiv> init_state params v0 \<guillemotright>= RETURN \<circ> update_visited (\<lambda>_. {})"

  definition "dfs \<equiv> do {
    s0 \<leftarrow> dfs_init;
    s \<leftarrow> dfs_aux s0 v0;
    RETURN (s0, s)
  }"
end

(* TODO: move *)
lemma REC_rule_arb':
  fixes x::"'x" and arb::'arb
  assumes M: "mono body"
  assumes I0: "\<Phi> arb x"
  assumes IS: "\<And>f arb x. \<lbrakk>
    \<And>arb' x. \<Phi> arb' x \<Longrightarrow> f x \<le> M arb' x; \<Phi> arb x
  \<rbrakk> \<Longrightarrow> body f x \<le> M arb x"
  shows "REC body x \<le> M arb x"
  apply (rule REC_rule_arb[where M=M and \<Phi>=\<Phi>])
  apply fact
  apply fact
  apply (rule IS)
  apply assumption+
  done

(* TODO: move *)
lemma weaken_SPEC:
  "m \<le> SPEC \<Phi> \<Longrightarrow> (\<And>x. \<Phi> x \<Longrightarrow> \<Psi> x) \<Longrightarrow> m \<le> SPEC \<Psi>"
  by (force elim!: order_trans)

(* TODO: move to Refine_Basic/convenience lemmas *)
lemma strengthen_SPEC: "m \<le> SPEC \<Phi> \<Longrightarrow> m \<le> SPEC(\<lambda>s. inres m s \<and> \<Phi> s)"
  -- "Strengthen SPEC by adding trivial upper bound for result"
  by (auto simp: pw_le_iff refine_pw_simps)

(* TODO: rename ... and if useful: move *)
lemma FOREACHc_mono':
  assumes "\<Phi> S s0"
  assumes "\<And>it x \<sigma>. \<lbrakk> \<Phi> it \<sigma>; x \<in> it; it \<subseteq> S; c \<sigma> \<rbrakk> \<Longrightarrow> f x \<sigma> \<le> f' x \<sigma>"
  assumes "\<And>it x \<sigma> \<sigma>'. \<lbrakk> \<Phi> it \<sigma>; x \<in> it; it \<subseteq> S; c \<sigma>; inres (f' x \<sigma>) \<sigma>' \<rbrakk> \<Longrightarrow> \<Phi> (it - {x}) \<sigma>'"
  shows "FOREACHc S c f s0 \<le> FOREACHc S c f' s0"
  using assms apply -
  apply (rule refine_IdD)
  apply (rule FOREACHc_refine[where \<alpha>=id and R=Id and \<Phi>''="\<lambda>S \<sigma> _ _. \<Phi> S \<sigma>"])
    apply simp_all
  apply (rule pw_ref_svI)
    apply (blast intro: single_valuedI)
  apply (auto simp add: refine_pw_simps pw_le_iff)
  done


(* TODO: if useful: move *)
lemma bind_eq:
  assumes "m \<le> SPEC \<Phi>"
  and "\<And>x. \<Phi> x \<Longrightarrow> f x = f' x"
  shows "m \<guillemotright>= f = m \<guillemotright>= f'"
  using assms
  by (auto simp add: pw_eq_iff pw_le_iff refine_pw_simps)

(*
 * REMARK
 *
 * Die Beweisverpflichtungen sind sehr ausführlich. Eventuell kann man sie auch noch
 * etwas leserlicher gestalten, indem ein Teil hinter Definitionen versteckt wird.
 *
 * Die Verpflichtungen sind auch recht lang, da der Zustand zum Beginn des Aufrufs "s"
 * immer mitgeführt wird. Es hat sich aber in der Erfahrung gezeigt, dass das in der Regel
 * recht praktisch ist.
 *
 * Namen sind vorläufig.
 *
 * Nachtrag: Wenn E und v0 zu den Invarianten dazukommen, können auch noch einige Reachability-Fakten
 * mit aufgenommen werden bei den Bedingungen (sowas wie v \<in> E\<^sup>* `` {v0})
 *
 * Terminierungs-Argument: finite reachable wäre ausreichend, man könnte sich aber auch schwächere Bedingungen
 *      vorstellen, wo die Parametrisierung eine Terminierung garantiert, obwohl "infinite reachable". (Zb Suche bis zur Tiefe 10)
 *
 *
 *)

locale DFS = DFS_def +
  assumes enter_node_inv: 
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; s' = update_visited (insert v) s\<rbrakk>
      \<Longrightarrow> enter_node params v s' \<le> SPEC (\<lambda>s''. 
            visited s'' = visited s' 
            \<and> fe_inv invs S v s (succ v) s'')"

  and fe_inv_recurse:
    "\<lbrakk> rpre_full S v s;
       v \<notin> visited s; 
       \<not>is_brk params s'; 
       rfe_inv_full S v s it s';
       v' \<in> it; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> pre_cond invs (insert v S) v' s'"

  and fe_inv_step:
    "\<lbrakk>rpre_full S v s;
      v \<notin> visited s; \<not>is_brk params s';
      v' \<in> it; it \<subseteq> succ v;
      rfe_inv_full S v s it s'; rpre_full (insert v S) v' s';
      rpost_full (insert v S) v' s' s''\<rbrakk>
      \<Longrightarrow> fe_inv invs S v s (it - {v'}) s''"

  and fe_inv_brk: 
     "\<lbrakk>rpre_full S v s; v \<notin> visited s;
       rfe_inv_full S v s it s'; 
       it \<subseteq> succ v; is_brk params s'\<rbrakk>
      \<Longrightarrow> post_cond invs S v s s'"

  and leave_node_inv:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      rfe_inv_full S v s {} s'; \<not> is_brk params s'\<rbrakk>
      \<Longrightarrow> leave_node params v s' \<le> SPEC (\<lambda>s''. visited s'' = visited s' \<and> post_cond invs S v s s'')"

  and known_node_inv:
    "\<lbrakk>rpre_full S v s; v \<in> visited s\<rbrakk> 
      \<Longrightarrow> known_node params v s \<le> SPEC (\<lambda>s'. 
           visited s' = visited s 
         \<and> post_cond invs S v s s')"

  and init_state_inv:
    "init_state params v \<le> SPEC (\<lambda>es. 
         let s = update_visited (\<lambda>_. {}) es in 
           \<not> is_brk params s \<and> pre_cond invs {} v s
         )"
begin

  lemma dfs_init_correct: "
    dfs_init \<le> SPEC (\<lambda>s0.
         visited s0 = {}
       \<and> rpre {} v0 s0 \<and> pre_cond invs {} v0 s0)"
    unfolding dfs_init_def
    apply (refine_rcg refine_vcg order_trans[OF strengthen_SPEC[OF init_state_inv]])
    by (simp_all add: rpre_def)

  lemma (in DFS_def) vc_enter_foreach: (* fe_inv initial *)
    assumes "v \<notin> visited s"
    and "rpre_full S v s"
    shows "rfe_inv S v s (succ v) (update_visited (insert v) s)"
    using assms unfolding rpre_def rfe_inv_def 
    by auto

  lemma vc_enter_foreach_full:
    assumes "v \<notin> visited s"
    and "rpre_full S v s"
    shows "enter_node params v (update_visited (insert v) s) \<le> SPEC (\<lambda>s'. rfe_inv_full S v s (succ v) s')"
      (* we can't use vc_enter_foreach here, as 'is_brk' may differ after enter_node *)
      apply (rule weaken_SPEC[OF enter_node_inv])
      using assms apply blast+
      using assms unfolding rpre_def rfe_inv_def apply fastforce
    done

  lemma (in DFS_def) vc_rec_pre: (* fe_inv ensures rpre *)
    assumes A: "v \<notin> visited s" "v' \<in> it" "it \<subseteq> succ v"
    and [simp]: "\<not>is_brk params s'"
    and FEI: "rfe_inv_full S v s it s'"
    and PRE: "rpre_full S v s"
    shows "rpre (insert v S) v' s'"
  proof - 
    from PRE have "v \<in> reachable" by (auto simp add: rpre_def)
    also from A have "v' \<in> succ v" by auto
    finally have [simp]: "v'\<in>reachable" .

    from FEI PRE show ?thesis
      by (auto simp add: rfe_inv_def rpre_def)
  qed

  lemma vc_rec_pre_full:
    assumes "v \<notin> visited s" "v' \<in> it" "it \<subseteq> succ v"
    and "\<not>is_brk params s'"
    and "rfe_inv_full S v s it s'"
    and "rpre_full S v s"
    shows "rpre_full (insert v S) v' s'"
    using assms vc_rec_pre
    by (fastforce intro!: fe_inv_recurse)

  lemma (in DFS_def) vc_iterate_foreach: (* rpost implies fe_inv *)
    assumes "v \<notin> visited s" "\<not>is_brk params s'"
    and "v' \<in> it" "it \<subseteq> succ v"
    and "rpre_full S v s" "rpre_full (insert v S) v' s'"
    and "rfe_inv_full S v s it s'"
    and "rpost_full (insert v S) v' s' s''"
    shows "rfe_inv S v s (it - {v'}) s''"
    using assms
    unfolding rpre_def rpost_def rfe_inv_def
    by clarsimp blast

  lemma vc_iterate_foreach_full:
    assumes "v \<notin> visited s" "\<not>is_brk params s'"
    and "v' \<in> it" "it \<subseteq> succ v"
    and "rpre_full S v s" "rpre_full (insert v S) v' s'"
    and "rfe_inv_full S v s it s'"
    and "rpost_full (insert v S) v' s' s''"
    shows "rfe_inv_full S v s (it - {v'}) s''"
    using assms vc_iterate_foreach
    by (blast intro: fe_inv_step)

  lemma (in DFS_def) vc_node_visited: (* Case: Node already visited *)
    assumes "rpre_full S v s"
    and "v \<in> visited s"
    shows "rpost S v s s"
    using assms
    unfolding rpost_def rpre_def
    by simp

  lemma vc_node_visited_full:
    assumes "rpre_full S v s"
    and "v \<in> visited s"
    shows "known_node params v s \<le> SPEC (rpost_full S v s)"
      (* we cannot use vc_node_visited here, as is_brk may differ after known_node *)
      apply (rule weaken_SPEC[OF known_node_inv])
      using assms apply blast+
      using assms apply (simp add: rpost_def rpre_def)
    done

  lemma (in DFS_def) vc_foreach_completed_imp_post: (* fe_inv (completed) implies rpost *)
    assumes "rpre_full S v s" 
    and "v \<notin> visited s" "\<not>is_brk params s'"
    and "rfe_inv_full S v s {} s'"
    shows "rpost S v s s'"
    using assms
    by (clarsimp simp: rfe_inv_def rpost_def)

  lemma vc_foreach_completed_imp_post_full:
    assumes "rpre_full S v s" 
    and "v \<notin> visited s" "\<not>is_brk params s'"
    and "rfe_inv_full S v s {} s'"
    shows "leave_node params v s' \<le> SPEC (rpost_full S v s)"
    using assms vc_foreach_completed_imp_post
    by (fastforce intro!: weaken_SPEC[OF leave_node_inv] simp add: rpost_def)
 
  lemma (in DFS_def) vc_foreach_interrupted_imp_post: (* fe_inv (interrupted) implies rpost *)
    assumes "rpre_full S v s" 
    and  "v \<notin> visited s" "it \<subseteq> succ v" "is_brk params s'"
    and  "rfe_inv_full S v s it s'"
    shows "rpost S v s s'" 
    using assms
    by (auto simp add: rpre_def rpost_def rfe_inv_def)
   
  lemma vc_foreach_interrupted_imp_post_full:
    assumes "rpre_full S v s" 
    and  "v \<notin> visited s" "it \<subseteq> succ v" "is_brk params s'"
    and  "rfe_inv_full S v s it s'"
    shows "rpost_full S v s s'"
    using assms
    by (metis fe_inv_brk vc_foreach_interrupted_imp_post)

  lemma dfs_aux_correct:
    assumes "rpre_full S v s"
    shows "dfs_aux s v \<le> SPEC (rpost_full S v s)"
  proof -
    (* add rpost', as refine_rcg else fails to generalize S, v, and s *)
    def rpost' \<equiv> "\<lambda>S (s,v). rpost_full S v s"
      
    have "dfs_aux s v \<le> SPEC (rpost' S (s,v))"
      unfolding dfs_aux_def dfs_aux_body_def
      apply (refine_rcg refine_vcg 
        REC_rule_arb'[where \<Phi>="\<lambda>S (s,v). rpre_full S v s"]
        FOREACHc_rule'[where I="rfe_inv_full S v s" for S v s]
        order_trans[OF vc_enter_foreach_full]
        )
      unfolding rpost'_def

      apply (vc_solve 
        solve: asm_rl vc_foreach_completed_imp_post_full vc_foreach_interrupted_imp_post_full
        simp: 
          assms 
          vc_node_visited_full
          rpre_finite_reachable
          vc_iterate_foreach_full
          )

      
      apply (rule order_trans)
      apply rprems
      apply (erule vc_rec_pre_full)
      apply blast+

      apply clarsimp
      apply (rule vc_iterate_foreach_full)
      apply blast+
      apply (rule vc_rec_pre_full)
      apply blast+
      done
    thus ?thesis by (simp add: rpost'_def)
  qed
    
  lemma dfs_correct: "dfs \<le> SPEC (\<lambda>(s0,s).
      (\<not> is_brk params s \<longrightarrow> visited s = reachable)
    \<and> rpost {} v0 s0 s \<and> post_cond invs {} v0 s0 s
    \<and> inres dfs_init s0)"
    unfolding dfs_def
    apply (refine_rcg refine_vcg order_trans[OF dfs_aux_correct] order_trans[OF strengthen_SPEC[OF dfs_init_correct]])
    apply (auto simp add: rpost_def)
    apply (force dest!: Image_closed_trancl simp: reachable_def)
    done

  (* This rule ignores the 'inres dfs_init s0'. If needed, uses dfs_correct_rule_full or dfs_correct directly *)
  lemma dfs_correct_rule:
    assumes P: "\<And>s0 s. \<lbrakk>rpost {} v0 s0 s; post_cond invs {} v0 s0 s; \<not> is_brk params s \<longrightarrow> visited s = reachable \<rbrakk> \<Longrightarrow> P s"
    shows "dfs \<le> SPEC (\<lambda>(_,s). P s)"
    by (auto intro: P weaken_SPEC[OF dfs_correct])

  lemma dfs_correct_extend:
    assumes P: "dfs \<le> SPEC (\<lambda>(_,s). P s)"
    and Q: "\<And>s0 s. \<lbrakk>rpost {} v0 s0 s; post_cond invs {} v0 s0 s; \<not> is_brk params s \<longrightarrow> visited s = reachable; P s \<rbrakk> \<Longrightarrow> Q s"
    shows "dfs \<le> SPEC (\<lambda>(_,s). Q s)"
    using SPEC_rule_conjI[OF P dfs_correct]
    by (auto intro: Q elim: weaken_SPEC)

  (* Also allows to assume something about the initial state.*)
  lemma dfs_correct_rule_full:
    assumes "\<And>s0. \<lbrakk> visited s0 = {}; rpre {} v0 s0; pre_cond invs {} v0 s0 \<rbrakk> \<Longrightarrow> \<Phi> s0"
    and "\<And>s0 s. \<lbrakk> rpost {} v0 s0 s; post_cond invs {} v0 s0 s; \<not> is_brk params s \<longrightarrow> visited s = reachable; \<Phi> s0 \<rbrakk>
          \<Longrightarrow> P (s0, s)"
    shows "dfs \<le> SPEC P"
    apply (rule weaken_SPEC[OF dfs_correct])
    by (auto simp add: assms inres_SPEC[OF _ dfs_init_correct])
end


locale finite_DFS = DFS +
  assumes finite_reachable: "finite reachable"
begin

  definition "dfs_aux_finite s v \<equiv> REC\<^sub>T dfs_aux_body (s,v)"

  definition upper_bound where
    "upper_bound = {((s,v),(s',v')). (visited s, visited s') \<in> (finite_psupset reachable)}"
  
  lemma upper_bound_alt: "upper_bound 
    = inv_image (finite_psupset reachable) (\<lambda>(s,v). visited s)"
    unfolding upper_bound_def by auto

  lemma wf_upper_bound:
    "wf upper_bound"
    unfolding upper_bound_alt
    by (blast intro: finite_reachable)

  (* Dieser Beweis spielt den Korrektheitsbeweis nach. Er könnte hier 
    schwächere Invarianten verwenden, muss er aber nicht. 
    Das genaue Setup, insbesondere die Form der RECT_eq_REC_test-Regel ist aber 
    noch vorläufig
  *)
  lemma dfs_aux_finite_eq:
    notes exI[simp] (* TODO: Should this be declared globally? *)
    assumes "rpre_full S v s"
    shows "dfs_aux_finite s v = dfs_aux s v"
    unfolding dfs_aux_finite_def dfs_aux_def
    apply (rule RECT_eq_REC_test[OF _ wf_upper_bound, 
        where \<Phi>="\<lambda>S (s,v). rpre_full S v s" and post = "\<lambda>S (s,v). rpost_full S v s"
        ])
    apply (simp add: dfs_aux_body_def, refine_mono)

    using assms apply auto []

    apply (subst dfs_aux_body_def)

    apply (refine_rcg refine_vcg 
      FOREACHc_rule'[where I="rfe_inv_full S v s" for S v s]
      order_trans[OF vc_enter_foreach_full] 
      )

    apply (vc_solve 
      solve: asm_rl vc_foreach_completed_imp_post_full
      simp: 
        assms 
        vc_node_visited_full
        rpre_finite_reachable   
        vc_iterate_foreach_full
        vc_foreach_interrupted_imp_post_full
        )

    apply (rule order_trans, rprems)
    apply (rule vc_rec_pre_full, blast+)[]
    apply assumption

    apply clarsimp
    apply (rule vc_iterate_foreach_full, blast+)[]
    apply (rule vc_rec_pre_full, blast+)
    apply (frule vc_rec_pre_full, assumption+)[]
    apply blast apply blast
    apply (auto simp add: upper_bound_def finite_psupset_def rfe_inv_def)[]
    apply (metis (full_types) vc_foreach_interrupted_imp_post_full)+
    done
   
  lemma dfs_aux_finite_eq_alternative_proof:
    assumes "rpre_full S v s"
    shows "dfs_aux_finite s v = dfs_aux s v"
  proof -
    {
      fix S v v' s \<sigma> \<sigma>' f it
      assume "rpre_full S v s"
        and AUX: "\<And>s v. f (s, v) \<le> dfs_aux s v"
        and  "v \<notin> visited s"
        and "rfe_inv_full S v s it \<sigma>"
        and "\<not> is_brk params \<sigma>"
        and "it \<subseteq> succ v" "v' \<in> it"
        and INR: "inres (f (\<sigma>, v')) \<sigma>'"

      have "rpre_full (insert v S) v' \<sigma>"
        by (rule vc_rec_pre_full) fact+
      with order_trans[OF AUX dfs_aux_correct] have "f (\<sigma>,v') \<le> SPEC (rpost_full (insert v S) v' \<sigma>)" .
      with INR have "rpost_full (insert v S) v' \<sigma> \<sigma>'" by (blast dest: inres_SPEC)
      
      have "rfe_inv_full S v s (it - {v'}) \<sigma>'"
        by (rule vc_iterate_foreach_full) fact+
    } note rfe_inv_step = this

    {
      fix S v s f
      assume "rpre_full S v s" "\<And>s v. f (s,v) \<le> dfs_aux s v"
      hence "dfs_aux_body (\<lambda>x'. if (case x' of (s',v') \<Rightarrow> \<exists>S. rpre_full S v' s') \<and> (x',s,v) \<in> upper_bound then f x' else FAIL) (s,v) \<le> dfs_aux_body f (s,v)"
        unfolding dfs_aux_body_def
        apply clarsimp
        apply (rule bind_mono[OF order_refl])
        apply (rule bind_mono)
          apply (drule order_trans[OF _ vc_enter_foreach_full], blast+)[]
          apply (rule FOREACHc_mono'[where \<Phi>="\<lambda>it s'. rfe_inv_full S v s it s'"])
            apply clarsimp_all
          apply (frule vc_rec_pre_full, blast+)[]
          apply (auto simp add: upper_bound_def finite_psupset_def rfe_inv_def)[]
        apply (metis rfe_inv_step)
        done
    } note term_rule = this

    show ?thesis
      unfolding dfs_aux_finite_def dfs_aux_def
      apply (rule RECT_eq_REC'[OF wf_upper_bound, where I="\<lambda>(s,v). \<exists>S. rpre_full S v s"])
      apply (auto intro!: term_rule assms simp: dfs_aux_def)
      done
  qed

  definition "dfs_finite \<equiv> do {
    s0 \<leftarrow> dfs_init;
    s \<leftarrow> dfs_aux_finite s0 v0;
    RETURN (s0, s)}"

  lemma dfs_finite_eq:
    "dfs_finite = dfs"
    unfolding dfs_finite_def dfs_def 
    by (rule bind_eq[OF dfs_init_correct]) (auto simp: dfs_aux_finite_eq)
  
  lemma dfs_finite_correct:
    "dfs_finite \<le> SPEC (\<lambda>(s0, s). 
       (\<not> is_brk params s \<longrightarrow> visited s = reachable)
       \<and> rpost {} v0 s0 s \<and> post_cond invs {} v0 s0 s  
       \<and> inres dfs_init s0)"
    by (simp add: dfs_correct dfs_finite_eq)
end

record 'v search_state =
  visited :: "'v set"
  brk :: bool 

locale search = fb_graph E v0 
  for E and v0 :: "'v" +
  fixes vd :: "'v" 
begin
  definition "ss_param \<equiv> \<lparr>
    enter_node = \<lambda>v s. RETURN (if v=vd then s\<lparr> brk := True \<rparr> else s\<lparr> brk := False \<rparr>),
    leave_node = NOOP,
    known_node = NOOP,
    init_state = \<lambda>_. RETURN \<lparr> visited = {}, brk = False \<rparr>,
    is_brk = brk
  \<rparr>"

  definition "ss_inv \<equiv> \<lparr>
     pre_cond = \<lambda>S v s. \<not>brk s \<and> vd \<notin> visited s,
     post_cond = \<lambda>S v s s'. brk s' \<longleftrightarrow> vd \<in> visited s',
     fe_inv = \<lambda>S v s it s'. brk s' \<longleftrightarrow> vd \<in> visited s'
  \<rparr>"

  sublocale DFS E v0 visited visited_update ss_param ss_inv
    apply unfold_locales
    apply (simp_all add: ss_param_def ss_inv_def)
    done

  lemma search_correct: "dfs \<le> SPEC (\<lambda>(_,s). brk s \<longleftrightarrow> vd\<in>reachable)"
    apply (rule dfs_correct_rule)
    unfolding rpost_def
    unfolding ss_inv_def ss_param_def
    by force
end

(*
 * REMARK
 *
 * Bei der Stackimplementierung wurde auf ein eigener _param und _inv Record verzichtet
 * (s.o.). Das macht die Definitionen von dfs_stack_params und dfs_stack_invs einfacher.
 *
 * dfs_stack_invs benötigt auch "params", da auf brk geprüft werden muss.
 *
 * Die Invarianten können in der momentanen Form nichts über den Zusammenhang Stack/Graph
 * aussagen, da der Graph ein Parameter des dfs und nicht der Locale ist. Eventuell sollte
 * man das ändern.
 *
 * Auch das Hinzufügen eines "start"-Parameters für den State könnte interessant sein,
 * um Eigenschaften auf dem Stack zu zeigen (ansonsten ist dies, ähnlich wie der Graph) nicht
 * erreichbar.
 *
 * Nachtrag zu den beiden oberen Punkten: Oder jeweils als Parameter für die Invarianten.
 *)

(*record 'v stack_state = "'v dfs_state" + 
  stack :: "'v list"

type_synonym ('v,'s) dfs_stack_params = "('v, ('v,'s) stack_state_ext,'s) dfs_params"
type_synonym ('v,'s) dfs_stack_invs = "('v, ('v,'s) stack_state_ext) dfs_invs"
*)

locale stack_DFS_raw = DFS_raw
  where update_visited = update_visited
  for update_visited :: "('v set \<Rightarrow> _) \<Rightarrow> 's \<Rightarrow> 's" +
  fixes stack :: "'s \<Rightarrow> 'v list"
  and update_stack :: "('v list \<Rightarrow> 'v list) \<Rightarrow> 's \<Rightarrow> 's"
  assumes update_stack_correct[simp]:
    "\<And>f. stack (update_stack f s) = f (stack s)"
    "\<And>f. visited (update_stack f s) = visited s"
    "\<And>f. stack (update_visited f s) = stack s"

locale stack_DFS_def =  stack_DFS_raw
  where update_stack = update_stack
  for update_stack :: "('v list \<Rightarrow> _) \<Rightarrow> 's \<Rightarrow> 's"+
  fixes params :: "('v,'s) dfs_params"
  fixes invs   :: "('v,'s) dfs_invs"
begin

  definition stack_params where
 "  stack_params = \<lparr>
    enter_node = \<lambda>v s. 
      let s = update_stack (Cons v) s
      in enter_node params v s,

    leave_node = \<lambda>v s.
      let s = update_stack tl s
      in leave_node params v s,

    known_node = known_node params,
    init_state = \<lambda>v. init_state params v \<guillemotright>= RETURN \<circ> update_stack (\<lambda>_. []),
    is_brk = is_brk params
   \<rparr>"

  abbreviation (input) "stack_common_inv s \<equiv> 
                                   distinct (stack s)
                                 \<and> set (stack s) \<subseteq> visited s
                                 \<and> (stack s \<noteq> [] \<longrightarrow> path E v0 (rev (tl (stack s))) (hd (stack s))
                                                  \<and> last (stack s) = v0)"

  definition "stack_pre \<equiv> \<lambda>S v s. stack_common_inv s
                    \<and> set (stack s) = S
                    \<and> (stack s = [] \<longleftrightarrow> visited s = {})
                    \<and> (stack s \<noteq> [] \<longrightarrow> (hd (stack s), v) \<in> E)"

  definition "stack_post \<equiv> \<lambda>S v s s'. stack_common_inv s'
                       \<and> set (stack s) = S
                       \<and> (\<exists>xs. stack s' = xs@stack s)
                       \<and> (\<not> is_brk params s' \<longrightarrow> stack s = stack s')"

  definition "stack_fe \<equiv> \<lambda>S v s it s'. stack s' \<noteq> [] 
                       \<and> set (stack s) = S
                       \<and> stack_common_inv s'
                       \<and> (\<exists>xs. stack s' = xs@stack s)
                       \<and> (\<not> is_brk params s' \<longrightarrow> stack s' = v#stack s)"

  definition "stack_invs = \<lparr>
    pre_cond = \<lambda>S v s. stack_pre S v s \<and> pre_cond invs S v s,
    post_cond = \<lambda>S v s s'. stack_post S v s s' \<and> post_cond invs S v s s',
    fe_inv = \<lambda>S v s it s'. stack_fe S v s it s' \<and> fe_inv invs S v s it s'
    \<rparr>"

  bundle stack_invs_def = stack_invs_def[simp]

  schematic_lemma stack_params_simps[simp]:
    "enter_node stack_params v s = ?EN"
    "leave_node stack_params v s = ?LN"
    "known_node stack_params v s = ?KN"
    "init_state stack_params v = ?IS"
    "is_brk stack_params s = ?IB"
    unfolding stack_params_def dfs_params.simps
    by (rule refl)+

  schematic_lemma stack_invs_simps':
    "pre_cond stack_invs S v s = ?PC"
    "post_cond stack_invs S v s s' = ?PoC"
    "fe_inv stack_invs S v s it s' = ?FEI"
    unfolding stack_invs_def stack_pre_def stack_post_def stack_fe_def dfs_invs.simps
    by (rule refl)+

  lemmas stack_invs_simps = stack_invs_simps' stack_pre_def stack_post_def stack_fe_def
  
  lemma stack_invs_pre_cond[intro]:
    "pre_cond stack_invs S v s \<Longrightarrow> pre_cond invs S v s"
    by (simp add: stack_invs_simps)

  lemma stack_invs_post_cond[intro]:
    "post_cond stack_invs S v s s' \<Longrightarrow> post_cond invs S v s s'"
    by (simp add: stack_invs_simps)

  lemma stack_invs_fe_inv[intro]:
    "fe_inv stack_invs S v s it s' \<Longrightarrow> fe_inv invs S v s it s'"
    by (simp add: stack_invs_simps)

  sublocale DFS_def 
    where params = stack_params and invs = stack_invs
    ..

  lemma (in -) path_append':
    assumes "s \<noteq> []"
    and "path E v (butlast s) (last s)"
    and "(last s, v') \<in> E"
    shows "path E v s v'"
    using assms
    by (cases s rule: rev_cases) (simp_all add: path_append)

  lemma pre_cond_imp_fe_inv:
    assumes "rpre S v s"
    and "stack_pre S v s"
    and "v \<notin> visited s"
    and "visited x = insert v (visited s)" "stack x = v # stack s"
    shows "stack_fe S v s (succ v) x"
    using assms path_append'[where E=E] 
    unfolding stack_invs_simps rpre_def
    by auto

  lemma fe_inv_imp_pre_cond:
    assumes "stack_pre S v s"
    and "v \<notin> visited s"
    and "\<not> is_brk stack_params s'"
    and "rfe_inv S v s it s'"
    and "stack_fe S v s it s'"
    and "v' \<in> it"
    and "it \<subseteq> succ v"
    shows "stack_pre (insert v S) v' s'"
    using assms
    apply (unfold stack_invs_simps stack_params_simps)
      apply (intro conjI)
      apply (simp_all add: rpre_def rfe_inv_def stack_invs_simps)
      apply fast
      apply (metis subset_trans)
      apply (intro conjI impI)
      apply clarsimp_all[3]
      apply fast+
      done

  lemma post_cond_imp_fe_inv:
    assumes "\<not> is_brk stack_params s'"
    and "stack_fe S v s it s'"
    and "stack_post (insert v S) v' s' s''"
    shows "stack_fe S v s (it - {v'}) s''"
    using assms
    by (simp add: stack_invs_simps) force

  lemma fe_inv_imp_post_cond_brk:
    assumes "stack_fe S v s it s'"
    and "is_brk stack_params s'"
    shows "stack_post S v s s'"
    using assms
    by (simp add: stack_invs_simps)

  lemma fe_inv_imp_post_cond_not_brk:
    assumes "stack_pre S v s"
    and "v \<notin> visited s"
    and "rfe_inv S v s {} s'"
    and "stack_fe S v s {} s'"
    and "\<not> is_brk params s'"
    and "visited x = visited s'"
    and "stack x = tl (stack s')"
    shows "stack_post S v s x"
    using assms
    by (auto simp add: rfe_inv_def stack_invs_simps)

  lemma pre_visited_imp_post_cond:
    assumes "rpre S v s"
    and "stack_pre S v s"
    and "v \<in> visited s"
    and "visited x = visited s" "stack x = stack s"
    shows "stack_post S v s x"
    using assms by (simp add: rpre_def stack_invs_simps)
end

locale stack_DFS = stack_DFS_def +
   assumes enter_node_inv: 
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; s' = update_stack (Cons v) (update_visited (insert v) s)\<rbrakk>
      \<Longrightarrow> enter_node params v s' \<le> SPEC (\<lambda>s''. 
            visited s'' = visited s' \<and> stack s'' = stack s'
          \<and> fe_inv invs S v s (succ v) s'')"

  and fe_inv_recurse:
    "\<lbrakk> rpre_full S v s;
       v \<notin> visited s; 
       \<not>is_brk params s'; 
       rfe_inv_full S v s it s';
       v' \<in> it; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> pre_cond invs (insert v S) v' s'"

  and fe_inv_step:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; \<not>is_brk params s';
      v' \<in> it; it \<subseteq> succ v;
      rfe_inv_full S v s it s'; rpre_full (insert v S) v' s'; 
      rpost_full (insert v S) v' s' s''\<rbrakk>
      \<Longrightarrow> fe_inv invs S v s (it - {v'}) s''"

  and fe_inv_brk: 
     "\<lbrakk>rpre_full S v s; v \<notin> visited s;
       rfe_inv_full S v s it s'; 
       is_brk params s'; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> post_cond invs S v s s'"

  and leave_node_inv:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      rfe_inv_full S v s {} s'; \<not> is_brk params s';
      s'' = update_stack tl s'\<rbrakk>
      \<Longrightarrow> leave_node params v s'' \<le> SPEC (\<lambda>x. 
            visited x = visited s'' \<and> stack x = stack s'' 
          \<and> post_cond invs S v s x)"

  and known_node_inv:
    "\<lbrakk>rpre_full S v s; v \<in> visited s\<rbrakk>
      \<Longrightarrow> known_node params v s \<le> SPEC (\<lambda>s'. 
            visited s' = visited s \<and> stack s' = stack s 
          \<and> post_cond invs S v s s')"

  and init_state_inv:
    "init_state params v \<le> SPEC (\<lambda>es. 
         let s = update_visited (\<lambda>_. {}) (update_stack (\<lambda>_. []) es) in
           \<not> is_brk params s \<and> pre_cond invs {} v s
         )"
begin

  sublocale dfs: DFS where params = stack_params and invs = stack_invs
    including stack_invs_def
    apply default
    apply (fastforce intro!: pre_cond_imp_fe_inv weaken_SPEC[OF enter_node_inv])
    apply (fastforce intro!: fe_inv_imp_pre_cond fe_inv_recurse)
    apply (fastforce intro!: post_cond_imp_fe_inv fe_inv_step)
    apply (fastforce intro!: fe_inv_imp_post_cond_brk fe_inv_brk)
    apply (fastforce intro!: fe_inv_imp_post_cond_not_brk weaken_SPEC[OF leave_node_inv])
    apply (fastforce intro!: pre_visited_imp_post_cond weaken_SPEC[OF known_node_inv])
    apply (simp)
    apply (refine_rcg refine_vcg order_trans[OF init_state_inv])
      apply (simp add: stack_invs_simps)
    done
  
  lemma stack_dfs_correct:
    "dfs \<le> SPEC (\<lambda>(_,s). \<not> is_brk params s \<longrightarrow> (stack s = [] \<and> visited s = reachable))"
    by (rule dfs_correct_rule) (simp add: stack_invs_simps ,metis)
end

locale timing_DFS_raw = DFS_raw
  where update_visited = update_visited
  for update_visited :: "('v set \<Rightarrow> _) \<Rightarrow> 's \<Rightarrow> 's" +
  fixes counter :: "'s \<Rightarrow> nat"
  and update_counter :: "(nat \<Rightarrow> nat) \<Rightarrow> 's \<Rightarrow> 's"
  fixes discover :: "'s \<Rightarrow> ('v, nat) map"
  and  update_discover :: "(('v,nat) map \<Rightarrow> ('v,nat) map) \<Rightarrow> 's \<Rightarrow> 's"
  fixes finish :: "'s \<Rightarrow> ('v, nat) map"
  and  update_finish :: "(('v,nat) map \<Rightarrow> ('v,nat) map) \<Rightarrow> 's \<Rightarrow> 's"
  assumes update_discover_correct[simp]:
    "\<And>f. discover (update_discover f s) = f (discover s)"
    "\<And>f. discover (update_visited f s) = discover s"
    "\<And>f. visited (update_discover f s) = visited s"
    "\<And>f. finish (update_discover f s) = finish s"
    "\<And>f. counter (update_discover f s) = counter s"
  assumes update_finish_correct[simp]:
    "\<And>f. finish (update_finish f s) = f (finish s)"
    "\<And>f. finish (update_visited f s) = finish s"
    "\<And>f. visited (update_finish f s) = visited s"
    "\<And>f. discover (update_finish f s) = discover s"
    "\<And>f. counter (update_finish f s) = counter s"
   assumes update_counter_correct[simp]:
    "\<And>f. counter (update_counter f s) = f (counter s)"
    "\<And>f. counter (update_visited f s) = counter s"
    "\<And>f. visited (update_counter f s) = visited s"
    "\<And>f. discover (update_counter f s) = discover s"
    "\<And>f. finish (update_counter f s) = finish s"
begin
  abbreviation disc ("\<delta>") where "\<delta> s v \<equiv> the (discover s v)"
  abbreviation fin ("\<phi>") where "\<phi> s v \<equiv> the (finish s v)"
end

abbreviation map_upd ("_ \<longmapsto> _" 40) where "k \<longmapsto> v \<equiv> \<lambda>m. m(k\<mapsto>v)"

locale timing_DFS_def =  timing_DFS_raw 
  where update_finish = update_finish
  for update_finish :: "(('v,nat) map \<Rightarrow> _) \<Rightarrow> 's \<Rightarrow> 's" +
  fixes params :: "('v,'s) dfs_params"
  fixes invs   :: "('v,'s) dfs_invs"
begin
  definition timing_params where
 "  timing_params = \<lparr>
    enter_node = \<lambda>v s. 
      let s = update_counter Suc (update_discover (v \<longmapsto> counter s) s)
      in enter_node params v s,

    leave_node = \<lambda>v s.
      let s = update_counter Suc (update_finish (v \<longmapsto> counter s) s)
      in leave_node params v s,

    known_node = known_node params,
    init_state = \<lambda>v. init_state params v \<guillemotright>= 
         RETURN \<circ> update_discover (\<lambda>_. Map.empty) \<circ> update_finish (\<lambda>_. Map.empty) \<circ> update_counter (\<lambda>_. 0),
    is_brk = is_brk params
   \<rparr>"

  definition "parenthesis s \<equiv>
    \<forall>v \<in> dom (discover s). \<forall>w \<in> dom (discover s). v \<noteq> w \<longrightarrow>
     \<delta> s v < \<delta> s w \<and> v \<in> dom (finish s) \<longrightarrow> (
             \<phi> s v < \<delta> s w (* disjoint *)
           \<or> (\<phi> s v > \<delta> s w \<and> w \<in> dom (finish s) \<and> \<phi> s w < \<phi> s v))"

  abbreviation (input) "timing_common_inv s \<equiv> 
                    (* \<delta> s v < \<phi> s v *)
                     (\<forall>v \<in> dom (finish s). \<delta> s v < \<phi> s v)

                    (* v \<noteq> w \<longrightarrow> \<delta> s v \<noteq> \<delta> s w \<and> \<phi> s v \<noteq> \<phi> s w
                       Can't use 'card dom = card ran' as the maps may be infinite ... *)
                    \<and> (\<forall>v \<in> dom (discover s). \<forall>w \<in> dom (discover s). v \<noteq> w \<longrightarrow> \<delta> s v \<noteq> \<delta> s w)
                    \<and> (\<forall>v \<in> dom (finish s). \<forall>w \<in> dom (finish s). v \<noteq> w \<longrightarrow> \<phi> s v \<noteq> \<phi> s w)

                    (* \<delta> s v < counter \<and> \<phi> s v < counter *)
                    \<and> (\<forall>v \<in> dom (discover s). \<delta> s v < counter s)
                    \<and> (\<forall>v \<in> dom (finish s). \<phi> s v < counter s)

                    \<and> (\<forall>v \<in> dom (finish s). \<forall>w \<in> succ v. \<delta> s w < \<phi> s v)

                    \<and> parenthesis s"

  definition "timing_pre \<equiv> \<lambda>S v s. 
                      dom (discover s) = visited s
                    \<and> dom (finish s) = visited s - S
                    \<and> timing_common_inv s"

  definition "timing_post \<equiv> \<lambda>S v s s'. 
                         dom (discover s') = visited s'
                       \<and> dom (finish s') \<subseteq> visited s' - S
                       \<and> (\<not> is_brk params s' \<longrightarrow> dom (finish s') = visited s' - S)

                       \<and> discover s \<subseteq>\<^sub>m discover s'
                       \<and> finish s \<subseteq>\<^sub>m finish s'
                       
                       \<and> (\<forall>v \<in> dom (discover s') - dom (discover s). \<delta> s' v \<ge> counter s)
                       \<and> (\<forall>v \<in> dom (finish s') - dom (finish s). \<phi> s' v \<ge> counter s)

                       \<and> timing_common_inv s'
                       \<and> counter s' \<ge> counter s"

  definition "timing_fe \<equiv> \<lambda>S v s it s'.
                       dom (discover s') = visited s'
                     \<and> dom (finish s') \<subseteq> visited s' - insert v S
                     \<and> (\<not> is_brk params s' \<longrightarrow> dom (finish s') = visited s' - insert v S)
                     \<and> discover s \<subseteq>\<^sub>m discover s'
                     \<and> finish s \<subseteq>\<^sub>m finish s'
                     \<and> the (discover s' v) = counter s

                     \<and> (\<forall>v \<in> dom (discover s') - dom (discover s). \<delta> s' v \<ge> counter s)
                     \<and> (\<forall>v \<in> dom (finish s') - dom (finish s). \<phi> s' v \<ge> counter s)

                     \<and> timing_common_inv s'
                     \<and> counter s' \<ge> counter s"

  definition "timing_invs = \<lparr>
    pre_cond = \<lambda>S v s. timing_pre S v s \<and> pre_cond invs S v s,
    post_cond = \<lambda>S v s s'. timing_post S v s s' \<and> post_cond invs S v s s',
    fe_inv = \<lambda>S v s it s'. timing_fe S v s it s' \<and> fe_inv invs S v s it s' \<rparr>"

  bundle timing_invs_def = timing_invs_def[simp]

  schematic_lemma timing_params_simps[simp]:
    "enter_node timing_params v s = ?EN"
    "leave_node timing_params v s = ?LN"
    "known_node timing_params v s = ?KN"
    "init_state timing_params v = ?IS"
    "is_brk timing_params s = ?IB"
    unfolding timing_params_def dfs_params.simps
    by (rule refl)+

  schematic_lemma timing_invs_simps':
    "pre_cond timing_invs S v s = ?PC"
    "post_cond timing_invs S v s s' = ?PoC"
    "fe_inv timing_invs S v s it s' = ?FEI"
    unfolding timing_invs_def dfs_invs.simps
    by (rule refl)+

  lemmas timing_invs_simps = timing_invs_simps' timing_pre_def timing_post_def timing_fe_def

  sublocale DFS_def 
    where params = timing_params and invs = timing_invs ..

  lemma timing_invs_pre_cond[intro]:
    "pre_cond timing_invs S v s \<Longrightarrow> pre_cond invs S v s"
    by (simp add: timing_invs_simps)

  lemma timing_invs_post_cond[intro]:
    "post_cond timing_invs S v s s' \<Longrightarrow> post_cond invs S v s s'"
    by (simp add: timing_invs_simps)

  lemma timing_invs_fe_inv[intro]:
    "fe_inv timing_invs S v s it s' \<Longrightarrow> fe_inv invs S v s it s'"
    by (simp add: timing_invs_simps)
  
  (* TODO: move *)
  lemma (in -) map_le_not_in_dom:
    "v \<notin> dom m \<Longrightarrow> m \<subseteq>\<^sub>m m(v \<mapsto> k)"
    by (simp add: map_le_def)

  lemma pre_cond_imp_fe_inv:
    assumes PRE: "rpre S v s"
    and TPRE: "timing_pre S v s"
    and V: "v \<notin> visited s"
    and EQ: "visited s' = insert v (visited s)" 
            "discover s' = (discover s)(v \<mapsto> counter s)" 
            "finish s' = finish s"
            "counter s' = Suc (counter s)"
    shows "timing_fe S v s it s'"
  proof -
    have *[dest]: "\<And>v m. v \<notin> dom m \<Longrightarrow> m v = None" by blast

    note TPRE' = TPRE[unfolded timing_invs_simps]
    
    with PRE V have V': "v \<notin> dom (discover s)" "v \<notin> dom (finish s)" by (auto simp add: rpre_def)

    {
      fix m and n :: nat
      assume "\<forall>v\<in>dom m. the (m v) < n"
      hence "\<forall>v\<in>dom m. the (m v) < Suc n" by force
    } note F = this

    {
      fix x y
      assume  dom: "x \<in> dom (discover s')" "y \<in> dom (discover s')"
        and neq: "x \<noteq> y" 
        and \<delta>: "\<delta> s' x < \<delta> s' y"
        and f: "x \<in> dom (finish s')"

      let ?DISJ = "\<phi> s' x < \<delta> s' y"
      let ?IN = "\<delta> s' y < \<phi> s' x \<and> y \<in> dom (finish s') \<and> \<phi> s' y < \<phi> s' x"
      
      from dom EQ V' f have x: "x \<in> dom (discover s)""x \<noteq> v" by auto
      
      have "?DISJ \<or> ?IN"
      proof (cases "y = v")
        case True with EQ have "\<delta> s' y = counter s" by simp
        moreover from TPRE' EQ x f have "\<phi> s' x < counter s" by auto
        ultimately have "?DISJ" by simp
        thus ?thesis ..
      next
        case False with EQ dom have y: "y \<in> dom (discover s)" by simp
        from TPRE' have "parenthesis s" by simp
        with False EQ \<delta> f y x neq show ?thesis by (simp add: parenthesis_def) 
      qed
    }
    hence P: "parenthesis s'" unfolding parenthesis_def by blast

    from EQ V V' show ?thesis
      apply (simp add: timing_invs_simps TPRE')
      using V V'
      apply (intro conjI)
        apply (force intro: map_le_not_in_dom)
        apply force
        using TPRE' apply simp apply (metis less_not_refl)
        using TPRE' apply simp apply (metis less_not_refl)
        using TPRE' apply simp
        apply (metis TPRE' F)
        apply (metis TPRE' F)
        using PRE[unfolded rpre_def] apply blast
        apply (rule P)
      done
  qed

  lemma fe_inv_imp_pre_cond:
    assumes "rfe_inv S v s it s'"
    and "timing_fe S v s it s'"
    and "\<not>is_brk params s'"
    shows "timing_pre (insert v S) v' s'"
    using assms
    unfolding rfe_inv_def timing_invs_simps
    by simp

  lemma post_cond_imp_fe_inv:
    assumes FEI: "rfe_inv S v s it s'"
    and TFEI: "timing_fe S v s it s'"
    and POST: "rpost (insert v S) v' s' s''"
    and TPOST: "timing_post (insert v S) v' s' s''"
    and brk: " \<not>is_brk params s'"
    shows "timing_fe S v s (it - {v'}) s''"
  proof -
    note timing_invs_simps[simp]

    { 
      fix m
      assume A: "\<forall>v \<in> dom (m s') - dom (m s). the (m s' v) \<ge> counter s"
      assume B: "\<forall>v \<in> dom (m s'') - dom (m s'). the (m s'' v) \<ge> counter s'"
      assume sub: "m s' \<subseteq>\<^sub>m m s''" "m s \<subseteq>\<^sub>m m s'"
      hence "dom (m s) \<subseteq> dom (m s')" "dom (m s') \<subseteq> dom (m s'')" by (metis map_le_implies_dom_le)+
      moreover
      from A sub have "\<forall>v \<in> dom (m s') - dom (m s). the (m s'' v) \<ge> counter s" by (auto simp add: map_le_def)
      moreover
      from TFEI have"counter s' \<ge> counter s" by simp
      with B have "\<forall>v \<in> dom (m s'') - dom (m s'). the (m s'' v) \<ge> counter s" by (metis order_trans)
      ultimately have "\<forall>v \<in> dom (m s'') - dom (m s). the (m s'' v) \<ge> counter s" "m s \<subseteq>\<^sub>m m s''" using sub by (blast intro: map_le_trans)+
    } note trans_m = this   

    from TPOST brk have "timing_common_inv s''" "dom (discover s'') = visited s''" 
                        "dom (finish s'') \<subseteq> visited s'' - insert v S"
                        "\<not> is_brk params s'' \<longrightarrow> dom (finish s'') = visited s'' - insert v S"  by simp_all
    moreover from TPOST have "counter s'' \<ge> counter s'" by simp
    with TFEI have "counter s'' \<ge> counter s" by simp
    moreover from FEI TFEI have "\<delta> s' v = counter s" and "v \<in> dom (discover s')" by (simp_all add: rfe_inv_def)
    with TPOST have "\<delta> s'' v = counter s" by (simp add: map_le_def)
    moreover from TFEI TPOST have "\<forall>v \<in> dom (discover s'') - dom (discover s). \<delta> s'' v \<ge> counter s" "discover s \<subseteq>\<^sub>m discover s''"
      using trans_m[of discover] by simp_all
    moreover from TFEI TPOST have "\<forall>v \<in> dom (finish s'') - dom (finish s). \<phi> s'' v \<ge> counter s" "finish s \<subseteq>\<^sub>m finish s''"
      using trans_m[of finish] unfolding timing_invs_simps by blast+
    ultimately show ?thesis by simp
  qed

  lemma fe_inv_imp_post_cond_brk:
    assumes "rfe_inv S v s it s'"
    and "timing_fe S v s it s'"
    and "is_brk params s'"
    shows "timing_post S v s s'"
  proof -
    have "\<And>A B C. A \<subseteq> B - (insert v C) \<Longrightarrow> A \<subseteq> B - C" by blast
    with assms show ?thesis
      unfolding rfe_inv_def timing_invs_simps
      by simp
  qed

  lemma fe_inv_imp_post_cond_not_brk:
    assumes PRE: "rpre S v s"
    and TPRE: "timing_pre S v s"
    and "v \<notin> visited s"
    and FEI: "rfe_inv S v s {} s'"
    and TFEI: "timing_fe S v s {} s'"
    and brk: "\<not> is_brk params s'"
    and EQ: "visited s'' = visited s'" "discover s'' = discover s'" 
        "counter s'' = Suc (counter s')" "finish s'' = finish s'(v \<mapsto> counter s')"
    shows "timing_post S v s s''"
  proof -
    note timing_invs_simps[simp]

    {
      fix x y
      assume  dom: "x \<in> dom (discover s'')" "y \<in> dom (discover s'')"
        and neq: "x \<noteq> y" 
        and \<delta>: "\<delta> s'' x < \<delta> s'' y"
        and f: "x \<in> dom (finish s'')"

      note assms = dom neq \<delta> f EQ

      let ?DISJ = "\<phi> s'' x < \<delta> s'' y"
      let ?IN = "\<delta> s'' y < \<phi> s'' x \<and> y \<in> dom (finish s'') \<and> \<phi> s'' y < \<phi> s'' x"

      from TFEI have P: "parenthesis s'" by simp

      have "?DISJ \<or> ?IN"
      proof (cases "x = v")
        case True with EQ have \<phi>x: "\<phi> s'' x = counter s'" by simp
        from True neq have "y \<noteq> v" by simp
        
        have "\<delta> s' y < \<phi> s'' x \<and> y \<in> dom (finish s') \<and> \<phi> s' y < \<phi> s'' x"
        proof (cases "y \<in> S")
          -- "y on stack is not possible: According to @{fact \<delta>} it is discovered _after_ x (= v)"
          case True with PRE TPRE have "y \<in> visited s" "y \<in> dom (discover s)" by (auto simp add: rpre_def)
          with TPRE TFEI EQ have "\<delta> s'' y < counter s" by (fastforce simp add: map_le_def)
          moreover from TFEI `x=v` EQ have "\<delta> s'' x = counter s" by simp
          ultimately have False using \<delta> by simp
          thus ?thesis ..
        next
          case False
          -- "y must be a successor of x (= v)"
          from dom EQ have *: "y \<in> dom (discover s')" by simp
          with False TFEI brk `y \<noteq> v` have "y \<in> dom (finish s')" by simp
          moreover with TFEI \<phi>x have "\<phi> s' y < \<phi> s'' x" by simp
          moreover from * TFEI \<phi>x have "\<delta> s' y < \<phi> s'' x" by simp
          ultimately show ?thesis by simp
        qed
        with EQ `y \<noteq> v` show ?thesis by simp
      next
        case False note [simp] = this
        show ?thesis
        proof (cases "y = v")
          case False with assms show ?thesis 
            using P[unfolded parenthesis_def] 
            by simp
        next
          case True with TFEI have "y \<notin> dom (finish s')" by auto
          with assms have "\<phi> s' x < \<delta> s' y" using P[unfolded parenthesis_def] by force
          with EQ have ?DISJ by simp
          thus ?thesis ..
        qed
      qed
    }
    hence P: "parenthesis s''" unfolding parenthesis_def by blast

    {
      fix w
      assume "w \<in> succ v"
      with FEI brk have vis: "w \<in> visited s'" unfolding rfe_inv_def by auto
      hence "\<delta> s' w < counter s'" using TFEI by simp
    } note S = this
      

    show ?thesis
      using assms
      unfolding rfe_inv_def rpre_def
      apply simp
      apply (intro conjI)
        apply (auto simp add: map_le_def)[5]
        apply metis
        apply auto[]
        apply (metis less_not_refl)
        apply fastforce
        apply (metis less_SucI)
        apply (metis S)
        apply (rule P)
      done
  qed

  lemma parenthesis_lift:
    assumes "discover x = discover s"
            "counter x = counter s" 
            "finish x = finish s"
    and "parenthesis s"
    shows "parenthesis x"
    using assms
    by (simp add: parenthesis_def)

  lemma pre_visited_imp_post_cond:
    assumes "rpre S v s"
    and "timing_pre S v s"
    and "v \<in> visited s"
    and "visited x = visited s" "discover x = discover s"
        "counter x = counter s" "finish x = finish s"
    shows "timing_post S v s x"
    using assms
    unfolding rpre_def timing_invs_simps
    by simp (metis parenthesis_lift)
end

locale timing_DFS = timing_DFS_def +
   assumes enter_node_inv: 
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      s' = update_counter Suc (update_discover (v \<longmapsto> counter s) (update_visited (insert v) s))\<rbrakk> 
      \<Longrightarrow> enter_node params v s' \<le> SPEC (\<lambda>s''. 
            visited s'' = visited s' 
          \<and> discover s'' = discover s' \<and> counter s'' = counter s' \<and> finish s'' = finish s'
          \<and> fe_inv invs S v s (succ v) s'')"

  and fe_inv_recurse:
    "\<lbrakk> rpre_full S v s; 
       v \<notin> visited s; 
       \<not>is_brk params s'; 
       rfe_inv_full S v s it s';
       v' \<in> it; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> pre_cond invs (insert v S) v' s'"

  and fe_inv_step:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; \<not>is_brk params s';
      v' \<in> it; it \<subseteq> succ v;
      rfe_inv_full S v s it s'; rpre_full (insert v S) v' s'; 
      rpost_full (insert v S) v' s' s''\<rbrakk>
      \<Longrightarrow> fe_inv invs S v s (it - {v'}) s''"

  and fe_inv_brk: 
     "\<lbrakk>rpre_full S v s; v \<notin> visited s;
       rfe_inv_full S v s it s'; is_brk params s'; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> post_cond invs S v s s'"

  and leave_node_inv:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      rfe_inv_full S v s {} s'; \<not> is_brk params s';
      s'' = update_counter Suc (update_finish (v \<longmapsto> counter s') s')\<rbrakk>
      \<Longrightarrow> leave_node params v s'' \<le> SPEC (\<lambda>x. 
            visited x = visited s'' 
          \<and> discover x = discover s'' \<and> counter x = counter s'' \<and> finish x = finish s''
          \<and> post_cond invs S v s x)"

  and known_node_inv:
    "\<lbrakk>rpre_full S v s; v \<in> visited s\<rbrakk> 
      \<Longrightarrow> known_node params v s \<le> SPEC (\<lambda>s'. 
            visited s' = visited s 
          \<and> discover s' = discover s \<and> counter s' = counter s \<and> finish s' = finish s
          \<and> post_cond invs S v s s')"

  and init_state_inv:
    "init_state params v \<le> SPEC (\<lambda>es. 
         let s = update_visited (\<lambda>_. {}) (update_discover (\<lambda>_. Map.empty) (update_finish (\<lambda>_. Map.empty) (update_counter (\<lambda>_. 0) es))) in
           \<not> is_brk params s \<and> pre_cond invs {} v s
         )"
begin
  
  sublocale dfs: DFS where params = timing_params and invs = timing_invs
    including timing_invs_def
    apply default
    apply (fastforce intro!: weaken_SPEC[OF enter_node_inv] pre_cond_imp_fe_inv)
    apply (force intro!: fe_inv_imp_pre_cond fe_inv_recurse)
    apply (force intro!: post_cond_imp_fe_inv fe_inv_step)
    apply (force intro!: fe_inv_imp_post_cond_brk fe_inv_brk)
    apply (fastforce intro!: weaken_SPEC[OF leave_node_inv] fe_inv_imp_post_cond_not_brk)
    apply (fastforce intro!: weaken_SPEC[OF known_node_inv] pre_visited_imp_post_cond)
    apply simp
      apply (refine_rcg refine_vcg order_trans[OF init_state_inv])
      apply (simp add: timing_invs_simps parenthesis_def)
    done

  (* Parenthesis Theorem *)
  lemma dfs_parenthesis:
    "dfs \<le> SPEC (\<lambda>(_,s). \<not> is_brk params s \<longrightarrow> ( 
       \<forall>v\<in>reachable. \<forall>w\<in>reachable. v \<noteq> w \<longrightarrow>
        \<delta> s v < \<delta> s w \<longrightarrow> (
                           \<delta> s v < \<phi> s v \<and> \<phi> s v < \<delta> s w \<and> \<delta> s w < \<phi> s w (* disjoint *)
                                          \<or> 
                           \<delta> s w < \<phi> s w \<and> \<phi> s w < \<phi> s v (* w contained in v *)
                          )))"
     (is "dfs \<le> SPEC ?P")
  proof -
    {
      fix v w s s0
      assume PC: "post_cond timing_invs {} v0 s0 s"
         and [simp]: "reachable = visited s" "\<not> is_brk params s"
      assume vw: "v \<in> reachable" "w \<in> reachable"
      and A: "v \<noteq> w" "\<delta> s v < \<delta> s w"
      
      note [simp] = PC[unfolded timing_invs_simps]

      from vw have 
        "\<delta> s v < \<phi> s v" "\<delta> s w < \<phi> s w"
        by simp_all
      moreover have "parenthesis s" by simp
      with vw A have "\<phi> s v < \<delta> s w \<or> \<phi> s w < \<phi> s v"
        by (auto simp add: parenthesis_def)
      ultimately have 
        "\<delta> s v < \<phi> s v \<and> \<phi> s v < \<delta> s w \<and> \<delta> s w < \<phi> s w (* disjoint *)
         \<or> 
         \<delta> s w < \<phi> s w \<and> \<phi> s w < \<phi> s v (* w contained in v *)"
        by metis
    }
    thus ?thesis 
      using dfs_correct_rule
      by fastforce
  qed
end

locale timing_stack_DFS_raw = timing_DFS_raw +
                              stack_DFS_raw +
   assumes [simp]: 
       "\<And>f. stack (update_finish f s) = stack s"
       "\<And>f. stack (update_discover f s) = stack s"
       "\<And>f. stack (update_counter f s) = stack s"
       "\<And>f. finish (update_stack f s) = finish s"
       "\<And>f. discover (update_stack f s) = discover s"
       "\<And>f. counter (update_stack f s) = counter s"

locale timing_stack_DFS_def = timing_stack_DFS_raw 
   where update_stack = update_stack
   for update_stack :: "('v list \<Rightarrow> _) \<Rightarrow> 's \<Rightarrow> 's" +
   fixes params :: "('v,'s) dfs_params"
   fixes invs   :: "('v,'s) dfs_invs"
begin

  (* We cannot simply inherit from both implementations at the same time.
     Instead we have to first apply one, then the other.
     The order is somewhat important, as only the properties of the last one
     can be lifted easily, the others have to be re-proven (mostly by copying). *)
  sublocale stack: stack_DFS_def .. (* apply stack first *)
  sublocale timing_DFS_def
    where params = stack_params and invs = stack_invs
    .. (* and timing thereafter *)

  abbreviation "timing_stack_params \<equiv> timing_params"
  
  abbreviation "timing_stack_pre \<equiv> \<lambda>S v s. timing_pre S v s \<and> stack_pre S v s"
  abbreviation "timing_stack_post \<equiv> \<lambda>S v s s'. timing_post S v s s' \<and> stack_post S v s s'"
  abbreviation "timing_stack_fe \<equiv> \<lambda>S v s it s'. timing_fe S v s it s' \<and> stack_fe S v s it s'"

  definition "timing_stack_invs = \<lparr>
     pre_cond = \<lambda>S v s. timing_stack_pre S v s \<and> pre_cond invs S v s,
     post_cond = \<lambda>S v s s'. timing_stack_post S v s s' \<and> post_cond invs S v s s',
     fe_inv = \<lambda>S v s it s'. timing_stack_fe S v s it s' \<and> fe_inv invs S v s it s'\<rparr>"

  lemma invs_eq [simp]:
    "timing_invs = timing_stack_invs"
    unfolding timing_stack_invs_def timing_invs_def
    unfolding stack_invs_def
    by simp
  
  bundle timing_stack_invs_def = timing_stack_invs_def[simp]
  
  schematic_lemma timing_stack_params_simps[simp]:
    "enter_node timing_stack_params v s = ?EN"
    "leave_node timing_stack_params v s = ?LN"
    "known_node timing_stack_params v s = ?KN"
    "init_state timing_stack_params v = ?IS"
    "is_brk timing_stack_params s = ?IB"
    unfolding timing_params_simps stack_params_simps[abs_def]
    by (rule refl)+

  schematic_lemma timing_stack_invs_simps':
    "pre_cond timing_stack_invs S v s = ?PC"
    "post_cond timing_stack_invs S v s s' = ?PoC"
    "fe_inv timing_stack_invs S v s it s' = ?FEI"
    unfolding timing_stack_invs_def timing_invs_simps stack_invs_simps dfs_invs.simps
    by (rule refl)+

  lemmas timing_stack_invs_simps = timing_stack_invs_simps' timing_pre_def stack_pre_def timing_post_def stack_post_def timing_fe_def stack_fe_def

  lemma rpost_lift[intro!]:
    "rpost S v s s' \<Longrightarrow> stack.rpost S v s s'"
    unfolding rpost_def stack.rpost_def
    by simp

  lemma rfe_inv_lift[intro!]:
    "rfe_inv S v s it s' \<Longrightarrow> stack.rfe_inv S v s it s'"
    unfolding rfe_inv_def stack.rfe_inv_def
    by simp
end

locale timing_stack_DFS = timing_stack_DFS_def +
   assumes enter_node_inv: 
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      s' = update_stack (Cons v) (update_counter Suc (update_discover (v \<longmapsto> counter s) (update_visited (insert v) s)))\<rbrakk> 
      \<Longrightarrow> enter_node params v s' \<le> SPEC (\<lambda>s''. 
            visited s'' = visited s' \<and> stack s'' = stack s'
          \<and> discover s'' = discover s' \<and> counter s'' = counter s' \<and> finish s'' = finish s'
          \<and> fe_inv invs S v s (succ v) s'')"

  and fe_inv_recurse:
    "\<lbrakk>rpre_full S v s; 
      v \<notin> visited s; 
      \<not>is_brk params s'; 
      rfe_inv_full S v s it s';
      v' \<in> it; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> pre_cond invs (insert v S) v' s'"

  and fe_inv_step:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; \<not>is_brk params s';
      it \<subseteq> succ v; v' \<in> it;
      rfe_inv_full S v s it s'; rpre_full (insert v S) v' s'; 
      rpost_full (insert v S) v' s' s''\<rbrakk>
      \<Longrightarrow> fe_inv invs S v s (it - {v'}) s''"

  and fe_inv_brk: 
     "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
       rfe_inv_full S v s it s'; is_brk params s'; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> post_cond invs S v s s'"

  and leave_node_inv:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      rfe_inv_full S v s {} s'; \<not> is_brk params s';
      s'' = update_stack tl (update_counter Suc (update_finish (v \<longmapsto> counter s') s'))\<rbrakk>
      \<Longrightarrow> leave_node params v s'' \<le> SPEC (\<lambda>x. 
            visited x = visited s'' \<and> stack x = stack s''
          \<and> discover x = discover s'' \<and> counter x = counter s'' \<and> finish x = finish s''
          \<and> post_cond invs S v s x)"

  and known_node_inv:
    "\<lbrakk>rpre_full S v s; v \<in> visited s\<rbrakk> 
      \<Longrightarrow> known_node params v s \<le> SPEC (\<lambda>s'. 
            visited s' = visited s \<and> stack s' = stack s
          \<and> discover s' = discover s \<and> counter s' = counter s \<and> finish s' = finish s
          \<and> post_cond invs S v s s')"

  and init_state_inv:
    "init_state params v \<le> SPEC (\<lambda>es. 
         let s = update_visited (\<lambda>_. {}) (update_discover (\<lambda>_. Map.empty) (update_finish (\<lambda>_. Map.empty) (update_counter (\<lambda>_. 0) (update_stack (\<lambda>_. []) es)))) in
           \<not> is_brk params s \<and> pre_cond invs {} v s
         )"
begin

  lemma (in stack_DFS_def) stack_invs_intros: 
    "post_cond invs S v s s' \<Longrightarrow> stack_post S v s s' \<Longrightarrow> post_cond stack_invs S v s s'"
    "pre_cond invs S v s \<Longrightarrow> stack_pre S v s \<Longrightarrow> pre_cond stack_invs S v s"
    "fe_inv invs S v s it s' \<Longrightarrow> stack_fe S v s it s' \<Longrightarrow> fe_inv stack_invs S v s it s'"
    by (simp_all add: stack_invs_def)

  bundle (in stack_DFS_def) stack_invs_intros = stack_invs_intros[intro!]

  sublocale dfs: timing_DFS where invs = stack_invs and params = stack_params
    including timing_stack_invs_def stack_invs_intros
    apply default
    apply (fastforce intro!: stack.pre_cond_imp_fe_inv weaken_SPEC[OF enter_node_inv])
    apply (fastforce intro!: stack.fe_inv_imp_pre_cond fe_inv_recurse)
    apply (fastforce intro!: stack.post_cond_imp_fe_inv fe_inv_step)
    apply (fastforce intro!: stack.fe_inv_imp_post_cond_brk fe_inv_brk)
    apply (fastforce intro!: stack.fe_inv_imp_post_cond_not_brk weaken_SPEC[OF leave_node_inv])
    apply (fastforce intro!: stack.pre_visited_imp_post_cond weaken_SPEC[OF known_node_inv])
    apply simp
      apply (refine_rcg refine_vcg order_trans[OF init_state_inv])
      apply (simp add: stack_invs_simps)
    done

  (* this cannot be lifted and has to be proven again *)
  lemma stack_dfs_correct:
    "dfs \<le> SPEC (\<lambda>(_,s). \<not> is_brk params s \<longrightarrow> (stack s = [] \<and> visited s = reachable))"
    by (rule dfs_correct_rule) (simp add: timing_stack_invs_simps, metis)

  (* this can nearly be taken as-is, only the 'is_brk stack_params' needs to be rewritten *)
  lemma dfs_parenthesis:
    "dfs \<le> SPEC (\<lambda>(_,s). \<not> is_brk params s \<longrightarrow> ( 
       \<forall>v\<in>reachable. \<forall>w\<in>reachable. v \<noteq> w \<longrightarrow>
        \<delta> s v < \<delta> s w \<longrightarrow> (
                           \<delta> s v < \<phi> s v \<and> \<phi> s v < \<delta> s w \<and> \<delta> s w < \<phi> s w (* disjoint *)
                                          \<or> 
                           \<delta> s w < \<phi> s w \<and> \<phi> s w < \<phi> s v (* w contained in v *)
                          )))"
    using dfs_parenthesis by simp
end

locale tree_DFS_raw = timing_stack_DFS_raw
  where update_stack = update_stack
  for update_stack :: "('v list \<Rightarrow> _) \<Rightarrow> 's \<Rightarrow> 's" +
  fixes tree :: "'s \<Rightarrow> 'v rel"
  and update_tree :: "('v rel \<Rightarrow> 'v rel) \<Rightarrow> 's \<Rightarrow> 's"
  assumes update_tree_correct[simp]:
    "\<And>f. tree (update_tree f s) = f (tree s)"
    "\<And>f. visited (update_tree f s) = visited s"
    "\<And>f. stack (update_tree f s) = stack s"
    "\<And>f. discover (update_tree f s) = discover s"
    "\<And>f. finish (update_tree f s) = finish s"
    "\<And>f. counter (update_tree f s) = counter s"
    "\<And>f. tree (update_visited f s) = tree s"
    "\<And>f. tree (update_stack f s) = tree s"
    "\<And>f. tree (update_discover f s) = tree s"
    "\<And>f. tree (update_finish f s) = tree s"
    "\<And>f. tree (update_counter f s) = tree s"

locale tree_DFS_def =  tree_DFS_raw
  where update_tree = update_tree
  for update_tree :: "('v rel \<Rightarrow> _) \<Rightarrow> 's \<Rightarrow> 's"+
  fixes params :: "('v,'s) dfs_params"
  fixes invs   :: "('v,'s) dfs_invs"
begin
  definition tree_params where
 "  tree_params = \<lparr>
    enter_node = \<lambda>v s.
      let s' = (if length (stack s) > 1 then update_tree (insert (hd (tl (stack s)), v)) s else s)
      in enter_node params v s',

    leave_node = leave_node params,
    known_node = known_node params,
    init_state = \<lambda>v. init_state params v \<guillemotright>= RETURN \<circ> update_tree (\<lambda>_. {}),
    is_brk = is_brk params
   \<rparr>"

  abbreviation (input) "tree_common_inv s \<equiv>
                       tree s \<subseteq> E
                     \<and> (tree s = {} \<longleftrightarrow> visited s \<subseteq> {v0})
                     \<and> Field (tree s) \<subseteq> visited s
                     \<and> (stack s \<noteq> [] \<longrightarrow> path (tree s) v0 (rev (tl (stack s))) (hd (stack s)))
                     \<and> (tree s \<noteq> {} \<longrightarrow> Field (tree s) = visited s
                                      \<and> (tree s)\<^sup>* `` {v0} = visited s)
                     \<and> (\<forall>(v,v') \<in> tree s. \<delta> s v < \<delta> s v' \<and> (v' \<in> dom (finish s) \<longrightarrow> v \<notin> dom (finish s) \<or> \<phi> s v' < \<phi> s v))
                     \<and> (\<forall>(v,v') \<in> tree s. \<forall>x. x\<noteq>v \<longrightarrow> (x,v') \<notin> tree s)
                     \<and> (\<forall>v \<in> dom (discover s). \<forall>w \<in> dom (finish s). \<delta> s v < \<delta> s w \<and> (v \<notin> dom (finish s) \<or> \<phi> s v > \<phi> s w) \<longrightarrow> (v,w) \<in> (tree s)\<^sup>+)"

  definition "tree_pre \<equiv> \<lambda>S v s. tree_common_inv s"
  definition "tree_post \<equiv> \<lambda>S v s s'. tree_common_inv s' \<and> tree s \<subseteq> tree s'"
  definition "tree_fe \<equiv> \<lambda>S v s it s'. 
                     tree_common_inv s' 
                  (* \<and> (\<forall>v' \<in> succ v - it. v' \<notin> visited s \<and> v' \<in> visited s' \<longrightarrow> (v,v') \<in> (tree s')\<^sup>+) *)
                   \<and> tree s \<subseteq> tree s'" 

  definition "tree_invs = \<lparr>
    pre_cond = \<lambda>S v s. tree_pre S v s \<and> pre_cond invs S v s,
    post_cond = \<lambda>S v s s'. tree_post S v s s' \<and> post_cond invs S v s s',
    fe_inv = \<lambda>S v s it s'. tree_fe S v s it s' \<and> fe_inv invs S v s it s' \<rparr>"
  
  schematic_lemma tree_invs_def':
    "pre_cond tree_invs S v s = ?PC"
    "post_cond tree_invs S v s s' = ?PoC"
    "fe_inv tree_invs S v s it s' = ?FEI"
    unfolding tree_invs_def  dfs_invs.simps
    by (rule refl)+

  bundle tree_invs_def = tree_invs_def'[simp]
  
  schematic_lemma tree_params_simps[simp]:
    "enter_node tree_params v s = ?EN"
    "leave_node tree_params v s = ?LN"
    "known_node tree_params v s = ?KN"
    "init_state tree_params v = ?IS"
    "is_brk tree_params s = ?IB"
    unfolding tree_params_def dfs_params.simps
    by (rule refl)+

  schematic_lemma tree_invs_simps':
    "pre_cond tree_invs S v s = ?PC"
    "post_cond tree_invs S v s s' = ?PoC"
    "fe_inv tree_invs S v s it s' = ?FEI"
    unfolding tree_invs_def tree_pre_def tree_post_def tree_fe_def dfs_invs.simps
    by (rule refl)+
  
  lemmas tree_invs_simps = tree_invs_simps' tree_pre_def tree_post_def tree_fe_def
  
  lemma tree_invs_pre_cond[intro]:
    "pre_cond tree_invs S v s \<Longrightarrow> pre_cond invs S v s"
    by (simp add: tree_invs_simps)

  lemma tree_invs_post_cond[intro]:
    "post_cond tree_invs S v s s' \<Longrightarrow> post_cond invs S v s s'"
    by (simp add: tree_invs_simps)

  lemma tree_invs_fe_inv[intro]:
    "fe_inv tree_invs S v s it s' \<Longrightarrow> fe_inv invs S v s it s'"
    by (simp add: tree_invs_simps)

  sublocale timing_stack_DFS_def 
    where params = tree_params and invs = tree_invs
    ..

  lemma pre_cond_imp_fe_inv_1:
    assumes "rpre S v s"
    and "timing_stack_pre S v s" "tree_pre S v s"
    and "stack s = []"
    and "visited x = insert v (visited s)" "stack x = v # stack s"
    and "finish x = finish s"
    and "tree x = tree s"
    shows "tree_fe S v s (succ v) x"
    using assms
    unfolding rpre_def timing_stack_invs_simps tree_invs_simps
    by simp

  lemma (in -) Field_not_elem:
    "v \<notin> Field R \<Longrightarrow> \<forall>(x,y) \<in> R. x \<noteq> v \<and> y \<noteq> v"
    unfolding Field_def by auto

  lemma (in -) trancl_Image_in_Range:
    "R^+ `` {v} \<subseteq> Range R"
    by (auto elim: trancl.induct)
  
  lemma (in -) rtrancl_Image_in_Field:
    "R^* `` {v} \<subseteq> Field R \<union> {v}"
  proof -
    from trancl_Image_in_Range have "R^+ `` {v} \<subseteq> Field R" unfolding Field_def by fast
    hence "R^+ `` {v} \<union> {v} \<subseteq> Field R \<union> {v}" by blast
    with trancl_image_by_rtrancl show ?thesis by metis
  qed

  lemma (in -) rtrancl_sub_insert_rtrancl:
    "R^* \<subseteq> (insert x R)^*"
    by (auto elim: rtrancl.induct rtrancl_into_rtrancl)

  lemma (in -) trancl_sub_insert_trancl:
    "R^+ \<subseteq> (insert x R)^+"
    by (auto elim: trancl.induct trancl_into_trancl)

  lemma pre_cond_imp_fe_inv_2:
    assumes PRE: "rpre S v s"
    and TSPRE: "timing_stack_pre S v s"
    and TPRE: "tree_pre S v s"
    and NE: "stack s \<noteq> []"
    and V: "v \<notin> visited s"
    and EQ: "visited x = insert v (visited s)" "stack x = v # stack s"
            "tree x = insert (hd (stack s), v) (tree s)"
            "discover x = (discover s)(v \<mapsto> counter s)" "counter x = Suc (counter s)"
            "finish x = finish s"
    shows "tree_fe S v s (succ v) x"
    unfolding tree_fe_def
  proof (intro conjI impI)
    from TSPRE have SPRE: "stack_pre S v s" by simp

    note pre_simps[simp] = tree_invs_simps timing_stack_invs_simps rpre_def
    note PRE' = PRE TSPRE TPRE

    have empty_tree: "tree s = {} \<Longrightarrow> visited s = {v0} \<and> stack s = [v0]"
    proof
      assume "tree s = {}"
      with PRE' NE show "visited s = {v0}" by auto
      with SPRE NE show "stack s = [v0]" by (cases "stack s") auto
    qed

    from V TPRE[unfolded pre_simps] have v_notin_tree: "\<forall>(x,x') \<in> tree s. x \<noteq> v \<and> x' \<noteq> v" by (blast intro!: Field_not_elem)
    from NE V SPRE have v_hd: "hd (stack s) \<noteq> v" using hd_in_set by auto

    from NE have RS: "rev (stack s) = (rev (tl (stack s))) @ [hd (stack s)]" by (cases "stack s") auto

    from EQ show "tree s \<subseteq> tree x" by auto

    from PRE' NE EQ show "tree x \<subseteq> E" by simp
    
    from TPRE NE have Ps: "path (tree s) v0 (rev (tl (stack s))) (hd (stack s))" by simp
    with EQ path_mono[OF _ this] have Px: "path (tree x) v0 (rev (tl (stack s))) (hd (stack s))" by blast
    with EQ NE show "path (tree x) v0 (rev (tl (stack x))) (hd (stack x))" by (fastforce simp: RS intro: path_append)

    show Field: "Field (tree x) = visited x"
    proof (cases "tree s = {}")
      case True with empty_tree EQ show ?thesis by auto
    next
      case False with TPRE have "visited s = Field (tree s)" by simp
      moreover from NE SPRE[unfolded pre_simps] have "hd (stack s) \<in> visited s" by (metis hd_in_set set_mp)
      ultimately show ?thesis using EQ by auto
    qed
    thus "Field (tree x) \<subseteq> visited x" by simp

    show "(tree x)\<^sup>* `` {v0} = visited x" (is "?TR = _")
    proof
      from PRE' NE EQ have "last (stack x) = v0" "stack x \<noteq> []" by simp_all
      with last_in_set have "v0 \<in> set (stack x)" by blast
      with SPRE EQ have "v0 \<in> visited x" by auto
      with Field rtrancl_Image_in_Field[where v=v0] show "?TR \<subseteq> visited x" by auto
    next
      show "visited x \<subseteq> ?TR"
      proof (cases "tree s = {}")
        case True with empty_tree EQ show ?thesis by auto
      next
        case False with TPRE have "visited s = (tree s)\<^sup>* `` {v0}" by simp
        moreover from Px have "(v0, hd (stack s)) \<in> (tree x)\<^sup>*" by (metis path_is_rtrancl)
        ultimately show ?thesis apply -
          apply rule
          apply (simp add: EQ)
          apply (erule disjE1)
            apply (erule rtrancl_into_rtrancl[where b="hd (stack s)"])
            apply simp
          using rtrancl_sub_insert_rtrancl apply blast
          done
      qed
    qed

    show "tree x = {} \<longleftrightarrow> visited x \<subseteq> {v0}"
    proof (cases "tree s = {}")
      case True with empty_tree PRE' EQ V show ?thesis by simp
    next
      case False with TPRE EQ show ?thesis by simp
    qed

    {
      fix a b
      assume T: "(a,b) \<in> tree x"

      have "b \<in> dom (finish x) \<longrightarrow> a \<notin> dom (finish x) \<or> \<phi> x b < \<phi> x a"
      proof (rule impI)
        assume F: "b \<in> dom (finish x)"

        from EQ TSPRE V have "v \<notin> dom (finish x)"  by simp
        with F have "b \<noteq> v" by blast
        with EQ v_notin_tree T have "(a,b) \<in> tree s" "a \<noteq> v" by auto
        with EQ TPRE F show "a \<notin> dom (finish x) \<or> \<phi> x b < \<phi> x a" by simp fast
      qed

      also have "\<delta> x a < \<delta> x b"
      proof (cases "b = v")
        case True with T EQ v_notin_tree have [simp]: "a = hd (stack s)" by auto
        with PRE'[unfolded pre_simps] have "a \<in> dom (discover s)" by (metis hd_in_set NE set_mp)
        with TSPRE EQ v_hd True show ?thesis by simp
      next
        case False with EQ v_notin_tree T have "(a,b) \<in> tree s" "a \<noteq> v" by auto
        with TPRE have "\<delta> s a < \<delta> s b" by auto
        with False `a\<noteq>v` EQ show ?thesis by simp
      qed
      also (conjI) note calculation
    } 
    thus "\<forall>(v, v')\<in>tree x. \<delta> x v < \<delta> x v' \<and> (v' \<in> dom (finish x) \<longrightarrow> v \<notin> dom (finish x) \<or> \<phi> x v' < \<phi> x v)"
      by blast
    
    {
      fix w w' y
      assume *: "(w,w') \<in> tree x"
      and "y \<noteq> w"

      from NE V SPRE have v_hd: "hd (stack s) \<noteq> v" using hd_in_set by auto
      from V TPRE[unfolded pre_simps] have v_notin_tree: "\<forall>(x,x') \<in> tree s. x \<noteq> v \<and> x' \<noteq> v" by (blast intro!: Field_not_elem)

      have "(y,w') \<notin> tree x"
      proof (cases "w = hd (stack s)")
        case True
        have "(y,v) \<notin> tree x"
        proof (rule notI)
          assume "(y,v) \<in> tree x" 
          with True `y\<noteq>w` EQ have "(y,v) \<in> tree s" by simp
          with v_notin_tree show False by auto
        qed
        with True * EQ `y\<noteq>w` v_hd show ?thesis 
          apply (cases "w = v")
          apply simp 
          using TPRE apply simp apply blast
          done
      next
        case False with v_notin_tree * EQ `y\<noteq>w` v_hd 
        show ?thesis
          apply (cases "w' = v")
          apply simp apply blast
          using TPRE apply simp apply blast
          done
      qed
    }
    thus "\<forall>(v,v') \<in> tree x. \<forall>y. y \<noteq> v \<longrightarrow> (y,v') \<notin> tree x" by blast
    
    show "\<forall>v\<in>dom (discover x). \<forall>w\<in>dom (finish x). \<delta> x v < \<delta> x w \<and> (v \<notin> dom (finish x) \<or> \<phi> x w < \<phi> x v) \<longrightarrow> (v, w) \<in> (tree x)\<^sup>+"
    proof (intro ballI impI)
      fix a b
      assume F:"a \<in> dom (discover x)" "b \<in> dom (finish x)"
      and D: "\<delta> x a < \<delta> x b \<and> (a \<notin> dom (finish x) \<or> \<phi> x a > \<phi> x b)"

      from F EQ have F': "b \<in> dom (finish s)" by simp_all
      with TSPRE V have "b \<noteq> v" by auto
      
      from EQ have sub: "(tree s)\<^sup>+ \<subseteq> (tree x)\<^sup>+" by (simp add: trancl_sub_insert_trancl)

      show "(a,b) \<in> (tree x)\<^sup>+"
      proof (cases "a = v")
        case True with D EQ `b\<noteq>v` have "counter s < \<delta> s b" by simp
        also from TSPRE EQ F `b\<noteq>v` have "\<delta> s b < counter s" by simp
        finally have False .
        thus ?thesis ..
      next
        case False with `b\<noteq>v` F TPRE D EQ have "(a,b) \<in> (tree s)\<^sup>+" by simp
        with sub show ?thesis by auto
      qed
    qed
    (*show "\<forall>v'\<in>succ v - succ v. v' \<notin> visited s \<and> v' \<in> visited x \<longrightarrow> (v, v') \<in> (tree x)\<^sup>+" by simp*)
  qed

  lemma fe_inv_imp_pre_cond:
    assumes "tree_fe S v s it s'"
    shows "tree_pre (insert v S) v' s'"
    using assms
    by (simp_all add: tree_invs_simps)

  lemma post_cond_imp_fe_inv:
    assumes "tree_fe S v s it s'"
    and "v' \<in> it" "it \<subseteq> succ v"
    and "\<not> is_brk params s'"
    and "tree_post (insert v S) v' s' s''"
    shows "tree_fe S v s (it - {v'}) s''"
    using assms
    unfolding tree_invs_simps
    by simp (metis subset_trans)

  lemma fe_inv_imp_post_cond_brk:
    assumes "tree_fe S v s it s'"
    and "is_brk params s'"
    shows "tree_post S v s s'"
    using assms
    by (simp add: tree_invs_simps)

  lemma (in -) path_last_is_edge:
    assumes "path E x p y"
    and "p \<noteq> []"
    shows "(last p, y) \<in> E"
    using assms
    by (induct p rule: rev_induct) (simp_all add: path_append_conv)

  lemma (in -) rev_nonempty_induct[consumes 1, case_names single snoc]:
    assumes ne: "xs \<noteq> []"
    and single: "\<And>x. P [x]"
    and snoc': "\<And>x xs. xs \<noteq> [] \<Longrightarrow> P xs \<Longrightarrow> P (xs@[x])"
    shows "P xs"
    using ne
    proof (induct xs rule: rev_induct)
      case (snoc x xs) thus ?case
      proof (cases xs)
        case Nil thus ?thesis by (simp add: single)
      next
        case Cons with snoc show ?thesis by (fastforce intro!: snoc')
      qed
    qed simp

  lemma (in -) path_member_reach_end:
    assumes P: "path E x p y"
    and v: "v \<in> set p"
    shows "(v,y) \<in> E^+"
  proof -
    from v have "p \<noteq> []" by auto
    thus ?thesis using assms
    proof (induction p arbitrary: y rule: rev_nonempty_induct)
      case (single u) with path_is_trancl have "(x,y) \<in> E\<^sup>+" by force
      with single show ?case by (simp add: path_cons_conv)
    next
      case (snoc u us)
      show ?case
      proof (cases "u=v")
        case True with snoc have "(v,y) \<in> E" by (simp add: path_append_conv)
        thus ?thesis by simp
      next
        case False with snoc have "v \<in> set us" by simp
        moreover from snoc have "path E x us u" "(u,y) \<in> E" by (simp_all add: path_append_conv)
        ultimately have "(v,u) \<in> E\<^sup>+" by (metis snoc.IH)
        also note `(u,y) \<in> E`
        finally show ?thesis .
      qed
    qed
  qed

  lemma (in -) rev_split:
    "l \<noteq> [] \<Longrightarrow> rev (tl l) @ [hd l] = rev l"
    by (induct l) simp_all

  lemma fe_inv_imp_post_cond_not_brk:
    assumes PRE: "rpre S v s"
    and TSPRE: "timing_stack_pre S v s"
    and TPRE: "tree_pre S v s"
    and "v \<notin> visited s"
    and FEI: "rfe_inv S v s {} s'"
    and TSFEI: "timing_stack_fe S v s {} s'"
    and TFEI: "tree_fe S v s {} s'"
    and NB: "\<not> is_brk params s'"
    and EQ[simp]: "visited x = visited s'" "stack x = tl (stack s')"
        "discover x = discover s'" "finish x = (finish s')(v \<mapsto> counter s')" "counter x = Suc (counter s')"
        "tree x = tree s'"
    shows "tree_post S v s x"
  proof -
    from NB TSFEI have s_eq: "stack s = stack x" by (simp add: stack_invs_simps)
    
    note simps [simp] = stack_invs_simps tree_invs_simps timing_invs_simps

    from TSFEI have NE: "stack s' \<noteq> []" by simp
    with TFEI have "path (tree s') v0 (rev (tl (stack s'))) (hd (stack s'))" by simp
    with NB TSFEI have P: "path (tree x) v0 (rev (stack x)) v" by simp
    hence in_t: "(stack x \<noteq> [] \<and> (hd (stack x), v) \<in> tree x) \<or> (stack x = [] \<and> v0 = v)"
      using path_last_is_edge
      by (cases "stack x = []") force+

    from TFEI have tree_rule: "\<And>a b z. (a,b) \<in> tree x \<Longrightarrow> (z,b) \<in> tree x \<Longrightarrow> z = a"
      by simp blast

    {
      fix a b
      assume "(a,b) \<in> tree x"
      hence "{a,b} \<subseteq> Field (tree x)" by (auto simp add: Field_def)
      also from TFEI TSFEI have "Field (tree x) \<subseteq> dom (discover x)" by simp
      finally have "a \<in> dom (discover x) \<and> b \<in> dom (discover x)" by blast
    } note tree_discover = this

    {
      fix a b
      assume A: "(a,b) \<in> tree x"
      with TFEI A have \<delta>: "\<delta> x a < \<delta> x b" (is ?C1) by auto
      moreover hence "a \<noteq> b" by auto
      moreover have "b \<in> dom (finish x) \<longrightarrow> a \<notin> dom (finish x) \<or> \<phi> x b < \<phi> x a" (is ?C2)
      proof (cases "b = v")
        case True
        show ?thesis
        proof (cases "stack x = []")
          case True with PRE TSPRE `stack s = stack x` have "dom (discover s) = {}" by (simp add: rpre_def)
          with `b=v` True tree_discover[OF A] TSFEI have "\<delta> x b \<le> \<delta> x a" by simp
          with \<delta> have False by simp
          thus ?thesis ..
        next
          case False 
          from EQ `b=v` have "b \<in> dom (finish x)" by simp
          moreover from False in_t A `b=v` have "a = hd (stack x)" by (metis tree_rule)
          with hd_in_set[OF False] s_eq have "a \<in> set (stack s)" by simp
          with TSFEI have "a \<notin> dom (finish s')" by auto
          with `b=v` `a\<noteq>b` have "a \<notin> dom (finish x)" by simp
          ultimately show ?thesis by simp
        qed
      next
        case False -- "b \<noteq> v"
        show ?thesis
        proof (cases "a = v")
          case False with `b \<noteq> v` A EQ show ?thesis
            using TFEI by simp blast
        next
          case True
          have "b \<notin> S"
          proof (rule notI)
            assume "b \<in> S" 
            with PRE TSPRE have "b \<in> visited s" "b \<in> dom (discover s)" by (simp_all add: rpre_def) blast+
            with TSPRE TSFEI have "\<delta> x b < counter s" by (simp add: map_le_def) metis
            moreover from TSFEI `a=v` have "\<delta> x a = counter s" by simp
            ultimately show False using \<delta> by simp
          qed
          hence "b \<in> dom (finish s')" using tree_discover[OF A] `b\<noteq>v` TSFEI NB by simp
          moreover with False TSFEI have "\<phi> x b < counter s'" by auto
          with True have "\<phi> x b < \<phi> x a" by simp
          ultimately show ?thesis by simp
        qed
      qed
      ultimately have "?C1 \<and> ?C2" by simp
    }
    hence "\<forall>(v, v')\<in>tree x. \<delta> x v < \<delta> x v' \<and> (v' \<in> dom (finish x) \<longrightarrow> v \<notin> dom (finish x) \<or> \<phi> x v' < \<phi> x v)"
      by blast

    moreover { 
      assume A:"stack x \<noteq> []"
      with in_t have "(hd (stack x), v) \<in> tree x" by simp
      moreover from P A have "path (tree x) v0 (rev (tl (stack x)) @ [hd (stack x)]) v"
        by (simp add: rev_split)
      ultimately have "path (tree x) v0 (rev (tl (stack x))) (hd (stack x))"
        by (simp add: path_append_conv)
    }

    moreover 
    have "\<forall>v \<in> dom (discover x). \<forall>w \<in> dom (finish x). \<delta> x v < \<delta> x w \<and> (v \<notin> dom (finish x) \<or> \<phi> x v > \<phi> x w) \<longrightarrow> (v,w) \<in> (tree x)\<^sup>+"
    proof (intro ballI impI)
      fix a b
      assume F: "a \<in> dom (discover x)" "b \<in> dom (finish x)"
      and paren: "\<delta> x a < \<delta> x b \<and> (a \<notin> dom (finish x) \<or> \<phi> x a > \<phi> x b)"
      hence "a \<noteq> b" by auto
      
      show "(a,b) \<in> (tree x)\<^sup>+"
      proof (cases "b = v")
        case True with `a\<noteq>b` F have \<phi>b: "\<phi> x b = counter s'" and F': "a \<in> dom (discover s')" by simp_all
        have "a \<notin> dom (finish x)"
        proof (rule notI)
          assume A: "a \<in> dom (finish x)"
          with True `a\<noteq>b` have "a \<in> dom (finish s')" "\<phi> x a = \<phi> s' a" by simp_all
          with TSFEI EQ have "\<phi> x a < counter s'" by simp
          with \<phi>b paren A show False by simp
        qed
        hence "a \<notin> dom (finish s')" by simp
        with NB TSFEI F' True `a\<noteq>b` have "a \<in> set (stack x)" by simp
        with path_member_reach_end[OF P] True show ?thesis by simp
      next
        case False
        show ?thesis
        proof (cases "a = v")
          case False with `b\<noteq>v` F paren show ?thesis 
            using TFEI by simp
        next
          case True with F `a\<noteq>b` paren TSFEI have "a \<in> dom (discover s')" "b \<in> dom (finish s')" "\<delta> s' a < \<delta> s' b" "a \<notin> dom (finish s')" by auto
          with TFEI have "(a,b) \<in> (tree s')\<^sup>+" by simp
          thus ?thesis by simp
        qed
      qed
    qed

    ultimately show ?thesis
      using TFEI unfolding tree_invs_simps
      by auto
  qed

  lemma pre_visited_imp_post_cond:
    assumes "tree_pre S v s"
    and "v \<in> visited s"
    and "visited x = visited s" "stack x = stack s"
        "discover x = discover s" "finish x = finish s" "counter x = counter s"
        "tree x = tree s"
    shows "tree_post S v s x"
    using assms by (simp add: tree_invs_simps)
end

locale tree_DFS = tree_DFS_def +
  assumes enter_node_inv: 
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      ps' = update_stack (Cons v) (update_counter Suc (update_discover (v \<longmapsto> counter s) (update_visited (insert v) s)));
      stack s = [] \<longrightarrow> s' = ps';
      stack s \<noteq> [] \<longrightarrow> s' = update_tree (insert (hd (stack s), v)) ps'
     \<rbrakk> 
      \<Longrightarrow> enter_node params v s' \<le> SPEC (\<lambda>s''. 
            visited s'' = visited s' \<and> stack s'' = stack s'
          \<and> discover s'' = discover s' \<and> counter s'' = counter s' \<and> finish s'' = finish s'
          \<and> tree s'' = tree s'
          \<and> fe_inv invs S v s (succ v) s'')"

  and fe_inv_recurse:
    "\<lbrakk> rpre_full S v s;
       v \<notin> visited s; 
       \<not>is_brk params s'; 
       rfe_inv_full S v s it s';
       v' \<in> it; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> pre_cond invs (insert v S) v' s'"

  and fe_inv_step:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; \<not>is_brk params s';
      v' \<in> it; it \<subseteq> succ v;
      rfe_inv_full S v s it s'; rpre_all (insert v S) v' s'; 
      rpost_full (insert v S) v' s' s''\<rbrakk>
      \<Longrightarrow> fe_inv invs S v s (it - {v'}) s''"

  and fe_inv_brk: 
     "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
       rfe_inv_full S v s it s'; is_brk params s'; it \<subseteq> succ v\<rbrakk>
      \<Longrightarrow> post_cond invs S v s s'"

  and leave_node_inv:
    "\<lbrakk>rpre_full S v s; v \<notin> visited s; 
      rfe_inv_full S v s {} s'; \<not>is_brk params s';
      s'' = update_stack tl (update_counter Suc (update_finish (v \<longmapsto> counter s') s'))\<rbrakk>
      \<Longrightarrow> leave_node params v s'' \<le> SPEC (\<lambda>x. 
            visited x = visited s'' \<and> stack x = stack s''
          \<and> discover x = discover s'' \<and> counter x = counter s'' \<and> finish x = finish s''
          \<and> tree x = tree s''
          \<and> post_cond invs S v s x)"

  and known_node_inv:
    "\<lbrakk>rpre_full S v s; v \<in> visited s\<rbrakk> 
      \<Longrightarrow> known_node params v s \<le> SPEC (\<lambda>s'. 
            visited s' = visited s \<and> stack s' = stack s
          \<and> discover s' = discover s \<and> counter s' = counter s \<and> finish s' = finish s
          \<and> tree s' = tree s
          \<and> post_cond invs S v s s')"

  and init_state_inv:
    "init_state params v \<le> SPEC (\<lambda>es. 
         let s = update_tree (\<lambda>_. {}) es in
         let s' = update_visited (\<lambda>_. {}) (update_discover (\<lambda>_. Map.empty) (update_finish (\<lambda>_. Map.empty) (update_counter (\<lambda>_. 0) (update_stack (\<lambda>_. []) s))))
         in  \<not> is_brk params s' \<and> pre_cond invs {} v s'
         )"
begin
  sublocale dfs: timing_stack_DFS where params = tree_params and invs = tree_invs
    including tree_invs_def timing_stack_invs_def
    apply default
    apply (rename_tac s s')
    apply (case_tac "stack s = []")
      apply (fastforce intro!: weaken_SPEC[OF enter_node_inv] pre_cond_imp_fe_inv_1)
      apply (fastforce intro!: weaken_SPEC[OF enter_node_inv] pre_cond_imp_fe_inv_2)
    apply (fastforce intro!: fe_inv_imp_pre_cond fe_inv_recurse)
    apply (force intro!: post_cond_imp_fe_inv fe_inv_step)
    apply (force intro!: fe_inv_imp_post_cond_brk fe_inv_brk)
    apply (fastforce intro!: weaken_SPEC[OF leave_node_inv] fe_inv_imp_post_cond_not_brk)
    apply (fastforce intro!: weaken_SPEC[OF known_node_inv] pre_visited_imp_post_cond) 
    apply simp
      apply (refine_rcg refine_vcg order_trans[OF init_state_inv])
      apply (simp add: tree_invs_simps)
    done

  (* It's a tree, Mario! *)
  definition "treeishP s \<equiv> \<forall>(v,w) \<in> tree s.
         (w,v) \<notin> tree s
       \<and> v \<in> (tree s)\<^sup>* `` {v0}
       \<and> v \<in> reachable \<and> w \<in> reachable
       \<and> \<not>(\<exists>z. (z,w) \<in> tree s \<and> z\<noteq>v)
       \<and> (v,v) \<notin> (tree s)\<^sup>+"


  lemma treeish_unequal:
    assumes tree: "treeishP s"
    and reach: "(v,w) \<in> (tree s)\<^sup>+"
    shows "v \<noteq> w"
  proof -
    from reach obtain x where "(v,x) \<in> tree s" by (metis converse_tranclE)
    with tree[unfolded treeishP_def] have "(v,v) \<notin> (tree s)\<^sup>+" by blast
    with reach show ?thesis by blast
  qed
    
  lemma treeish:
    "dfs \<le> SPEC (\<lambda>(_,s). treeishP s)"
  proof -
    {
      fix s s0
      assume RP: "rpost {} v0 s0 s"
      and TSP:   "timing_stack_post {} v0 s0 s"
      and TP:    "tree_post {} v0 s0 s"

      {
        fix v w
        assume A: "(v,w) \<in> tree s"
        hence NE: "tree s \<noteq> {}" by auto
      
        note [simp] = tree_invs_simps
        
        from A TP have "\<delta> s v < \<delta> s w" by auto
        hence "(w,v) \<notin> tree s" using TP by auto
      
        also from TP A have V: "v \<in> visited s" "w \<in> visited s" by (auto simp add: Field_def)
        with TP NE have "v \<in> (tree s)\<^sup>* `` {v0}" by simp

        also (conjI) from V RP have "v \<in> reachable \<and> w \<in> reachable" by (auto simp add: rpost_def)
        
        also (conjI) from TP have "\<And>a b z. (a,b) \<in> tree s \<Longrightarrow> (z,b) \<in> tree s \<Longrightarrow> z = a"
          by simp blast
        with A have "\<not>(\<exists>z. (z,w) \<in> tree s \<and> z\<noteq>v)" by blast

        also (conjI) {
          fix x y
          assume "(x,y) \<in> (tree s)\<^sup>+"
          hence "\<delta> s x < \<delta> s y"
          proof (induction)
            case (base a) with TP show ?case by auto
          next
            case (step a b) with TP have "\<delta> s a < \<delta> s b" by auto
            with step.IH show ?case by simp
          qed
        }
        hence "(v,v) \<notin> (tree s)\<^sup>+" by blast

        also (conjI) note calculation
      } hence "treeishP s" unfolding treeishP_def by blast
    } 
    thus ?thesis
      including timing_stack_invs_def tree_invs_def
      apply (rule dfs_correct_rule)
      by force+
  qed

  lemma extended_parenthesis:
    "dfs \<le> SPEC (\<lambda>(_,s). \<forall>x \<in> reachable. \<forall>y \<in> reachable. \<not> is_brk params s \<longrightarrow> 
            (x,y) \<in> (tree s)\<^sup>+ \<longleftrightarrow> \<delta> s x < \<delta> s y \<and> \<phi> s y < \<phi> s x)"
  proof -
    {
      fix s s0
      assume RP: "rpost {} v0 s0 s"
      and TSP:   "timing_stack_post {} v0 s0 s"
      and TP:    "tree_post {} v0 s0 s"
      and tree:  "treeishP s"
      and vis_eq: "\<not> is_brk params s \<longrightarrow> visited s = reachable"

      {
        fix x y
        assume nb: "\<not> is_brk params s"
          and reach:  "x \<in> reachable" "y \<in> reachable"

        from nb TSP vis_eq have dom_f_eq: "dom (finish s) = reachable" "dom (discover s) = reachable" by (simp_all add: timing_stack_invs_simps)
        with reach have F: "x \<in> dom (finish s)" "y \<in> dom (finish s)" by simp_all

        have "(x,y) \<in> (tree s)\<^sup>+ \<longleftrightarrow> \<delta> s x < \<delta> s y \<and> \<phi> s y < \<phi> s x" (is "?L \<longleftrightarrow> ?R")
        proof 
          assume "?L" thus ?R using F
          proof (induction)
            case (base y) with TP show ?case 
              unfolding tree_invs_simps 
              by fast
          next
            case (step y z) with tree dom_f_eq have "y \<in> dom (finish s)" by (auto simp add: treeishP_def)
            moreover with TP step have "\<delta> s y < \<delta> s z \<and> \<phi> s z < \<phi> s y"
              unfolding tree_invs_simps
              by fast
            ultimately show ?case
              using step by simp
          qed
        next
          assume "?R" with reach dom_f_eq  TP show ?L by (simp add: tree_invs_simps)
        qed
      }
      hence "\<forall>x \<in> reachable. \<forall>y \<in> reachable. \<not> is_brk params s \<longrightarrow> 
            (x,y) \<in> (tree s)\<^sup>+ \<longleftrightarrow> \<delta> s x < \<delta> s y \<and> \<phi> s y < \<phi> s x" by blast
    }
    thus ?thesis
      including timing_stack_invs_def tree_invs_def
      apply (rule dfs_correct_extend[OF treeish])
      by force+
  qed

  definition white_path where
    "white_path s x y \<equiv> x\<noteq>y \<longrightarrow> (\<exists>p. path E x p y \<and> \<delta> s x < \<delta> s y \<and> (\<forall> v \<in> set (tl p). \<delta> s x < \<delta> s v))"
  
  lemma (in -) path_tl_induct[consumes 2, case_names single step]:
    assumes P: "path E x p y"
    and NE: "x \<noteq> y"
    and S: "\<And>u. (x,u) \<in> E \<Longrightarrow> P x u"
    and ST: "\<And>u v. (x,u) \<in> E\<^sup>+ \<Longrightarrow> (u,v) \<in> E \<Longrightarrow> P x u \<Longrightarrow> P x v"
    shows "P x y \<and> (\<forall> v \<in> set (tl p). P x v)"
  proof -
    from P NE have "p \<noteq> []" by auto
    thus ?thesis using P
    proof (induction p arbitrary: y rule: rev_nonempty_induct)
      case (single u) hence "(x,y) \<in> E" by (simp add: path_cons_conv)
      with S show ?case by simp
    next
      case (snoc u us) hence "path E x us u" by (simp add: path_append_conv)
      with snoc path_is_trancl have "P x u" "(x,u) \<in> E\<^sup>+" "\<forall>v \<in> set (tl us). P x v" by simp_all
      moreover with snoc have "\<forall>v \<in> set (tl (us@[u])). P x v" by simp
      moreover from snoc have "(u,y) \<in> E" by (simp add: path_append_conv)
      ultimately show ?case by (auto intro: ST)
    qed
  qed

  lemma white_path:
    "dfs \<le> SPEC(\<lambda>(_,s). \<forall>x \<in> reachable. \<forall>y \<in> reachable. \<not> is_brk params s \<longrightarrow> 
            white_path s x y \<longleftrightarrow> (x,y) \<in> (tree s)\<^sup>*)"
  proof -
    {
      fix s s0
      assume RP: "rpost {} v0 s0 s"
      and TSP:   "timing_stack_post {} v0 s0 s"
      and TP:    "tree_post {} v0 s0 s"
      and vis_eq: "\<not> is_brk params s \<longrightarrow> visited s = reachable"
      and ext_paren: "\<forall>x \<in> reachable. \<forall>y \<in> reachable. \<not> is_brk params s \<longrightarrow> (x,y) \<in> (tree s)\<^sup>+ \<longleftrightarrow> \<delta> s x < \<delta> s y \<and> \<phi> s y < \<phi> s x"

      {
        note [simp] = timing_invs_simps

        fix x y
        assume nb: "\<not> is_brk params s"
          and reach:  "x \<in> reachable" "y \<in> reachable"
          and ne: "x \<noteq> y"

        from nb vis_eq TSP have dom_eq_r: "dom (finish s) = reachable" "dom (discover s) = reachable" by simp_all
        from TSP have paren: "parenthesis s" by simp

        {
          fix x y
          assume xreach:  "x \<in> reachable"
            and s: "y \<in> succ x"
            and \<delta>: "\<delta> s x < \<delta> s y"
          with dom_eq_r have "x \<in> dom (finish s)" by simp
          with TSP s have \<delta>\<phi>: "\<delta> s y < \<phi> s x" by simp
          
          from xreach s have "y \<in> reachable" by (metis reachable_trans)
          note reach = xreach this

          moreover from paren reach dom_eq_r \<delta> have "\<phi> s x < \<delta> s y \<or> \<phi> s x > \<delta> s y \<and> \<phi> s y < \<phi> s x"
            unfolding parenthesis_def
            by auto
          with \<delta>\<phi> have "\<phi> s y < \<phi> s x" by simp
          
          with nb reach ext_paren \<delta> have "(x,y) \<in> (tree s)\<^sup>+"
            by simp
        } note succ_tree = this

        have "white_path s x y \<longleftrightarrow> (x,y) \<in> (tree s)\<^sup>*"
        proof
          assume "(x,y) \<in> (tree s)\<^sup>*"
          with ne have T: "(x,y) \<in> (tree s)\<^sup>+" by (metis rtranclD)
          then obtain p where P: "path (tree s) x p y" by (metis trancl_is_path)
          with TP have "path E x p y" using path_mono[where E="tree s"] by (simp add: tree_invs_simps)
          moreover 
          from P have "\<delta> s x < \<delta> s y \<and> (\<forall> v \<in> set (tl p). \<delta> s x < \<delta> s v)"
            using ne
          proof (induct rule: path_tl_induct)
            case (single u) with TP show ?case by (auto simp add: tree_invs_simps)
          next
            case (step u v) note `\<delta> s x < \<delta> s u`
            also from step TP have "\<delta> s u < \<delta> s v" by (auto simp add: tree_invs_simps)
            finally show ?case .
          qed
          ultimately show "white_path s x y" 
            by (auto simp add: ne white_path_def)
        next
          assume "white_path s x y"
          with ne obtain p where P:"path E x p y" and
                                 white: "\<delta> s x < \<delta> s y \<and> (\<forall> v \<in> set (tl p). \<delta> s x < \<delta> s v)"
            unfolding white_path_def
            by blast
          hence "p \<noteq> []" by auto

          thus "(x,y) \<in> (tree s)\<^sup>*" using P white reach(2)
          proof (induction p arbitrary: y rule: rev_nonempty_induct)
            case (single u) hence "y \<in> succ x" by (simp add: path_cons_conv)
            with succ_tree[of x y] reach single show ?case by simp
          next
            case (snoc u us) hence "path E x us u" by (simp add: path_append_conv)
            moreover hence "(x,u) \<in> E\<^sup>*" by (simp add: path_is_rtrancl)
            with reach have ureach: "u \<in> reachable" unfolding reachable_def by auto
            moreover from snoc have "\<delta> s x < \<delta> s u" "(\<forall>v\<in>set (tl us). \<delta> s x < \<delta> s v)" by simp_all
            ultimately have x_u: "(x,u) \<in> (tree s)\<^sup>*" by (metis snoc.IH)
            
            from snoc have "y \<in> succ u" by (simp add: path_append_conv)
            hence yreach: "y \<in> reachable" by (metis reachable_trans ureach)
            
            from `y \<in> succ u` ureach dom_eq_r TSP have "\<delta> s y < \<phi> s u" by auto
            also from `\<delta> s x < \<delta> s u` have "x \<noteq> u" by auto
            with x_u have "(x,u) \<in> (tree s)\<^sup>+" by (metis rtrancl_eq_or_trancl)
            with nb ureach reach ext_paren have "\<phi> s u < \<phi> s x" by simp
            finally have "\<phi> s y < \<phi> s x" 
              using yreach
              using snoc dom_eq_r
              using paren nb reach
              unfolding parenthesis_def
              by force
            with ext_paren yreach reach nb snoc have "(x,y) \<in> (tree s)\<^sup>+" by simp
            thus ?case by auto
          qed
        qed
      }
      moreover have "\<And>x. white_path s x x \<longleftrightarrow> (x,x) \<in> (tree s)\<^sup>*" by (simp add: white_path_def)

      ultimately have " \<forall>x \<in> reachable. \<forall>y \<in> reachable. \<not> is_brk params s \<longrightarrow> 
            white_path s x y \<longleftrightarrow> (x,y) \<in> (tree s)\<^sup>*" 
        by blast
    }
    thus ?thesis
      including timing_stack_invs_def tree_invs_def
      apply (rule dfs_correct_extend[OF extended_parenthesis])
      by force+
  qed
end
end
