theory Nested_DFS_Neumann
imports 
  "Lasso_Util"
  "../../Neumann_DFS/SE_DFS_Impl"
  "~~/src/HOL/Library/Code_Target_Nat"
begin

interpretation hpy!: HPY_SE_def "\<lambda>x. bag1_E G `` {x}" "bag1_v0 G" "bag1_F G"
  for G :: "'a option ba_graph1" .

fun extract_lasso where
  "extract_lasso v\<^sub>0 NO_LASSO = None"
| "extract_lasso v\<^sub>0 (LASSO x ys cyc) = Some (make_lasso v\<^sub>0 (butlast ys) (butlast cyc))"

definition "nested_dfs G = do { l \<leftarrow> hpy.nested_dfs G (bag1_v0 G); RETURN (extract_lasso (bag1_v0 G) l) }"

lemma lift_reachable:
  "hpy.pd_reachable G (bag1_v0 G) y \<longleftrightarrow> (bag1_v0 G,y)\<in>(bag1_E G)\<^sup>*"
using hpy.reachable_succs_rtrancl
by auto

lemma lift_reachable1:
  assumes "(bag1_v0 G, y) \<in> (bag1_E G)\<^sup>*"
  shows "hpy.reachable1 G y y \<longleftrightarrow> (y,y)\<in>(bag1_E G)\<^sup>+"
proof -
  from assms lift_reachable hpy.v\<^sub>0_reach_eq_vert  have "y \<in> hpy.V G" by metis
  with hpy.reachable1_succs_trancl show ?thesis by auto
qed

lemma lift_has_lasso1:
  "has_lasso1 G \<longleftrightarrow> (\<exists>y. y \<in> bag1_F G \<and> hpy.pd_reachable G (bag1_v0 G) y \<and> hpy.reachable1 G y y)"
unfolding has_lasso1_alt
by (metis lift_reachable1 lift_reachable)

lemma vwalk_to_path:
  assumes H:"hd xs = x"
  and L:"last xs = y"
  and VW: "vwalk xs (hpy.sd_to_graph G)"
  shows "Ba_Graph.path (bag1_E G) x (butlast xs) y"
using VW H L
proof (induct arbitrary: x rule: vwalk_induct)
  case Base thus ?case by simp
next
  case (Cons u v es) hence "Ba_Graph.path (bag1_E G) v (butlast (v#es)) y" by simp
  with Cons have "Ba_Graph.path (bag1_E G) x (x#butlast (v#es)) y"
    by (auto intro!: path_prepend[of x v]) (simp_all add: hpy.E_def)
  with Cons show ?case by simp
qed

lemma lift_is_lasso:
  "hpy.is_lasso G (bag1_v0 G) y to_y cyc \<Longrightarrow> is_lasso1 G (make_lasso (bag1_v0 G) (butlast to_y) (butlast cyc))"
  unfolding hpy.is_lasso_def
  apply (rule make_lasso1I[where va = y])
    apply (simp_all add: vwalk_to_path)
    apply (cases cyc)
      apply simp_all
      done

lemma finite_V:
  assumes G: "ba_graph1 G"
  shows "finite (hpy.V G)"
proof -
  interpret ba_graph1 G by (fact G)
  show ?thesis by (auto simp add: hpy.V_alt2_def)
qed

theorem nested_dfs_correct:
  assumes G: "ba_graph1 G"
  shows "nested_dfs G \<le> SPEC (\<lambda>r. 
    case r of None \<Rightarrow> \<not>has_lasso1 G
  | Some L \<Rightarrow> is_lasso1 G L)"
proof -
  interpret hpy!: HPY_SE "\<lambda>x. bag1_E G `` {x}" "bag1_v0 G" "bag1_F G"
    by unfold_locales (simp add: finite_V[OF G])

  have v0: "bag1_v0 G \<in> hpy.V G" by simp
  note hpy_sound = inres_SPEC[OF _ hpy.nested_dfs_sound[OF v0]]
  note hpy_complete = inres_SPEC[OF _ hpy.nested_dfs_complete[OF v0]]
  note hpy_NF = hpy.nested_dfs_NF[OF v0]

  show ?thesis
    unfolding nested_dfs_def
    apply (intro refine_vcg pw_leI)
    apply (auto simp add: hpy_NF)
    apply (case_tac x)
      apply (fastforce simp add: lift_has_lasso1 dest: hpy_complete)
      apply (fastforce simp add: lift_is_lasso dest: hpy_sound)
    done
qed

abbreviation "G1succs \<equiv> \<lambda>(_,_,succs) _ x. (succs x, ())"
abbreviation "G1v0 \<equiv> \<lambda>(v\<^sub>0,_,_). v\<^sub>0"
abbreviation "G1F \<equiv> \<lambda>(_,F,_). F"

interpretation hpy!: HPY_SE_Impl_def "G1succs Gi" "G1v0 Gi" "()" "\<lambda>_. True" "G1F Gi" hs_ops hs_ops hs_ops
  for Gi :: "('a::hashable) option ba_graph1_impl" .

lemma lsi_\<alpha>_set:
  "lsi.invar S \<Longrightarrow> lsi.\<alpha> S = set S"
  unfolding lsi_ops_def
  unfolding StdBasicSetDefs.dflt_ops_def
  unfolding lsi_basic_ops_def
  by simp

lemma \<alpha>succs_alt:
  assumes REL: "(Gi,G)\<in>ba_graph1_rel"
  shows "hpy.\<alpha>succs Gi = (\<lambda>x. bag1_E G `` {x})"
proof -
  obtain I F E where [simp]: "Gi = (I,F,E)" by (cases Gi)
  with REL[unfolded ba_graph1_rel_as_br_def] have INV: "\<And>x. lsi_invar (E x)" and [simp]: "bag1_E G = {(u,v). v \<in> lsi_\<alpha> (E u)}" 
    unfolding br_def ba_graph1_\<alpha>_def ba_graph1_invar_def slg_invar_def slg_\<alpha>_def
    by simp_all
  thus ?thesis  unfolding hpy.\<alpha>succs_def[abs_def] by (simp add: lsi_\<alpha>_set)
qed

definition "nested_dfs_code Gi = (let l = hpy.nested_dfs_code Gi (G1v0 Gi) in extract_lasso (G1v0 Gi) l)"

lemma nested_dfs_code_refine:
  assumes G: "ba_graph1 G"
  assumes REL: "(Gi,G)\<in>ba_graph1_rel"
  shows "RETURN (nested_dfs_code Gi) \<le> nested_dfs G"
proof -
  from REL have DEFS [simp]: "G1v0 Gi = bag1_v0 G" "Collect (G1F Gi) = bag1_F G"
            and INV: "\<And>v. lsi.invar (fst (G1succs Gi () v))"
    unfolding ba_graph1_rel_as_br_def br_def ba_graph1_\<alpha>_def ba_graph1_invar_def slg_invar_def
    by auto

  note [simp] = \<alpha>succs_alt[OF REL]

  interpret hpy!: HPY_SE_Impl hs_ops hs_ops hs_ops "G1succs Gi" "G1v0 Gi" "()" "\<lambda>_. True" "G1F Gi"
    by unfold_locales (simp_all add: finite_V[OF G] INV)

  have "G1v0 Gi \<in> hpy.V Gi" by simp
  note ref = hpy.nested_dfs_code_refine[OF this, simplified]

  show ?thesis
    unfolding nested_dfs_code_def nested_dfs_def
    by (refine_transfer) (auto simp add: ref)
qed

theorem nested_dfs_code_correct:
  assumes G: "ba_graph1 G"
  assumes REL: "(Gi,G)\<in>ba_graph1_rel"
  shows "RETURN (nested_dfs_code Gi) \<le> SPEC (\<lambda>r. 
    case r of None \<Rightarrow> \<not>has_lasso1 G
  | Some L \<Rightarrow> is_lasso1 G L)"
proof -
  note nested_dfs_code_refine[OF assms]
  also note nested_dfs_correct[OF G]
  finally show ?thesis .
qed

lemma bag1_find_lasso_neumann_impl:
  "(nested_dfs_code,bag1_find_lasso_spec)
  \<in>ba_graph1_rel \<rightarrow> \<langle>\<langle>\<langle>Id\<rangle>lasso_rel\<rangle>option_rel\<rangle>plain_nres_rel"
  apply (intro fun_relI plain_nres_relI)
  unfolding bag1_find_lasso_spec_def
  apply (refine_rcg refine_vcg)
  apply (rule order_trans[OF nested_dfs_code_correct], assumption+)
  apply auto
  done

definition "bag_find_lasso_neumann_ndfs_impl 
  \<equiv> mk_one_initial_find_lasso_impl nested_dfs_code"
  
lemma bag_find_lasso_neumann_ndfs_impl_refine_aux:
  "(bag_find_lasso_neumann_ndfs_impl, bag_find_lasso_spec) 
  \<in> ba_graph_rel \<rightarrow> \<langle>\<langle>\<langle>Id\<rangle>lasso_rel\<rangle>option_rel\<rangle>plain_nres_rel"
  unfolding bag_find_lasso_neumann_ndfs_impl_def
  apply (rule mk_one_initial_find_lasso_impl_correct)
  apply (rule bag1_find_lasso_neumann_impl)
  done

lemma bag_find_lasso_neumann_ndfs_impl_refine:
  "(bag_find_lasso_neumann_ndfs_impl, bag_find_ce_spec) 
  \<in> ba_graph_rel \<rightarrow> \<langle>\<langle>lasso_run_rel\<rangle>option_rel\<rangle>plain_nres_rel"
proof -
  note bag_find_lasso_neumann_ndfs_impl_refine_aux[param_fo, THEN plain_nres_relD]
  also note bag_find_lasso_spec_refine[param_fo, THEN nres_relD]
  finally show ?thesis
    by (fastforce intro!: plain_nres_relI)
qed

(* Code mappings *)

definition "hpy_restr_succs G = hpy.restr_\<gamma>succs G"
lemmas[code] = hpy.restr_\<gamma>succs_def[folded hpy_restr_succs_def]

interpretation red!: RedDFS_Impl_def R hs_ops hs_ops "hpy_restr_succs G R" "()" "G1v0 G" "\<lambda>_. True" ss
  for G R ss x .

lemma iss_unit[code_unfold]:
  "iss (s :: ('a,'b,'c, unit) impl_sws) = ()" by simp

definition "red_visit_impl ss = red.red_visit_impl ss"
lemmas[code] = red.red_visit_impl_def[folded red_visit_impl_def, unfolded red.mk_ret_impl_def]

definition "red_remove_impl ss = red.red_remove_impl ss"
lemmas[code] = red.red_remove_impl_def[folded red_remove_impl_def, unfolded red.mk_ret_impl_def]

definition "red_dfs_code G R ss = red.red_dfs_code R G ss"
prepare_code_thms red.dfs_code_def[folded red_dfs_code_def, unfolded red.red_dfs_impl_simps red.red_post_impl_def red.succ_step_upd_impl_def, folded red_visit_impl_def red_remove_impl_def]

definition "mk_cycle = hpy.mk_cycle"
lemmas[code] = hpy.mk_cycle.simps[folded mk_cycle_def]

definition "hpy_visit = hpy.blue_visit_impl"
lemmas[code] = hpy.blue_visit_impl_def[folded hpy_visit_def]

definition "hpy_run_red G = hpy.run_red_dfs_impl G"
lemmas[code] = hpy.run_red_dfs_impl_def[folded hpy_run_red_def hpy_restr_succs_def, unfolded iss_unit, folded red_dfs_code_def]

definition "hpy_post G = hpy.blue_post_impl G"
lemmas[code] = hpy.blue_post_impl_def[folded hpy_post_def mk_cycle_def hpy_run_red_def, unfolded hpy.hpyM_impl_def, simplified]

definition "hpy_remove G = hpy.se_remove_impl G"
lemmas[code] = hpy.se_remove_impl_def[folded hpy_remove_def mk_cycle_def, unfolded iss_unit]

definition "hpy_code G = hpy.hpyse_code G"
prepare_code_thms hpy.dfs_code_def[folded hpy_code_def, unfolded hpy.se_dfs_impl_simps hpy.succ_step_upd_impl_def, folded hpy_remove_def hpy_post_def hpy_visit_def]

lemmas[code] = nested_dfs_code_def[unfolded hpy.nested_dfs_code_def, folded hpy_code_def]

export_code nested_dfs_code checking SML

end