theory Ample_Analysis
imports
  "Ample_Abstract"
begin

  locale system_ample =
    system_sticky en ex init int sticky +
    system_interpreted_traces en ex int ind
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
    and int :: "'state \<Rightarrow> 'interpretation"
    and sticky :: "'action set"
    and ind :: "'action \<Rightarrow> 'action \<Rightarrow> bool"
  begin

    sublocale ample_base "en" "ex" "int" "ind" "scut\<inverse>\<inverse>" by default

    lemma restrict_ample_set:
      assumes "s \<in> reachable"
      assumes "A \<inter> en s \<noteq> {}" "A \<inter> en s \<inter> sticky = {}"
      assumes "Ind (A \<inter> en s) (executable - A)"
      assumes "\<And> w. w \<in> words_fin s \<Longrightarrow> A \<inter> en s \<inter> set w = {} \<Longrightarrow> A \<inter> set w = {}"
      shows "ample_set s (A \<inter> en s)"
    unfolding ample_set_def
    proof (intro conjI allI impI)
      show "A \<inter> en s \<subseteq> en s" by simp
    next
      show "A \<inter> en s \<noteq> {}" using assms(2) by this
    next
      fix a
      assume 1: "a \<in> A \<inter> en s"
      show "scut\<inverse>\<inverse> (ex a s) s"
      proof (rule no_cut_scut)
        show "s \<in> reachable" using assms(1) by this
        show "a \<in> en s" using 1 by simp
        show "a \<notin> sticky" using assms(3) 1 by auto
      qed
    next
      have 1: "A \<inter> en s \<subseteq> executable" using executable assms(1) by blast
      show "A \<inter> en s \<subseteq> invisible" using executable_visible_sticky assms(3) 1 by blast
    next
      fix w
      assume 1: "w \<in> words_fin s" "A \<inter> en s \<inter> set w = {}"
      have 2: "A \<inter> set w = {}" using assms(5) 1 by this
      have 3: "set w \<subseteq> executable" using assms(1) 1(1) by rule
      show "Ind (A \<inter> en s) (set w)" using assms(4) 2 3 by blast
    qed

  end

  locale system_concurrent =
    system_initial en ex init
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
    +
    fixes procs :: "'state \<Rightarrow> 'process set"
    fixes pac :: "'process \<Rightarrow> 'action set"
    fixes psen :: "'process \<Rightarrow> 'state \<Rightarrow> 'action set"
    assumes procs_finite: "s \<in> reachable \<Longrightarrow> finite (procs s)"
    assumes psen_en: "s \<in> reachable \<Longrightarrow> pac p \<inter> en s \<subseteq> psen p s"
    assumes psen_ex: "s \<in> reachable \<Longrightarrow> a \<in> en s - pac p \<Longrightarrow> psen p (ex a s) = psen p s"
  begin

    lemma psen_fin_word:
      assumes "s \<in> reachable" "w \<in> words_fin s" "pac p \<inter> set w = {}"
      shows "psen p (fold ex w s) = psen p s"
    using assms(2, 1, 3)
    proof induct
      case (empty s)
      show ?case by simp
    next
      case (step a s w)
      have 1: "ex a s \<in> reachable" using step(4, 1) by rule
      have "psen p (fold ex (a # w) s) = psen p (fold ex w (ex a s))" by simp
      also have "\<dots> = psen p (ex a s)" using step 1 by simp
      also have "\<dots> = psen p s" using psen_ex step by simp
      finally show ?case by this
    qed

    lemma en_fin_word:
      assumes "\<And> r a b. r \<in> reachable \<Longrightarrow> a \<in> psen p s - en s \<Longrightarrow> b \<in> en r - pac p \<Longrightarrow>
        a \<in> en (ex b r) \<Longrightarrow> a \<in> en r"
      assumes "s \<in> reachable" "w \<in> words_fin s" "pac p \<inter> set w = {}"
      shows "pac p \<inter> en (fold ex w s) \<subseteq> pac p \<inter> en s"
    using assms
    proof (induct w rule: rev_induct)
      case Nil
      show ?case by simp
    next
      case (snoc b w)
      show ?case
      proof (safe, rule ccontr)
        fix a
        assume 2: "a \<in> pac p" "a \<in> en (fold ex (w @ [b]) s)" "a \<notin> en s"
        have 3: "a \<in> psen p s"
        proof -
          have 3: "psen p (fold ex (w @ [b]) s) = psen p s" using psen_fin_word snoc(3, 4, 5) by this
          have 4: "fold ex (w @ [b]) s \<in> reachable" using snoc(3, 4) by rule
          have 5: "a \<in> psen p (fold ex (w @ [b]) s)" using psen_en 4 2(1, 2) by auto
          show ?thesis using 2(1) 3 5 by auto
        qed
        have 4: "a \<in> en (fold ex w s)"
        proof (rule snoc(2))
          show "fold ex w s \<in> reachable" using snoc(3, 4) by auto
          show "a \<in> psen p s - en s" using 2(3) 3 by simp
          show "b \<in> en (fold ex w s) - pac p" using snoc(4, 5) by auto
          show "a \<in> en (ex b (fold ex w s))" using 2(2) by simp
        qed
        have 5: "pac p \<inter> en (fold ex w s) \<subseteq> pac p \<inter> en s"
        proof (rule snoc(1))
          show "\<And> r a b. r \<in> reachable \<Longrightarrow> a \<in> psen p s - en s \<Longrightarrow> b \<in> en r - pac p \<Longrightarrow>
            a \<in> en (ex b r) \<Longrightarrow> a \<in> en r" using snoc(2) by this
          show "s \<in> reachable" using snoc(3) by this
          show "w \<in> words_fin s" using snoc(4) by auto
          show "pac p \<inter> set w = {}" using snoc(5) by auto
        qed
        have 6: "a \<in> en s" using 2(1) 4 5 by auto
        show "False" using 2(3) 6 by simp
      qed
    qed

    lemma pac_en_blocked:
      assumes "\<And> r a b. r \<in> reachable \<Longrightarrow> a \<in> psen p s - en s \<Longrightarrow> b \<in> en r - pac p \<Longrightarrow>
        a \<in> en (ex b r) \<Longrightarrow> a \<in> en r"
      assumes "s \<in> reachable" "w \<in> words_fin s" "pac p \<inter> en s \<inter> set w = {}"
      shows "pac p \<inter> set w = {}"
      using words_fin_blocked en_fin_word assms by metis

    abbreviation "proc a \<equiv> {p. a \<in> pac p}"
    abbreviation "Proc A \<equiv> \<Union> a \<in> A. proc a"

    lemma psen_simple:
      assumes "Proc (psen p s) = {p}"
      assumes "\<And> r a b. r \<in> reachable \<Longrightarrow> a \<in> psen p s - en s \<Longrightarrow> b \<in> en r \<Longrightarrow>
        proc a \<inter> proc b = {} \<Longrightarrow> a \<in> en (ex b r) \<Longrightarrow> a \<in> en r"
      shows "\<And> r a b. r \<in> reachable \<Longrightarrow> a \<in> psen p s - en s \<Longrightarrow> b \<in> en r - pac p \<Longrightarrow>
        a \<in> en (ex b r) \<Longrightarrow> a \<in> en r"
      using assms by fast

  end

end
