theory Word_Extensions
imports
  "../../Modelchecker/CAVA_Automata/Words"
  "List_Extensions"
begin

  declare conc_fst[simp]
  declare conc_snd[simp]
  declare conc_conc[simp]
  declare prefix_suffix[symmetric, simp]

  abbreviation "prefix k w \<equiv> map w [0 ..< k]"
  abbreviation "scan f w a \<equiv> \<lambda> i. fold f (prefix i w) a"

  lemma set_prefix[intro, simp]: "set (prefix i w) \<subseteq> range w" by auto
  lemma range_suffix[intro, simp]: "range (suffix i w) \<subseteq> range w" unfolding suffix_def by blast

  primrec build :: "'a \<Rightarrow> 'a word \<Rightarrow> 'a word" (infixr "##" 65)
    where "(a ## w) 0 = a" | "(a ## w) (Suc i) = w i"

  lemma build_eq[iff]: "a\<^sub>1 ## w\<^sub>1 = a\<^sub>2 ## w\<^sub>2 \<longleftrightarrow> a\<^sub>1 = a\<^sub>2 \<and> w\<^sub>1 = w\<^sub>2"
  proof
    assume 1: "a\<^sub>1 ## w\<^sub>1 = a\<^sub>2 ## w\<^sub>2"
    have 2: "\<And> i. (a\<^sub>1 ## w\<^sub>1) i = (a\<^sub>2 ## w\<^sub>2) i" using 1 by auto
    show "a\<^sub>1 = a\<^sub>2 \<and> w\<^sub>1 = w\<^sub>2"
    proof (intro conjI ext)
      show "a\<^sub>1 = a\<^sub>2" using 2[of "0"] by simp
    next
      fix i
      show "w\<^sub>1 i = w\<^sub>2 i" using 2[of "Suc i"] by simp
    qed
  next
    assume 1: "a\<^sub>1 = a\<^sub>2 \<and> w\<^sub>1 = w\<^sub>2"
    show "a\<^sub>1 ## w\<^sub>1 = a\<^sub>2 ## w\<^sub>2" using 1 by simp
  qed
  lemma build_cons[simp]: "(a # u) \<frown> v = a ## u \<frown> v"
  proof
    fix i
    show "((a # u) \<frown> v) i = (a ## u \<frown> v) i"
    proof (cases i)
      case 0
      show ?thesis unfolding 0 by simp
    next
      case (Suc j)
      show ?thesis unfolding Suc by (cases "j < length u", simp+)
    qed
  qed
  lemma build_append[simp]: "(w @ a # u) \<frown> v = w \<frown> a ## u \<frown> v"
    unfolding conc_conc[symmetric] by simp
  lemma build_first[simp]: "w 0 ## suffix (Suc 0) w = w"
  proof
    fix i
    show "(w 0 ## suffix (Suc 0) w) i = w i" by (cases i, simp+)
  qed
  lemma build_split[intro]: "w = w 0 ## suffix 1 w" by simp
  lemma build_range[simp]: "range (a ## w) = insert a (range w)"
  proof safe
    fix i
    show "(a ## w) i \<notin> range w \<Longrightarrow> (a ## w) i = a" by (cases i, auto)
  next
    show "a \<in> range (a ## w)"
    proof (rule range_eqI)
      show "a = (a ## w) 0" by simp
    qed
  next
    fix i
    show "w i \<in> range (a ## w)"
    proof (rule range_eqI)
      show "w i = (a ## w) (Suc i)" by simp
    qed
  qed

  lemma range_conc[simp]: "range (w\<^sub>1 \<frown> w\<^sub>2) = set w\<^sub>1 \<union> range w\<^sub>2"
  proof (intro equalityI subsetI)
    case (goal1 a)
    obtain i where 1: "a = (w\<^sub>1 \<frown> w\<^sub>2) i" using goal1 by auto
    show ?case unfolding 1 by (cases "i < length w\<^sub>1", simp+)
  next
    case (goal2 a)
    show ?case
    using goal2
    proof
      case (goal1)
      obtain i where 1: "i < length w\<^sub>1" "a = w\<^sub>1 ! i" using goal1 in_set_conv_nth by metis
      show ?thesis
      proof
        show "a = (w\<^sub>1 \<frown> w\<^sub>2) i" using 1 by auto
        show "i \<in> UNIV" by rule
      qed
    next
      case (goal2)
      obtain i where 1: "a = w\<^sub>2 i" using goal2 by auto
      show ?thesis
      proof
        show "a = (w\<^sub>1 \<frown> w\<^sub>2) (length w\<^sub>1 + i)" using 1 by simp
        show "length w\<^sub>1 + i \<in> UNIV" by rule
      qed
    qed
  qed

  lemma prefix_conc_fst[simp]:
    assumes "n \<le> length u"
    shows "prefix n (u \<frown> v) = take n u"
  proof (intro nth_equalityI allI impI)
    case (goal1)
    show ?case using assms by simp
  next
    case (goal2 i)
    show ?case using assms goal2 by simp
  qed
  lemma prefix_conc_snd[simp]:
    assumes "n \<ge> length u"
    shows "prefix n (u \<frown> v) = u @ prefix (n - length u) v"
  proof (intro nth_equalityI allI impI)
    case (goal1)
    show ?case using assms by simp
  next
    case (goal2 i)
    show ?case using goal2 by (cases "i < length u", auto)
  qed

  lemma suffix_conc_fst[simp]:
    assumes "n \<le> length u"
    shows "suffix n (u \<frown> v) = drop n u \<frown> v"
  proof
    case (goal1 i)
    show ?case using assms by (cases "n + i < length u", auto simp: algebra_simps)
  qed
  lemma suffix_conc_snd[simp]:
    assumes "n \<ge> length u"
    shows "suffix n (u \<frown> v) = suffix (n - length u) v"
  proof
    case (goal1 i)
    show ?case using assms by simp
  qed

  lemma suffix_comp: "suffix i v = v \<circ> op + i" unfolding suffix_def by force
  lemma map_suffix[simp]: "map (suffix i v) [j ..< j + n] = map v [i + j ..< i + j + n]"
  proof -
    have "map (suffix i v) [j ..< j + n] = map (v \<circ> op + i) [j ..< j + n]"
      unfolding suffix_comp by simp
    also have "\<dots> = (map v \<circ> map (op + i)) [j ..< j + n]" by simp
    also have "\<dots> =  map v (map (op + i) [j ..< j + n])" unfolding comp_apply by simp
    also have "\<dots> =  map v [i + j ..< i + j + n]" by (simp add: algebra_simps)
    finally show ?thesis by this
  qed
  lemma suffix_prefix_suffix[simp]: "map v [i ..< i + n] \<frown> suffix (i + n) v = suffix i v"
  proof -
    have "suffix i v = map (suffix i v) [0 ..< n] \<frown> suffix n (suffix i v)"
      using prefix_suffix
      by blast
    also have "\<dots> = map (suffix i v) [0 ..< 0 + n] \<frown> suffix (i + n) v" by simp
    also have "\<dots> = map v [i ..< i + n] \<frown> suffix (i + n) v" unfolding map_suffix by simp
    finally show ?thesis by simp
  qed
  lemma suffix_singleton_suffix[simp]: "w i ## suffix (Suc i) w = suffix i w"
    using suffix_prefix_suffix[of w i 1] upt.simps by simp

  lemma word_split:
    obtains v\<^sub>1 v\<^sub>2
    where "v = v\<^sub>1 \<frown> v\<^sub>2" "length v\<^sub>1 = k"
  proof
    show "v = map v [0 ..< k] \<frown> suffix k v" using prefix_suffix by this
    show "length (map v [0 ..< k]) = k" by simp
  qed

  lemma concat_eq[iff]:
    assumes "length v\<^sub>1 = length v\<^sub>2"
    shows "v\<^sub>1 \<frown> u\<^sub>1 = v\<^sub>2 \<frown> u\<^sub>2 \<longleftrightarrow> v\<^sub>1 = v\<^sub>2 \<and> u\<^sub>1 = u\<^sub>2"
  proof
    case (goal1)
    have 1: "\<And> i. (v\<^sub>1 \<frown> u\<^sub>1) i = (v\<^sub>2 \<frown> u\<^sub>2) i" using goal1 by auto
    show ?case
    proof (intro conjI ext nth_equalityI allI impI)
      show "length v\<^sub>1 = length v\<^sub>2" using assms(1) by this
    next
      fix i
      assume 2: "i < length v\<^sub>1"
      have 3: "i < length v\<^sub>2" using assms(1) 2 by simp
      show "v\<^sub>1 ! i = v\<^sub>2 ! i" using 1[of i] 2 3 by simp
    next
      fix i
      show "u\<^sub>1 i = u\<^sub>2 i" using 1[of "length v\<^sub>1 + i"] assms(1) by simp
    qed
  next
    case (goal2)
    show ?case using goal2 by simp
  qed
  lemma same_concat_eq[iff]: "u \<frown> v = u \<frown> w \<longleftrightarrow> v = w" by simp

  lemma comp_concat[simp]: "f \<circ> u \<frown> v = map f u \<frown> (f \<circ> v)"
  proof
    fix i
    show "(f \<circ> u \<frown> v) i = (map f u \<frown> (f \<circ> v)) i" by (cases "i < length u", simp+)
  qed

  lemma word_first_split_set:
    assumes "A \<inter> range w \<noteq> {}"
    obtains u a v
    where "w = u \<frown> [a] \<frown> v" "A \<inter> set u = {}" "a \<in> A"
  proof -
    def i \<equiv> "LEAST i. w i \<in> A"
    show ?thesis
    proof
      show "w = prefix i w \<frown> [w i] \<frown> suffix (Suc i) w" by simp
      show "A \<inter> set (prefix i w) = {}"
      proof safe
        case (goal1 a)
        obtain k where 3: "k < i" "w k = a" using goal1(2) by auto
        have 4: "w k \<notin> A" using not_less_Least 3(1) unfolding i_def by this
        show ?case using goal1(1) 3(2) 4 by auto
      qed
      show "w i \<in> A" using LeastI assms(1) unfolding i_def by fast
    qed
  qed

end