theory System
imports
  "Basics/Word_Prefixes"
  "Extensions/Set_Extensions"
  "Extensions/Relation_Extensions"
begin

  locale system =
    fixes en :: "'state \<Rightarrow> 'action set"
    fixes ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
  begin

    inductive words_finp :: "'state \<Rightarrow> 'action list \<Rightarrow> bool"
      where "words_finp p []" | "a \<in> en p \<Longrightarrow> words_finp (ex a p) w \<Longrightarrow> words_finp p (a # w)"
    definition words_fin :: "'state \<Rightarrow> 'action list set"
      where "words_fin p \<equiv> Collect (words_finp p)"

    lemma words_finI_empty[intro!, simp]: "[] \<in> words_fin p"
      unfolding words_fin_def by (auto intro: words_finp.intros)
    lemma words_finI_step[intro!]:
      assumes "a \<in> en p" "w \<in> words_fin (ex a p)"
      shows "a # w \<in> words_fin p"
      using assms unfolding words_fin_def by (auto intro: words_finp.intros)
    lemma words_finE_step[elim!]:
      assumes "a # w \<in> words_fin p"
      obtains "a \<in> en p" "w \<in> words_fin (ex a p)"
      using assms unfolding words_fin_def by (auto elim: words_finp.cases)

    lemma words_fin_induct[case_names empty step, induct set: words_fin]:
      assumes "w \<in> words_fin p"
      assumes "\<And> p. P p []"
      assumes "\<And> a p w. a \<in> en p \<Longrightarrow> w \<in> words_fin (ex a p) \<Longrightarrow> P (ex a p) w \<Longrightarrow> P p (a # w)"
      shows "P p w"
      using words_finp.induct assms unfolding words_fin_def mem_Collect_eq by metis

    lemma words_finI_append[intro!]:
      assumes "u \<in> words_fin p" "v \<in> words_fin (fold ex u p)"
      shows "u @ v \<in> words_fin p"
      using assms by (induct u arbitrary: p rule: list.induct, auto)
    lemma words_finE_append[elim!]:
      assumes "u @ v \<in> words_fin p"
      obtains "u \<in> words_fin p" "v \<in> words_fin (fold ex u p)"
      using assms by (induct u arbitrary: p rule: list.induct, auto)

    lemma prefix_words_fin: "u \<le> v \<Longrightarrow> v \<in> words_fin p \<Longrightarrow> u \<in> words_fin p" by auto

    coinductive words_infp :: "'state \<Rightarrow> 'action word \<Rightarrow> bool"
      where "a \<in> en p \<Longrightarrow> words_infp (ex a p) w \<Longrightarrow> words_infp p (a ## w)"
    definition words_inf :: "'state \<Rightarrow> 'action word set"
      where "words_inf p \<equiv> Collect (words_infp p)"

    lemma words_infI_step[intro!]:
      assumes "a \<in> en p" "w \<in> words_inf (ex a p)"
      shows "a ## w \<in> words_inf p"
      using assms unfolding words_inf_def by (auto intro: words_infp.intros)
    lemma words_infE_step[elim!]:
      assumes "a ## w \<in> words_inf p"
      obtains "a \<in> en p" "w \<in> words_inf (ex a p)"
      using assms unfolding words_inf_def by (auto elim: words_infp.cases)

    lemma words_inf_coinduct[case_names en words_inf, coinduct set: words_inf]:
      assumes "Q p w"
      assumes "\<And> p a w. Q p (a ## w) \<Longrightarrow> a \<in> en p \<and> (Q (ex a p) w \<or> w \<in> words_inf (ex a p))"
      shows "w \<in> words_inf p"
      using build_split words_infp.coinduct assms unfolding words_inf_def mem_Collect_eq by metis

    lemma words_infI_append[intro!]:
      assumes "u \<in> words_fin p" "v \<in> words_inf (fold ex u p)"
      shows "u \<frown> v \<in> words_inf p"
      using assms by (induct u arbitrary: p rule: list.induct, auto)
    lemma words_infE_append[elim!]:
      assumes "u \<frown> v \<in> words_inf p"
      obtains "u \<in> words_fin p" "v \<in> words_inf (fold ex u p)"
      using assms by (induct u arbitrary: p rule: list.induct, auto)

    lemma prefix_words_inf: "u \<le>\<^sub>F\<^sub>I v \<Longrightarrow> v \<in> words_inf p \<Longrightarrow> u \<in> words_fin p" by auto

    lemma words_infI_construct[rule_format, intro?]:
      assumes "\<forall> v. v \<le>\<^sub>F\<^sub>I w \<longrightarrow> v \<in> words_fin p"
      shows "w \<in> words_inf p"
      using assms by (coinduct, auto)

    lemma words_inf_alt_def: "w \<in> words_inf p \<longleftrightarrow> (\<forall> v. v \<le>\<^sub>F\<^sub>I w \<longrightarrow> v \<in> words_fin p)"
      using prefix_words_inf words_infI_construct by metis

    lemma words_infI_construct':
      assumes "\<And> k. \<exists> v. v \<le>\<^sub>F\<^sub>I w \<and> k < length v \<and> v \<in> words_fin p"
      shows "w \<in> words_inf p"
    proof
      fix u
      assume 1: "u \<le>\<^sub>F\<^sub>I w"
      obtain v where 2: "v \<le>\<^sub>F\<^sub>I w" "length u < length v" "v \<in> words_fin p" using assms(1) by auto
      have 3: "length u \<le> length v" using 2(2) by simp
      have 4: "u \<le> v" using prefix_fininf_length 1 2(1) 3 by this
      show "u \<in> words_fin p" using prefix_words_fin 4 2(3) by this
    qed

    lemma words_infI_construct_chain[intro?]:
      assumes "chain w" "\<And> k. w k \<in> words_fin p"
      shows "limit w \<in> words_inf p"
    proof (rule words_infI_construct')
      fix k
      obtain l where 1: "k < length (w l)" using assms(1) by rule
      show "\<exists> v. v \<le>\<^sub>F\<^sub>I limit w \<and> k < length v \<and> v \<in> words_fin p"
      proof (intro exI conjI)
        show "w l \<le>\<^sub>F\<^sub>I limit w" using chain_prefix_limit assms(1) by this
        show "k < length (w l)" using 1 by this
        show "w l \<in> words_fin p" using assms(2) by this
      qed
    qed

    lemma words_inf_run:
      assumes "\<And> i. w i \<in> en (r i)" "\<And> i. r (Suc i) = ex (w i) (r i)"
      shows "w \<in> words_inf (r 0)" "r = scan ex w (r 0)"
    proof -
      have 1: "r = scan ex w (r 0)"
      proof
        fix i
        show "r i = scan ex w (r 0) i" using assms(2) by (induct i, auto)
      qed
      show "w \<in> words_inf (r 0)"
      proof
        fix v
        assume "v \<le>\<^sub>F\<^sub>I w"
        then show "v \<in> words_fin (r 0)"
        proof (induct v rule: rev_induct)
          case Nil
          show ?case by simp
        next
          case (snoc a v)
          show ?case
          proof safe
            show "v \<in> words_fin (r 0)" using snoc by auto
            have "a = w (length v)" using snoc(2) by auto
            also have "\<dots> \<in> en (r (length v))" using assms(1) by this
            also have "\<dots> = en (scan ex w (r 0) (length v))" using 1 by metis
            also have "\<dots> = en (fold ex v (r 0))" using snoc(2) by auto
            finally show "a \<in> en (fold ex v (r 0))" by this
          qed
        qed
      qed
      show "r = scan ex w (r 0)" using 1 by this
    qed

    definition cycles :: "'state \<Rightarrow> 'action list set"
      where "cycles p \<equiv> {w. w \<in> words_fin p \<and> fold ex w p = p}"

    lemma cyclesI[intro!]:
      assumes "w \<in> words_fin p" "fold ex w p = p"
      shows "w \<in> cycles p"
      using assms unfolding cycles_def by auto
    lemma cyclesE[elim!]:
      assumes "w \<in> cycles p"
      obtains "w \<in> words_fin p" "fold ex w p = p"
      using assms unfolding cycles_def by auto

    lemma words_fin_blocked:
      assumes "\<And> w. w \<in> words_fin p \<Longrightarrow> A \<inter> set w = {} \<Longrightarrow> A \<inter> en (fold ex w p) \<subseteq> A \<inter> en p"
      assumes "w \<in> words_fin p" "A \<inter> en p \<inter> set w = {}"
      shows "A \<inter> set w = {}"
      using assms by (induct w rule: rev_induct, auto)

    definition en' :: "'state \<Rightarrow> 'action option set"
      where "en' p \<equiv> if en p = {} then {None} else Some ` en p"
    definition ex' :: "'action option \<Rightarrow> 'state \<Rightarrow> 'state"
      where "ex' \<equiv> \<lambda> None \<Rightarrow> id | Some a \<Rightarrow> ex a"

  end

  locale system_initial =
    system en ex
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
  begin

    inductive_set reachable :: "'state set"
      where reachable: "p \<in> init \<Longrightarrow> w \<in> words_fin p \<Longrightarrow> fold ex w p \<in> reachable"

    lemma reachableI_init[intro]:
      assumes "p \<in> init"
      shows "p \<in> reachable"
      using reachable assms by force
    lemma reachableI_step[intro!]:
      assumes "p \<in> reachable" "a \<in> en p"
      shows "ex a p \<in> reachable"
      using assms reachable by induct force
    lemma reachableI_words_fin[intro!]:
      assumes "p \<in> reachable" "w \<in> words_fin p"
      shows "fold ex w p \<in> reachable"
      using assms reachable by induct force
    lemma reachableE[elim?]:
      assumes "q \<in> reachable"
      obtains p w
      where "p \<in> init" "w \<in> words_fin p" "fold ex w p = q"
      using assms by induct force
    lemma reachable_induct[case_names init step, induct set: reachable]:
      assumes "q \<in> reachable"
      assumes "\<And> p. p \<in> init \<Longrightarrow> P p"
      assumes "\<And> p a. p \<in> reachable \<Longrightarrow> P p \<Longrightarrow> a \<in> en p \<Longrightarrow> P (ex a p)"
      shows "P q"
    using assms(1)
    proof induct
      fix p w
      assume "w \<in> words_fin p" "p \<in> init"
      then show "P (fold ex w p)"
      proof (induct w rule: rev_induct)
        case Nil
        show ?case using assms(2) Nil by simp
      next
        case (snoc a w)
        have "P (ex a (fold ex w p))" using assms(3) snoc by auto
        then show ?case using snoc by simp
      qed
    qed

    inductive_set executable :: "'action set"
      where executable: "p \<in> reachable \<Longrightarrow> a \<in> en p \<Longrightarrow> a \<in> executable"

    lemma executableI_step[intro!]:
      assumes "p \<in> reachable" "a \<in> en p"
      shows "a \<in> executable"
      using executable assms by this
    lemma executableI_words_fin[intro!]:
      assumes "p \<in> reachable" "w \<in> words_fin p"
      shows "set w \<subseteq> executable"
      using assms by (induct w arbitrary: p, auto del: subsetI)
    lemma executableE[elim?]:
      assumes "a \<in> executable"
      obtains p
      where "p \<in> reachable" "a \<in> en p"
      using assms by induct auto

  end

  locale system_interpreted =
    system en ex
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and int :: "'state \<Rightarrow> 'interpretation"
  begin

    definition visible :: "'action set"
      where "visible \<equiv> {a. \<exists> q. a \<in> en q \<and> int q \<noteq> int (ex a q)}"

    lemma visibleI[intro]:
      assumes "a \<in> en q" "int q \<noteq> int (ex a q)"
      shows "a \<in> visible"
      using assms unfolding visible_def by auto
    lemma visibleE[elim]:
      assumes "a \<in> visible"
      obtains q
      where "a \<in> en q" "int q \<noteq> int (ex a q)"
      using assms unfolding visible_def by auto

    abbreviation "invisible \<equiv> - visible"

    lemma execute_fin_word_invisible:
      assumes "w \<in> words_fin p" "set w \<subseteq> invisible"
      shows "int (fold ex w p) = int p"
      using assms by (induct w arbitrary: p rule: list.induct, auto)
    lemma execute_inf_word_invisible:
      assumes "w \<in> words_inf p" "k \<le> l" "\<And> i. k \<le> i \<Longrightarrow> i < l \<Longrightarrow> w i \<notin> visible"
      shows "int (scan ex w p k) = int (scan ex w p l)"
    proof -
      have 1: "w = prefix k w \<frown> map w [k ..< l] \<frown> suffix l w"
        using assms(2) by (metis le_Suc_ex prefix_suffix suffix_prefix_suffix)
      have 2: "prefix l w = prefix k w @ map w [k ..< l]"
        using assms(2) by (metis le0 le_Suc_ex map_append upt_add_eq_append)
      have 3: "prefix k w \<frown> map w [k ..< l] \<frown> suffix l w \<in> words_inf p" using assms(1) 1 by auto
      have 4: "int (fold ex (map w [k ..< l]) (scan ex w p k)) = int (scan ex w p k)"
      proof (rule execute_fin_word_invisible)
        show "map w [k ..< l] \<in> words_fin (scan ex w p k)" using 3 by auto
        show "set (map w [k ..< l]) \<subseteq> invisible" using assms(3) by auto
      qed
      have "int (scan ex w p k) = int (fold ex (map w [k ..< l]) (scan ex w p k))" using 4 by simp
      also have "\<dots> = int (fold ex (prefix k w @ map w [k ..< l]) p)" by simp
      also have "\<dots> = int (fold ex (prefix l w) p)" using 2 by simp
      finally show ?thesis by this
    qed

  end

  locale system_complete =
    system_initial en ex init +
    system_interpreted en ex int
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
    and int :: "'state \<Rightarrow> 'interpretation"
  begin

    definition language :: "'interpretation word set"
      where "language \<equiv> {int \<circ> scan ex w p |p w. p \<in> init \<and> w \<in> words_inf p}"

    lemma languageI[intro!]:
      assumes "w = int \<circ> scan ex v p" "p \<in> init" "v \<in> words_inf p"
      shows "w \<in> language"
      using assms unfolding language_def by auto
    lemma languageE[elim!]:
      assumes "w \<in> language"
      obtains p v
      where "w = int \<circ> scan ex v p" "p \<in> init" "v \<in> words_inf p"
      using assms unfolding language_def by auto

  end

  locale system_finite_reachable =
    system_initial en ex init
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
    +
    assumes reachable_finite: "finite reachable"

  locale system_cut =
    system_finite_reachable en ex init
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
    +
    fixes cuts :: "'action set"
    assumes cycles_cut: "p \<in> reachable \<Longrightarrow> w \<in> cycles p \<Longrightarrow> w \<noteq> [] \<Longrightarrow> set w \<inter> cuts \<noteq> {}"
  begin

    inductive scut :: "'state \<Rightarrow> 'state \<Rightarrow> bool"
      where scut: "p \<in> reachable \<Longrightarrow> a \<in> en p \<Longrightarrow> a \<notin> cuts \<Longrightarrow> scut p (ex a p)"

    declare scut.intros[intro!]
    declare scut.cases[elim!]

    lemma scut_reachable:
      assumes "scut p q"
      shows "p \<in> reachable" "q \<in> reachable"
      using assms by auto
    lemma scut_trancl:
      assumes "scut\<^sup>+\<^sup>+ p q"
      obtains w
      where "w \<in> words_fin p" "fold ex w p = q" "set w \<inter> cuts = {}" "w \<noteq> []"
    using assms
    proof (induct arbitrary: thesis)
      case (base q)
      show ?case using base by force
    next
      case (step q r)
      obtain w where 1: "w \<in> words_fin p" "fold ex w p = q" "set w \<inter> cuts = {}" "w \<noteq> []"
        using step(3) by this
      obtain a where 2: "a \<in> en q" "a \<notin> cuts" "ex a q = r" using step(2) by auto
      show ?case
      proof (rule step(4))
        show "w @ [a] \<in> words_fin p" using 1 2 by auto
        show "fold ex (w @ [a]) p = r" using 1 2 by auto
        show "set (w @ [a]) \<inter> cuts = {}" using 1 2 by auto
        show "w @ [a] \<noteq> []" by auto
      qed
    qed

    sublocale wellfounded_relation "scut\<inverse>\<inverse>"
    proof (unfold_locales, intro finite_acyclic_wf_converse[to_pred] acyclicI[to_pred], safe)
      have 1: "{(p, q). scut p q} \<subseteq> reachable \<times> reachable" using scut_reachable by blast
      have 2: "finite (reachable \<times> reachable)"
        using finite_cartesian_product reachable_finite by blast
      show "finite {(p, q). scut p q}" using 1 2 by blast
    next
      fix p
      assume 1: "scut\<^sup>+\<^sup>+ p p"
      have 2: "p \<in> reachable" using 1 tranclE[to_pred] scut_reachable by metis
      obtain w where 3: "w \<in> words_fin p" "fold ex w p = p" "set w \<inter> cuts = {}" "w \<noteq> []"
        using scut_trancl 1 by this
      have 4: "w \<in> cycles p" using 3(1, 2) by auto
      have 5: "set w \<inter> cuts \<noteq> {}" using cycles_cut 2 4 3(4) by this
      show "False" using 3(3) 5 by simp
    qed

    lemma no_cut_scut:
      assumes "p \<in> reachable" "a \<in> en p" "a \<notin> cuts"
      shows "scut\<inverse>\<inverse> (ex a p) p"
      using assms by auto

  end

  locale system_sticky =
    system_complete en ex init int +
    system_cut en ex init sticky
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
    and int :: "'state \<Rightarrow> 'interpretation"
    and sticky :: "'action set"
    +
    assumes executable_visible_sticky: "executable \<inter> visible \<subseteq> sticky"

end
