theory System_Traces
imports
  "System"
  "Traces"
begin

  locale system_traces =
    system en ex +
    traces ind
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and ind :: "'action \<Rightarrow> 'action \<Rightarrow> bool"
    +
    assumes en: "ind a b \<Longrightarrow> a \<in> en p \<Longrightarrow> b \<in> en p \<longleftrightarrow> b \<in> en (ex a p)"
    assumes ex: "ind a b \<Longrightarrow> a \<in> en p \<Longrightarrow> b \<in> en p \<Longrightarrow> ex b (ex a p) = ex a (ex b p)"
  begin

    lemma diamond_bottom:
      assumes "ind a b"
      assumes "a \<in> en p" "b \<in> en p"
      shows "a \<in> en (ex b p)" "b \<in> en (ex a p)" "ex b (ex a p) = ex a (ex b p)"
      using assms independence_symmetric en ex by metis+
    lemma diamond_right:
      assumes "ind a b"
      assumes "a \<in> en p" "b \<in> en (ex a p)"
      shows "a \<in> en (ex b p)" "b \<in> en p" "ex b (ex a p) = ex a (ex b p)"
      using assms independence_symmetric en ex by metis+
    lemma diamond_left:
      assumes "ind a b"
      assumes "a \<in> en (ex b p)" "b \<in> en p"
      shows "a \<in> en p" "b \<in> en (ex a p)" "ex b (ex a p) = ex a (ex b p)"
      using assms independence_symmetric en ex by metis+

    lemma eq_swap_word:
      assumes "w\<^sub>1 =\<^sub>S w\<^sub>2" "w\<^sub>1 \<in> words_fin p"
      shows "w\<^sub>2 \<in> words_fin p"
      using assms diamond_right by (induct, auto)
    lemma eq_fin_word:
      assumes "w\<^sub>1 =\<^sub>F w\<^sub>2" "w\<^sub>1 \<in> words_fin p"
      shows "w\<^sub>2 \<in> words_fin p"
      using assms eq_swap_word by (induct, auto)
    lemma le_fin_word:
      assumes "w\<^sub>1 \<preceq>\<^sub>F w\<^sub>2" "w\<^sub>2 \<in> words_fin p"
      shows "w\<^sub>1 \<in> words_fin p"
      using assms eq_fin_word by blast
    lemma le_fininf_word:
      assumes "w\<^sub>1 \<preceq>\<^sub>F\<^sub>I w\<^sub>2" "w\<^sub>2 \<in> words_inf p"
      shows "w\<^sub>1 \<in> words_fin p"
      using assms le_fin_word by blast
    lemma le_inf_word:
      assumes "w\<^sub>2 \<preceq>\<^sub>I w\<^sub>1" "w\<^sub>1 \<in> words_inf p"
      shows "w\<^sub>2 \<in> words_inf p"
      using assms le_fininf_word by (intro words_infI_construct, blast)
    lemma eq_inf_word:
      assumes "w\<^sub>1 =\<^sub>I w\<^sub>2" "w\<^sub>1 \<in> words_inf p"
      shows "w\<^sub>2 \<in> words_inf p"
      using assms le_inf_word by auto

    lemma eq_swap_execute:
      assumes "w\<^sub>1 \<in> words_fin p" "w\<^sub>1 =\<^sub>S w\<^sub>2"
      shows "fold ex w\<^sub>1 p = fold ex w\<^sub>2 p"
      using assms(2, 1) diamond_right by (induct, auto)
    lemma eq_fin_execute:
      assumes "w\<^sub>1 \<in> words_fin p" "w\<^sub>1 =\<^sub>F w\<^sub>2"
      shows "fold ex w\<^sub>1 p = fold ex w\<^sub>2 p"
      using assms(2, 1) eq_fin_word eq_swap_execute by (induct, auto)

    lemma diamond_fin_word_step:
      assumes "Ind {a} (set v)" "a \<in> en p" "v \<in> words_fin p"
      shows "v \<in> words_fin (ex a p)"
      using diamond_bottom assms by (induct v arbitrary: p, auto, metis)
    lemma diamond_inf_word_step:
      assumes "Ind {a} (range w)" "a \<in> en p" "w \<in> words_inf p"
      shows "w \<in> words_inf (ex a p)"
      using diamond_fin_word_step assms by (intro words_infI_construct, fastforce)
    lemma diamond_fin_word_inf_word:
      assumes "Ind (set v) (range w)" "v \<in> words_fin p" "w \<in> words_inf p"
      shows "w \<in> words_inf (fold ex v p)"
      using diamond_inf_word_step assms by (induct v arbitrary: p, auto)
    lemma diamond_fin_word_inf_word':
      assumes "Ind (set v) (range w)" "u @ v \<in> words_fin p" "u \<frown> w \<in> words_inf p"
      shows "u \<frown> v \<frown> w \<in> words_inf p"
      using assms diamond_fin_word_inf_word by auto

  end

end
