open RegularExpression
open NFA_string
open ThompsonString

fun nth l 0 = hd l
  | nth l n = nth (tl l) (n-1)

fun getParameterString n default =
	if length (CommandLine.arguments ()) <= n
	then default
	else nth (CommandLine.arguments ()) n

fun getParameterInteger n default =
	if length (CommandLine.arguments ()) <= n
	then default
	else
		case Int.fromString (nth (CommandLine.arguments ()) n)
		of SOME k => k
		 | NONE => default

fun naturalNumbers 0 = []
  | naturalNumbers n = naturalNumbers (n - 1) @ [n - 1]

fun measureTime f x =
	let
		val startTime = Time.now ()
		val result = f x
		val endTime = Time.now ()
	in (result, Time.- (endTime, startTime))
	end

fun translateRegexp regexp =
	let
		val (nfa, duration) = measureTime taTranslate regexp
		val duration = Time.toReal duration
		val stateCount = Real.fromInt (NFA_string.nfa_states_no nfa)
		val transitionCount = Real.fromInt (NFA_string.nfa_trans_no nfa)
	in (duration, stateCount, transitionCount)
	end
fun translateRegexps length count =
	let
		val results = map (fn n => translateRegexp (liftToString (generateRandom length))) (naturalNumbers count)
		val totalDuration = foldl (fn ((duration, stateCount, transitionCount), current) => current + duration) 0.0 results
		val totalStateCount = foldl (fn ((duration, stateCount, transitionCount), current) => current + stateCount) 0.0 results
		val totalTransitionCount = foldl (fn ((duration, stateCount, transitionCount), current) => current + transitionCount) 0.0 results
		val _ = print (Int.toString length)
		val _ = print " "
		val _ = print (Int.toString count)
		val _ = print " "
		val _ = print (Real.toString totalDuration)
		val _ = print " "
		val _ = print (Real.toString (totalDuration / Real.fromInt count))
		val _ = print " "
		val _ = print (Real.toString (totalStateCount / Real.fromInt count))
		val _ = print " "
		val _ = print (Real.toString (totalTransitionCount / Real.fromInt count))
	in print "\n"
	end

val regexpLength = getParameterInteger 0 512
val regexpCount = getParameterInteger 1 100
val _ = translateRegexps regexpLength regexpCount
