theory NFA_Extensions
imports
  "../../Automata_Malik_Tuerk/NFA"
  LTS_Extensions
begin

  lemma (in NFA) LTS_fin_path_consistent_f:
    assumes "LTS_is_fin_path (\<Delta> \<A>) f ts l"
    assumes "l \<in> \<Q> \<A>"
    shows "f \<in> \<Q> \<A>"
    using assms \<Delta>_consistent
    by (induct rule: LTS_is_fin_path_induct, auto)
  lemma (in NFA) LTS_fin_path_consistent_l:
    assumes "LTS_is_fin_path (\<Delta> \<A>) f ts l"
    assumes "f \<in> \<Q> \<A>"
    shows "l \<in> \<Q> \<A>"
    using assms \<Delta>_consistent
    by (induct rule: LTS_is_fin_path_induct, auto)

  lemma NFA_language_mono:
    assumes a: "\<Delta> \<A>1 \<subseteq> \<Delta> \<A>2" "\<I> \<A>1 \<subseteq> \<I> \<A>2" "\<F> \<A>1 \<subseteq> \<F> \<A>2"
    shows "\<L> \<A>1 \<subseteq> \<L> \<A>2"
  proof
    fix w
    assume b: "w \<in> \<L> \<A>1"
    have c: "NFA_accept \<A>1 w" using b unfolding \<L>_def by auto
    obtain p q where d: "LTS_is_reachable (\<Delta> \<A>1) p w q" "p \<in> \<I> \<A>1" "q \<in> \<F> \<A>1"
      using c unfolding NFA_accept_def by auto
    have e: "LTS_is_reachable (\<Delta> \<A>2) p w q" "p \<in> \<I> \<A>2" "q \<in> \<F> \<A>2"
      using LTS_is_reachable_mono[OF a(1) d(1)] d(2) a(2) d(3) a(3) by auto
    have f: "NFA_accept \<A>2 w" using e unfolding NFA_accept_def by auto
    show "w \<in> \<L> \<A>2" using f unfolding \<L>_def by auto
  qed

  lemma NFA_empty_word: "\<I> \<A> \<inter> \<F> \<A> \<noteq> {} \<longleftrightarrow> [] \<in> \<L> \<A>"
    by (auto simp: \<L>_def NFA_accept_def)

  locale disjoint_NFAs =
    NFA_\<A>1: NFA \<A>1 + NFA_\<A>2: NFA \<A>2
    for \<A>1 :: "('q, 'a, _) NFA_rec_scheme"
    and \<A>2 :: "('q, 'a, _) NFA_rec_scheme"
    +
    assumes disjoint_\<Q>: "\<Q> \<A>1 \<inter> \<Q> \<A>2 = {}"

  sublocale disjoint_NFAs < reversed: disjoint_NFAs \<A>2 \<A>1
    using disjoint_\<Q>
    by (unfold_locales, auto)

  lemma (in disjoint_NFAs) disjoint_\<I>: "\<I> \<A>1 \<inter> \<I> \<A>2 = {}"
    using NFA_\<A>1.\<I>_consistent
    using NFA_\<A>2.\<I>_consistent
    using disjoint_\<Q>
    by auto

  lemma (in disjoint_NFAs) disjoint_\<F>: "\<F> \<A>1 \<inter> \<F> \<A>2 = {}"
    using NFA_\<A>1.\<F>_consistent
    using NFA_\<A>2.\<F>_consistent
    using disjoint_\<Q>
    by auto

end
