theory TA_Definitions
imports
  "../../Libs/Regular-Sets/Regular_Exp"
  "../../Automata_Malik_Tuerk/NFA"
begin

  definition TA_bridges :: "'q set \<Rightarrow> 'q set \<Rightarrow> ('q \<times> 'a \<times> 'q) set \<Rightarrow>
    ('q \<times> 'a \<times> 'q) set \<Rightarrow> ('q \<times> 'a \<times> 'q) set"
  where "TA_bridges \<F>1 \<I>2 \<Delta>2 \<Delta>0 =
    \<Delta>0 \<union> {(p1, a2, q2) | p1 p2 a2 q2. p1 \<in> \<F>1 \<and> p2 \<in> \<I>2 \<and> (p2, a2, q2) \<in> \<Delta>2}"

  type_synonym 'a TA_state = "(nat, 'a) NFA_rec \<times> nat \<times> bool"
  type_synonym 'a TA_adder = "'a TA_state \<Rightarrow> 'a TA_state"

  fun TA_zero_NFA :: "'a TA_adder"
  where "TA_zero_NFA (\<A>0, n0, e0) = (\<A>0, n0, e0)"
  fun TA_one_NFA :: "'a TA_adder"
  where
    "TA_one_NFA (\<A>0, n0, e0) =
    (
      \<lparr>
        \<Q> = insert n0 (\<Q> \<A>0),
        \<Sigma> = \<Sigma> \<A>0,
        \<Delta> = \<Delta> \<A>0,
        \<I> = insert n0 (\<I> \<A>0),
        \<F> = insert n0 (\<F> \<A>0)
      \<rparr>,
      Suc n0,
      e0 \<or> True
    )"
  fun TA_atom_NFA :: "'a \<Rightarrow> 'a TA_adder"
  where
    "TA_atom_NFA a (\<A>0, n0, e0) =
    (
      \<lparr>
        \<Q> = insert (Suc n0) (insert n0 (\<Q> \<A>0)),
        \<Sigma> = insert a (\<Sigma> \<A>0),
        \<Delta> = insert (n0, a, Suc n0) (\<Delta> \<A>0),
        \<I> = insert n0 (\<I> \<A>0),
        \<F> = insert (Suc n0) (\<F> \<A>0)
      \<rparr>,
      Suc (Suc n0),
      e0 \<or> False
    )"
  fun TA_plus_NFA :: "'a TA_adder \<Rightarrow> 'a TA_adder \<Rightarrow> 'a TA_adder"
  where
    "TA_plus_NFA add_\<A>1 add_\<A>2 (\<A>0, n0, e0) = add_\<A>2 (add_\<A>1 (\<A>0, n0, e0))"
  fun TA_times_NFA :: "'a TA_adder \<Rightarrow> 'a TA_adder \<Rightarrow> 'a TA_adder"
  where
    "TA_times_NFA add_\<A>1 add_\<A>2 (\<A>0, n0, e0) =
    (
      let
        (\<A>1, n1, e1) =
          add_\<A>1 (\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>, n0, False);
        (\<A>2, n2, e2) =
          add_\<A>2 (\<lparr> \<Q> = \<Q> \<A>1, \<Sigma> = \<Sigma> \<A>1, \<Delta> = \<Delta> \<A>1, \<I> = {}, \<F> = {} \<rparr>, n1, False)
      in
      (
        \<lparr>
          \<Q> = \<Q> \<A>2,
          \<Sigma> = \<Sigma> \<A>2,
          \<Delta> = TA_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2) (\<Delta> \<A>2),
          \<I> = \<I> \<A>0 \<union> \<I> \<A>1,
          \<F> = \<F> \<A>0 \<union> (if e2 then \<F> \<A>1 \<union> \<F> \<A>2 else \<F> \<A>2)
        \<rparr>,
        n2,
        e0 \<or> e1 \<and> e2
      )
    )"
  fun TA_transcl_NFA :: "'a TA_adder \<Rightarrow> 'a TA_adder"
  where
    "TA_transcl_NFA add_\<A> (\<A>0, n0, e0) =
    (
      let
        (\<A>1, n1, e1) =
          add_\<A> (\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>, n0, False)
      in
      (
        \<lparr>
          \<Q> = \<Q> \<A>1,
          \<Sigma> = \<Sigma> \<A>1,
          \<Delta> = TA_bridges (\<F> \<A>1) (\<I> \<A>1) (\<Delta> \<A>1) (\<Delta> \<A>1),
          \<I> = \<I> \<A>0 \<union> \<I> \<A>1,
          \<F> = \<F> \<A>0 \<union> \<F> \<A>1
        \<rparr>,
        n1,
        e0 \<or> e1
      )
    )"
  fun TA_reflcl_NFA :: "'a TA_adder \<Rightarrow> 'a TA_adder"
  where
    "TA_reflcl_NFA add_\<A> (\<A>0, n0, e0) =
    (
      let
        (\<A>1, n1, e1) =
          add_\<A> (\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>, n0, False);
        add_reflcl = if e1 then TA_zero_NFA else TA_one_NFA;
        (\<A>2, n2, e2) = add_reflcl (\<A>1, n1, e1)
      in
      (
        \<lparr>
          \<Q> = \<Q> \<A>2,
          \<Sigma> = \<Sigma> \<A>2,
          \<Delta> = \<Delta> \<A>2,
          \<I> = \<I> \<A>0 \<union> \<I> \<A>2,
          \<F> = \<F> \<A>0 \<union> \<F> \<A>2
        \<rparr>,
        n2,
        e0 \<or> e2
      )
    )"
  fun TA_star_NFA :: "'a TA_adder \<Rightarrow> 'a TA_adder"
  where "TA_star_NFA add_\<A> (\<A>0, n0, e0) =
    TA_reflcl_NFA (TA_transcl_NFA add_\<A>) (\<A>0, n0, e0)"

  fun TA_translate_NFA :: "'a rexp \<Rightarrow> 'a TA_adder"
  where
    "TA_translate_NFA Regular_Exp.Zero = TA_zero_NFA" |
    "TA_translate_NFA Regular_Exp.One = TA_one_NFA" |
    "TA_translate_NFA (Regular_Exp.Atom a) = TA_atom_NFA a" |
    "TA_translate_NFA (Regular_Exp.Plus r1 r2) =
      TA_plus_NFA (TA_translate_NFA r1) (TA_translate_NFA r2)" |
    "TA_translate_NFA (Regular_Exp.Times r1 r2) =
      TA_times_NFA (TA_translate_NFA r1) (TA_translate_NFA r2)" |
    "TA_translate_NFA (Regular_Exp.Star r) =
      TA_star_NFA (TA_translate_NFA r)"
  fun TA_translate_NFA_wrapper :: "'a rexp \<Rightarrow> (nat, 'a) NFA_rec"
  where "TA_translate_NFA_wrapper r = fst (TA_translate_NFA r
    (\<lparr> \<Q> = {}, \<Sigma> = {}, \<Delta> = {}, \<I> = {}, \<F> = {} \<rparr>, 0, False))"

end
