theory TA_EqualityTM
imports
  Common
  TM_Correctness
  TA_Definitions
begin

  lemma TA_bridges_eq_TM: "TA_bridges \<F>1 \<I>2 \<Delta>2 \<Delta>0 = \<Delta>0 \<union> TM_bridges \<F>1 \<I>2 \<Delta>2"
    unfolding TA_bridges_def TM_bridges_def
    by auto

  locale TA_adder_empty =
    fixes add_\<A> :: "'a TA_adder"
  locale TA_adder_empty1 =
    add_\<A>: TA_adder_empty add_\<A>
    for add_\<A> :: "'a TA_adder"
  locale TA_adder_empty2 =
    add_\<A>1: TA_adder_empty add_\<A>1 +
    add_\<A>2: TA_adder_empty add_\<A>2
    for add_\<A>1 :: "'a TA_adder"
    and add_\<A>2 :: "'a TA_adder"

  locale TA_adder_inc =
    TA_adder_empty add_\<A>
    for add_\<A> :: "'a TA_adder"
    +
    assumes inc:
      "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow> \<Q> \<A>1 \<supseteq> \<Q> \<A>0"
      "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow> \<Sigma> \<A>1 \<supseteq> \<Sigma> \<A>0"
      "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow> \<Delta> \<A>1 \<supseteq> \<Delta> \<A>0"
      "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow> \<I> \<A>1 \<supseteq> \<I> \<A>0"
      "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow> \<F> \<A>1 \<supseteq> \<F> \<A>0"
  locale TA_adder_inc1 =
    TA_adder_empty1 add_\<A> +
    add_\<A>: TA_adder_inc add_\<A>
    for add_\<A> :: "'a TA_adder"
  locale TA_adder_inc2 =
    TA_adder_empty2 add_\<A>1 add_\<A>2 +
    add_\<A>1: TA_adder_inc add_\<A>1 +
    add_\<A>2: TA_adder_inc add_\<A>2
    for add_\<A>1 :: "'a TA_adder"
    and add_\<A>2 :: "'a TA_adder"

  interpretation TA_zero_NFA: TA_adder_inc "TA_zero_NFA"
    by (unfold_locales, auto)
  interpretation TA_one_NFA: TA_adder_inc "TA_one_NFA"
    by (unfold_locales, auto)
  interpretation TA_atom_NFA: TA_adder_inc "TA_atom_NFA a"
    by (unfold_locales, auto)
  sublocale TA_adder_inc2 \<subseteq> TA_plus_NFA:
    TA_adder_inc "TA_plus_NFA add_\<A>1 add_\<A>2"
  proof (unfold_locales)
    fix \<A>0 n0 e0 \<A>2 n2 e2
    assume a: "(\<A>2, n2, e2) = TA_plus_NFA add_\<A>1 add_\<A>2 (\<A>0, n0, e0)"
    obtain \<A>1 n1 e1 where b1: "(\<A>1, n1, e1) = add_\<A>1 (\<A>0, n0, e0)"
      using PairE by metis
    have b2: "(\<A>2, n2, e2) = add_\<A>2 (\<A>1, n1, e1)" using a b1 by simp
    show
      "\<Q> \<A>2 \<supseteq> \<Q> \<A>0"
      "\<Sigma> \<A>2 \<supseteq> \<Sigma> \<A>0"
      "\<Delta> \<A>2 \<supseteq> \<Delta> \<A>0"
      "\<I> \<A>2 \<supseteq> \<I> \<A>0"
      "\<F> \<A>2 \<supseteq> \<F> \<A>0"
      using add_\<A>1.inc[OF b1] add_\<A>2.inc[OF b2]
      by auto
  qed
  sublocale TA_adder_inc2 \<subseteq> TA_times_NFA:
    TA_adder_inc "TA_times_NFA add_\<A>1 add_\<A>2"
  proof (unfold_locales)
    fix \<A>0 n0 e0 \<A>5 n5 e5
    assume a: "(\<A>5, n5, e5) = TA_times_NFA add_\<A>1 add_\<A>2 (\<A>0, n0, e0)"
    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>2 n2 e2 where b1: "(\<A>2, n2, e2) = add_\<A>1 (?\<A>1, n0, False)"
      using PairE by metis
    let ?\<A>3 = "\<lparr> \<Q> = \<Q> \<A>2, \<Sigma> = \<Sigma> \<A>2, \<Delta> = \<Delta> \<A>2, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>4 n4 e4 where b2: "(\<A>4, n4, e4) = add_\<A>2 (?\<A>3, n2, False)"
      using PairE by metis
    show
      "\<Q> \<A>5 \<supseteq> \<Q> \<A>0"
      "\<Sigma> \<A>5 \<supseteq> \<Sigma> \<A>0"
      "\<Delta> \<A>5 \<supseteq> \<Delta> \<A>0"
      "\<I> \<A>5 \<supseteq> \<I> \<A>0"
      "\<F> \<A>5 \<supseteq> \<F> \<A>0"
      using a add_\<A>1.inc[OF b1] add_\<A>2.inc[OF b2]
      by (auto simp: b1[symmetric] b2[symmetric] TA_bridges_def)
  qed
  sublocale TA_adder_inc1 \<subseteq> TA_transcl_NFA:
    TA_adder_inc "TA_transcl_NFA add_\<A>"
  proof (unfold_locales)
    fix \<A>0 n0 e0 \<A>3 n3 e3
    assume a: "(\<A>3, n3, e3) = TA_transcl_NFA add_\<A> (\<A>0, n0, e0)"
    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>2 n2 e2 where b1: "(\<A>2, n2, e2) = add_\<A> (?\<A>1, n0, False)"
      using PairE by metis
    show
      "\<Q> \<A>3 \<supseteq> \<Q> \<A>0"
      "\<Sigma> \<A>3 \<supseteq> \<Sigma> \<A>0"
      "\<Delta> \<A>3 \<supseteq> \<Delta> \<A>0"
      "\<I> \<A>3 \<supseteq> \<I> \<A>0"
      "\<F> \<A>3 \<supseteq> \<F> \<A>0"
      using a add_\<A>.inc[OF b1]
      by (auto simp: b1[symmetric] TA_bridges_def)
  qed
  sublocale TA_adder_inc1 \<subseteq> TA_reflcl_NFA:
    TA_adder_inc "TA_reflcl_NFA add_\<A>"
  proof (unfold_locales)
    fix \<A>0 n0 e0 \<A>4 n4 e4
    assume a: "(\<A>4, n4, e4) = TA_reflcl_NFA add_\<A> (\<A>0, n0, e0)"
    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>2 n2 e2 where b1: "(\<A>2, n2, e2) = add_\<A> (?\<A>1, n0, False)"
      using PairE by metis
    show
      "\<Q> \<A>4 \<supseteq> \<Q> \<A>0"
      "\<Sigma> \<A>4 \<supseteq> \<Sigma> \<A>0"
      "\<Delta> \<A>4 \<supseteq> \<Delta> \<A>0"
      "\<I> \<A>4 \<supseteq> \<I> \<A>0"
      "\<F> \<A>4 \<supseteq> \<F> \<A>0"
      using a add_\<A>.inc[OF b1]
      by (cases e2, auto simp: b1[symmetric])+
  qed
  sublocale TA_adder_inc1 \<subseteq> TA_star_NFA:
    TA_adder_inc "TA_star_NFA add_\<A>"
  proof (unfold_locales)
    fix \<A>0 n0 e0 \<A>1 n1 e1
    assume a: "(\<A>1, n1, e1) = TA_star_NFA add_\<A> (\<A>0, n0, e0)"
    have b1: "(\<A>1, n1, e1) = TA_reflcl_NFA (TA_transcl_NFA add_\<A>) (\<A>0, n0, e0)"
      using a by auto
    interpret TA_transcl_NFA: TA_adder_inc1 "TA_transcl_NFA add_\<A>"
      by unfold_locales
    show
      "\<Q> \<A>1 \<supseteq> \<Q> \<A>0"
      "\<Sigma> \<A>1 \<supseteq> \<Sigma> \<A>0"
      "\<Delta> \<A>1 \<supseteq> \<Delta> \<A>0"
      "\<I> \<A>1 \<supseteq> \<I> \<A>0"
      "\<F> \<A>1 \<supseteq> \<F> \<A>0"
      using TA_transcl_NFA.TA_reflcl_NFA.inc[OF b1]
      by this
  qed

  interpretation TA_translate_NFA: TA_adder_inc "TA_translate_NFA r"
  proof (induct r)
    case Zero thus ?case by (simp, unfold_locales)
    case One thus ?case by (simp, unfold_locales)
    case Atom thus ?case by (simp, unfold_locales)
  next
    case (Plus r1 r2)
    interpret TA_adder_inc "TA_translate_NFA r1" using Plus(1) by this
    interpret TA_adder_inc "TA_translate_NFA r2" using Plus(2) by this
    interpret TA_adder_inc2 "TA_translate_NFA r1" "TA_translate_NFA r2"
      by intro_locales
    show ?case by (simp, intro_locales)
  next
    case (Times r1 r2)
    interpret TA_adder_inc "TA_translate_NFA r1" using Times(1) by this
    interpret TA_adder_inc "TA_translate_NFA r2" using Times(2) by this
    interpret TA_adder_inc2 "TA_translate_NFA r1" "TA_translate_NFA r2"
      by intro_locales
    show ?case by (simp, intro_locales)
  next
    case (Star r)
    interpret TA_adder_inc "TA_translate_NFA r" using Star by this
    interpret TA_adder_inc1 "TA_translate_NFA r"
      by intro_locales
    show ?case by (simp, intro_locales)
  qed

  locale TA_adder_well_formed =
    TA_adder_inc add_\<A>
    for add_\<A> :: "'a TA_adder"
    +
    assumes preserves_well_formed: "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow>
      NFA \<A>0 \<Longrightarrow> NFA \<A>1"
  locale TA_adder_well_formed1 =
    TA_adder_inc1 add_\<A> +
    add_\<A>: TA_adder_well_formed add_\<A>
    for add_\<A> :: "'a TA_adder"
  locale TA_adder_well_formed2 =
    TA_adder_inc2 add_\<A>1 add_\<A>2 +
    add_\<A>1: TA_adder_well_formed add_\<A>1 +
    add_\<A>2: TA_adder_well_formed add_\<A>2
    for add_\<A>1 :: "'a TA_adder"
    and add_\<A>2 :: "'a TA_adder"

  interpretation TA_zero_NFA: TA_adder_well_formed "TA_zero_NFA"
    apply unfold_locales
    unfolding NFA_full_def
    apply auto
  done
  interpretation TA_one_NFA: TA_adder_well_formed "TA_one_NFA"
    apply unfold_locales
    unfolding NFA_full_def
    apply auto
  done
  interpretation TA_atom_NFA: TA_adder_well_formed "TA_atom_NFA a"
    apply unfold_locales
    unfolding NFA_full_def
    apply auto
  done
  sublocale TA_adder_well_formed2 \<subseteq> TA_plus_NFA:
    TA_adder_well_formed "TA_plus_NFA add_\<A>1 add_\<A>2"
  apply intro_locales
  unfolding TA_adder_well_formed_axioms_def
  apply clarify
  proof -
    fix \<A>0 n0 e0 \<A>2 n2 e2
    assume a: "(\<A>2, n2, e2) = TA_plus_NFA add_\<A>1 add_\<A>2 (\<A>0, n0, e0)"
    assume n: "NFA \<A>0" 
    obtain \<A>1 n1 e1 where b1: "(\<A>1, n1, e1) = add_\<A>1 (\<A>0, n0, e0)"
      using PairE by metis
    have b2: "(\<A>2, n2, e2) = add_\<A>2 (\<A>1, n1, e1)" using a b1 by simp
    show "NFA \<A>2"
      using n
      using add_\<A>1.preserves_well_formed[OF b1]
      using add_\<A>2.preserves_well_formed[OF b2]
      by simp
  qed
  sublocale TA_adder_well_formed2 \<subseteq> TA_times_NFA:
    TA_adder_well_formed "TA_times_NFA add_\<A>1 add_\<A>2"
  apply intro_locales
  unfolding TA_adder_well_formed_axioms_def
  apply clarify
  proof -
    fix \<A>0 n0 e0 \<A>5 n5 e5
    assume a: "NFA \<A>0" "(\<A>5, n5, e5) = TA_times_NFA add_\<A>1 add_\<A>2 (\<A>0, n0, e0)"

    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>2 n2 e2 where b1: "(\<A>2, n2, e2) = add_\<A>1 (?\<A>1, n0, False)"
      using PairE by metis
    have r1': "NFA ?\<A>1" using a(1) unfolding NFA_full_def by simp
    note r1 = add_\<A>1.preserves_well_formed[OF b1 r1']

    let ?\<A>3 = "\<lparr> \<Q> = \<Q> \<A>2, \<Sigma> = \<Sigma> \<A>2, \<Delta> = \<Delta> \<A>2, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>4 n4 e4 where b2: "(\<A>4, n4, e4) = add_\<A>2 (?\<A>3, n2, False)"
      using PairE by metis
    have r2': "NFA ?\<A>3" using r1 unfolding NFA_full_def by simp
    note r2 = add_\<A>2.preserves_well_formed[OF b2 r2']

    have 1: "\<A>5 = \<lparr> \<Q> = \<Q> \<A>4, \<Sigma> = \<Sigma> \<A>4, \<Delta> = TA_bridges (\<F> \<A>2) (\<I> \<A>4) (\<Delta> \<A>4) (\<Delta> \<A>4),
      \<I> = \<I> \<A>0 \<union> \<I> \<A>2, \<F> = \<F> \<A>0 \<union> (if e4 then \<F> \<A>2 \<union> \<F> \<A>4 else \<F> \<A>4) \<rparr>" 
      using a(2) by (simp add: b1[symmetric] b2[symmetric])

    interpret \<A>0: NFA \<A>0 by fact
    interpret \<A>2: NFA \<A>2 by fact
    interpret \<A>4: NFA \<A>4 by fact

    show "NFA \<A>5"
    proof (default, unfold 1 SemiAutomaton.simps NFA.simps)
      case (goal1 p a q)
      show ?case
        using goal1 \<A>2.\<F>_consistent \<A>4.\<Delta>_subset add_\<A>2.inc(1)[OF b2]
        unfolding TA_bridges_def
        by auto
    next
      case (goal2)
      show ?case
        using \<A>0.\<I>_consistent \<A>2.\<I>_consistent add_\<A>1.inc(1)[OF b1] add_\<A>2.inc(1)[OF b2]
        by auto
    next
      case (goal3)
      show ?case
        using \<A>0.\<F>_consistent \<A>2.\<F>_consistent \<A>4.\<F>_consistent
        using add_\<A>1.inc(1)[OF b1] add_\<A>2.inc(1)[OF b2]
        by auto
    next
      case (goal4)
      show ?case using \<A>4.finite_\<Q> by this
    next
      case (goal5)
      show ?case
      proof (rule finite_subset)
        show "TA_bridges (\<F> \<A>2) (\<I> \<A>4) (\<Delta> \<A>4) (\<Delta> \<A>4) \<subseteq> \<Q> \<A>4 \<times> \<Sigma> \<A>4 \<times> \<Q> \<A>4"
          unfolding TA_bridges_def
          using \<A>4.\<Delta>_subset \<A>2.\<F>_consistent add_\<A>2.inc(1)[OF b2]
          by auto
        show "finite (\<Q> \<A>4 \<times> \<Sigma> \<A>4 \<times> \<Q> \<A>4)" using \<A>4.finite_\<Q> \<A>4.finite_\<Sigma> by auto
      qed
    next
      case (goal6)
      show ?case using \<A>4.finite_\<Sigma> by this
    qed
  qed
  sublocale TA_adder_well_formed1 \<subseteq> TA_transcl_NFA:
    TA_adder_well_formed "TA_transcl_NFA add_\<A>"
  apply intro_locales
  unfolding TA_adder_well_formed_axioms_def
  apply clarify
  proof -
    fix \<A>0 n0 e0 \<A>3 n3 e3
    assume a: "NFA \<A>0" "(\<A>3, n3, e3) = TA_transcl_NFA add_\<A> (\<A>0, n0, e0)"

    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>2 n2 e2 where b1: "(\<A>2, n2, e2) = add_\<A> (?\<A>1, n0, False)"
      using PairE by metis
    have r1': "NFA ?\<A>1" using a(1) unfolding NFA_full_def by simp
    note r1 = add_\<A>.preserves_well_formed[OF b1 r1']

    have 1: "\<A>3 = \<lparr> \<Q> = \<Q> \<A>2, \<Sigma> = \<Sigma> \<A>2, \<Delta> = TA_bridges (\<F> \<A>2) (\<I> \<A>2) (\<Delta> \<A>2) (\<Delta> \<A>2),
      \<I> = \<I> \<A>0 \<union> \<I> \<A>2, \<F> = \<F> \<A>0 \<union> \<F> \<A>2 \<rparr>"
      using a(2) by (simp add: b1[symmetric])


    interpret \<A>0: NFA \<A>0 by fact
    interpret \<A>2: NFA \<A>2 by fact

    show "NFA \<A>3"
    proof (default, unfold 1 SemiAutomaton.simps NFA.simps)
      case (goal1 p a q)
      show ?case
        using goal1 \<A>2.\<F>_consistent \<A>2.\<Delta>_subset add_\<A>.inc(1)[OF b1]
        unfolding TA_bridges_def
        by auto
    next
      case (goal2)
      show ?case using \<A>0.\<I>_consistent \<A>2.\<I>_consistent add_\<A>.inc(1)[OF b1] by auto
    next
      case (goal3)
      show ?case using \<A>0.\<F>_consistent \<A>2.\<F>_consistent using add_\<A>.inc(1)[OF b1] by auto
    next
      case (goal4)
      show ?case using \<A>2.finite_\<Q> by this
    next
      case (goal5)
      show ?case
      proof (rule finite_subset)
        show "TA_bridges (\<F> \<A>2) (\<I> \<A>2) (\<Delta> \<A>2) (\<Delta> \<A>2) \<subseteq> \<Q> \<A>2 \<times> \<Sigma> \<A>2 \<times> \<Q> \<A>2"
          unfolding TA_bridges_def using \<A>2.\<Delta>_subset \<A>2.\<F>_consistent by auto
        show "finite (\<Q> \<A>2 \<times> \<Sigma> \<A>2 \<times> \<Q> \<A>2)" using \<A>2.finite_\<Q> \<A>2.finite_\<Sigma> by auto
      qed
    next
      case (goal6)
      show ?case using \<A>2.finite_\<Sigma> by this
    qed
  qed
  sublocale TA_adder_well_formed1 \<subseteq> TA_reflcl_NFA:
    TA_adder_well_formed "TA_reflcl_NFA add_\<A>"
  apply intro_locales
  unfolding TA_adder_well_formed_axioms_def
  apply clarify
  proof -
    fix \<A>0 n0 e0 \<A>4 n4 e4
    assume a: "NFA \<A>0" "(\<A>4, n4, e4) = TA_reflcl_NFA add_\<A> (\<A>0, n0, e0)"

    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    obtain \<A>2 n2 e2 where b1: "(\<A>2, n2, e2) = add_\<A> (?\<A>1, n0, False)"
      using PairE by metis
    have r1': "NFA ?\<A>1" using a(1) unfolding NFA_full_def by simp
    note r1 = add_\<A>.preserves_well_formed[OF b1 r1']

    show "NFA \<A>4"
      using a r1 add_\<A>.inc[OF b1]
      apply (cases e2)
      apply (auto simp: b1[symmetric] NFA_full_def)
    done
  qed
  sublocale TA_adder_well_formed1 \<subseteq> TA_star_NFA:
    TA_adder_well_formed "TA_star_NFA add_\<A>"
  apply intro_locales
  unfolding TA_adder_well_formed_axioms_def
  apply clarify
  proof -
    fix \<A>0 n0 e0 \<A>1 n1 e1
    assume a: "NFA \<A>0" "(\<A>1, n1, e1) = TA_star_NFA add_\<A> (\<A>0, n0, e0)"
    have b1: "(\<A>1, n1, e1) = TA_reflcl_NFA (TA_transcl_NFA add_\<A>) (\<A>0, n0, e0)"
      using a by auto
    interpret TA_transcl_NFA: TA_adder_well_formed1 "TA_transcl_NFA add_\<A>"
      by unfold_locales
    show "NFA \<A>1"
      using TA_transcl_NFA.TA_reflcl_NFA.preserves_well_formed[OF b1 a(1)]
      by this
  qed

  interpretation TA_translate_NFA: TA_adder_well_formed "TA_translate_NFA r"
  proof (induct r)
    case Zero thus ?case by (simp, unfold_locales)
    case One thus ?case by (simp, unfold_locales)
    case Atom thus ?case by (simp, unfold_locales)
  next
    case (Plus r1 r2)
    interpret TA_adder_well_formed "TA_translate_NFA r1" using Plus(1) by this
    interpret TA_adder_well_formed "TA_translate_NFA r2" using Plus(2) by this
    interpret TA_adder_well_formed2 "TA_translate_NFA r1" "TA_translate_NFA r2"
      by intro_locales
    show ?case by (simp, intro_locales)
  next
    case (Times r1 r2)
    interpret TA_adder_well_formed "TA_translate_NFA r1" using Times(1) by this
    interpret TA_adder_well_formed "TA_translate_NFA r2" using Times(2) by this
    interpret TA_adder_well_formed2 "TA_translate_NFA r1" "TA_translate_NFA r2"
      by intro_locales
    show ?case by (simp, intro_locales)
  next
    case (Star r)
    interpret TA_adder_well_formed "TA_translate_NFA r" using Star by this
    interpret TA_adder_well_formed1 "TA_translate_NFA r"
      by intro_locales
    show ?case by (simp, intro_locales)
  qed

  locale TA_invar =
    fixes \<A> :: "(nat, 'a) NFA_rec"
    fixes n :: "nat"
    fixes e :: "bool"
    assumes well_formed: "NFA \<A>"
    assumes states: "\<Q> \<A> = {0 ..< n}"
    assumes empty: "e \<longleftrightarrow> \<I> \<A> \<inter> \<F> \<A> \<noteq> {}"

  sublocale TA_invar \<subseteq> NFA_\<A>!: NFA \<A> using well_formed by this

  locale TA_adder =
    TA_adder_well_formed add_\<A>
    for add_\<A> :: "'a TA_adder"
    +
    fixes \<A> :: "(nat, 'a) NFA_rec"
    fixes n :: "nat"
    assumes well_formed: "NFA \<A>"
    assumes states: "\<Q> \<A> = {n ..< n + card (\<Q> \<A>)}"
    assumes adds: "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow>
      TA_invar \<A>0 n0 e0 \<Longrightarrow> n0 = n \<Longrightarrow> \<A>1 = TM_plus_NFA \<A>0 \<A>"
    assumes invar: "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0) \<Longrightarrow>
      TA_invar \<A>0 n0 e0 \<Longrightarrow> n0 = n \<Longrightarrow> TA_invar \<A>1 n1 e1"
  begin

    lemma add_step_n:
      assumes b: "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0)"
      assumes TA_invar0: "TA_invar \<A>0 n0 e0"
      assumes n0: "n0 = n"
      shows "n1 = n + card (\<Q> \<A>)"
    proof -
      interpret TA_invar0!: TA_invar \<A>0 n0 e0 using TA_invar0 by this
      interpret TA_invar1!: TA_invar \<A>1 n1 e1 using invar[OF b TA_invar0 n0] by this

      have "\<Q> \<A>1 = {0 ..< n + card (\<Q> \<A>)}"
        using TA_invar0.states states
        using TM_plus_NFA_states[of \<A>0 _ _ \<A>]
        using adds[OF b TA_invar0.TA_invar_axioms n0]
        unfolding n0
        by auto
      thus "n1 = n + card (\<Q> \<A>)"
        using TA_invar1.states
        by force
    qed
    lemma add_step_e:
      assumes b: "(\<A>1, n1, e1) = add_\<A> (\<A>0, n0, e0)"
      assumes TA_invar0: "TA_invar \<A>0 n0 e0"
      assumes n0: "n0 = n"
      shows "e1 \<longleftrightarrow> e0 \<or> \<I> \<A> \<inter> \<F> \<A> \<noteq> {}"
    proof -
      interpret TA_invar0!: TA_invar \<A>0 n0 e0 using TA_invar0 by this
      interpret TA_invar1!: TA_invar \<A>1 n1 e1 using invar[OF b TA_invar0 n0] by this
      interpret disjoint_NFAs_\<A>0_\<A>!: disjoint_NFAs \<A>0 \<A>
        apply (rule disjoint_NFAs.intro)
        apply (rule TA_invar0.well_formed)
        apply (rule well_formed)
        unfolding disjoint_NFAs_axioms_def
        using TA_invar0.states states
        unfolding n0
        using ivl_disj_int(11)
        apply blast
      done

      show "e1 \<longleftrightarrow> e0 \<or> \<I> \<A> \<inter> \<F> \<A> \<noteq> {}"
        using TA_invar1.empty
        unfolding adds[OF b TA_invar0.TA_invar_axioms n0]
        unfolding disjoint_NFAs_\<A>0_\<A>.TM_plus_NFA_empty_word[symmetric]
        unfolding TA_invar0.empty[symmetric]
        by this
    qed

  end

  locale TA_adder1 =
    TA_adder_well_formed1 add_\<A>' +
    add_\<A>'!: TA_adder "add_\<A>'" "\<A>'" "n"
    for add_\<A>' :: "'a TA_adder" and \<A>' :: "(nat, 'a) NFA_rec"
    and n :: "nat"
  locale TA_adder2 =
    TA_adder_well_formed2 add_\<A>1' add_\<A>2' +
    add_\<A>1'!: TA_adder "add_\<A>1'" "\<A>1'" "n" +
    add_\<A>2'!: TA_adder "add_\<A>2'" "\<A>2'" "n + card (\<Q> \<A>1')"
    for add_\<A>1' :: "'a TA_adder" and \<A>1' :: "(nat, 'a) NFA_rec"
    and add_\<A>2' :: "'a TA_adder" and \<A>2' :: "(nat, 'a) NFA_rec"
    and n :: "nat"

  sublocale TA_adder1 \<subseteq> NFA_\<A>'!: NFA \<A>' using add_\<A>'.well_formed by this
  sublocale TA_adder2 \<subseteq> disjoint_NFAs_\<A>1'_\<A>2'!: disjoint_NFAs \<A>1' \<A>2'
    apply (rule disjoint_NFAs.intro)
    apply (rule add_\<A>1'.well_formed)
    apply (rule add_\<A>2'.well_formed)
    unfolding disjoint_NFAs_axioms_def
    using add_\<A>1'.states add_\<A>2'.states ivl_disj_int(11)
    by blast

  interpretation TA_zero_NFA!: TA_adder "TA_zero_NFA" "TM_zero_NFA" "n"
    apply intro_locales
    unfolding TA_adder_axioms_def
    apply auto
  done
  interpretation TA_one_NFA!: TA_adder "TA_one_NFA" "TM_one_NFA n" "n"
    apply intro_locales
    unfolding TA_adder_axioms_def TA_invar_def NFA_full_def
    apply force
  done
  interpretation TA_atom_NFA!: TA_adder "TA_atom_NFA a" "TM_atom_NFA n a (Suc n)" "n"
    apply intro_locales
    unfolding TA_adder_axioms_def TA_invar_def NFA_full_def TM_plus_NFA_def
    apply force
  done
  sublocale TA_adder2 \<subseteq> TA_plus_NFA!:
    TA_adder "TA_plus_NFA add_\<A>1' add_\<A>2'" "TM_plus_NFA \<A>1' \<A>2'"
  apply intro_locales
  unfolding TA_adder_axioms_def
  proof (intro allI impI conjI)
    show well_formed: "NFA (TM_plus_NFA \<A>1' \<A>2')"
      using TM_plus_NFA_well_formed[OF add_\<A>1'.well_formed add_\<A>2'.well_formed]
      by this
    show states: "\<Q> (TM_plus_NFA \<A>1' \<A>2') = {n ..< n + card (\<Q> (TM_plus_NFA \<A>1' \<A>2'))}"
      using disjoint_NFAs_\<A>1'_\<A>2'.disjoint_\<Q>
      using TM_plus_NFA_states[OF add_\<A>1'.states add_\<A>2'.states]
      by auto

    fix \<A>0 n0 e0 \<A>3 n3 e3
    assume a: "(\<A>3, n3, e3) = TA_plus_NFA add_\<A>1' add_\<A>2' (\<A>0, n0, e0)"
    assume b: "TA_invar \<A>0 n0 e0"
    assume n0: "n0 = n"
    interpret TA_invar0!: TA_invar \<A>0 n0 e0 using b by this

    obtain \<A>1 n1 e1 where b1: "(\<A>1, n1, e1) = add_\<A>1' (\<A>0, n0, e0)"
      using PairE by metis
    interpret TA_invar1!: TA_invar \<A>1 n1 e1
      using add_\<A>1'.invar[OF b1 TA_invar0.TA_invar_axioms n0]
      by this
    note n1 = add_\<A>1'.add_step_n[OF b1 TA_invar0.TA_invar_axioms n0]

    obtain \<A>2 n2 e2 where b2: "(\<A>2, n2, e2) = add_\<A>2' (\<A>1, n1, e1)"
      using PairE by metis
    interpret TA_invar2!: TA_invar \<A>2 n2 e2
      using add_\<A>2'.invar[OF b2 TA_invar1.TA_invar_axioms n1]
      by this

    interpret disjoint_NFAs_\<A>0_\<A>1'_\<A>2'!: disjoint_NFAs "\<A>0" "TM_plus_NFA \<A>1' \<A>2'"
      apply (rule disjoint_NFAs.intro)
      apply (rule TA_invar0.well_formed)
      apply (rule well_formed)
      unfolding disjoint_NFAs_axioms_def
      using TA_invar0.states states
      unfolding n0
      using ivl_disj_int(11)
      apply blast
    done

    have p:
      "\<A>3 = \<A>2"
      "n3 = n2"
      "e3 = e2"
      using a
      by (auto simp: b1[symmetric] b2[symmetric])

    show adds: "\<A>3 = TM_plus_NFA \<A>0 (TM_plus_NFA \<A>1' \<A>2')"
      unfolding p
      using add_\<A>1'.adds[OF b1 TA_invar0.TA_invar_axioms n0]
      using add_\<A>2'.adds[OF b2 TA_invar1.TA_invar_axioms n1]
      using TM_plus_NFA_associative
      by auto

    show invar: "TA_invar \<A>3 n3 e3"
    unfolding TA_invar_def
    proof (intro conjI)
      show "NFA \<A>3"
        using TM_plus_NFA_well_formed[OF TA_invar0.NFA_\<A>.NFA_axioms well_formed]
        unfolding adds
        by this
      show "\<Q> \<A>3 = {0 ..< n3}"
        unfolding p
        using TA_invar2.states
        by this
      show "e3 \<longleftrightarrow> \<I> \<A>3 \<inter> \<F> \<A>3 \<noteq> {}"
        unfolding adds
        unfolding p
        unfolding disjoint_NFAs_\<A>0_\<A>1'_\<A>2'.TM_plus_NFA_empty_word[symmetric]
        unfolding disjoint_NFAs_\<A>1'_\<A>2'.TM_plus_NFA_empty_word[symmetric]
        using TA_invar0.empty
        using add_\<A>1'.add_step_e[OF b1 TA_invar0.TA_invar_axioms n0]
        using add_\<A>2'.add_step_e[OF b2 TA_invar1.TA_invar_axioms n1]
        by auto
    qed
  qed
  sublocale TA_adder2 \<subseteq> TA_times_NFA!:
    TA_adder "TA_times_NFA add_\<A>1' add_\<A>2'" "TM_times_NFA \<A>1' \<A>2'"
  apply intro_locales
  unfolding TA_adder_axioms_def
  proof (intro allI impI conjI)
    show well_formed: "NFA (TM_times_NFA \<A>1' \<A>2')"
      using TM_times_NFA_well_formed[OF add_\<A>1'.well_formed add_\<A>2'.well_formed]
      by this
    show states: "\<Q> (TM_times_NFA \<A>1' \<A>2') = {n ..< n + card (\<Q> (TM_times_NFA \<A>1' \<A>2'))}"
      using disjoint_NFAs_\<A>1'_\<A>2'.disjoint_\<Q>
      using TM_times_NFA_states[OF add_\<A>1'.states add_\<A>2'.states]
      by auto

    fix \<A>0 n0 e0 \<A>5 n5 e5
    assume a: "(\<A>5, n5, e5) = TA_times_NFA add_\<A>1' add_\<A>2' (\<A>0, n0, e0)"
    assume b: "TA_invar \<A>0 n0 e0"
    assume n0: "n0 = n"
    interpret TA_invar0!: TA_invar \<A>0 n0 e0 using b by this

    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    interpret TA_invar1!: TA_invar ?\<A>1 n0 False
      using TA_invar0.well_formed TA_invar0.states
      unfolding TA_invar_def NFA_full_def
      by auto

    obtain \<A>2 n2 e2 where b2: "(\<A>2, n2, e2) = add_\<A>1' (?\<A>1, n0, False)"
      using PairE by metis
    interpret TA_invar2!: TA_invar \<A>2 n2 e2
      using add_\<A>1'.invar[OF b2 TA_invar1.TA_invar_axioms n0]
      by this
    note n2 = add_\<A>1'.add_step_n[OF b2 TA_invar1.TA_invar_axioms n0]

    let ?\<A>3 = "\<lparr> \<Q> = \<Q> \<A>2, \<Sigma> = \<Sigma> \<A>2, \<Delta> = \<Delta> \<A>2, \<I> = {}, \<F> = {} \<rparr>"
    interpret TA_invar3!: TA_invar ?\<A>3 n2 False
      using TA_invar2.well_formed TA_invar2.states
      unfolding TA_invar_def NFA_full_def
      by auto

    obtain \<A>4 n4 e4 where b4: "(\<A>4, n4, e4) = add_\<A>2' (?\<A>3, n2, False)"
      using PairE by metis
    interpret TA_invar4!: TA_invar \<A>4 n4 e4
      using add_\<A>2'.invar[OF b4 TA_invar3.TA_invar_axioms n2]
      by this
    note n4 = add_\<A>2'.add_step_n[OF b4 TA_invar3.TA_invar_axioms n2]

    interpret disjoint_NFAs_\<A>0_\<A>1'_\<A>2'!: disjoint_NFAs "\<A>0" "TM_times_NFA \<A>1' \<A>2'"
      apply (rule disjoint_NFAs.intro)
      apply (rule TA_invar0.well_formed)
      apply (rule well_formed)
      unfolding disjoint_NFAs_axioms_def
      using TA_invar0.states states
      unfolding n0
      using ivl_disj_int(11)
      apply blast
    done

    have f': "\<And> p a q. (p, a, q) \<in> \<Delta> \<A>0 \<union> \<Delta> \<A>1' \<Longrightarrow> p \<notin> \<I> \<A>2'"
    proof -
      fix p a q
      assume "(p, a, q) \<in> \<Delta> \<A>0 \<union> \<Delta> \<A>1'"
      thus "p \<notin> \<I> \<A>2'"
      proof
        assume "(p, a, q) \<in> \<Delta> \<A>0"
        thus "p \<notin> \<I> \<A>2'"
          using TA_invar0.NFA_\<A>.\<Delta>_consistent
          using disjoint_NFAs_\<A>1'_\<A>2'.NFA_\<A>2.\<I>_consistent
          using disjoint_NFAs_\<A>0_\<A>1'_\<A>2'.disjoint_\<Q>
          by fastforce
      next
        assume "(p, a, q) \<in> \<Delta> \<A>1'"
        thus "p \<notin> \<I> \<A>2'"
          using disjoint_NFAs_\<A>1'_\<A>2'.NFA_\<A>1.\<Delta>_consistent
          using disjoint_NFAs_\<A>1'_\<A>2'.NFA_\<A>2.\<I>_consistent
          using disjoint_NFAs_\<A>1'_\<A>2'.disjoint_\<Q>
          by fastforce
      qed
    qed
    have f: "TM_bridges (\<F> \<A>1') (\<I> \<A>2') (\<Delta> \<A>0 \<union> \<Delta> \<A>1' \<union> \<Delta> \<A>2') =
      TM_bridges (\<F> \<A>1') (\<I> \<A>2') (\<Delta> \<A>2')"
      unfolding TM_bridges_def
      using f'
      by blast

    have p:
      "\<A>5 =
      \<lparr>
        \<Q> = \<Q> \<A>4,
        \<Sigma> = \<Sigma> \<A>4,
        \<Delta> = TA_bridges (\<F> \<A>2) (\<I> \<A>4) (\<Delta> \<A>4) (\<Delta> \<A>4),
        \<I> = \<I> \<A>0 \<union> \<I> \<A>2,
        \<F> = \<F> \<A>0 \<union> (if e4 then \<F> \<A>2 \<union> \<F> \<A>4 else \<F> \<A>4)
      \<rparr>"
      "n5 = n4"
      "e5 \<longleftrightarrow> e0 \<or> e2 \<and> e4"
      using a
      by (auto simp: b2[symmetric] b4[symmetric])

    show adds: "\<A>5 = TM_plus_NFA \<A>0 (TM_times_NFA \<A>1' \<A>2')"
      unfolding p TA_bridges_eq_TM TM_plus_NFA_def
      using add_\<A>1'.adds[OF b2 TA_invar1.TA_invar_axioms n0]
      using add_\<A>2'.adds[OF b4 TA_invar3.TA_invar_axioms n2]
      using add_\<A>2'.add_step_e[OF b4 TA_invar3.TA_invar_axioms n2]
      using f
      by auto

    show invar: "TA_invar \<A>5 n5 e5"
    unfolding TA_invar_def
    proof (intro conjI)
      show "NFA \<A>5"
        using TM_plus_NFA_well_formed[OF TA_invar0.NFA_\<A>.NFA_axioms well_formed]
        unfolding adds
        by this
      show "\<Q> \<A>5 = {0 ..< n5}"
        unfolding p
        using TA_invar4.states
        by simp
      show "e5 \<longleftrightarrow> \<I> \<A>5 \<inter> \<F> \<A>5 \<noteq> {}"
        unfolding adds
        unfolding p
        unfolding disjoint_NFAs_\<A>0_\<A>1'_\<A>2'.TM_plus_NFA_empty_word[symmetric]
        unfolding disjoint_NFAs_\<A>1'_\<A>2'.TM_times_NFA_empty_word[symmetric]
        using TA_invar0.empty
        using add_\<A>1'.add_step_e[OF b2 TA_invar1.TA_invar_axioms n0]
        using add_\<A>2'.add_step_e[OF b4 TA_invar3.TA_invar_axioms n2]
        by simp
    qed
  qed
  sublocale TA_adder1 \<subseteq> TA_transcl_NFA!:
    TA_adder "TA_transcl_NFA add_\<A>'" "TM_transcl_NFA \<A>'"
  apply intro_locales
  unfolding TA_adder_axioms_def
  proof (intro allI impI conjI)
    show well_formed: "NFA (TM_transcl_NFA \<A>')"
      using NFA_\<A>'.TM_transcl_NFA_well_formed
      by this
    show states: "\<Q> (TM_transcl_NFA \<A>') = {n ..< n + card (\<Q> (TM_transcl_NFA \<A>'))}"
      using add_\<A>'.states
      by simp

    fix \<A>0 n0 e0 \<A>3 n3 e3
    assume a: "(\<A>3, n3, e3) = TA_transcl_NFA add_\<A>' (\<A>0, n0, e0)"
    assume b: "TA_invar \<A>0 n0 e0"
    assume n0: "n0 = n"
    interpret TA_invar0!: TA_invar \<A>0 n0 e0 using b by this

    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    interpret TA_invar1!: TA_invar ?\<A>1 n0 False
      using TA_invar0.well_formed TA_invar0.states
      unfolding TA_invar_def NFA_full_def
      by auto

    obtain \<A>2 n2 e2 where b2: "(\<A>2, n2, e2) = add_\<A>' (?\<A>1, n0, False)"
      using PairE by metis
    interpret TA_invar2!: TA_invar \<A>2 n2 e2
      using add_\<A>'.invar[OF b2 TA_invar1.TA_invar_axioms n0]
      by this
    note n2 = add_\<A>'.add_step_n[OF b2 TA_invar1.TA_invar_axioms n0]

    interpret disjoint_NFAs_\<A>0_\<A>1'_\<A>2'!: disjoint_NFAs "\<A>0" "TM_transcl_NFA \<A>'"
      apply (rule disjoint_NFAs.intro)
      apply (rule TA_invar0.well_formed)
      apply (rule well_formed)
      unfolding disjoint_NFAs_axioms_def
      using TA_invar0.states states
      unfolding n0
      using ivl_disj_int(11)
      apply blast
    done

    have f': "\<And> p a q. (p, a, q) \<in> \<Delta> \<A>0 \<Longrightarrow> p \<notin> \<I> \<A>'"
      using TA_invar0.NFA_\<A>.\<Delta>_consistent
      using NFA_\<A>'.\<I>_consistent
      using disjoint_NFAs_\<A>0_\<A>1'_\<A>2'.disjoint_\<Q>
      by fastforce
    have f: "TM_bridges (\<F> \<A>') (\<I> \<A>') (\<Delta> \<A>0 \<union> \<Delta> \<A>') =
      TM_bridges (\<F> \<A>') (\<I> \<A>') (\<Delta> \<A>')"
      unfolding TM_bridges_def
      using f'
      by blast

    have p:
      "\<A>3 =
      \<lparr>
        \<Q> = \<Q> \<A>2,
        \<Sigma> = \<Sigma> \<A>2,
        \<Delta> = TA_bridges (\<F> \<A>2) (\<I> \<A>2) (\<Delta> \<A>2) (\<Delta> \<A>2),
        \<I> = \<I> \<A>0 \<union> \<I> \<A>2,
        \<F> = \<F> \<A>0 \<union> \<F> \<A>2
      \<rparr>"
      "n3 = n2"
      "e3 \<longleftrightarrow> e0 \<or> e2"
      using a
      by (auto simp: b2[symmetric])

    show adds: "\<A>3 = TM_plus_NFA \<A>0 (TM_transcl_NFA \<A>')"
      unfolding p TA_bridges_eq_TM TM_plus_NFA_def
      using add_\<A>'.adds[OF b2 TA_invar1.TA_invar_axioms n0]
      using f add_\<A>'.add_step_e[OF b2 TA_invar1.TA_invar_axioms n0]
      by auto

    show invar: "TA_invar \<A>3 n3 e3"
    unfolding TA_invar_def
    proof (intro conjI)
      show "NFA \<A>3"
        using TM_plus_NFA_well_formed[OF TA_invar0.NFA_\<A>.NFA_axioms well_formed]
        unfolding adds
        by this
      show "\<Q> \<A>3 = {0 ..< n3}"
        unfolding p
        using TA_invar2.states
        by simp
      show "e3 \<longleftrightarrow> \<I> \<A>3 \<inter> \<F> \<A>3 \<noteq> {}"
        unfolding adds
        unfolding p
        unfolding disjoint_NFAs_\<A>0_\<A>1'_\<A>2'.TM_plus_NFA_empty_word[symmetric]
        unfolding TM_transcl_NFA_empty_word[symmetric]
        using TA_invar0.empty
        using add_\<A>'.add_step_e[OF b2 TA_invar1.TA_invar_axioms n0]
        by simp
    qed
  qed
  sublocale TA_adder1 \<subseteq> TA_reflcl_NFA!:
    TA_adder "TA_reflcl_NFA add_\<A>'" "TM_reflcl_NFA \<A>' (n + card (\<Q> \<A>'))" "n"
  apply intro_locales
  unfolding TA_adder_axioms_def
  proof (intro allI impI conjI)
    show well_formed: "NFA (TM_reflcl_NFA \<A>' (n + card (\<Q> \<A>')))"
      using NFA_\<A>'.TM_reflcl_NFA_well_formed
      by this
    show states: "\<Q> (TM_reflcl_NFA \<A>' (n + card (\<Q> \<A>'))) =
      {n ..< n + card (\<Q> (TM_reflcl_NFA \<A>' (n + card (\<Q> \<A>'))))}"
      using insert_interval[of "n" "n + card (\<Q> \<A>')"]
      using add_\<A>'.states
      by (cases "\<I> \<A>' \<inter> \<F> \<A>' = {}", auto)

    fix \<A>0 n0 e0 \<A>4 n4 e4
    assume a: "(\<A>4, n4, e4) = TA_reflcl_NFA add_\<A>' (\<A>0, n0, e0)"
    assume b: "TA_invar \<A>0 n0 e0"
    assume n0: "n0 = n"
    interpret TA_invar0!: TA_invar \<A>0 n0 e0 using b by this

    let ?\<A>1 = "\<lparr> \<Q> = \<Q> \<A>0, \<Sigma> = \<Sigma> \<A>0, \<Delta> = \<Delta> \<A>0, \<I> = {}, \<F> = {} \<rparr>"
    interpret TA_invar1!: TA_invar ?\<A>1 n0 False
      using TA_invar0.well_formed TA_invar0.states
      unfolding TA_invar_def NFA_full_def
      by auto

    obtain \<A>2 n2 e2 where b2: "(\<A>2, n2, e2) = add_\<A>' (?\<A>1, n0, False)"
      using PairE by metis
    interpret TA_invar2!: TA_invar \<A>2 n2 e2
      using add_\<A>'.invar[OF b2 TA_invar1.TA_invar_axioms n0]
      by this
    note n2 = add_\<A>'.add_step_n[OF b2 TA_invar1.TA_invar_axioms n0]

    def add_reflcl \<equiv> "if e2 then TA_zero_NFA else TA_one_NFA :: 'a TA_adder" 
    def reflcl \<equiv> "if e2
      then TM_zero_NFA
      else TM_one_NFA (n + card (\<Q> \<A>')) :: (nat, 'a) NFA_rec"
    interpret add_reflcl!: TA_adder "add_reflcl" "reflcl" "n + card (\<Q> \<A>')"
      unfolding add_reflcl_def reflcl_def
      by (cases e2, simp_all, intro_locales)

    obtain \<A>3 n3 e3 where b3: "(\<A>3, n3, e3) = add_reflcl (\<A>2, n2, e2)"
      using PairE by metis
    interpret TA_invar3!: TA_invar \<A>3 n3 e3
      using add_reflcl.invar[OF b3 TA_invar2.TA_invar_axioms n2]
      by this
    note n3 = add_reflcl.add_step_n[OF b3 TA_invar2.TA_invar_axioms n2]

    interpret disjoint_NFAs_\<A>0_\<A>1'_\<A>2'!:
      disjoint_NFAs "\<A>0" "TM_reflcl_NFA \<A>' (n + card (\<Q> \<A>'))"
      apply (rule disjoint_NFAs.intro)
      apply (rule TA_invar0.well_formed)
      apply (rule well_formed)
      unfolding disjoint_NFAs_axioms_def
      using TA_invar0.states states
      unfolding n0
      using ivl_disj_int(11)
      apply blast
    done

    have p:
      "\<A>4 =
      \<lparr>
        \<Q> = \<Q> \<A>3,
        \<Sigma> = \<Sigma> \<A>3,
        \<Delta> = \<Delta> \<A>3,
        \<I> = \<I> \<A>0 \<union> \<I> \<A>3,
        \<F> = \<F> \<A>0 \<union> \<F> \<A>3
      \<rparr>"
      "n4 = n3"
      "e4 \<longleftrightarrow> e0 \<or> e3"
      apply (cases e2)
      using a add_reflcl_def b3
      apply (auto simp: b2[symmetric])[2]

      apply (cases e2)
      using a add_reflcl_def b3
      apply (auto simp: b2[symmetric])[2]
      
      apply (cases e2)
      using a add_reflcl_def b3
      apply (auto simp: b2[symmetric])[2]
    done

    show adds: "\<A>4 = TM_plus_NFA \<A>0 (TM_reflcl_NFA \<A>' (n + card (\<Q> \<A>')))"
      unfolding p TM_plus_NFA_def
      using add_\<A>'.adds[OF b2 TA_invar1.TA_invar_axioms n0]
      using add_reflcl.adds[OF b3 TA_invar2.TA_invar_axioms n2]
      unfolding reflcl_def
      using add_\<A>'.add_step_e[OF b2 TA_invar1.TA_invar_axioms n0]
      by auto

    show invar: "TA_invar \<A>4 n4 e4"
    unfolding TA_invar_def
    proof (intro conjI)
      show "NFA \<A>4"
        using TM_plus_NFA_well_formed[OF TA_invar0.NFA_\<A>.NFA_axioms well_formed]
        unfolding adds
        by this
      show "\<Q> \<A>4 = {0 ..< n4}"
        unfolding p
        using TA_invar3.states
        by simp
      show "e4 \<longleftrightarrow> \<I> \<A>4 \<inter> \<F> \<A>4 \<noteq> {}"
        unfolding adds
        unfolding p
        unfolding disjoint_NFAs_\<A>0_\<A>1'_\<A>2'.TM_plus_NFA_empty_word[symmetric]
        using TM_reflcl_NFA_empty_word[of "\<A>'" "n + card (\<Q> \<A>')"]
        using add_reflcl.add_step_e[OF b3 TA_invar2.TA_invar_axioms n2]
        unfolding reflcl_def
        by simp
    qed
  qed
  sublocale TA_adder1 \<subseteq> TA_star_NFA!:
    TA_adder "TA_star_NFA add_\<A>'" "TM_star_NFA \<A>' (n + card (\<Q> \<A>'))" "n"
  proof -
    interpret a: TA_adder1 "TA_transcl_NFA add_\<A>'" "TM_transcl_NFA \<A>'" "n"
      by intro_locales
    have b: "TA_star_NFA add_\<A>' = TA_reflcl_NFA (TA_transcl_NFA add_\<A>')"
      using TA_star_NFA.simps
      by auto
    show "TA_adder (TA_star_NFA add_\<A>') (TM_star_NFA \<A>' (n + card (\<Q> \<A>'))) n"
      using a.TA_reflcl_NFA.TA_adder_axioms b
      by auto
  qed

  interpretation TA_translate_NFA!:
    TA_adder "TA_translate_NFA r" "TM_translate_NFA r n" "n"
  proof (induct r arbitrary: n)
    case Zero
    show ?case by (simp, intro_locales)
  next
    case One
    show ?case by (simp, intro_locales)
  next
    case (Atom a)
    show ?case by (simp, intro_locales)
  next
    case (Plus r1 r2)
    let ?n1 = "n"
    let ?n2 = "n + card (\<Q> (TM_translate_NFA r1 n))"
    interpret TA_adder2
      "TA_translate_NFA r1" "TM_translate_NFA r1 ?n1"
      "TA_translate_NFA r2" "TM_translate_NFA r2 ?n2"
      "n"
      apply (rule TA_adder2.intro)
      apply intro_locales[1]
      apply (rule Plus(1))
      apply (rule Plus(2))
    done
    show ?case by (simp add: Let_def, intro_locales)
  next
    case (Times r1 r2)
    let ?n1 = "n"
    let ?n2 = "n + card (\<Q> (TM_translate_NFA r1 n))"
    interpret TA_adder2
      "TA_translate_NFA r1" "TM_translate_NFA r1 ?n1"
      "TA_translate_NFA r2" "TM_translate_NFA r2 ?n2"
      "n"
      apply (rule TA_adder2.intro)
      apply intro_locales[1]
      apply (rule Times(1))
      apply (rule Times(2))
    done
    show ?case by (simp add: Let_def, intro_locales)
  next
    case (Star r)
    interpret TA_adder1 "TA_translate_NFA r" "TM_translate_NFA r n" "n"
      apply (rule TA_adder1.intro)
      apply intro_locales[1]
      apply (rule Star(1))
    done
    show ?case by (simp add: Let_def del: TM_star_NFA_def, intro_locales)
  qed

end
