theory TA_Implementation
imports
  TA_Correctness
  Thompson_Implementation
begin

  context Thompson_impl
  begin

    definition TA_bridges_impl :: "'q_set \<Rightarrow> 'q_set \<Rightarrow> 'd \<Rightarrow> 'd \<Rightarrow> 'd"
    where "TA_bridges_impl F1 I2 D2 D0 \<equiv>
    (
      let
        add_transitions_from_p2_to_p1 = \<lambda> p1 p2. d_succ_label_it D2
          p2 (\<lambda> _. True) (\<lambda> (a2, q2). d.add p1 a2 q2);
        add_transitions_from_all_p2_to_p1 = \<lambda> p1. s_it I2
          (\<lambda> _. True) (\<lambda> p2. add_transitions_from_p2_to_p1 p1 p2);
        add_transitions_from_all_p2_to_all_p1 = s_it F1
          (\<lambda> _. True) (\<lambda> p1. add_transitions_from_all_p2_to_p1 p1)
      in
        add_transitions_from_all_p2_to_all_p1 D0
    )"

    lemma TA_bridges_impl_correct:
      assumes "s.invar F1" "s.invar I2" "d.invar D2" "d.invar D0"
      shows
        "d.invar (TA_bridges_impl F1 I2 D2 D0)"
        "d.\<alpha> (TA_bridges_impl F1 I2 D2 D0) =
          TA_bridges (s.\<alpha> F1) (s.\<alpha> I2) (d.\<alpha> D2) (d.\<alpha> D0)"
    proof -
      let ?add_transitions_from_p2_to_p1 =
        "\<lambda> p1 p2. d_succ_label_it D2 p2 (\<lambda>_. True) (\<lambda> (a2, q2). d.add p1 a2 q2)"
      let ?add_transitions_from_all_p2_to_p1 =
        "\<lambda> p1. s_it I2 (\<lambda>_. True) (?add_transitions_from_p2_to_p1 p1)"
      let ?add_transitions_from_all_p2_to_all_p1 =
        "s_it F1 (\<lambda>_. True) ?add_transitions_from_all_p2_to_p1"

      note s_it_\<A>1_rule = set_iterator_rule_insert_P[OF iteratei_rule[OF assms(1)]]
      note s_it_\<A>2_rule = set_iterator_rule_insert_P[OF iteratei_rule[OF assms(2)]]
      note d_succ_label_it_\<A>2_rule = set_iterator_rule_insert_P
        [OF lts_succ_label_it_correct[OF assms(3)]]

      have a1: "\<And> p1 p2 d. d.invar d \<Longrightarrow>
        d.invar (?add_transitions_from_p2_to_p1 p1 p2 d)"
        by (rule_tac I = "\<lambda> S d'. d.invar d'"
          in d_succ_label_it_\<A>2_rule, auto simp: d.correct)
      have b1: "\<And> p1 d. d.invar d \<Longrightarrow>
        d.invar (?add_transitions_from_all_p2_to_p1 p1 d)"
        using a1
        by (rule_tac I = "\<lambda> S d'. d.invar d'"
          in s_it_\<A>2_rule, auto simp: d.correct)
      have c1: "\<And> d. d.invar d \<Longrightarrow>
        d.invar (?add_transitions_from_all_p2_to_all_p1 d)"
        using b1
        by (rule_tac I = "\<lambda> S d'. d.invar d'"
          in s_it_\<A>1_rule, auto simp: d.correct)

      have a2: "\<And> p1 p2 d. d.invar d \<Longrightarrow>
        d.\<alpha> (?add_transitions_from_p2_to_p1 p1 p2 d) =
          d.\<alpha> d \<union> {(p1, a2, q2) | a2 q2. (p2, a2, q2) \<in> d.\<alpha> D2}"
        by (rule_tac I = "\<lambda> S d'. d.invar d' \<and> d.\<alpha> d' =
          d.\<alpha> d \<union> {(p1, a2, q2) | a2 q2. (a2, q2) \<in> S}"
          in d_succ_label_it_\<A>2_rule, auto simp: d.correct)
      have b2: "\<And> p1 d. d.invar d \<Longrightarrow>
        d.\<alpha> (?add_transitions_from_all_p2_to_p1 p1 d) =
          d.\<alpha> d \<union> {(p1, a2, q2) | p2 a2 q2. p2 \<in> s.\<alpha> I2 \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
        using a1 a2
        by (rule_tac I = "\<lambda> S d'. d.invar d' \<and> d.\<alpha> d' =
          d.\<alpha> d \<union> {(p1, a2, q2) | p2 a2 q2. p2 \<in> S \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
          in s_it_\<A>2_rule, auto simp: d.correct)
      have c2: "\<And> d. d.invar d \<Longrightarrow>
        d.\<alpha> (?add_transitions_from_all_p2_to_all_p1 d) =
          d.\<alpha> d \<union> {(p1, a2, q2) | p1 p2 a2 q2. p1 \<in> s.\<alpha> F1 \<and>
          p2 \<in> s.\<alpha> I2 \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
        using b1 b2
        by (rule_tac I = "\<lambda> S d'. d.invar d' \<and> d.\<alpha> d' =
          d.\<alpha> d \<union> {(p1, a2, q2) | p1 p2 a2 q2. p1 \<in> S \<and>
          p2 \<in> s.\<alpha> I2 \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
          in s_it_\<A>1_rule, auto simp: d.correct)

      show "d.invar (TA_bridges_impl F1 I2 D2 D0)"
        using c1 assms(4)
        by (auto simp: d.correct TA_bridges_impl_def)
      show "d.\<alpha> (TA_bridges_impl F1 I2 D2 D0) =
        TA_bridges (s.\<alpha> F1) (s.\<alpha> I2) (d.\<alpha> D2) (d.\<alpha> D0)"
        using c2 assms(4)
        by (auto simp: d.correct TA_bridges_def TA_bridges_impl_def)
    qed

  end
  type_synonym ('q_set, 'a_set, 'd) NFA_impl_no_props = 
    "'q_set \<times> 'a_set \<times> 'd \<times> 'q_set \<times> 'q_set"
  type_synonym ('q_set, 'a_set, 'd) TA_state_impl =
    "('q_set, 'a_set, 'd) NFA_impl_no_props \<times> nat \<times> bool"
  type_synonym ('q_set, 'a_set, 'd) TA_adder_impl =
    "('q_set, 'a_set, 'd) TA_state_impl \<Rightarrow> ('q_set, 'a_set, 'd) TA_state_impl"

  context Thompson_nat_impl
  begin

    fun TA_zero_NFA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl"
    where
      "TA_zero_NFA_impl ((Q0, A0, D0, I0, F0), n0, e0) =
      ((Q0, A0, D0, I0, F0), n0, e0)"
    fun TA_one_NFA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl"
    where 
      "TA_one_NFA_impl ((Q0, A0, D0, I0, F0), n0, e0) =
      (
        (
          s.ins n0 Q0,
          A0,
          D0,
          s.ins n0 I0,
          s.ins n0 F0
        ),
        Suc n0,
        e0 \<or> True
      )"
    fun TA_atom_NFA_impl :: "'a \<Rightarrow> ('nat_set, 'a_set, 'd) TA_adder_impl"
    where
      "TA_atom_NFA_impl a ((Q0, A0, D0, I0, F0), n0, e0) =
      (
        (
          s.ins (Suc n0) (s.ins n0 Q0),
          l.ins a A0,
          d.add n0 a (Suc n0) D0,
          s.ins n0 I0,
          s.ins (Suc n0) F0
        ),
        Suc (Suc n0),
        e0 \<or> False
      )"
    fun TA_plus_NFA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow>
      ('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow> ('nat_set, 'a_set, 'd) TA_adder_impl"
    where
      "TA_plus_NFA_impl add_\<A>1 add_\<A>2 ((Q0, A0, D0, I0, F0), n0, e0) =
        add_\<A>2 (add_\<A>1 ((Q0, A0, D0, I0, F0), n0, e0))"
    fun TA_times_NFA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow>
      ('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow> ('nat_set, 'a_set, 'd) TA_adder_impl"
    where 
      "TA_times_NFA_impl \<A>1 \<A>2 ((Q0, A0, D0, I0, F0), n0, e0) =
      (
        let
          ((Q1, A1, D1, I1, F1), n1, e1) =
            \<A>1 ((Q0, A0, D0, s.empty (), s.empty ()), n0, False);
          ((Q2, A2, D2, I2, F2), n2, e2) =
            \<A>2 ((Q1, A1, D1, s.empty (), s.empty ()), n1, False)
        in
        (
          (
            Q2,
            A2,
            TA_bridges_impl F1 I2 D2 D2,
            s.union I0 I1,
            s.union F0 (if e2 then s.union F1 F2 else F2)
          ),
          n2,
          e0 \<or> e1 \<and> e2
        )
      )"
    fun TA_transcl_NFA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow>
      ('nat_set, 'a_set, 'd) TA_adder_impl"
    where
      "TA_transcl_NFA_impl \<A> ((Q0, A0, D0, I0, F0), n0, e0) =
      (
        let
          ((Q1, A1, D1, I1, F1), n1, e1) =
            \<A> ((Q0, A0, D0, s.empty (), s.empty ()), n0, False)
        in
        (
          (
            Q1,
            A1,
            TA_bridges_impl F1 I1 D1 D1,
            s.union I0 I1,
            s.union F0 F1
          ),
          n1,
          e0 \<or> e1
        )
      )"
    fun TA_reflcl_NFA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow>
      ('nat_set, 'a_set, 'd) TA_adder_impl"
    where 
      "TA_reflcl_NFA_impl \<A> ((Q0, A0, D0, I0, F0), n0, e0) =
      (
        let
          ((Q1, A1, D1, I1, F1), n1, e1) =
            \<A> ((Q0, A0, D0, s.empty (), s.empty ()), n0, False);
          add_reflcl = if e1 then TA_zero_NFA_impl else TA_one_NFA_impl;
          ((Q2, A2, D2, I2, F2), n2, e2) =
            add_reflcl ((Q1, A1, D1, I1, F1), n1, e1)
        in
        (
          (
            Q2,
            A2,
            D2,
            s.union I0 I2,
            s.union F0 F2
          ),
          n2,
          e0 \<or> e2
        )
      )"
    fun TA_star_NFA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow>
      ('nat_set, 'a_set, 'd) TA_adder_impl"
    where
      "TA_star_NFA_impl add_\<A> ((Q0, A0, D0, I0, F0), n0, e0) =
        TA_reflcl_NFA_impl (TA_transcl_NFA_impl add_\<A>)
        ((Q0, A0, D0, I0, F0), n0, e0)"

    definition TA_impl :: "('nat_set, 'a_set, 'd) TA_adder_impl \<Rightarrow> 'a TA_adder \<Rightarrow> bool"
    where "TA_impl add_\<A> add_\<A>' \<equiv> \<forall> Q0 A0 D0 I0 F0 n0 e0.
      nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial) \<longrightarrow>
      (
        let ((Q1, A1, D1, I1, F1), n1, e1) = add_\<A> ((Q0, A0, D0, I0, F0), n0, e0)
        in
          nfa_invar (Q1, A1, D1, I1, F1, nfa_props_trivial) \<and>
          (nfa_\<alpha> (Q1, A1, D1, I1, F1, nfa_props_trivial), n1, e1) =
            add_\<A>' (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)
      )"

    lemma TA_impl_consequences:
      assumes "TA_impl add_\<A> add_\<A>'"
      assumes "((Q1, A1, D1, I1, F1), n1, e1) = add_\<A> ((Q0, A0, D0, I0, F0), n0, e0)"
      assumes "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"
      shows
        "nfa_invar (Q1, A1, D1, I1, F1, nfa_props_trivial)"
        "(nfa_\<alpha> (Q1, A1, D1, I1, F1, nfa_props_trivial), n1, e1) =
          add_\<A>' (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
      using assms(1)[unfolded TA_impl_def Let_def, rule_format, of Q0 A0 D0 I0 F0 n0 e0]
      unfolding assms(2)[symmetric]
      using assms(3-3)
      by auto
    lemma TA_impl_intro[intro]:
      assumes "\<And> Q0 A0 D0 I0 F0 n0 e0 Q1 A1 D1 I1 F1 n1 e1.
        ((Q1, A1, D1, I1, F1), n1, e1) = add_\<A> ((Q0, A0, D0, I0, F0), n0, e0) \<Longrightarrow>
        nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial) \<Longrightarrow>
          nfa_invar (Q1, A1, D1, I1, F1, nfa_props_trivial) \<and>
          (nfa_\<alpha> (Q1, A1, D1, I1, F1, nfa_props_trivial), n1, e1) =
            add_\<A>' (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
      shows "TA_impl add_\<A> add_\<A>'"
      unfolding TA_impl_def
      unfolding Let_def
      apply clarify
      using assms
      by metis

    lemma TA_zero_NFA_impl_correct: "TA_impl TA_zero_NFA_impl TA_zero_NFA"
      by (rule, auto)
    lemma TA_one_NFA_impl_correct: "TA_impl TA_one_NFA_impl TA_one_NFA"
    proof (rule, intro conjI)
      fix Q0 A0 D0 I0 F0 n0 e0
      fix Q1 A1 D1 I1 F1 n1 e1
      assume a:
        "((Q1, A1, D1, I1, F1), n1, e1) = TA_one_NFA_impl ((Q0, A0, D0, I0, F0), n0, e0)"
        "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"
      show b: "(nfa_\<alpha> (Q1, A1, D1, I1, F1, nfa_props_trivial), n1, e1) =
        TA_one_NFA (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
        using a
        unfolding nfa_invar_full_def
        by (auto simp: s.correct l.correct d.correct)
      show "nfa_invar (Q1, A1, D1, I1, F1, nfa_props_trivial)"
        using a
        using nfa_invar_consequences(6)[OF a(2)]
        using TA_one_NFA.preserves_well_formed[OF b]
        unfolding nfa_invar_full_def
        by (auto simp: s.correct l.correct d.correct)
    qed
    lemma TA_atom_NFA_impl_correct: "TA_impl (TA_atom_NFA_impl a) (TA_atom_NFA a)"
    proof (rule, intro conjI)
      fix Q0 A0 D0 I0 F0 n0 e0
      fix Q1 A1 D1 I1 F1 n1 e1
      assume a:
        "((Q1, A1, D1, I1, F1), n1, e1) = TA_atom_NFA_impl a ((Q0, A0, D0, I0, F0), n0, e0)"
        "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"
      show b: "(nfa_\<alpha> (Q1, A1, D1, I1, F1, nfa_props_trivial), n1, e1) =
        TA_atom_NFA a (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
        using a
        unfolding nfa_invar_full_def
        by (auto simp: s.correct l.correct d.correct)
      show "nfa_invar (Q1, A1, D1, I1, F1, nfa_props_trivial)"
        using a
        using nfa_invar_consequences(6)[OF a(2)]
        using TA_atom_NFA.preserves_well_formed[OF b]
        unfolding nfa_invar_full_def
        by (auto simp: s.correct l.correct d.correct)
    qed
    lemma TA_plus_NFA_impl_correct: 
      assumes "TA_adder_well_formed2 add_\<A>1 add_\<A>2"
      assumes "TA_impl add_\<A>1' add_\<A>1" "TA_impl add_\<A>2' add_\<A>2"
      shows "TA_impl (TA_plus_NFA_impl add_\<A>1' add_\<A>2') (TA_plus_NFA add_\<A>1 add_\<A>2)"
    proof (rule, intro conjI)
      fix Q0 A0 D0 I0 F0 n0 e0
      fix Q2 A2 D2 I2 F2 n2 e2
      assume a:
        "((Q2, A2, D2, I2, F2), n2, e2) =
          TA_plus_NFA_impl add_\<A>1' add_\<A>2' ((Q0, A0, D0, I0, F0), n0, e0)"
        "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"
      obtain Q1 A1 D1 I1 F1 n1 e1 where b1: "((Q1, A1, D1, I1, F1), n1, e1) =
        add_\<A>1' ((Q0, A0, D0, I0, F0), n0, e0)" by (metis PairE)
      have b2: "((Q2, A2, D2, I2, F2), n2, e2) = add_\<A>2' ((Q1, A1, D1, I1, F1), n1, e1)"
        using a(1) b1 by auto
      note r1 = TA_impl_consequences[OF assms(2) b1 a(2)]
      note r2 = TA_impl_consequences[OF assms(3) b2 r1(1)]
      show "(nfa_\<alpha> (Q2, A2, D2, I2, F2, nfa_props_trivial), n2, e2) =
        TA_plus_NFA add_\<A>1 add_\<A>2 (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"  
        using r1(2) r2(2) by auto
      show "nfa_invar (Q2, A2, D2, I2, F2, nfa_props_trivial)" using r2(1) by this
    qed
    lemma TA_times_NFA_impl_correct:
      assumes "TA_adder_well_formed2 add_\<A>1 add_\<A>2"
      assumes "TA_impl add_\<A>1' add_\<A>1" "TA_impl add_\<A>2' add_\<A>2"
      shows "TA_impl (TA_times_NFA_impl add_\<A>1' add_\<A>2') (TA_times_NFA add_\<A>1 add_\<A>2)"
    proof (rule, intro conjI)
      fix Q0 A0 D0 I0 F0 n0 e0
      fix Q5 A5 D5 I5 F5 n5 e5
      assume a:
        "((Q5, A5, D5, I5, F5), n5, e5) =
          TA_times_NFA_impl add_\<A>1' add_\<A>2' ((Q0, A0, D0, I0, F0), n0, e0)"
        "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"

      let ?\<A>1 = "(Q0, A0, D0, s.empty (), s.empty ())"
      obtain Q2 A2 D2 I2 F2 n2 e2 where b1: "((Q2, A2, D2, I2, F2), n2, e2) =
        add_\<A>1' (?\<A>1, n0, False)" by (metis PairE)
      have r1': "nfa_invar (Q0, A0, D0, s.empty (), s.empty (), nfa_props_trivial)"
        using a(2)
        unfolding nfa_invar_full_def NFA_full_def
        by (auto simp: s.correct l.correct d.correct)
      note r1 = TA_impl_consequences[OF assms(2) b1 r1']
      have r1c: "nfa_\<alpha> (Q0, A0, D0, s.empty (), s.empty (), nfa_props_trivial) =
        \<lparr> \<Q> = s.\<alpha> Q0, \<Sigma> = l.\<alpha> A0, \<Delta> = d.\<alpha> D0, \<I> = {}, \<F> = {} \<rparr>"
        by (auto simp: s.correct)

      let ?\<A>3 = "(Q2, A2, D2, s.empty (), s.empty ())"
      obtain Q4 A4 D4 I4 F4 n4 e4 where b2: "((Q4, A4, D4, I4, F4), n4, e4) =
        add_\<A>2' (?\<A>3, n2, False)" by (metis PairE)
      have r2': "nfa_invar (Q2, A2, D2, s.empty (), s.empty (), nfa_props_trivial)"
        using r1(1)
        unfolding nfa_invar_full_def NFA_full_def
        by (auto simp: s.correct l.correct d.correct)
      note r2 = TA_impl_consequences[OF assms(3) b2 r2']
      have r2c: "nfa_\<alpha> (Q2, A2, D2, s.empty (), s.empty (), nfa_props_trivial) =
        \<lparr> \<Q> = s.\<alpha> Q2, \<Sigma> = l.\<alpha> A2, \<Delta> = d.\<alpha> D2, \<I> = {}, \<F> = {} \<rparr>"
        by (auto simp: s.correct)

      interpret TA_adder_well_formed2 add_\<A>1 add_\<A>2 by fact

      show b: "(nfa_\<alpha> (Q5, A5, D5, I5, F5, nfa_props_trivial), n5, e5) =
        TA_times_NFA add_\<A>1 add_\<A>2 (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
        using a r1(1) r2(1) nfa_invar_consequences
        by (auto simp: b1[symmetric] b2[symmetric]
          r1(2)[unfolded r1c, symmetric] r2(2)[unfolded r2c, symmetric]
          s.correct l.correct d.correct TA_bridges_impl_correct)
      show "nfa_invar (Q5, A5, D5, I5, F5, nfa_props_trivial)"
        using a r1(1) r2(1) nfa_invar_consequences
        using TA_times_NFA.preserves_well_formed[OF b nfa_invar_consequences(6)[OF a(2)]]
        unfolding nfa_invar_full_def
        by (auto simp: b1[symmetric] b2[symmetric]
          r1(2)[unfolded r1c, symmetric] r2(2)[unfolded r2c, symmetric]
          s.correct l.correct d.correct TA_bridges_impl_correct)
    qed
    lemma TA_transcl_NFA_impl_correct:
      assumes "TA_adder_well_formed1 add_\<A>"
      assumes "TA_impl add_\<A>' add_\<A>"
      shows "TA_impl (TA_transcl_NFA_impl add_\<A>') (TA_transcl_NFA add_\<A>)"
    proof (rule, intro conjI)
      fix Q0 A0 D0 I0 F0 n0 e0
      fix Q3 A3 D3 I3 F3 n3 e3
      assume a:
        "((Q3, A3, D3, I3, F3), n3, e3) =
          TA_transcl_NFA_impl add_\<A>' ((Q0, A0, D0, I0, F0), n0, e0)"
        "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"

      let ?\<A>1 = "(Q0, A0, D0, s.empty (), s.empty ())"
      obtain Q2 A2 D2 I2 F2 n2 e2 where b1: "((Q2, A2, D2, I2, F2), n2, e2) =
        add_\<A>' (?\<A>1, n0, False)" by (metis PairE)
      have r1': "nfa_invar (Q0, A0, D0, s.empty (), s.empty (), nfa_props_trivial)"
        using a(2)
        unfolding nfa_invar_full_def NFA_full_def
        by (auto simp: s.correct l.correct d.correct)
      note r1 = TA_impl_consequences[OF assms(2) b1 r1']
      have r1c: "nfa_\<alpha> (Q0, A0, D0, s.empty (), s.empty (), nfa_props_trivial) =
        \<lparr> \<Q> = s.\<alpha> Q0, \<Sigma> = l.\<alpha> A0, \<Delta> = d.\<alpha> D0, \<I> = {}, \<F> = {} \<rparr>"
        by (auto simp: s.correct)

      interpret TA_adder_well_formed1 add_\<A> by fact

      show b: "(nfa_\<alpha> (Q3, A3, D3, I3, F3, nfa_props_trivial), n3, e3) =
        TA_transcl_NFA add_\<A> (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
        using a r1(1) nfa_invar_consequences
        by (auto simp: b1[symmetric]
          r1(2)[unfolded r1c, symmetric]
          s.correct l.correct d.correct TA_bridges_impl_correct)
      show "nfa_invar (Q3, A3, D3, I3, F3, nfa_props_trivial)"
        using a r1(1) nfa_invar_consequences
        using TA_transcl_NFA.preserves_well_formed[OF b nfa_invar_consequences(6)[OF a(2)]]
        unfolding nfa_invar_full_def
        by (auto simp: b1[symmetric]
          r1(2)[unfolded r1c, symmetric]
          s.correct l.correct d.correct TA_bridges_impl_correct)
    qed
    lemma TA_reflcl_NFA_impl_correct:
      assumes "TA_adder_well_formed1 add_\<A>"
      assumes "TA_impl add_\<A>' add_\<A>"
      shows "TA_impl (TA_reflcl_NFA_impl add_\<A>') (TA_reflcl_NFA add_\<A>)"
    proof (rule, intro conjI)
      fix Q0 A0 D0 I0 F0 n0 e0
      fix Q4 A4 D4 I4 F4 n4 e4
      assume a:
        "((Q4, A4, D4, I4, F4), n4, e4) =
          TA_reflcl_NFA_impl add_\<A>' ((Q0, A0, D0, I0, F0), n0, e0)"
        "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"

      let ?\<A>1 = "(Q0, A0, D0, s.empty (), s.empty ())"
      obtain Q2 A2 D2 I2 F2 n2 e2 where b1: "((Q2, A2, D2, I2, F2), n2, e2) =
        add_\<A>' (?\<A>1, n0, False)" by (metis PairE)
      have r1': "nfa_invar (Q0, A0, D0, s.empty (), s.empty (), nfa_props_trivial)"
        using a(2)
        unfolding nfa_invar_full_def NFA_full_def
        by (auto simp: s.correct l.correct d.correct)
      note r1 = TA_impl_consequences[OF assms(2) b1 r1']
      have r1c: "nfa_\<alpha> (Q0, A0, D0, s.empty (), s.empty (), nfa_props_trivial) =
        \<lparr> \<Q> = s.\<alpha> Q0, \<Sigma> = l.\<alpha> A0, \<Delta> = d.\<alpha> D0, \<I> = {}, \<F> = {} \<rparr>"
        by (auto simp: s.correct)

      let ?add_reflcl' = "if e2 then TA_zero_NFA_impl else TA_one_NFA_impl"
      let ?add_reflcl = "if e2 then TA_zero_NFA else TA_one_NFA"
      obtain Q3 A3 D3 I3 F3 n3 e3 where b2: "((Q3, A3, D3, I3, F3), n3, e3) =
        ?add_reflcl' ((Q2, A2, D2, I2, F2), n2, e2)" by (metis PairE)
      have r2': "TA_impl ?add_reflcl' ?add_reflcl"
        apply (cases e2)
        using TA_zero_NFA_impl_correct apply simp
        using TA_one_NFA_impl_correct apply simp
      done
      note r2 = TA_impl_consequences[OF r2' b2 r1(1)]

      interpret TA_adder_well_formed1 add_\<A> by fact

      show b: "(nfa_\<alpha> (Q4, A4, D4, I4, F4, nfa_props_trivial), n4, e4) =
        TA_reflcl_NFA add_\<A> (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
        using a r1(1) nfa_invar_consequences
        by (auto simp: b1[symmetric]
          r1(2)[unfolded r1c, symmetric]
          s.correct l.correct d.correct)
      show "nfa_invar (Q4, A4, D4, I4, F4, nfa_props_trivial)"
        using a r1(1) r2(1) nfa_invar_consequences
        using TA_reflcl_NFA.preserves_well_formed[OF b nfa_invar_consequences(6)[OF a(2)]]
        unfolding nfa_invar_full_def
        by (auto simp: b1[symmetric] b2[symmetric]
          r1(2)[unfolded r1c, symmetric]
          s.correct l.correct d.correct)
    qed
    lemma TA_star_NFA_impl_correct:
      assumes "TA_adder_well_formed1 add_\<A>"
      assumes "TA_impl add_\<A>' add_\<A>"
      shows "TA_impl (TA_star_NFA_impl add_\<A>') (TA_star_NFA add_\<A>)"
    proof (rule, intro conjI)
      fix Q0 A0 D0 I0 F0 n0 e0
      fix Q1 A1 D1 I1 F1 n1 e1
      assume a:
        "((Q1, A1, D1, I1, F1), n1, e1) =
          TA_star_NFA_impl add_\<A>' ((Q0, A0, D0, I0, F0), n0, e0)"
        "nfa_invar (Q0, A0, D0, I0, F0, nfa_props_trivial)"

      have b1: "((Q1, A1, D1, I1, F1), n1, e1) = TA_reflcl_NFA_impl
        (TA_transcl_NFA_impl add_\<A>') ((Q0, A0, D0, I0, F0), n0, e0)"
        using a(1) by auto
      interpret TA_adder_well_formed1 "add_\<A>" by fact
      note r1'' = TA_transcl_NFA_impl_correct[OF TA_adder_well_formed1_axioms assms(2)]
      interpret TA_adder_well_formed1 "TA_transcl_NFA add_\<A>" by intro_locales
      note r1' = TA_reflcl_NFA_impl_correct[OF TA_adder_well_formed1_axioms r1'']
      note r1 = TA_impl_consequences[OF r1' b1 a(2)]

      show b: "(nfa_\<alpha> (Q1, A1, D1, I1, F1, nfa_props_trivial), n1, e1) =
        TA_star_NFA add_\<A> (nfa_\<alpha> (Q0, A0, D0, I0, F0, nfa_props_trivial), n0, e0)"
        using r1(2) by auto
      show "nfa_invar (Q1, A1, D1, I1, F1, nfa_props_trivial)"
        using r1(1) by this
    qed

    fun TA_translate_NFA_impl :: "'a rexp \<Rightarrow> ('nat_set, 'a_set, 'd) TA_adder_impl"
    where
      "TA_translate_NFA_impl Regular_Exp.Zero = TA_zero_NFA_impl" |
      "TA_translate_NFA_impl Regular_Exp.One = TA_one_NFA_impl" |
      "TA_translate_NFA_impl (Regular_Exp.Atom a) = TA_atom_NFA_impl a" |
      "TA_translate_NFA_impl (Regular_Exp.Plus r1 r2) =
        TA_plus_NFA_impl (TA_translate_NFA_impl r1) (TA_translate_NFA_impl r2)" |
      "TA_translate_NFA_impl (Regular_Exp.Times r1 r2) =
        TA_times_NFA_impl (TA_translate_NFA_impl r1) (TA_translate_NFA_impl r2)" |
      "TA_translate_NFA_impl (Regular_Exp.Star r) =
        TA_star_NFA_impl (TA_translate_NFA_impl r)"
    fun TA_translate_NFA_wrapper_impl :: "'a rexp \<Rightarrow> ('nat_set, 'a_set, 'd) NFA_impl"
    where "TA_translate_NFA_wrapper_impl r = 
      (
        let
          S = (s.empty (), l.empty (), d.empty (), s.empty (), s.empty ());
          ((Q, A, D, I, F), n, e) = TA_translate_NFA_impl r (S, 0, False)
        in (Q, A, D, I, F, nfa_props_trivial)
      )"

    lemma TA_translate_NFA_impl_correct:
      "TA_impl (TA_translate_NFA_impl r) (TA_translate_NFA r)"
    proof (induct r)
      case Zero thus ?case using TA_zero_NFA_impl_correct by simp
      case One thus ?case using TA_one_NFA_impl_correct by simp
      case Atom thus ?case using TA_atom_NFA_impl_correct by simp
    next
      case (Plus r1 r2)
      interpret TA_adder_well_formed2 "TA_translate_NFA r1" "TA_translate_NFA r2"
        by intro_locales
      show ?case
        using TA_plus_NFA_impl_correct[OF TA_adder_well_formed2_axioms Plus]
        by simp
    next
      case (Times r1 r2)
      interpret TA_adder_well_formed2 "TA_translate_NFA r1" "TA_translate_NFA r2"
        by intro_locales
      show ?case
        using TA_times_NFA_impl_correct[OF TA_adder_well_formed2_axioms Times]
        by simp
    next
      case (Star r)
      interpret TA_adder_well_formed1 "TA_translate_NFA r"
        by intro_locales
      show ?case
        using TA_star_NFA_impl_correct[OF TA_adder_well_formed1_axioms Star]
        by simp
    qed

    lemma TA_translate_NFA_wrapper_impl_correct:
      "nfa_invar (TA_translate_NFA_wrapper_impl r)"
      "nfa_\<alpha> (TA_translate_NFA_wrapper_impl r) = TA_translate_NFA_wrapper r"
    proof -
      have a:
        "nfa_invar (s.empty (), l.empty (), d.empty (), s.empty (), s.empty (), nfa_props_trivial)"
        "nfa_\<alpha> (s.empty (), l.empty (), d.empty (), s.empty (), s.empty (), nfa_props_trivial) =
          \<lparr>\<Q> = {}, \<Sigma> = {}, \<Delta> = {}, \<I> = {}, \<F> = {}\<rparr>"
        by (auto simp: s.correct l.correct d.correct nfa_invar_full_def)
      obtain Q A D I F n e where b: "((Q, A, D, I, F), n, e) = TA_translate_NFA_impl r
        ((s.empty (), l.empty (), d.empty (), s.empty (), s.empty ()), 0, False)"
        using PairE
        by metis
      note c = TA_impl_consequences[OF TA_translate_NFA_impl_correct b a(1)]
      show "nfa_invar (TA_translate_NFA_wrapper_impl r)"
        using c(1)
        by (simp add: b[symmetric])
      show "nfa_\<alpha> (TA_translate_NFA_wrapper_impl r) = TA_translate_NFA_wrapper r"
        using c(2)
        unfolding a(2)
        by (simp add: b[symmetric], metis fst_conv)
    qed

  end

  definition "rs_TA_bridges_impl \<equiv> rs_Thompson_impl.TA_bridges_impl"
  definition "rs_TA_zero_NFA_impl \<equiv> rs_Thompson_impl.TA_zero_NFA_impl"
  definition "rs_TA_one_NFA_impl \<equiv> rs_Thompson_impl.TA_one_NFA_impl"
  definition "rs_TA_atom_NFA_impl \<equiv> rs_Thompson_impl.TA_atom_NFA_impl"
  definition "rs_TA_plus_NFA_impl \<equiv> rs_Thompson_impl.TA_plus_NFA_impl"
  definition "rs_TA_times_NFA_impl \<equiv> rs_Thompson_impl.TA_times_NFA_impl"
  definition "rs_TA_transcl_NFA_impl \<equiv> rs_Thompson_impl.TA_transcl_NFA_impl"
  definition "rs_TA_reflcl_NFA_impl \<equiv> rs_Thompson_impl.TA_reflcl_NFA_impl"
  definition "rs_TA_star_NFA_impl \<equiv> rs_Thompson_impl.TA_star_NFA_impl"
  definition "rs_TA_translate_NFA_impl \<equiv> rs_Thompson_impl.TA_translate_NFA_impl"
  definition "rs_TA_translate_NFA_wrapper_impl \<equiv> rs_Thompson_impl.TA_translate_NFA_wrapper_impl"

  lemmas rs_defs =
    rs_nfa_defs
    rs_TA_bridges_impl_def
    rs_TA_zero_NFA_impl_def
    rs_TA_one_NFA_impl_def
    rs_TA_atom_NFA_impl_def
    rs_TA_plus_NFA_impl_def
    rs_TA_times_NFA_impl_def
    rs_TA_transcl_NFA_impl_def
    rs_TA_reflcl_NFA_impl_def
    rs_TA_star_NFA_impl_def
    rs_TA_translate_NFA_impl_def
    rs_TA_translate_NFA_wrapper_impl_def

  lemmas [code] = 
    rs_Thompson_impl.TA_bridges_impl_def[folded rs_defs]
    rs_Thompson_impl.TA_zero_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_one_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_atom_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_plus_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_times_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_transcl_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_reflcl_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_star_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_translate_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TA_translate_NFA_wrapper_impl.simps[folded rs_defs]

  definition "rs_TA_translate_NFA_wrapper_normalize_impl r \<equiv>
    rs_nfa_normalise (rs_TA_translate_NFA_wrapper_impl r)"

  theorem rs_TA_translate_NFA_wrapper_correct:
    "rs_nfa_invar (rs_TA_translate_NFA_wrapper_impl r)"
    "\<L> (rs_nfa_\<alpha> (rs_TA_translate_NFA_wrapper_impl r)) = lang r"
    unfolding rs_nfa_invar_def rs_nfa_\<alpha>_def
    unfolding rs_TA_translate_NFA_wrapper_impl_def
    using TA_translate_NFA_wrapper_correct(2)
    using rs_Thompson_impl.TA_translate_NFA_wrapper_impl_correct[of r]
    by auto

  theorem rs_TA_translate_NFA_wrapper_normalize_correct:
    "rs_nfa_invar (rs_TA_translate_NFA_wrapper_normalize_impl r)"
    "\<L> (rs_nfa_\<alpha> (rs_TA_translate_NFA_wrapper_normalize_impl r)) = lang r"
    unfolding rs_TA_translate_NFA_wrapper_normalize_impl_def
    using rs_TA_translate_NFA_wrapper_correct(1)
    using rs_nfa.normalise_correct(1)
    apply metis
    using rs_TA_translate_NFA_wrapper_correct
    using NFA_isomorphic_wf_\<L> NFA_remove_unreachable_states_\<L>
    using rs_nfa.normalise_correct(2)
    apply metis
  done

end
