theory TM_Correctness
imports
  TM_Definitions
  TM_Simple
  TM_Plus
  TM_Times
  TM_Star
begin

  lemma TM_translate_NFA_correct:
    "\<Q> (TM_translate_NFA r n) = {n ..< n + card (\<Q> (TM_translate_NFA r n))}"
    "NFA (TM_translate_NFA r n)"
    "\<L> (TM_translate_NFA r n) = lang r"
  proof (induct r arbitrary: n)
    case Zero
    case 1 thus ?case using TM_zero_NFA_states by simp
    case 2 thus ?case using TM_zero_NFA_well_formed by simp
    case 3 thus ?case using TM_zero_NFA_language by simp
  next
    case One
    case 1 thus ?case using TM_one_NFA_states by simp
    case 2 thus ?case using TM_one_NFA_well_formed by simp
    case 3 thus ?case using TM_one_NFA_language by simp
  next
    case (Atom a)
    case 1 thus ?case using TM_atom_NFA_states by simp
    case 2 thus ?case using TM_atom_NFA_well_formed by simp
    case 3 thus ?case using TM_atom_NFA_language[of "n" "Suc n"] by simp
  next
    case (Plus r1 r2)
    fix n
    let ?\<A>1 = "TM_translate_NFA r1 n"
    let ?n' = "n + card (\<Q> ?\<A>1)"
    let ?\<A>2 = "TM_translate_NFA r2 ?n'"
    have TM_translate_NFA_case: "TM_translate_NFA (Plus r1 r2) n = TM_plus_NFA ?\<A>1 ?\<A>2"
      by (simp add: Let_def)
    interpret disjoint_NFAs ?\<A>1 ?\<A>2
      using Plus(2) Plus(5) Plus(1) Plus(4) ivl_disj_int(11)
      unfolding disjoint_NFAs_def disjoint_NFAs_axioms_def
      by metis

    show "\<Q> (TM_translate_NFA (Plus r1 r2) n) = {n ..< n + card (\<Q> (TM_translate_NFA (Plus r1 r2) n))}"
      using TM_translate_NFA_case TM_plus_NFA_states[OF Plus(1) Plus(4)]
      by auto
    show "NFA (TM_translate_NFA (Plus r1 r2) n)"
      using TM_translate_NFA_case TM_plus_NFA_well_formed[OF Plus(2) Plus(5)]
      by simp
    show "\<L> (TM_translate_NFA (Plus r1 r2) n) = lang (Plus r1 r2)"
      using TM_translate_NFA_case TM_plus_NFA_language Plus(3) Plus(6)
      by simp
  next
    case (Times r1 r2)
    fix n
    let ?\<A>1 = "TM_translate_NFA r1 n"
    let ?n' = "n + card (\<Q> ?\<A>1)"
    let ?\<A>2 = "TM_translate_NFA r2 ?n'"
    have TM_translate_NFA_case: "TM_translate_NFA (Times r1 r2) n = TM_times_NFA ?\<A>1 ?\<A>2"
      by (simp add: Let_def)
    interpret disjoint_NFAs ?\<A>1 ?\<A>2
      using Times(2) Times(5) Times(1) Times(4) ivl_disj_int(11)
      unfolding disjoint_NFAs_def disjoint_NFAs_axioms_def
      by metis

    show "\<Q> (TM_translate_NFA (Times r1 r2) n) = {n ..< n + card (\<Q> (TM_translate_NFA (Times r1 r2) n))}"
      using TM_translate_NFA_case TM_times_NFA_states[OF Times(1) Times(4)]
      by auto
    show "NFA (TM_translate_NFA (Times r1 r2) n)"
      using TM_translate_NFA_case TM_times_NFA_well_formed[OF Times(2) Times(5)]
      by simp
    show "\<L> (TM_translate_NFA (Times r1 r2) n) = lang (Times r1 r2)"
      using TM_translate_NFA_case TM_times_NFA_language Times(3) Times(6)
      by simp
  next
    case (Star r)
    fix n
    let ?\<A> = "TM_translate_NFA r n"
    let ?n' = "n + card (\<Q> ?\<A>)"
    have TM_translate_NFA_case: "TM_translate_NFA (Star r) n = TM_star_NFA ?\<A> ?n'"
      by (simp add: Let_def)
    interpret NFA ?\<A> using Star(2) by simp
    have a: "?n' \<notin> \<Q> ?\<A>" using Star(1)[of n] by auto

    have b: "\<And> S n m. S = {n :: nat ..< n + m} \<Longrightarrow> insert (n + m) S = {n ..< n + m + 1}"
      using Star(1) by auto
    show "\<Q> (TM_translate_NFA (Star r) n) = {n ..< n + card (\<Q> (TM_translate_NFA (Star r) n))}"
      unfolding TM_translate_NFA_case
      using b[OF Star(1)] Star(1)
      by (cases "\<I> (TM_translate_NFA r n) \<inter> \<F> (TM_translate_NFA r n) = {}", auto)
    show "NFA (TM_translate_NFA (Star r) n)"
      using TM_translate_NFA_case TM_star_NFA_well_formed
      by simp
    show "\<L> (TM_translate_NFA (Star r) n) = lang (Star r)"
      using TM_translate_NFA_case TM_star_NFA_language[OF a] Star(3)
      by simp
  qed
  theorem TM_translate_NFA_wrapper_correct:
    "NFA (TM_translate_NFA_wrapper r)"
    "\<L> (TM_translate_NFA_wrapper r) = lang r"
    using TM_translate_NFA_correct[of r 0]
    by auto

end
