theory TM_Definitions
imports
  "../../Libs/Regular-Sets/Regular_Exp"
  "../../Automata_Malik_Tuerk/NFA"
begin

  definition TM_bridges ::
    "'q set \<Rightarrow> 'q set \<Rightarrow> ('q \<times> 'a \<times> 'q) set \<Rightarrow> ('q \<times> 'a \<times> 'q) set"
  where "TM_bridges \<F>1 \<I>2 \<Delta>2 \<equiv>
    {(p1, a2, q2) | p1 p2 a2 q2. p1 \<in> \<F>1 \<and> p2 \<in> \<I>2 \<and> (p2, a2, q2) \<in> \<Delta>2}"

  definition TM_zero_NFA :: "('q, 'a, _) NFA_rec_scheme"
  where "TM_zero_NFA \<equiv> \<lparr> \<Q> = {}, \<Sigma> = {}, \<Delta> = {}, \<I> = {}, \<F> = {} \<rparr>"
  definition TM_one_NFA :: "'q \<Rightarrow> ('q, 'a, _) NFA_rec_scheme"
  where "TM_one_NFA p \<equiv> \<lparr> \<Q> = {p}, \<Sigma> = {}, \<Delta> = {}, \<I> = {p}, \<F> = {p} \<rparr>"
  definition TM_atom_NFA :: "'q \<Rightarrow> 'a \<Rightarrow> 'q \<Rightarrow> ('q, 'a, _) NFA_rec_scheme"
  where "TM_atom_NFA p a q \<equiv> \<lparr> \<Q> = {p, q}, \<Sigma> = {a}, \<Delta> = {(p, a, q)}, \<I> = {p}, \<F> = {q} \<rparr>"
  definition TM_plus_NFA :: "('q, 'a, _) NFA_rec_scheme \<Rightarrow>
    ('q, 'a, _) NFA_rec_scheme \<Rightarrow> ('q, 'a, _) NFA_rec_scheme"
  where
    "TM_plus_NFA \<A>1 \<A>2 \<equiv>
    \<lparr>
      \<Q> = \<Q> \<A>1 \<union> \<Q> \<A>2,
      \<Sigma> = \<Sigma> \<A>1 \<union> \<Sigma> \<A>2,
      \<Delta> = \<Delta> \<A>1 \<union> \<Delta> \<A>2,
      \<I> = \<I> \<A>1 \<union> \<I> \<A>2,
      \<F> = \<F> \<A>1 \<union> \<F> \<A>2
    \<rparr>"
  definition TM_times_NFA :: "('q, 'a, _) NFA_rec_scheme \<Rightarrow>
    ('q, 'a, _) NFA_rec_scheme \<Rightarrow> ('q, 'a, _) NFA_rec_scheme"
  where
    "TM_times_NFA \<A>1 \<A>2 \<equiv>
    \<lparr>
      \<Q> = \<Q> \<A>1 \<union> \<Q> \<A>2,
      \<Sigma> = \<Sigma> \<A>1 \<union> \<Sigma> \<A>2,
      \<Delta> = \<Delta> \<A>1 \<union> \<Delta> \<A>2 \<union> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2),
      \<I> = \<I> \<A>1,
      \<F> = if \<I> \<A>2 \<inter> \<F> \<A>2 = {} then \<F> \<A>2 else \<F> \<A>1 \<union> \<F> \<A>2
    \<rparr>"
  definition TM_transcl_NFA :: "('q, 'a, _) NFA_rec_scheme \<Rightarrow>
    ('q, 'a, _) NFA_rec_scheme"
  where "TM_transcl_NFA \<A> \<equiv> TM_times_NFA \<A> \<A>"
  definition TM_reflcl_NFA :: "('q, 'a, _) NFA_rec_scheme \<Rightarrow> 'q \<Rightarrow>
    ('q, 'a, _) NFA_rec_scheme"
  where "TM_reflcl_NFA \<A> p \<equiv>
    let reflcl = if \<I> \<A> \<inter> \<F> \<A> = {} then TM_one_NFA p else TM_zero_NFA
    in TM_plus_NFA \<A> reflcl"
  definition TM_star_NFA :: "('q, 'a, _) NFA_rec_scheme \<Rightarrow> 'q \<Rightarrow>
    ('q, 'a, _) NFA_rec_scheme"
  where "TM_star_NFA \<A> p \<equiv> TM_reflcl_NFA (TM_transcl_NFA \<A>) p"

  fun TM_translate_NFA :: "'a rexp \<Rightarrow> nat \<Rightarrow> (nat, 'a, _) NFA_rec_scheme"
  where
    "TM_translate_NFA Regular_Exp.Zero n = TM_zero_NFA" |
    "TM_translate_NFA Regular_Exp.One n = TM_one_NFA n" |
    "TM_translate_NFA (Regular_Exp.Atom a) n = TM_atom_NFA n a (Suc n)" |
    "TM_translate_NFA (Regular_Exp.Plus r1 r2) n =
    (
      let
        \<A>1 = TM_translate_NFA r1 n;
        n' = n + card (\<Q> \<A>1);
        \<A>2 = TM_translate_NFA r2 n'
      in TM_plus_NFA \<A>1 \<A>2
    )" |
    "TM_translate_NFA (Regular_Exp.Times r1 r2) n =
    (
      let
        \<A>1 = TM_translate_NFA r1 n;
        n' = n + card (\<Q> \<A>1);
        \<A>2 = TM_translate_NFA r2 n'
      in TM_times_NFA \<A>1 \<A>2
    )" |
    "TM_translate_NFA (Regular_Exp.Star r) n =
    (
      let
        \<A> = TM_translate_NFA r n;
        n' = n + card (\<Q> \<A>)
      in TM_star_NFA \<A> n'
    )"
  fun TM_translate_NFA_wrapper :: "'a rexp \<Rightarrow> (nat, 'a, _) NFA_rec_scheme"
  where "TM_translate_NFA_wrapper r = TM_translate_NFA r 0"

end
