theory TM_Implementation
imports
  TM_Correctness
  Thompson_Implementation
begin

  context Thompson_impl
  begin

    definition TM_bridges_impl :: "'q_set \<Rightarrow> 'q_set \<Rightarrow> 'd \<Rightarrow> 'd"
    where "TM_bridges_impl F1 I2 D2 \<equiv>
    (
      let
        add_transitions_from_p2_to_p1 = \<lambda> p1 p2. d_succ_label_it D2
          p2 (\<lambda> _. True) (\<lambda> (a2, q2). d.add p1 a2 q2);
        add_transitions_from_all_p2_to_p1 = \<lambda> p1. s_it I2
          (\<lambda> _. True) (\<lambda> p2. add_transitions_from_p2_to_p1 p1 p2);
        add_transitions_from_all_p2_to_all_p1 = s_it F1
          (\<lambda> _. True) (\<lambda> p1. add_transitions_from_all_p2_to_p1 p1)
      in
        add_transitions_from_all_p2_to_all_p1 (d.empty ())
    )"

    definition TM_zero_NFA_impl :: "('q_set, 'a_set, 'd) NFA_impl"
    where "TM_zero_NFA_impl \<equiv> nfa_construct ([], [], [], [], [])"
    definition TM_one_NFA_impl :: "'q \<Rightarrow> ('q_set, 'a_set, 'd) NFA_impl"
    where "TM_one_NFA_impl p \<equiv> nfa_construct ([p], [], [], [p], [p])"
    definition TM_atom_NFA_impl :: "'q \<Rightarrow> 'a \<Rightarrow> 'q \<Rightarrow> ('q_set, 'a_set, 'd) NFA_impl"
    where "TM_atom_NFA_impl p a q \<equiv> nfa_construct ([p, q], [a], [(p, [a], q)], [p], [q])"
    definition TM_plus_NFA_impl :: "('q_set, 'a_set, 'd) NFA_impl \<Rightarrow>
      ('q_set, 'a_set, 'd) NFA_impl \<Rightarrow> ('q_set, 'a_set, 'd) NFA_impl"
    where
      "TM_plus_NFA_impl \<A>1 \<A>2 \<equiv>
      (
        s.union (nfa_states \<A>1) (nfa_states \<A>2),
        l.union (nfa_labels \<A>1) (nfa_labels \<A>2),
        d.union (nfa_trans \<A>1) (nfa_trans \<A>2),
        s.union (nfa_initial \<A>1) (nfa_initial \<A>2),
        s.union (nfa_accepting \<A>1) (nfa_accepting \<A>2),
        nfa_props_trivial
      )"
    definition TM_times_NFA_impl :: "('q_set, 'a_set, 'd) NFA_impl \<Rightarrow>
      ('q_set, 'a_set, 'd) NFA_impl \<Rightarrow> ('q_set, 'a_set, 'd) NFA_impl"
    where
      "TM_times_NFA_impl \<A>1 \<A>2 \<equiv>
      (
        s.union (nfa_states \<A>1) (nfa_states \<A>2),
        l.union (nfa_labels \<A>1) (nfa_labels \<A>2),
        d.union (d.union (nfa_trans \<A>1) (nfa_trans \<A>2)) (TM_bridges_impl (nfa_accepting \<A>1) (nfa_initial \<A>2) (nfa_trans \<A>2)),
        nfa_initial \<A>1,
        if s.isEmpty (s.inter (nfa_initial \<A>2) (nfa_accepting \<A>2))
          then nfa_accepting \<A>2
          else s.union (nfa_accepting \<A>1) (nfa_accepting \<A>2),
        nfa_props_trivial
      )"
    definition TM_transcl_NFA_impl :: "('q_set, 'a_set, 'd) NFA_impl \<Rightarrow>
      ('q_set, 'a_set, 'd) NFA_impl"
    where "TM_transcl_NFA_impl \<A> \<equiv> TM_times_NFA_impl \<A> \<A>"
    definition TM_reflcl_NFA_impl :: "('q_set, 'a_set, 'd) NFA_impl \<Rightarrow>
      'q \<Rightarrow> ('q_set, 'a_set, 'd) NFA_impl"
    where "TM_reflcl_NFA_impl \<A> p \<equiv> 
      let reflcl =
        if s.isEmpty (s.inter (nfa_initial \<A>) (nfa_accepting \<A>))
          then TM_one_NFA_impl p
          else TM_zero_NFA_impl
      in TM_plus_NFA_impl \<A> reflcl"
    definition TM_star_NFA_impl :: "('q_set, 'a_set, 'd) NFA_impl \<Rightarrow>
      'q \<Rightarrow> ('q_set, 'a_set, 'd) NFA_impl"
    where "TM_star_NFA_impl \<A> p \<equiv> TM_reflcl_NFA_impl (TM_transcl_NFA_impl \<A>) p"

    lemma TM_bridges_impl_correct:
      assumes "s.invar F1" "s.invar I2" "d.invar D2"
      shows
        "d.invar (TM_bridges_impl F1 I2 D2)"
        "d.\<alpha> (TM_bridges_impl F1 I2 D2) = TM_bridges (s.\<alpha> F1) (s.\<alpha> I2) (d.\<alpha> D2)"
    proof -
      let ?add_transitions_from_p2_to_p1 =
        "\<lambda> p1 p2. d_succ_label_it D2 p2 (\<lambda>_. True) (\<lambda> (a2, q2). d.add p1 a2 q2)"
      let ?add_transitions_from_all_p2_to_p1 =
        "\<lambda> p1. s_it I2 (\<lambda>_. True) (?add_transitions_from_p2_to_p1 p1)"
      let ?add_transitions_from_all_p2_to_all_p1 =
        "s_it F1 (\<lambda>_. True) ?add_transitions_from_all_p2_to_p1"

      note s_it_\<A>1_rule = set_iterator_rule_insert_P[OF iteratei_rule[OF assms(1)]]
      note s_it_\<A>2_rule = set_iterator_rule_insert_P[OF iteratei_rule[OF assms(2)]]
      note d_succ_label_it_\<A>2_rule = set_iterator_rule_insert_P
        [OF lts_succ_label_it_correct[OF assms(3)]]

      have a1: "\<And> p1 p2 d. d.invar d \<Longrightarrow>
        d.invar (?add_transitions_from_p2_to_p1 p1 p2 d)"
        by (rule_tac I = "\<lambda> S d'. d.invar d'"
          in d_succ_label_it_\<A>2_rule, auto simp: d.correct)
      have b1: "\<And> p1 d. d.invar d \<Longrightarrow>
        d.invar (?add_transitions_from_all_p2_to_p1 p1 d)"
        using a1
        by (rule_tac I = "\<lambda> S d'. d.invar d'"
          in s_it_\<A>2_rule, auto simp: d.correct)
      have c1: "\<And> d. d.invar d \<Longrightarrow>
        d.invar (?add_transitions_from_all_p2_to_all_p1 d)"
        using b1
        by (rule_tac I = "\<lambda> S d'. d.invar d'"
          in s_it_\<A>1_rule, auto simp: d.correct)

      have a2: "\<And> p1 p2 d. d.invar d \<Longrightarrow>
        d.\<alpha> (?add_transitions_from_p2_to_p1 p1 p2 d) =
          d.\<alpha> d \<union> {(p1, a2, q2) | a2 q2. (p2, a2, q2) \<in> d.\<alpha> D2}"
        by (rule_tac I = "\<lambda> S d'. d.invar d' \<and> d.\<alpha> d' =
          d.\<alpha> d \<union> {(p1, a2, q2) | a2 q2. (a2, q2) \<in> S}"
          in d_succ_label_it_\<A>2_rule, auto simp: d.correct)
      have b2: "\<And> p1 d. d.invar d \<Longrightarrow>
        d.\<alpha> (?add_transitions_from_all_p2_to_p1 p1 d) =
          d.\<alpha> d \<union> {(p1, a2, q2) | p2 a2 q2. p2 \<in> s.\<alpha> I2 \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
        using a1 a2
        by (rule_tac I = "\<lambda> S d'. d.invar d' \<and> d.\<alpha> d' =
          d.\<alpha> d \<union> {(p1, a2, q2) | p2 a2 q2. p2 \<in> S \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
          in s_it_\<A>2_rule, auto simp: d.correct)
      have c2: "\<And> d. d.invar d \<Longrightarrow>
        d.\<alpha> (?add_transitions_from_all_p2_to_all_p1 d) =
          d.\<alpha> d \<union> {(p1, a2, q2) | p1 p2 a2 q2. p1 \<in> s.\<alpha> F1 \<and>
          p2 \<in> s.\<alpha> I2 \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
        using b1 b2
        by (rule_tac I = "\<lambda> S d'. d.invar d' \<and> d.\<alpha> d' =
          d.\<alpha> d \<union> {(p1, a2, q2) | p1 p2 a2 q2. p1 \<in> S \<and>
          p2 \<in> s.\<alpha> I2 \<and> (p2, a2, q2) \<in> d.\<alpha> D2}"
          in s_it_\<A>1_rule, auto simp: d.correct)

      show "d.invar (TM_bridges_impl F1 I2 D2)"
        using c1 by (auto simp: d.correct TM_bridges_impl_def)
      show "d.\<alpha> (TM_bridges_impl F1 I2 D2) = TM_bridges (s.\<alpha> F1) (s.\<alpha> I2) (d.\<alpha> D2)"
        using c2 by (auto simp: d.correct TM_bridges_def TM_bridges_impl_def)
    qed

    lemma TM_zero_NFA_impl_correct: 
      "nfa_invar TM_zero_NFA_impl"
      "nfa_\<alpha> TM_zero_NFA_impl = TM_zero_NFA"
      unfolding TM_zero_NFA_impl_def
      using nfa_from_list.nfa_from_list_correct[OF nfa_construct_correct]
      by (auto simp: NFA_construct.simps SemiAutomaton_construct.simps)
    lemma TM_one_NFA_impl_correct: 
      "nfa_invar (TM_one_NFA_impl p)"
      "nfa_\<alpha> (TM_one_NFA_impl p) = TM_one_NFA p"
      unfolding TM_one_NFA_impl_def
      using nfa_from_list.nfa_from_list_correct[OF nfa_construct_correct]
      by (auto simp: NFA_construct.simps SemiAutomaton_construct.simps)
    lemma TM_atom_NFA_impl_correct: 
      "nfa_invar (TM_atom_NFA_impl p a q)"
      "nfa_\<alpha> (TM_atom_NFA_impl p a q) = TM_atom_NFA p a q"
      unfolding TM_atom_NFA_impl_def TM_atom_NFA_def
      using nfa_from_list.nfa_from_list_correct[OF nfa_construct_correct]
      by (auto simp: NFA_construct.simps SemiAutomaton_construct.simps
        SemiAutomaton_to_NFA_def)
    lemma TM_plus_NFA_impl_correct: 
      assumes "nfa_invar \<A>1" "nfa_invar \<A>2"
      shows
        "nfa_invar (TM_plus_NFA_impl \<A>1 \<A>2)"
        "nfa_\<alpha> (TM_plus_NFA_impl \<A>1 \<A>2) = TM_plus_NFA (nfa_\<alpha> \<A>1) (nfa_\<alpha> \<A>2)"
    proof -
      show a: "nfa_\<alpha> (TM_plus_NFA_impl \<A>1 \<A>2) = TM_plus_NFA (nfa_\<alpha> \<A>1) (nfa_\<alpha> \<A>2)" 
        using assms
        by (auto simp: s.correct l.correct d.correct nfa_invar_consequences
          TM_plus_NFA_impl_def)
      have b: "NFA (nfa_\<alpha> (TM_plus_NFA_impl \<A>1 \<A>2))"
        using assms
        by (simp add: a nfa_invar_consequences(6) TM_plus_NFA_well_formed)
      show "nfa_invar (TM_plus_NFA_impl \<A>1 \<A>2)"
        using assms b
        by (auto simp: s.correct l.correct d.correct nfa_invar_full_def
          TM_plus_NFA_impl_def)
    qed
    lemma TM_times_NFA_impl_correct:
      assumes "nfa_invar \<A>1" "nfa_invar \<A>2"
      shows
        "nfa_invar (TM_times_NFA_impl \<A>1 \<A>2)"
        "nfa_\<alpha> (TM_times_NFA_impl \<A>1 \<A>2) = TM_times_NFA (nfa_\<alpha> \<A>1) (nfa_\<alpha> \<A>2)"
    proof -
      show a: "nfa_\<alpha> (TM_times_NFA_impl \<A>1 \<A>2) = TM_times_NFA (nfa_\<alpha> \<A>1) (nfa_\<alpha> \<A>2)" 
        using assms
        by (auto simp: s.correct l.correct d.correct TM_bridges_impl_correct
          nfa_invar_consequences TM_times_NFA_impl_def)
      have b: "NFA (nfa_\<alpha> (TM_times_NFA_impl \<A>1 \<A>2))"
        using assms
        by (simp add: a nfa_invar_consequences(6) TM_times_NFA_well_formed)
      show "nfa_invar (TM_times_NFA_impl \<A>1 \<A>2)"
        using assms b
        by (auto simp: s.correct l.correct d.correct TM_bridges_impl_correct
          nfa_invar_full_def TM_times_NFA_impl_def)
    qed
    lemma TM_transcl_NFA_impl_correct:
      assumes "nfa_invar \<A>"
      shows
        "nfa_invar (TM_transcl_NFA_impl \<A>)"
        "nfa_\<alpha> (TM_transcl_NFA_impl \<A>) = TM_transcl_NFA (nfa_\<alpha> \<A>)"
      using assms
      by (auto simp: TM_transcl_NFA_impl_def TM_times_NFA_impl_correct)
    lemma TM_reflcl_NFA_impl_correct:
      assumes "nfa_invar \<A>"
      shows
        "nfa_invar (TM_reflcl_NFA_impl \<A> p)"
        "nfa_\<alpha> (TM_reflcl_NFA_impl \<A> p) = TM_reflcl_NFA (nfa_\<alpha> \<A>) p"
      using assms
      by (auto simp: s.correct nfa_invar_consequences TM_reflcl_NFA_impl_def
        TM_plus_NFA_impl_correct TM_one_NFA_impl_correct TM_zero_NFA_impl_correct)
    lemma TM_star_NFA_impl_correct:
      assumes "nfa_invar \<A>"
      shows
        "nfa_invar (TM_star_NFA_impl \<A> p)"
        "nfa_\<alpha> (TM_star_NFA_impl \<A> p) = TM_star_NFA (nfa_\<alpha> \<A>) p"
      using assms
      by (auto simp: TM_star_NFA_impl_def TM_reflcl_NFA_impl_correct TM_transcl_NFA_impl_correct)

  end

  context Thompson_nat_impl
  begin

    fun TM_translate_NFA_impl :: "'a rexp \<Rightarrow> nat \<Rightarrow> ('nat_set, 'a_set, 'd) NFA_impl"
    where
      "TM_translate_NFA_impl Regular_Exp.Zero n = TM_zero_NFA_impl" |
      "TM_translate_NFA_impl Regular_Exp.One n = TM_one_NFA_impl n" |
      "TM_translate_NFA_impl (Regular_Exp.Atom a) n = TM_atom_NFA_impl n a (Suc n)" |
      "TM_translate_NFA_impl (Regular_Exp.Plus r1 r2) n =
      (
        let
          \<A>1 = TM_translate_NFA_impl r1 n;
          n' = n + s.size (nfa_states \<A>1);
          \<A>2 = TM_translate_NFA_impl r2 n'
        in TM_plus_NFA_impl \<A>1 \<A>2
      )" |
      "TM_translate_NFA_impl (Regular_Exp.Times r1 r2) n =
      (
        let
          \<A>1 = TM_translate_NFA_impl r1 n;
          n' = n + s.size (nfa_states \<A>1);
          \<A>2 = TM_translate_NFA_impl r2 n'
        in TM_times_NFA_impl \<A>1 \<A>2
      )" |
      "TM_translate_NFA_impl (Regular_Exp.Star r) n =
      (
        let
          \<A> = TM_translate_NFA_impl r n;
          n' = n + s.size (nfa_states \<A>)
        in TM_star_NFA_impl \<A> n'
      )"
    fun TM_translate_NFA_wrapper_impl ::
      "'a rexp \<Rightarrow> ('nat_set, 'a_set, 'd) NFA_impl"
    where "TM_translate_NFA_wrapper_impl r = TM_translate_NFA_impl r 0"

    lemma TM_translate_NFA_impl_correct:
      "nfa_invar (TM_translate_NFA_impl r n)"
      "nfa_\<alpha> (TM_translate_NFA_impl r n) = TM_translate_NFA r n"
    proof (induct r arbitrary: n)
      case Zero
        case 1 show ?case by (auto simp: TM_zero_NFA_impl_correct)
        case 2 show ?case by (auto simp: TM_zero_NFA_impl_correct)
      case One
        case 1 show ?case by (auto simp: TM_one_NFA_impl_correct)
        case 2 show ?case by (auto simp: TM_one_NFA_impl_correct)
      case (Atom a)
        case 1 show ?case by (auto simp: TM_atom_NFA_impl_correct)
        case 2 show ?case by (auto simp: TM_atom_NFA_impl_correct)
      case (Plus r1 r2)
        case 1 show ?case using Plus by (auto simp: Let_def TM_plus_NFA_impl_correct)
        case 2
        have a: "s.size (nfa_states (TM_translate_NFA_impl r1 n)) = card (\<Q> (TM_translate_NFA r1 n))"
          unfolding s.size_correct[OF nfa_invar_consequences(1)[OF Plus(1)]]
          unfolding nfa_\<alpha>_simps(1)[symmetric]
          unfolding Plus(2)
          by (rule refl)
        show ?case using Plus a by (auto simp: Let_def TM_plus_NFA_impl_correct)
      case (Times r1 r2)
        case 1 show ?case using Times by (auto simp: Let_def TM_times_NFA_impl_correct)
        case 2
        have a: "s.size (nfa_states (TM_translate_NFA_impl r1 n)) = card (\<Q> (TM_translate_NFA r1 n))"
          unfolding s.size_correct[OF nfa_invar_consequences(1)[OF Times(1)]]
          unfolding nfa_\<alpha>_simps(1)[symmetric]
          unfolding Times(2)
          by (rule refl)
        show ?case using Times a by (auto simp: Let_def TM_times_NFA_impl_correct)
      case (Star r)
        case 1 show ?case using Star by (auto simp: Let_def TM_star_NFA_impl_correct)
        case 2
        have a: "s.size (nfa_states (TM_translate_NFA_impl r n)) = card (\<Q> (TM_translate_NFA r n))"
          unfolding s.size_correct[OF nfa_invar_consequences(1)[OF Star(1)]]
          unfolding nfa_\<alpha>_simps(1)[symmetric]
          unfolding Star(2)
          by (rule refl)
        show ?case using Star a by (auto simp: Let_def TM_star_NFA_impl_correct)
    qed

    lemma TM_translate_NFA_wrapper_impl_correct:
      "nfa_invar (TM_translate_NFA_wrapper_impl r)"
      "nfa_\<alpha> (TM_translate_NFA_wrapper_impl r) = TM_translate_NFA_wrapper r"
      using TM_translate_NFA_impl_correct[of r 0]
      by auto

  end

  definition "rs_TM_bridges_impl \<equiv> rs_Thompson_impl.TM_bridges_impl"
  definition "rs_TM_zero_NFA_impl \<equiv> rs_Thompson_impl.TM_zero_NFA_impl"
  definition "rs_TM_one_NFA_impl \<equiv> rs_Thompson_impl.TM_one_NFA_impl"
  definition "rs_TM_atom_NFA_impl \<equiv> rs_Thompson_impl.TM_atom_NFA_impl"
  definition "rs_TM_plus_NFA_impl \<equiv> rs_Thompson_impl.TM_plus_NFA_impl"
  definition "rs_TM_times_NFA_impl \<equiv> rs_Thompson_impl.TM_times_NFA_impl"
  definition "rs_TM_transcl_NFA_impl \<equiv> rs_Thompson_impl.TM_transcl_NFA_impl"
  definition "rs_TM_reflcl_NFA_impl \<equiv> rs_Thompson_impl.TM_reflcl_NFA_impl"
  definition "rs_TM_star_NFA_impl \<equiv> rs_Thompson_impl.TM_star_NFA_impl"
  definition "rs_TM_translate_NFA_impl \<equiv> rs_Thompson_impl.TM_translate_NFA_impl"
  definition "rs_TM_translate_NFA_wrapper_impl \<equiv> rs_Thompson_impl.TM_translate_NFA_wrapper_impl"

  lemmas rs_defs =
    rs_nfa_defs
    rs_TM_bridges_impl_def
    rs_TM_zero_NFA_impl_def
    rs_TM_one_NFA_impl_def
    rs_TM_atom_NFA_impl_def
    rs_TM_plus_NFA_impl_def
    rs_TM_times_NFA_impl_def
    rs_TM_transcl_NFA_impl_def
    rs_TM_reflcl_NFA_impl_def
    rs_TM_star_NFA_impl_def
    rs_TM_translate_NFA_impl_def
    rs_TM_translate_NFA_wrapper_impl_def

  lemmas [code] = 
    rs_Thompson_impl.TM_bridges_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_zero_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_one_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_atom_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_plus_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_times_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_transcl_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_reflcl_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_star_NFA_impl_def[folded rs_defs]
    rs_Thompson_impl.TM_translate_NFA_impl.simps[folded rs_defs]
    rs_Thompson_impl.TM_translate_NFA_wrapper_impl.simps[folded rs_defs]

  definition "rs_TM_translate_NFA_wrapper_normalize_impl r \<equiv>
    rs_nfa_normalise (rs_TM_translate_NFA_wrapper_impl r)"

  theorem rs_TM_translate_NFA_wrapper_correct:
    "rs_nfa_invar (rs_TM_translate_NFA_wrapper_impl r)"
    "\<L> (rs_nfa_\<alpha> (rs_TM_translate_NFA_wrapper_impl r)) = lang r"
    unfolding rs_nfa_invar_def rs_nfa_\<alpha>_def
    unfolding rs_TM_translate_NFA_wrapper_impl_def
    using TM_translate_NFA_wrapper_correct(2)
    using rs_Thompson_impl.TM_translate_NFA_wrapper_impl_correct[of r]
    by auto

  theorem rs_TM_translate_NFA_wrapper_normalize_correct:
    "rs_nfa_invar (rs_TM_translate_NFA_wrapper_normalize_impl r)"
    "\<L> (rs_nfa_\<alpha> (rs_TM_translate_NFA_wrapper_normalize_impl r)) = lang r"
    unfolding rs_TM_translate_NFA_wrapper_normalize_impl_def
    using rs_TM_translate_NFA_wrapper_correct(1)
    using rs_nfa.normalise_correct(1)
    apply metis
    using rs_TM_translate_NFA_wrapper_correct
    using NFA_isomorphic_wf_\<L> NFA_remove_unreachable_states_\<L>
    using rs_nfa.normalise_correct(2)
    apply metis
  done

end
