theory TM_Plus
imports
  TM_Definitions
  NFA_Extensions
  LTS_Extensions
begin

  lemma TM_plus_NFA_simps[simp]:
    "\<Q> (TM_plus_NFA \<A>1 \<A>2) = \<Q> \<A>1 \<union> \<Q> \<A>2"
    "\<Sigma> (TM_plus_NFA \<A>1 \<A>2) = \<Sigma> \<A>1 \<union> \<Sigma> \<A>2"
    "\<Delta> (TM_plus_NFA \<A>1 \<A>2) = \<Delta> \<A>1 \<union> \<Delta> \<A>2"
    "\<I> (TM_plus_NFA \<A>1 \<A>2) = \<I> \<A>1 \<union> \<I> \<A>2"
    "\<F> (TM_plus_NFA \<A>1 \<A>2) = \<F> \<A>1 \<union> \<F> \<A>2"
    "TM_plus_NFA TM_zero_NFA \<A> = \<A>"
    "TM_plus_NFA \<A> TM_zero_NFA = \<A>"
    unfolding TM_plus_NFA_def TM_zero_NFA_def
    by auto

  lemma TM_plus_NFA_deg_idempotence[simp]: "TM_plus_NFA \<A> \<A> = \<A>" by simp
  lemma TM_plus_NFA_associative:
    "TM_plus_NFA (TM_plus_NFA \<A>1 \<A>2) \<A>3 = TM_plus_NFA \<A>1 (TM_plus_NFA \<A>2 \<A>3)"
    unfolding TM_plus_NFA_def
    by auto
  lemma TM_plus_NFA_commutative: "TM_plus_NFA \<A>1 \<A>2 = TM_plus_NFA \<A>2 \<A>1"
    unfolding TM_plus_NFA_def
    by auto

  lemma TM_plus_NFA_states[simp]:
    assumes "\<Q> \<A>1 = {n ..< n + m1}" "\<Q> \<A>2 = {n + m1 ..< n + m1 + m2}"
    shows "\<Q> (TM_plus_NFA \<A>1 \<A>2 :: (nat, 'a) NFA_rec) = {n ..< n + m1 + m2}"
    using assms ivl_disj_un(17)
    by auto

  lemma TM_plus_NFA_well_formed: "NFA \<A>1 \<Longrightarrow> NFA \<A>2 \<Longrightarrow> NFA (TM_plus_NFA \<A>1 \<A>2)"
    unfolding NFA_full_def
    by auto

  lemma (in disjoint_NFAs) TM_plus_NFA_no_empty_word:
    "\<I> \<A>1 \<inter> \<F> \<A>1 = {} \<and> \<I> \<A>2 \<inter> \<F> \<A>2 = {} \<longleftrightarrow>
      \<I> (TM_plus_NFA \<A>1 \<A>2) \<inter> \<F> (TM_plus_NFA \<A>1 \<A>2) = {}"
    using disjoint_\<Q>
    using NFA_\<A>1.\<I>_consistent NFA_\<A>1.\<F>_consistent
    using NFA_\<A>2.\<I>_consistent NFA_\<A>2.\<F>_consistent
    apply auto
  done

  lemma (in disjoint_NFAs) TM_plus_NFA_empty_word:
    "\<I> \<A>1 \<inter> \<F> \<A>1 \<noteq> {} \<or> \<I> \<A>2 \<inter> \<F> \<A>2 \<noteq> {} \<longleftrightarrow>
      \<I> (TM_plus_NFA \<A>1 \<A>2) \<inter> \<F> (TM_plus_NFA \<A>1 \<A>2) \<noteq> {}"
    using disjoint_\<Q>
    using NFA_\<A>1.\<I>_consistent NFA_\<A>1.\<F>_consistent
    using NFA_\<A>2.\<I>_consistent NFA_\<A>2.\<F>_consistent
    apply auto
  done

  lemma (in disjoint_NFAs) TM_plus_NFA_no_bridge:
    assumes "LTS_is_fin_path (\<Delta> (TM_plus_NFA \<A>1 \<A>2)) f ts l" "f \<in> \<Q> \<A>1"
    shows "LTS_is_fin_path (\<Delta> \<A>1) f ts l"
  using assms
  proof (induct rule: LTS_is_fin_path_induct)
    case (Reflexive l)
    show ?case by simp
  next
    case (Step p a q ts l)
    have a: "(p, a, q) \<in> \<Delta> \<A>1"
      using Step(1) Step(4) NFA_\<A>2.\<Delta>_subset disjoint_\<Q> by auto
    have b: "q \<in> \<Q> \<A>1" using a NFA_\<A>1.\<Delta>_subset by auto
    show ?case using a Step(3)[OF b] by auto
  qed

  lemma (in disjoint_NFAs) TM_plus_NFA_language: "\<L> (TM_plus_NFA \<A>1 \<A>2) = \<L> \<A>1 \<union> \<L> \<A>2"
  proof
    have a: "\<L> \<A>1 \<subseteq> \<L> (TM_plus_NFA \<A>1 \<A>2)"
      using NFA_language_mono[of "\<A>1" "TM_plus_NFA \<A>1 \<A>2"]
      unfolding TM_plus_NFA_def
      by auto
    have b: "\<L> \<A>2 \<subseteq> \<L> (TM_plus_NFA \<A>1 \<A>2)"
      using NFA_language_mono[of "\<A>2" "TM_plus_NFA \<A>1 \<A>2"]
      unfolding TM_plus_NFA_def
      by auto
    show "\<L> \<A>1 \<union> \<L> \<A>2 \<subseteq> \<L> (TM_plus_NFA \<A>1 \<A>2)" using a b by auto
    show "\<L> (TM_plus_NFA \<A>1 \<A>2) \<subseteq> \<L> \<A>1 \<union> \<L> \<A>2"
    proof
      fix w
      assume a: "w \<in> \<L> (TM_plus_NFA \<A>1 \<A>2)"
      have b: "NFA_accept (TM_plus_NFA \<A>1 \<A>2) w" using a unfolding \<L>_def by auto
      obtain f l where f:
        "LTS_is_reachable (\<Delta> (TM_plus_NFA \<A>1 \<A>2)) f w l"
        "f \<in> \<I> (TM_plus_NFA \<A>1 \<A>2)"
        "l \<in> \<F> (TM_plus_NFA \<A>1 \<A>2)"
        using b unfolding NFA_accept_def by auto
      obtain ts where g:
        "LTS_is_fin_path (\<Delta> (TM_plus_NFA \<A>1 \<A>2)) f ts l"
        "LTS_fin_path_word ts = w"
        using f(1) unfolding LTS_is_reachable_fin_path by auto
      have h: "f \<in> \<Q> \<A>1 \<or> f \<in> \<Q> \<A>2"
        using f(2) NFA_\<A>1.\<I>_consistent NFA_\<A>2.\<I>_consistent by auto
      have i: "NFA_accept \<A>1 w \<or> NFA_accept \<A>2 w" using h
      proof
        assume j: "f \<in> \<Q> \<A>1"
        have k: "LTS_is_fin_path (\<Delta> \<A>1) f ts l"
          using TM_plus_NFA_no_bridge[OF g(1) j] by this
        have l: "l \<in> \<Q> \<A>1"
          using NFA_\<A>1.LTS_fin_path_consistent_l[OF k j] by this
        have m: "f \<in> \<I> \<A>1" "l \<in> \<F> \<A>1"
          using j l f(2) f(3)
          using NFA_\<A>2.\<I>_consistent NFA_\<A>2.\<F>_consistent disjoint_\<Q>
          by auto
        have n: "LTS_is_reachable (\<Delta> \<A>1) f w l"
          using k g(2) unfolding LTS_is_reachable_fin_path by auto
        show "NFA_accept \<A>1 w \<or> NFA_accept \<A>2 w"
          using n m unfolding NFA_accept_def by auto
      next
        assume j: "f \<in> \<Q> \<A>2"
        have k: "LTS_is_fin_path (\<Delta> \<A>2) f ts l"
          using reversed.TM_plus_NFA_no_bridge g(1) j TM_plus_NFA_commutative[of \<A>1 \<A>2]
          by simp
        have l: "l \<in> \<Q> \<A>2"
          using NFA_\<A>2.LTS_fin_path_consistent_l[OF k j] by this
        have m: "f \<in> \<I> \<A>2" "l \<in> \<F> \<A>2"
          using j l f(2) f(3)
          using NFA_\<A>1.\<I>_consistent NFA_\<A>1.\<F>_consistent disjoint_\<Q>
          by auto
        have n: "LTS_is_reachable (\<Delta> \<A>2) f w l"
          using k g(2) unfolding LTS_is_reachable_fin_path by auto
        show "NFA_accept \<A>1 w \<or> NFA_accept \<A>2 w"
          using n m unfolding NFA_accept_def by auto
      qed
      show "w \<in> \<L> \<A>1 \<union> \<L> \<A>2" using i unfolding \<L>_def by auto
    qed
  qed

end
