theory TM_Simple
imports
  TM_Definitions
begin

  lemma TM_zero_NFA_simps[simp]:
    "\<Q> TM_zero_NFA = {}"
    "\<Sigma> TM_zero_NFA = {}"
    "\<Delta> TM_zero_NFA = {}"
    "\<I> TM_zero_NFA = {}"
    "\<F> TM_zero_NFA = {}"
    unfolding TM_zero_NFA_def
    by auto

  lemma TM_one_NFA_simps[simp]:
    "\<Q> (TM_one_NFA p) = {p}"
    "\<Sigma> (TM_one_NFA p) = {}"
    "\<Delta> (TM_one_NFA p) = {}"
    "\<I> (TM_one_NFA p) = {p}"
    "\<F> (TM_one_NFA p) = {p}"
    unfolding TM_one_NFA_def
    by auto

  lemma TM_atom_NFA_simps[simp]:
    "\<Q> (TM_atom_NFA p a q) = {p, q}"
    "\<Sigma> (TM_atom_NFA p a q) = {a}"
    "\<Delta> (TM_atom_NFA p a q) = {(p, a, q)}"
    "\<I> (TM_atom_NFA p a q) = {p}"
    "\<F> (TM_atom_NFA p a q) = {q}"
    unfolding TM_atom_NFA_def
    by auto

  lemma TM_zero_NFA_states[simp]: "\<Q> (TM_zero_NFA :: (nat, 'a) NFA_rec) = {n ..< n}"
    by simp
  lemma TM_one_NFA_states[simp]: "\<Q> (TM_one_NFA n :: (nat, 'a) NFA_rec) = {n ..< n + 1}"
    by simp
  lemma TM_atom_NFA_states[simp]: "\<Q> (TM_atom_NFA n a (Suc n)) = {n ..< n + 2}"
    by auto

  lemma TM_zero_NFA_well_formed: "NFA TM_zero_NFA" by auto
  lemma TM_one_NFA_well_formed: "NFA (TM_one_NFA n)" by auto
  lemma TM_atom_NFA_well_formed: "NFA (TM_atom_NFA p a q)" by auto

  lemma TM_zero_NFA_language: "\<L> TM_zero_NFA = {}" by (simp add: \<L>_def NFA_accept_def)
  lemma TM_one_NFA_language: "\<L> (TM_one_NFA n) = {[]}"
  proof
    show "\<L> (TM_one_NFA n) \<subseteq> {[]}"
    proof
      fix w :: "'c list"
      assume "w \<in> \<L> (TM_one_NFA n)"
      hence "NFA_accept (TM_one_NFA n) w" unfolding \<L>_def by simp
      hence "LTS_is_reachable {} n w n" unfolding NFA_accept_def by simp
      hence "w = []" by (cases w, simp+)
      thus "w \<in> {[]}" by simp
    qed
    show "{[]} \<subseteq> \<L> (TM_one_NFA n)" by (simp add: \<L>_def NFA_accept_def)
  qed
  lemma TM_atom_NFA_language: "p \<noteq> q \<Longrightarrow> \<L> (TM_atom_NFA p a q) = {[a]}"
  proof
    assume ne_p_q: "p \<noteq> q"
    show "\<L> (TM_atom_NFA p a q) \<subseteq> {[a]}"
    proof
      fix w
      assume "w \<in> \<L> (TM_atom_NFA p a q)"
      hence "NFA_accept (TM_atom_NFA p a q) w" unfolding \<L>_def by simp
      hence "LTS_is_reachable {(p, a, q)} p w q" unfolding NFA_accept_def by simp
      hence "w = [a]"
      proof (cases w)
        case Nil
        thus "LTS_is_reachable {(p, a, q)} p w q \<Longrightarrow> w = [a]"
          using ne_p_q by auto
      next
        case (Cons a' w')
        thus "LTS_is_reachable {(p, a, q)} p w q \<Longrightarrow> w = [a]"
          using ne_p_q by (cases w', auto)
      qed
      thus "w \<in> {[a]}" by simp
    qed
    show "{[a]} \<subseteq> \<L> (TM_atom_NFA p a q)" by (simp add: \<L>_def NFA_accept_def)
  qed

end
