theory TM_Star
imports
  TM_Definitions
  TM_Simple
  TM_Plus
  TM_Times
  Regular_Set_Extensions
begin

  lemmas TM_transcl_NFA_simps[simp] = TM_transcl_NFA_def
  lemma TM_reflcl_NFA_simps1[simp]:
    assumes "\<I> \<A> \<inter> \<F> \<A> = {}"
    shows
      "\<Q> (TM_reflcl_NFA \<A> n) = insert n (\<Q> \<A>)"
      "\<Sigma> (TM_reflcl_NFA \<A> n) = \<Sigma> \<A>"
      "\<Delta> (TM_reflcl_NFA \<A> n) = \<Delta> \<A>"
      "\<I> (TM_reflcl_NFA \<A> n) = insert n (\<I> \<A>)"
      "\<F> (TM_reflcl_NFA \<A> n) = insert n (\<F> \<A>)"
    using assms
    unfolding TM_reflcl_NFA_def
    by auto
  lemma TM_reflcl_NFA_simps2[simp]:
    assumes "\<I> \<A> \<inter> \<F> \<A> \<noteq> {}"
    shows "TM_reflcl_NFA \<A> n = \<A>"
    using assms
    unfolding TM_reflcl_NFA_def
    by auto
  lemmas TM_star_NFA_simps[simp] = TM_star_NFA_def

  lemma (in NFA) TM_transcl_NFA_well_formed: "NFA (TM_transcl_NFA \<A>)"
    unfolding TM_transcl_NFA_def
    using TM_times_NFA_well_formed[OF NFA_axioms NFA_axioms]
    by this
  lemma (in NFA) TM_reflcl_NFA_well_formed: "NFA (TM_reflcl_NFA \<A> n)"
    unfolding TM_reflcl_NFA_def
    apply (cases "\<I> \<A> \<inter> \<F> \<A> = {}")

    apply simp
    using TM_plus_NFA_well_formed[OF NFA_axioms TM_one_NFA_well_formed[of n]]
    apply this

    apply simp
    using TM_plus_NFA_well_formed[OF NFA_axioms TM_zero_NFA_well_formed]
    apply this
  done
  lemma (in NFA) TM_star_NFA_well_formed: "NFA (TM_star_NFA \<A> n)"
    unfolding TM_star_NFA_def
    using NFA.TM_reflcl_NFA_well_formed[OF TM_transcl_NFA_well_formed]
    by this

  lemma TM_transcl_NFA_no_empty_word:
    "\<I> \<A> \<inter> \<F> \<A> = {} \<longleftrightarrow> \<I> (TM_transcl_NFA \<A>) \<inter> \<F> (TM_transcl_NFA \<A>) = {}"
    unfolding TM_transcl_NFA_def
    by simp
  lemma TM_transcl_NFA_empty_word:
    "\<I> \<A> \<inter> \<F> \<A> \<noteq> {} \<longleftrightarrow> \<I> (TM_transcl_NFA \<A>) \<inter> \<F> (TM_transcl_NFA \<A>) \<noteq> {}"
    unfolding TM_transcl_NFA_def
    by simp

  lemma TM_reflcl_NFA_empty_word:
    "\<I> (TM_reflcl_NFA \<A> n) \<inter> \<F> (TM_reflcl_NFA \<A> n) \<noteq> {}"
    unfolding TM_reflcl_NFA_def
    by (cases "\<I> \<A> \<inter> \<F> \<A> = {}", auto)

  lemma TM_star_NFA_empty_word:
    "\<I> (TM_star_NFA \<A> n) \<inter> \<F> (TM_star_NFA \<A> n) \<noteq> {}"
    apply simp
    using TM_reflcl_NFA_empty_word[of "TM_times_NFA \<A> \<A>"]
    apply this
  done

  lemma TM_transcl_NFA_language: "\<L> (TM_transcl_NFA \<A>) = transcl (\<L> \<A>)"
    unfolding TM_transcl_NFA_def
    using TM_times_NFA_deg_language[of \<A>]
    by this
  lemma (in NFA) TM_reflcl_NFA_language:
    assumes "n \<notin> \<Q> \<A>"
    shows "\<L> (TM_reflcl_NFA \<A> n) = reflcl (\<L> \<A>)"
  proof -
    interpret disjoint_zero: disjoint_NFAs "\<A>" "TM_zero_NFA"
      unfolding disjoint_NFAs_def disjoint_NFAs_axioms_def
      using NFA_axioms TM_zero_NFA_well_formed assms
      by simp
    interpret disjoint_one: disjoint_NFAs "\<A>" "TM_one_NFA n"
      unfolding disjoint_NFAs_def disjoint_NFAs_axioms_def
      using NFA_axioms TM_one_NFA_well_formed[of n] assms
      by simp
    show ?thesis
      apply (cases "\<I> \<A> \<inter> \<F> \<A> = {}")

      unfolding TM_reflcl_NFA_def reflcl_alt_def
      apply simp
      unfolding disjoint_one.TM_plus_NFA_language TM_one_NFA_language
      apply simp

      unfolding TM_reflcl_NFA_def reflcl_alt_def
      apply simp
      unfolding disjoint_zero.TM_plus_NFA_language TM_zero_NFA_language
      unfolding NFA_empty_word[of \<A>]
      apply auto
    done
  qed
  lemma (in NFA) TM_star_NFA_language:
    assumes "n \<notin> \<Q> \<A>"
    shows "\<L> (TM_star_NFA \<A> n) = star (\<L> \<A>)"
  proof -
    interpret NFA_TM_transcl_NFA: NFA "TM_transcl_NFA \<A>"
      using TM_transcl_NFA_well_formed
      by this
    have a: "n \<notin> \<Q> (TM_transcl_NFA \<A>)" using assms by simp
    show ?thesis
      unfolding TM_star_NFA_def
      unfolding NFA_TM_transcl_NFA.TM_reflcl_NFA_language[OF a]
      unfolding TM_transcl_NFA_language
      unfolding reflcl_transcl_eq_star
      by simp
  qed

end
