theory TM_Times
imports
  TM_Definitions
  NFA_Extensions
  LTS_Extensions
  Regular_Set_Extensions
begin

  lemma TM_times_NFA_simps[simp]:
    "\<Q> (TM_times_NFA \<A>1 \<A>2) = \<Q> \<A>1 \<union> \<Q> \<A>2"
    "\<Sigma> (TM_times_NFA \<A>1 \<A>2) = \<Sigma> \<A>1 \<union> \<Sigma> \<A>2"
    "\<Delta> (TM_times_NFA \<A>1 \<A>2) = \<Delta> \<A>1 \<union> \<Delta> \<A>2 \<union> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2)"
    "\<I> (TM_times_NFA \<A>1 \<A>2) = \<I> \<A>1"
    "\<I> \<A>2 \<inter> \<F> \<A>2 = {} \<Longrightarrow> \<F> (TM_times_NFA \<A>1 \<A>2) = \<F> \<A>2"
    "\<I> \<A>2 \<inter> \<F> \<A>2 \<noteq> {} \<Longrightarrow> \<F> (TM_times_NFA \<A>1 \<A>2) = \<F> \<A>1 \<union> \<F> \<A>2"
    unfolding TM_times_NFA_def
    by auto

  lemma TM_times_NFA_deg_simps[simp]:
    "\<Q> (TM_times_NFA \<A> \<A>) = \<Q> \<A>"
    "\<Sigma> (TM_times_NFA \<A> \<A>) = \<Sigma> \<A>"
    "\<Delta> (TM_times_NFA \<A> \<A>) = \<Delta> \<A> \<union> TM_bridges (\<F> \<A>) (\<I> \<A>) (\<Delta> \<A>)"
    "\<I> (TM_times_NFA \<A> \<A>) = \<I> \<A>"
    "\<F> (TM_times_NFA \<A> \<A>) = \<F> \<A>"
    unfolding TM_times_NFA_def
    by auto

  lemma TM_times_NFA_absorption[simp]: "TM_times_NFA \<A> (TM_times_NFA \<A> \<A>) = TM_times_NFA \<A> \<A>"
    unfolding TM_times_NFA_def TM_bridges_def
    by auto

  lemma TM_times_NFA_states[simp]:
    assumes "\<Q> \<A>1 = {n ..< n + m1}" "\<Q> \<A>2 = {n + m1 ..< n + m1 + m2}"
    shows "\<Q> (TM_times_NFA \<A>1 \<A>2 :: (nat, 'a) NFA_rec) = {n ..< n + m1 + m2}"
    using assms ivl_disj_un(17)
    by auto

  lemma TM_times_NFA_well_formed:
    assumes "NFA \<A>1" "NFA \<A>2"
    shows "NFA (TM_times_NFA \<A>1 \<A>2)"
  proof -
    interpret \<A>1: NFA \<A>1 by fact
    interpret \<A>2: NFA \<A>2 by fact
    show ?thesis
    proof
      case (goal1 p a q)
      show ?case
        using goal1 \<A>1.\<Delta>_subset \<A>2.\<Delta>_subset \<A>1.\<F>_consistent
        by (auto simp add: TM_bridges_def)
    next
      case (goal2)
      show ?case using \<A>1.\<I>_consistent by auto
    next
      case (goal3)
      show ?case using \<A>1.\<F>_consistent \<A>2.\<F>_consistent by (cases "\<I> \<A>2 \<inter> \<F> \<A>2 = {}", auto)
    next
      case (goal4)
      show ?case using \<A>1.finite_\<Q> \<A>2.finite_\<Q> by auto
    next
      case (goal5)
      have 1: "finite (TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2))"
      proof (rule finite_subset)
        show "TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2) \<subseteq> \<Q> \<A>1 \<times> \<Sigma> \<A>2 \<times> \<Q> \<A>2"
          unfolding TM_bridges_def using \<A>1.\<F>_consistent \<A>2.\<Delta>_subset by auto
        show "finite (\<Q> \<A>1 \<times> \<Sigma> \<A>2 \<times> \<Q> \<A>2)"
          using \<A>1.finite_\<Q> \<A>2.finite_\<Sigma> \<A>2.finite_\<Q> by auto
      qed
      show ?case using \<A>1.finite_\<Delta> \<A>2.finite_\<Delta> 1 by auto
    next
      case (goal6)
      show ?case using \<A>1.finite_\<Sigma> \<A>2.finite_\<Sigma> by auto
    qed
  qed

  lemma (in disjoint_NFAs) TM_times_NFA_no_empty_word:
    "\<I> \<A>1 \<inter> \<F> \<A>1 = {} \<or> \<I> \<A>2 \<inter> \<F> \<A>2 = {} \<longleftrightarrow>
      \<I> (TM_times_NFA \<A>1 \<A>2) \<inter> \<F> (TM_times_NFA \<A>1 \<A>2) = {}"
    apply (cases "\<I> \<A>2 \<inter> \<F> \<A>2 = {}")
    using disjoint_\<Q> NFA_\<A>1.\<I>_consistent NFA_\<A>2.\<F>_consistent
    apply auto
  done

  lemma (in disjoint_NFAs) TM_times_NFA_empty_word:
    "\<I> \<A>1 \<inter> \<F> \<A>1 \<noteq> {} \<and> \<I> \<A>2 \<inter> \<F> \<A>2 \<noteq> {} \<longleftrightarrow>
      \<I> (TM_times_NFA \<A>1 \<A>2) \<inter> \<F> (TM_times_NFA \<A>1 \<A>2) \<noteq> {}"
    apply (cases "\<I> \<A>2 \<inter> \<F> \<A>2 = {}")
    using disjoint_\<Q> NFA_\<A>1.\<I>_consistent NFA_\<A>2.\<F>_consistent
    apply auto
  done

  lemma (in disjoint_NFAs) TM_times_NFA_step_\<A>1_\<A>1:
    assumes "(p, a, q) \<in> \<Delta> (TM_times_NFA \<A>1 \<A>2)" "p \<in> \<Q> \<A>1" "q \<in> \<Q> \<A>1"
    shows "(p, a, q) \<in> \<Delta> \<A>1"
  proof -
    have "(p, a, q) \<in> \<Delta> \<A>1 \<union> \<Delta> \<A>2 \<union> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2)"
      using assms(1) by auto
    thus ?thesis 
      using assms(2) assms(3)
      using NFA_\<A>1.\<Delta>_subset NFA_\<A>1.\<I>_consistent NFA_\<A>1.\<F>_consistent
      using NFA_\<A>2.\<Delta>_subset NFA_\<A>2.\<I>_consistent NFA_\<A>2.\<F>_consistent
      using disjoint_\<Q>
      unfolding TM_bridges_def
      by auto
  qed
  lemma (in disjoint_NFAs) TM_times_NFA_step_\<A>2_\<A>2:
    assumes "(p, a, q) \<in> \<Delta> (TM_times_NFA \<A>1 \<A>2)" "p \<in> \<Q> \<A>2" "q \<in> \<Q> \<A>2"
    shows "(p, a, q) \<in> \<Delta> \<A>2"
  proof -
    have "(p, a, q) \<in> \<Delta> \<A>1 \<union> \<Delta> \<A>2 \<union> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2)"
      using assms(1) by auto
    thus ?thesis
      using assms(2) assms(3)
      using NFA_\<A>1.\<Delta>_subset NFA_\<A>1.\<I>_consistent NFA_\<A>1.\<F>_consistent
      using NFA_\<A>2.\<Delta>_subset NFA_\<A>2.\<I>_consistent NFA_\<A>2.\<F>_consistent
      using disjoint_\<Q>
      unfolding TM_bridges_def
      by auto
  qed
  lemma (in disjoint_NFAs) TM_times_NFA_step_\<A>1_\<A>2:
    assumes "(p, a, q) \<in> \<Delta> (TM_times_NFA \<A>1 \<A>2)" "p \<in> \<Q> \<A>1" "q \<in> \<Q> \<A>2"
    shows "(p, a, q) \<in> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2)"
  proof -
    have "(p, a, q) \<in> \<Delta> \<A>1 \<union> \<Delta> \<A>2 \<union> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2)"
      using assms(1) by auto
    thus ?thesis
      using assms(2) assms(3)
      using NFA_\<A>1.\<Delta>_subset NFA_\<A>1.\<I>_consistent NFA_\<A>1.\<F>_consistent
      using NFA_\<A>2.\<Delta>_subset NFA_\<A>2.\<I>_consistent NFA_\<A>2.\<F>_consistent
      using disjoint_\<Q>
      unfolding TM_bridges_def
      by auto
  qed
  lemma (in disjoint_NFAs) TM_times_NFA_step_\<A>2_\<A>1:
    assumes "(p, a, q) \<in> \<Delta> (TM_times_NFA \<A>1 \<A>2)" "p \<in> \<Q> \<A>2" "q \<in> \<Q> \<A>1"
    shows "False"
  proof -
    have "(p, a, q) \<in> \<Delta> \<A>1 \<union> \<Delta> \<A>2 \<union> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2)"
      using assms(1) by auto
    thus ?thesis
      using assms(2) assms(3)
      using NFA_\<A>1.\<Delta>_subset NFA_\<A>1.\<I>_consistent NFA_\<A>1.\<F>_consistent
      using NFA_\<A>2.\<Delta>_subset NFA_\<A>2.\<I>_consistent NFA_\<A>2.\<F>_consistent
      using disjoint_\<Q>
      unfolding TM_bridges_def
      by auto
  qed

  lemma (in disjoint_NFAs) TM_times_NFA_path_\<A>2_\<A>1:
    assumes "LTS_is_fin_path (\<Delta> (TM_times_NFA \<A>1 \<A>2)) f ts l" "f \<in> \<Q> \<A>2" "l \<in> \<Q> \<A>1"
    shows "False"
  using assms
  proof (induct rule: LTS_is_fin_path_induct)
    case (Reflexive p)
    show ?case using Reflexive disjoint_\<Q> by auto
  next
    case (Step p a q ts l)
    interpret NFA_\<A>1_\<A>2: NFA "TM_times_NFA \<A>1 \<A>2"
      using TM_times_NFA_well_formed[OF NFA_\<A>1.NFA_axioms NFA_\<A>2.NFA_axioms]
      by this
    have a: "q \<in> \<Q> \<A>1 \<union> \<Q> \<A>2" using Step(1) NFA_\<A>1_\<A>2.\<Delta>_subset by auto
    show ?case using a
    proof
      assume a: "q \<in> \<Q> \<A>1"
      show ?thesis using TM_times_NFA_step_\<A>2_\<A>1[OF Step(1) Step(4) a] by this
    next
      assume a: "q \<in> \<Q> \<A>2"
      show ?thesis using Step(3)[OF a Step(5)] by this
    qed
  qed
  lemma (in disjoint_NFAs) TM_times_NFA_path_\<A>1_\<A>1:
    assumes "LTS_is_fin_path (\<Delta> (TM_times_NFA \<A>1 \<A>2)) f ts l" "f \<in> \<Q> \<A>1" "l \<in> \<Q> \<A>1"
    shows "LTS_is_fin_path (\<Delta> \<A>1) f ts l"
  using assms
  proof (induct rule: LTS_is_fin_path_induct)
    case (Reflexive p)
    show ?case by simp
  next
    case (Step p a q ts l)
    interpret NFA_\<A>1_\<A>2: NFA "TM_times_NFA \<A>1 \<A>2"
      using TM_times_NFA_well_formed[OF NFA_\<A>1.NFA_axioms NFA_\<A>2.NFA_axioms]
      by this
    have a: "q \<in> \<Q> \<A>1 \<union> \<Q> \<A>2" using Step(1) NFA_\<A>1_\<A>2.\<Delta>_subset by auto
    show ?case using a
    proof
      assume a: "q \<in> \<Q> \<A>1"
      show ?thesis
        using TM_times_NFA_step_\<A>1_\<A>1[OF Step(1) Step(4) a] Step(3)[OF a Step(5)]
        by simp
    next
      assume a: "q \<in> \<Q> \<A>2"
      show ?thesis using TM_times_NFA_path_\<A>2_\<A>1[OF Step(2) a Step(5)] by simp
    qed
  qed
  lemma (in disjoint_NFAs) TM_times_NFA_path_\<A>2_\<A>2:
    assumes "LTS_is_fin_path (\<Delta> (TM_times_NFA \<A>1 \<A>2)) f ts l" "f \<in> \<Q> \<A>2" "l \<in> \<Q> \<A>2"
    shows "LTS_is_fin_path (\<Delta> \<A>2) f ts l"
  using assms
  proof (induct rule: LTS_is_fin_path_induct)
    case (Reflexive p)
    show ?case by simp
  next
    case (Step p a q ts l)
    interpret NFA_\<A>1_\<A>2: NFA "TM_times_NFA \<A>1 \<A>2"
      using TM_times_NFA_well_formed[OF NFA_\<A>1.NFA_axioms NFA_\<A>2.NFA_axioms]
      by this
    have a: "q \<in> \<Q> \<A>1 \<union> \<Q> \<A>2" using Step(1) NFA_\<A>1_\<A>2.\<Delta>_subset by auto
    show ?case using a
    proof
      assume a: "q \<in> \<Q> \<A>1"
      show ?thesis using TM_times_NFA_step_\<A>2_\<A>1[OF Step(1) Step(4) a] by simp
    next
      assume a: "q \<in> \<Q> \<A>2"
      show ?thesis
        using TM_times_NFA_step_\<A>2_\<A>2[OF Step(1) Step(4) a] Step(3)[OF a Step(5)]
        by simp
    qed
  qed
  lemma (in disjoint_NFAs) TM_times_NFA_path_\<A>1_\<A>2:
    assumes "LTS_is_fin_path (\<Delta> (TM_times_NFA \<A>1 \<A>2)) f ts l" "f \<in> \<Q> \<A>1" "l \<in> \<Q> \<A>2"
    shows "\<exists> ts1 l' f' ts2.
      LTS_fin_path_word ts1 @ LTS_fin_path_word ts2 = LTS_fin_path_word ts \<and>
      LTS_is_fin_path (\<Delta> \<A>1) f ts1 l' \<and> l' \<in> \<F> \<A>1 \<and>
      LTS_is_fin_path (\<Delta> \<A>2) f' ts2 l \<and> f' \<in> \<I> \<A>2"
  using assms
  proof (induct rule: LTS_is_fin_path_induct)
    case (Reflexive p)
    show ?case using Reflexive disjoint_\<Q> by auto
  next
    case (Step p a q ts l)
    interpret NFA_\<A>1_\<A>2: NFA "TM_times_NFA \<A>1 \<A>2"
      using TM_times_NFA_well_formed[OF NFA_\<A>1.NFA_axioms NFA_\<A>2.NFA_axioms]
      by this
    have a: "q \<in> \<Q> \<A>1 \<union> \<Q> \<A>2" using Step(1) NFA_\<A>1_\<A>2.\<Delta>_subset by auto
    show ?case using a
    proof
      assume a: "q \<in> \<Q> \<A>1"
      obtain ts1 l' f' ts2
      where
        l1: "LTS_fin_path_word ts1 @ LTS_fin_path_word ts2 = LTS_fin_path_word ts" and
        l2: "LTS_is_fin_path (\<Delta> \<A>1) q ts1 l'" "l' \<in> \<F> \<A>1" and
        l3: "LTS_is_fin_path (\<Delta> \<A>2) f' ts2 l" "f' \<in> \<I> \<A>2"
        using Step(3)[OF a Step(5)] by auto
      have b: "LTS_fin_path_word ((p, a, q) # ts1) @ LTS_fin_path_word ts2 =
        LTS_fin_path_word ((p, a, q) # ts)" using l1 by auto
      have c: "LTS_is_fin_path (\<Delta> \<A>1) p ((p, a, q) # ts1) l'" "l' \<in> \<F> \<A>1"
        using TM_times_NFA_step_\<A>1_\<A>1[OF Step(1) Step(4) a] l2 by auto
      have d: "LTS_is_fin_path (\<Delta> \<A>2) f' ts2 l" "f' \<in> \<I> \<A>2" using l3 by this
      show ?thesis using b c d by blast
    next
      assume a: "q \<in> \<Q> \<A>2"
      have b: "(p, a, q) \<in> TM_bridges (\<F> \<A>1) (\<I> \<A>2) (\<Delta> \<A>2)"
        using TM_times_NFA_step_\<A>1_\<A>2[OF Step(1) Step(4) a] by this
      obtain p' where c: "(p', a, q) \<in> \<Delta> \<A>2" "p' \<in> \<I> \<A>2"
        using b unfolding TM_bridges_def by auto
      have d: "LTS_fin_path_word [] @ LTS_fin_path_word ((p', a, q) # ts) =
        LTS_fin_path_word ((p, a, q) # ts)" by auto
      have e: "LTS_is_fin_path (\<Delta> \<A>1) p [] p" "p \<in> \<F> \<A>1"
        using b unfolding TM_bridges_def by auto
      have f: "LTS_is_fin_path (\<Delta> \<A>2) p' ((p', a, q) # ts) l" "p' \<in> \<I> \<A>2"
        using c(1) TM_times_NFA_path_\<A>2_\<A>2[OF Step(2) a Step(5)] c(2) by auto
      show ?thesis using d e f by blast
    qed
  qed

  lemma TM_times_NFA_deg_path:
    assumes "LTS_is_fin_path (\<Delta> (TM_times_NFA \<A> \<A>)) f ts l" "l \<in> \<F> (TM_times_NFA \<A> \<A>)"
    shows "\<exists> ts' l' tss.
      LTS_is_fin_path (\<Delta> \<A>) f ts' l' \<and> l' \<in> \<F> \<A> \<and>
      LTS_fin_path_word ts' @ concat (map (LTS_fin_path_word \<circ> fst \<circ> snd) tss) =
        LTS_fin_path_word ts \<and>
      (\<forall> (f, ts, l) \<in> set tss. LTS_is_fin_path (\<Delta> \<A>) f ts l \<and> f \<in> \<I> \<A> \<and> l \<in> \<F> \<A>)"
  using assms
  proof (induct rule: LTS_is_fin_path_induct)
    case (Reflexive l)
    let ?ts' = "[]"
    let ?l' = "l"
    let ?tss = "[]"
    have a': "LTS_is_fin_path (\<Delta> \<A>) l ?ts' ?l'" "?l' \<in> \<F> \<A>"
      using Reflexive by auto
    have b': "LTS_fin_path_word ?ts' @ concat (map (LTS_fin_path_word \<circ> fst \<circ> snd) ?tss) =
      LTS_fin_path_word []" by auto
    have c': "\<And> f ts l. (f, ts, l) \<in> set ?tss \<Longrightarrow>
      LTS_is_fin_path (\<Delta> \<A>) f ts l \<and> f \<in> \<I> \<A> \<and> l \<in> \<F> \<A>" by auto
    show ?case using a' b' c' by fast
  next
    case (Step p a q ts l)
    obtain ts' l' tss
    where
      a: "LTS_is_fin_path (\<Delta> \<A>) q ts' l'" "l' \<in> \<F> \<A>" and
      b: "LTS_fin_path_word ts' @ concat (map (LTS_fin_path_word \<circ> fst \<circ> snd) tss) =
        LTS_fin_path_word ts" and
      c: "\<And> f ts l. (f, ts, l) \<in> set tss \<Longrightarrow>
        LTS_is_fin_path (\<Delta> \<A>) f ts l \<and> f \<in> \<I> \<A> \<and> l \<in> \<F> \<A>"
      using Step(3)[OF Step(4)] by fast
    have d: "(p, a, q) \<in> \<Delta> \<A> \<union> TM_bridges (\<F> \<A>) (\<I> \<A>) (\<Delta> \<A>)"
      using Step(1) unfolding TM_times_NFA_def by auto
    show ?case using d
    proof
      assume e: "(p, a, q) \<in> \<Delta> \<A>"
      let ?ts' = "(p, a, q) # ts'"
      let ?l' = "l'"
      let ?tss = "tss"
      have a': "LTS_is_fin_path (\<Delta> \<A>) p ?ts' ?l'" "?l' \<in> \<F> \<A>" using e a by auto
      have b': "LTS_fin_path_word ?ts' @ concat (map (LTS_fin_path_word \<circ> fst \<circ> snd) ?tss) =
        LTS_fin_path_word ((p, a, q) # ts)" using b by auto
      have c': "\<And> f ts l. (f, ts, l) \<in> set ?tss \<Longrightarrow>
        LTS_is_fin_path (\<Delta> \<A>) f ts l \<and> f \<in> \<I> \<A> \<and> l \<in> \<F> \<A>" using c by simp
      show ?thesis using a' b' c' by blast
    next
      assume e: "(p, a, q) \<in> TM_bridges (\<F> \<A>) (\<I> \<A>) (\<Delta> \<A>)"
      obtain p' where f: "p' \<in> \<I> \<A>" "(p', a, q) \<in> \<Delta> \<A>"
        using e unfolding TM_bridges_def by auto
      let ?ts' = "[]"
      let ?l' = "p"
      let ?tss = "(p', (p', a, q) # ts', l') # tss"
      have a': "LTS_is_fin_path (\<Delta> \<A>) p ?ts' ?l'" "?l' \<in> \<F> \<A>"
        using e unfolding TM_bridges_def by auto
      have b': "LTS_fin_path_word ?ts' @ concat (map (LTS_fin_path_word \<circ> fst \<circ> snd) ?tss) =
        LTS_fin_path_word ((p, a, q) # ts)" using b by simp
      have c': "\<And> f ts l. (f, ts, l) \<in> set ?tss \<Longrightarrow>
        LTS_is_fin_path (\<Delta> \<A>) f ts l \<and> f \<in> \<I> \<A> \<and> l \<in> \<F> \<A>" using f a c by auto
      show ?thesis using a' b' c' by blast
    qed
  qed

  lemma TM_times_NFA_\<L>_forward: "\<L> \<A>1 @@ \<L> \<A>2 \<subseteq> \<L> (TM_times_NFA \<A>1 \<A>2)"
  proof
    fix w
    assume a: "w \<in> \<L> \<A>1 @@ \<L> \<A>2"
    obtain w1 w2 where b: "w1 @ w2 = w" "w1 \<in> \<L> \<A>1" "w2 \<in> \<L> \<A>2"
      using a concE by auto
    show "w \<in> \<L> (TM_times_NFA \<A>1 \<A>2)"
    proof (cases w2)
      case Nil
      have c: "\<Delta> \<A>1 \<subseteq> \<Delta> (TM_times_NFA \<A>1 \<A>2)" by auto
      have d: "\<I> \<A>1 \<subseteq> \<I> (TM_times_NFA \<A>1 \<A>2)" by auto
      have e: "\<F> \<A>1 \<subseteq> \<F> (TM_times_NFA \<A>1 \<A>2)"
        using NFA_empty_word[of \<A>2] b(3) Nil by auto
      show ?thesis using NFA_language_mono[OF c d e] b(1) b(2) Nil by auto
    next
      case (Cons a w2')
      have c: "NFA_accept \<A>1 w1" using b(2) unfolding \<L>_def by auto
      obtain p1 q1 where d:
        "LTS_is_reachable (\<Delta> \<A>1) p1 w1 q1" "p1 \<in> \<I> \<A>1" "q1 \<in> \<F> \<A>1"
        using c unfolding NFA_accept_def by auto
      have e: "NFA_accept \<A>2 w2" using b(3) unfolding \<L>_def by auto
      obtain p2 q2 where f:
        "LTS_is_reachable (\<Delta> \<A>2) p2 w2 q2" "p2 \<in> \<I> \<A>2" "q2 \<in> \<F> \<A>2"
        using e unfolding NFA_accept_def by auto
      obtain p2' where g:
        "(p2, a, p2') \<in> \<Delta> \<A>2" "LTS_is_reachable (\<Delta> \<A>2) p2' w2' q2"
        using Cons f(1) by auto
      have h: "p1 \<in> \<I> (TM_times_NFA \<A>1 \<A>2)"
        using d(2) unfolding TM_times_NFA_def by auto
      have i: "q2 \<in> \<F> (TM_times_NFA \<A>1 \<A>2)"
        using f(3) unfolding TM_times_NFA_def by auto
      have j: "LTS_is_reachable (\<Delta> (TM_times_NFA \<A>1 \<A>2)) p1 w1 q1"
        using LTS_is_reachable_mono[OF _ d(1), of "\<Delta> (TM_times_NFA \<A>1 \<A>2)"]
        by auto
      have k: "(q1, a, p2') \<in> \<Delta> (TM_times_NFA \<A>1 \<A>2)" using d(3) g(1) f(2)
        unfolding TM_times_NFA_def TM_bridges_def by auto
      have l: "LTS_is_reachable (\<Delta> (TM_times_NFA \<A>1 \<A>2)) p2' w2' q2"
        using LTS_is_reachable_mono[OF _ g(2), of "\<Delta> (TM_times_NFA \<A>1 \<A>2)"]
        by auto
      have m: "LTS_is_reachable (\<Delta> (TM_times_NFA \<A>1 \<A>2)) p1 (w1 @ [a] @ w2') q2"
        using j k l LTS_is_reachable_concat by fastforce
      have m: "LTS_is_reachable (\<Delta> (TM_times_NFA \<A>1 \<A>2)) p1 w q2"
        using m b(1) Cons by auto
      have m: "NFA_accept (TM_times_NFA \<A>1 \<A>2) w"
        using h i m unfolding NFA_accept_def by auto
      show ?thesis using m unfolding \<L>_def by auto
    qed
  qed

  lemma (in disjoint_NFAs) TM_times_NFA_\<L>_reverse: "\<L> (TM_times_NFA \<A>1 \<A>2) \<subseteq> \<L> \<A>1 @@ \<L> \<A>2"
  proof
    fix w
    assume a: "w \<in> \<L> (TM_times_NFA \<A>1 \<A>2)"
    have b: "NFA_accept (TM_times_NFA \<A>1 \<A>2) w" using a unfolding \<L>_def by simp
    obtain f l where c:
      "f \<in> \<I> (TM_times_NFA \<A>1 \<A>2)"
      "l \<in> \<F> (TM_times_NFA \<A>1 \<A>2)"
      "LTS_is_reachable (\<Delta> (TM_times_NFA \<A>1 \<A>2)) f w l"
      using b unfolding NFA_accept_def by auto
    obtain ts where d:
      "LTS_is_fin_path (\<Delta> (TM_times_NFA \<A>1 \<A>2)) f ts l"
      "LTS_fin_path_word ts = w"
      using c(3) unfolding LTS_is_reachable_fin_path by auto
    have h: "f \<in> \<I> \<A>1" using c(1) by auto
    have j: "f \<in> \<Q> \<A>1" using h NFA_\<A>1.\<I>_consistent by auto
    have e: "l \<in> \<F> \<A>1 \<union> \<F> \<A>2" using c(2) by (cases "\<I> \<A>2 \<inter> \<F> \<A>2 \<noteq> {}", auto)
    show "w \<in> \<L> \<A>1 @@ \<L> \<A>2" using e
    proof
      assume f: "l \<in> \<F> \<A>1"
      have k: "l \<in> \<Q> \<A>1" using f NFA_\<A>1.\<F>_consistent by auto
      have r: "LTS_is_fin_path (\<Delta> \<A>1) f ts l"
        using TM_times_NFA_path_\<A>1_\<A>1[OF d(1) j k] by this
      have o: "LTS_is_reachable (\<Delta> \<A>1) f w l"
        using r d(2) unfolding LTS_is_reachable_fin_path by auto
      have p: "NFA_accept \<A>1 w"
        using f h o unfolding NFA_accept_def by auto
      have m: "\<I> \<A>2 \<inter> \<F> \<A>2 \<noteq> {}" using f c(2) disjoint_\<F> by auto
      have n: "NFA_accept \<A>2 []" using m unfolding NFA_accept_def by auto
      have q: "w \<in> \<L> \<A>1" "[] \<in> \<L> \<A>2" using p n unfolding \<L>_def by auto
      show ?thesis using concI[OF q] by simp
    next
      assume f: "l \<in> \<F> \<A>2"
      have k: "l \<in> \<Q> \<A>2" using f NFA_\<A>2.\<F>_consistent by auto
      obtain ts1 l' f' ts2
      where
        l1: "LTS_fin_path_word ts1 @ LTS_fin_path_word ts2 = LTS_fin_path_word ts" and
        l2: "LTS_is_fin_path (\<Delta> \<A>1) f ts1 l'" "l' \<in> \<F> \<A>1" and
        l3: "LTS_is_fin_path (\<Delta> \<A>2) f' ts2 l" "f' \<in> \<I> \<A>2"
        using TM_times_NFA_path_\<A>1_\<A>2[OF d(1) j k] by auto
      have j: "LTS_is_reachable (\<Delta> \<A>1) f (LTS_fin_path_word ts1) l'"
        using l2(1) unfolding LTS_is_reachable_fin_path by auto
      have k: "LTS_is_reachable (\<Delta> \<A>2) f' (LTS_fin_path_word ts2) l"
        using l3(1) unfolding LTS_is_reachable_fin_path by auto
      have n: "NFA_accept \<A>1 (LTS_fin_path_word ts1)"
        using j h l2(2) unfolding NFA_accept_def by auto
      have o: "NFA_accept \<A>2 (LTS_fin_path_word ts2)"
        using k l3(2) f unfolding NFA_accept_def by auto
      have p: "LTS_fin_path_word ts1 \<in> \<L> \<A>1" "LTS_fin_path_word ts2 \<in> \<L> \<A>2"
        using n o unfolding \<L>_def by auto
      show ?thesis using concI[OF p] l1 d(2) by auto
    qed
  qed

  lemma (in disjoint_NFAs) TM_times_NFA_language: "\<L> (TM_times_NFA \<A>1 \<A>2) = \<L> \<A>1 @@ \<L> \<A>2"
  using TM_times_NFA_\<L>_forward
  using TM_times_NFA_\<L>_reverse
  by auto

  lemma TM_times_NFA_deg_language_mono: "\<L> \<A> \<subseteq> \<L> (TM_times_NFA \<A> \<A>)"
  using NFA_language_mono[of "\<A>" "TM_times_NFA \<A> \<A>"]
  by auto

  lemma TM_times_NFA_deg_language_forward: "transcl (\<L> \<A>) \<subseteq> \<L> (TM_times_NFA \<A> \<A>)"
  proof
    fix wc
    assume a: "wc \<in> transcl (\<L> \<A>)"
    show "wc \<in> \<L> (TM_times_NFA \<A> \<A>)" using a unfolding transcl_def
    proof
      fix n :: "nat"
      assume b: "n \<in> {1..}"
      assume c: "wc \<in> \<L> \<A> ^^ n"
      have d: "1 \<le> n" using b by simp
      show "wc \<in> \<L> (TM_times_NFA \<A> \<A>)" using d c
      proof (induct n arbitrary: wc)
        case 0
        show ?case using 0(1) by simp
      next
        case (Suc n)
        note Suc' = Suc
        show ?case
        proof (cases n)
          case 0
          show ?thesis using TM_times_NFA_deg_language_mono Suc'(3) 0 by auto
        next
          case (Suc n')
          have d: "1 \<le> n" using Suc by auto
          obtain w1 wc' where c: "w1 @ wc' = wc" "w1 \<in> \<L> \<A>" "wc' \<in> \<L> \<A> ^^ n"
            using Suc'(3) by auto
          have e: "wc' \<in> \<L> (TM_times_NFA \<A> \<A>)" using Suc Suc'(1)[OF d c(3)] by auto
          have f: "(w1 @ wc') \<in> \<L> (TM_times_NFA \<A> (TM_times_NFA \<A> \<A>))"
            using c(2) e TM_times_NFA_\<L>_forward[of "\<A>" "TM_times_NFA \<A> \<A>"] by auto
          show ?thesis using f c(1) by simp
        qed
      qed
    qed
  qed

  lemma TM_times_NFA_deg_language_reverse: "\<L> (TM_times_NFA \<A> \<A>) \<subseteq> transcl (\<L> \<A>)"
  proof
    fix wc
    assume b: "wc \<in> \<L> (TM_times_NFA \<A> \<A>)"
    have c: "NFA_accept (TM_times_NFA \<A> \<A>) wc" using b unfolding \<L>_def by auto
    obtain f l where d:
      "f \<in> \<I> (TM_times_NFA \<A> \<A>)"
      "l \<in> \<F> (TM_times_NFA \<A> \<A>)"
      "LTS_is_reachable (\<Delta> (TM_times_NFA \<A> \<A>)) f wc l"
      using c unfolding NFA_accept_def by auto
    obtain ts where e:
      "LTS_is_fin_path (\<Delta> (TM_times_NFA \<A> \<A>)) f ts l"
      "LTS_fin_path_word ts = wc"
      using d unfolding LTS_is_reachable_fin_path by auto
    obtain ts' l' tss
      where
      g1: "LTS_is_fin_path (\<Delta> \<A>) f ts' l'" "l' \<in> \<F> \<A>" and
      g2: "LTS_fin_path_word ts' @ concat (map (LTS_fin_path_word \<circ> fst \<circ> snd) tss) =
        LTS_fin_path_word ts" and
      g3: "\<And> f ts l. (f, ts, l) \<in> set tss \<Longrightarrow>
        LTS_is_fin_path (\<Delta> \<A>) f ts l \<and> f \<in> \<I> \<A> \<and> l \<in> \<F> \<A>"
      using TM_times_NFA_deg_path[OF e(1) d(2)] by fast
    have i:
      "LTS_is_reachable (\<Delta> \<A>) f (LTS_fin_path_word ts') l'"
      "f \<in> \<I> \<A>"
      "l' \<in> \<F> \<A>"
      using g1 d(1) unfolding LTS_is_reachable_fin_path by auto
    have i: "NFA_accept \<A> (LTS_fin_path_word ts')"
      using i unfolding NFA_accept_def by auto
    have i: "LTS_fin_path_word ts' \<in> \<L> \<A>" using i unfolding \<L>_def by auto
    have h: "\<And> f ts l. (f, ts, l) \<in> set tss \<Longrightarrow>
      LTS_is_reachable (\<Delta> \<A>) f (LTS_fin_path_word ts) l \<and> f \<in> \<I> \<A> \<and> l \<in> \<F> \<A>"
      using g3 unfolding LTS_is_reachable_fin_path by auto
    have h: "\<And> f ts l. (f, ts, l) \<in> set tss \<Longrightarrow> NFA_accept \<A> (LTS_fin_path_word ts)"
      using h unfolding NFA_accept_def by blast
    have h: "\<And> f ts l. (f, ts, l) \<in> set tss \<Longrightarrow> LTS_fin_path_word ts \<in> \<L> \<A>"
      using h unfolding \<L>_def by auto
    have h: "\<And> w. w \<in> set (map (LTS_fin_path_word \<circ> fst \<circ> snd) tss) \<Longrightarrow> w \<in> \<L> \<A>"
      using h by auto
    have k: "concat (map (LTS_fin_path_word \<circ> fst \<circ> snd) tss) \<in>
      \<L> \<A> ^^ length (map (LTS_fin_path_word \<circ> fst \<circ> snd) tss)"
      using h set_pow_lang[of "map (LTS_fin_path_word \<circ> fst \<circ> snd) tss" "\<L> \<A>"]
      by auto
    have l: "wc \<in> \<L> \<A> ^^ Suc (length (map (LTS_fin_path_word \<circ> fst \<circ> snd) tss))"
      using i k g2 e(2) by auto
    show "wc \<in> transcl (\<L> \<A>)" using l unfolding transcl_def by fastforce
  qed

  lemma TM_times_NFA_deg_language: "\<L> (TM_times_NFA \<A> \<A>) = transcl (\<L> \<A>)"
  using TM_times_NFA_deg_language_forward
  using TM_times_NFA_deg_language_reverse
  by auto

end
