theory Enumerate_LGraph
imports "../../DFS_Framework/Examples/Reachable_Nodes"
begin

(* Define a state, based on the DFS-state. 
  In our case, we just add a break-flag.
*)
record 'v elg_state = "'v state" +
  num :: nat
  nmap :: "'v \<rightharpoonup> nat"
  edges :: "(nat \<times> nat) set"

(* Some utility lemmas for the simplifier, to handle idiosyncrasies of
  the record package. *)
lemma elg_more_cong: 
  "state.more s = state.more s' \<Longrightarrow> nmap s = nmap s'"
  "state.more s = state.more s' \<Longrightarrow> edges s = edges s'"
  "state.more s = state.more s' \<Longrightarrow> num s = num s'"
  by (cases s, cases s', simp)+

lemma [simp]: 
  "s\<lparr> state.more := \<lparr> num = x1, nmap = x2, edges = x3 \<rparr> \<rparr> = s \<lparr> num := x1, nmap := x2, edges := x3  \<rparr>"
  by (cases s) simp

definition "elg_on_new_edge u v s \<equiv> do {
  ASSERT (u \<in> dom (nmap s));
  let
    v'=num s;
    u'=the (nmap s u);
    nmap' = nmap s(v \<mapsto> v');
    num'=num s + 1;
    edges' = insert (u',v') (edges s)
  in 
    RETURN \<lparr> num = num', nmap = nmap', edges = edges' \<rparr>
}"

definition "elg_on_known_edge u v s \<equiv> do {
  ASSERT (u\<in>dom (nmap s));
  ASSERT (v\<in>dom (nmap s));
  let u'=the (nmap s u);
  let v'=the (nmap s v);
  let edges' = insert (u',v') (edges s);
  RETURN \<lparr>num = num s, nmap = nmap s, edges = edges'\<rparr>
}"



definition elg_params :: "('v,('v,unit) elg_state_ext) parameterization"
where "elg_params \<equiv> dflt_parametrization state.more 
  (RETURN \<lparr> num = 0, nmap=Map.empty, edges={} \<rparr>) \<lparr>
    on_cross_edge := elg_on_known_edge,
    on_back_edge := elg_on_known_edge,
    on_discover := elg_on_new_edge
  \<rparr>"
lemmas elg_params_simp[simp] = 
  gen_parameterization.simps[mk_record_simp, OF elg_params_def[simplified]]

interpretation elg: param_DFS_defs where param=elg_params for E V0 .

*** Argh, do not have DFS for labeled graphs!
Does Rene have something?


(* Find feedback arc set *)
definition "find_fas E V0 \<equiv> do {
  ASSERT (finite (E\<^sup>*``V0));
  s \<leftarrow> fas.it_dfsT E V0;
  RETURN (fas_state.fas s)
}"

locale fas = param_DFS E V0 fas_params for E and V0 :: "'v set" +
  assumes finite_reachable[simp, intro!]: "finite (E\<^sup>*``V0)"
begin
  sublocale DFS E V0 fas_params
    apply unfold_locales
    apply (simp_all add: fas_params_def)
    done

end

lemma fasI: 
  assumes "finite (E\<^sup>*``V0)"
  shows "fas E V0"
proof -
  interpret fb_graph E V0 by (rule finite_reachable_is_fb_graph[OF assms])
  show ?thesis by unfold_locales fact
qed

locale fas_invar = DFS_invar where param = fas_params + fas
begin

  lemma (in fas) i_fas_eq_back: "is_invar (\<lambda>s. fas_state.fas s = back_edges s)"
    apply (induct rule: establish_invarI)
    apply (simp_all add: cond_def cong: fas_more_cong)
    apply (simp add: empty_state_def)
    done
  lemmas fas_eq_back = i_fas_eq_back[THEN make_invar_thm]

  lemma find_fas_correct_aux:
    assumes NC: "\<not>cond s"
    shows "is_fas E V0 (fas_state.fas s)"
  proof -
    note [simp] = fas_eq_back  

    from nc_edges_covered[OF NC] edges_disjoint have 
      "E \<inter> reachable \<times> UNIV - back_edges s = tree s \<union> cross_edges s"
      by auto
    with tree_cross_acyclic show "is_fas E V0 (fas_state.fas s)"
      unfolding is_fas_alt reachable_def by simp
  qed    

end

lemma find_fas_correct: 
  assumes "finite (E\<^sup>*``V0)"
  shows "find_fas E V0 \<le> SPEC (is_fas E V0)"
  unfolding find_fas_def
proof (refine_vcg le_ASSERTI order_trans[OF DFS.it_dfsT_correct], clarsimp_all)
  assume "finite (E\<^sup>*``V0)"
  then interpret fb_graph E V0 by (rule finite_reachable_is_fb_graph)
  interpret fas by unfold_locales fact
  show "DFS E V0 fas_params" by unfold_locales
next
  fix s
  assume "DFS_invar E V0 fas_params s"
  then interpret DFS_invar E V0 fas_params s .
  interpret fas_invar E V0 s by unfold_locales fact
  assume "\<not>fas.cond V0 s"
  thus "is_fas E V0 (fas_state.fas s)"
    by (rule find_fas_correct_aux)
qed (rule assms | unfold fb_graph_defs.reachable_def)+

subsection {* Implementation *}

(* Implementation with stack and sso_visited set *)
record 'v fas_state_impl = "'v simple_state" +
  fas :: "('v\<times>'v) set"

(* Definition of refinement relation: The break-flag is refined by identity.*)
definition "fas_erel \<equiv> { 
  (\<lparr> fas_state_impl.fas = f \<rparr>, \<lparr> fas_state.fas = f\<rparr>) | f. True }"
abbreviation "fas_rel \<equiv> \<langle>fas_erel\<rangle>simple_state_rel"

(* Implementation of the parameters *)
definition fas_params_impl 
  :: "('v,'v fas_state_impl,('v,unit) fas_state_impl_ext) gen_parameterization"
where "fas_params_impl 
  \<equiv> dflt_parametrization simple_state.more (RETURN \<lparr> fas = {} \<rparr>) \<lparr>
  on_back_edge := \<lambda>u v s. RETURN \<lparr> fas = insert (u,v) (fas s) \<rparr>\<rparr>"
lemmas fas_params_impl_simp[simp,DFS_code_unfold] = 
  gen_parameterization.simps[mk_record_simp, OF fas_params_impl_def[simplified]]

lemma fas_impl: "(si,s)\<in>fas_rel 
  \<Longrightarrow> fas_state_impl.fas si = fas_state.fas s"
  by (cases si, cases s, simp add: simple_state_rel_def fas_erel_def)

interpretation fas_impl: simple_impl_defs E V0 fas_params_impl fas_params 
  for E V0 .

(* The above locale creates an iterative and a recursive implementation *)
term fas_impl.tailrec_impl term fas_impl.tailrec_implT term fas_impl.rec_impl

definition [DFS_code_unfold]: "find_fas_impl E V0 \<equiv> do {
  ASSERT (finite (E\<^sup>*``V0));
  s \<leftarrow> fas_impl.tailrec_implT V0 E;
  RETURN (fas s)
}"


context fas begin
  (* Derive the implementation *)
  sublocale simple_impl E V0 fas_params fas_params_impl fas_erel 
    apply unfold_locales
    apply (intro fun_relI, clarsimp simp: simple_state_rel_def, parametricity) []
    apply (auto simp: fas_erel_def fas_impl simple_state_rel_def)
    done

  lemmas impl_refine = simple_tailrec_refine simple_tailrecT_refine simple_rec_refine
  thm simple_refine
end

lemma find_fas_impl_refine: "find_fas_impl E V0 \<le> \<Down>Id (find_fas E V0)"
  unfolding find_fas_impl_def find_fas_def
  apply (refine_vcg fas.impl_refine)
  apply (simp_all add: fas_impl fasI)
  done
  
    

(* Autoref *)
record ('si,'nsi,'fsi)fas_state_impl' = "('si,'nsi)simple_state_impl" +
  fas_impl :: 'fsi

definition [to_relAPP]: "fas_state_erel frel erel \<equiv> {
  (\<lparr>fas_impl = fi, \<dots> =  mi\<rparr>,\<lparr>fas = f, \<dots> = m\<rparr>) | fi mi f m.
    (fi,f)\<in>frel \<and> (mi,m)\<in>erel}"

consts 
  i_fas_state_ext :: "interface \<Rightarrow> interface \<Rightarrow> interface"

lemmas [autoref_rel_intf] = REL_INTFI[of fas_state_erel i_fas_state_ext]


term fas_update
term fas_state_impl'.fas_impl_update
lemma [autoref_rules]:
  fixes ns_rel vis_rel frel erel
  defines "R \<equiv> \<langle>ns_rel,vis_rel,\<langle>frel,erel\<rangle>fas_state_erel\<rangle>ss_impl_rel"
  shows 
    "(fas_state_impl'_ext, fas_state_impl_ext) \<in> frel \<rightarrow> erel \<rightarrow> \<langle>frel,erel\<rangle>fas_state_erel"
    "(fas_impl, fas_state_impl.fas) \<in> R \<rightarrow> frel"
    "(fas_state_impl'.fas_impl_update, fas_update) \<in> (frel \<rightarrow> frel) \<rightarrow> R \<rightarrow> R"
  unfolding fas_state_erel_def ss_impl_rel_def R_def
  by (auto, parametricity)

schematic_lemma find_fas_impl:
  fixes V :: "('vi\<times>'v) set"
  assumes [autoref_ga_rules]: "is_bounded_hashcode V eq bhc"
  assumes [autoref_rules]: "(eq,op =) \<in> V \<rightarrow> V \<rightarrow> bool_rel"
  assumes [autoref_ga_rules]: "is_valid_def_hm_size TYPE ('vi) sz"
  assumes [autoref_rules]: 
    "(succi,E)\<in>\<langle>V\<rangle>slg_rel"
    "(V0',V0)\<in>\<langle>V\<rangle>list_set_rel"
  notes [autoref_tyrel] = 
    TYRELI[where R="\<langle>V\<rangle>ahs_rel bhc"]
    TYRELI[where R="\<langle>V \<times>\<^sub>r V\<rangle>ahs_rel (prod_bhc bhc bhc)"]
    TYRELI[where R="\<langle>V \<times>\<^sub>r \<langle>V\<rangle>list_set_rel\<rangle>ras_rel"]
  shows "RETURN (?c::?'c) \<le>\<Down>?R (find_fas_impl E V0)"
  unfolding DFS_code_unfold
  using [[autoref_trace_failed_id, goals_limit=1]]
  apply (autoref_monadic (trace))
  done
concrete_definition find_fas_code for eq bhc sz succi V0' uses find_fas_impl
thm find_fas_code.refine
export_code find_fas_code checking SML

thm find_fas_code.refine

lemma find_fas_code_refine[refine]:
  fixes V :: "('vi\<times>'v) set"
  assumes "is_bounded_hashcode V eq bhc"
  assumes "(eq,op =) \<in> V \<rightarrow> V \<rightarrow> bool_rel"
  assumes "is_valid_def_hm_size TYPE ('vi) sz"
  assumes 2: "(Ei, E) \<in> \<langle>V\<rangle>slg_rel"
  assumes 3: "(V0i, V0) \<in> \<langle>V\<rangle>list_set_rel"
  shows "RETURN (find_fas_code eq bhc sz Ei V0i) \<le> \<Down>(\<langle>V\<times>\<^sub>rV\<rangle>ahs_rel (prod_bhc bhc bhc)) (find_fas E V0)"
proof -
  note find_fas_code.refine[OF assms]
  also note find_fas_impl_refine
  finally show ?thesis .
qed

context begin interpretation autoref_syn .

lemma find_fas_code_autoref[autoref_rules]:
  fixes V :: "('vi\<times>'v) set" and bhc
  defines "RR \<equiv> \<langle>\<langle>V\<times>\<^sub>rV\<rangle>ahs_rel (prod_bhc bhc bhc)\<rangle>nres_rel"
  assumes BHC: "SIDE_GEN_ALGO (is_bounded_hashcode V eq bhc)"
  assumes EQ: "GEN_OP eq op = (V \<rightarrow> V \<rightarrow> bool_rel)"
  assumes VDS: "SIDE_GEN_ALGO (is_valid_def_hm_size TYPE ('vi) sz)"
  assumes 2: "(Ei, E) \<in> \<langle>V\<rangle>slg_rel"
  assumes 3: "(V0i, V0) \<in> \<langle>V\<rangle>list_set_rel"
  shows "(RETURN (find_fas_code eq bhc sz Ei V0i),
    (OP find_fas 
      ::: \<langle>V\<rangle>slg_rel \<rightarrow> \<langle>V\<rangle>list_set_rel \<rightarrow> RR)$E$V0)\<in>RR"
  unfolding RR_def
  apply (rule nres_relI)
  using assms 
  by (simp add: find_fas_code_refine)  

end





end

