theory Pid_Scheduler
imports "Gen_Scheduler_Refine" "../../Partial_Order_Reduction/System"
begin

  lemma system_complete_language_eq_lsystem_accept:
    fixes en ex init int
    defines "S \<equiv> \<lparr> g_V = UNIV, g_E = rel_of_enex (en, ex), g_V0 = init, sa_L = int \<rparr>"
    assumes "sa S"
    shows "system_complete.language en ex init int = {w. sa.accept S w}"
  proof -
    interpret jsys?: system_complete en ex init int by this
    interpret psys?: sa S by fact
    show ?thesis
    unfolding jsys.language_def
    unfolding psys.accept_def
    proof safe
      fix p w
      assume 1: "p \<in> init" "w \<in> words_inf p"
      show "\<exists> r. is_run r \<and> int \<circ> scan ex w p = sa_L S \<circ> r"
      unfolding is_run_def ipath_def
      unfolding S_def graph_rec.simps sa_rec.simps
      unfolding rel_of_enex_def Let_def prod.case
      proof (intro allI exI conjI)
        fix i
        show "int \<circ> scan ex w p = int \<circ> scan ex w p" by rule
        show "scan ex w p 0 \<in> init" using 1(1) by simp
        have "prefix (Suc i) w \<in> words_fin p" using 1(2) prefix_words_inf by blast
        then show "(scan ex w p i, scan ex w p (Suc i)) \<in> {(s, ex a s) |s a. a \<in> en s}" by auto
      qed
    next
      fix r
      assume "is_run r"
      then obtain w where "r 0 \<in> init" "\<And> i. w i \<in> en (r i)" "\<And> i. r (Suc i) = ex (w i) (r i)"
        unfolding is_run_def ipath_def
        unfolding S_def graph_rec.simps
        unfolding rel_of_enex_def Let_def prod.case
        by simp metis
      then have "r 0 \<in> init" "w \<in> words_inf (r 0)" "r = scan ex w (r 0)"
        using words_inf_run by blast+
      then show "\<exists> p w. sa_L S \<circ> r = int \<circ> scan ex w p \<and> p \<in> init \<and> w \<in> words_inf p"
        unfolding S_def sa_rec.simps
        by metis
    qed
  qed

  (* TODO: Move *)  
  lemma prod_splitE:
    assumes "\<forall>a b. p=(a,b) \<longrightarrow> P a b"
    obtains a b where "p=(a,b)" "P a b"
    using assms
    by (cases p) auto

  lemma sngp_sym_conv: "(\<lambda>p. p=x) = op = x" by auto  


  lemma in_multiset_of_conv_nth: "x \<in># multiset_of l \<longleftrightarrow> (\<exists>i<length l. x = l!i)"
    by (metis in_set_conv_nth mem_set_multiset_eq)
    
  lemma multiset_of_em_conv_nth: "multiset_of l = {#x#} + m' 
    \<longleftrightarrow> (\<exists>i<length l. x = l!i \<and> m' = multiset_of l - {#x#})"
    by (metis add_diff_cancel_left' in_multiset_of_conv_nth 
      insert_DiffM multi_member_this)

  
  type_synonym pid = nat
  record ('c,'ls,'gs) pid_global_config =
    processes :: "('c,'ls) local_config list"  
    state :: 'gs

  definition pidgc_\<alpha> 
    :: "('c,'ls,'gs) pid_global_config \<Rightarrow> ('c,'ls,'gs) global_config" where
    "pidgc_\<alpha> pgc \<equiv> \<lparr>     
      global_config.processes = multiset_of (pid_global_config.processes pgc),
      global_config.state = pid_global_config.state pgc
    \<rparr>"

  abbreviation pid_valid :: "('c,'ls,'gs) pid_global_config \<Rightarrow> pid \<Rightarrow> bool" 
    where
    "pid_valid gc pid \<equiv> pid < length (pid_global_config.processes gc)"
    
  abbreviation "lc_of_pid gc pid 
    \<equiv> pid_global_config.processes gc ! pid"
  
  abbreviation "cmd_of_pid gc pid 
    \<equiv> local_config.command (lc_of_pid gc pid)"
    
  abbreviation "ls_of_pid gc pid 
    \<equiv> local_config.state (lc_of_pid gc pid)"


  context Gen_Scheduler'
  begin

    definition pid_gstep_succ 
      :: "('c,'ls,'gs) pid_global_config \<Rightarrow> ('c,'ls,'gs) pid_global_config set" 
      where
      "pid_gstep_succ gc \<equiv> do {
        (* Select process *)
        let lcs = pid_global_config.processes gc;
        pid \<leftarrow> {0..<length lcs};
        let lc = lcs!pid;
        (* Focus state *)
        let gs = pid_global_config.state gc;
        let ls = local_config.state lc;
  
        (* Select action *)
        let cmd = local_config.command lc;
        (a,cmd') \<leftarrow> {(a,cmd'). cstep cmd a cmd'};
  
        if en (ls,gs) a then {do { 
          (* Execute *)
          let (ls,gs) = ex (ls,gs) a;
          (* Unfocus state. *)
          let lc = \<lparr> local_config.command = cmd', local_config.state = ls\<rparr>;
          \<lparr> 
            pid_global_config.processes = lcs[pid:=lc],
            pid_global_config.state = gs  
          \<rparr>
        }}
      else {}
    }"

    lemma pid_gstep_succE:
      assumes "gc' \<in> pid_gstep_succ gc"
      obtains pid a c' ls' gs' where
        "pid_valid gc pid"
        "cstep (cmd_of_pid gc pid) a c'"
        "en (ls_of_pid gc pid, pid_global_config.state gc) a"
        "ex (ls_of_pid gc pid, pid_global_config.state gc) a = (ls',gs')"
        "gc' = \<lparr> 
          pid_global_config.processes = pid_global_config.processes gc 
            [pid := 
              \<lparr> local_config.command = c', local_config.state = ls'\<rparr>
            ],
          pid_global_config.state = gs'\<rparr>"
      using assms
      unfolding pid_gstep_succ_def
      by (auto 
        split: split_if_asm prod.splits 
        elim!: prod_splitE
        simp: all_conj_distrib)

  lemma pid_gstep_succI:
    assumes
      "pid_valid gc pid"
      "cstep (cmd_of_pid gc pid) a c'"
      "en (ls_of_pid gc pid, pid_global_config.state gc) a"
      "ex (ls_of_pid gc pid, pid_global_config.state gc) a = (ls',gs')"
    shows "
      \<lparr> 
        pid_global_config.processes = pid_global_config.processes gc 
          [pid := 
            \<lparr> local_config.command = c', local_config.state = ls'\<rparr>
          ],
        pid_global_config.state = gs'
      \<rparr> \<in> pid_gstep_succ gc"
    using assms
    unfolding pid_gstep_succ_def
    by fastforce

  lemma pid_gc_state_eq: 
    "pid_global_config.state pgc = global_config.state (pidgc_\<alpha> pgc)"
    unfolding pidgc_\<alpha>_def by auto  
    
  lemma pid_gc_proc_eq:
    assumes "pid_valid gc pid"  
    shows "
      global_config.processes (pidgc_\<alpha> gc) 
      = {#lc_of_pid gc pid#} 
      + (global_config.processes (pidgc_\<alpha> gc) - {#lc_of_pid gc pid#})"
    using assms
    by (auto simp: pidgc_\<alpha>_def multiset_of_em_conv_nth)


  lemma pid_gstep_correct: 
    "gstep_succ (pidgc_\<alpha> pgc) = pidgc_\<alpha>`pid_gstep_succ pgc"
    apply safe
    apply (erule gstep_succE)
    apply (clarsimp simp: pidgc_\<alpha>_def multiset_of_em_conv_nth)
    apply (frule (3) pid_gstep_succI)
    apply (erule image_eqI[rotated])
    apply (auto simp: pidgc_\<alpha>_def multiset_of_update add_ac) []

    apply (erule pid_gstep_succE)
    apply (clarsimp simp: pidgc_\<alpha>_def multiset_of_em_conv_nth multiset_of_update)
    apply (simp add: pid_gc_state_eq)
    apply (frule (2) gstep_succI[rotated])
    apply (rule pid_gc_proc_eq, assumption)
    apply (auto simp: add_ac pidgc_\<alpha>_def) []
    done

  abbreviation "pid_gstep \<equiv> rel_of_succ pid_gstep_succ"
  definition "pid_step \<equiv> stutter_extend_edges UNIV pid_gstep"

end

locale Pid_Gen_Scheduler_linit = 
  Gen_Scheduler'_linit cstep en ex init label
  for cstep :: "'c \<Rightarrow> 'a \<Rightarrow> 'c \<Rightarrow> bool"
  and en :: "('ls\<times>'gs) \<Rightarrow> 'a \<Rightarrow> bool"
  and ex :: "('ls\<times>'gs) \<Rightarrow> 'a \<Rightarrow> ('ls\<times>'gs)"
  and init :: "('c,'ls,'gs) global_config set"
  and label :: "('c,'ls,'gs) global_config \<Rightarrow> 'l" +
  fixes pid_init :: "('c,'ls,'gs) pid_global_config"
  and pid_label :: "('c,'ls,'gs) pid_global_config \<Rightarrow> 'l"
  assumes pid_init_eq: "gc \<in> init \<longleftrightarrow> (gc=pidgc_\<alpha> pid_init)"
  assumes pid_label_eq: "label (pidgc_\<alpha> pgc) = pid_label pgc"
begin

  definition pid_system_automaton' :: "(('c,'ls,'gs) pid_global_config, 'l) sa_rec"
    where "pid_system_automaton' \<equiv>
      \<lparr> g_V = UNIV, g_E = pid_gstep, g_V0 = {pid_init}, sa_L = pid_label \<rparr>"

  lemma pid_system_automaton'_simps[simp]:
    "g_V pid_system_automaton' = UNIV"
    "g_E pid_system_automaton' = pid_gstep"
    "g_V0 pid_system_automaton' = {pid_init}"
    "sa_L pid_system_automaton' = pid_label"
    unfolding pid_system_automaton'_def by auto

  definition pid_system_automaton :: "(('c,'ls,'gs) pid_global_config, 'l) sa_rec"
    where "pid_system_automaton \<equiv>
      \<lparr> g_V = UNIV, g_E = pid_step, g_V0 = {pid_init}, sa_L = pid_label \<rparr>"

  lemma pid_system_automaton_simps[simp]:
    "g_V pid_system_automaton = UNIV"
    "g_E pid_system_automaton = pid_step"
    "g_V0 pid_system_automaton = {pid_init}"
    "sa_L pid_system_automaton = pid_label"
    unfolding pid_system_automaton_def by auto

  lemma pid_system_automaton_alt_def: "pid_system_automaton = stutter_extend pid_system_automaton'"
    unfolding pid_step_def pid_system_automaton'_def pid_system_automaton_def stutter_extend_def
    by simp

  sublocale pid': sa pid_system_automaton' by unfold_locales auto
  sublocale pid: sa pid_system_automaton
    unfolding pid_system_automaton_alt_def using stutter_extend_sa by this

  sublocale bisim!: lbisimulation "br pidgc_\<alpha> (\<lambda>_. True)" pid_system_automaton system_automaton
  proof -
    interpret bisim!: lbisimulation "br pidgc_\<alpha> (\<lambda>_. True)" pid_system_automaton' system_automaton'
      apply unfold_locales
      apply (auto simp: build_rel_def pid_init_eq pid_gstep_correct pid_label_eq)
      done
    interpret bisim!: lbisimulation "br pidgc_\<alpha> (\<lambda>_. True)" pid_system_automaton system_automaton
      unfolding system_automaton_alt_def pid_system_automaton_alt_def
      using bisim.lstutter_extend by this
    show "lbisimulation (br pidgc_\<alpha> (\<lambda> _. True)) pid_system_automaton system_automaton"
      by unfold_locales
  qed

  lemma pid_accept_eq: "pid.accept = sa.accept" by (rule bisim.accept_bisim)

  lemma pid_run_conv: "sa.g.is_run r \<longleftrightarrow> (\<exists>r'. r=pidgc_\<alpha> o r' \<and> pid.g.is_run r')"
    by (rule bisim.br_run_conv[OF refl])   

  lemma pid_run_sim: "pid.g.is_run r \<Longrightarrow> sa.g.is_run (pidgc_\<alpha> o r)"
    by (rule bisim.s1.br_run_sim[OF refl])

section "Global Actions"

text {*
  A global action consists of a PID and a control flow graph edge
*}
type_synonym (in -) ('c,'a) global_action = "pid \<times> 'c \<times> 'a \<times> 'c"


abbreviation (in -) "fs_of_pid gc pid \<equiv> (ls_of_pid gc pid,pid_global_config.state gc)"

  definition ga_gen 
    :: "('c,'ls,'gs) pid_global_config \<Rightarrow> ('c,'a) global_action set"
  where
    "ga_gen gc \<equiv> { (pid,c,a,c') .
      cstep c a c' 
    \<and> pid_valid gc pid 
    \<and> c=cmd_of_pid gc pid
    \<and> en (fs_of_pid gc pid) a }"

  definition ga_gex
    :: "('c,'a) global_action \<Rightarrow> ('c,'ls,'gs) pid_global_config \<Rightarrow> ('c,'ls,'gs) pid_global_config"
  where
    "ga_gex ga gc \<equiv> let 
      (pid,c,a,c') = ga;
      fs = fs_of_pid gc pid;
      (ls',gs') = ex fs a;
      gc' = \<lparr> 
          pid_global_config.processes = pid_global_config.processes gc 
            [pid := 
              \<lparr> local_config.command = c', local_config.state = ls'\<rparr>
            ],
          pid_global_config.state = gs'\<rparr>
    in
      gc'"

  definition "ga_gstep \<equiv> rel_of_enex (ga_gen,ga_gex)"
  definition "ga_en \<equiv> stutter_extend_en ga_gen"
  definition "ga_ex \<equiv> stutter_extend_ex ga_gex"
  definition "ga_step \<equiv> stutter_extend_edges UNIV ga_gstep"
  lemma ga_step_alt: "ga_step = rel_of_enex (ga_en,ga_ex)"
    unfolding ga_step_def ga_gstep_def ga_en_def ga_ex_def
    by (auto simp: stutter_extend_pred_of_enex_conv)

  lemma some_ga_en_eq: 
    "Some a \<in> ga_en gc \<longleftrightarrow> a \<in> ga_gen gc"
    by (auto simp: ga_en_def)
  lemma some_ga_ex_eq: 
    "ga_ex (Some a) gc = ga_gex a gc"
    "ga_ex \<circ> Some = ga_gex"
    by (auto simp: ga_ex_def intro!: ext)
  lemma none_ga_en_eq: "None \<in> ga_en gc \<longleftrightarrow> ga_gen gc = {}"
    by (auto simp: ga_en_def stutter_extend_en_def)
  lemma ga_ex_none[simp]: "ga_ex None = id"   
    by (auto simp: ga_ex_def)

  lemma ga_gstep_eq: "ga_gstep = pid_gstep"
    apply safe
    apply (auto 
      simp: ga_gstep_def ga_gen_def ga_gex_def 
      split: prod.splits elim!: prod_splitE
      intro: pid_gstep_succI) []
     
    apply (fastforce 
      simp: pred_of_succ_def ga_gstep_def ga_gen_def ga_gex_def
      split: prod.splits
      elim!: pid_gstep_succE)
    done

  lemma ga_step_eq: "ga_step = pid_step" 
    unfolding ga_step_def pid_step_def ga_gstep_eq by rule

  definition ga_automaton :: "(('c,'ls,'gs) pid_global_config, 'l) sa_rec"
    where "ga_automaton \<equiv> \<lparr> g_V = UNIV, g_E = ga_step, g_V0 = {pid_init}, sa_L = pid_label \<rparr>"

  lemma ga_automaton_simps[simp]:
    "g_V ga_automaton = UNIV"
    "g_E ga_automaton = ga_step"
    "g_V0 ga_automaton = {pid_init}"
    "sa_L ga_automaton = pid_label"
    unfolding ga_automaton_def by simp+

  lemma ga_pid_system_automaton: "ga_automaton = pid_system_automaton"
    unfolding ga_automaton_def pid_system_automaton_def ga_step_eq by rule

  sublocale jsys!: system ga_en ga_ex .
  sublocale ga!: sa ga_automaton unfolding ga_pid_system_automaton by unfold_locales

  sublocale ga_bisim!: lbisimulation Id ga_automaton pid_system_automaton
    unfolding ga_pid_system_automaton by auto

  lemma ga_accept_eq: "ga.accept = sa.accept" 
    by (simp add: ga_bisim.accept_bisim pid_accept_eq) 

  lemma ga_run_eq: "pid.g.is_run r = ga.is_run r"
    using ga_bisim.br_run_conv[of id "\<lambda>_. True" r]
    by (auto simp: build_rel_def[abs_def])

  sublocale jsys!: system_initial ga_en ga_ex "{pid_init}" . 

  lemma reachable_alt: "jsys.reachable = ga.E\<^sup>* `` ga.V0"
  proof safe
    fix gc
    assume "gc \<in> jsys.reachable"
    thus "gc \<in> ga.E\<^sup>* `` ga.V0"
      unfolding ga_automaton_def ga_step_alt rel_of_enex_def
      by (induct, auto intro: rtrancl_into_rtrancl)
  next
    fix gc gc'
    assume "(gc, gc') \<in> ga.E\<^sup>*" "gc \<in> ga.V0"
    thus "gc' \<in> jsys.reachable" unfolding ga_automaton_def ga_step_alt by induct auto
  qed

  lemma rtrancl_step_exec_word_conv:  
    "(gc, gc') \<in> ga_step\<^sup>* \<longleftrightarrow> (\<exists>w\<in>jsys.words_fin gc. gc' = fold ga_ex w gc)"
  proof safe
    assume "(gc, gc') \<in> ga_step\<^sup>*"
    thus "\<exists>w\<in>jsys.words_fin gc. gc' = fold ga_ex w gc"
      apply induction
      apply force
      apply (force simp: ga_step_alt pred_of_enex_def)
      done
  next
    fix w
    assume "w\<in>jsys.words_fin gc"
    thus "(gc, fold ga_ex w gc) \<in> ga_step\<^sup>*"
      apply induction
      apply (auto simp: ga_step_alt rel_of_enex_def
        intro: converse_rtrancl_into_rtrancl)
      done
  qed

  sublocale jsys!: system_complete ga_en ga_ex "{pid_init}" pid_label .

  lemma accept_eq_lang: "jsys.language = Collect sa.accept"
    unfolding ga_accept_eq[symmetric]
    unfolding ga_automaton_def
    unfolding ga_step_alt
    apply (rule system_complete_language_eq_lsystem_accept)
    unfolding ga_step_alt[symmetric]
    unfolding ga_automaton_def[symmetric]
    by unfold_locales

end

end

