theory Simple
imports Main Word_Misc Multiset Monad_Syntax 
  CAVA_Base "../Modelchecker/CAVA_Automata/Digraph"
  "../DFS_Framework/Misc/DFS_Framework_Misc"
  "../DFS_Framework/Examples/Reachable_Nodes"
  "../Partial_Order_Reduction/Ample_Analysis"
begin

lemma finite_ImageI:
  assumes "finite A"  
  assumes "\<And>a. a\<in>A \<Longrightarrow> finite (R``{a})"
  shows "finite (R``A)"
proof -
  note [[simproc add: finite_Collect]]
  have "R``A = \<Union>{R``{a} | a. a:A}"
    by auto
  also have "finite (\<Union>{R``{a} | a. a:A})"
    apply (rule finite_Union)
    apply (simp add: assms)
    apply (clarsimp simp: assms)
    done
  finally show ?thesis .
qed




  definition "build_relp \<alpha> I x y \<equiv> y=\<alpha> x \<and> I x"
  abbreviation "brp \<equiv> build_relp"

  lemma brp_right_unique[simp, intro!]: "right_unique (brp \<alpha> I)"
    apply (rule right_uniqueI)
    unfolding build_relp_def
    by auto

  definition "the_brp_\<alpha> R \<equiv> (\<lambda>x. SOME y. R x y)"
  abbreviation (input) "the_brp_invar \<equiv> Domainp"

  lemma the_brp[simp]: 
    assumes "right_unique R"  
    shows "brp (the_brp_\<alpha> R) (the_brp_invar R) = R"
      apply (rule sym)
      apply (intro ext)
      unfolding build_relp_def the_brp_\<alpha>_def
      apply auto
      apply (metis assms right_uniqueD someI)
      by (blast intro: someI)

  lemma obtain_brp:
    assumes "right_unique R"  
    obtains \<alpha> I where "R=brp \<alpha> I"
    using the_brp[OF assms, THEN sym] ..

  lemma the_brp_brp[simp]: 
    "I x \<Longrightarrow> the_brp_\<alpha> (brp \<alpha> I) x = \<alpha> x"
    "the_brp_invar (brp \<alpha> I) = I"
    unfolding the_brp_\<alpha>_def build_relp_def[abs_def]
    by (auto)

  lemma brp_comp[simp]: "brp \<alpha>1 I1 OO brp \<alpha>2 I2 = brp (\<alpha>2 o \<alpha>1) (\<lambda>x. I1 x \<and> I2 (\<alpha>1 x))"
    unfolding build_relp_def[abs_def] by auto


  definition is_run_IS :: "('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a word \<Rightarrow> bool"
    where "is_run_IS V0 step r \<equiv> V0 (r 0) \<and> (\<forall>i. step (r i) (r (Suc i)))"

  definition run_rel :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> 'a word \<Rightarrow> 'b word \<Rightarrow> bool" where
    "run_rel R ra rb \<equiv> \<forall>i. R (ra i) (rb i)"

  lemma run_rel_converse[simp]: "run_rel (R\<inverse>\<inverse>) r1 r2 = run_rel R r2 r1"  
    unfolding run_rel_def by auto

  lemma run_rel_right_unique: "right_unique R 
    \<Longrightarrow> run_rel R ra rb \<longleftrightarrow> ((\<forall>i. the_brp_invar R (ra i)) \<and> rb = the_brp_\<alpha> R o ra)"
    unfolding run_rel_def the_brp_\<alpha>_def
    apply (auto intro!: ext)
    apply (metis (full_types) right_uniqueD someI)
    by (metis Domainp_iff someI)

  locale system =
    fixes V0 :: "'a \<Rightarrow> bool"
    fixes step :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  begin
    abbreviation "is_run \<equiv> is_run_IS V0 step"

    definition "is_reachable s \<equiv> \<exists>s0. V0 s0 \<and> step\<^sup>*\<^sup>* s0 s"

    lemma is_reachable_induct[induct pred:is_reachable, case_names init step]:
      assumes "is_reachable s"
      assumes "\<And>s. V0 s \<Longrightarrow> P s"
      assumes "\<And>s s'. step s s' \<Longrightarrow> P s \<Longrightarrow> P s'"
      shows "P s"
      using assms
      unfolding is_reachable_def
      by (auto intro: rtranclp_induct)

    definition "reachable \<equiv> Collect (is_reachable)"
    
    lemma reachable_induct[induct set:reachable, case_names init step]:
      assumes "s\<in>reachable"
      assumes "\<And>s. V0 s \<Longrightarrow> P s"
      assumes "\<And>s s'. step s s' \<Longrightarrow> P s \<Longrightarrow> P s'"
      shows "P s"
      using assms
      unfolding reachable_def 
      by (auto intro: is_reachable_induct)

    lemma is_reachable0: "V0 s \<Longrightarrow> is_reachable s"
      unfolding is_reachable_def by auto

    lemma is_reachable_step: "is_reachable s \<Longrightarrow> step s s' \<Longrightarrow> is_reachable s'"
      unfolding is_reachable_def by auto

    lemma reachable0: "V0 s \<Longrightarrow> s\<in>reachable"
      unfolding reachable_def by (auto intro: is_reachable0)

    lemma reachable_step: "s\<in>reachable \<Longrightarrow> step s s' \<Longrightarrow> s'\<in>reachable"
      unfolding reachable_def by (auto intro: is_reachable_step)

  end

  locale simulation = a!: system A0 astep + b!: system B0 bstep
    for R :: "'a \<Rightarrow> 'b \<Rightarrow> bool"
    and A0 :: "'a \<Rightarrow> bool" and astep :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
    and B0 :: "'b \<Rightarrow> bool" and bstep :: "'b \<Rightarrow> 'b \<Rightarrow> bool" +
    assumes init_sim: "\<And>a0. A0 a0 \<Longrightarrow> \<exists>b0. B0 b0 \<and> R a0 b0"
    assumes step_sim: "\<And>a a' b. \<lbrakk>astep a a'; R a b\<rbrakk> \<Longrightarrow> (\<exists>b'. bstep b b' \<and> R a' b')"
  begin
    lemma simulation_this: "simulation R A0 astep B0 bstep"
      by unfold_locales

    abbreviation "Rrel \<equiv> {(a,b). R a b}"  

    lemma run_sim: 
      assumes arun: "a.is_run ra"
      obtains rb where "b.is_run rb" "run_rel R ra rb"
    proof -
      from arun have ainit: "A0 (ra 0)" 
        and astep: "\<forall>i. astep (ra i) (ra (Suc i))"
        unfolding is_run_IS_def by auto
      from init_sim obtain rb0 where rel0: "R (ra 0) rb0" and binit: "B0 rb0"
        by (auto intro: ainit)
  
      def rb \<equiv> "rec_nat rb0 (\<lambda>i rbi. SOME rbsi. bstep rbi rbsi \<and> R (ra (Suc i)) rbsi)"
      
      have [simp]: 
        "rb 0 = rb0" 
        "\<And>i. rb (Suc i) = (SOME rbsi. bstep (rb i) rbsi \<and> R (ra (Suc i)) rbsi)"
        unfolding rb_def by auto
  
      {
        fix i
        have "bstep (rb i) (rb (Suc i)) \<and> R (ra (Suc i)) (rb (Suc i))"
        proof (induction i)
          case 0
          from step_sim astep rel0 obtain rb1 where "bstep (rb 0) rb1" and "R (ra 1) rb1"
            by fastforce
          thus ?case by (auto intro!: someI)
        next
          case (Suc i)
          with step_sim astep obtain rbss where "bstep (rb (Suc i)) rbss" and "R (ra (Suc (Suc i))) rbss"
            by fastforce
          thus ?case by (auto intro!: someI)
        qed
      } note aux=this
      
      from aux binit have "is_run_IS B0 bstep rb"
        unfolding is_run_IS_def by auto
      moreover from aux rel0 have "run_rel R ra rb"
        unfolding run_rel_def apply - apply rule apply (case_tac i) by auto
      ultimately show ?thesis by rule
    qed

    lemma stuck_sim: 
      assumes "R a b"
      assumes "\<not>Domainp bstep b"
      shows "\<not>Domainp astep a"
      using assms
      by (auto dest: step_sim)

    lemma run_Domain: "a.is_run r \<Longrightarrow> Domainp R (r i)"
      by (erule run_sim) (auto simp: run_rel_def)

    lemma brp_run_sim:
      assumes "R = brp \<alpha> I"
      assumes "a.is_run r"
      shows "b.is_run (\<alpha> o r)"
      using assms
      apply -
      apply (erule run_sim)
      apply (auto simp: run_rel_def build_relp_def is_run_IS_def)
      done

    lemma is_reachable_sim: "a.is_reachable a \<Longrightarrow> \<exists>b. R a b \<and> b.is_reachable b"
      apply (induction rule: a.is_reachable_induct)
      apply (metis b.is_reachable0 init_sim)
      apply (metis b.is_reachable_step step_sim)
      done
      
    lemma reachable_sim: "a.reachable \<subseteq> Rrel\<inverse>``b.reachable"
      using is_reachable_sim
      by (auto simp: a.reachable_def b.reachable_def)
        
    lemma reachable_finite_sim:
      assumes "finite b.reachable"
      assumes "\<And>b. b\<in>b.reachable \<Longrightarrow> finite (Rrel\<inverse>``{b})"
      shows "finite a.reachable"
      apply (rule finite_subset[OF reachable_sim])
      apply (rule finite_ImageI)
      apply fact+
      done

  end  

  lemma simulation_trans[trans]: 
    assumes "simulation R1 a0 as b0 bs" 
    assumes "simulation R2 b0 bs c0 cs"
    shows "simulation (R1 OO R2) a0 as c0 cs"
  proof -
    interpret s1: simulation R1 a0 as b0 bs by fact
    interpret s2: simulation R2 b0 bs c0 cs by fact
    show ?thesis
      apply unfold_locales
      using s1.init_sim s2.init_sim apply blast
      using s1.step_sim s2.step_sim apply blast
      done
  qed

  locale bisimulation = 
    a!: system A0 astep + b!: system B0 bstep +
    s1!: simulation R A0 astep B0 bstep + 
    s2!: simulation "conversep R" B0 bstep A0 astep
    for R :: "'a \<Rightarrow> 'b \<Rightarrow> bool"
    and A0 :: "'a \<Rightarrow> bool" and astep :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
    and B0 :: "'b \<Rightarrow> bool" and bstep :: "'b \<Rightarrow> 'b \<Rightarrow> bool"
  begin

    lemma bisimulation_this: "bisimulation R A0 astep B0 bstep"
      by unfold_locales

    lemma converse: "bisimulation (conversep R) B0 bstep A0 astep"
      apply intro_locales
      apply simp
      by unfold_locales

    lemma brp_run_conv:
      assumes "R = brp \<alpha> I"
      shows "b.is_run rb \<longleftrightarrow> (\<exists>ra. rb=\<alpha> o ra \<and> a.is_run ra)"
      using assms
      apply safe
      apply (erule s2.run_sim, auto 
        intro!: ext
        simp: run_rel_def build_relp_def) []
      apply (erule s1.brp_run_sim, assumption)
      done

    lemma inj_map_run_eq:
      assumes "inj \<alpha>"
      assumes E: "\<alpha> o r1 = \<alpha> o r2"
      shows "r1 = r2"
    proof (rule ext)
      fix i
      from E have "\<alpha> (r1 i) = \<alpha> (r2 i)"
        by (simp add: comp_def) metis
      with `inj \<alpha>` show "r1 i = r2 i" 
        by (auto dest: injD)
    qed

    lemma brp_inj_run_conv:
      assumes INJ: "inj \<alpha>"
      assumes [simp]: "R = brp \<alpha> I"
      shows "b.is_run (\<alpha> o ra) \<longleftrightarrow> a.is_run ra"
      apply (subst brp_run_conv[OF assms(2)])
      apply (auto dest: inj_map_run_eq[OF INJ])
      done

    lemma right_unique_run_conv:
      assumes "right_unique R"
      shows "b.is_run rb 
        \<longleftrightarrow> (\<exists>ra. rb=the_brp_\<alpha> R o ra \<and> a.is_run ra)"
      using assms
      apply safe
      apply (erule s2.run_sim)
      apply (auto simp add: run_rel_right_unique)
      apply (erule s1.run_sim)
      apply (auto simp add: run_rel_right_unique)
      done

    lemma stuck_bisim: 
      assumes A: "R a b"
      shows "Domainp astep a \<longleftrightarrow> Domainp bstep b"
      using s1.stuck_sim[OF A] 
      using s2.stuck_sim[OF A[THEN conversepI[of R]]]
      by blast

  end

  lemma bisimulation_trans[trans]:
    assumes "bisimulation R1 a0 as b0 bs" 
    assumes "bisimulation R2 b0 bs c0 cs"
    shows "bisimulation (R1 OO R2) a0 as c0 cs"
  proof -
    interpret s1: bisimulation R1 a0 as b0 bs by fact
    interpret s2: bisimulation R2 b0 bs c0 cs by fact
    show ?thesis
      apply intro_locales
      apply (rule simulation_trans[OF s1.s1.simulation_this s2.s1.simulation_this])
      apply (rule simulation_trans[OF s2.s2.simulation_this s1.s2.simulation_this, folded converse_relcompp])
      done
  qed

  lemma simulation_refl[simp]: "simulation op= a0 as a0 as"
    by (unfold_locales) auto

  lemma bisimulation_refl[simp]: "bisimulation op= a0 as a0 as"
    by (unfold_locales) auto

  locale lsystem = system V0 step
    for V0 :: "'a \<Rightarrow> bool" and step and L :: "'a \<Rightarrow> 'l"
  begin
    definition "accept w \<equiv> \<exists>r. is_run r \<and> w=L o r"
  end

  locale lsimulation = 
    a!: lsystem A0 astep La +
    b!: lsystem B0 bstep Lb +
    simulation R A0 astep B0 bstep for
      R :: "'a \<Rightarrow> 'b \<Rightarrow> bool" and 
      A0 :: "'a \<Rightarrow> bool" and astep and La :: "'a \<Rightarrow> 'l" and 
      B0 :: "'b \<Rightarrow> bool" and bstep and Lb :: "'b \<Rightarrow> 'l" +
    assumes labeling_consistent: "\<And>a b. R a b \<Longrightarrow> La a = Lb b"
  begin
    lemma lsimulation_this: "lsimulation R A0 astep La B0 bstep Lb"
      by unfold_locales

    lemma run_rel_consistent: "run_rel R ra rb \<Longrightarrow> La o ra = Lb o rb"
      using labeling_consistent unfolding run_rel_def
      by auto

    lemma accept_sim: "a.accept w \<Longrightarrow> b.accept w"
      unfolding a.accept_def b.accept_def
      apply clarsimp
      apply (erule run_sim)
      apply (auto simp: run_rel_consistent)
      done
  end

  locale lbisimulation = 
    a!: lsystem A0 astep La +
    b!: lsystem B0 bstep Lb +
    s1!: lsimulation R A0 astep La B0 bstep Lb +
    s2!: lsimulation "R\<inverse>\<inverse>" B0 bstep Lb A0 astep La +
    bisimulation R A0 astep B0 bstep
    for R A0 astep and La :: "'a \<Rightarrow>'l" and B0 bstep and Lb :: "'b \<Rightarrow> 'l"
  begin
    lemma lbisimulation_this: "lbisimulation R A0 astep La B0 bstep Lb"
      by unfold_locales

    lemma accept_bisim: "a.accept = b.accept"
      apply (rule ext)
      using s1.accept_sim s2.accept_sim by blast
  end


  lemma lsimulation_trans[trans]: 
    assumes "lsimulation R1 a0 as La b0 bs Lb"
    assumes "lsimulation R2 b0 bs Lb c0 cs Lc"
    shows "lsimulation (R1 OO R2) a0 as La c0 cs Lc"
  proof -
    interpret s1: lsimulation R1 a0 as La b0 bs Lb by fact
    interpret s2: lsimulation R2 b0 bs Lb c0 cs Lc by fact
    show ?thesis
      apply intro_locales
      apply (rule simulation_trans[OF s1.simulation_this s2.simulation_this])
      apply unfold_locales
      using s1.labeling_consistent s2.labeling_consistent 
      by auto
  qed

  lemma lbisimulation_trans[trans]:
    assumes "lbisimulation R1 a0 as La b0 bs Lb" 
    assumes "lbisimulation R2 b0 bs Lb c0 cs Lc"
    shows "lbisimulation (R1 OO R2) a0 as La c0 cs Lc"
  proof -
    interpret s1: lbisimulation R1 a0 as La b0 bs Lb by fact
    interpret s2: lbisimulation R2 b0 bs Lb c0 cs Lc by fact

    from lsimulation_trans[OF s1.s1.lsimulation_this s2.s1.lsimulation_this]
    interpret t1: lsimulation "(R1 OO R2)" a0 as La c0 cs Lc .

    from lsimulation_trans[OF s2.s2.lsimulation_this s1.s2.lsimulation_this, folded converse_relcompp]
    interpret t2: lsimulation "(R1 OO R2)\<inverse>\<inverse>" c0 cs Lc a0 as La .

    show ?thesis by unfold_locales
  qed


  lemma lsimulation_refl[simp]: "lsimulation op= a0 as La a0 as La"
    by (unfold_locales) auto

  lemma lbisimulation_refl[simp]: "lbisimulation op= a0 as La a0 as La"
    by (unfold_locales) auto


  (* Different ways of representing transitions, 
    and functions to convert between them:

    rel :: ('a \<times> 'b) set
    pred :: 'a \<Rightarrow> 'b \<Rightarrow> bool
    succ :: 'a \<Rightarrow> 'b set
  *)

  definition "rel_of_pred s \<equiv> {(a,b). s a b}"
  definition "rel_of_succ s \<equiv> {(a,b). b\<in>s a}"

  definition "pred_of_rel s \<equiv> \<lambda>a b. (a,b)\<in>s"
  definition "pred_of_succ s \<equiv> \<lambda>a b. b\<in>s a"

  definition "succ_of_rel s \<equiv> \<lambda>a. {b. (a,b)\<in>s}"
  definition "succ_of_pred s \<equiv> \<lambda>a. {b. s a b}"

  lemma rps_expand[simp]:
    "(a,b)\<in>rel_of_pred p \<longleftrightarrow> p a b"
    "(a,b)\<in>rel_of_succ s \<longleftrightarrow> b \<in> s a"

    "pred_of_rel r a b \<longleftrightarrow> (a,b)\<in>r"
    "pred_of_succ s a b \<longleftrightarrow> b \<in> s a"

    "b\<in>succ_of_rel r a \<longleftrightarrow> (a,b)\<in>r"
    "b\<in>succ_of_pred p a \<longleftrightarrow> p a b"
    unfolding rel_of_pred_def pred_of_rel_def
    unfolding rel_of_succ_def succ_of_rel_def
    unfolding pred_of_succ_def succ_of_pred_def
    by auto

  lemma rps_conv[simp]:
    "rel_of_pred (pred_of_rel r) = r"
    "rel_of_pred (pred_of_succ s) = rel_of_succ s"

    "rel_of_succ (succ_of_rel r) = r"
    "rel_of_succ (succ_of_pred p) = rel_of_pred p"

    "pred_of_rel (rel_of_pred p) = p"
    "pred_of_rel (rel_of_succ s) = pred_of_succ s"

    "pred_of_succ (succ_of_pred p) = p"
    "pred_of_succ (succ_of_rel r) = pred_of_rel r"

    "succ_of_rel (rel_of_succ s) = s"
    "succ_of_rel (rel_of_pred p) = succ_of_pred p"

    "succ_of_pred (pred_of_succ s) = s"
    "succ_of_pred (pred_of_rel r) = succ_of_rel r"
    unfolding rel_of_pred_def pred_of_rel_def
    unfolding rel_of_succ_def succ_of_rel_def
    unfolding pred_of_succ_def succ_of_pred_def
    by auto

  (* Lifting transitions from option monad to option\<times>option *)
  definition m2r_rel :: "('a \<times> 'a option) set \<Rightarrow> 'a option rel"
    where "m2r_rel r \<equiv> {(Some a,b)|a b. (a,b)\<in>r}"

  definition m2r_pred :: "('a \<Rightarrow> 'a option \<Rightarrow> bool) \<Rightarrow> 'a option \<Rightarrow> 'a option \<Rightarrow> bool"
    where "m2r_pred p \<equiv> \<lambda>None \<Rightarrow> \<lambda>_. False | Some a \<Rightarrow> p a"

  definition m2r_succ :: "('a \<Rightarrow> 'a option set) \<Rightarrow> 'a option \<Rightarrow> 'a option set"
    where "m2r_succ s \<equiv> \<lambda>None \<Rightarrow> {} | Some a \<Rightarrow> s a"

  lemma m2r_expand[simp]:
    "(a,b)\<in>m2r_rel r \<longleftrightarrow> (\<exists>a'. a=Some a' \<and> (a',b)\<in>r)"
    "m2r_pred p a b \<longleftrightarrow> (\<exists>a'. a=Some a' \<and> p a' b)"
    "b\<in>m2r_succ s a \<longleftrightarrow> (\<exists>a'. a=Some a' \<and> b \<in> s a')"
    unfolding m2r_rel_def m2r_succ_def m2r_pred_def
    by (auto split: option.splits)

  lemma m2r_conv[simp]:
    "m2r_rel (rel_of_succ s) = rel_of_succ (m2r_succ s)"
    "m2r_rel (rel_of_pred p) = rel_of_pred (m2r_pred p)"

    "m2r_pred (pred_of_succ s) = pred_of_succ (m2r_succ s)"
    "m2r_pred (pred_of_rel r) = pred_of_rel (m2r_rel r)"

    "m2r_succ (succ_of_pred p) = succ_of_pred (m2r_pred p)"
    "m2r_succ (succ_of_rel r) = succ_of_rel (m2r_rel r)"
    unfolding rel_of_pred_def pred_of_rel_def
    unfolding rel_of_succ_def succ_of_rel_def
    unfolding pred_of_succ_def succ_of_pred_def
    unfolding m2r_rel_def m2r_succ_def m2r_pred_def
    by (auto split: option.splits)

  
  (* Stuttering extension. A stuck state is repeated infinitely. *)
  inductive stutter_extend :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> bool" for step where
    step: "\<lbrakk> step c c' \<rbrakk> \<Longrightarrow> stutter_extend step c c'"
  | stutter: "\<lbrakk> \<not>Domainp step c \<rbrakk> \<Longrightarrow> stutter_extend step c c"
    
  lemma stutter_extend_no_stuck[simp, intro!]: "Domainp (stutter_extend step) a"
    by (cases "Domainp step a") (auto intro: step stutter)

  (* Stuttering extension preserves bisimulation. *)
  lemma (in bisimulation) stutter_extend:
    shows "bisimulation R A0 (stutter_extend astep) B0 (stutter_extend bstep)"
    apply unfold_locales
    apply (auto intro: s1.init_sim dest: s2.init_sim)

    apply (erule stutter_extend.cases)
    apply (auto intro: stutter_extend.intros 
      dest: s1.step_sim simp: stuck_bisim)
    apply (auto 
      elim!: stutter_extend.cases
      intro: stutter_extend.intros 
      dest: s2.step_sim dest: s2.stuck_sim s1.stuck_sim)
    done

  (* Does not work: Nonterminating sublocale relationship :(
  sublocale (in bisimulation) stutter_extend!: 
    bisimulation R A0 "stutter_extend astep" B0 "stutter_extend bstep"
    by (rule stutter_extend)
  *)

  lemma (in lbisimulation) lstutter_extend:
    shows "lbisimulation R A0 (stutter_extend astep) La B0 (stutter_extend bstep) Lb"
  proof -
    interpret se!: bisimulation R A0 "stutter_extend astep" B0 "stutter_extend bstep"
      by (rule stutter_extend)

    show ?thesis
      apply unfold_locales
      by (auto simp: s1.labeling_consistent)
  qed


  (* Set monad utilities *)

  lemma do_set_push_Image:
    "\<And>g f m. g`(do {x\<leftarrow>m; f x}) = do {x\<leftarrow>m; g`f x}"
    "\<And>g f m. g`(do {let x = m; f x}) = (do {let x=m; g`f x})"
    by fastforce+
    
  lemma case_option_distrib: "f (case x of Some v \<Rightarrow> fs v | None \<Rightarrow> fn) 
    = (case x of Some v \<Rightarrow> f (fs v) | None \<Rightarrow> f fn)" 
    by (auto split: option.split)

  lemma case_sum_distrib: "f (case x of Inl x \<Rightarrow> fl x | Inr x \<Rightarrow> fr x) = 
    (case x of Inl x \<Rightarrow> f (fl x) | Inr x \<Rightarrow> f (fr x))"
    by (auto split: sum.split)

  lemma do_set_eq_bind:
    assumes "m'=m"
    assumes "\<And>x. x\<in>m \<Longrightarrow> f x = g x"
    shows "do {x\<leftarrow>m; f x} = do {x\<leftarrow>m'; g x}"  
    using assms by auto







  (* Goal: Design simple concurrent modelling language
    Features: Processes, local/global integer variables
    *)

(* Grammar *)

  (* Literals *)
  type_synonym ident = string             (* Identifier *)
  type_synonym numeric = int              (* Numeric value *)

  datatype_new bin_op =                   (* Binary operators *)
  (* Standard arith *)
    bo_plus | bo_minus | bo_mul | bo_div | bo_mod  
  (* Comparison *)
  | bo_less | bo_less_eq | bo_eq
  (* Bit-Wise *)
  | bo_and | bo_or | bo_xor

  datatype_new un_op =                    (* Unary operators *)
    uo_minus | uo_not 

  (* Expressions *)
  datatype_new exp =
      e_var ident                       (* Variable *)
    | e_localvar ident                  (* Local variable *)
    | e_globalvar ident                 (* Global variable *)
    | e_const numeric                   (* Numeric constant *)
    | e_bin bin_op exp exp              (* Binary operators *) 
    | e_un un_op exp                    (* Unary operators *)    
  
  (* Commands *)
  datatype_new cmd =
    Assign ident exp 
  | Assign_local ident exp | Assign_global ident exp
  | Test exp 
  | Skip      (* Basic commands *)
  | Seq cmd cmd | Or cmd cmd | Loop cmd       (* Control flow *)
  | Break | Continue                          (* More control flow*) 
  | Iterate cmd cmd                           (* Iterate command *) 
  | Invalid                                   (* Internal: Invalid command *)  

  type_synonym var_decl = ident
  record proc_decl =
    body :: "cmd"
    local_vars :: "var_decl list"

  record program =
    processes :: "proc_decl list"
    global_vars :: "var_decl list"

(* Semantics *)

  (* Actions (edge labels) *)
  datatype_new action = 
      AAssign ident exp | AAssign_local ident exp | AAssign_global ident exp
    | ATest exp | ASkip

  (* Control flow graph, aka SOS-semantics ;) *)
  (* Internal actions *)  
  datatype_new brk_ctd = AIBreak | AIContinue


  inductive cfg :: "cmd \<Rightarrow> (action + brk_ctd) \<Rightarrow> cmd \<Rightarrow> bool" where
    Ass: "cfg (Assign x e) (Inl (AAssign x e)) Skip"
  | Assl: "cfg (Assign_local x e) (Inl (AAssign_local x e)) Skip"
  | Assg: "cfg (Assign_global x e) (Inl (AAssign_global x e)) Skip"
  | Test: "cfg (Test e) (Inl (ATest e)) Skip"

  | Seq1: "cfg (Seq Skip c2) (Inl ASkip) c2"
  | Seq2: "\<lbrakk> cfg c1 e c1' \<rbrakk> \<Longrightarrow> cfg (Seq c1 c2) e (Seq c1' c2)"

  | Or1: "cfg (Or c1 c2) (Inl ASkip) c1"
  | Or2: "cfg (Or c1 c2) (Inl ASkip) c2"

  | Loop: "cfg (Loop c) (Inl ASkip) (Iterate c c)"

  | Break: "cfg (Break) (Inr AIBreak) Invalid"
  | Continue: "cfg (Continue) (Inr AIContinue) Invalid"
  
  | Iterate1: "cfg (Iterate Skip c) (Inl ASkip) (Iterate c c)"
  | Iterate2: "\<lbrakk> cfg c (Inl e) c' \<rbrakk> \<Longrightarrow> cfg (Iterate c cb) (Inl e) (Iterate c' cb)"
  | IterateBrk: "\<lbrakk> cfg c (Inr AIBreak) _  \<rbrakk> \<Longrightarrow> cfg (Iterate c cb) (Inl ASkip) Skip"
  | IterateCtd: "\<lbrakk> cfg c (Inr AIContinue) _  \<rbrakk> \<Longrightarrow> cfg (Iterate c cb) (Inl ASkip) (Iterate cb cb)"

  inductive_cases Seq_cases: "cfg (Seq c1 c2) e c'"

  (* Well-formedness check: Break/Continue inside loops *)
  primrec wf_cmd_aux :: "cmd \<Rightarrow> bool" where
    "wf_cmd_aux (Assign _ _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Assign_local _ _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Assign_global _ _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Test _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Skip) \<longleftrightarrow> True"
  | "wf_cmd_aux (Seq c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd_aux (Or c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd_aux (Loop c1) \<longleftrightarrow> wf_cmd_aux c1"
  | "wf_cmd_aux (Break) \<longleftrightarrow> True"
  | "wf_cmd_aux (Continue) \<longleftrightarrow> True"
  | "wf_cmd_aux (Iterate c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd_aux (Invalid) \<longleftrightarrow> False"

  primrec wf_cmd :: "cmd \<Rightarrow> bool" where
    "wf_cmd (Assign _ _) \<longleftrightarrow> True"
  | "wf_cmd (Assign_local _ _) \<longleftrightarrow> True"
  | "wf_cmd (Assign_global _ _) \<longleftrightarrow> True"
  | "wf_cmd (Test _) \<longleftrightarrow> True"
  | "wf_cmd (Skip) \<longleftrightarrow> True"
  | "wf_cmd (Seq c1 c2) \<longleftrightarrow> wf_cmd c1 \<and> wf_cmd c2"
  | "wf_cmd (Or c1 c2) \<longleftrightarrow> wf_cmd c1 \<and> wf_cmd c2"
  | "wf_cmd (Loop c1) \<longleftrightarrow> wf_cmd_aux c1"
  | "wf_cmd (Break) \<longleftrightarrow> False"
  | "wf_cmd (Continue) \<longleftrightarrow> False"
  | "wf_cmd (Iterate c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd (Invalid) \<longleftrightarrow> False"
  

  (* Well-formedness is preserved *)
  lemma cfg_wf_aux1:
    assumes "cfg c (Inl e) c'"
    assumes "wf_cmd_aux c"
    shows "wf_cmd_aux c'"
    using assms
    by (induction c "Inl e::action+brk_ctd" c') auto

  lemma cfg_wf_aux2:
    assumes "wf_cmd c" 
    assumes "cfg c e c'" 
    shows "wf_cmd c' \<and> isl e"
    using assms(2,1)
    by induction (auto simp: cfg_wf_aux1)

  theorem cfg_wf: "\<lbrakk> wf_cmd c; cfg c e c' \<rbrakk> \<Longrightarrow> wf_cmd c'"
    -- "Steps in CFG preserve well-formedness"
    using cfg_wf_aux2 by simp
   
  theorem cfg_no_internal: "\<lbrakk> wf_cmd c; cfg c e c' \<rbrakk> \<Longrightarrow> isl e"
    -- "No internal actions on CFG-edges"
    using cfg_wf_aux2 by simp

  (* Commands do not get stuck (except on skip) *)
  lemma cfg_no_stuck_aux: "\<lbrakk>wf_cmd_aux c; c\<noteq>Skip\<rbrakk> \<Longrightarrow> \<exists>e c'. cfg c e c'"
    apply (induction c)
    apply (auto intro!: cfg.intros)
    apply (case_tac "c1=Skip")
    apply (auto intro!: cfg.intros) []
    apply clarsimp
    apply (case_tac e)
    apply (auto intro!: cfg.intros) []
    apply (case_tac b)
    apply (auto intro!: cfg.intros) [2]
    done

  theorem cfg_no_stuck: "\<lbrakk>wf_cmd c; c\<noteq>Skip\<rbrakk> \<Longrightarrow> \<exists>e c'. cfg c e c'"
    -- "Well-formed commands have outgoing edges, except skip"
    using assms
    by (induction c) (auto intro: cfg.intros cfg_no_stuck_aux)

  (* Values and evaluation *)  
  type_synonym word_size = 32 
  type_synonym signed = "word_size Word.word"
  
  abbreviation "word_len \<equiv> len_of TYPE(word_size)"
  abbreviation "signeds \<equiv> sints (len_of TYPE(word_size))"

  definition min_signed :: int where "min_signed \<equiv> -(2^(word_len - 1))"
  definition max_signed :: int where "max_signed \<equiv> 2^(word_len - 1) - 1"

  definition signed_of_int :: "int \<Rightarrow> signed" 
    where "signed_of_int i \<equiv> word_of_int i"

  definition int_of_signed :: "signed \<Rightarrow> int"
    where "int_of_signed i == sint i"

  lemma si_inv[simp]: "signed_of_int (int_of_signed i) = i"
    unfolding signed_of_int_def int_of_signed_def
    by simp

  lemma int_of_signed_in_range[simp]: 
    "int_of_signed i \<ge> min_signed"
    "int_of_signed i \<le> max_signed"
    unfolding int_of_signed_def min_signed_def max_signed_def
    apply -
    apply (rule sint_ge)
    using sint_lt[of i]
    by simp

  lemma is_inv[simp]: 
    "\<lbrakk> i\<ge>min_signed; i\<le>max_signed \<rbrakk> \<Longrightarrow> (int_of_signed (signed_of_int i)) = i"
    unfolding int_of_signed_def signed_of_int_def min_signed_def max_signed_def
    by (simp add: sints_num word_sint.Abs_inverse)

  type_synonym val = signed  
  
  primrec val_of_bool :: "bool \<Rightarrow> val" where
    "val_of_bool False = 0" | "val_of_bool True = 1"

  definition bool_of_val :: "val \<Rightarrow> bool" where
    "bool_of_val v \<equiv> v\<noteq>0"

  lemma bool_of_val_invs[simp]: 
    "bool_of_val (val_of_bool b) = b"  
    "val_of_bool (bool_of_val v) = (if v=0 then 0 else 1)"
    unfolding bool_of_val_def
    by (cases b) auto

  (* Attention: Silently overflows, using 2's complement.
    This should match Java-semantics. In C, signed overflow is undefined! *)
  primrec eval_bin_op :: "bin_op \<Rightarrow> val \<Rightarrow> val \<Rightarrow> val" where
    "eval_bin_op bo_plus v1 v2 = v1+v2"
  | "eval_bin_op bo_minus v1 v2 = v1-v2"
  | "eval_bin_op bo_mul v1 v2 = v1*v2"
  | "eval_bin_op bo_div v1 v2 = v1 sdiv v2"
  | "eval_bin_op bo_mod v1 v2 = v1 smod v2"
  | "eval_bin_op bo_less v1 v2 = val_of_bool (v1 < v2)"
  | "eval_bin_op bo_less_eq v1 v2 = val_of_bool (v1 \<le> v2)"
  | "eval_bin_op bo_eq v1 v2 = val_of_bool (v1 = v2)"
  | "eval_bin_op bo_and v1 v2 = v1 AND v2"
  | "eval_bin_op bo_or v1 v2 = v1 OR v2"
  | "eval_bin_op bo_xor v1 v2 = v1 XOR v2"

  primrec eval_un_op :: "un_op \<Rightarrow> val \<Rightarrow> val" where
    "eval_un_op uo_minus v = -v"   (*Attention: Silently overflows. -min_int = min_int! *)
  | "eval_un_op uo_not v = NOT v"
  
  (* Valuations, configurations *)
  type_synonym valuation = "ident \<rightharpoonup> val"

  record local_state = 
    variables :: valuation

  record local_config = 
    command :: cmd
    state :: local_state

  record global_state = 
    variables :: valuation

  record global_config =  
    processes :: "local_config multiset"
    state :: global_state

  type_synonym focused_state = "local_state \<times> global_state"

  type_synonym focused_config = "cmd \<times> focused_state \<times> local_config multiset"

  primrec assert_option :: "bool \<Rightarrow> unit option" where
    "assert_option True = Some ()" | "assert_option False = None"

  lemma assert_option_eqs[simp]:
    "assert_option \<Phi> = Some x \<longleftrightarrow> \<Phi>"  
    "assert_option \<Phi> = None \<longleftrightarrow> \<not>\<Phi>"  
    by (cases \<Phi>) auto

  abbreviation exists_var :: "valuation \<Rightarrow> ident \<Rightarrow> bool" where
    "exists_var s x \<equiv> x \<in> dom s"

  abbreviation update_var :: "ident \<Rightarrow> val \<Rightarrow> valuation \<Rightarrow> valuation" where
    "update_var x v s \<equiv> s(x\<mapsto>v)"

  fun eval_exp :: "exp \<Rightarrow> focused_state \<rightharpoonup> val" where
    "eval_exp (e_var x) (ls,gs) = do {
      let lv = local_state.variables ls; let gv = global_state.variables gs;
      case lv x of
        Some v \<Rightarrow> Some v
      | None \<Rightarrow> (case gv x of
          Some v \<Rightarrow> Some v
        | None \<Rightarrow> None)
    }"
  | "eval_exp (e_localvar x) (ls,gs) = local_state.variables ls x"
  | "eval_exp (e_globalvar x) (ls,gs) = global_state.variables gs x"
  | "eval_exp (e_const n) fs = do {
      assert_option (n\<ge>min_signed \<and> n\<le>max_signed);
      Some (signed_of_int n)
    }"
  | "eval_exp (e_bin bop e1 e2) fs = do {
      v1\<leftarrow>eval_exp e1 fs;
      v2\<leftarrow>eval_exp e2 fs;
      Some (eval_bin_op bop v1 v2)
      }"
  | "eval_exp (e_un uop e) fs = do {
      v\<leftarrow>eval_exp e fs;
      Some (eval_un_op uop v)
      }"

  (* Enabledness and effects of actions *)  
  primrec enabled_fs :: "action \<Rightarrow> focused_state \<rightharpoonup> bool" where
    "enabled_fs (AAssign _ _) _ = Some True"
  | "enabled_fs (AAssign_local _ _) _ = Some True"
  | "enabled_fs (AAssign_global _ _) _ = Some True"
  | "enabled_fs (ATest e) fs = do {
      v \<leftarrow> eval_exp e fs;
      Some (bool_of_val v)}"
  | "enabled_fs (ASkip) _ = Some True"

  (* Extension (for run, handshake): Let action have scheduler effect *)
  fun effect_fs :: "action \<Rightarrow> focused_state \<rightharpoonup> focused_state" where
    "effect_fs (AAssign x e) (ls,gs) = do {
      v \<leftarrow> eval_exp e (ls,gs);
      if exists_var (local_state.variables ls) x then do {
        let ls = local_state.variables_update (update_var x v) ls;
        Some (ls,gs)
      } else do {
        assert_option (exists_var (global_state.variables gs) x);
        let gs = global_state.variables_update (update_var x v) gs;
        Some (ls,gs)
      }
    }"
  | "effect_fs (AAssign_local x e) (ls,gs) = do {
      v \<leftarrow> eval_exp e (ls,gs);
      assert_option (exists_var (local_state.variables ls) x);
      let ls = local_state.variables_update (update_var x v) ls;
      Some (ls,gs)
    }"
  | "effect_fs (AAssign_global x e) (ls,gs) = do {
      v \<leftarrow> eval_exp e (ls,gs);
      assert_option (exists_var (global_state.variables gs) x);
      let gs = global_state.variables_update (update_var x v) gs;
      Some (ls,gs)
    }"
  | "effect_fs (ATest e) fs = do {
      v \<leftarrow> eval_exp e fs;
      assert_option (bool_of_val v); 
      Some fs
    }"
  | "effect_fs ASkip fs = Some fs"

  (* We now define a single global step, which may return
    a new configuration, or an error.
    We choose to define the set of successor states, using the set-monad.
  *)
  definition gstep_succ :: "global_config \<Rightarrow> global_config option set" where
    "gstep_succ gc \<equiv> do {
      (* Select process *)
      let lcs = global_config.processes gc;
      (lc,lcs') \<leftarrow> {(lc,lcs'). lcs = {#lc#} + lcs'};
      (* Focus state *)
      let gs = global_config.state gc;
      let ls = local_config.state lc;

      (* Select action *)
      let cmd = local_config.command lc;
      (a,cmd') \<leftarrow> {(a,cmd'). cfg cmd a cmd'};

      case a of
        Inr _ \<Rightarrow> {None} (* Internal action: Should not be visible \<longrightarrow> Error *)
      | Inl a \<Rightarrow> do {   (* Visible action *)
        (* Enabled or error *)
        case enabled_fs a (ls,gs) of
          None \<Rightarrow> {None} (* Error *)
        | Some False \<Rightarrow> {} (* Not enabled *)  
        | Some True \<Rightarrow> {do { (* Enabled *)
            (* Execute *)
            (ls,gs) \<leftarrow> effect_fs a (ls,gs);
            (* Unfocus state. *)
            Some \<lparr> 
              global_config.processes = {# 
                \<lparr> local_config.command = cmd', local_config.state = ls\<rparr> 
              #} + lcs',
              global_config.state = gs  
            \<rparr>
          }}
        }
    }"

  (* We extend the step function to a state of type option\<times>option, and
    convert to predicate representation *)  
  abbreviation gstep :: "global_config option \<Rightarrow> global_config option \<Rightarrow> bool" where
    "gstep \<equiv> pred_of_succ (m2r_succ gstep_succ)"

  (* For the final step function, we apply stutter extension *)
  definition "step \<equiv> stutter_extend gstep"

  (* Initial state *)
  definition init_valuation :: "var_decl list \<Rightarrow> valuation" where
    "init_valuation vd x \<equiv> if x \<in> set vd then Some 0 else None"

  definition init_pc :: "proc_decl \<Rightarrow> local_config" where
    "init_pc pd \<equiv> \<lparr>
      local_config.command = proc_decl.body pd,
      local_config.state = \<lparr>
        local_state.variables = init_valuation (proc_decl.local_vars pd)
      \<rparr>
    \<rparr>"

  definition init_gc :: "program \<Rightarrow> global_config" where
    "init_gc prog \<equiv> \<lparr>
      global_config.processes = multiset_of (map init_pc (program.processes prog)),
      global_config.state = \<lparr>
        global_state.variables = init_valuation (program.global_vars prog)
      \<rparr>
    \<rparr>"

  (* The final set of initial states: A singleton set *)
  definition init :: "program \<Rightarrow> global_config option \<Rightarrow> bool" where
    "init prog \<equiv> op = (Some (init_gc prog))"

  (* Labeling function: We are interested in global states, i.e.,
    the global variables, but no information about processes *)
  definition label where "label \<equiv> map_option global_config.state"

  (*******************************************)  
  (* Reference point semantics.
    All the remaining lemmas will refine this reference point.
  *)
  interpretation sys: lsystem "init prog" step label for prog .

  abbreviation "is_run \<equiv> sys.is_run"
  abbreviation "accept \<equiv> sys.accept"
  (*******************************************)  

  (* CFG successors *)
  primrec cfg_succ :: "cmd \<Rightarrow> ((action+brk_ctd) \<times> cmd) set" where
    "cfg_succ (Assign x e) = {((Inl (AAssign x e)), Skip)}"
  | "cfg_succ (Assign_local x e) = {((Inl (AAssign_local x e)), Skip)}"
  | "cfg_succ (Assign_global x e) = {((Inl (AAssign_global x e)), Skip)}"
  | "cfg_succ (Test e) = {((Inl (ATest e)), Skip)}"
  | "cfg_succ (Seq c1 c2) =
      (if c1=Skip then {((Inl ASkip), c2)} else {})
    \<union> { (e,Seq c1' c2) | c1' e.  (e,c1')\<in>cfg_succ c1}"
  | "cfg_succ (Or c1 c2) = {(Inl ASkip, c1), (Inl ASkip, c2)}"
  | "cfg_succ (Loop c) =  {((Inl ASkip), (Iterate c c))}"
  | "cfg_succ (Break) = {(Inr AIBreak, Invalid)}"
  | "cfg_succ (Continue) = {(Inr AIContinue, Invalid)}"
  | cfg_succ_iterate: 
    "cfg_succ (Iterate c cb) =
    (if c=Skip then {(Inl ASkip, Iterate cb cb)} else {})
    \<union> (
      \<lambda> (Inl e,c') \<Rightarrow> (Inl e, Iterate c' cb)
      | (Inr AIBreak,_) \<Rightarrow> (Inl ASkip,Skip)
      | (Inr AIContinue,_) \<Rightarrow> (Inl ASkip, Iterate cb cb)
    )`cfg_succ c"
  | "cfg_succ Skip = {}"
  | "cfg_succ Invalid = {}"


  declare cfg_succ_iterate[simp del]
  lemma cfg_succ_iterate'[simp]: "cfg_succ (Iterate c cb) = 
    (if c=Skip then {(Inl ASkip, Iterate cb cb)} else {})
    \<union> { (Inl e, Iterate c' cb) | e c'. (Inl e, c') \<in> cfg_succ c }
    \<union> (if \<exists>c'. (Inr AIBreak,c')\<in> cfg_succ c then { (Inl ASkip, Skip)} else {})
    \<union> (if \<exists>c'. (Inr AIContinue,c')\<in> cfg_succ c then { (Inl ASkip, Iterate cb cb)} else {})"
    unfolding cfg_succ_iterate
    apply (rule sym)
    apply (cases "c=Skip")
    apply simp
    apply safe
    apply simp_all
    apply force
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (auto split: split_if_asm sum.splits brk_ctd.splits)
    done


  export_code cfg_succ in SML

  lemma cfg_succ_correct': "(a,c')\<in>cfg_succ c \<longleftrightarrow> cfg c a c'"
    apply rule
    apply (induction c arbitrary: a c')
    apply (auto intro: cfg.intros split: split_if_asm)
    
    apply (induction rule: cfg.induct)
    apply auto
    done

  lemma cfg_succ_correct: "cfg_succ c = {(a,c'). cfg c a c'}"
    using cfg_succ_correct' by auto

  (* WF checks *)

  primrec fold_cmd :: "(cmd \<Rightarrow> 's \<Rightarrow> 's) \<Rightarrow> cmd \<Rightarrow> 's \<Rightarrow> 's" where
    "fold_cmd f (Assign x e) s = f (Assign x e) s"
  | "fold_cmd f (Assign_local x e) s = f (Assign_local x e) s"
  | "fold_cmd f (Assign_global x e) s = f (Assign_global x e) s"
  | "fold_cmd f (Test e) s = f (Test e) s"
  | "fold_cmd f Skip s = f Skip s"
  | "fold_cmd f (Seq c1 c2) s = (let 
      s = fold_cmd f c1 s;
      s = fold_cmd f c2 s;
      s = f (Seq c1 c2) s
      in s)"
  | "fold_cmd f (Or c1 c2) s = (let 
      s = fold_cmd f c1 s;
      s = fold_cmd f c2 s;
      s = f (Or c1 c2) s
      in s)"
  | "fold_cmd f (Loop c) s = (let 
      s = fold_cmd f c s;
      s = f (Loop c) s
      in s)"
  | "fold_cmd f Break s = f Break s"
  | "fold_cmd f Continue s = f Continue s"
  | "fold_cmd f (Iterate c1 c2) s = (let 
      s = fold_cmd f c1 s;
      s = fold_cmd f c2 s;
      s = f (Or c1 c2) s
      in s)"
  | "fold_cmd f Invalid s = f Invalid s"
  
  primrec fold_exp :: "(exp \<Rightarrow> 's \<Rightarrow> 's) \<Rightarrow> exp \<Rightarrow> 's \<Rightarrow> 's" where
    "fold_exp f (e_var x) s = f (e_var x) s"
  | "fold_exp f (e_localvar x) s = f (e_localvar x) s"
  | "fold_exp f (e_globalvar x) s = f (e_globalvar x) s"
  | "fold_exp f (e_const x) s = f (e_const x) s"
  | "fold_exp f (e_bin bop e1 e2) s = (let
      s = fold_exp f e1 s;
      s = fold_exp f e2 s;
      s = f (e_bin bop e1 e2) s
    in s)"
  | "fold_exp f (e_un uop e) s = (let
      s = fold_exp f e s;
      s = f (e_un uop e) s
    in s)"

  primrec subexps :: "exp \<Rightarrow> exp set" where
    "subexps (e_var x) = {e_var x}"
  | "subexps (e_localvar x) = {e_localvar x}"
  | "subexps (e_globalvar x) = {e_globalvar x}"
  | "subexps (e_const x) = {e_const x}"
  | "subexps (e_bin bop e1 e2) = insert (e_bin bop e1 e2) (subexps e1 \<union> subexps e2)"
  | "subexps (e_un uop e) = insert (e_un uop e) (subexps e)"

  primrec subcmds :: "cmd \<Rightarrow> cmd set" where
    "subcmds (Assign x e) = {Assign x e}"
  | "subcmds (Assign_local x e) = {Assign_local x e}"
  | "subcmds (Assign_global x e) = {Assign_global x e}"
  | "subcmds (Test e) = {Test e}"
  | "subcmds (Skip) = {Skip}"
  | "subcmds (Seq c1 c2) = insert (Seq c1 c2) (subcmds c1 \<union> subcmds c2)"
  | "subcmds (Or c1 c2) = insert (Or c1 c2) (subcmds c1 \<union> subcmds c2)"
  | "subcmds (Loop c) = insert (Loop c) (subcmds c)"
  | "subcmds (Break) = {Break}"
  | "subcmds (Continue) = {Continue}"
  | "subcmds (Iterate c1 c2) = insert (Iterate c1 c2) (subcmds c1 \<union> subcmds c2)"
  | "subcmds (Invalid) = {Invalid}"

  (* TODO: Move *)
  lemma set_bind_insert[simp]: "insert x s \<guillemotright>= f = f x \<union> (s\<guillemotright>=f)"
    by auto

  lemma set_bind_union[simp]: "(s1\<union>s2) \<guillemotright>= f = (s1\<guillemotright>=f) \<union> (s2\<guillemotright>=f)"
    by auto


  primrec udvars_of_exp :: "exp \<Rightarrow> ident set" where
      "udvars_of_exp (e_var x) = {x}"
    | "udvars_of_exp (e_localvar x) = {}"
    | "udvars_of_exp (e_globalvar x) = {}"
    | "udvars_of_exp (e_const c) = {}"
    | "udvars_of_exp (e_bin bop e1 e2) = (udvars_of_exp e1) \<union> (udvars_of_exp e2)"
    | "udvars_of_exp (e_un uop e) = (udvars_of_exp e)"

  primrec lvars_of_exp :: "exp \<Rightarrow> ident set" where
      "lvars_of_exp (e_var x) = {}"
    | "lvars_of_exp (e_localvar x) = {x}"
    | "lvars_of_exp (e_globalvar x) = {}"
    | "lvars_of_exp (e_const c) = {}"
    | "lvars_of_exp (e_bin bop e1 e2) = (lvars_of_exp e1) \<union> (lvars_of_exp e2)"
    | "lvars_of_exp (e_un uop e) = (lvars_of_exp e)"
      
  primrec gvars_of_exp :: "exp \<Rightarrow> ident set" where
      "gvars_of_exp (e_var x) = {}"
    | "gvars_of_exp (e_localvar x) = {}"
    | "gvars_of_exp (e_globalvar x) = {x}"
    | "gvars_of_exp (e_const c) = {}"
    | "gvars_of_exp (e_bin bop e1 e2) = (gvars_of_exp e1) \<union> (gvars_of_exp e2)"
    | "gvars_of_exp (e_un uop e) = (gvars_of_exp e)"

  primrec join_cmd_exp :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> (exp \<Rightarrow> 'a) \<Rightarrow> cmd \<Rightarrow> 'a" where  
    "join_cmd_exp z join f (Assign x e) = f e"
  | "join_cmd_exp z join f (Assign_local x e) = f e"
  | "join_cmd_exp z join f (Assign_global x e) = f e"
  | "join_cmd_exp z join f (Test e) = f e"
  | "join_cmd_exp z join f (Skip) = z"
  | "join_cmd_exp z join f (Seq c1 c2) 
      = join (join_cmd_exp z join f c1) (join_cmd_exp z join f c2)"
  | "join_cmd_exp z join f (Or c1 c2) 
      = join (join_cmd_exp z join f c1) (join_cmd_exp z join f c2)"
  | "join_cmd_exp z join f (Loop c) = (join_cmd_exp z join f c)"
  | "join_cmd_exp z join f (Break) = z"
  | "join_cmd_exp z join f (Continue) = z"
  | "join_cmd_exp z join f (Iterate c1 c2) 
      = join (join_cmd_exp z join f c1) (join_cmd_exp z join f c2)"
  | "join_cmd_exp z join f (Invalid) = z"

  abbreviation "union_cmd_exp \<equiv> join_cmd_exp {} op \<union>"
  abbreviation "all_cmd_exp \<equiv> join_cmd_exp True op \<and>"
  abbreviation "ex_cmd_exp \<equiv> join_cmd_exp False op \<or>"

  abbreviation "ud_rvars_of_cmd == union_cmd_exp udvars_of_exp"
  abbreviation "g_rvars_of_cmd == union_cmd_exp gvars_of_exp"
  abbreviation "l_rvars_of_cmd == union_cmd_exp lvars_of_exp"

  primrec ud_wvars_of_cmd :: "cmd \<Rightarrow> ident set" where  
    "ud_wvars_of_cmd (Assign x e) = {x}"
  | "ud_wvars_of_cmd (Assign_local x e) = {}"
  | "ud_wvars_of_cmd (Assign_global x e) = {}"
  | "ud_wvars_of_cmd (Test e) = {}"
  | "ud_wvars_of_cmd (Skip) = {}"
  | "ud_wvars_of_cmd (Seq c1 c2) = (ud_wvars_of_cmd c1 \<union> ud_wvars_of_cmd c2)"
  | "ud_wvars_of_cmd (Or c1 c2) = (ud_wvars_of_cmd c1 \<union> ud_wvars_of_cmd c2)"
  | "ud_wvars_of_cmd (Loop c) = (ud_wvars_of_cmd c)"
  | "ud_wvars_of_cmd (Break) = {}"
  | "ud_wvars_of_cmd (Continue) = {}"
  | "ud_wvars_of_cmd (Iterate c1 c2) = (ud_wvars_of_cmd c1 \<union> ud_wvars_of_cmd c2)"
  | "ud_wvars_of_cmd (Invalid) = {}"

  primrec l_wvars_of_cmd :: "cmd \<Rightarrow> ident set" where  
    "l_wvars_of_cmd (Assign x e) = {}"
  | "l_wvars_of_cmd (Assign_local x e) = {x}"
  | "l_wvars_of_cmd (Assign_global x e) = {}"
  | "l_wvars_of_cmd (Test e) = {}"
  | "l_wvars_of_cmd (Skip) = {}"
  | "l_wvars_of_cmd (Seq c1 c2) = (l_wvars_of_cmd c1 \<union> l_wvars_of_cmd c2)"
  | "l_wvars_of_cmd (Or c1 c2) = (l_wvars_of_cmd c1 \<union> l_wvars_of_cmd c2)"
  | "l_wvars_of_cmd (Loop c) = (l_wvars_of_cmd c)"
  | "l_wvars_of_cmd (Break) = {}"
  | "l_wvars_of_cmd (Continue) = {}"
  | "l_wvars_of_cmd (Iterate c1 c2) = (l_wvars_of_cmd c1 \<union> l_wvars_of_cmd c2)"
  | "l_wvars_of_cmd (Invalid) = {}"

  primrec g_wvars_of_cmd :: "cmd \<Rightarrow> ident set" where  
    "g_wvars_of_cmd (Assign x e) = {}"
  | "g_wvars_of_cmd (Assign_local x e) = {}"
  | "g_wvars_of_cmd (Assign_global x e) = {x}"
  | "g_wvars_of_cmd (Test e) = {}"
  | "g_wvars_of_cmd (Skip) = {}"
  | "g_wvars_of_cmd (Seq c1 c2) = (g_wvars_of_cmd c1 \<union> g_wvars_of_cmd c2)"
  | "g_wvars_of_cmd (Or c1 c2) = (g_wvars_of_cmd c1 \<union> g_wvars_of_cmd c2)"
  | "g_wvars_of_cmd (Loop c) = (g_wvars_of_cmd c)"
  | "g_wvars_of_cmd (Break) = {}"
  | "g_wvars_of_cmd (Continue) = {}"
  | "g_wvars_of_cmd (Iterate c1 c2) = (g_wvars_of_cmd c1 \<union> g_wvars_of_cmd c2)"
  | "g_wvars_of_cmd (Invalid) = {}"


  (*
  definition "udvars_of_exp e \<equiv> {x. e_var x \<in> subexps e}"
  definition "lvars_of_exp e \<equiv> {x. e_localvar x \<in> subexps e}"
  definition "gvars_of_exp e \<equiv> {x. e_globalvar x \<in> subexps e}"
  definition "exps_of_cmd c \<equiv> 
    \<Union>((\<lambda>Assign _ e \<Rightarrow> {e} 
      | Assign_local _ e \<Rightarrow> {e} 
      | Assign_global _ e \<Rightarrow> {e} 
      | Test e \<Rightarrow> {e} 
      | _ \<Rightarrow> {})`subcmds c)"

  definition "ud_rvars_of_cmd c \<equiv> exps_of_cmd c \<guillemotright>= udvars_of_exp"
  definition "ud_wvars_of_cmd c \<equiv> subcmds c \<guillemotright>= (\<lambda>Assign x _ \<Rightarrow> {x} | _ \<Rightarrow> {})"
  *)
  definition "udvars_of_cmd c \<equiv> ud_rvars_of_cmd c \<union> ud_wvars_of_cmd c"

  (*definition "l_rvars_of_cmd c \<equiv> exps_of_cmd c \<guillemotright>= lvars_of_exp"
  definition "l_wvars_of_cmd c \<equiv> subcmds c \<guillemotright>= (\<lambda>Assign_local x _ \<Rightarrow> {x} | _ \<Rightarrow> {})"*)
  definition "lvars_of_cmd c \<equiv> l_rvars_of_cmd c \<union> l_wvars_of_cmd c"

  (*definition "g_rvars_of_cmd c \<equiv> exps_of_cmd c \<guillemotright>= gvars_of_exp"
  definition "g_wvars_of_cmd c \<equiv> subcmds c \<guillemotright>= (\<lambda>Assign_global x _ \<Rightarrow> {x} | _ \<Rightarrow> {})"*)
  definition "gvars_of_cmd c \<equiv> g_rvars_of_cmd c \<union> g_wvars_of_cmd c"

  (* Decide variable locality *)
  fun dloc_exp :: "ident set \<Rightarrow> exp \<Rightarrow> exp" where
    "dloc_exp L (e_var x) = (if x\<in>L then e_localvar x else e_globalvar x)"
  | "dloc_exp _ (e_localvar x) = e_localvar x"  
  | "dloc_exp _ (e_globalvar x) = e_globalvar x"  
  | "dloc_exp _ (e_const c) = e_const c"
  | "dloc_exp L (e_bin bop e1 e2) = e_bin bop (dloc_exp L e1) (dloc_exp L e2)"
  | "dloc_exp L (e_un uop e) = e_un uop (dloc_exp L e)"

  fun dloc_cmd :: "ident set \<Rightarrow> cmd \<Rightarrow> cmd" where
    "dloc_cmd L (Assign x e) = 
      (if x\<in>L then Assign_local else Assign_global) x (dloc_exp L e)"
  | "dloc_cmd L (Assign_local x e) = Assign_local x (dloc_exp L e)"
  | "dloc_cmd L (Assign_global x e) = Assign_global x (dloc_exp L e)"
  | "dloc_cmd L (Test e) = Test (dloc_exp L e)"
  | "dloc_cmd L Skip = Skip"
  | "dloc_cmd L (Seq c1 c2) = Seq (dloc_cmd L c1) (dloc_cmd L c2)"
  | "dloc_cmd L (Or c1 c2) = Or (dloc_cmd L c1) (dloc_cmd L c2)"
  | "dloc_cmd L (Loop c) = Loop (dloc_cmd L c)"
  | "dloc_cmd L Break = Break"
  | "dloc_cmd L Continue = Continue"
  | "dloc_cmd L (Iterate c1 c2) = Iterate (dloc_cmd L c1) (dloc_cmd L c2)"
  | "dloc_cmd L Invalid = Invalid"

  definition dloc_proc :: "proc_decl \<Rightarrow> proc_decl" where
    "dloc_proc pd \<equiv> proc_decl.body_update (dloc_cmd (set (proc_decl.local_vars pd))) pd"

  definition dloc :: "program \<Rightarrow> program" where 
    "dloc \<equiv> program.processes_update (map dloc_proc)"

  abbreviation "ls_vars ls \<equiv> (dom (local_state.variables ls))"
  abbreviation "lc_vars lc \<equiv> ls_vars (local_config.state lc)"

  abbreviation "gs_vars gs \<equiv> (dom (global_state.variables gs))"
  abbreviation "gc_vars gc \<equiv> gs_vars (global_config.state gc)"

  definition dloc_lc :: "local_config \<Rightarrow> local_config" where
    "dloc_lc lc \<equiv> \<lparr> 
      local_config.command 
        = dloc_cmd 
            (lc_vars lc) 
            (local_config.command lc),
      local_config.state = local_config.state lc\<rparr>"
  
  definition dloc_gc :: "global_config \<Rightarrow> global_config" where
    "dloc_gc gc \<equiv> \<lparr>
      global_config.processes = image_mset dloc_lc (global_config.processes gc),
      global_config.state = global_config.state gc
    \<rparr>"

  lemma dom_init_valuation_eq_set[simp]: "dom (init_valuation l) = set l"
    unfolding init_valuation_def[abs_def]
    by (auto split: split_if_asm)

  lemma dloc_rel_init: "dloc_gc (init_gc prog) = init_gc (dloc prog)"
    unfolding dloc_gc_def init_gc_def dloc_lc_def[abs_def] init_pc_def[abs_def]
      dloc_def dloc_proc_def[abs_def] 
    apply (auto simp: multiset_of_map image_mset.compositionality)
    apply (fo_rule fun_cong arg_cong)+
    apply (auto)
    done

  lemma bind_set_map: "\<lbrakk> g`m=m'; \<And>x. x\<in>m \<Longrightarrow> f x = f' (g x) \<rbrakk> \<Longrightarrow> m\<guillemotright>=f = m'\<guillemotright>=f'"
    by auto

  lemma image_mset_eq_add_conv:
    "image_mset f m = a'+b' \<longleftrightarrow> (\<exists>a b. m=a+b \<and> image_mset f a = a' \<and> image_mset f b = b')"
    apply auto
    apply (induction m arbitrary: a' b')
    apply auto
    apply (erule mset_un_single_un_cases)
    apply auto
    apply (smt2 add.assoc add.commute image_mset_single image_mset_union mset_diff_union_s_inverse) 
    apply (smt2 add.assoc add.commute image_mset_single image_mset_union mset_diff_union_s_inverse) 
    done

  lemma image_mset_eq_sng_conv:
    "image_mset f m = {#y#} \<longleftrightarrow> (\<exists>x. m={#x#} \<and> y = f x)"
    by (cases m) auto

  fun dloc_a :: "ident set \<Rightarrow> action \<Rightarrow> action" where
    "dloc_a L ((AAssign x e)) = ((if x\<in>L then AAssign_local else AAssign_global) x (dloc_exp L e))"
  | "dloc_a L ((AAssign_local x e)) = (AAssign_local x (dloc_exp L e))"
  | "dloc_a L ((AAssign_global x e)) = (AAssign_global x (dloc_exp L e))"
  | "dloc_a L ((ATest e)) = (ATest (dloc_exp L e))"
  | "dloc_a L (ASkip) = ASkip"

  fun dloc_ai :: "ident set \<Rightarrow> action+brk_ctd \<Rightarrow> action+brk_ctd" where
    "dloc_ai L (Inr x) = Inr x"
  | "dloc_ai L (Inl x) = Inl (dloc_a L x)"

  lemma cfg_dloc: "cfg (dloc_cmd L c) da dc' \<longleftrightarrow> (\<exists>c' a. dc'=dloc_cmd L c' \<and> da=dloc_ai L a \<and> cfg c a c')"
  proof (safe)
    have aux1: "\<And>c1 c2 c' S. Seq c1 c2 = dloc_cmd S c' 
    \<longleftrightarrow> (\<exists>c1' c2'. c'=Seq c1' c2' \<and> c1=dloc_cmd S c1' \<and> c2=dloc_cmd S c2')"
      by (case_tac "(S,c')" rule: dloc_cmd.cases) auto

    have [simp]: "\<And>L c. Skip=dloc_cmd L c \<longleftrightarrow> c=Skip"
      by (case_tac c) auto

    assume A: "cfg (dloc_cmd L c) da dc'"
    thus "\<exists>c' a. dc'=dloc_cmd L c' \<and> da=dloc_ai L a \<and> cfg c a c'"
    proof (induction "dloc_cmd L c" da dc' arbitrary: c)
      case Ass thus ?case by (cases c) (auto intro: cfg.intros split: split_if_asm)
    next
      case (Assl x e) thus ?case by (cases c) (auto intro: cfg.intros intro!: exI split: split_if_asm)
    next
      case (Assg x e) thus ?case by (cases c) (auto intro: cfg.intros intro!: exI split: split_if_asm)
    next
      case (Test e) thus ?case by (cases c) (auto intro: cfg.intros intro!: exI split: split_if_asm)
    next
      case (Seq1 dc2 c) thus ?case
        apply (auto simp add: aux1)
        apply (intro exI conjI)
        apply (auto intro: cfg.Seq1)
        done
    next
      case (Seq2 dc1 da dc1' dc2 c)
      with aux1 obtain c1 c2 where [simp]: "c=Seq c1 c2" "dc1=dloc_cmd L c1" "dc2=dloc_cmd L c2"
        by auto
      from Seq2.hyps(2) obtain c1' a where 
        [simp]: "dc1' = dloc_cmd L c1'" "da = dloc_ai L a" 
        and 1: "cfg c1 a c1'"
        by fastforce
      from cfg.Seq2[OF 1] show ?case by force
    next
      case (Iterate2 dc1 da dc1' dc2)
      from Iterate2.hyps(3)
      obtain c1 c2 where [simp]: "c = Iterate c1 c2" "dc1 = dloc_cmd L c1" "dc2 = dloc_cmd L c2"
        by (cases c) (auto split: split_if_asm) 
      from Iterate2.hyps(2) obtain c1' a where 
        S1: "dc1' = dloc_cmd L c1'" "Inl da = dloc_ai L a" 
        and 1: "cfg c1 a c1'"
        by fastforce
      then obtain aa where [simp]: "a=Inl aa" by (cases a) (auto split: split_if_asm)
      note [simp] = S1
      from cfg.Iterate2[OF 1[simplified]] show ?case by force
    next
      case (IterateBrk dc1 dc1' dc2) 
      from IterateBrk.hyps(3)
      obtain c1 c2 where [simp]: "c = Iterate c1 c2" "dc1 = dloc_cmd L c1" "dc2 = dloc_cmd L c2"
        by (cases c) (auto split: split_if_asm) 
      from IterateBrk.hyps(2) obtain c1' a where 
        S1: "dc1' = dloc_cmd L c1'" and "Inr AIBreak = dloc_ai L a" 
        and 1: "cfg c1 a c1'"
        by fastforce
      hence [simp]: "a=Inr AIBreak" by (cases "(L,a)" rule: dloc_ai.cases) simp_all
      note [simp] = S1
      from cfg.IterateBrk[OF 1[simplified]] show ?case by force
    next
      case (IterateCtd dc1 dc1' dc2) 
      from IterateCtd.hyps(3)
      obtain c1 c2 where [simp]: "c = Iterate c1 c2" "dc1 = dloc_cmd L c1" "dc2 = dloc_cmd L c2"
        by (cases c) (auto split: split_if_asm) 
      from IterateCtd.hyps(2) obtain c1' a where 
        S1: "dc1' = dloc_cmd L c1'" and "Inr AIContinue = dloc_ai L a" 
        and 1: "cfg c1 a c1'"
        by fastforce
      hence [simp]: "a=Inr AIContinue" by (cases "(L,a)" rule: dloc_ai.cases) simp_all
      note [simp] = S1
      from cfg.IterateCtd[OF 1[simplified]] show ?case by force
    next
      apply_end (case_tac c, auto intro: cfg.intros intro!: exI split: split_if_asm) []
      apply_end (case_tac c, auto intro: cfg.intros intro!: exI split: split_if_asm) []
      apply_end (case_tac ca, auto intro: cfg.intros intro!: exI split: split_if_asm) []
      apply_end (case_tac c, auto intro: cfg.intros intro!: exI split: split_if_asm) []
      apply_end (case_tac c, auto intro: cfg.intros intro!: exI split: split_if_asm) []
      apply_end (case_tac ca, auto intro: cfg.intros intro!: exI split: split_if_asm) []
    qed
  next
    fix c' a
    assume "cfg c a c'"
    thus "cfg (dloc_cmd L c) (dloc_ai L a) (dloc_cmd L c')"
      by (induction) (auto intro: cfg.intros)
  qed
      
  lemma choose_action_as_map_aux: " 
    (map_prod (dloc_ai (lc_vars lc)) (dloc_cmd (lc_vars lc)))`{(a, c'). cfg (command lc) a c'}
    = {(a, c'). cfg (command (dloc_lc lc)) a c'}"
    by (auto simp: dloc_lc_def map_prod_def image_def cfg_dloc)

  lemma dloc_lc_state[simp]: "local_config.state (dloc_lc lc) = local_config.state lc"  
    unfolding dloc_lc_def by simp

  lemma dloc_gc_state[simp]: "global_config.state (dloc_gc gc) = global_config.state gc"  
    unfolding dloc_gc_def by simp

  lemma dloc_exp_correct[simp]: 
    "eval_exp (dloc_exp (ls_vars ls) e) (ls, gs) = eval_exp e (ls,gs)"
    by (induction e) (auto split: option.splits)

  lemma dloc_a_en_correct[simp]: 
    "enabled_fs (dloc_a (ls_vars ls) a) (ls, gs) = enabled_fs a (ls, gs)"
    by (induction a) auto

  lemma dloc_a_eff_correct[simp]:
    "effect_fs (dloc_a (ls_vars ls) a) (ls,gs) = effect_fs a (ls,gs)"
    by (induction a) auto

  lemma eff_pres_ls_vars[simp]: 
    "effect_fs a (ls,gs) = Some (ls',gs') \<Longrightarrow> ls_vars ls' = ls_vars ls"
    by (induction a) (auto split: split_if_asm option_bind_splits)

  lemma eff_pres_gs_vars[simp]: 
    "effect_fs a (ls,gs) = Some (ls',gs') \<Longrightarrow> gs_vars gs' = gs_vars gs"
    by (induction a) (auto split: split_if_asm option_bind_splits)

  lemma dloc_rel_gstep_succ: 
    "gstep_succ (dloc_gc gc) = map_option dloc_gc ` gstep_succ gc"
    apply (rule sym)
    unfolding gstep_succ_def 
    apply (simp add: do_set_push_Image case_prod_distrib 
      case_sum_distrib case_option_distrib)
    apply (rule bind_set_map[where g="map_prod dloc_lc (image_mset dloc_lc)"])
    apply (auto simp: dloc_gc_def image_def image_mset_eq_add_conv image_mset_eq_sng_conv) [] 
    apply (simp split: prod.splits)
    apply (rule bind_set_map[OF choose_action_as_map_aux])
    apply (simp split: prod.splits sum.splits, intro allI impI)
    apply (simp add: case_option_distrib)
    apply (auto split: option.splits bool.splits option_bind_splits
      simp: do_set_push_Image dloc_gc_def dloc_lc_def)
    done

  lemma dloc_rel_label: "label (map_option dloc_gc a) = label a"
    apply (cases a, simp_all)
    apply (auto simp: dloc_gc_def label_def)
    done

  interpretation dloc_sim: lbisimulation 
    "brp (map_option dloc_gc) (%_. True)" 
    "init prog" step label
    "init (dloc prog)" step label for prog
    unfolding step_def
    apply (rule lbisimulation.lstutter_extend)
    apply unfold_locales
    apply (auto simp: build_relp_def init_def
      simp: dloc_rel_init dloc_rel_gstep_succ dloc_rel_label)
    done

  thm dloc_sim.accept_bisim
  thm dloc_sim.brp_run_conv[OF refl]


  definition no_ovc_exp :: "exp \<Rightarrow> bool" where
    "no_ovc_exp e \<equiv> \<forall>c. e_const c \<in> subexps e \<longrightarrow> min_signed \<le> c \<and> c \<le> max_signed"

  abbreviation "no_ovc_cmd \<equiv> all_cmd_exp no_ovc_exp"
   
  (* Variable access wf-checks*)
  abbreviation vars_of_pd :: "proc_decl \<Rightarrow> ident set" 
    where "vars_of_pd pd \<equiv> set (proc_decl.local_vars pd)"

  definition vars_ok_cmd :: "ident set \<Rightarrow> ident set \<Rightarrow> cmd \<Rightarrow> bool" where
    "vars_ok_cmd G L cmd \<equiv>       
      gvars_of_cmd cmd \<subseteq> G 
    \<and> lvars_of_cmd cmd \<subseteq> L
    \<and> udvars_of_cmd cmd \<subseteq> G \<union> L"

  definition vars_ok_proc :: "ident set \<Rightarrow> proc_decl \<Rightarrow> bool" where
    "vars_ok_proc G pd \<equiv> vars_ok_cmd G (vars_of_pd pd) (proc_decl.body pd)"

  definition vars_ok :: "program \<Rightarrow> bool" where
    "vars_ok prog \<equiv> let
      gv = set (program.global_vars prog);
      pds = set (program.processes prog)
    in
      \<forall>pd\<in>pds. vars_ok_proc gv pd"

  abbreviation lvars_of_fs :: "focused_state \<Rightarrow> ident set" where 
    "lvars_of_fs fs \<equiv> ls_vars (fst fs)"

  abbreviation gvars_of_fs :: "focused_state \<Rightarrow> ident set" where 
    "gvars_of_fs fs \<equiv> gs_vars (snd fs)"

  definition cmds_of_program :: "program \<Rightarrow> cmd set" where
    "cmds_of_program prog \<equiv> proc_decl.body ` set (program.processes prog)"

  definition wf_program :: "program \<Rightarrow> bool" where
    "wf_program prog \<equiv> 
      (\<forall>cmd\<in>cmds_of_program prog. wf_cmd cmd \<and> no_ovc_cmd cmd)
    \<and> vars_ok prog"

  definition wf_global_config :: "global_config \<Rightarrow> bool" where
    "wf_global_config gc \<equiv> let
      gs = global_config.state gc
    in
      (\<forall>lc \<in> set_of (global_config.processes gc). 
        let cmd = local_config.command lc;
            ls = local_config.state lc
        in 
          wf_cmd cmd 
        \<and> no_ovc_cmd cmd 
        \<and> vars_ok_cmd (gs_vars gs) (ls_vars ls) cmd
      )"

  (* Initial configuration is well-formed *)
  theorem wf_init_gc: "wf_program prog \<Longrightarrow> wf_global_config (init_gc prog)"
    unfolding wf_program_def wf_global_config_def init_gc_def
    by (fastforce
      simp: init_pc_def cmds_of_program_def 
      simp: vars_ok_def vars_ok_proc_def init_valuation_def
      split: split_if_asm
      )

  lemma cfg_no_ovc: 
    assumes "cfg c e c'" 
    assumes "no_ovc_cmd c"  
    shows "no_ovc_cmd c'"
    using assms 
    by induction (auto simp: )

  (*  
  lemma l_rvars_simps: 
    "l_rvars_of_cmd (Seq c1 c2) = l_rvars_of_cmd c1 \<union> l_rvars_of_cmd c2"
    "l_rvars_of_cmd (Or c1 c2) = l_rvars_of_cmd c1 \<union> l_rvars_of_cmd c2"
    "l_rvars_of_cmd (Iterate c1 c2) = l_rvars_of_cmd c1 \<union> l_rvars_of_cmd c2"
    unfolding l_rvars_of_cmd_def
    by (auto simp: exps_of_cmd_def)
*)

  lemma cfg_vars_of_cmd_subset:
    notes [[goals_limit = 1]]
    assumes "cfg c a c'"
    shows 
      "lvars_of_cmd c' \<subseteq> lvars_of_cmd c"
      "gvars_of_cmd c' \<subseteq> gvars_of_cmd c"
      "udvars_of_cmd c' \<subseteq> udvars_of_cmd c"
      using assms apply induction
      apply (simp_all add: lvars_of_cmd_def gvars_of_cmd_def udvars_of_cmd_def)
      apply blast+
      done

  (* Well-formedness preserved by gsteps *)  
  lemma wf_gstep_succ: 
    assumes "Some gc' \<in> gstep_succ gc"    
    assumes "wf_global_config gc"
    shows "wf_global_config gc'"
    using assms
    unfolding gstep_succ_def
    apply clarsimp
    apply (split sum.split_asm, simp_all)
    apply (split option.split_asm, simp_all)
    apply (split bool.split_asm, simp_all)
    apply (split option_bind_split_asm, simp_all split: prod.split_asm)
    apply (auto simp: wf_global_config_def)
    apply (simp add: cfg_wf)
    apply (simp add: cfg_no_ovc)

    apply (frule cfg_vars_of_cmd_subset(1))
    apply (frule cfg_vars_of_cmd_subset(2))
    apply (frule cfg_vars_of_cmd_subset(3))
    apply (auto simp add: vars_ok_cmd_def) []
    done


  lemma eval_exp_no_err:
    "\<lbrakk>no_ovc_exp e; 
      lvars_of_exp e \<subseteq> lvars_of_fs fs;
      gvars_of_exp e \<subseteq> gvars_of_fs fs;
      udvars_of_exp e \<subseteq> gvars_of_fs fs \<union> lvars_of_fs fs
    \<rbrakk> \<Longrightarrow> eval_exp e fs \<noteq> None"
    apply (cases fs)
    apply (induction e)
    apply (auto split: option.splits option_bind_splits 
      simp: no_ovc_exp_def Collect_disj_eq)
    apply fastforce
    done
    
  (* CFG has no internal actions *)  
  definition "cfg' c a c' \<equiv> cfg c (Inl a) c'"
  lemma cfg'_eq: "wf_cmd c \<Longrightarrow> cfg c a c' \<longleftrightarrow> (\<exists>aa. a=Inl aa \<and> cfg' c aa c')"
    by (auto simp: cfg'_def dest!: cfg_no_internal simp: isl_def)

  (* Action enabledness check succeeds *)
  (*
  definition "enabled_fs' a fs \<equiv> the (enabled_fs a fs)"
  definition "effect_fs' a fs \<equiv> the (effect_fs a fs)"
  *)

  definition "enabled_fs' a fs \<equiv> case enabled_fs a fs of Some b \<Rightarrow> b | None \<Rightarrow> False"
  definition "effect_fs' a fs \<equiv> case effect_fs a fs of Some fs' \<Rightarrow> fs' | None \<Rightarrow> fs"


  lemma enabled_no_err:
    assumes "cfg' c a c'"
    assumes "vars_ok_cmd (gvars_of_fs fs) (lvars_of_fs fs) c"
    assumes "no_ovc_cmd c"
    shows "enabled_fs a fs = Some (enabled_fs' a fs)"
    using assms unfolding enabled_fs'_def cfg'_def
    apply (induction c "Inl a :: action + brk_ctd" c')
    apply simp_all
    apply (auto 
      split: option_bind_splits 
      simp: vars_ok_cmd_def gvars_of_cmd_def lvars_of_cmd_def udvars_of_cmd_def
      dest!: eval_exp_no_err)
    done

  (* Action effect succeeds *)
  lemma effect_no_err:
    assumes "cfg' c a c'"
    assumes "vars_ok_cmd (gvars_of_fs fs) (lvars_of_fs fs) c"
    assumes "no_ovc_cmd c"
    assumes "enabled_fs' a fs = True"
    shows "effect_fs a fs = Some (effect_fs' a fs)"
    using assms unfolding effect_fs'_def cfg'_def enabled_fs'_def
    apply (induction c "Inl a :: action + brk_ctd" c')
    apply (simp_all add: vars_ok_cmd_def)
    apply (auto 
      split: option_bind_splits 
      simp: vars_ok_cmd_def gvars_of_cmd_def lvars_of_cmd_def udvars_of_cmd_def
      dest!: eval_exp_no_err)
    apply (cases fs, auto dest!: eval_exp_no_err split: option_bind_splits) []
    apply (cases fs, auto dest!: eval_exp_no_err split: option_bind_splits) []
    apply (cases fs, auto dest!: eval_exp_no_err split: option_bind_splits) []
    apply (cases fs, auto dest!: eval_exp_no_err split: option_bind_splits) []
    done


  definition gstep_succ' :: "global_config \<Rightarrow> global_config set" where
    "gstep_succ' gc \<equiv> do {
      (* Select process *)
      let lcs = global_config.processes gc;
      (lc,lcs') \<leftarrow> {(lc,lcs'). lcs = {#lc#} + lcs'};
      (* Focus state *)
      let gs = global_config.state gc;
      let ls = local_config.state lc;

      (* Select action *)
      let cmd = local_config.command lc;
      (a,cmd') \<leftarrow> {(a,cmd'). cfg' cmd a cmd'};

      if enabled_fs' a (ls,gs) then {do { 
        (* Execute *)
        let (ls,gs) = effect_fs' a (ls,gs);
        (* Unfocus state. *)
        \<lparr> 
          global_config.processes = {# 
            \<lparr> local_config.command = cmd', local_config.state = ls\<rparr> 
          #} + lcs',
          global_config.state = gs  
        \<rparr>
      }}
    else {}
  }"

  definition "label' \<equiv> Some o global_config.state"
  definition "init' prog \<equiv> op = (init_gc prog)"
  definition "step' \<equiv> stutter_extend (pred_of_succ gstep_succ')"

  interpretation sys': lsystem "init' prog" step' label' for prog .

  abbreviation "accept' \<equiv> sys'.accept"
  abbreviation "is_run' \<equiv> sys'.is_run"

  lemma wf_select_action_aux: "wf_cmd c \<Longrightarrow> do { (a,c') \<leftarrow> {(a,c') . cfg c a c'}; f a c'}
    = do { (a,c') \<leftarrow> {(a,c') . cfg' c a c'}; f (Inl a) c'}"
    apply auto
    apply (drule (1) iffD1[OF cfg'_eq])
    apply auto
    apply (auto intro: iffD2[OF cfg'_eq])
    done

  lemma gstep_succ_no_err: 
    "wf_global_config gc \<Longrightarrow> gstep_succ gc = Some`gstep_succ' gc"
    unfolding gstep_succ_def gstep_succ'_def
    apply (simp add: do_set_push_Image case_prod_distrib 
      if_distrib cong: if_cong)
    apply (rule do_set_eq_bind, clarsimp_all)
    apply (clarsimp simp: wf_global_config_def)
    apply (subst wf_select_action_aux, simp_all)
    apply (rule do_set_eq_bind, clarsimp_all split del: split_if)
    apply (subst enabled_no_err)
      apply assumption apply auto [2]
    apply (clarsimp)  
    apply (subst effect_no_err)
      apply assumption apply auto [3]
    apply (auto split: prod.split) []
    done

  lemma step'_bisim_aux1:
    assumes "wf_global_config gc"
    assumes "s' \<in> gstep_succ gc"
    obtains gc' where "s'=Some gc'" "wf_global_config gc'" "gc'\<in>gstep_succ' gc"
    by (metis assms image_iff in_these_eq not_Some_eq gstep_succ_no_err 
      these_image_Some_eq wf_gstep_succ)

  locale wf_program_loc =  
    fixes prog
    assumes WFP[simp, intro!]: "wf_program prog"
  begin
    sublocale bisim': lbisimulation "brp Some wf_global_config" 
      "init' prog" "step'" label'
      "init prog" "step" label
      unfolding step'_def step_def
      apply (rule lbisimulation.lstutter_extend)
      apply unfold_locales
      apply (auto 
        simp: build_relp_def wf_init_gc[OF WFP]
        simp: init_def init'_def label'_def label_def
        dest: gstep_succ_no_err intro: wf_gstep_succ
        )
      apply (blast elim: step'_bisim_aux1)
      done

    lemma accept'_eq: "accept' prog = accept prog"
      using bisim'.accept_bisim .

    lemma is_run'_eq: "is_run' prog r = is_run prog (Some \<circ> r)"
      using bisim'.brp_inj_run_conv[of Some wf_global_config, simplified]
      ..
  end


(* A scheduler with process IDs *)
  type_synonym pid = nat

  record pid_global_config =  
    processes :: "local_config list"
    state :: global_state

  definition pidgc_\<alpha> :: "pid_global_config \<Rightarrow> global_config" where
    "pidgc_\<alpha> pgc \<equiv> \<lparr>     
      global_config.processes = multiset_of (pid_global_config.processes pgc),
      global_config.state = pid_global_config.state pgc
    \<rparr>"

  definition pid_init_gc :: "program \<Rightarrow> pid_global_config" where
    "pid_init_gc prog \<equiv> \<lparr>
      pid_global_config.processes = (map init_pc (program.processes prog)),
      pid_global_config.state = \<lparr>
        global_state.variables = init_valuation (program.global_vars prog)
      \<rparr>
    \<rparr>"

  definition pid_gstep_succ :: "pid_global_config \<Rightarrow> pid_global_config set" where
    "pid_gstep_succ gc \<equiv> do {
      (* Select process *)
      let lcs = pid_global_config.processes gc;
      pid \<leftarrow> {0..<length lcs};
      let lc = lcs!pid;
      (* Focus state *)
      let gs = pid_global_config.state gc;
      let ls = local_config.state lc;

      (* Select action *)
      let cmd = local_config.command lc;
      (a,cmd') \<leftarrow> {(a,cmd'). cfg' cmd a cmd'};

      if enabled_fs' a (ls,gs) then {do { 
        (* Execute *)
        let (ls,gs) = effect_fs' a (ls,gs);
        (* Unfocus state. *)
        let lc = \<lparr> local_config.command = cmd', local_config.state = ls\<rparr>;
        \<lparr> 
          pid_global_config.processes = lcs[pid:=lc],
          pid_global_config.state = gs  
        \<rparr>
      }}
    else {}
  }"

  lemma pid_init_correct: "pidgc_\<alpha> (pid_init_gc prog) = init_gc prog"
    unfolding pidgc_\<alpha>_def pid_init_gc_def init_gc_def
    by auto

  lemma in_multiset_of_conv_nth: "x \<in># multiset_of l \<longleftrightarrow> (\<exists>i<length l. x = l!i)"
    by (metis in_set_conv_nth mem_set_multiset_eq)
    
  lemma multiset_of_em_conv_nth: "multiset_of l = {#x#} + m' 
    \<longleftrightarrow> (\<exists>i<length l. x = l!i \<and> m' = multiset_of l - {#x#})"
    by (metis add_diff_cancel_left' in_multiset_of_conv_nth 
      insert_DiffM multi_member_this)
    

  lemma mset_select_couple_pid_select:
    assumes [simp]: "m = multiset_of m_impl"
    assumes F: "\<And>x m' i. \<lbrakk> m = {#x#} + m'; i<length m_impl; x=m_impl!i \<rbrakk> \<Longrightarrow> 
      (f x m') = (f_impl i)"
    shows " 
      (do { 
        (x,m') \<leftarrow> { (x,m'). m = {#x#} + m' };
        f x m'
      }) =
      (do { 
        i \<leftarrow> {0..<length m_impl};
        f_impl i
      })"
      using F
      by (fastforce simp: multiset_of_em_conv_nth)


  lemma pid_gstep_correct: 
    "gstep_succ' (pidgc_\<alpha> pgc) = pidgc_\<alpha>`pid_gstep_succ pgc"
    unfolding pidgc_\<alpha>_def[abs_def] gstep_succ'_def pid_gstep_succ_def
    apply (simp add: do_set_push_Image case_prod_distrib if_distrib 
      cong: if_cong)
    apply (rule mset_select_couple_pid_select, simp_all)
    apply (rule do_set_eq_bind, simp_all)
    apply (auto split: prod.splits simp: multiset_of_update algebra_simps)
    done

  definition "pid_step \<equiv> stutter_extend (pred_of_succ pid_gstep_succ)"
  definition "pid_init prog \<equiv> op = (pid_init_gc prog)"
  definition "pid_label \<equiv> Some o pid_global_config.state"

  interpretation pid_sys: lsystem "pid_init prog" pid_step pid_label for prog .
  abbreviation "pid_accept \<equiv> pid_sys.accept"
  abbreviation "pid_is_run \<equiv> pid_sys.is_run"

  lemma pid_label_correct: "pid_label a = label' (pidgc_\<alpha> a)"
    unfolding pid_label_def label'_def pidgc_\<alpha>_def
    by auto

  interpretation pid_bisim': lbisimulation "brp pidgc_\<alpha> (\<lambda>_. True)" 
    "pid_init prog" "pid_step" "pid_label"
    "init' prog" step' label'
    unfolding pid_step_def step'_def
    apply (rule lbisimulation.lstutter_extend)
    apply unfold_locales
    apply (auto simp: build_relp_def 
      simp: pid_init_correct pid_init_def init'_def
      simp: pid_gstep_correct
      simp: pid_label_correct
      )
    done

  context wf_program_loc
  begin
    sublocale pid_bisim: lbisimulation "brp (Some \<circ> pidgc_\<alpha>) (\<lambda>x. wf_global_config (pidgc_\<alpha> x))"
      "pid_init prog" pid_step pid_label 
      "init prog" step label
      using lbisimulation_trans[
        OF pid_bisim'.lbisimulation_this bisim'.lbisimulation_this,
        simplified] .

    thm pid_bisim.accept_bisim
    thm pid_bisim.right_unique_run_conv[simplified]  
  end

(** A scheduler that returns sets of actions **)

(* Global action: Consists of a process id, a local action, and a new location
  that is reached after the action.

  Note: This will not work out for atomic blocks, as the new location
  after the action is not immediately clear, without actually executing the
  atomic block, which is too expensive for just collecting enabled actions.
  *)

type_synonym global_action = "pid \<times> cmd \<times> action \<times> cmd"

definition ga_is_enabled :: "global_action \<Rightarrow> pid_global_config \<Rightarrow> bool" where
  "ga_is_enabled ga gc \<equiv> do {
    let (pid,_,a,_) = ga;
    let lcs = pid_global_config.processes gc;
    let lc = lcs!pid;
    (* Focus state *)
    let gs = pid_global_config.state gc;
    let ls = local_config.state lc;
    enabled_fs' a (ls,gs)
  }"

definition pid_gexecute :: "global_action \<Rightarrow> pid_global_config \<Rightarrow> pid_global_config" where
  "pid_gexecute ga gc \<equiv> do {
    let (pid,_,a,cmd') = ga;
    let lcs = pid_global_config.processes gc;
    let lc = lcs!pid;
    (* Focus state *)
    let gs = pid_global_config.state gc;
    let ls = local_config.state lc;

    (* Execute *)
    let (ls,gs) = effect_fs' a (ls,gs);

    (* Unfocus state. *)
    let lc = \<lparr> local_config.command = cmd', local_config.state = ls\<rparr>;
    \<lparr> 
      pid_global_config.processes = lcs[pid:=lc],
      pid_global_config.state = gs  
    \<rparr>
  }"


definition pid_get_actions_of_pid 
  :: "pid \<Rightarrow> pid_global_config \<Rightarrow> global_action set"
  where "pid_get_actions_of_pid pid gc \<equiv> do {
    let lcs = pid_global_config.processes gc;
    let lc = lcs!pid;

    (* Select action *)
    let cmd = local_config.command lc;
    (a,cmd') \<leftarrow> {(a,cmd'). cfg' cmd a cmd'};
    {(pid,cmd,a,cmd')}
  }"

definition pid_get_actions :: "pid_global_config \<Rightarrow> global_action set" where
  "pid_get_actions gc \<equiv> do {
    (* Select process *)
    let lcs = pid_global_config.processes gc;
    pid \<leftarrow> {0..<length lcs};
    let lc = lcs!pid;

    (* Select action *)
    let cmd = local_config.command lc;
    (a,cmd') \<leftarrow> {(a,cmd'). cfg' cmd a cmd'};
    {(pid,cmd,a,cmd')}
  }"
  
lemma pid_get_actions_alt: "pid_get_actions gc = do {
    (* Select process *)
    let lcs = pid_global_config.processes gc;
    pid \<leftarrow> {0..<length lcs};
    pid_get_actions_of_pid pid gc
  }"
  unfolding pid_get_actions_of_pid_def pid_get_actions_def
  by simp

lemma pid_get_actions_of_pid_pid:
  "(pid',cac)\<in>pid_get_actions_of_pid pid gc \<Longrightarrow> pid'=pid"
  by (auto simp: pid_get_actions_of_pid_def)


definition pid_genabled :: "pid_global_config \<Rightarrow> global_action set" where
  "pid_genabled gc \<equiv> {ga\<in>pid_get_actions gc.  ga_is_enabled ga gc}"

definition "pred_of_enex enex \<equiv> \<lambda>s s'. let (en,ex) = enex in \<exists>a\<in>en s. s'=ex a s"
definition "succ_of_enex enex \<equiv> \<lambda>s. let (en,ex) = enex in {s'. \<exists>a\<in>en s. s'=ex a s}"

lemma x_of_enex_expand[simp]: 
  "pred_of_enex (en,ex) s s' \<longleftrightarrow> (\<exists>a\<in>en s. s'=ex a s)"
  "s'\<in>succ_of_enex (en,ex) s \<longleftrightarrow> (\<exists>a\<in>en s. s'=ex a s)"
  unfolding pred_of_enex_def succ_of_enex_def by auto

lemma x_of_enex_conv[simp]: 
  "succ_of_pred (pred_of_enex enex) = succ_of_enex enex"
  "pred_of_succ (succ_of_enex enex) = pred_of_enex enex"
  unfolding pred_of_enex_def succ_of_enex_def 
  unfolding succ_of_pred_def pred_of_succ_def
  by auto

definition stutter_extend_en :: "('s\<Rightarrow>'a set) \<Rightarrow> ('s \<Rightarrow> 'a option set)" where
  "stutter_extend_en en \<equiv> \<lambda>s. let as = en s in if as={} then {None} else Some`as"

definition stutter_extend_ex :: "('a \<Rightarrow> 's \<Rightarrow> 's) \<Rightarrow> ('a option \<Rightarrow> 's \<Rightarrow> 's)" where
  "stutter_extend_ex ex \<equiv> \<lambda>None \<Rightarrow> id | Some a \<Rightarrow> ex a"

abbreviation stutter_extend_enex 
  :: "('s\<Rightarrow>'a set) \<times> ('a \<Rightarrow> 's \<Rightarrow> 's) \<Rightarrow> ('s \<Rightarrow> 'a option set) \<times> ('a option \<Rightarrow> 's \<Rightarrow> 's)"
where
 "stutter_extend_enex \<equiv> map_prod stutter_extend_en stutter_extend_ex"

lemma stutter_extend_pred_of_enex_conv:
  "stutter_extend (pred_of_enex enex) = pred_of_enex (stutter_extend_enex enex)"
  apply (intro ext)
  unfolding pred_of_enex_def 
  apply (auto simp: stutter_extend.simps stutter_extend_en_def stutter_extend_ex_def
    split: prod.splits)
  done

lemma stutter_extend_en_Some_eq[simp]:
  "Some a \<in> stutter_extend_en en gc \<longleftrightarrow> a \<in> en gc"
  "stutter_extend_ex ex (Some a) gc = ex a gc"
  unfolding stutter_extend_en_def stutter_extend_ex_def
  by auto

lemma stutter_extend_ex_None_eq[simp]:
  "stutter_extend_ex ex None = id"
  unfolding stutter_extend_ex_def
  by auto
  

definition "pid_enabled \<equiv> stutter_extend_en pid_genabled"
definition "pid_execute \<equiv> stutter_extend_ex pid_gexecute"

definition "pida_step \<equiv> pred_of_enex (pid_enabled,pid_execute)"

lemma pid_execute_None[simp]: "pid_execute None = (\<lambda>x. x)"
  unfolding pid_execute_def by auto

lemma pida_step_eq_pid_step: "pida_step = pid_step"
  unfolding pida_step_def pid_step_def
  unfolding pid_enabled_def pid_execute_def
  apply (subst stutter_extend_pred_of_enex_conv[
    of "(pid_genabled, pid_gexecute)",
    simplified, symmetric
  ])
  apply (fo_rule arg_cong)
  apply (intro ext)
  apply (simp 
    add: pid_genabled_def pid_get_actions_def pid_gexecute_def pid_gstep_succ_def
    add: ga_is_enabled_def)
  apply fastforce
  done  


(* Finite set of reachable CFG nodes *)

primrec reachable_term_order_aux :: "cmd \<Rightarrow> nat" where
  "reachable_term_order_aux (Assign x e) = 0"
| "reachable_term_order_aux (Assign_local x e) = 0"
| "reachable_term_order_aux (Assign_global x e) = 0"
| "reachable_term_order_aux (Test e) = 0"
| "reachable_term_order_aux (Skip) = 0"
| "reachable_term_order_aux (Seq c1 c2) = reachable_term_order_aux c1 + reachable_term_order_aux c2"
| "reachable_term_order_aux (Or c1 c2) = reachable_term_order_aux c1 + reachable_term_order_aux c2"
| "reachable_term_order_aux (Loop c) = 1 + 2*reachable_term_order_aux c"
| "reachable_term_order_aux (Break) = 0"
| "reachable_term_order_aux (Continue) = 0"
| "reachable_term_order_aux (Iterate c1 c2) = reachable_term_order_aux c1 + reachable_term_order_aux c2"
| "reachable_term_order_aux Invalid = 0"

function reachable :: "cmd \<Rightarrow> cmd set" where
  "reachable (Assign x e) = {Assign x e, Skip}"
| "reachable (Assign_local x e) = {Assign_local x e, Skip}"
| "reachable (Assign_global x e) = {Assign_global x e, Skip}"
| "reachable (Test e) = {Test e, Skip}"
| "reachable (Skip) = {Skip}"
| "reachable (Seq c1 c2) = (reachable c2) \<union> (do { c1\<leftarrow>reachable c1; {Seq c1 c2} })"
| "reachable (Or c1 c2) = insert (Or c1 c2) (reachable c1 \<union> reachable c2)"
| "reachable (Loop c) = insert (Loop c) (reachable (Iterate c c))"
| "reachable (Break) = {Break,Invalid}"
| "reachable (Continue) = {Continue,Invalid}"
| "reachable (Iterate c1 c2) = insert Skip
    (do { c1\<leftarrow>reachable c1; {Iterate c1 c2} })
  \<union> (do { c2'\<leftarrow>reachable c2; {Iterate c2' c2} })"
| "reachable Invalid = {Invalid}"
by pat_completeness auto

termination
  apply (relation "reachable_term_order_aux <*mlex*> measure size")
  apply (metis wf_measure wf_mlex)
  apply (auto simp: mlex_prod_def)
  done

lemma finite_bindI[intro]:
  assumes "finite m"
  assumes "\<And>x. x\<in>m \<Longrightarrow> finite (f x)"
  shows "finite (do { x\<leftarrow>m; f x })"
proof -
  have S: "do { x\<leftarrow>m; f x } = \<Union>(f`m)"
    by auto

  show ?thesis  
    apply (subst S)
    using assms
    by blast
qed        


lemma finite_reachable[simp, intro!]: "finite (reachable c)"
  by (induction c rule: reachable.induct) auto

lemma reachable_refl[simp]: "c\<in>reachable c"
  by (induction c rule: reachable.induct) auto

(*lemma reachable_closed: 
  assumes "cfg c a c'" 
  shows "c'\<in>reachable c"
  using assms
  by induction simp_all
*)

lemma reachable_closed:
  assumes "c\<in>reachable c0"
  assumes "cfg c a c'"
  shows "c'\<in>reachable c0"
  using assms
  apply (induction c0 arbitrary: c c' rule: reachable.induct)
  apply (auto elim: cfg.cases) (* Takes quite long. *)
  done


(* Set of reachable states is finite *)
(* Approximation of reachable CFG-states *)
definition "cfg_approx prog \<equiv> \<Union>{reachable (proc_decl.body p) | p. p\<in>set (program.processes prog)}"

definition "local_var_approx prog \<equiv> \<Union>{  vars_of_pd p | p. p\<in>set (program.processes prog) }"
definition "global_var_approx prog \<equiv> set (program.global_vars prog)"

definition "lc_approx prog \<equiv> { 
  \<lparr> 
    local_config.command = c, 
    local_config.state = \<lparr> local_state.variables = v\<rparr> \<rparr> 
  | c v. c \<in> cfg_approx prog \<and> dom v \<subseteq> local_var_approx prog }"

definition "gc_approx prog \<equiv> {
  \<lparr> global_config.processes = lcs,
    global_config.state = \<lparr> global_state.variables = v \<rparr>
  \<rparr> | lcs v.
    size lcs \<le> length (program.processes prog)
  \<and> set_of lcs \<subseteq> lc_approx prog
  \<and> dom v \<subseteq> global_var_approx prog
}"

definition "gco_approx prog \<equiv> insert None (Some`gc_approx prog)"



lemma init_approx: "init prog gc \<Longrightarrow> gc\<in>gco_approx prog"
  unfolding init_def gco_approx_def
  apply hypsubst
  apply simp  
  apply (rule imageI)
  apply (thin_tac "?a")
  unfolding init_gc_def gc_approx_def
  apply (auto simp: 
    init_pc_def lc_approx_def cfg_approx_def local_var_approx_def global_var_approx_def)
  apply force
  apply force
  done
  
lemma cfg_approx:
  assumes "c\<in>cfg_approx prog"
  assumes "cfg c a c'"
  shows "c'\<in>cfg_approx prog"
  using assms
  unfolding cfg_approx_def
  by (auto dest: reachable_closed)


lemma step_approx: 
  assumes "gc\<in>gco_approx prog" 
  assumes "step gc gc'"
  shows "gc'\<in>gco_approx prog"
  using assms
  unfolding step_def
  apply (cases gc, simp_all add: gco_approx_def)
  apply (auto simp:  elim!: stutter_extend.cases) []
  apply (cases gc', simp_all)
  apply (rule imageI)
  apply (auto elim!: stutter_extend.cases) []
  
  unfolding gstep_succ_def
  apply clarsimp
  apply (clarsimp split: sum.splits option.splits bool.splits option_bind_splits)
  apply (clarsimp simp: gco_approx_def gc_approx_def)
  apply (case_tac bb, clarsimp_all)
  apply (frule eff_pres_gs_vars, clarsimp)
  apply (clarsimp simp: lc_approx_def)
  apply (case_tac ac, clarsimp_all)
  apply (frule eff_pres_ls_vars, clarsimp)

  apply (auto simp: cfg_approx) []
  done  
  
lemma "finite { (x,y). P x y } \<Longrightarrow> finite { f x y | x y. P x y }"
  using [[simproc add: finite_Collect]]
  by (simp)

(* TODO: Move to Misc, near lists_of_len_fin1 - lemma *)
lemma lists_le_len_fin: "finite P \<Longrightarrow> finite (lists P \<inter> { l. length l \<le> n })"
proof (induct n)
  case 0 thus ?case by auto
next
  case (Suc n)
  have "lists P \<inter> { l. length l \<le> Suc n } 
        = (lists P \<inter> {l. length l \<le> n}) 
        \<union> (\<lambda>(a,l). a#l) ` (P \<times> (lists P \<inter> {l. length l \<le> n}))"
    apply auto
    apply (case_tac x)
    apply auto
    done
  moreover from Suc have "finite \<dots>" by auto
  ultimately show ?case by simp
qed



lemma obtain_list_of_multiset: 
  obtains l where "multiset_of l = m"
proof -
  have "\<exists>l. multiset_of l = m"
    apply (induction m)
    apply auto
    apply (rule_tac x="x#l" in exI)
    apply auto
    done
  thus ?thesis using that by blast
qed

lemma finite_size_bounded_msets:
  assumes "finite S"
  shows "finite { m . size m \<le> len \<and> set_of m \<subseteq> S }"
proof -
  have "{ m . size m \<le> len \<and> set_of m \<subseteq> S } 
    \<subseteq> multiset_of`(lists S \<inter> {l. length l \<le> len})"
    apply clarsimp
    apply (rule_tac m=x in obtain_list_of_multiset)
    apply (auto intro!: imageI)
    done  
  also have "finite (multiset_of`(lists S \<inter> {l. length l \<le> len}))"
    apply (rule finite_imageI)
    apply (rule lists_le_len_fin)
    by (rule assms)
  finally (finite_subset) show ?thesis .
qed
  
(*
lemma "map_to_set` (dom -` Pow d) \<subseteq> (Pow (d \<times> UNIV))"
  apply (simp add: vimage_def)
  apply (auto simp: map_to_set_def)
*)  

lemma finite_local_var_approx[simp, intro!]: "finite (local_var_approx prog)"
  using [[simproc add: finite_Collect]]
  unfolding local_var_approx_def
  apply (rule finite_Union)
  apply simp
  apply clarsimp
  done

lemma finite_global_var_approx[simp, intro!]: "finite (global_var_approx prog)"
  using [[simproc add: finite_Collect]]
  unfolding global_var_approx_def
  by simp

lemma finite_maps_to_finite_type:
  assumes "finite (D::'a set)" "finite (UNIV::'v set)" 
  shows "finite ((dom::(_ \<rightharpoonup> 'v) \<Rightarrow> _) -` Pow D)"
proof -
  {fix DD::"'a set"
    have "{x. dom x \<subseteq> DD} = \<Union>{ {x. dom x = D } | D.  D\<subseteq>DD }"  
    by auto
  } note aux2=this

  from assms show ?thesis
    apply (simp add: vimage_def)
    apply (subst aux2)
    apply (rule finite_Union)
    using [[simproc add: finite_Collect]]
    apply simp
    apply clarsimp
    apply (rule finite_set_of_finite_maps[where B="UNIV::'v set", simplified])
    apply (erule finite_subset)
    apply simp
    by simp
qed

lemma gco_approx_finite[simp, intro!]: "finite (gco_approx prog)"  
proof -
  have aux1: "({lcs.
        size lcs
        \<le> length (program.processes prog)} \<inter>
       set_of -` Pow (lc_approx prog)) 
    = {lcs. size lcs \<le> length (program.processes prog) \<and> set_of lcs \<subseteq> lc_approx prog }"
    by auto

  {fix DD::"ident set"
    have "{x. dom x \<subseteq> DD} = \<Union>{ {x. dom x = D } | D.  D\<subseteq>DD }"  
    by auto
  } note aux2=this

  show ?thesis
    unfolding gco_approx_def
    apply simp
    apply (rule finite_imageI)
    unfolding gc_approx_def
    using [[simproc add: finite_Collect]]
    apply simp
    apply (rule finite_imageI)
    apply (subst aux1)
    apply (rule finite_cartesian_product)
    apply (rule finite_size_bounded_msets)
    unfolding lc_approx_def
    apply simp
    apply (rule finite_imageI)
    apply (rule finite_cartesian_product)
    unfolding cfg_approx_def
    apply (rule finite_Union)
    apply simp
    apply auto []
    apply (simp add: finite_maps_to_finite_type)
    apply (simp add: finite_maps_to_finite_type)
    done
qed

lemma sys_finite_reachable: "finite (sys.reachable prog)"
proof -
  {fix x
    assume "x\<in>sys.reachable prog"
    hence "x\<in>gco_approx prog"
      apply induction
      apply (auto intro: step_approx init_approx)
      done
  } hence "sys.reachable prog \<subseteq> gco_approx prog" by blast
  also note gco_approx_finite
  finally (finite_subset) show ?thesis .
qed

term dloc_sys

lemma "finite (sys.reachable (dloc prog))"
  by (rule sys_finite_reachable)


lemma finitely_many_lists_for_multiset[simp, intro!]: 
  "finite {l . multiset_of l = m}"
proof -
  have "{l . multiset_of l = m} \<subseteq> {l. set l \<subseteq> set_of m \<and> length l = size m}"
    by auto
  also note finite_lists_length_eq
  finally (finite_subset) show ?thesis by auto
qed
  

context wf_program_loc begin
  lemma sys'_finite_reachable: "finite (sys'.reachable prog)"
  proof -
    show ?thesis
      apply (rule bisim'.s1.reachable_finite_sim)
      apply (rule sys_finite_reachable)
      apply (clarsimp simp: build_relp_def)
      apply (case_tac b)
      apply simp
      apply simp
      done
  qed

  lemma pid_sys_finite_reachable: "finite (pid_sys.reachable prog)"
  proof -
    { fix lcs s
      have "{gc. 
        multiset_of (pid_global_config.processes gc) = lcs 
        \<and> pid_global_config.state gc = s}
      = ((\<lambda>a. pid_global_config.make a s))`{a. multiset_of a = lcs}"
        apply (auto simp: pid_global_config.make_def)
        apply (case_tac x, auto)
        done
    } note aux=this

    show ?thesis
      apply (rule pid_bisim'.s1.reachable_finite_sim)
      apply (rule sys'_finite_reachable)
      apply (clarsimp simp: build_relp_def pidgc_\<alpha>_def)
      apply (case_tac b)
      apply clarsimp
      apply (subst aux)
      apply (rule finite_imageI)
      by simp
  qed
end



(* Program analysis: Set of variables in actions *)
fun read_globals :: "action \<Rightarrow> ident set" where
  "read_globals (AAssign _ e) = gvars_of_exp e \<union> udvars_of_exp e"
| "read_globals (AAssign_local _ e) = gvars_of_exp e \<union> udvars_of_exp e"
| "read_globals (AAssign_global _ e) = gvars_of_exp e \<union> udvars_of_exp e"
| "read_globals (ATest e) = gvars_of_exp e \<union> udvars_of_exp e"
| "read_globals (ASkip) = {}"

fun write_globals :: "action \<Rightarrow> ident set" where
  "write_globals (AAssign x _) = {x}"
| "write_globals (AAssign_local _ _) = {}"
| "write_globals (AAssign_global x _) = {x}"
| "write_globals (ATest e) = {}"
| "write_globals (ASkip) = {}"

fun is_local_action :: "global_action option \<Rightarrow> bool" where
  "is_local_action None \<longleftrightarrow> False"
| "is_local_action (Some (_,_,a,_)) \<longleftrightarrow> 
    read_globals a = {} \<and> write_globals a = {}"

fun is_action_pid_indep :: "global_action option \<Rightarrow> global_action option \<Rightarrow> bool"
where
  "is_action_pid_indep None _ \<longleftrightarrow> False"
| "is_action_pid_indep _ None \<longleftrightarrow> False"
| "is_action_pid_indep (Some (pid1,_)) (Some (pid2,_)) \<longleftrightarrow> pid1\<noteq>pid2"

definition eq_on_pid :: "pid \<Rightarrow> pid_global_config \<Rightarrow> pid_global_config \<Rightarrow> bool" where
  "eq_on_pid pid gc gc' \<equiv> 
    (pid_global_config.processes gc!pid = pid_global_config.processes gc'!pid)
    \<and> pid<length (pid_global_config.processes gc) 
    \<and> pid<length (pid_global_config.processes gc')"

lemma local_exp_indep_g:
  assumes "gvars_of_exp e = {}" "udvars_of_exp e = {}"
  shows "eval_exp e (ls, gs) = eval_exp e (ls, gs')"
  using assms
  by (induction e) (auto)

lemma local_act_en_indep_g:
  assumes "read_globals a = {}"
  shows "enabled_fs' a (ls,gs) \<longleftrightarrow> enabled_fs' a (ls,gs')"
  using assms
  unfolding enabled_fs'_def
  by (cases a) (auto split: option_bind_splits 
    simp: local_exp_indep_g[where gs=gs and gs'=gs'])
          
lemma local_act_ex_indep_g1:
  assumes "write_globals a = {}"
  shows "\<exists>ls'. effect_fs' a (ls,gs) = (ls',gs)"
  using assms
  unfolding effect_fs'_def
  by (cases a) (auto split: option_bind_splits)

lemma local_act_ex_indep_g2:
  assumes "read_globals a = {}"
  assumes "effect_fs' a (ls,gs) = (ls',gs')"
  shows "\<exists>gs'''. effect_fs' a (ls,gs'') = (ls',gs''')"
  using assms
  unfolding effect_fs'_def
  apply (cases a) 
  apply (auto split: option_bind_splits option.splits split_if_asm
    simp: local_exp_indep_g[where gs=gs and gs'=gs''])
  done



lemma 
  assumes "is_local_action (Some (pid,c,a,c'))"
  assumes "eq_on_pid pid gc gc'"
  shows "(Some (pid,c,a,c')) \<in> pid_enabled gc 
  \<longleftrightarrow> (Some (pid,c,a,c')) \<in> pid_enabled gc'"
  using assms
  unfolding eq_on_pid_def pid_enabled_def 
  by (auto simp: pid_genabled_def pid_get_actions_def ga_is_enabled_def
    local_act_en_indep_g[THEN iffD1])

lemma exec_indep_preserves_lc:
  assumes "pid' < length (pid_global_config.processes gc)"
  assumes "pid\<noteq>pid'"
  shows "eq_on_pid pid' (pid_execute (Some (pid,c,a,c')) gc) gc"
  using assms
  unfolding eq_on_pid_def pid_enabled_def pid_execute_def 
  by (auto simp: pid_gexecute_def split: prod.splits)
  
(*lemma exec_local_preserves_gc:
  assumes "is_local_action (Some (pid,c,a,c'))"
  assumes "(Some (pid,c,a,c')) \<in> pid_enabled gc"
  assumes "gc'=pid_execute (Some (pid,c,a,c')) gc"
  shows "pid_global_config.state gc' = pid_global_config.state gc"
  using assms
  unfolding eq_on_pid_def pid_enabled_def pid_execute_def 
  apply (auto 
    simp: pid_gexecute_def pid_genabled_def ga_is_enabled_def
    split: prod.splits
    dest: local_act_ex_indep_g1)
*)

fun local_enabled :: "global_action option \<Rightarrow> pid_global_config \<Rightarrow> bool" where
  "local_enabled None _ \<longleftrightarrow> False"
| "local_enabled (Some (pid,c,a,c')) gc \<longleftrightarrow> (let
    nump = length (pid_global_config.processes gc);
    lc = pid_global_config.processes gc ! pid;
    ls = local_config.state lc
  in 
    pid<nump 
  \<and> c = command lc
  \<and> cfg' c a c'
  \<and> enabled_fs' a (ls,undefined))"

lemma local_enabled_eq:
  assumes "is_local_action a" 
  shows "a\<in>pid_enabled gc \<longleftrightarrow> local_enabled a gc"
  apply rule
  using assms
  unfolding pid_enabled_def stutter_extend_en_def
  apply (cases a rule: is_local_action.cases, simp_all) []
  apply (auto split: split_if_asm 
    simp: pid_genabled_def pid_get_actions_def ga_is_enabled_def
    dest: local_act_en_indep_g[where gs'=undefined, THEN iffD1]
    ) []
  
  using assms
  apply (cases a rule: is_local_action.cases, simp_all) []
  apply (auto
    simp: pid_genabled_def pid_get_actions_def ga_is_enabled_def
    dest: local_act_en_indep_g[where gs'=undefined, THEN iffD2]
    )
  done

fun local_execute 
  :: "global_action option \<Rightarrow> pid_global_config \<Rightarrow> pid_global_config"
  where 
  "local_execute None gc = gc"
| "local_execute (Some (pid,c,a,c')) gc = (let
    lc = pid_global_config.processes gc ! pid;
    ls = local_config.state lc;
    (ls',_) = effect_fs' a (ls, undefined);
    lc' = \<lparr> local_config.command=c', local_config.state = ls' \<rparr>;
    gc' = gc\<lparr> pid_global_config.processes := pid_global_config.processes gc[pid := lc']  \<rparr>
  in gc'
)"

lemma local_execute_eq:
  assumes "is_local_action a" 
  shows "pid_execute a gc = local_execute a gc"
  using assms
  apply (cases a rule: is_local_action.cases)
  apply (auto 
    split: prod.splits
    simp: pid_execute_def pid_gexecute_def
    )
  by (smt2 local_act_ex_indep_g1 local_act_ex_indep_g2 old.unit.exhaust 
    pid_global_config.surjective pid_global_config.update_convs(1) prod.inject)

lemma effect_fs_write_globals:
  assumes "effect_fs' a (l,g) = (l',g')"
  shows "global_state.variables g' |` (-write_globals a) = 
    global_state.variables g |` (-write_globals a)"
  using assms
  unfolding effect_fs'_def
  apply (auto split: option.splits)
  apply (cases a)
  apply (auto split: split_if_asm option_bind_splits)
  done

context system_initial begin
  definition step where "step = pred_of_enex (en,ex)"

  sublocale peter_system!: system "\<lambda>x. x\<in>init" step .

  lemma reachable_eq_step_rtrancl: "s \<in> reachable \<longleftrightarrow> (\<exists>s0\<in>init. step\<^sup>*\<^sup>* s0 s)"
  proof safe
    assume "s \<in> reachable"
    thus "\<exists>s0\<in>init. step\<^sup>*\<^sup>* s0 s"
      apply (induction rule: reachable.induct)
      apply auto [2]
      apply (metis rtranclp.rtrancl_into_rtrancl system_initial.step_def 
        x_of_enex_expand(1))
      done
  next
    fix s0
    assume "step\<^sup>*\<^sup>* s0 s" "s0\<in>init"
    thus "s \<in> reachable"
      apply (induct rule: rtranclp_induct)
      apply (auto intro: reachable.intros)
      by (metis (no_types, hide_lams) system_initial.reachable.step 
        system_initial.step_def x_of_enex_expand(1))
  qed

  lemma compat_reachable: "reachable = peter_system.reachable"
    unfolding peter_system.reachable_def peter_system.is_reachable_def[abs_def]
    using reachable_eq_step_rtrancl
    by blast

  lemma step_enex_conv: "step s s' \<longleftrightarrow> (\<exists>a. a\<in>en s \<and> s'=ex a s)"
    unfolding step_def pred_of_enex_def
    by auto

  lemma rtrancl_step_exec_word_conv:  
    "step\<^sup>*\<^sup>* s s' \<longleftrightarrow> (\<exists>w\<in>words_fin s. s' = fold ex w s)"
  proof safe
    assume "step\<^sup>*\<^sup>* s s'"
    thus "\<exists>w\<in>words_fin s. s' = fold ex w s"
      apply induction
      apply force
      apply (force simp: step_def pred_of_enex_def)
      done
  next
    fix w
    assume "w\<in>words_fin s"
    thus "local.step\<^sup>*\<^sup>* s (fold ex w s)"
      apply induction
      apply (auto simp: step_def pred_of_enex_def
        intro: converse_rtranclp_into_rtranclp
        )
      done      
  qed


  corollary step_of_enexD:
    assumes "a\<in>en s"
    shows "step s (ex a s)"
    using assms
    by (auto simp: step_enex_conv)

  corollary step_of_exec_wordD:
    assumes "w\<in>words_fin s"
    shows "step\<^sup>*\<^sup>* s (fold ex w s)"
    using assms
    by (auto simp: rtrancl_step_exec_word_conv)



  definition "lstep s a s' \<equiv> a\<in>en s \<and> s'=ex a s"
  definition "lpath s w s' \<equiv> w\<in>words_fin s \<and> s'=fold ex w s"

  lemma lpath_empty[simp]: "lpath s [] s' \<longleftrightarrow> s=s'"
    unfolding lpath_def by auto

  lemma lpath_cons: "lpath s (a#w) s' \<longleftrightarrow> (\<exists>sh. lstep s a sh \<and> lpath sh w s')"
    unfolding lpath_def lstep_def by auto

  lemma lpath_conc: "lpath s (w1@w2) s' \<longleftrightarrow> (\<exists>sh. lpath s w1 sh \<and> lpath sh w2 s')"
    unfolding lpath_def by auto
    
  lemmas lpath_simps = lpath_empty lpath_cons lpath_conc

  lemma lpath_induct[induct pred: lpath, case_names empty step]:
    assumes "lpath s w s'"
    assumes "P [] s'"
    assumes "\<And>s a sh wh. \<lbrakk> lstep s a sh; lpath sh wh s'; P wh sh \<rbrakk> \<Longrightarrow> P (a#wh) s"
    shows "P w s"
    using assms
    apply (induction w arbitrary: s)
    by (auto simp: lpath_simps)

  lemma step_lstep_conv: "step s s' \<longleftrightarrow> (\<exists>a. lstep s a s')"
    unfolding step_def lstep_def pred_of_enex_def
    by auto

  lemma steps_lpath_conv: "step\<^sup>*\<^sup>* s s' \<longleftrightarrow> (\<exists>w. lpath s w s')"
    unfolding lpath_def
    using rtrancl_step_exec_word_conv by blast

  lemma fold_lstep_iff: "a\<in>en s \<longleftrightarrow> lstep s a (ex a s)"
    unfolding lstep_def by simp

  lemma fold_lpath_iff: "w\<in>words_fin s \<longleftrightarrow> lpath s w (fold ex w s)"
    unfolding lpath_def by simp

  lemmas fold_lstepD = fold_lstep_iff[THEN iffD1]
  lemmas fold_lpathD = fold_lpath_iff[THEN iffD1]

end    

locale wf_program_vis_vars_loc = wf_program_loc +
  fixes vis_vars :: "ident set"
begin

definition interp :: "pid_global_config \<Rightarrow> (ident \<rightharpoonup> val)" where
  "interp gc \<equiv> global_state.variables (pid_global_config.state gc) |` vis_vars"

fun is_invisible_action :: "global_action option \<Rightarrow> bool" where
  "is_invisible_action None \<longleftrightarrow> False"
| "is_invisible_action (Some (pid,c,a,c')) \<longleftrightarrow> write_globals a \<inter> vis_vars = {}"

term reachable


lemma invisible_ex_interp:
  "\<lbrakk>is_invisible_action a\<rbrakk> \<Longrightarrow> interp (pid_execute a gc) = interp gc"
  unfolding interp_def
  apply (cases a rule: is_invisible_action.cases)
  apply (auto 
    simp: pid_execute_def pid_gexecute_def restrict_map_def intro!: ext
    dest!: effect_fs_write_globals
    split: prod.splits)
  by (metis IntI empty_iff)
  
  
fun ind :: "global_action option \<Rightarrow> global_action option \<Rightarrow> bool" where
  "ind None None \<longleftrightarrow> True"
| "ind (Some _) None \<longleftrightarrow> False"
| "ind None (Some _) \<longleftrightarrow> False"
| "ind (Some (pid,cac)) (Some (pid',cac')) 
  \<longleftrightarrow> 
    pid\<noteq>pid'
  \<and> (is_local_action (Some (pid,cac)) 
     \<or> is_local_action (Some (pid',cac')))
  \<and> (is_invisible_action (Some (pid,cac)) 
     \<or> is_invisible_action (Some (pid',cac')))"

lemma ind_symmetric: "ind a b \<longleftrightarrow> ind b a"
  by (cases "(a,b)" rule: ind.cases) auto

abbreviation pid_valid :: "pid_global_config \<Rightarrow> pid \<Rightarrow> bool" where
  "pid_valid gc pid \<equiv> pid < length (pid_global_config.processes gc)"

lemma enabled_indep_other_processes:
  assumes "pid_global_config.state gc' = pid_global_config.state gc"
  assumes "pid_global_config.processes gc'!pid = pid_global_config.processes gc!pid"
  assumes "pid_valid gc pid"
  assumes "Some (pid,cac)\<in>pid_enabled gc'"
  shows "Some (pid,cac)\<in>pid_enabled gc"
  using assms
  unfolding pid_enabled_def
  by (auto simp: pid_genabled_def pid_get_actions_def ga_is_enabled_def)

lemma enabled_pid_valid[simp, intro]: 
  "Some (pid,cac)\<in>pid_enabled gc \<Longrightarrow> pid_valid gc pid"
  unfolding pid_enabled_def
  by (auto simp: pid_genabled_def pid_get_actions_def)

lemma enabled_indep_change_other_process[simp]:
  assumes "pid\<noteq>pid'"
  shows "Some (pid', a)
           \<in> pid_enabled
               (gc\<lparr>pid_global_config.processes :=
                    pid_global_config.processes gc[pid := lc']\<rparr>) 
      \<longleftrightarrow> Some (pid', a) \<in> pid_enabled gc"
  using assms
  apply safe
  apply (rule enabled_indep_other_processes)
  apply (auto dest: enabled_pid_valid) [4]
  
  apply (rule enabled_indep_other_processes)
  apply (auto) [4]
  done

lemma pid_execute_preserves_valid_pids[intro, simp]:
  assumes "pid_valid gc pid"
  shows "pid_valid (pid_execute a gc) pid"
  using assms
  unfolding pid_execute_def
  apply (cases a)
  apply simp
  apply simp
  apply (auto simp: pid_gexecute_def split: prod.splits)
  done

lemma pid_execute_preserves_valid_pids_sym[simp]:
  (* Warning: Only if no dynamic process creation. 
    Otherwise:
      Tune ind: Store pid of created process in action 
        or make proc-creation actions dependent on everything (over-approx)
  *)
  shows "pid_valid (pid_execute a gc) pid \<longleftrightarrow> pid_valid gc pid"
  using assms
  unfolding pid_execute_def
  apply (cases a)
  apply simp
  apply simp
  apply (auto simp: pid_gexecute_def split: prod.splits)
  done



lemma pid_execute_preserves_other_procs[simp]:
  assumes "pid\<noteq>pid'"
  shows "pid_global_config.processes (pid_execute (Some (pid,cac)) gc) ! pid' 
    = pid_global_config.processes gc ! pid'"
  using assms
  by (auto simp: pid_execute_def pid_gexecute_def split: prod.splits)

abbreviation "gc_upd_lc gc pid lc 
  \<equiv> gc\<lparr> pid_global_config.processes := pid_global_config.processes gc[pid:=lc] \<rparr>"

lemma pid_execute_upd_other_lc_eq:
  assumes "pid\<noteq>pid'"
  shows "pid_execute (Some (pid,cac)) (gc_upd_lc gc pid' lc) 
    = gc_upd_lc (pid_execute (Some (pid,cac)) gc) pid' lc"
  using assms
  by (auto simp: pid_execute_def pid_gexecute_def list_update_swap split: prod.splits)

lemma ind_en: "\<lbrakk>ind a b; a \<in> pid_enabled p\<rbrakk>
       \<Longrightarrow> (b \<in> pid_enabled (pid_execute a p)) = (b \<in> pid_enabled p)"
  apply (cases "(a,b)" rule: ind.cases, simp_all)
  apply (safe, simp_all 
    del: is_local_action.simps 
    add: local_execute_eq local_enabled_eq)
  apply (auto split: prod.splits)
  done

lemma ind_ex:
  assumes IND: "ind a b" 
  (*assumes EN: "a \<in> pid_enabled p" "b \<in> pid_enabled p"*)
  shows "pid_execute b (pid_execute a p) = pid_execute a (pid_execute b p)"
  using IND
  apply (cases "(a,b)" rule: ind.cases, simp_all)
  apply (safe, simp_all 
    del: is_local_action.simps 
    add: local_execute_eq local_enabled_eq)
  apply (auto split: prod.splits simp: pid_execute_upd_other_lc_eq)
  done
  

sublocale pid_system!: system_complete 
  pid_enabled pid_execute "Collect (pid_init prog)" interp
  .

sublocale pid_system!: system_finite_reachable 
  pid_enabled pid_execute "Collect (pid_init prog)"
  apply (unfold_locales)
  unfolding pid_system.compat_reachable system_initial.step_def
  apply simp
  apply (fold pida_step_def)
  unfolding pida_step_eq_pid_step
  by (rule pid_sys_finite_reachable)
    

sublocale pid_system!: system_traces pid_enabled pid_execute ind
  apply unfold_locales
  apply (simp_all add: ind_symmetric ind_en ind_ex)
  done
    
sublocale pid_system: system_interpreted_traces 
  pid_enabled pid_execute interp
  ind
  apply unfold_locales
  apply (case_tac "(a,b)" rule: ind.cases)
  apply (auto simp: pid_system.visible_def invisible_ex_interp)
  done

lemma pid_system_step_is_pida_step: "pid_system.step = pida_step"
  unfolding pida_step_def pid_system.step_def ..



definition "cfg_E \<equiv> {(c,c'). \<exists>a. cfg' c a c'}"

definition "vis_edges cmd \<equiv> {
  (c,c'). \<exists>a. c\<in>reachable cmd \<and> cfg' c a c' 
    \<and> write_globals a \<inter> vis_vars \<noteq> {} }"

definition collect_sticky_local :: "(cmd \<times> action \<times> cmd) set nres"
where
"collect_sticky_local \<equiv> do {
  let pds = program.processes prog;
  sticky \<leftarrow> FOREACHi 
    (\<lambda>it s. \<forall>pd\<in>set pds - it. is_fas cfg_E {proc_decl.body pd} s) 
    (set pds) 
    (\<lambda>pd sticky. do {
      let cmd = proc_decl.body pd;
      let vis = vis_edges cmd;
      new_sticky \<leftarrow> find_fas_init (cfg_E) {cmd} vis;
      RETURN (sticky \<union> new_sticky)
    })
    {};
  RETURN {(c,a,c'). (c,c')\<in>sticky \<or> write_globals a \<inter> vis_vars \<noteq> {}}
}"   (* TODO/FIXME: the write_globals a \<inter> vis_vars \<noteq> {} is redundant!  *)


(*definition collect_sticky_edges :: "global_action option set nres"
where
"collect_sticky_edges \<equiv> do {
  let pds = program.processes prog;
  sticky \<leftarrow> FOREACHi 
    (\<lambda>it s. \<forall>pd\<in>set pds - it. is_fas cfg_E {proc_decl.body pd} s) 
    (set pds) 
    (\<lambda>pd sticky. do {
      let cmd = proc_decl.body pd;
      let vis = vis_edges cmd;
      new_sticky \<leftarrow> find_fas_init (cfg_E) {cmd} vis;
      RETURN (sticky \<union> new_sticky)
    })
    {};
  let sticky_actions = Collect (
    \<lambda> None \<Rightarrow> True 
  | Some (pid,c,a,c') \<Rightarrow> (c,c')\<in>sticky \<or> \<not>is_invisible_action (Some (pid,c,a,c')));
  RETURN sticky_actions
}"*)

definition actions_of_sticky_local 
  :: "(cmd \<times> action \<times> cmd) set \<Rightarrow> global_action option set"
  where
  "actions_of_sticky_local sl \<equiv> insert None { Some (pid,cac) | pid cac. cac\<in>sl}"

definition collect_sticky_edges :: "global_action option set nres"
where "collect_sticky_edges = do {
  sl \<leftarrow> collect_sticky_local;
  RETURN (actions_of_sticky_local sl)
}"


lemma cfg_succ_finite[simp, intro!]: "finite (cfg_succ c)"
  apply (induction c)
  using [[simproc add: finite_Collect]]
  by (simp_all del: cfg_succ_iterate' add: cfg_succ_iterate)

lemma cfg_E_finitely_branching[simp, intro!]: "finite (cfg_E``{c})"
proof -
  have "cfg_E``{c} \<subseteq> snd`cfg_succ c"
    unfolding cfg_E_def cfg'_def cfg_succ_correct
    by auto
  also have "finite (snd`cfg_succ c)" by auto
  finally (finite_subset) show ?thesis .
qed

sublocale sys_graph!: fb_graph cfg_E "{cmd}" for cmd
  apply unfold_locales
  apply simp_all
  done

lemma cfg_E_ss_reachable: "cfg_E\<^sup>*``{c} \<subseteq> reachable c"
proof safe 
  fix c'
  assume "(c,c')\<in>cfg_E\<^sup>*"
  thus "c'\<in>reachable c"
    by (induction) (auto simp: cfg_E_def cfg'_def intro: reachable_closed)
qed
    
lemma is_fas_mono: 
  assumes "is_fas E V0 F" "F\<subseteq>F'" 
  shows "is_fas E V0 F'"
proof -
  from assms have 1: "E - F' \<subseteq> E - F" by blast
  show ?thesis
    using assms unfolding is_fas_def 
    by (auto dest: trancl_mono[OF _ 1])
qed

abbreviation "lc_of_pid gc pid 
  \<equiv> pid_global_config.processes gc ! pid"

abbreviation "cmd_of_pid gc pid 
  \<equiv> local_config.command (lc_of_pid gc pid)"

lemma pida_step_preserves_valid: 
  "\<lbrakk> pid_valid gc pid; pida_step gc gc' \<rbrakk> \<Longrightarrow> pid_valid gc' pid"
  unfolding pida_step_def pred_of_enex_def
  by auto

lemma pida_steps_preserve_valid: 
  "\<lbrakk> pid_valid gc pid; pida_step\<^sup>*\<^sup>* gc gc' \<rbrakk> \<Longrightarrow> pid_valid gc' pid"
  apply (rotate_tac)
  apply (induction rule: rtranclp_induct)
  apply (auto intro: pida_step_preserves_valid)
  done

lemma pida_no_stutter_proj_cfg_E:
  assumes "Some (pid,c,a,c') \<in> pid_enabled gc"
  shows "(cmd_of_pid gc pid, cmd_of_pid (pid_execute (Some (pid,c,a,c')) gc) pid)
    \<in> cfg_E \<inter> {(c,c')}"
  using assms
  apply (auto
    simp: pid_enabled_def pid_execute_def pid_genabled_def
    simp: pid_gexecute_def 
    split: prod.splits
    )
  apply (simp_all add: pid_get_actions_def cfg'_def cfg_E_def)
  apply auto
  done

definition cfg_E_of_lpath :: "global_action option list \<Rightarrow> (cmd\<times>cmd) set"
where "cfg_E_of_lpath w \<equiv> {(c,c'). \<exists>pid a. Some (pid,c,a,c') \<in> set w}"

lemma cfg_E_of_lpath_simps[simp]:
  "cfg_E_of_lpath [] = {}"
  "cfg_E_of_lpath (a#w) = (case a of Some (_,c,_,c') \<Rightarrow> {(c,c')} | _ \<Rightarrow> {}) 
    \<union> cfg_E_of_lpath w"
  "cfg_E_of_lpath (w1@w2) = cfg_E_of_lpath w1 \<union> cfg_E_of_lpath w2"
  unfolding cfg_E_of_lpath_def
  by (auto split: option.splits)

(* TODO: Move *)
lemma pid_execute_None[simp]: "pid_execute None = id"
  unfolding pid_execute_def by simp

(* TODO: Move *)
lemma reflcl_rtrancl_trans[trans]: "\<lbrakk> (a,b)\<in>R\<^sup>=; (b,c)\<in>R\<^sup>* \<rbrakk> \<Longrightarrow> (a,c)\<in>R\<^sup>*"
  by auto

lemma reflcl_trancl_trans[trans]: "\<lbrakk> (a,b)\<in>R\<^sup>=; (b,c)\<in>R\<^sup>+ \<rbrakk> \<Longrightarrow> (a,c)\<in>R\<^sup>+"
  by auto

lemma project_lstep_cfg_E:
  assumes "pid_system.lstep gc a gc'"
  shows "(cmd_of_pid gc pid, cmd_of_pid gc' pid)\<in>(cfg_E \<inter> cfg_E_of_lpath [a])\<^sup>="
  using assms
  unfolding pid_system.lstep_def
  apply (auto 
    simp: pid_enabled_def pid_execute_def
    simp: pid_genabled_def pid_gexecute_def
    simp: ga_is_enabled_def cfg_E_def
    split: option.splits prod.split)
  apply (simp_all add:  cfg'_def pid_get_actions_def)
  apply (case_tac "pid=aa", auto) []
  apply (case_tac "pid=aa", auto) []
  apply (case_tac "pid=aa", auto) []
  done

lemma project_lpath_cfg_E:
  assumes "pid_system.lpath gc w gc'"
  assumes "pid_valid gc pid" (* TODO: This assumption is probably not required *)
  shows "(cmd_of_pid gc pid, cmd_of_pid gc' pid)
    \<in> (cfg_E \<inter> cfg_E_of_lpath w)\<^sup>*"
  using assms
proof (induction w arbitrary: gc)
  case Nil thus ?case by auto
next
  case (Cons a w)
  then obtain gch where 
    A: "pid_system.lstep gc a gch" and W: "pid_system.lpath gch w gc'"
    by (auto simp: pid_system.lpath_simps)

  from A have "(cmd_of_pid gc pid, cmd_of_pid gch pid)\<in>(cfg_E \<inter> cfg_E_of_lpath [a])\<^sup>="
    by (rule project_lstep_cfg_E)
  hence "(cmd_of_pid gc pid, cmd_of_pid gch pid)\<in>(cfg_E \<inter> cfg_E_of_lpath (a#w))\<^sup>="
    by auto
  also  
  from Cons.prems A have "pid_valid gch pid"
    by (auto simp: pid_system.lstep_def)
  from Cons.IH[OF W this] have 
    "(cmd_of_pid gch pid, cmd_of_pid gc' pid)\<in>(cfg_E \<inter> cfg_E_of_lpath w)\<^sup>*" .
  hence "(cmd_of_pid gch pid, cmd_of_pid gc' pid)\<in>(cfg_E \<inter> cfg_E_of_lpath (a#w))\<^sup>*"
    by (rule set_rev_mp[OF _ rtrancl_mono]) auto
  finally show ?case .
qed
    

(* TODO: Move *)
lemma r_rtrancl_trans_union: "\<lbrakk> (a,b)\<in>R; (b,c)\<in>S\<^sup>* \<rbrakk> \<Longrightarrow> (a,c)\<in>(R\<union>S)\<^sup>+"
proof -
  assume "(a,b)\<in>R"
  hence "(a,b)\<in>R\<union>S" by auto
  also assume "(b,c)\<in>S\<^sup>*"
  hence "(b,c)\<in>(R\<union>S)\<^sup>*" 
    by (rule rtrancl_mono[THEN set_mp, rotated]) auto
  finally (rtrancl_into_trancl2) show ?thesis .
qed
  
fun action_case_split :: "global_action option \<Rightarrow> unit" where
  "action_case_split None = ()"
| "action_case_split (Some (_,_,_,_)) = ()"


lemma collect_sticky_edges_correct:
  "collect_sticky_edges \<le> SPEC (\<lambda>cuts. 
    (\<forall>gc\<in>pid_system.reachable. \<forall>w\<in>pid_system.cycles gc. w\<noteq>[] \<longrightarrow> set w \<inter> cuts \<noteq>{})
  \<and> (Collect (Not o is_invisible_action) \<subseteq> cuts)  
  )" 
proof -
  interpret unit_DFS cfg_E "{cmd}" for cmd by unfold_locales


  {
    fix gc pid c a c' gc' w
    assume 1: "Some (pid,c,a,c') \<in> pid_enabled gc"
    assume 2: "pid_system.lpath (pid_execute (Some (pid,c,a,c')) gc) w gc'"
    
    have "pid_valid gc pid" using enabled_pid_valid[OF 1] .
    hence PV': "pid_valid (pid_execute (Some (pid,c,a,c')) gc) pid"
      using pid_execute_preserves_valid_pids by blast

    note pida_no_stutter_proj_cfg_E[OF 1]
    also note project_lpath_cfg_E[OF 2 PV']
    finally (r_rtrancl_trans_union) have 
      "(cmd_of_pid gc pid, cmd_of_pid gc' pid) \<in> 
        (cfg_E \<inter> (insert (c,c') (cfg_E_of_lpath w)))\<^sup>+"
      by (simp add: Int_Un_distrib[symmetric])
  } note aux1 = this


  { (* No processes created. Has to changed for process creation to:
      pid either already there, or at initial cfg-node *)
    fix gc a gc'
    assume "pid_system.lstep gc a gc'"
    hence "length (pid_global_config.processes gc') 
      = length (pid_global_config.processes gc)"
      unfolding pida_step_def pred_of_enex_def pid_system.lstep_def
      by (auto 
          simp: pid_execute_def stutter_extend_ex_def
          simp: pid_gexecute_def
          split: option.splits prod.split)
            
  } note no_cr_proc=this

  { (* For every valid pid in reachable config, 
      command is reachable from an initial command
    *)
    have A1: "\<And>x a b. x\<in>(a \<inter> b)\<^sup>= \<Longrightarrow> x\<in>a\<^sup>*"
      by auto

    fix gc0 gc pid w
    assume INIT: "pid_init prog gc0"
    assume VALID: "pid_valid gc pid"
    assume STEP: "pid_system.lpath gc0 w gc"
    have "cmd_of_pid gc pid \<in> cfg_E\<^sup>*``(proc_decl.body`set (program.processes prog))"
      using STEP VALID
      apply (induction w arbitrary: gc rule: rev_induct)
      using INIT 
      apply (auto 
        simp: pid_init_def pid_init_gc_def init_pc_def
      ) []

      apply (auto simp: no_cr_proc pid_system.lpath_simps)
      apply (drule project_lstep_cfg_E[where pid=pid], drule A1)
      apply (rule rtrancl_image_advance_rtrancl)
      apply (rprems, assumption+)
      done
  } note aux2 = this

  {
    fix w :: "global_action option list" and cuts
    have 
      "set w \<inter> {x. \<forall>pid c a c'. x = Some (pid, c, a, c') \<longrightarrow> (c, c') \<in> cuts} = {} 
      \<Longrightarrow> cfg_E_of_lpath w \<inter> cuts = {}"
      by (induction w) (auto split: option.splits)        
  } note aux3=this

  { 
    fix c F F'
    assume 1: "c\<in>cfg_E\<^sup>*``body`set (program.processes prog)"
    assume 2: "(c,c)\<in>(cfg_E \<inter> F')\<^sup>+"
    assume 3: "\<forall>pd\<in>set (program.processes prog). is_fas cfg_E {body pd} F"

    from 1 obtain pd where 
      PDI: "pd\<in>set (program.processes prog)" and PDR: "(body pd,c)\<in>cfg_E\<^sup>*"
      by auto
    from bspec[OF 3 PDI] have FAS: "is_fas cfg_E {body pd} F" .

    have "F' \<inter> F \<noteq> {}" proof
      assume "F'\<inter>F={}"
      with 2 have "(c,c)\<in>(cfg_E - F)\<^sup>+"
        by (auto elim: trancl_mono)
      with PDR FAS show False unfolding is_fas_def by auto
    qed
  } note aux4=this


  show ?thesis
    unfolding collect_sticky_edges_def collect_sticky_local_def
    apply (refine_vcg order_trans[OF find_fas_init_correct])
    apply simp
    apply simp
    apply (rule finite_subset[OF cfg_E_ss_reachable])
    apply simp
    apply auto []
    apply (metis (poly_guards_query) DiffI Un_subset_iff is_fas_mono subsetI)
    apply (metis (poly_guards_query) Un_subset_iff is_fas_mono subsetI)

    (* TODO: Clean up this proof *)
    apply (auto split: option.splits 
      simp: pid_system.reachable_words_fin_conv neq_Nil_conv
      simp: pid_system.fold_lpath_iff 
      simp: actions_of_sticky_local_def
      ) []
    apply (case_tac y rule: action_case_split.cases)
    apply simp
    apply simp
    apply (frule enabled_pid_valid)
    apply (drule (1) aux1)
    apply (drule (2) aux2)
    apply (drule aux4, assumption, blast)
    apply (auto simp: cfg_E_of_lpath_def) []

    apply (auto split: option.splits 
      simp: actions_of_sticky_local_def) []
    done
    
qed

(* TODO: Move Refine_Basic.convenience*)
lemma RETURN_SPECD:"RETURN x \<le> SPEC \<Phi> \<Longrightarrow> \<Phi> x"
  by simp

lemma visible_approx: 
  "pid_system.visible \<subseteq> Collect (Not o is_invisible_action)"
  unfolding pid_system.visible_def
  by (auto simp: invisible_ex_interp)


  definition pid_procs :: "pid_global_config \<Rightarrow> pid set" where
    "pid_procs gc \<equiv> {0..<length (pid_global_config.processes gc)}"
  
  definition pid_pac :: "pid \<Rightarrow> global_action option set" where
    "pid_pac pid \<equiv> Collect (\<lambda>Some (pid',_) \<Rightarrow> pid'=pid | None \<Rightarrow> False)"
  
  definition pid_psen :: "pid \<Rightarrow> pid_global_config \<Rightarrow> global_action option set"
    where "pid_psen pid gc \<equiv> Some`pid_get_actions_of_pid pid gc"

sublocale pid_system: system_concurrent
  pid_enabled pid_execute "(Collect (pid_init prog))"
  pid_procs pid_pac pid_psen
  apply unfold_locales
  
  apply (auto simp: pid_procs_def) []

  apply (auto 
    simp: pid_pac_def pid_psen_def pid_enabled_def pid_genabled_def
    simp: 
    split: option.splits
    ) []
  apply (auto simp: pid_get_actions_alt pid_get_actions_of_pid_def) []
  
  apply (auto 
    simp: pid_pac_def pid_psen_def pid_get_actions_of_pid_def
    split: option.splits
    ) []
  done

end

locale wf_program_sticky_loc = wf_program_vis_vars_loc +
  fixes cut_edges
  assumes cut_edges_sticky: "RETURN cut_edges \<le> collect_sticky_local"
begin

  definition "cut_actions \<equiv> actions_of_sticky_local cut_edges"

  lemma cut_actions_sticky: "RETURN cut_actions \<le> collect_sticky_edges"
    unfolding cut_actions_def collect_sticky_edges_def
    using cut_edges_sticky
    by (simp add: refine_pw_simps pw_le_iff, blast)

  sublocale pid_system: system_sticky 
    pid_enabled pid_execute "Collect (pid_init prog)" 
    interp cut_actions
    apply unfold_locales
    using order_trans[OF cut_actions_sticky collect_sticky_edges_correct, THEN RETURN_SPECD]
    apply -
    apply simp  
    using visible_approx
    apply auto []
    done
    
  sublocale pid_system: system_ample pid_enabled pid_execute 
    "Collect (pid_init prog)" interp cut_actions ind
    by unfold_locales


  
  definition is_ample_static :: "cmd \<Rightarrow> bool" where
    "is_ample_static c \<equiv> (\<forall>c' a. cfg' c a c' 
      \<longrightarrow> read_globals a = {} \<and> write_globals a = {} \<and> (c,a,c')\<notin>cut_edges) 
      \<and> (\<exists>a c'. cfg' c a c')"
      (* TODO: Can we do a finer approx here? *)

  primrec find_min_idx :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<rightharpoonup> nat" where
    "find_min_idx P [] = None"
  | "find_min_idx P (x#xs) = (
      if P x then Some 0 else
      map_option Suc (find_min_idx P xs)
    )"
  
  lemma find_min_idx_None_conv: 
    "find_min_idx P l = None \<longleftrightarrow> (\<forall>a\<in>set l. \<not>P a)"
    apply (induction l)
    apply auto
    done

  lemma find_min_idx_SomeD:
    "find_min_idx P l = Some i \<Longrightarrow> P (l!i) \<and> i < length l"
    by (induction l arbitrary: i) (auto split: split_if_asm)

  lemma find_min_idx_SomeD_complete: 
    "find_min_idx P l = Some i \<Longrightarrow> (P (l!i) \<and> i < length l \<and> (\<forall>j<i. \<not>P (l!j)))"
    apply (induction l arbitrary: i) 
    apply (auto split: split_if_asm)
    apply (case_tac j)
    apply auto
    done
    

  definition "pid_is_ample_dynamic lc gs \<equiv> 
    (\<exists>c' a. cfg' (local_config.command lc) a c' 
          \<and> enabled_fs' a (local_config.state lc, gs))"

  definition pid_genabled_ample :: "pid_global_config \<Rightarrow> global_action set" where
    "pid_genabled_ample gc \<equiv> let
      lcs = pid_global_config.processes gc;
      pid = find_min_idx (\<lambda>lc.
          is_ample_static (local_config.command lc) 
        \<and> pid_is_ample_dynamic lc (pid_global_config.state gc)
      ) lcs
    in
      case pid of   
        None \<Rightarrow> pid_genabled gc
      | Some pid \<Rightarrow> {ga\<in>pid_get_actions_of_pid pid gc.  ga_is_enabled ga gc}
    "

  definition "pid_enabled_ample \<equiv> stutter_extend_en pid_genabled_ample"

  lemma get_actions_of_pid_subset: "pid_valid gc pid \<Longrightarrow> 
    pid_get_actions_of_pid pid gc \<subseteq> pid_get_actions gc"
    by (force simp: pid_get_actions_alt)

  lemma pid_enabled_ample_cases[consumes 0, case_names no_red red_on_pid]:
    obtains
      "pid_enabled_ample gc = pid_enabled gc"
    | pid where "pid_enabled_ample gc = pid_pac pid \<inter> pid_enabled gc"
        "pid_valid gc pid"
        "is_ample_static (cmd_of_pid gc pid)"
        "pid_is_ample_dynamic (lc_of_pid gc pid) (pid_global_config.state gc)"
  proof -  
    let ?lcs = "pid_global_config.processes gc"
    let ?pid = "find_min_idx (\<lambda>lc.
          is_ample_static (local_config.command lc) 
        \<and> pid_is_ample_dynamic lc (pid_global_config.state gc)) ?lcs"

    show ?thesis proof (cases "?pid")
      case None
      thus ?thesis
        apply -
        apply (rule that(1))
        apply (auto 
          simp: pid_enabled_ample_def pid_enabled_def pid_genabled_ample_def
          simp: stutter_extend_en_def)
        done
    next
      case (Some pid)[simp]

      let ?res = "{ga\<in>pid_get_actions_of_pid pid gc.  ga_is_enabled ga gc}"

      from find_min_idx_SomeD[OF Some] have
        ample_static: "is_ample_static (cmd_of_pid gc pid)" 
        and ample_dyn: "pid_is_ample_dynamic (lc_of_pid gc pid) (pid_global_config.state gc)"
        and pid_valid: "pid_valid gc pid"
        by auto

      from ample_dyn have res_ne: "?res \<noteq> {}"
        by (auto 
          simp: pid_is_ample_dynamic_def pid_get_actions_of_pid_def
          simp: ga_is_enabled_def)

      have [simp]: "pid_enabled gc = Some`pid_genabled gc" 
      proof -
        from get_actions_of_pid_subset[OF pid_valid] res_ne 
        have "pid_genabled gc \<noteq> {}"
          by (force simp: pid_genabled_def)
        thus ?thesis
          unfolding pid_enabled_def
          by (auto simp: stutter_extend_en_def)
      qed
        
      have [simp]: "pid_enabled_ample gc = Some`pid_genabled_ample gc"
        using res_ne
        by (auto simp: 
          simp: pid_enabled_ample_def pid_genabled_ample_def stutter_extend_en_def)

      show ?thesis
        apply -
        apply (rule that(2)[of pid])
        apply safe apply (clarsimp simp: pid_pac_def inj_image_mem_iff)
        
        apply (auto simp: pid_genabled_ample_def pid_get_actions_of_pid_def) []
        
        using get_actions_of_pid_subset[OF pid_valid]
        apply (auto simp: pid_genabled_ample_def pid_genabled_def inj_image_mem_iff) []
        
        apply (auto 
         simp: pid_pac_def inj_image_mem_iff
         simp: pid_genabled_ample_def pid_genabled_def 
         ) []
        apply (auto simp: pid_get_actions_alt) []
        apply (frule pid_get_actions_of_pid_pid)
        apply auto []
        apply fact+
        done
    qed
  qed
  
  lemma [simp, intro!]: "stutter_extend_en en s \<noteq> {}"
    unfolding stutter_extend_en_def
    by auto

  lemma 
    assumes [simp, intro!]: "gc \<in> pid_system.reachable"
    shows "ample_set gc (pid_enabled_ample gc)"
  proof (cases rule: pid_enabled_ample_cases[of gc])
    case (no_red) [simp]
    show ?thesis by auto
  next
    case (red_on_pid pid)
    show ?thesis
      unfolding red_on_pid(1)
      apply (rule restrict_ample_set)

      apply fact

      apply (subst red_on_pid(1)[symmetric])
      apply (simp add: pid_enabled_ample_def)

      apply (auto simp: pid_pac_def pid_enabled_def pid_genabled_def
        split: option.splits) []
      using red_on_pid(2,3)
      apply (auto 
        simp: pid_get_actions_def is_ample_static_def
        simp: cut_actions_def actions_of_sticky_local_def) []

.x      xxx, ind is defined to be actions on different procs

      
                        
      




.xxx, ctd here:
  Clean up. Seperate stuff into:

  General
    * Systems and simulation

  Reference point semantics
    * CFG, local/global state
      * Finiteness: Of reachable CFG, of reachable local/global state
    * Scheduler (Indep of concrete semantics)
      * Combine finiteness arguments
    * Stutter extension (Indep of concrete semantics)
      * Transfer finiteness

  Refinements of semantics
    * Deciding local/global variables
    * Well-formedness checks, refined semantics without explicit error states
    * pid-based scheduler (Independent of concrete semantics)
    * action based scheduler (Indep)

  Program Analysis
    * Locals/globals read/written by actions
    * Independence overapproximation
    * Visible overapproximation
    * Sticky edge analysis

  
.x  xxx, ctd here: Clean up. Then ample set stuff


.xxx, ctd here:
  return ample set candidates.
  For each process
    for each set of transitions
      if no transitions disabled that could be enabled by other processes:
        ample-set candidate

  The program analysis part is: no transitions disabled that could be enabled by
    other processes

    Can a transition be enabled by another process:
      Depends on global var?
        No: Safe
        Yes: Unsafe (over-approx!)





end

