#!/bin/bash

die () {
    echo "Error: $@"
    exit 1
}

if [[ $0 == *run_test.sh ]]; then
    # we are not executing a link -- request the name on the cmdline
    [[ -z $1 ]] && die "No promela script given"

    name=${1/.pml/}
    shift
else
    name=`basename $0`
fi

[[ $# -eq 0 ]] && die "No argument size given"

num=$1

if [[ -n $2 ]]; then
    ltl=$2
else
    ltl=triv
fi

add_zero=$3

pml="${name}.pml"
trail="${pml}.trail"
pan="pan_${name}"

# compile Promela model
# -o1 : turn off data-flow optimization
# -o2 : turn off dead-code elimination
# -o3 : turn off statement merging
# -a : generate pan
spin -a -o1 -o2 -o3 -DN="$num" -DM="$(( num + 1 ))" "$pml" || die "Spin failed"

# compile generated pan
# -DNOREDUCE : disables the partial order reduction algorithm
gcc -DNOREDUCE pan.c -o "$pan" || die "GCC failed"

# cleanup compiling products from spin
rm pan.* _spin_nvr.tmp

# run pan
# -a : look for accepting run
# -n : do not print unreached states
# -m : search depth
./"${pan}" -a -n -m3000000${add_zero} -N"${ltl}" || die "PAN failed"

if [[ -e "${trail}" ]]; then
    echo
    echo -n "Trail generated -- do you want to evaluate it? [y/n] "
    read -N 1 y
    [[ $y == "y" ]] && ./"${pan}" -N"${ltl}" -r "${trail}" | less
    rm "${trail}"
fi

