theory Autoref_Basic_Demo
imports Tutorial_Basic
begin
  text \<open>Most basic example\<close>

  schematic_lemma "(?c::?'c, {1,2,3::nat}) \<in> ?R"
    apply autoref
    done

  schematic_lemma "(?c::?'c list, {1,2,3::nat}) \<in> \<langle>?R\<rangle>list_set_rel"  
    apply autoref
    done

  text \<open>But wait, lists where distinct? And we inserted? Where's the @{text list.insert}? \<close>  

  schematic_lemma "(?c::?'c list, {1,2,3,3::nat}) \<in> \<langle>?R\<rangle>list_set_rel"  
    apply autoref
    done
  
  text \<open>There is a specialized rule for inserting into list-set, 
    if disjointness can be proved\<close>
  thm list_set_autoref_insert_dj

  text \<open>We can handle maps (although @{typ "'a\<rightharpoonup>'b"} is not a single type-constructor)\<close>

  
  schematic_lemma "(?c::?'c, [ 1::nat \<mapsto> 2, 3\<mapsto>4  ]) \<in> ?R"  
    apply autoref
    done

  context begin interpretation autoref_syn .  
  schematic_lemma "(?c::?'c, [ 1::nat \<mapsto> 2, 3\<mapsto>4 ] ::: \<langle>nat_rel,Id\<rangle>dflt_rm_rel ) \<in> ?R"  
    apply autoref
    done
  end  

  schematic_lemma "(?c::?'c, [ 1::nat \<mapsto> 2, 3\<mapsto>4 ] :::\<^sub>r \<langle>nat_rel,Id\<rangle>dflt_ahm_rel ) \<in> ?R"  
    apply autoref
    done


  text \<open>Lets try nesting\<close>

  schematic_lemma "(?c::?'c, {{1,2,3::nat},{}}) : ?R"
    apply autoref
    done
  text \<open>It generates a list for the outer set\<close>  

  schematic_lemma "(?c::?'c, {{1,2,3::nat},{},{}}) : ?R"
    apply autoref
    done
 text \<open>And it generates an equality-operator for sets, 
   from a nested seq of generic algorithms\<close>


  text \<open> This fails, there is no default linear ordering on sets! \<close>  
  schematic_lemma "(?c, {{1,2,3::nat},{}}) : \<langle>\<langle>nat_rel\<rangle>dflt_rs_rel\<rangle>dflt_rs_rel"
    apply (autoref (keep_goal))
    oops

  text \<open> But you can make autoref define a custom one.
    Note that sets are derived from maps, via a generic algorithm. 
    Thus, the relator transformer @{term map2set_rel} occurs here.\<close>  
  schematic_lemma "(?c, {{1,2,3::nat},{}}) : \<langle>\<langle>nat_rel\<rangle>dflt_rs_rel\<rangle>map2set_rel (rbt_map_rel ?CMP)"
    apply (autoref)
    done

  text \<open>Lets try parameters\<close>
  schematic_lemma 
    fixes f :: "nat \<Rightarrow> int"
    shows "(?c::?'c, f`{1,2,3::nat})\<in>?R"
    using [[autoref_trace_failed_id]]
    apply (autoref (trace, keep_goal))  (* This command fails! *)
    oops text \<open>It does not know the parameter\<close>


  schematic_lemma 
    fixes f :: "nat \<Rightarrow> int"
    assumes [autoref_rules]: "(fi,f)\<in>nat_rel \<rightarrow> int_rel"
    shows "(?c::?'c, f`{1,2,3::nat})\<in>?R"
    apply (autoref)
    done

  text \<open>Note, iterator is not yet efficient. 
    But fortunately, we have deforestation rules.
    However, not in basic autoref-tool. 
    By default, they are applied in the post-processor of refine-transfer.

    Most conveniently, we use a combination of Autoref and the refinement framework.
    \<close>  

  schematic_lemma 
    fixes f :: "nat \<Rightarrow> int"
    assumes [autoref_rules]: "(fi,f)\<in>nat_rel \<rightarrow> int_rel"
    shows "RETURN (?c::?'c) \<le> \<Down>?R (RETURN (f`{1,2,3::nat}))"
    apply (autoref_monadic)
    done
  text \<open>Note, deforestation done! \<close>  

  text \<open>The plain-option avoids using the monad at all. 
    (Only for total-correct tail-recursion)\<close>
  schematic_lemma 
    fixes f :: "nat \<Rightarrow> int"
    assumes [autoref_rules]: "(fi,f)\<in>nat_rel \<rightarrow> int_rel"
    shows "RETURN (?c::?'c) \<le> \<Down>?R (RETURN (f`{1,2,3::nat}))"
    apply (autoref_monadic (plain))
    done

  text \<open>If we want to explicitly return value of dres-monad: \<close>  
  schematic_lemma 
    fixes f :: "nat \<Rightarrow> int"
    assumes [autoref_rules]: "(fi,f)\<in>nat_rel \<rightarrow> int_rel"
    shows "nres_of (?c::?'c dres) \<le> \<Down>?R (RETURN (f`{1,2,3::nat}))"
    apply (autoref_monadic)
    done

  text \<open>Going more advanced: Nondeterminism\<close> 
  schematic_lemma 
    shows "RETURN (?c::?'c) \<le> \<Down>?R (do {
      let s = {1,2,3::nat};
      ASSERT (s \<noteq> {});   (* Be careful: Autoref is very stupid here. 
        It forgets everything about a let-bound variable that is not explicitly stated*)
      v \<leftarrow> SPEC (\<lambda>x. x\<in>s);
      RETURN v
    })"
    apply autoref_monadic
    done

  text \<open>Debugging the above problem: A troubleshooting overview\<close>
  schematic_lemma 
    shows "RETURN (?c::?'c) \<le> \<Down>?R (do {
      let s = {1,2,3::nat};
      v \<leftarrow> SPEC (\<lambda>x. x\<in>s);
      RETURN v
    })"
    apply autoref_monadic

    (* Autoref leaves you with a goal about op_set_pick. This is 
      the operation name for selecting an arbitrary element from a set.
      It was inserted by the id-phase, due to a pattern rule for SPEC (\<lambda>x. x\<in>s)
    *)

    apply autoref_trans_step (* This command fails! *)
    (* Applying further translation steps fails.
      To debug this problem, we have to realize that autoref has already
      refined the goal too far.

      The refinement rules that have side conditions are written 
      in first - order style. But, when Autoref found that none of them 
      applies (it could not discharge the side condition), it fell back
      to the parametricity rule for function application, and
      left us with the above goal.

      In order to check why the side-condition does not apply, 
      we have to go back to the point where the parameters are still attached.
    *)
    oops

  (* Currently, the only way to achieve that is to go step-by-step
    through the whole translation phase *)
  schematic_lemma 
    shows "RETURN (?c::?'c) \<le> \<Down>?R (do {
      let s = {1,2,3::nat};
      v \<leftarrow> SPEC (\<lambda>x. x\<in>s);
      RETURN v
    })"
    (* First, do the initialization that the autoref_monadic - wrapper would have applied *)
    apply (rule autoref_monadicI)
    (* Then, execute the other phases. It's always a good idea to pass the 
      trace and keep_goal options when debugging. The trace-option shows you the current 
      progress of the phases, the keep_goal option leaves you with the unsolved 
      subgoals of a failed phase, instead of failing the whole tactic.
      Note: keep_goal does not work for the id_op-phase, this always fails the 
      whole tactic! It's debugging should be done via 
        using [[autoref_trace_failed_id]]
      if this does not work, you have to dive into the internals.  
      *)
    apply (autoref (trace, keep_goal) phases: id_op rel_inf fix_rel)  

    (* Now, we have executed all phases but translation. We see the internal goal
      of autoref, containing a lot of tagging (to cope with ho-unification madness),
      and annotations of relators to each operation.

      To make the goal more readable, you may selectively remove some of the tagging.
      But attention: The translation steps require all the tagging, so only remove the
      tagging temporarily for inspecting the goal:
      *)
    (*unfolding autoref_tag_defs (* Remove all tagging *)
    unfolding ANNOT_def (* Remove relator annotations *) 
    unfolding OP_def PROTECT_def APP_def (* Remove everything else but relator annotations *) 
    *)

    (* The translation is executed step by step *)
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step  (* At this point, we found the problem. *) 
    oops


  (* Note that autoref_trans_step already does quite a lot.
    It recursively instantiates generic algorithms, and solves side-conditions 
    if it can, only presenting you the unsolved ones. 
    Moreover, it does NOT allow backtracking, so it will present you the 
    unsolvable side-conditions of the first rule that matches.

    Now, lets use autoref_trans_step_only, which only does a single step,
    and allows backtracking
  *)  
  schematic_lemma 
    shows "RETURN (?c::?'c) \<le> \<Down>?R (do {
      let s = {1,2,3::nat};
      v \<leftarrow> SPEC (\<lambda>x. x\<in>s);
      RETURN v
    })"
    apply (rule autoref_monadicI)
    apply (autoref (trace, keep_goal) phases: id_op rel_inf fix_rel)  
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step
    apply autoref_trans_step_only   

    apply autoref_side  (* (Try to) Solve side-condition *)
    apply autoref_side  (* This one is the unsolvable side condition. 
      It leaves you with a subgoal tagged with PRECOND_tag. 
      If you still want to inspect why you cannot solve the goal, you may use
      its introduction rule, and inspect the bare side condition: *)
    apply (rule PRECOND_tagI)
    oops


    

end

